#ifdef REV_INFO
static char rcsid[] = "$Id: goInit.cc,v 1.19 1992/12/30 13:32:43 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  Go Editor
FILE:    goInit.cc
AUTHOR:  Aline Jacomet
DATE:    May 14th 1991

DESCRIPTION:
    Go initialization

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France

  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.

  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include "goMain.h"

#include <Go/Tools/FilePath.h>
#include <Go/Tools/Alloc.h>
#include <Go/Tools/FileMedium.h>
#include <Go/Tools/Dash.h>
#include <Go/Tools/Array.h>
#include <Go/Tools/StringAssoc.h>
#include <Go/Tools/Bind.h>
#include <Go/FullSensor.h>
#include <Go/Translator.h>
#include <Go/Init.h>

#include <Go/Sensor.h>
#include <Go/PointSensor.h>

#include <Go/Rectangle.h>
#include <Go/Circle.h>
#include <Go/Curve.h>
#include <Go/RegPoly.h>
#include <Go/Bezier.h>
#include <Go/BSpline.h>
#include <Go/Cardinal.h>
#include <Go/Beta.h>
#include <Go/Tau.h>
#include <Go/Label.h>
#include <Go/Text.h>
#include <Go/SText.h>
#include <Go/mTextual.h>
#include <Go/SimpleForm.h>

#ifdef GoLICENSE
#include <Go/ScText.h>
#include <Go/MCForm.h>
#include <Go/MBForm.h>
#include <Go/CircleForm.h>
#endif

#include <Go/Image.h>
#include <Go/Composite.h>
#include <Go/GoAscii.h>

#ifdef GoLICENSE
#include <Go/GoBinary.h>
#endif

#include <Go/XGoInterf.h>
#include <Go/XtInDisp.h>
#include <Go/BGoUIEd.h>
#include "classes.h"


/*
 * VARIABLES
 */

GoRoot* goroot = NULL;
GoRoot* edit_goroot = NULL;
GoRoot* named_goroot = NULL;
GoRoot* help_goroot = NULL;

#ifdef GoLICENSE
GXtClipboardInterf* cb_interf = NULL;
#else
GoDrawClipboard* cb_interf = NULL;
#endif


/*
 * the following interfaces are created by the Graphic widget
 * so : they are XmGoInterf interface except interf_getimage
 * which is an XGoInterf.
 * The input dispatcher for any XmGoInterf is
 *          XtNextEvent(), XtDispatchEvent()    (cf XmGoInterf::handleNextEvent()).
 * so It is OK for any XmGoInterf.
 *
 * If the current_input_interface is NULL we will use the first created XmGoInterf.
 * see HandleNextEvent static function.
 * NOTE that we create an XGoInterf for  interf_getimage in place of creation of
 * a Graphic widget because is is simpler.
 */

XGoInterf* interf;
XGoInterf* edit_interf;
XGoInterf* named_interf;
XGoInterf* help_interf;
    
char* font_name;
float pixel_size;

GDash* goDash;

extern PenAttributes pen;
extern PenAttributes named_pen;
extern TextAttributes text;
extern TextAttributes named_text;
extern AllAttributes all_attrib;

extern GStringAssoc NamedBrush;
extern GStringAssoc NamedPen;
extern GStringAssoc NamedText;

extern GSet SaveShared;
extern GSet HasDestroyPostcond;



/*
 * FUNCTIONS
 */

// in file goGoUI.cc;
extern void Move(GoRoot*, const GEvent*);
extern void MoveThis(Go*, const GEvent*);
extern void Rescale(GoRoot*, const GEvent*);
extern void RescaleThis(Go*, const GEvent*);
extern void Rotate(GoRoot*, const GEvent*);
extern void Destroy(Go*);
extern void MoveClosestPolyPoint(Go*, const GEvent*);

// in file goLoadSave.cc;
extern void RestoreAssoc(void*, Go*, GoStream*);
extern int  saveNumber(const GNumber*, GoStream*);
extern int  restoreNumber(GNumber**, GoStream*);
extern void sharedDestroyed(void*, const GSharedObject*);

// in file goUndo.cc;
extern void UndoTextModification(void*, GoTextual*, GModifyValueStruct*);

// in file goPopup.cc;
extern void MapPopupMenu(GoRoot*, const GPoint*);

// in file goTools.cc;
extern void GoFlush();

// in file goStructure.cc;
extern GBool HandleEmptyComposite(Go*);

// in file goSelection.cc;
extern void TextSelectPortion(class GoTextual*, const class GEvent*);

// in file goMiscel.cc;
extern void dup_assoc(void*, Go*, Go*);


static void _compositeWithNoChild(void*, Go* composite, Go*, Go*)
{
    if (composite->Child() == NULL) {
	// handle destroy with undo;
	HandleEmptyComposite(composite);
    }
}

static void textColorInitialize (void*, GoTextual* go)
{
    /* Text Selection Brush */
    GShColor* c = interf->colorFromName ("gray");
    go->selectionBrush().color(c);
    c->destroy();
}


/**
FUNCTION
    RunEventLoop()
DESCRIPTION
    We use here the GInputDispatcher(): which is an XtInputDispatcher.
**/
void RunEventLoop()
{
    (void) GInputDispatcher()->runEventLoop();
}

/**
FUNCTION
    flush()
DESCRIPTION
   This function is called at idle time when no event is left
   in the input queue
   Flush the request on each goroot;
   Flush X request for motif request;
**/
static GIdle flush(void* /* unused */)
{
    // flush Go reguests;
    GoFlush();

    // flush Motif requests;
    interf->flush();
    return GIDLE_AFTER_NEXTEVENT;
}    


void SelectChild (GoRoot* root)
{
    BGoUIEditor* bgoui = (BGoUIEditor*) GoUIFrom (root);
    if (bgoui != NULL)
	bgoui->selectFirstChild ();
}

void SelectNext (GoRoot* root)
{
    BGoUIEditor* bgoui = (BGoUIEditor*) GoUIFrom (root);
    if (bgoui != NULL)
	bgoui->selectNext ();
}

void SelectPrevious (GoRoot* root)
{
    BGoUIEditor* bgoui = (BGoUIEditor*) GoUIFrom (root);
    if (bgoui != NULL)
	bgoui->selectPrevious ();
}


GEvent* godraw_remove_release_events (void*, GoUI*, const GEvent* event)
{
    return
	(event->type() == GBUTTON_RELEASE_EVENT)
	    ? NULL
		: (GEvent*) event;
}


void InitGo(void* i)
{
    interf = (XGoInterf*)i;
    goroot = GoRoot::create();
    goroot->interf(interf);
    goroot->map();

    GoExtensionsInit();

    /*
     * Init the global input dispatcher;
     * I make an Xt one;
     */

    GInputDispatcher() = new XtInputDispatcher
	(XtDisplayToApplicationContext (interf->display()));
    GInputDispatcher()->idleProc (flush, NULL);

    GoStream::bind("GNumber",
		   (int (*) (const void*, GoStream*))saveNumber,
		   (int (*) (void**, GoStream*))restoreNumber);

#ifdef GoLICENSE
    /* GXtClipboardInterf creation */
    cb_interf = new GXtClipboardInterf((XtGoInterf*)interf);
#else    
    /* GoDrawClipboard creation */
    cb_interf = new GoDrawClipboard(&undoMgr);
#endif
    
    /* GoUI file */

    char* all_name = GGetFilePath("app-files/godraw/goui", GOUIFILE);
    if (all_name)
    {
	GLoadTranslatorPgms(all_name);
	GFREE(all_name);
    }
    
    (void) new BGoUIEditor (goroot);
    GoAlwaysFirstChild (GoFullSensor::create ("goui", goroot, NULL, NULL, NULL));
    extern GBool noFinger;
    if (noFinger)
	GoUIFrom (goroot)->filter (godraw_remove_release_events, NULL);
    

    GBind ("Rescale", Rescale);
    GBind ("RescaleThis", RescaleThis);
    GBind ("Rotate", Rotate);
    GBind ("Move", Move);
    GBind ("MoveThis", MoveThis);
    GBind ("Destroy", Destroy);
    GBind ("SelectChild", SelectChild);
    GBind ("SelectNext", SelectNext);
    GBind ("SelectPrevious", SelectPrevious);
    GBind ("MoveClosestPolyPoint", MoveClosestPolyPoint);
    GBind ("MapPopupMenu", MapPopupMenu);
    GBind ("TextSelectPortion", TextSelectPortion);

    (GoComposite::metaclass())->removeChildPostcond
	(_compositeWithNoChild, NULL, GADD);

    (Go::metaclass())->dupPostcond (dup_assoc, NULL, GADD);
    
    // run a "textEditor" translator when a textual is selected;
    // destroy it when the textual is unselected;

    GoTextualMakeSensitive ("textEditor");


    /* formatter initialisations */
    GSimpleFormatter::init();
#ifdef GoLICENSE
    GMCFormatter::init();
    GMBFormatter::init();
    GCircleFormatter::init();
#endif    

    (GoTextual::metaclass())->createPostcond(
	(void(*)(void*, Go*))textColorInitialize, NULL, GADD);
    ((GoTextualMetaclass*)(GoTextual::metaclass()))->
	modifyValuePostcond(UndoTextModification, NULL, GADD);

    (Go::metaclass())->restorePostcond(RestoreAssoc, NULL);

    (GoPointSensor::metaclass())->radius(0.8);
    
    pixel_size = max(((GoInterf*)interf)->pixelWidth(),
		     ((GoInterf*)interf)->pixelHeight());

    int inter = (int)((pixel_size * 100) + 1);
    double pixel = (double)(inter / 100.);
    char buf[20];
    sprintf(buf, "%.2f", pixel);
    GFREE(pen.LineWidth);
    pen.LineWidth = (char*)GMALLOC(strlen(buf) + 1);
    strcpy(pen.LineWidth, buf);
    GFREE(named_pen.LineWidth);
    named_pen.LineWidth = (char*)GMALLOC(strlen(buf) + 1);
    strcpy(named_pen.LineWidth, buf);
    GFREE(all_attrib.pen.LineWidth);
    all_attrib.pen.LineWidth =(char*) GMALLOC(strlen(buf) + 1);
    strcpy(all_attrib.pen.LineWidth, buf);

    GTextBrush tb;
    font_name = (char*)interf->fontName(GUnitCoordSys(),
					BGOUI_LINE_HEIGHT_AT_CREATION_TIME,
					(const GShFont*)tb.font());
    GFREE(text.Font);
    text.Font = (char*)GMALLOC(strlen(font_name) + 1);
    strcpy(text.Font, font_name);
    GFREE(named_text.Font);
    named_text.Font = (char*)GMALLOC(strlen(font_name) + 1);
    strcpy(named_text.Font, font_name);
    GFREE(all_attrib.text.Font);
    all_attrib.text.Font = (char*)GMALLOC(strlen(font_name) + 1);
    strcpy(all_attrib.text.Font, font_name);

    // dash initialization;
    GNumber dash_list[2];
    dash_list[0] = 2.0;
    dash_list[1] = 2.0;

    GDashValues values;
    values.dash_offset = 0;
    values.dash_number = 2;
    values.dash_list = dash_list;

    goDash = new GDash(GDASH_OFFSET|GDASH_LIST, &values);

    /*
     * Load named objects
     */
    all_name = GGetFilePath("app-files/godraw", NAMED_OBJECTS_FILE);
    if (all_name) {
	GFileMedium* medium = new GFileMedium();
	if (! medium->open(all_name, "r")) {
	    GoStream* gostream = NULL;
#ifdef GoLICENSE	    
	    if (isGoBinary(medium))
		gostream = new GoBinary(medium);
	    else
#endif		
		gostream = new GoAscii(medium);
	    if (gostream->openRead()) {
		char* class_name;
		void* obj;
		do {
		    if (! gostream->restoreObj((void**)&obj, &class_name))
			break;
		    else 
			if (obj) {
			    if (strcmp(class_name, GShPen::classname()) == 0) {
				((GShPen*)obj)->dup();
				SaveShared.add(obj);
				if (((GShPen*)obj)->isNamed())
				    NamedPen.set(((GShPen*)obj)->name(), obj);
				((GShPen*)obj)->destroyPostcond((void(*)(void*, const GShPen*))sharedDestroyed, NULL);
				HasDestroyPostcond.add(obj);
			    } else
			    if (strcmp(class_name,
				     GShBrush::classname()) == 0) {
				((GShBrush*)obj)->dup();
				SaveShared.add(obj);
				if (((GShBrush*)obj)->isNamed())
				    NamedBrush.set(((GShBrush*)obj)->name(),
						   obj);
				((GShBrush*)obj)->destroyPostcond((void(*)(void*, const GShBrush*))sharedDestroyed, NULL);
				HasDestroyPostcond.add(obj);
			    } else
			    if (strcmp(class_name,
				       GShTextBrush::classname())== 0) {
				((GShTextBrush*)obj)->dup();
				SaveShared.add(obj);
				TextStruct str;
				str.sh = (GShTextBrush*)obj;
				void* num;
				if (!gostream->restoreObj((void**)&num,
							  &class_name))
				    break;
				str.height = *((GNumber*)num);
				NamedText.set
				    (((GShTextBrush*)obj)->name(),
				     sizeof(TextStruct), &str);
				((GShTextBrush*)obj)->destroyPostcond((void(*)(void*, const GShTextBrush*))sharedDestroyed, NULL);
				HasDestroyPostcond.add(obj);
			    }
			}
		} while (obj);
		gostream->close();
	    }
	    delete gostream;
	    medium->close();
	}
	delete medium;
	GFREE(all_name);
    }
}


#define OBJECT_STREAM "attributes.go"

void InitNamedGo(void* i)
{
    named_interf = (XGoInterf*)i;
    named_goroot = GoRoot::create();
    named_goroot->interf(named_interf);
    named_goroot->map();

    
    GFileMedium* medium = new GFileMedium();
    char* all_name = GGetFilePath("app-files/godraw", OBJECT_STREAM);
    if (all_name) {
	if (! medium->open(all_name, "r")) {
	    GoStream* gostream = NULL;
#ifdef GoLICENSE
	    if (isGoBinary(medium))
		gostream = new GoBinary(medium);
	    else
#endif		
		gostream = new GoAscii(medium);
	    gostream->openRead();
	    Go* go;
	    char* class_name;
	    do {
		if (!gostream->restoreObj((void**)&go, &class_name))
		    break;
		else
		    if (go)
			go->reparent(named_goroot);
	    } while (go);
	
	    gostream->close();
	    delete gostream;
	    medium->close();
	} else
	    printf("cannot open the file attributes.go !!\n");
	GFREE(all_name);
    } else
	printf("cannot find the file attributes.go !!\n");
    delete medium;
}



void InitEditGo(void* i)
{
    edit_interf = (XGoInterf*)i;
    edit_goroot = GoRoot::create();
    edit_goroot->interf(edit_interf);
    edit_goroot->map();

    
    GFileMedium* medium = new GFileMedium();
    char* all_name = GGetFilePath("app-files/godraw", OBJECT_STREAM);
    if (all_name) {
	if (! medium->open(all_name, "r")) {
	    GoStream* gostream = NULL;
#ifdef GoLICENSE
	    if (isGoBinary(medium))
		gostream = new GoBinary(medium);
	    else
#endif		
		gostream = new GoAscii(medium);
	    gostream->openRead();
	    Go* go;
	    char* class_name;
	    do {
		if (!gostream->restoreObj((void**)&go, &class_name))
		    break;
		else
		    if (go)
			go->reparent(edit_goroot);
	    } while (go);
	    
	    
	    gostream->close();
	    delete gostream;
	    medium->close();
	} else
	printf("cannot open the file attributes.go !!\n");
	delete medium;
    }
}

void InitGoHelp(void* interf)
{
    help_goroot = GoRoot::create();
    help_goroot->interf((XGoInterf*)interf);
    help_goroot->map();

    
    (void)GoRegPolygon::metaclass();
    (void)GoRectangle::metaclass();
    (void)GoCircle::metaclass();
    (void)GoCurve::metaclass();
    (void)GoBezier::metaclass();
    (void)GoText::metaclass();
    (void)GoSimpleText::metaclass();
    (void)GoLabel::metaclass();
#ifdef GoLICENSE
    (void)GoScrolledText::metaclass();
#endif    
}
