#ifdef REV_INFO
static char rcsid[] = "$Id: clock.cc,v 1.6 1992/10/22 08:35:55 andre Exp $";
#endif

/*****************************************************************************
MODULE:  Examples
FILE:    clock.cc
AUTHOR:  Michel Andre
DATE:    07-01-92

DESCRIPTION: Analogic and digital clock objects.
    

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France

  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.

  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <Go/GoInterf.h>
#include "clock.h"


goClock::goClock()
{
    _timeDifference = 0;
}


void goClock::shift (int n)
{
    _timeDifference = n;
    _firstTime = GTRUE;
    update();
}


/***************************************************/
/***************************************************/
/*                                                 */
/*                ANALOGIC CLOCK                   */
/*                                                 */
/***************************************************/
/***************************************************/


analogicGoClock::analogicGoClock (GNumber clockRadius)
{
    /* circle */
    _circle = GoCircle::create();
    _circle->reparent (this);
    _circle->background (GTRUE);

    /* label */
    _label = GoLabel::create();
    _label->reparent (_circle);
    _label->mapCursor (GFALSE);
    
    /* hour hand */
    GPoint pts[3];
    _hourHand = GoPolygon::create (3, pts);
    _hourHand->reparent (_circle);
    _hourHand->background (GTRUE);
    
    /* minute hand */
    _minuteHand = GoPolygon::create (3, pts);
    _minuteHand->reparent (_circle);
    _minuteHand->background (GTRUE);
    
    /* second ring */
    for (int loopCounter = 0; loopCounter < 60; loopCounter++){
        _secondsRing [loopCounter] = GoRectangle::create();
	_secondsRing [loopCounter]->reparent (_circle);
	_secondsRing [loopCounter]->outline (GFALSE);
	_secondsRing [loopCounter]->background (GTRUE);
    }

    /* initialisations */
    _lastSecond = 0;
    _firstTime = GTRUE;
    update();
    reframe (clockRadius);
}


void
analogicGoClock::update()
{
    goClock::update();

    /* update seconds */
    GShColor* c = interf()->colorFromName ("black");
    _secondsRing [_lastSecond]->brush().color (c);
    c->destroy();
    c = (GShColor*) _highLightCol.shColor();
    _secondsRing [_timeBuf->tm_sec]->brush().color (c);

    /* new minute */
    if (_lastSecond > _timeBuf->tm_sec || _firstTime){
	/* update minutes */
	GNumber angleM = (GNumber)PI*_timeBuf->tm_min/30;
	angleM -= PI / 2;
	GCoordSys cs = _minuteHand->Relcs();
	cs.angle (angleM);
	_minuteHand->setCoordsys (cs, _circle);

	/* update hours */
	_timeBuf->tm_hour += _timeDifference;
	_timeBuf->tm_hour %= 12;
	GNumber angleH = (_timeBuf->tm_hour*60 + _timeBuf->tm_min)
	    * PI /360;        
	angleH -= PI / 2;
	cs = _hourHand->Relcs();
	cs.angle (angleH);
        _hourHand->setCoordsys (cs, _circle);

        /* update day (add the time difference seconds) */
	if (_timeBuf->tm_min == 0 || _firstTime){
	    _time += _timeDifference * 60 * 60;
            char* currentTime = ctime (&_time);
	    char tmp [16];
	    strncpy (tmp, currentTime, 4);
	    strncpy (tmp + 4, currentTime + 8, 3);
            _label->setString (tmp, 7);
	}

	_firstTime = GFALSE;
    }

    /* update _lastSecond */
    _lastSecond = _timeBuf->tm_sec;
}


void
analogicGoClock::reframe (GNumber clockRadius)
{
    GPoint origin = Abscs().O;
    _clockRadius = clockRadius;

    /* circle */
    _circle->reset (clockRadius);
    _circle->move (origin, this);

    /* hour hand */
    GPoint pts[3];
    pts[0].x = -clockRadius/8;
    pts[0].y = -clockRadius/16;
    pts[1].x = -clockRadius/8;
    pts[1].y = clockRadius/16;
    pts[2].x = clockRadius/2;
    pts[2].y = 0;
    _hourHand->reset (3, pts);
    _hourHand->move (origin, this);
    
    /* minute hand */
    pts[0].x = -clockRadius/6;
    pts[0].y = -clockRadius/24;
    pts[1].x = -clockRadius/6;
    pts[1].y = _clockRadius/24;
    pts[2].x = _clockRadius*3/4;
    pts[2].y = 0;
    _minuteHand->reset (3, pts);
    _minuteHand->move (origin, this);

    /* seconds ring */
    GNumber len1 = _clockRadius / 5;
    GNumber len2 = _clockRadius / 12;
    GNumber len3 = _clockRadius / 30;
    GNumber len4;
    for (int loopCounter = 0; loopCounter < 60; loopCounter++){
        len4 = (loopCounter%5) ? len2 : len1;
	_secondsRing[loopCounter]->reset (len3, len4);
        _secondsRing[loopCounter]->move (GPoint (0, -clockRadius), this);
        _secondsRing[loopCounter]->rotate (PI*loopCounter/30, origin);
    }

    /* label */
    _label->lineHeight (_clockRadius * 0.2);
    _label->move (GPoint (_clockRadius/4, -_clockRadius*0.1), this);
}


/***************************************************/
/***************************************************/
/*                                                 */
/*                DIGITAL CLOCK                    */
/*                                                 */
/***************************************************/
/***************************************************/


/* digit coding */
/****************/

/*
	   1
        --------
        |      |
      2 |      | 3
        |  4   |
        --------
        |      |
      5 |      | 6
        |      |
        --------
           7
*/


const unsigned char codingTable[] = {
    0xee,	/* 0 */
    0x24,	/* 1 */
    0xba,	/* 2 */
    0xb6,	/* 3 */
    0x74,	/* 4 */
    0xd6,	/* 5 */
    0xde,	/* 6 */
    0xa4,	/* 7 */
    0xfe,	/* 8 */
    0xf6	/* 9 */
};


digit::digit (unsigned char value)
{
    /* initialisation */
    _value = value;
    _segmentSize = 0;

    /* internal reparenting */
    for (int loopCounter = 0; loopCounter < 7; loopCounter++){
	_segments [loopCounter] = GoRectangle::create();
	_segments [loopCounter]->reparent (this, NULL);
	_segments [loopCounter]->background (GTRUE);
    }

    /* segments resize */
    segmentsSize (5);

    /* segments update */
    _segmentsUpdate();
}


void
digit::setValue (unsigned char value)
{
    if (_value == value) return;

    _value = value % 10;
    _segmentsUpdate();
}


void
digit::segmentsSize (GNumber size)
{
    if (_segmentSize == size) return;

    _segments [0]->move (GPoint (size/4, 0), this);
    _segments [0]->reset (size, size/4);

    _segments [1]->move (GPoint (0, size/4), this);
    _segments [1]->reset (size/4, size);

    _segments [2]->move (GPoint (5*size/4, size/4), this);
    _segments [2]->reset (size/4, size);

    _segments [3]->move (GPoint (size/4, 5*size/4), this);
    _segments [3]->reset (size, size/4);

    _segments [4]->move (GPoint (0, 6*size/4), this);
    _segments [4]->reset (size/4, size);

    _segments [5]->move (GPoint (5*size/4, 6*size/4), this);
    _segments [5]->reset (size/4, size);

    _segments [6]->move (GPoint (size/4, 10*size/4), this);
    _segments [6]->reset (size, size/4);
}


void
digit::_segmentsUpdate()
{
    unsigned char valueCode = codingTable [_value];
    unsigned char tester = 1 << 7;

    for (int loopCounter = 0; loopCounter < 7; loopCounter++){
	if (tester & valueCode)
	    _segments [loopCounter]->map();
	else
	    _segments [loopCounter]->unmap();

	tester >>= 1;
    }
}


digitalGoClock::digitalGoClock (GNumber clockWidth)
{
    _background = GoRectangle::create();
    _background->reparent (this);
    _background->background (GTRUE);
    _background->move (GPoint (0, 0), this);
    _hour1 = new digit();
    _hour1->reparent (_background);
    _hour2 = new digit();
    _hour2->reparent (_background);
    _minute1 = new digit();
    _minute1->reparent (_background);
    _minute2 = new digit();
    _minute2->reparent (_background);
    _second1 = new digit();
    _second1->reparent (_background);
    _second2 = new digit();
    _second2->reparent (_background);

    reframe (clockWidth);
    update();
}


void
digitalGoClock::reframe (GNumber clockWidth)
{
    _hour1->segmentsSize (clockWidth/16);
    _hour1->move (GPoint (clockWidth/32, clockWidth/32), _background);

    _hour2->segmentsSize (clockWidth/16);
    _hour2->move (GPoint (clockWidth/8, 0), _hour1);

    _minute1->segmentsSize (clockWidth/16);
    _minute1->move (GPoint (clockWidth/6, 0), _hour2);

    _minute2->segmentsSize (clockWidth/16);
    _minute2->move (GPoint (clockWidth/8, 0), _minute1);

    _second1->segmentsSize (clockWidth/24);
    _second1->move (GPoint (clockWidth/6, 0), _minute2);

    _second2->segmentsSize (clockWidth/24);
    _second2->move (GPoint (clockWidth/12, 0), _second1);

    GNumber w = _second2->Relcs().O.x +
	_second2->clip().bigbox().width() + clockWidth / 32;
    _background->reset (w, clockWidth / 4);
    _clockWidth = w;
}


void
digitalGoClock::update()
{
    goClock::update();
    _timeBuf->tm_hour += 24 + _timeDifference;
    _timeBuf->tm_hour %= 24;

    _hour1->setValue ((unsigned char)(_timeBuf->tm_hour/10));
    _hour2->setValue ((unsigned char)(_timeBuf->tm_hour%10));
    _minute1->setValue ((unsigned char)(_timeBuf->tm_min/10));
    _minute2->setValue ((unsigned char)(_timeBuf->tm_min%10));
    _second1->setValue ((unsigned char)(_timeBuf->tm_sec/10));
    _second2->setValue ((unsigned char)(_timeBuf->tm_sec%10));
}
