#ifdef REV_INFO
static char rcsid[] = "$Id: demo_graph.cc,v 1.12 1992/12/31 14:03:09 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  Examples
FILE:    demo_graph.cc
AUTHOR:  Aline Jacomet
DATE:    March 91

DESCRIPTION:
    Test program for GoTree and GoGraph objects.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <math.h>
#include <stdio.h>
#include <Go/GoRoot.h>
#include <Go/XGoInterf.h>
#include <Go/XInDisp.h>
#include <Go/Rectangle.h>
#include <Go/Circle.h>
#include <Go/Curve.h>
#include <Go/Bezier.h>
#include <Go/RegPoly.h>
#include <Go/Label.h>
#include <Go/SText.h>

#include <Go/Tree.h>
#include <Go/TreeNode.h>
#include <Go/Graph.h>
#include <Go/GraphNode.h>
#include <Go/Link.h>
#include <Go/SimLink.h>
#include <Go/BusLink.h>

#include <Go/BGoUIEd.h>
#include <Go/Sensor.h>
#include <Go/FullSensor.h>
#include <Go/ClipSensor.h>
#include <Go/Event.h>
#include <Go/ChoosePoint.h>
#include <Go/Translator.h>
#include <Go/Tools/Bind.h>
#include <Go/Tools/Error.h>

#include <Go/Line.h>
#include <Go/EdLine.h>

static XGoInterf* xinterf;
static GoRoot* goroot;

#define NB_TREE_NODES  12
#define NB_GRAPH_NODES 7


static void SetGoGeometricForeground (Go* go, const char* color)
{
    GShColor* c = xinterf->colorFromName(color);
    ((GoGeometric*)go)->pen().color(c);
    c->destroy();
}

static void SetGoGeometricBackground (Go* go, const char* color)
{
    if (!go->getMetaclass()->isSubclass(GoGeometric::metaclass()))
    {
	GWarning("%s is not a subclass of GoGeometric");
	return;
    }

    GoGeometric* geom = (GoGeometric*)go;
    GShColor* c = xinterf->colorFromName(color);
    geom->brush().color(c);
    c->destroy();
}

static void SetGoGeometricLineWidth (Go* go, const int lw)
{
    ((GoGeometric*)go)->pen().lineWidth(lw);
}

static void SetGoLinkLineWidth (Go* go, const int lw)
{
    ((GoLink*)go)->pen().lineWidth(lw);
}

static void SetGoLinkForeground (Go* go, const char* color)
{
    GShColor* c = xinterf->colorFromName(color);
    ((GoLink*)go)->pen().color(c);
    c->destroy();
}

static void SetGoGraphStructBackground (Go* go, const char* color)
{
    GShColor* c = xinterf->colorFromName(color);
    ((GoGraphStruct*)go)->brush().color(c);
    c->destroy();
}

static void SetGoNodeBackground (Go* go, const char* color)
{
    GShColor* c = xinterf->colorFromName(color);
    ((GoNode*)go)->brush().color(c);
    c->destroy();
}

static void SetTextColor(Go* go, const char* color)
{
    GShColor* c = xinterf->colorFromName(color);
    ((GoTextual*)go)->stringTextBrush().color(c);
    c->destroy();
}


static void SetTextBackColor(Go* go, const char* color)
{
    GShColor* c = xinterf->colorFromName(color);
    ((GoTextual*)go)->backgroundBrush().color(c);
    c->destroy();
}


static GoTree* GetTree(Go* go)
{
    if (go) {
	Go* ret = go->parent();
	while (ret && !ret->getMetaclass()->isSubclass(GoTree::metaclass()))
	    ret = ret->parent();
	return (GoTree*)ret;
    } else
	return NULL;
}

static GoGraph* GetGraph(Go* go)
{
    if (go) {
	Go* ret = go->parent();
	while (ret && !ret->getMetaclass()->isSubclass(GoGraph::metaclass()))
	    ret = ret->parent();
	return (GoGraph*)ret;
    } else
	return NULL;
}

static GoNode* GetNode(Go* go)
{
    if (go) {
	Go* ret = go->parent();
	while (ret && !ret->getMetaclass()->isSubclass(GoNode::metaclass()))
	    ret = ret->parent();
	return (GoNode*)ret;
    } else
	return NULL;
}

static GoGraphStruct* GetGraphStr(Go* go)
{
    if (go) {
	Go* ret = go->parent();
	while (ret && !ret->getMetaclass()->isSubclass(GoGraphStruct::metaclass()))
	    ret = ret->parent();
	return (GoGraphStruct*)ret;
    } else
	return NULL;
}

static GoGraphNode* GetGraphNode(Go* go)
{
    if (go) {
	Go* ret = go->parent();
	while (ret && !ret->getMetaclass()->isSubclass(GoGraphNode::metaclass()))
	    ret = ret->parent();
	return (GoGraphNode*)ret;
    } else
	return NULL;
}

static void SelectNode(GoRoot* root, const GPoint* abspt)
{
    Go* go = root->goAt(*abspt, 0.5);
    GoNode* node = GetNode(go);
    if (node)
	node->select();
}

static void SelectGraphStr(GoRoot* root, const GPoint* abspt)
{
    Go* go = root->goAt(*abspt, 0.5);
    GoGraphStruct* g = GetGraphStr(go);
    if (g)
	g->select();
}


static void SetTextLineHeight(Go* go, GNumber l)
{
    ((GoTextual*)go)->lineHeight (l);
}


static void change_rect(GoRectangle* rect, GoTextual* text)
{
    GRegion tmp_reg;
    text->addMyClip(tmp_reg);
    GBox box = tmp_reg.bigbox();

    GCoordSys cs(GPoint(box.xmin, box.ymin) - GVector(1,1),
		GVector(box.width() + 2, 0),
		GVector(0, box.height() + 2));
    rect->SetThisAbscs(cs);
}


static void MakeTree()
{
    GoTreeNode* nodes[NB_TREE_NODES];
    GoTree* tree;
    GoRectangle* rects[NB_TREE_NODES];
    int i;

    /*
     * Create tree and nodes
     */
    tree = GoTree::create();
    tree->reparent(goroot);
    tree->defaultLinkType(GoBusLink::metaclass());
    nodes[0] = tree->addNode(NULL);
    for (i = 1; i < 4; i++)
	nodes[i] = tree->addNode(nodes[0]);
    for (i = 4; i < 7; i++)
	nodes[i] = tree->addNode(nodes[1]);
    nodes[7] = tree->addNode(nodes[2]);
    for (i = 8; i < 10; i++)
	nodes[i] = tree->addNode(nodes[3]);
    for (i = 10; i < NB_TREE_NODES; i++)
	nodes[i] = tree->addNode(nodes[4]);

    /*
     * Create node's contents
     */
    for (i = 0; i < NB_TREE_NODES; i++) {
	rects[i] = GoRectangle::create();
	rects[i]->reparent(nodes[i]);
	rects[i]->map();
    }

    GoLabel* texts[NB_TREE_NODES];
    for (i = 0; i < NB_TREE_NODES; i++) {
	texts[i] = GoLabel::create();
	texts[i]->reparent(rects[i]);
	texts[i]->map();
	texts[i]->clipPostcond((void(*)(void*, Go*))change_rect, rects[i],
				 GADD);
	texts[i]->lineHeight(7);
	// texts[i]->underlinePos((GNumber)2/3);
    }
    texts[0]->setString("This", strlen("This"));
    texts[1]->setString("is", strlen("is"));
    texts[2]->setString("a", strlen("a"));
    texts[3]->setString("GoTree", strlen("GoTree"));
    texts[4]->setString("object", strlen("object"));
    texts[5]->setString("with", strlen("with"));
    texts[6]->setString("twelve", strlen("twelve"));
    texts[7]->setString("nodes", strlen("nodes"));
    texts[8]->setString("and", strlen("and"));
    texts[9]->setString("eleven", strlen("eleven"));
    texts[10]->setString("GoBusLink", strlen("GoBusLink"));
    texts[11]->setString("objects.", strlen("objects."));

    tree->anchoringType(GMIDDLE_WIDTH);
    tree->map();
    goroot->callClipPostconditions();
    tree->makePosition();
}

static void MakeGraph()
{
    GoGraphNode* nodes[NB_GRAPH_NODES];

    /*
     * Create the graph
     */
    GoGraph* graph = GoGraph::create();
    graph->defaultLinkType(GoSimpleLink::metaclass());
    graph->reparent(goroot);
    nodes[0] = graph->addNode(NULL);
    for (int i = 1; i < 4; i++)
	nodes[i] = graph->addNode(nodes[0]);
    nodes[4] = graph->addNode(nodes[1]);
    nodes[5] = graph->addNode(nodes[3]);
    nodes[6] = graph->addNode(nodes[4]);
    graph->addLink(nodes[2], nodes[4]);
    graph->addLink(nodes[2], nodes[5]);
    graph->addLink(nodes[5], nodes[6]);

    /*
     * Create node's contents
     */
    GoEllipse* ellipses[2];
    GoRectangle* rects[2];
    GoPolygon* polygons[2];
    GoCircle* circle;

    ellipses[0] = GoEllipse::create(27, 10);
    ellipses[0]->map();
    ellipses[0]->reparent(nodes[0]);
    ellipses[1] = GoEllipse::create(10, 27);
    ellipses[1]->map();
    ellipses[1]->reparent(nodes[4]);
    GShColor* c = xinterf->colorFromName("white");
    for (i = 0; i < 2; i++) {
	ellipses[i]->background(GTRUE);
	(ellipses[i]->brush()).color(c);
    }
    
    rects[0] = GoRectangle::create(15, 30);
    rects[0]->map();
    rects[0]->reparent(nodes[2]);
    rects[1] = GoRectangle::create(30, 15);
    rects[1]->map();
    rects[1]->reparent(nodes[6]);
    for (i = 0; i < 2; i++) {
	rects[i]->background(GTRUE);
	(rects[i]->brush()).color(c);
    }

    circle = GoCircle::create(15);
    circle->reparent(nodes[3]);
    circle->map();
    circle->background(GTRUE);
    (circle->brush()).color(c);
    c->destroy();

    GPoint points[6];
    points[0] = GPoint(-15, -5);
    points[1] = GPoint(15, -05);
    points[2] = GPoint(-15, 15);
    points[3] = GPoint(0, -15);
    points[4] = GPoint(15, 15);
    polygons[0] = GoPolygon::create(5, points);
    polygons[0]->map();
    polygons[0]->reparent(nodes[1]);
    points[0] = GPoint(-2.5, -15);
    points[1] = GPoint(12.5, -20);
    points[2] = GPoint(-12.5, 0);
    points[3] = GPoint(12.5, -1.0);
    points[4] = GPoint(-12.5, 20);
    points[5] = GPoint(12.5, 10);
    polygons[1] = GoPolygon::create(6, points);
    polygons[1]->map();
    polygons[1]->reparent(nodes[5]);
    c = xinterf->colorFromName("white");
   for (i = 0; i < 2; i++) {
	polygons[i]->background(GTRUE);
	(polygons[i]->brush()).color(c);
    }
    c->destroy();

    nodes[0]->move(GPoint(90, 23.0));
    nodes[1]->move(GPoint(25, 55));
    nodes[2]->move(GPoint(90, 55));
    nodes[3]->move(GPoint(150, 55));
    nodes[4]->move(GPoint(65, 90));
    nodes[5]->move(GPoint(130, 100));
    nodes[6]->move(GPoint(90, 130));
    
    graph->anchoringType(GCENTER);
    graph->map();
}

static void  _line_edited(void* closure, GoSensor* sensor)
{
    GoLine* line = (GoLine*)sensor->parent();

    if (sensor->interactor()->state() != GITOR_DONE)
    {
	line->destroy();
	return;
    }

    GoRoot* goroot = sensor->root();

    GPoint rel;
    if (line->first() == GPoint(0,0))
	rel = line->second();
    else
	rel = line->first();
	
    // compute in absolute;
    GPoint point = line->coordsys().to(rel);
    
    line->destroy();
    sensor = NULL;
    line = NULL;
    
    Go* go = goroot->goAt(point, 0.5);
    if (!go)
	return;
    GoGraphNode* next = GetGraphNode(go);
    if (!next)
	return;

    // add link;
    GoGraphNode* prev = (GoGraphNode*)closure;
    GoGraph* graph = GetGraph(prev);
    if  (graph)
    {
	(void)graph->addLink(prev, next);
	return;
    }
    GWarning("You can't add a link to a tree object.");
}

static void _point_designed(void* /* closure */, GoSensor* sensor)
{
    if (sensor->interactor()->state() != GITOR_DONE)
	return;

    GPoint point;
    point = ((GChoosePoint*)sensor->interactor())->choosenPoint();
    Go* go = sensor->root()->goAt(point, 0.5);
    if (!go)
	return;
    GoGraphNode* prev = GetGraphNode(go);
    if (!prev)
	return;

    // choose next;
    // by editing a line;
    GoLine* line = GoLine::create();
    line->move(point);
    line->reparent(sensor->root());
	
    printf("Clicks on next node.\n");
    (void)GoFullSensor::create("edit", line, NULL,
			       _line_edited, prev);
}

			      
static void AddLink()
{
    /*
     * designer les deux extremites du lien
     */
    printf("Clicks on previous node.\n");
    (void)GoFullSensor::create("choosePoint", goroot, NULL,
			       _point_designed, NULL);
}

static void RemoveNode(Go* go)
{
    GoTree* tree = NULL;
    GoGraph* graph = GetGraph(go);
    if (!graph) {
	tree = GetTree(go);
	tree->removeNode((GoTreeNode*)go);
    } else
	graph->removeNode((GoGraphNode*)go);
}

static void RemoveLink(GoLink* l)
{
    if (l) {
	GoGraph* g = GetGraph(l);
	if (g)
	    g->removeLink(l);
	else
	    GWarning("You can't remove a link in a tree object.");
    }
}

void* selectKey = (void*)"";

static GBool _clip_sensor_reparent_precond(void* /* closure */,
					   const Go* /* sensor */,
					   const Go* newpar,
					   const Go*(*next))
{
    if (newpar) {
	// put  the sensor ;
	*next = newpar->Child();
    }
    return GTRUE;
}

static void _select_postcond(void* /* closure */, Go* go)
{
    if (go->selected())
    {
	GoSensor* sensor;
	if (go->getMetaclass() == GoTree::metaclass() ||
	    go->getMetaclass() == GoGraph::metaclass() ||
	    go->getMetaclass()->isSubclass(GoNode::metaclass()))
	{
	    sensor = GoClipSensor::create();
	    sensor->program("selection");
	    // put it a the beginning of the tree/graph;
	    sensor->reparent(go, go->Child());
	    sensor->reparentPrecond(_clip_sensor_reparent_precond, NULL);
	}
	else
	    sensor = GoSensor::create("selection", go);

	go->assoc()->set(selectKey, sensor);
    }
    else
    {
	GoSensor* sensor = (GoSensor*)go->assoc()->get(selectKey);
	(void)sensor->destroy();
	go->assoc()->rem(selectKey);
    }
}

// END of MyGrabber definition;

static void Exit()
{
    GInputDispatcher()->stopEventLoop(0);
}

main()
{
    GTraceLevel() = 1;
    
    xinterf = new XGoInterf;
    xinterf->resize((GNumber)240.0,(GNumber)170.0);
    goroot =  (GoRoot*)GoRoot::create(xinterf);
    
    
    xinterf->map();
    goroot->map();
    xinterf->flush();

    (void)GoRegPolygon::metaclass();
    (void)GoRectangle::metaclass();
    (void)GoCircle::metaclass();
    (void)GoCurve::metaclass();
    (void)GoBezier::metaclass();
    (void)GoSimpleText::metaclass();
    (void)GoLabel::metaclass();
    (void)GoLine::metaclass();
    (void)GoClipSensor::metaclass();
    (void)Go::metaclass()->selectPostcond(_select_postcond, NULL);

    // load user programs;
    GLoadTranslatorPgms("demo.goui");

    // Create GoFullSensor referenced in the translator file;
    Go* sensor = GoFullSensor::create("goui", goroot, NULL, NULL, NULL);
    GoAlwaysFirstChild(sensor, GADD);

    // bind the correcponding actions;

    GBind("SetGoGeometricForeground", &SetGoGeometricForeground);
    GBind("SetGoGeometricBackground", &SetGoGeometricBackground);
    GBind("SetGoGeometricLineWidth", &SetGoGeometricLineWidth);
    GBind("SetGoLinkLineWidth", &SetGoLinkLineWidth);
    GBind("SetGoLinkForeground", &SetGoLinkForeground);
    GBind("SetGoGraphStructBackground",
	  &SetGoGraphStructBackground);
    GBind("SetGoNodeBackground",
	  &SetGoNodeBackground);
    GBind("MakeTree", &MakeTree);
    GBind("MakeGraph", &MakeGraph);
    GBind("AddLink", &AddLink);
    GBind("RemoveNode", &RemoveNode);
    GBind("RemoveLink", &RemoveLink);
    GBind("SelectNode", &SelectNode);
    GBind("SelectGraphStr", &SelectGraphStr);
    GBind("SetTextColor", &SetTextColor);
    GBind("SetTextBackColor", &SetTextBackColor);
    GBind("SetTextLineHeight", &SetTextLineHeight);
    GBind("Exit", &Exit);

    // edit line program not created in BGoUIEditor;
    (void)GEditLinePgm::create();
    
    // show a graph onto the window;
    MakeGraph();
    
    // INIT and RUN THE GLOBAL INPUT DISPATCHER;
    // I choose the X one;

    GInputDispatcher() = new XInputDispatcher(xinterf->display());

    // create the event to sensor dispatcher;
    // the BGoUIEditor (superclass of GoUI) handles;
    // also "selection" and "creation";

    // GoUI will rrun the GoRoot.goui translator
    BGoUIEditor* ed = new BGoUIEditor(goroot);
    
    // specific to XInputDispatcher;
    ((XInputDispatcher*)GInputDispatcher())->addXGoInterf(xinterf);
   
    // main loop; 
    (void)GInputDispatcher()->runEventLoop();

    goroot->destroy();
    delete xinterf;
    delete ed;
    delete GInputDispatcher();
}

