/* $Id: Go.h,v 1.12 1992/12/29 15:09:07 davy Exp $ */
/*****************************************************************************
MODULE:  Go Kernel
FILE:    Go.h
AUTHOR:  Oliver Jojic
DATE:    November 90

DESCRIPTION:
  The Go class is the basic class for each graphic object.
  The Go class provide methods to manage:
  	- go creation and destruction,
	- the go tree structure,
	- go coordinate system,
	- go drawings,
	- go saving and restoring,
	- and pre- and post- conditions on every modification.

  The Go class has almost no data storage. This allows implementation of
  "light weight gos".
  To implement new Gos you can use "Go...DS" classes (DS means data storage).

  See also mGo.h for Go metaclass informations.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/
#ifndef _Go_
#define _Go_

#include <Go/Copyright.h>
#include <Go/machdef.h>
#include <Go/Tools/CoordSys.h>
#include <Go/Tools/Callback.h>
#include <Go/Tools/Assoc.h>
#include <Go/Tools/Set.h>
#include <Go/Tools/Stack.h>
#include <Go/Tools/Queue.h>
#include <Go/Tools/Matrix.h>
#include <Go/Tools/GCdef.h>

class Go;
class GoMetaclass;
class GoRoot;
class GoInterf;
class GoSensor;
class GoStream;
class GPen;


/*
 * used by methods that navigate
 * in the go tree structure:
 */
#define GUNKNOWN_INDEX -111 	/* why not ? */


/*
 * Each go is a node of a go tree.
 * The go may have a cross type:
 *	- CrossAsNode:
 *		this go and its children are accessible by public methods,
 *	- CrossAsLeaf:
 *		the go is accessible by public methods but its children are not
 *	- CrossAsNothing:
 *		public methods cannot access this go.
 */
enum GCrossType
{
    GCROSS_AS_NODE,
    GCROSS_AS_LEAF,
    GCROSS_AS_NOTHING
};


/***********************************************
*
*		    Go class
*
************************************************/

class Go
{
  public:

    /*** go creation and destruction ***/

    Go* 	 dup ();
    Go*		 dupThis ();
    GBool	 callDupPrecond () const;

    static Go*	 create (const char* name);
    
    GBool 	 destroy ();
    GBool	 callDestroyPrecond ();

    /*** tree structure ***/

    Go*		 parent () const;
    Go*		 previous () const;
    Go*		 next () const;
    Go*		 child () const;
    Go*		 lastChild () const;
    int		 childIndex (const Go*) const;

    GBool        leaf () const;
    Go*          goAt (const GPoint& pos, GNumber maxdist) const;
    GoRoot*      root () const;
    GBool	 isAncestor (const Go*) const;
    GBool	 isUpper (const Go*) const;
    GBool	 isLower (const Go*) const;
    friend Go*   GoCommonAncestor (const Go*, const Go*);

    GBool        reparent (Go* newparent, Go* nextbrother=NULL);
    GBool        parentRelativeReparent (Go* newpar, Go* newnext=NULL);
    GBool	 shift (Go* nextbrother);
    GBool        raise ();
    GBool	 lower ();
    GBool        leaf (GBool);

    Go* 	 Previous (int myindex = GUNKNOWN_INDEX) const;
    Go*		 Next (int myindex = GUNKNOWN_INDEX) const;
    Go*          Child (int child_index = 0) const;
    Go*		 LastChild () const;
    GCrossType    crossType () const;
    Go*          GoAt (const GPoint& abspos, GNumber maxdist) const;
    Go*          Root () const;

    GBool        crossType (GCrossType);

    GBool 	callReparentPrecond (const Go* newParent, const Go*(&nextBrother)) const;
    GBool	callCrossTypePrecond (GCrossType) const;

    void*        forEveryGo (void(*)(void*, Go*), void* closure);
    void*	 forEveryMappedGo (void(*)(void*, Go*), void* closure);
    void*	 forEverySelectedGo (void(*)(void*, Go*), void* closure);
    void*        forEverySelectedSubtree (void(*)(void*, Go*), void*);

    void*        forEveryGoContainingPoint
			(void(*)(void*, Go*), void* closure,
			 const GPoint&, GNumber maxdist=0);
    void*        forEverySelectedGoContainingPoint
			(void(*)(void*, Go*), void*, const GPoint&, GNumber=0);
    void*        forEverySelectedSubtreeContainingPoint
			(void(*)(void*, Go*), void*, const GPoint&, GNumber=0);
    void         stopRecursion(void* return_value);

    void*        ForEveryGo (void(*)(void*, Go*), void* closure);
    void*	 ForEveryMappedGo (void(*)(void*, Go*), void* closure);
    void*	 ForEverySelectedGo (void(*)(void*, Go*), void* closure);
    void*        ForEverySelectedSubtree (void(*)(void*, Go*), void*);
    void*        ForEverySensor (void(*)(void*, GoSensor*), void*);
    void*        ForEveryGoContainingPoint
			(void(*)(void*, Go*), void* closure,
			 const GPoint&, GNumber maxdist=0);
    void*	 ForEverySensorContainingPoint
			(void(*)(void*, GoSensor*), void* closure,
			 const GPoint&, GNumber maxdist=0);


    /*** coordinate system ***/

    GCoordSys     coordsys (const Go* relative_to = NULL) const;
    GPoint        origin (const Go* relto = NULL) const;
    GMatrix       matrix (const Go* relto = NULL) const;
    GCoordSys     parentCoordsys () const;

    GBool         setCoordsys (const GCoordSys&, Go* relto = NULL);
    GBool         move (const GPoint&, Go* relto = NULL);
    GBool         setMatrix (const GMatrix&, Go* relto = NULL);
    GBool         rescale (GNumber xsize, GNumber ysize, Go* relto = NULL);
    GBool         rotate (GNumber angle, const GPoint& center, Go* relto = NULL);

    GCoordSys     Abscs (const GCoordSys& parabscs = *((GCoordSys*)NULL)) const;
    GCoordSys     Relcs (const GCoordSys& parabscs = *((GCoordSys*)NULL)) const;
    GBool         SetThisAbscs (const GCoordSys&);

    GBool	 callCSPrecond (GCoordSys& abscs) const;


    /*** about drawings ***/

    void	 draw () const;

    GBool 	 mapped () const;
    GBool        mappable () const;
    GBool        selected () const;
    GBool        contains (const GPoint&, GNumber maxdist=0) const;
    GBool        subtreeContains (const GPoint&, GNumber maxdist=0) const;

    //do not care about cross types:
    GBool        Contains (const GPoint& p, GNumber maxdist=0) const 
    				{ return m_contains (p, maxdist); }

    void         wiredDraw
	(const GCoordSys&, const GPen&, GQuality = GMAX_QUALITY) const;

    GRegion      myClip () const;
    GRegion      myOpaqueClip () const;
    GRegion      clip () const;
    GRegion      opaqueClip () const;
    void         addMyClip (GRegion&) const;
    void 	 addMyOpaqueClip (GRegion&) const;
    void	 addClip (GRegion&) const;

    //do not care about leaves:
    GRegion MyClip () const;
    GRegion MyOpaqueClip () const;
    void AddMyClip (GRegion&) const;
    void AddMyOpaqueClip (GRegion&) const;

    GBool        map ();
    GBool        unmap ();
    GBool        select ();
    GBool        unselect ();

    GBool 	callMapPrecond (GBool mapOrUnmap) const;
    GBool	callSelectPrecond (GBool selectOrUnselect) const;
    

    /*** misc ***/

    //the following static method must be redefined in each go subclass:
    static GoMetaclass* metaclass ();

    static GoMetaclass* metaclass (const char* classname);
    GoMetaclass* getMetaclass () const;
    void 	 check () const;
    GAssoc*	 assoc () const;
    GoInterf*    interf () const;
    GBool        isSensor () const;
    GBool        isToBeDestroyed() const;
    static int	 save (const Go*, GoStream*);
    static int	 readParams (Go*, GoStream*); // restores only;

    GBool	callSavePrecond () const;
    
    static void  differDestroy () { DifferDestroy(); }
    static void  handleDestroy () { HandleDestroy(); }

  protected:
    		 Go ();
    void	 MyClipHasChanged ();
    void	 RedrawForeOnly (const GRegion& = *(GRegion*)NULL);
    void         RedrawBackAndFore (const GRegion& = *(GRegion*)NULL);
    void	 TempMap ();
    void	 TempUnmap ();

  public:    //should be protected: bug compilo;

    static Go*	 Create (GoMetaclass*);
    Go*	 	 Dup (Go* parcopy, Go* prevcopy, GBool dupNextBrother, GBool dupChilren);
    void         DoReparent (Go* newParent, Go* newNext = NULL);

  protected:

    static void	 DifferDestroy();
    static void	 HandleDestroy ();
    GBool        WillBeDestroyed() const;
    
    GBool	 AddOrRemoveCondition (const char*, void* proc, void* closure,
				       GAddOrRemove);
    
    GBool	 CallPreconditions (const char*) const;
    GBool 	 CallPreconditions (const char*, void*) const;
    GBool	 CallPreconditions (const char*, void*, void*) const;
    GBool	 CallPreconditions (const char*, void*, void*, void*) const;
    GBool	 CallPreconditions (const char*,
				    void*, void*, void*, void*) const;
    GBool	 CallPreconditions (const char*,
				    void*, void*, void*, void*, void*) const;
    void	 CallPostconditions (const char*);
    void 	 CallPostconditions (const char*, void*);
    void 	 CallPostconditions (const char*, void*, void*);
    void 	 CallPostconditions (const char*, void*, void*, void*);
    void 	 CallPostconditions (const char*, void*, void*, void*, void*);
    void 	 CallPostconditions (const char*,
				     void*, void*, void*, void*, void*);

  public:

    /*** pre and post conditions ***/

    // creation and destruction conditions:

    void         dupPrecond (GBool(*)(void*, const Go*),
			     void* closure, GAddOrRemove=GADD);
    void         dupPostcond (void(*)(void*, Go* source, Go* copy),
			      void* closure, GAddOrRemove=GADD);
    void	 destroyPrecond	(GBool(*)(void*, const Go*), void* closure,
				 GAddOrRemove=GADD);
    void 	 destroyPostcond (void(*)(void*, const Go*), void* closure,
				  GAddOrRemove=GADD);


    // tree structure conditions:

    void         shiftPrecond (GBool(*)(void*, const Go*, const Go*(*pnextbr)),
			       void* closure, GAddOrRemove=GADD);
    void         shiftPostcond (void(*)(void*, Go*, Go* oldnextbr),
				void* closure, GAddOrRemove=GADD);
    void	 removeChildPrecond (GBool(*)(void*,const Go*,const Go*chtorem),
				     void* closure, GAddOrRemove=GADD);
    void         removeChildPostcond (void(*)(void*,Go*,Go* remch, Go* onext),
				      void* closure, GAddOrRemove=GADD);
    void         addChildPrecond (GBool(*)(void*, const Go*,
					   const Go* addch, const Go*(*pnext)),
				  void* closure, GAddOrRemove=GADD);
    void         addChildPostcond (void(*)(void*, Go*, Go* addch),
				   void* closure, GAddOrRemove=GADD);
    void         reparentPrecond (GBool(*)(void*, const Go*,
					  const Go* newpar, const Go*(*pnext)),
				  void* closure, GAddOrRemove=GADD);
    void	 reparentPostcond (void(*)(void*, Go*, Go* oldp, Go* oldnext),
				   void* closure, GAddOrRemove=GADD);
    void         rootPrecond (GBool(*)(void*, const Go*, const Go* newroot),
			      void* closure, GAddOrRemove=GADD);
    void	 rootPostcond (void(*)(void*, Go*, Go* oldroot),
			       void* closure, GAddOrRemove=GADD);
    void         crossTypePrecond (GBool(*)(void*, const Go*, GCrossType*),
				   void* closure, GAddOrRemove=GADD);
    void	 crossTypePostcond (void(*)(void*, Go*, const GCrossType* old),
				    void* closure, GAddOrRemove=GADD);

    // coordinate system conditions:

    void 	 relcsPrecond (GBool(*)(void*, const Go*, GCoordSys* newrel,
				       const GCoordSys* oldparabs,
				       const GCoordSys* newparabs),
			       void* closure, GAddOrRemove=GADD);
    void 	 relcsPostcond (void(*)(void*, Go*, const GCoordSys* parabs),
				void* closure, GAddOrRemove=GADD);
    void 	 abscsPrecond (GBool(*)(void*, const Go*, GCoordSys* newabs,
				       const GCoordSys* oldparabs,
				       const GCoordSys* newparabs),
			       void* closure, GAddOrRemove=GADD);
    void	 abscsPostcond (void(*)(void*, Go*, const GCoordSys* parabs),
				void* closure, GAddOrRemove=GADD);



    // drawing conditions:

    void         mapPrecond (GBool(*)(void*, const Go*),
			     void* closure, GAddOrRemove=GADD);
    void	 mapPostcond (void(*)(void*, Go*),
			      void* closure, GAddOrRemove=GADD);
    void         selectPrecond (GBool(*)(void*, const Go*),
				void* closure, GAddOrRemove=GADD);
    void         selectPostcond (void(*)(void*, Go*),
				 void* closure, GAddOrRemove=GADD);
    void	 myClipPostcond (void(*)(void*, Go*),
				 void* closure, GAddOrRemove=GADD);
    void	 clipPostcond (void(*)(void*, Go*),
			       void* closure, GAddOrRemove=GADD);


    // misc. conditions:

    void         interfPrecond (GBool(*)(void*, const Go*, const GoInterf*),
				void*, GAddOrRemove=GADD);
    void	 interfPostcond (void(*)(void*, Go*, GoInterf* oldinterf),
				 void* closure, GAddOrRemove=GADD);
    void         savePrecond (GBool(*)(void*, const Go*),
			      void* closure, GAddOrRemove=GADD);
    void         savePostcond (void(*)(void*, const Go*, GoStream*),
			       void* closure, GAddOrRemove=GADD);

  private:

#if !defined(__BORLANDC__) && !defined (__MSC7__)
    virtual     ~Go ();    // you have to use ::destroy();
#endif

    /*** virtual Go class definition ***/

    virtual GoMetaclass* m_metaclass () const;

  protected:

    virtual GBool 	 m_contains (const GPoint&, GNumber maxdist) const;
    virtual void	 m_add_clip (GRegion&) const;
    virtual void	 m_add_opaque_clip (GRegion&) const;
    virtual void 	 m_draw (const GRegion&) const;
    virtual void 	 m_wired_draw
			    (const GCoordSys&, const GPen&, GQuality) const;
    virtual void	 m_dup (const Go* orig, Go* parentCopy, Go* prevCopy,
				GBool dupNextBrother, GBool dupChildren);

#if defined(__BORLANDC__) || defined (__MSC7__)
    // BORLAND and MSC7 makes a mistake for the private interpretation;
    // The destructor shoul be private: it is called only by;
    // the Go class itself;
    // you have to use ::destroy();
    virtual     ~Go ();   
#endif

  private:

    /*** virtual data management ***/

    virtual GAssoc*	d_get_assoc () const = 0;

    virtual Go*		d_get_parent () const = 0;
    virtual Go*		d_get_previous () const = 0;
    virtual Go*		d_get_next () const = 0;
    virtual Go*		d_get_first_child () const = 0;

    virtual GBool	d_store_children_array () const = 0;
    virtual Go**	d_get_children (int& count) const;

    virtual void	d_set_previous (const Go*) = 0;
    virtual void	d_set_next (const Go*) = 0;
    virtual void	d_remove_child (const Go*, int child_index) = 0;
    virtual void 	d_add_child (const Go*, int child_index) = 0;
    virtual void	d_set_tree (const Go* parent,
				    const Go* prev, const Go* next) = 0;

    virtual GBool	d_store_abscs () const = 0;
    virtual GBool	d_store_relcs () const = 0;
    virtual const GCoordSys& d_get_relcs (const GCoordSys* parabs=NULL) const;
    virtual const GCoordSys& d_get_abscs (const GCoordSys* parabs=NULL) const;

    virtual void 	d_set_coordsys (const GCoordSys& relcs,
					const GCoordSys& abscs,
					const GCoordSys& parabs) = 0;

    virtual GRegion*	d_get_clip () const;


    /*** private virtual methods ***/

    virtual GBool	is_root () const;
    virtual GBool	is_sensor () const;


    /*** private methods ***/

    friend GoRoot;
    
    static GBool call_class_create_precond(const GoMetaclass*);
    void call_class_create_postcond ();
    GBool call_dup_precond () const;
    void call_dup_postcond (Go*);
    GBool call_class_dup_precond () const;
    void call_class_dup_postcond (Go*);
    GBool call_destroy_precond () const;
    GBool call_class_destroy_precond () const;
    void call_destroy_postcond ();
    void call_class_destroy_postcond ();
    GBool call_shift_precond (const Go*&, int, int) const;
    GBool call_class_shift_precond (const Go*&, int, int) const;
    void call_shift_postcond ();
    void call_class_shift_postcond ();
    GBool call_reparent_precond (const Go*, const Go*&) const;
    GBool call_class_reparent_precond (const Go*, const Go*&) const;
    void call_reparent_postcond (Go*, int);
    void call_class_reparent_postcond (Go*, int);
    GBool call_remove_child_precond (const Go*, int) const;
    GBool call_class_remove_child_precond (const Go*, int) const;
    void call_remove_child_postcond (Go*, int);
    void call_class_remove_child_postcond (Go*, int);
    GBool call_add_child_precond (const Go*, const Go*&, int) const;
    GBool call_class_add_child_precond (const Go*, const Go*&, int) const;
    void call_add_child_postcond (Go*, int);
    void call_class_add_child_postcond (Go*, int);
    GBool call_root_precond (const Go*) const;
    GBool call_class_root_precond (const Go*) const;
    void call_root_postcond (Go*);
    void call_class_root_postcond (Go*);
    GBool call_interf_precond (const GoInterf*) const;
    GBool call_class_interf_precond (const GoInterf*) const;
    void call_interf_postcond (GoInterf*);
    void call_class_interf_postcond (GoInterf*);
    GBool call_root_and_interf_precond_on_subtree (const Go*,
						  const GoInterf*) const;
    void call_root_and_interf_postcond_on_subtree (Go*, GoInterf*);
    GBool call_cross_type_precond (GCrossType&) const;
    GBool call_class_cross_type_precond (GCrossType&) const;
    void call_cross_type_postcond (GCrossType);
    void call_class_cross_type_postcond (GCrossType);
    GBool call_relcs_precond (GCoordSys&, const GCoordSys&, const GCoordSys&)const;
    GBool call_class_relcs_precond (GCoordSys&, const GCoordSys&,
				   const GCoordSys&) const;
    void call_relcs_postcond (const GCoordSys&);
    void call_class_relcs_postcond (const GCoordSys&);
    GBool call_abscs_precond (GCoordSys&, const GCoordSys&, const GCoordSys&)const;
    GBool call_class_abscs_precond (GCoordSys&, const GCoordSys&,
				   const GCoordSys&) const;
    void call_abscs_postcond (const GCoordSys&);
    void call_class_abscs_postcond (const GCoordSys&);
    GBool call_map_precond () const;
    GBool call_class_map_precond () const;
    void call_map_postcond ();
    void call_class_map_postcond ();
    GBool call_select_precond () const;
    GBool call_class_select_precond () const;
    void call_select_postcond ();
    void call_class_select_postcond ();
    void call_clip_postcond ();
    void call_my_clip_postcond ();
    void call_class_clip_postcond ();
    void call_class_my_clip_postcond ();
    void call_save_postcond (GoStream*) const;
    void call_class_restore_postcond (GoStream*);

    GBool call_interf_preconditions_on_subtree (const GoInterf*) const;
    void call_interf_postconditions_on_subtree (GoInterf*);

    GBool set_abscs (const GCoordSys&, GBool);
    void clip_not_ready ();
    GBool reparent_call_preconditions (int, const Go*,
				      int&, const Go*, const Go*&);
    void reparent_call_postconditions (int, Go*, int, Go*, Go*);
    void do_reparent (int, Go*, Go*, int, Go*, Go*);
    void force_reparent (Go*, Go* =NULL);
    void virtual_destroy_recur();
    GBool virtual_destroy(GBool);
    void effective_destroy_recur();
    void abs_reparent_setup_relcs (Go*, Go*, GCoordSys&, GCoordSys&);
    void abs_reparent_call_cs_postconditions ();
    void rel_reparent_setup_abscs (Go*, Go*, GCoordSys&, GCoordSys&);
    void rel_reparent_call_cs_postconditions ();
    void relrep_setcs_on_subtree (GCoordSys&, GCoordSys&);
    void set_cs_asis();
    void build_setthiscs_list_on_subtree (struct SetCSRec*&,
					  const GCoordSys&, const GCoordSys&);
    void build_setcs_list_on_subtree (struct SetCSRec*&,
				      const GCoordSys&, const GCoordSys&);
    void push_every_go(GStack&, GBool, GBool, GBool, GBool, GBool);
    void push_every_sensor(GStack&, const GPoint&, GNumber);
    void for_every_go_containing_point_recur
	(void(*)(void*,Go*), void*, const GPoint&, GNumber, int, GBool);
    void for_every_selected_subtree_containing_point_recur
	(void(*)(void*,Go*), void*, const GPoint&, GNumber);
    void for_every_sensor_containing_point_recur
	(void(*)(void*,GoSensor*), void*, const GPoint&, GNumber, int);
    GBool do_not_have_drawings () const;


    /*** data ***/

    static int differ_destroy /** = 0 **/;

    unsigned cross_type :2;
    GBool _mappable :1;
    GBool _selected :1;
    GBool stop_recursion :1;
    GBool clip_is_ready :1;
    GBool clip_was_updated :1;
    GBool my_clip_was_updated :1;
    GBool common_ancestor_mark :1;

    GBool has_dup_precond :1;
    GBool has_destroy_precond :1;
    GBool has_save_precond :1;
    GBool has_shift_precond :1;
    GBool has_reparent_precond :1;
    GBool has_remove_child_precond :1;
    GBool has_add_child_precond :1;
    GBool has_root_precond :1;
    GBool has_interf_precond :1;
    GBool has_cross_type_precond :1;
    GBool has_relcs_precond :1;
    GBool has_abscs_precond :1;
    GBool has_map_precond :1;
    GBool has_select_precond :1;

    GBool has_dup_postcond :1;
    GBool has_destroy_postcond :1;
    GBool has_save_postcond :1;
    GBool has_shift_postcond :1;
    GBool has_reparent_postcond :1;
    GBool has_remove_child_postcond :1;
    GBool has_add_child_postcond :1;
    GBool has_root_postcond :1;
    GBool has_interf_postcond :1;
    GBool has_cross_type_postcond :1;
    GBool has_relcs_postcond :1;
    GBool has_abscs_postcond :1;
    GBool has_map_postcond :1;
    GBool has_select_postcond :1;
    GBool has_clip_postcond :1;
    GBool has_my_clip_postcond :1;

    GBool draw_this:1;
    GBool draw_subtree:1;
    GBool is_to_be_destroyed:1;
};


void GoAlwaysFirstChild (Go* go, GAddOrRemove = GADD);
void GoAlwaysLastChild (Go* go, GAddOrRemove = GADD);



 	/***********************************************\
	*						*
	*		inline methods			*
	*						*
	\***********************************************/


INLINE Go* Go::dup ()
{
    return Dup(NULL, NULL, GFALSE, GTRUE);
}

INLINE Go* Go::dupThis ()
{
    return Dup (NULL, NULL, GFALSE, GFALSE);
}
    

INLINE Go* Go::parent () const
{
    return d_get_parent();
}

INLINE GCrossType Go::crossType () const
{
    return (GCrossType)cross_type;
}


INLINE GBool Go::leaf () const
{
    return (GBool)(crossType() == GCROSS_AS_LEAF);
}


INLINE Go* Go::goAt (const GPoint& pos, GNumber maxdist) const
{
    extern void go_at_proc(void*, Go*);
    return (Go*)((Go*)this)->forEveryGoContainingPoint
				(go_at_proc, NULL, pos, maxdist);
}

INLINE Go* Go::GoAt (const GPoint& pos, GNumber maxdist) const
{
    extern void go_at_proc(void*, Go*);
    return (Go*)((Go*)this)->ForEveryGoContainingPoint
				(go_at_proc, NULL, pos, maxdist);
}

INLINE GoRoot* Go::root () const
{
    Go* goroot = Root();
    return (GoRoot*)((goroot->is_root()) ? goroot : NULL);
}


INLINE GBool Go::shift (Go* nextbrother)
{
    return reparent(parent(), nextbrother);
}

INLINE GBool Go::raise ()
{
    return reparent(parent(), NULL);
}

INLINE GBool Go::lower ()
{
    Go* par = parent();
    if (par && par->Child() != this)
	return reparent(par, par->Child());
    return GTRUE;
}

INLINE GBool Go::leaf (GBool b)
{
    return crossType((b) ? GCROSS_AS_LEAF : GCROSS_AS_NODE);
}

INLINE GPoint Go::origin (const Go* relto) const
{
    return coordsys(relto).O;
}


INLINE GMatrix Go::matrix (const Go* relto) const
{
    GCoordSys cs(coordsys(relto));
    return GMatrix(cs.I, cs.J);
}


INLINE GCoordSys Go::parentCoordsys () const
{
    if (parent())
	return parent()->Abscs();

    return GUnitCoordSys();
}


INLINE GBool Go::setCoordsys (const GCoordSys& cs, Go* relto)
{
    if (relto)
	return set_abscs(relto->Abscs().to(cs), GFALSE);
    else
	return set_abscs(cs, GFALSE);
}

INLINE GBool Go::move (const GPoint& p, Go* relto)
{
    GCoordSys cs(coordsys(relto));
    cs.O = p;
    return setCoordsys(cs, relto);
}

INLINE GBool Go::setMatrix (const GMatrix& m, Go* relto)
{
    GCoordSys cs(coordsys(relto));
    cs.I = m.vx();
    cs.J = m.vy();
    return setCoordsys(cs, relto);
}

INLINE GBool Go::rescale (GNumber xsize, GNumber ysize, Go* relto)
{
    GCoordSys cs(coordsys(relto));
    cs.I.resize(xsize);
    cs.J.resize(ysize);
    return setCoordsys(cs, relto);
}

INLINE GBool Go::SetThisAbscs (const GCoordSys& abscs)
{
    return set_abscs(abscs, GTRUE);
}

INLINE GBool Go::mappable () const
{
    return _mappable;
}


INLINE GBool Go::isToBeDestroyed() const
{
    return is_to_be_destroyed;
}


INLINE GBool Go::selected () const
{
    return _selected;
}


INLINE void Go::wiredDraw
    (const GCoordSys& cs, const GPen& pen, GQuality q) const
{
    m_wired_draw(cs, pen, q);
}


INLINE void Go::AddMyOpaqueClip (GRegion& r) const
{
    m_add_opaque_clip(r);
}


INLINE GRegion Go::clip () const
{
    GRegion clip;
    addClip(clip);
    return clip;
}


INLINE GAssoc* Go::assoc () const
{
    return d_get_assoc();
}

INLINE GBool Go::isSensor () const
{
    return is_sensor();
}

INLINE void Go::clip_not_ready ()
{
    clip_is_ready = GFALSE;
    clip_was_updated = GTRUE;
}

INLINE GoMetaclass* Go::getMetaclass () const
{
    return m_metaclass();
}


INLINE void Go::set_cs_asis ()
{
    setCoordsys(Abscs());
}

INLINE GBool Go::isLower (const Go* go) const
{
    return go->isUpper(this);
}

#endif /* _Go_ */
