/* $Id: GoInterf.h,v 1.6 1992/12/11 08:34:45 jojic Exp $ */
/*****************************************************************************
MODULE:  GoInterface
FILE:    GoInterf.h
AUTHOR:  Jacques Davy
DATE:    August 1990

DESCRIPTION:
    The GoInterf object is the graphic system interface for output
    input or query operations.

    The GoInterf class is independent of the Go environment, it may be
    used by other environments. It offers a set of output methods, a
    way to get events and a set of query methods when device information
    is needed (e.g. font metrics, pixel size, ...).

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#ifndef _GoInterf_h
#define _GoInterf_h


#include <Go/machdef.h>
#include <Go/Tools/GCdef.h>
#include <Go/Event.h>
#include <Go/Tools/CoordSys.h>
#include <Go/Tools/Region.h>
#include <Go/Tools/Assoc.h>
#include <Go/Tools/Callback.h>

// public interfaces uses:

#include <Go/Tools/Color.h>
#include <Go/Tools/Colormap.h>
#include <Go/Tools/Pen.h>
#include <Go/Tools/Brush.h>
#include <Go/Tools/Font.h>
#include <Go/Tools/TextBrush.h>
#include <Go/Tools/ImageBrush.h>
#include <Go/Tools/PixelArray.h>
    
extern GoInterf*         GoDefaultGoInterf();


/*
 * Postcond flags
 */
struct GoInterfCondFlags
{
    GBool _has_expose_postcond :1;
    GBool _has_event_postcond :1;
    GBool _has_destroy_postcond :1;
    GBool _has_resize_postcond :1;
    GoInterfCondFlags() {
	_has_expose_postcond = GFALSE;
	_has_event_postcond = GFALSE;
	_has_destroy_postcond = GFALSE;
	_has_resize_postcond = GFALSE;
    }
};

/*
 * GoInterf class
 */

#define GCURRENT_TIME 0

class GoInterf
{
  public:
    virtual ~GoInterf();

    virtual void map();
    virtual void unmap();
    virtual void flush();
    virtual void synchronize (GBool sync);


    inline void setClip(const GRegion& region) { _clip = region; _clipCount++;}
    inline void resetClip() { _clip = GBiggestRegion(); _clipCount++; }

    // OUTPUT METHODS:

    // The following MUST be written in a subclass;
    // *******************************************;
	
    virtual void clear (const GRegion&) = 0;

    virtual void polyline (const GShPen*, const GShBrush*,
			   const GCoordSys&,
			   int npoint, const GPoint*,
			   GQuality = GMAX_QUALITY) = 0;
    inline  void polyline (const GPen*, const GBrush*,
			   const GCoordSys&,
			   int npoint, const GPoint*,
			   GQuality = GMAX_QUALITY);

    virtual void polygon  (const GShPen*, const GShBrush*,
			   const GCoordSys&, int npoint,
			   const GPoint*, GQuality = GMAX_QUALITY) = 0;
    inline  void polygon  (const GPen*, const GBrush*,
			   const GCoordSys&, int npoint,
			   const GPoint*, GQuality = GMAX_QUALITY);

    virtual void image     (const GShImageBrush*,
			    const GCoordSys&,  const GPoint&,
			    GNumber width, GNumber height,
			    const GShPixelArray*,
			    GQuality = GMAX_QUALITY) = 0;
    inline  void image     (const GImageBrush*,
			    const GCoordSys&,  const GPoint&,
			    GNumber width, GNumber height,
			    const GPixelArray*,
			    GQuality = GMAX_QUALITY);

    virtual void string    (const GShTextBrush*, const GShBrush*, 
			    const GCoordSys&, const GPoint&,
			    GNumber height, int nchar, const char*,
			    GQuality = GMAX_QUALITY) = 0;
    inline  void string    (const GTextBrush*, const GBrush*, 
			    const GCoordSys&, const GPoint&,
			    GNumber height, int nchar, const char*,
			    GQuality = GMAX_QUALITY);

    // Output done by the GoInterf using the above functions;
	
    virtual void line      (const GShPen*, const GCoordSys&,
			    const GPoint&, const GPoint&,
			    GQuality = GMAX_QUALITY);
    inline  void line      (const GPen*, const GCoordSys&,
			    const GPoint&, const GPoint&,
			    GQuality = GMAX_QUALITY);

    virtual void rectangle (const GShPen*, const GShBrush*,
			    const GCoordSys&, const GPoint&,
			    GNumber width, GNumber height,
			    GQuality = GMAX_QUALITY);
    inline  void rectangle (const GPen*, const GBrush*,
			    const GCoordSys&, const GPoint&,
			    GNumber width, GNumber height,
			    GQuality = GMAX_QUALITY);


    virtual void ellipse      (const GShPen*, const GShBrush*,
			       const GCoordSys&, const GPoint&,
			       GNumber radiusX,
			       GNumber radiusY,
			       GNumber angle1 = 0, GNumber angle2 = 2*PI,
			       GQuality = GMAX_QUALITY);
    inline  void ellipse      (const GPen*, const GBrush*,
			       const GCoordSys&, const GPoint&,
			       GNumber radiusX,
			       GNumber radiusY,
			       GNumber angle1 = 0, GNumber angle2 = 2*PI,
			       GQuality = GMAX_QUALITY);

    virtual void ellipseChord (const GShPen*, const GShBrush*,
			       const GCoordSys&,const GPoint&,
			       GNumber radiusX,   GNumber radiusY,
			       GNumber angle1, GNumber angle2, GBool closed,
			       GQuality = GMAX_QUALITY);
    inline  void ellipseChord (const GPen*, const GBrush*,
			       const GCoordSys&,const GPoint&,
			       GNumber radiusX,   GNumber radiusY,
			       GNumber angle1, GNumber angle2, GBool closed,
			       GQuality = GMAX_QUALITY);

    virtual void ellipsePie   (const GShPen*, const GShBrush*,
			       const GCoordSys&,const GPoint&,
			       GNumber radiusX, GNumber radiusY,
			       GNumber angle1, GNumber angle2, GBool closed,
			       GQuality = GMAX_QUALITY);
    inline  void ellipsePie   (const GPen*, const GBrush*,
			       const GCoordSys&,const GPoint&,
			       GNumber radiusX, GNumber radiusY,
			       GNumber angle1, GNumber angle2, GBool closed,
			       GQuality = GMAX_QUALITY);
    
    virtual void circle      (const GShPen*, const GShBrush*,
			      const GCoordSys&, const GPoint&,
			      GNumber radius,
			      GNumber angle1 = 0, GNumber angle2 = 2*PI,
			      GQuality = GMAX_QUALITY);
    inline  void circle      (const GPen*, const GBrush*,
			      const GCoordSys&, const GPoint&,
			      GNumber radius,
			      GNumber angle1 = 0, GNumber angle2 = 2*PI,
			      GQuality = GMAX_QUALITY);
    
    virtual void circleChord (const GShPen*, const GShBrush*,
			      const GCoordSys&, const GPoint&,
			      GNumber radius,
			      GNumber angle1, GNumber angle2, GBool closed,
			      GQuality = GMAX_QUALITY);
    inline  void circleChord (const GPen*, const GBrush*,
			      const GCoordSys&, const GPoint&,
			      GNumber radius,
			      GNumber angle1, GNumber angle2, GBool closed,
			      GQuality = GMAX_QUALITY);

    virtual void circlePie   (const GShPen*, const GShBrush*,
			      const GCoordSys&, const GPoint&,
			      GNumber radius,
			      GNumber angle1 , GNumber angle2, GBool closed,
			      GQuality = GMAX_QUALITY);
    inline  void circlePie   (const GPen*, const GBrush*,
			      const GCoordSys&, const GPoint&,
			      GNumber radius,
			      GNumber angle1 , GNumber angle2, GBool closed,
			      GQuality = GMAX_QUALITY);

    virtual void bezier     (const GShPen*, const GShBrush*,
			     const GCoordSys&,
			     int npoint, const GPoint*,
			     GQuality = GMAX_QUALITY);
    inline  void bezier     (const GPen*, const GBrush*,
			     const GCoordSys&,
			     int npoint, const GPoint*,
			     GQuality = GMAX_QUALITY);

    virtual void cubicCurve (const GShPen*, const GShBrush*,
			     const GCoordSys&,
			     int npoint, const GPoint*,
			     const GNumber (*basisMatrix)[4] /* [4][4] */,
			     int evalShift = 1, GQuality = GMAX_QUALITY);
    inline  void cubicCurve (const GPen*, const GBrush*,
			     const GCoordSys&,
			     int npoint, const GPoint*,
			     const GNumber (*basisMatrix)[4] /* [4][4] */,
			     int evalShift = 1, GQuality = GMAX_QUALITY);

    // POSTCOND MANAGEMENT;
	
    void exposePostcond (void (*)(void*, GoInterf*, const GExposeEvent*),
			 const void* closure, GAddOrRemove = GADD);
    void eventPostcond (void (*)(void*, GoInterf*, const GEvent*),
			const void* closure, GAddOrRemove = GADD);
    void destroyPostcond (void (*)(void*, GoInterf*),
			  const void* closure, GAddOrRemove = GADD);
    void resizePostcond (void (*)(void*, GoInterf*),
			 const void* closure, GAddOrRemove = GADD);

    // SETTING MISC METHODS:

    virtual void resize (GNumber width, GNumber height);
    virtual void background (const GShColor*);
    inline  void background (const GColor* bg) {
	background(bg->shColor());
    }
    
    // INPUT METHODS:

    // The following MUST be written in a subclass;
    // *******************************************;

    virtual GPoint getPointer(const GCoordSys* = NULL) = 0;
    virtual void   setPointer(const GPoint&, const GCoordSys* = NULL) = 0;
    virtual GBool  captureEvent(GEventMask input_mask, GBool confined_in_window,
				GTimeStamp reference = GCURRENT_TIME) = 0;
    virtual void   unCaptureEvent(GTimeStamp reference = GCURRENT_TIME) = 0;
    virtual ulong  inputFd() = 0;

    // Input done by the GoInterf using the above functions;
    
    GEventMask inputMask () const;

    virtual void addInputMask (GEventMask);
    virtual void removeInputMask (GEventMask);

    virtual void handle(GEvent*);
    
    // QUERY ON OUTPUT METHODS returns the absolute clip and opaque clip:
    // or the contains info;

    virtual void polylineClips (const GShPen*, const GShBrush*,
				const GCoordSys&,
				int npoint, const GPoint*,
				GRegion* clip, GRegion* opaque_clip,
				GQuality = GMAX_QUALITY) ;
    inline  void polylineClips (const GPen*, const GBrush*, const GCoordSys&,
				int npoint, const GPoint*,
				GRegion* clip, GRegion* opaque_clip,
				GQuality = GMAX_QUALITY) ;

    virtual GBool polylineContains (const GShPen*, const GShBrush*,
				   const GCoordSys&,
				   int npoint, const GPoint*,
				   const GPoint& absolute, GNumber maxDistance,
				   GQuality = GMAX_QUALITY) ;
    inline  GBool polylineContains (const GPen*, const GBrush*,
				   const GCoordSys&,
				   int npoint, const GPoint*,
				   const GPoint& absolute, GNumber maxDistance,
				   GQuality = GMAX_QUALITY) ;


    virtual void polygonClips  (const GShPen*, const GShBrush*,
				const GCoordSys&, int npoint,
				const GPoint*,
				GRegion* clip, GRegion* opaque_clip,
				GQuality = GMAX_QUALITY) ;
    inline  void polygonClips  (const GPen*, const GBrush*,
				const GCoordSys&, int npoint,
				const GPoint*,
				GRegion* clip, GRegion* opaque_clip,
				GQuality = GMAX_QUALITY) ;

    virtual GBool polygonContains  (const GShPen*, const GShBrush*,
				   const GCoordSys&, int npoint,
				   const GPoint*,
				   const GPoint& absolute, GNumber maxDistance,
				   GQuality = GMAX_QUALITY) ;
    inline  GBool polygonContains  (const GPen*, const GBrush*,
				   const GCoordSys&, int npoint,
				   const GPoint*,
				   const GPoint& absolute, GNumber maxDistance,
				   GQuality = GMAX_QUALITY) ;


    virtual void ellipseClips   (const GShPen*, const GShBrush*,
				 const GCoordSys&, const GPoint&,
				 GNumber radiusX, GNumber radiusY,
				 GRegion* clip, GRegion* opaque_clip,
				 GNumber angle1 = 0, GNumber angle2 = 2*PI,
				 GQuality = GMAX_QUALITY) ;
    inline  void ellipseClips   (const GPen*, const GBrush*,
				 const GCoordSys&, const GPoint&,
				 GNumber radiusX, GNumber radiusY,
				 GRegion* clip, GRegion* opaque_clip,
				 GNumber angle1 = 0, GNumber angle2 = 2*PI,
				 GQuality = GMAX_QUALITY) ;

    virtual GBool ellipseContains(const GShPen*, const GShBrush*,
				  const GCoordSys&, const GPoint&,
				  GNumber radiusX,  GNumber radiusY,
				  const GPoint& absolute, GNumber maxDistance,
				  GNumber angle1 = 0, GNumber angle2 = 2*PI,
				  GQuality = GMAX_QUALITY) ;
    inline  GBool ellipseContains(const GPen*, const GBrush*,
				  const GCoordSys&, const GPoint&,
				  GNumber radiusX,  GNumber radiusY,
				  const GPoint& absolute, GNumber maxDistance,
				  GNumber angle1 = 0, GNumber angle2 = 2*PI,
				  GQuality = GMAX_QUALITY) ;


    virtual void imageClips     (const GShImageBrush*,
				 const GCoordSys&,const GPoint&,
				 GNumber width, GNumber height,
				 GRegion* clip, GRegion* opaque_clip,
				 GQuality = GMAX_QUALITY) ;
    inline  void imageClips     (const GImageBrush*,
				 const GCoordSys&,const GPoint&,
				 GNumber width, GNumber height,
				 GRegion* clip, GRegion* opaque_clip,
				 GQuality = GMAX_QUALITY) ;

    virtual GBool imageContains  (const GShImageBrush*,
				  const GCoordSys&, const GPoint&,
				  GNumber width, GNumber height,
				  const GPoint& absolute, GNumber maxDistance,
				  GQuality = GMAX_QUALITY) ;
    inline  GBool imageContains  (const GImageBrush*,
				  const GCoordSys&, const GPoint&,
				  GNumber width, GNumber height,
				  const GPoint& absolute, GNumber maxDistance,
				  GQuality = GMAX_QUALITY) ;


    virtual void stringClips    (const GShTextBrush*, const GShBrush*, 
				 const GCoordSys&, const GPoint&,
				 GNumber height, int nchar, const char*,
				 GRegion* clip, GRegion* opaque_clip,
				 GQuality = GMAX_QUALITY) ;
    inline  void stringClips    (const GTextBrush*, const GBrush*, 
				 const GCoordSys&, const GPoint&,
				 GNumber height, int nchar, const char*,
				 GRegion* clip, GRegion* opaque_clip,
				 GQuality = GMAX_QUALITY) ;

    virtual GBool stringContains(const GShTextBrush*, const GShBrush*, 
				 const GCoordSys&, const GPoint&,
				 GNumber height, int nchar, const char*,
				 const GPoint& absolute, GNumber maxDistance,
				 GQuality = GMAX_QUALITY) ;
    inline  GBool stringContains(const GTextBrush*, const GBrush*, 
				 const GCoordSys&, const GPoint&,
				 GNumber height, int nchar, const char*,
				 const GPoint& absolute, GNumber maxDistance,
				 GQuality = GMAX_QUALITY) ;

    
    virtual void lineClips      (const GShPen*, const GCoordSys&,
				 const GPoint&, const GPoint&,
				 GRegion* clip, GRegion* opaque_clip,
				 GQuality = GMAX_QUALITY);
    inline  void lineClips      (const GPen*, const GCoordSys&,
				 const GPoint&, const GPoint&,
				 GRegion* clip, GRegion* opaque_clip,
				 GQuality = GMAX_QUALITY);

    virtual GBool lineContains  (const GShPen*, const GCoordSys&,
				 const GPoint&, const GPoint&,
				 const GPoint& absolute, GNumber maxDistance,
				 GQuality = GMAX_QUALITY);
    inline  GBool lineContains  (const GPen*, const GCoordSys&,
				 const GPoint&, const GPoint&,
				 const GPoint& absolute, GNumber maxDistance,
				 GQuality = GMAX_QUALITY);


    virtual void rectangleClips (const GShPen*, const GShBrush*,
				 const GCoordSys&, const GPoint&,
				 GNumber width, GNumber height,
				 GRegion* clip, GRegion* opaque_clip,
				 GQuality = GMAX_QUALITY);
    inline  void rectangleClips (const GPen*, const GBrush*,
				 const GCoordSys&, const GPoint&,
				 GNumber width, GNumber height,
				 GRegion* clip, GRegion* opaque_clip,
				 GQuality = GMAX_QUALITY);

    virtual GBool rectangleContains (const GShPen*, const GShBrush*,
				     const GCoordSys&, const GPoint&,
				     GNumber width, GNumber height,
				     const GPoint& absolute,
				     GNumber maxDistance,
				     GQuality = GMAX_QUALITY);
    inline  GBool rectangleContains (const GPen*, const GBrush*,
				     const GCoordSys&, const GPoint&,
				     GNumber width, GNumber height,
				     const GPoint& absolute,
				     GNumber maxDistance,
				     GQuality = GMAX_QUALITY);
    
    
    virtual void ellipseChordClips (const GShPen*, const GShBrush*,
				    const GCoordSys&, const GPoint&,
				    GNumber radiusX,  GNumber radiusY,
				    GNumber angle1, GNumber angle2,
				    GBool closed, GRegion* clip,
				    GRegion* opaque_clip,
				    GQuality = GMAX_QUALITY);
    inline  void ellipseChordClips (const GPen*, const GBrush*,
				    const GCoordSys&, const GPoint&,
				    GNumber radiusX,  GNumber radiusY,
				    GNumber angle1, GNumber angle2,
				    GBool closed, GRegion* clip,
				    GRegion* opaque_clip,
				    GQuality = GMAX_QUALITY);

    virtual GBool ellipseChordContains (const GShPen*, const GShBrush*,
					const GCoordSys&, const GPoint&,
					GNumber radiusX, GNumber radiusY,
					GNumber angle1, GNumber angle2,
					GBool closed, const GPoint& absolute,
					GNumber maxDistance,
					GQuality = GMAX_QUALITY);
    inline  GBool ellipseChordContains (const GPen*, const GBrush*,
					const GCoordSys&, const GPoint&,
					GNumber radiusX, GNumber radiusY,
					GNumber angle1, GNumber angle2,
					GBool closed, const GPoint& absolute,
					GNumber maxDistance,
					GQuality = GMAX_QUALITY);

    
    virtual void ellipsePieClips   (const GShPen*, const GShBrush*,
				    const GCoordSys&, const GPoint&,
				    GNumber radiusX, GNumber radiusY,
				    GNumber angle1, GNumber angle2,
				    GBool closed, GRegion* clip,
				    GRegion* opaque_clip,
				    GQuality = GMAX_QUALITY);
    inline  void ellipsePieClips   (const GPen*, const GBrush*,
				    const GCoordSys&, const GPoint&,
				    GNumber radiusX, GNumber radiusY,
				    GNumber angle1, GNumber angle2,
				    GBool closed, GRegion* clip,
				    GRegion* opaque_clip,
				    GQuality = GMAX_QUALITY);

    virtual GBool ellipsePieContains   (const GShPen*, const GShBrush*,
					const GCoordSys&, const GPoint&,
					GNumber radiusX, GNumber radiusY,
					GNumber angle1, GNumber angle2,
					GBool closed, const GPoint& absolute,
					GNumber maxDistance,
					GQuality = GMAX_QUALITY);
    inline  GBool ellipsePieContains   (const GPen*, const GBrush*,
					const GCoordSys&, const GPoint&,
					GNumber radiusX, GNumber radiusY,
					GNumber angle1, GNumber angle2,
					GBool closed, const GPoint& absolute,
					GNumber maxDistance,
					GQuality = GMAX_QUALITY);
    
    
    virtual void circleClips      (const GShPen*, const GShBrush*,
				   const GCoordSys&,  const GPoint&,
				   GNumber radius,
				   GRegion* clip, GRegion* opaque_clip,
				   GNumber angle1 = 0, GNumber angle2 = 2*PI,
				   GQuality = GMAX_QUALITY);
    inline  void circleClips      (const GPen*, const GBrush*,
				   const GCoordSys&,  const GPoint&,
				   GNumber radius,
				   GRegion* clip, GRegion* opaque_clip,
				   GNumber angle1 = 0, GNumber angle2 = 2*PI,
				   GQuality = GMAX_QUALITY);

    virtual GBool circleContains      (const GShPen*, const GShBrush*,
				       const GCoordSys&, const GPoint&,
				       GNumber radius,
				       const GPoint& absolute,
				       GNumber maxDistance, GNumber angle1 = 0,
				       GNumber angle2 = 2*PI,
				       GQuality = GMAX_QUALITY);
    inline  GBool circleContains      (const GPen*, const GBrush*,
				       const GCoordSys&, const GPoint&,
				       GNumber radius,
				       const GPoint& absolute,
				       GNumber maxDistance, GNumber angle1 = 0,
				       GNumber angle2 = 2*PI,
				       GQuality = GMAX_QUALITY);
    
    
    virtual void circleChordClips (const GShPen*, const GShBrush*,
				   const GCoordSys&,  const GPoint&,
				   GNumber radius,
				   GNumber angle1, GNumber angle2,
				   GBool closed, GRegion* clip,
				   GRegion* opaque_clip,
				   GQuality = GMAX_QUALITY);
    inline  void circleChordClips (const GPen*, const GBrush*,
				   const GCoordSys&,  const GPoint&,
				   GNumber radius,
				   GNumber angle1, GNumber angle2,
				   GBool closed, GRegion* clip,
				   GRegion* opaque_clip,
				   GQuality = GMAX_QUALITY);

    virtual GBool circleChordContains (const GShPen*, const GShBrush*,
				       const GCoordSys&,  const GPoint&,
				       GNumber radius,
				       GNumber angle1, GNumber angle2,
				       GBool closed, const GPoint& absolute,
				       GNumber maxDistance,
				       GQuality = GMAX_QUALITY);
    inline  GBool circleChordContains (const GPen*, const GBrush*,
				       const GCoordSys&,  const GPoint&,
				       GNumber radius,
				       GNumber angle1, GNumber angle2,
				       GBool closed, const GPoint& absolute,
				       GNumber maxDistance,
				       GQuality = GMAX_QUALITY);

    
    virtual void circlePieClips   (const GShPen*, const GShBrush*,
				   const GCoordSys&,  const GPoint&,
				   GNumber radius,
				   GNumber angle1 , GNumber angle2,
				   GBool closed, GRegion* clip,
				   GRegion* opaque_clip,
				   GQuality = GMAX_QUALITY);
    inline  void circlePieClips   (const GPen*, const GBrush*,
				   const GCoordSys&,  const GPoint&,
				   GNumber radius,
				   GNumber angle1 , GNumber angle2,
				   GBool closed, GRegion* clip,
				   GRegion* opaque_clip,
				   GQuality = GMAX_QUALITY);

    virtual GBool circlePieContains (const GShPen*, const GShBrush*,
				     const GCoordSys&,  const GPoint&,
				     GNumber radius,
				     GNumber angle1, GNumber angle2,
				     GBool closed, const GPoint& absolute,
				     GNumber maxDistance,
				     GQuality = GMAX_QUALITY);
    inline  GBool circlePieContains (const GPen*, const GBrush*,
				     const GCoordSys&,  const GPoint&,
				     GNumber radius,
				     GNumber angle1, GNumber angle2,
				     GBool closed, const GPoint& absolute,
				     GNumber maxDistance,
				     GQuality = GMAX_QUALITY);


    virtual void bezierClips (const GShPen*, const GShBrush*,
			      const GCoordSys&,
			      int npoint, const GPoint*,
			      GRegion* clip, GRegion* opaque_clip,
			      GQuality = GMAX_QUALITY);
    inline  void bezierClips (const GPen*, const GBrush*,
			      const GCoordSys&,
			      int npoint, const GPoint*,
			      GRegion* clip, GRegion* opaque_clip,
			      GQuality = GMAX_QUALITY);

    virtual GBool bezierContains (const GShPen*, const GShBrush*,
				  const GCoordSys&,
				  int npoint, const GPoint*,
				  const GPoint& absolute, GNumber maxDistance,
				  GQuality = GMAX_QUALITY);
    inline  GBool bezierContains (const GPen*, const GBrush*,
				  const GCoordSys&,
				  int npoint, const GPoint*,
				  const GPoint& absolute, GNumber maxDistance,
				  GQuality = GMAX_QUALITY);


    virtual void cubicCurveClips (const GShPen*, const GShBrush*,
				  const GCoordSys&,
				  int npoint, const GPoint*,
				  const GNumber (*basisMatrix)[4] /* [4][4] */,
				  GRegion* clip, GRegion* opaque_clip,
				  int evalShift = 1, GQuality = GMAX_QUALITY);
    inline  void cubicCurveClips (const GPen*, const GBrush*,
				  const GCoordSys&,
				  int npoint, const GPoint*,
				  const GNumber (*basisMatrix)[4] /* [4][4] */,
				  GRegion* clip, GRegion* opaque_clip,
				  int evalShift = 1, GQuality = GMAX_QUALITY);

    virtual GBool cubicCurveContains(const GShPen*, const GShBrush*,
				     const GCoordSys&,
				     int npoint, const GPoint*,
				     const GNumber (*basisMatrix)[4] /*[4][4]*/,
				     const GPoint& absolute,
				     GNumber maxDistance, int evalShift = 1,
				     GQuality = GMAX_QUALITY);
    inline  GBool cubicCurveContains(const GPen*, const GBrush*,
				     const GCoordSys&,
				     int npoint, const GPoint*,
				     const GNumber (*basisMatrix)[4] /*[4][4]*/,
				     const GPoint& absolute,
				     GNumber maxDistance, int evalShift = 1,
				     GQuality = GMAX_QUALITY);


    
    // QUERY MISC METHODS:
	
    // methods to be written in the subclass:

    virtual GNumber pixelWidth () const = 0;
    virtual GNumber pixelHeight () const = 0;
    virtual GShColor* colorFromName(const char* color_name) const = 0;

    // implementation using the above methods:
	
    inline  GNumber width() const { return _width; }
    inline  GNumber height() const { return _height; }
    virtual int toPixelWidth (GNumber mm) const;
    virtual int toPixelHeight (GNumber mm) const;
    inline  ushort depth() const { return _depth; }
    inline  const GColor* background() const { return _background; }
    
    // QUERY ON FONT AND SHFONT OBJECTS:
    // These functions are not const because they load the font;
    // They are to be written in the subclass;
    
    virtual const char*  fontName      (const GCoordSys&, GNumber height,
					const GShFont* font) = 0;
    inline  const char*  fontName      (const GCoordSys&, GNumber height,
					const GFont* font) ;


    virtual GNumber stringLeftBearing  (const char* string,
					int size, GNumber height,
					const GShFont* font) = 0;
    inline  GNumber stringLeftBearing  (const char* string,
					int size,
					GNumber height, const GFont* font) ;

    virtual GNumber stringRightBearing (const char* string,
					int size, GNumber height,
					const GShFont* font) = 0;
    inline  GNumber stringRightBearing (const char* string,
					int size,
					GNumber height, const GFont* font) ;

    virtual GNumber stringAscent       (const char* string,
					int size, GNumber height,
					const GShFont* font) = 0;
    inline  GNumber stringAscent       (const char* string,
					int size,
					GNumber height, const GFont* font) ;

    virtual GNumber stringDescent      (const char* string,
					int size, GNumber height,
					const GShFont* font) = 0;
    inline  GNumber stringDescent      (const char* string,
					int size,
					GNumber height, const GFont* font) ;

    virtual GNumber stringWidth        (const char* string,
					int size, GNumber height,
					const GShFont* font) = 0;
    inline  GNumber stringWidth        (const char* string,
					int size,
					GNumber height, const GFont* font) ;

    
    virtual GNumber fontAscent             (GNumber height,
					    const GShFont* font) = 0;
    inline  GNumber fontAscent             (GNumber height,
					    const GFont* font) ;

    virtual GNumber fontDescent            (GNumber height,
					    const GShFont* font) = 0;
    inline  GNumber fontDescent            (GNumber height,
					    const GFont* font) ;

    virtual GNumber fontUnderlineThickness (GNumber height,
					    const GShFont* font) = 0;
    inline  GNumber fontUnderlineThickness (GNumber height,
					    const GFont* font) ;

    virtual GNumber fontUnderlinePosition  (GNumber height,
					    const GShFont* font) = 0;
    inline  GNumber fontUnderlinePosition  (GNumber height,
					    const GFont* font) ;

    virtual GNumber fontItalicAngle        (const GShFont* font) = 0;
    inline  GNumber fontItalicAngle        (const GFont* font) ;

    virtual GNumber fontStrikeoutAscent    (GNumber height,
					    const GShFont* font) = 0;
    inline  GNumber fontStrikeoutAscent    (GNumber height,
					    const GFont* font) ;

    virtual GNumber fontStrikeoutDescent   (GNumber height,
					    const GShFont* font) = 0;
    inline  GNumber fontStrikeoutDescent   (GNumber height,
					    const GFont* font) ;

    virtual GNumber fontMinSpace         (GNumber height,
					  const GShFont* font) = 0;
    inline  GNumber fontMinSpace         (GNumber height,
					  const GFont* font) ;

    virtual GNumber fontMaxSpace         (GNumber height,
					  const GShFont* font) = 0;
    inline  GNumber fontMaxSpace         (GNumber height,
					  const GFont* font) ;

    virtual GNumber fontNormSpace        (GNumber height,
					  const GShFont* font) = 0;
    inline  GNumber fontNormSpace        (GNumber height,
					  const GFont* font) ;

    virtual GNumber fontEndSpace         (GNumber height,
					  const GShFont* font) = 0;
    inline  GNumber fontEndSpace         (GNumber height,
					  const GFont* font) ;
    
    // GET IMAGE:
    // to be written in the subclass;
    
    virtual void   getImage (GImageBrushValues& imagebrush_result,
			     GPixelArrayValues& pixelarray_result,
			     const GCoordSys& cs, const GPoint& origin,
			     GNumber width, GNumber height, ushort depth = 0,
			     GBool compress_colormap = GTRUE,
			     ushort x_resolution = 0, ushort y_resolution = 0,
			     GQuality quality = GMAX_QUALITY) = 0;
    
    // CURSOR;
    
    virtual void pushCursor (long cursorIdent);
    virtual void pushCursor (const class GCursor* cursor);
    virtual void pushCursor (const class GShCursor* shcursor);

    virtual void popCursor  ();

    // BELL;
    virtual void bell (int percent = 0);

  protected:

    GNumber _width;  //in mm;
    GNumber _height;
    ushort _depth;
    GCoordSys _device_cs;

    class GColor* _background;

    ushort _eventMaskCount[GN_EVENT_MASK];

    GRegion _clip;
    uint _clipCount;

    // conversions;
    
    inline int mm_to_linew(GNumber mm) const {
	int lw =  (int) (mm /((pixelWidth()+pixelHeight())/2));
	if (lw == 0) lw = 1;
	return lw;
    }

    // GCallback management;
    GoInterfCondFlags _cond;
    GAssoc _callbacks;

    int _addCallback (const char* type, GCBProc proc,
		      const void* closure) const;
    int _remCallback (const char* type, GCBProc proc,
		      const void* closure) const;

    void _call_expose_postcond (const GExposeEvent* event);
    void _call_event_postcond (const GEvent* event);
    void _call_destroy_postcond ();
    void _call_resize_postcond ();

    void _handle (const GEvent*);


    virtual void _transformPixelArrayValues(GPixelArrayValues* result,
					    ushort* xOrigin_result,
					    ushort* yOrigin_result,
					    const GShPixelArray* src,
					    const GCoordSys&   cs,
					    const GArray* pixel_translation,
					    uchar bits_per_pixel, uchar depth,
					    uchar bit_order,  uchar byte_order,
					    GQuality quality) const;

    virtual void _generalImageDeformation(GPixelArrayValues* result,
					  ushort* xOrigin_result,
					  ushort* yOrigin_result,
					  const GShPixelArray* src,
					  const GCoordSys&   cs,
					  const GArray* pixel_translation,
					  uchar bits_per_pixel, uchar depth,
					  uchar bit_order,  uchar byte_order,
					  GQuality quality) const;
    
    virtual GBool _resizeImage_bitsPerPixel1(GPixelArrayValues*,
					     const GShPixelArray*,
					     int new_width, int new_height,
					     uchar bits_per_pixel, uchar depth,
					     uchar bit_order, uchar byte_order,
					     GQuality quality) const;
    virtual GBool _resizeImage_bitsPerPixel4(GPixelArrayValues*, const GShPixelArray*,
					     const GArray* pixels_translation,
					     int new_width, int new_height,
					     uchar bits_per_pixel, uchar depth,
					     uchar bit_order, uchar byte_order,
					     GQuality quality) const;
    virtual GBool _resizeImage_bitsPerPixel8(GPixelArrayValues*,
					     const GShPixelArray*,
					     const GArray* pixels_translation,
					     int new_width, int new_height,
					     uchar bits_per_pixel, uchar depth,
					     uchar bit_order, uchar byte_order,
					     GQuality quality) const;
    virtual GBool _resizeImage(GPixelArrayValues*, const GShPixelArray*,
			       const GArray* pixels_translation,
			       int new_width, int new_height,
			       uchar bits_per_pixel, uchar depth,
			       uchar bit_order, uchar byte_order,
			       GQuality quality) const;
    
    /**
      next functions on image.
      the function does not lose information so
      they do not need the GQuality parameter
      **/
    virtual GBool _mirorXaxe_bitsPerPixel1(GPixelArrayValues*,
					   const GShPixelArray*,
					   uchar bits_per_pixel, uchar depth,
					   uchar bit_order, uchar byte_order) const;
    virtual GBool _mirorXaxe_bitsPerPixel4(GPixelArrayValues*,
					   const GShPixelArray*,
					   const GArray* pixels_translation,
					   uchar bits_per_pixel, uchar depth,
					   uchar bit_order, uchar byte_order) const;
    virtual GBool _mirorXaxe_bitsPerPixel8(GPixelArrayValues*,
					   const GShPixelArray*,
					   const GArray* pixels_translation,
					   uchar bits_per_pixel, uchar depth,
					   uchar bit_order, uchar byte_order) const;
    virtual GBool _mirorXaxe(GPixelArrayValues*, const GShPixelArray*,
			     const GArray* pixels_translation,
			     uchar bits_per_pixel, uchar depth,
			     uchar bit_order, uchar byte_order) const;
    
    virtual GBool _mirorYaxe_bitsPerPixel1(GPixelArrayValues*,
					   const GShPixelArray*,
					   uchar bits_per_pixel, uchar depth,
					   uchar bit_order, uchar byte_order) const;
    virtual GBool _mirorYaxe_bitsPerPixel4(GPixelArrayValues*,
					   const GShPixelArray*,
					   const GArray* pixels_translation,
					   uchar bits_per_pixel, uchar depth,
					   uchar bit_order, uchar byte_order) const;
    virtual GBool _mirorYaxe_bitsPerPixel8(GPixelArrayValues*,
					   const GShPixelArray*,
					   const GArray* pixels_translation,
					   uchar bits_per_pixel, uchar depth,
					   uchar bit_order, uchar byte_order) const;
    virtual GBool _mirorYaxe(GPixelArrayValues*, const GShPixelArray*,
			     const GArray* pixels_translation,
			     uchar bits_per_pixel, uchar depth,
			     uchar bit_order, uchar byte_order) const;
    
    virtual GBool _rotation180_bitsPerPixel1(GPixelArrayValues*,
					     const GShPixelArray*,
					     uchar bits_per_pixel, uchar depth,
					     uchar bit_order, uchar byte_order) const;
    virtual GBool _rotation180_bitsPerPixel4(GPixelArrayValues*,
					     const GShPixelArray*,
					     const GArray* pixels_translation,
					     uchar bits_per_pixel, uchar depth,
					     uchar bit_order, uchar byte_order) const;
    virtual GBool _rotation180_bitsPerPixel8(GPixelArrayValues*,
					     const GShPixelArray*,
					     const GArray* pixels_translation,
					     uchar bits_per_pixel, uchar depth,
					     uchar bit_order, uchar byte_order) const;
    virtual GBool _rotation180(GPixelArrayValues*, const GShPixelArray*,
			      const GArray* pixels_translation,
			      uchar bits_per_pixel, uchar depth,
			      uchar bit_order, uchar byte_order) const;


    virtual GBool _rotation90_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					   uchar bits_per_pixel, uchar depth,
					   uchar bit_order, uchar byte_order) const;
    virtual GBool _rotation90_bitsPerPixel4(GPixelArrayValues*, const GShPixelArray*,
					   const GArray* pixels_translation,
					   uchar bits_per_pixel, uchar depth,
					   uchar bit_order, uchar byte_order) const;
    virtual GBool _rotation90_bitsPerPixel8(GPixelArrayValues*, const GShPixelArray*,
					   const GArray* pixels_translation,
					   uchar bits_per_pixel, uchar depth,
					   uchar bit_order, uchar byte_order) const;
    virtual GBool _rotation90(GPixelArrayValues*, const GShPixelArray*,
			     const GArray* pixels_translation,
			     uchar bits_per_pixel, uchar depth,
			     uchar bit_order, uchar byte_order) const;

    /**
      these resize function are used for good performance
      for 2,3,4 reduction and 2,3,4 extension
      the quality is given in parameter.
      **/
    virtual GBool _reduceImage_R2_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					       int wanted_width, int wanted_height,
					       GQuality) const;

    virtual GBool _reduceImage_R3_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					       int wanted_width, int wanted_height,
					       GQuality) const;

    virtual GBool _reduceImage_R4_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					       int wanted_width, int wanted_height,
					       GQuality) const;
    
    virtual GBool _extendImage_E2_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					       int wanted_width, int wanted_height,
					       GQuality) const;

    virtual GBool _extendImage_E3_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					       int wanted_width, int wanted_height,
					       GQuality) const;

    virtual GBool _extendImage_E4_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					       int wanted_width, int wanted_height,
					       GQuality) const;

    GoInterf();

};



/*
 * INLINE METHODS:
 */
inline void GoInterf::polyline (const GPen* pen, const GBrush* brush,
				const GCoordSys& cs,
				int npoint, const GPoint* pts,
				GQuality quality)
{
    polyline (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
	      cs, npoint, pts, quality);
}

inline void GoInterf::polygon (const GPen* pen, const GBrush* brush,
			       const GCoordSys& cs, int npoint,
			       const GPoint* pts, GQuality quality)
{
    polygon (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
	     cs, npoint, pts, quality);
}

inline  void GoInterf::image (const GImageBrush* ibrush,
			      const GCoordSys& cs,  const GPoint& p,
			      GNumber width, GNumber height,
			      const GPixelArray* pxl_array,
			      GQuality quality)
{
    image (ibrush ? ibrush->shImageBrush():NULL,
	   cs, p, width, height,
	   pxl_array ? pxl_array->shPixelArray():NULL, quality);
}

inline void GoInterf::string (const GTextBrush* tbrush, const GBrush* brush, 
			      const GCoordSys& cs, const GPoint& p,
			      GNumber height, int nchar, const char* str,
			      GQuality quality)
{
    string (tbrush ? tbrush->shTextBrush():NULL, brush ? brush->shBrush():NULL,
	    cs, p, height, nchar, str, quality);
}

inline void GoInterf::line (const GPen* pen, const GCoordSys& cs,
			    const GPoint& p1, const GPoint& p2,
			    GQuality quality)
{
    line (pen ? pen->shPen():NULL, cs, p1, p2, quality);
}

inline void GoInterf::rectangle (const GPen* pen, const GBrush* brush,
				 const GCoordSys& cs, const GPoint& p,
				 GNumber width, GNumber height,
				 GQuality quality)
{
    rectangle (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
	       cs, p, width, height, quality);
}

inline void GoInterf::ellipse (const GPen* pen, const GBrush* brush,
			       const GCoordSys& cs, const GPoint& p,
			       GNumber radiusX, GNumber radiusY,
			       GNumber angle1, GNumber angle2,
			       GQuality quality)
{
    ellipse (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
	     cs, p, radiusX, radiusY, angle1, angle2, quality);
}

inline void GoInterf::ellipseChord (const GPen* pen, const GBrush* brush,
				    const GCoordSys& cs,const GPoint& p,
				    GNumber radiusX,   GNumber radiusY,
				    GNumber angle1, GNumber angle2,
				    GBool closed, GQuality quality)
{
    ellipseChord (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		  cs, p, radiusX, radiusY, angle1, angle2, closed, quality);
}


inline void GoInterf::ellipsePie (const GPen* pen, const GBrush* brush,
				  const GCoordSys& cs,const GPoint& p,
				  GNumber radiusX, GNumber radiusY,
				  GNumber angle1, GNumber angle2, GBool closed,
				  GQuality quality)
{
    ellipsePie (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		cs, p, radiusX, radiusY, angle1, angle2, closed, quality);
}

inline void GoInterf::circle (const GPen* pen, const GBrush* brush,
			      const GCoordSys& cs, const GPoint& p,
			      GNumber radius, GNumber angle1, GNumber angle2,
			      GQuality quality)
{
    circle (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
	    cs, p, radius, angle1, angle2, quality);
}

inline void GoInterf::circleChord (const GPen* pen, const GBrush* brush,
				   const GCoordSys& cs, const GPoint& p,
				   GNumber radius,
				   GNumber angle1, GNumber angle2,
				   GBool closed, GQuality quality)
{
    circleChord (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		 cs, p, radius, angle1, angle2, closed, quality);
}

inline void GoInterf::circlePie (const GPen* pen, const GBrush* brush,
				 const GCoordSys& cs, const GPoint& p,
				 GNumber radius,
				 GNumber angle1 , GNumber angle2, GBool closed,
				 GQuality quality)
{
    circlePie (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
	       cs, p, radius, angle1, angle2, closed, quality);
}

inline void GoInterf::bezier (const GPen* pen, const GBrush* brush,
			      const GCoordSys& cs,
			      int npoint, const GPoint* pts,
			      GQuality quality)
{
    bezier (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
	    cs, npoint, pts, quality);
}

inline void GoInterf::cubicCurve (const GPen* pen, const GBrush* brush,
				  const GCoordSys& cs,
				  int npoint, const GPoint* pts,
				  const GNumber (*basisMatrix)[4] /* [4][4] */,
				  int evalShift, GQuality quality)
{
    cubicCurve (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		cs, npoint, pts, basisMatrix, evalShift, quality);
}

/**/
inline void GoInterf::polylineClips (const GPen* pen, const GBrush* brush,
				     const GCoordSys& cs,
				     int npoint, const GPoint* pts,
				     GRegion* clip, GRegion* opaque_clip,
				     GQuality quality)
{
    polylineClips (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		   cs, npoint, pts, clip, opaque_clip, quality);
}

inline GBool GoInterf::polylineContains (const GPen* pen, const GBrush* brush,
					 const GCoordSys& cs,
					 int npoint, const GPoint* pts,
					 const GPoint& absolute,
					 GNumber maxDistance,
					 GQuality quality)
{
    return polylineContains (pen ? pen->shPen():NULL,
			     brush ? brush->shBrush():NULL,
			     cs, npoint, pts, absolute, maxDistance, quality);
}

inline void GoInterf::polygonClips (const GPen* pen, const GBrush* brush,
				    const GCoordSys& cs, int npoint,
				    const GPoint* pts,
				    GRegion* clip, GRegion* opaque_clip,
				    GQuality quality)
{
    polygonClips (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		  cs, npoint, pts, clip, opaque_clip, quality);
}

inline GBool GoInterf::polygonContains (const GPen* pen, const GBrush* brush,
					const GCoordSys& cs, int npoint,
					const GPoint* pts,
					const GPoint& absolute,
					GNumber maxDistance,
					GQuality quality)
{
    return polygonContains (pen ? pen->shPen():NULL,
			    brush ? brush->shBrush():NULL,
			    cs, npoint, pts, absolute, maxDistance, quality);
}

inline void GoInterf::ellipseClips (const GPen* pen, const GBrush* brush,
				    const GCoordSys& cs, const GPoint& p,
				    GNumber radiusX, GNumber radiusY,
				    GRegion* clip, GRegion* opaque_clip,
				    GNumber angle1, GNumber angle2,
				    GQuality quality)
{
    ellipseClips (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		  cs, p, radiusX, radiusY, clip, opaque_clip,
		  angle1, angle2, quality);
}

inline GBool GoInterf::ellipseContains (const GPen* pen, const GBrush* brush,
					const GCoordSys& cs, const GPoint& p,
					GNumber radiusX,  GNumber radiusY,
					const GPoint& absolute,
					GNumber maxDistance,
					GNumber angle1, GNumber angle2,
					GQuality quality)
{
    return ellipseContains (pen ? pen->shPen():NULL,
			    brush ? brush->shBrush():NULL,
			    cs, p, radiusX, radiusY, absolute, maxDistance,
			    angle1, angle2, quality);
}

inline void GoInterf::imageClips (const GImageBrush* ibrush,
				  const GCoordSys& cs, const GPoint& p,
				  GNumber width, GNumber height,
				  GRegion* clip, GRegion* opaque_clip,
				  GQuality quality)
{
    imageClips (ibrush ? ibrush->shImageBrush():NULL,
		cs, p, width, height, clip, opaque_clip, quality);
}

inline GBool GoInterf::imageContains (const GImageBrush* ibrush,
				      const GCoordSys& cs, const GPoint& p,
				      GNumber width, GNumber height,
				      const GPoint& absolute,
				      GNumber maxDistance,
				      GQuality quality)
{
    return imageContains (ibrush ? ibrush->shImageBrush():NULL,
			  cs, p, width, height, absolute,
			  maxDistance, quality);
}

inline void GoInterf::stringClips (const GTextBrush* tbrush,
				   const GBrush* brush,
				   const GCoordSys& cs, const GPoint& p,
				   GNumber height, int nchar, const char* str,
				   GRegion* clip, GRegion* opaque_clip,
				   GQuality quality)
{
    stringClips (tbrush ? tbrush->shTextBrush():NULL,
		 brush ? brush->shBrush():NULL,
		 cs, p, height, nchar, str, clip, opaque_clip, quality);
}

inline GBool GoInterf::stringContains (const GTextBrush* tbrush,
				       const GBrush* brush,
				       const GCoordSys& cs, const GPoint& p,
				       GNumber height, int nchar,
				       const char* str, const GPoint& absolute,
				       GNumber maxDistance,
				       GQuality quality)
{
    return stringContains (tbrush ? tbrush->shTextBrush():NULL,
			   brush ? brush->shBrush():NULL,
			   cs, p, height, nchar, str, absolute, maxDistance,
			   quality);
}

inline void GoInterf::lineClips (const GPen* pen, const GCoordSys& cs,
				 const GPoint& p1, const GPoint& p2,
				 GRegion* clip, GRegion* opaque_clip,
				 GQuality quality)
{
    lineClips (pen ? pen->shPen():NULL,
	       cs, p1, p2, clip, opaque_clip, quality);
}

inline GBool GoInterf::lineContains (const GPen* pen, const GCoordSys& cs,
				     const GPoint& p1, const GPoint& p2,
				     const GPoint& absolute,
				     GNumber maxDistance,
				     GQuality quality)
{
    return lineContains (pen ? pen->shPen():NULL,
			 cs, p1, p2, absolute, maxDistance, quality);
}

inline void GoInterf::rectangleClips (const GPen* pen, const GBrush* brush,
				      const GCoordSys& cs, const GPoint& p,
				      GNumber width, GNumber height,
				      GRegion* clip, GRegion* opaque_clip,
				      GQuality quality)
{
    rectangleClips (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		    cs, p, width, height, clip, opaque_clip, quality);
}

inline GBool GoInterf::rectangleContains (const GPen* pen, const GBrush* brush,
					  const GCoordSys& cs, const GPoint& p,
					  GNumber width, GNumber height,
					  const GPoint& absolute,
					  GNumber maxDistance,
					  GQuality quality)
{
    return rectangleContains (pen ? pen->shPen():NULL,
			      brush ? brush->shBrush():NULL,
			      cs, p, width, height, absolute, maxDistance,
			      quality);
}

inline void GoInterf::ellipseChordClips (const GPen* pen, const GBrush* brush,
					 const GCoordSys& cs, const GPoint& p,
					 GNumber radiusX,  GNumber radiusY,
					 GNumber angle1, GNumber angle2,
					 GBool closed, GRegion* clip,
					 GRegion* opaque_clip,
					 GQuality quality)
{
    ellipseChordClips (pen ? pen->shPen():NULL,
		       brush ? brush->shBrush():NULL,
		       cs, p, radiusX, radiusY, angle1, angle2,
		       closed, clip, opaque_clip, quality);
}
    
inline GBool GoInterf::ellipseChordContains (const GPen* pen,
					     const GBrush* brush,
					     const GCoordSys& cs,
					     const GPoint& p,
					     GNumber radiusX, GNumber radiusY,
					     GNumber angle1, GNumber angle2,
					     GBool closed,
					     const GPoint& absolute,
					     GNumber maxDistance,
					     GQuality quality)
{
    return ellipseChordContains (pen ? pen->shPen():NULL,
				 brush ? brush->shBrush():NULL,
				 cs, p, radiusX, radiusY, angle1, angle2,
				 closed, absolute, maxDistance, quality);
}

inline void GoInterf::ellipsePieClips (const GPen* pen, const GBrush* brush,
				       const GCoordSys& cs, const GPoint& p,
				       GNumber radiusX, GNumber radiusY,
				       GNumber angle1, GNumber angle2,
				       GBool closed, GRegion* clip,
				       GRegion* opaque_clip,
				       GQuality quality)
{
    ellipsePieClips (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		     cs, p, radiusX, radiusY, angle1, angle2,
		     closed, clip, opaque_clip, quality);
}

inline GBool GoInterf::ellipsePieContains(const GPen* pen, const GBrush* brush,
					  const GCoordSys& cs, const GPoint& p,
					  GNumber radiusX, GNumber radiusY,
					  GNumber angle1, GNumber angle2,
					  GBool closed, const GPoint& absolute,
					  GNumber maxDistance,
					  GQuality quality)
{
    return ellipsePieContains(pen ? pen->shPen():NULL,
			      brush ? brush->shBrush():NULL,
			      cs, p, radiusX, radiusY, angle1, angle2,
			      closed, absolute, maxDistance, quality);
}

inline void GoInterf::circleClips (const GPen* pen, const GBrush* brush,
				   const GCoordSys& cs, const GPoint& p,
				   GNumber radius,
				   GRegion* clip, GRegion* opaque_clip,
				   GNumber angle1, GNumber angle2,
				   GQuality quality)
{
    circleClips (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		 cs, p, radius, clip, opaque_clip, angle1, angle2, quality);
}

inline GBool GoInterf::circleContains (const GPen* pen, const GBrush* brush,
				       const GCoordSys& cs, const GPoint& p,
				       GNumber radius,
				       const GPoint& absolute,
				       GNumber maxDistance, GNumber angle1,
				       GNumber angle2, GQuality quality)
{
    return circleContains (pen ? pen->shPen():NULL,
			   brush ? brush->shBrush():NULL,
			   cs, p, radius, absolute, maxDistance,
			   angle1, angle2, quality);
}

inline void GoInterf::circleChordClips (const GPen* pen, const GBrush* brush,
					const GCoordSys& cs, const GPoint& p,
					GNumber radius,
					GNumber angle1, GNumber angle2,
					GBool closed, GRegion* clip,
					GRegion* opaque_clip,
					GQuality quality)
{
    circleChordClips (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		      cs, p, radius, angle1, angle2, closed,
		      clip, opaque_clip, quality);
}

inline GBool GoInterf::circleChordContains (const GPen* pen,
					    const GBrush* brush,
					    const GCoordSys& cs,
					    const GPoint& p,
					    GNumber radius,
					    GNumber angle1, GNumber angle2,
					    GBool closed,
					    const GPoint& absolute,
					    GNumber maxDistance,
					    GQuality quality)
{
    return circleChordContains (pen ? pen->shPen():NULL,
				brush ? brush->shBrush():NULL,
				cs, p, radius, angle1, angle2, closed,
				absolute, maxDistance, quality);
}

inline void GoInterf::circlePieClips (const GPen* pen, const GBrush* brush,
				      const GCoordSys& cs, const GPoint& p,
				      GNumber radius,
				      GNumber angle1 , GNumber angle2,
				      GBool closed, GRegion* clip,
				      GRegion* opaque_clip,
				      GQuality quality)
{
    circlePieClips (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		    cs, p, radius, angle1, angle2, closed, clip, opaque_clip,
		    quality);
}

inline GBool GoInterf::circlePieContains (const GPen* pen, const GBrush* brush,
					  const GCoordSys& cs, const GPoint& p,
					  GNumber radius,
					  GNumber angle1, GNumber angle2,
					  GBool closed, const GPoint& absolute,
					  GNumber maxDistance,
					  GQuality quality)
{
    return circlePieContains (pen ? pen->shPen():NULL,
			      brush ? brush->shBrush():NULL,
			      cs, p, radius, angle1, angle2, closed,
			      absolute, maxDistance, quality);
}

inline void GoInterf::bezierClips (const GPen* pen, const GBrush* brush,
				   const GCoordSys& cs,
				   int npoint, const GPoint* pts,
				   GRegion* clip, GRegion* opaque_clip,
				   GQuality quality)
{
    bezierClips (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		 cs, npoint, pts, clip, opaque_clip, quality);
}

inline GBool GoInterf::bezierContains (const GPen* pen, const GBrush* brush,
				       const GCoordSys& cs,
				       int npoint, const GPoint* pts,
				       const GPoint& absolute,
				       GNumber maxDistance,
				       GQuality quality)
{
    return bezierContains (pen ? pen->shPen():NULL,
			   brush ? brush->shBrush():NULL,
			   cs, npoint, pts, absolute, maxDistance, quality);
}

inline void GoInterf::cubicCurveClips (const GPen* pen, const GBrush* brush,
				       const GCoordSys& cs,
				       int npoint, const GPoint* pts,
				       const GNumber (*basisMatrix)[4],
				       GRegion* clip, GRegion* opaque_clip,
				       int evalShift, GQuality quality)
{
    cubicCurveClips (pen ? pen->shPen():NULL, brush ? brush->shBrush():NULL,
		     cs, npoint, pts, basisMatrix, clip, opaque_clip,
		     evalShift, quality);
}

inline GBool GoInterf::cubicCurveContains(const GPen* pen, const GBrush* brush,
					  const GCoordSys& cs,
					  int npoint, const GPoint* pts,
					  const GNumber (*basisMatrix)[4],
					  const GPoint& absolute,
					  GNumber maxDistance, int evalShift,
					  GQuality quality)
{
    return cubicCurveContains (pen ? pen->shPen():NULL,
			       brush ? brush->shBrush():NULL,
			       cs, npoint, pts, basisMatrix, absolute,
			       maxDistance, evalShift, quality);
}


/**/
inline GNumber GoInterf::stringLeftBearing (const char* string,  int size,
					    GNumber height, const GFont* font) 
{
    return stringLeftBearing (string, size, height, font->shFont());
}

inline GNumber GoInterf::stringRightBearing (const char* string,   int size,
					     GNumber height, const GFont* font)
{
    return stringRightBearing (string, size, height, font->shFont());
}

inline GNumber GoInterf::stringAscent (const char* string,  int size,
				       GNumber height, const GFont* font) 
{
    return stringAscent (string, size, height, font->shFont());
}

inline GNumber GoInterf::stringDescent (const char* string, int size,
					GNumber height, const GFont* font) 
{
    return stringDescent (string, size, height, font->shFont());
}

inline GNumber GoInterf::stringWidth (const char* string,  int size,
				      GNumber height, const GFont* font) 
{
    return stringWidth (string, size, height, font->shFont());
}

inline const char* GoInterf::fontName (const GCoordSys& cs, GNumber height,
				       const GFont* font)
{
    return fontName (cs, height, font->shFont());
}

inline GNumber GoInterf::fontAscent (GNumber height,
				     const GFont* font) 
{
    return fontAscent (height, font->shFont());
}

inline GNumber GoInterf::fontDescent (GNumber height,
				      const GFont* font) 
{
    return fontDescent (height, font->shFont());
}

inline GNumber GoInterf::fontUnderlineThickness (GNumber height,
						 const GFont* font) 
{
    return fontUnderlineThickness (height, font->shFont());
}

inline GNumber GoInterf::fontUnderlinePosition (GNumber height,
						const GFont* font) 
{
    return fontUnderlinePosition (height, font->shFont());
}

inline GNumber GoInterf::fontItalicAngle (const GFont* font) 
{
    return fontItalicAngle (font->shFont());
}

inline GNumber GoInterf::fontStrikeoutAscent (GNumber height,
					      const GFont* font) 
{
    return fontStrikeoutAscent (height, font->shFont());
}

inline GNumber GoInterf::fontStrikeoutDescent (GNumber height,
					       const GFont* font) 
{
    return fontStrikeoutDescent (height, font->shFont());
}

inline GNumber GoInterf::fontMinSpace (GNumber height,
				       const GFont* font) 
{
    return fontMinSpace (height, font->shFont());
}

inline GNumber GoInterf::fontMaxSpace (GNumber height,
				       const GFont* font) 
{
    return fontMaxSpace (height, font->shFont());
}

inline GNumber GoInterf::fontNormSpace (GNumber height,
					const GFont* font) 
{
    return fontNormSpace (height, font->shFont());
}

inline GNumber GoInterf::fontEndSpace (GNumber height,
				       const GFont* font) 
{
    return fontEndSpace (height, font->shFont());
}

    
#endif /* _GoInterf_h */

