/* $Id: Interactor.h,v 1.1 1992/12/10 15:37:39 jojic Exp $ */
/*****************************************************************************
MODULE:  Go Kernel (User Interface)
FILE:    Interactor.h
AUTHOR:  Oliver Jojic
DATE:    June 92

DESCRIPTION:

  The GInteractor class is used for the programming and execution of user
  interactions.

  Examples of user interactions:
	creation of an ellipse arc,
	edition of a text,
	dialing a phone number by using voice recognition.

  We want to allow the programation of complex interactions (i.e. creation
  of an ellipse arc) without making it difficult to program obvious ones
  (ala event translation table). This is why the basic GInteractor class
  is virtual. In fact the GInteractor class permits the execution of
  interactions, and GInteractor subclasses provide different programations
  of user interactions.

  The fact that there is a common superclass to all objects serving the
  programation of user interactions allows also their grouping
  (their execution in parallel or one after the other).

  Each new execution of the user interaction defined in an interactor
  must be preceded by the call of the virtual ``start()'' method.
  This method has two parameters: an event and a closure.
  The event may be NULL, the closure (of type void*) is specific to each
  interactor subclass.
  The ``handle(event)'' method interprets ``event'' using rules
  defined in this interactor instance.
  The ``end()'' method must always be called when the execution of an
  interactor is finished.

  In general, the programation of interactions is done a long time before
  their use. It is then usefull to have a way to retrieve them.  This is why
  you may set a name to an interactor. This name is unique. The function
  ``GInteractorFromName(name)'' returns the interactor whos name is ``name''.
  For instance, you build an interactor asking the user to choose a Go on its
  screen. You call this interactor "ChooseAGo". Later, anybody will be able
  to use this interactor knowing only its name.
  Therefore a named interactor is public, i.e. it may be accessed by many
  people at a time, but a running interactor may update its internal data.
  To avoid this conflict, a named interactor (name != NULL) is ``const'',
  i.e. you cannot call ``start'', ``handle'' or ``end'' methods
  (of course, you may continue to program it).
  To execute the user interaction defined in a named interactor you have to
  make a copy of the interactor using the ``dup'' method. The interactor
  returned by the ``dup'' method has a NULL name, and therefore you can
  execute its interaction.
  If you change the programation of the original interactor, the copy is not
  changed.

  It is possible to have a more structured use of interactor names
  (see Sensor(3Go)).

  An interactor may have a return value. When it is finished, you can get it
  by calling the virtual ``returnValue()'' method.

  In conclusion, you will learn different methods of programing user
  interactions by reading the documentation of differant subclasses
  of GInteractor.


COPYRIGHT:
  Copyright (c) 1990, 1991, 1992 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.

*****************************************************************************/
#ifndef _Interactor_h
#define _Interactor_h


#include <Go/machdef.h>
#include <Go/EventDefs.h>
#include <Go/Tools/Bool.h>

class Go;
class GEvent;
class GInteractorPgm;


/* Possible states of an interactor */
enum GItorState
{
	GITOR_RUNNING,
	GITOR_DONE,
	GITOR_CANCELLED
};


/*
 *-------------------------------------------------------
 * The GInteractor class
 *-------------------------------------------------------
 */

class GInteractor
{
  public:

    friend GInteractor* GCreateInteractor
	(const char* actionName, const GEvent* startEvent, Go* go);
    virtual ~GInteractor ();

    GEventMask inputMask () const;
    GItorState state () const { return _state; }
    GInteractorPgm* pgm () const { return (GInteractorPgm*)_pgm; }
    void* returnValue () const { return return_value; }
    Go* go () const { return _go; }
    
    virtual GBool careAbout (const GEvent*) const;
    virtual GBool handle (const GEvent*) = 0;
    virtual void doit (); //superclass first
    virtual void cancel (); //superclass first

  protected:

    GInteractor (const GInteractorPgm*, Go*);
    virtual void Initialize (); //superclass last;

    /* protected data */
    void* return_value;
    Go* _go;
    
  private:

    const GInteractorPgm* _pgm;
    GItorState _state;

    friend class GInteractorPgm;
};


#endif /* _Interactor_h */






