/* $Id: Sensor.h,v 1.8 1992/12/30 13:30:45 jojic Exp $ */
/*****************************************************************************
MODULE: Graphic Objects
FILE:   Sensor.h
AUTHOR: jojic
DATE:   Jun 29 1992

DESCRIPTION :
  To avoid mixing graphics and input management while implementing new
  graphic objects, Gos are not sensitive to events.

  GoSensor and its subclasses are the only sensitive objects inside the
  Go world.

  A GoSensor is a Go. The basic GoSensor class is sensitive onto the full
  output of its parent. GoSensor subclasses implement a different
  sensitivity (onto the full screen, onto a rectangular area, ...).

  The execution of a behavior with regard to user events is done by
  the GInteractor object. Each sensor has an associated interactor.
  GoUI looks for the destination of an event, i.e. a sensor, and
  this sensor sends that event to its interactor.

  A sensor has a name defining what it has to do. This name is used
  to retreive the good interactor. In fact, the name of the class of the go
  parent of the sensor is also used.

  A little example will clarify this:
     GoCircle is a subclass of GoEllipse. The interaction to rescale
     a GoCircle is the same than the one used to rescale an ellipse.
     Therefore, there is an interactor named "GoEllipse.rescale" and
     there are no interactors named "GoCircle.rescale".
     If you create a sensor named "rescale" as a child of a GoCircle,
     the sensor will not find any interaction named "GoCircle.rescale".
     It will then try with the name of the superclass. actually, it finds
     an interaction named "GoEllipse.rescale". Afterwards, each time
     the sensor would have to handle an event, it will send it to
     this interactor.

  Of course, each time the sensor's name is changed or the sensor
  is reparented, the sensor looks for a new corresponding interactor.
  This is also done when a sensor gets a differant GoInterf.

  The basic GoSensor class implements a light weight Go. GoSensor is a direct
  subclass of Go. It cannot have children, it has always the coordinate system
  of its parent, it cannot make drawings.

  Sensors crossing type is CrossAsNothing. To access them you must
  use tree crossing methods which start with an upper case (Child(),
  Next(), ... and not child(), next(), ...).

  If you create sensors, it is to start well defined interactions.
  This is why a sensor destroys itself when the interaction is finished.
  
COPYRIGHT :    
  Copyright (c) 1992 Bull S.A.
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull S.A. may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull S.A..
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull S.A. under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull S.A..

  Bull S.A. assumes no responsibility for the use or inability to use
  any of its software. The software is provided "AS IS" without warranty
  of any kind, and Bull S.A. expressly disclaims all implied warranties,
  including but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull S.A. be liable for any special, incidental,
  indirect or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/
#ifndef _Sensor_h
#define _Sensor_h

#include <Go/Event.h>
#include <Go/Go.h>
#include <Go/mGo.h>


class GoSensorMetaclass;
class GInteractor;



/*
 *----------------------
 * The GoSensor class
 *----------------------
 */

class GoSensor : public Go
{
  public:

    /* specific public methods: */
	
    static GoSensor* create
	(const char* program, Go* parent, const GEvent* start_event = NULL,
	 void (*)(void* cl, GoSensor*) = NULL, void* closure = NULL);

    const char* program () const { return _program; }
    GBool program (const char* new_program);

    void postcond (void(*p)(void*, GoSensor*), void* cl) { _postcond = p; _closure = cl; }

    GBool careAbout (const GEvent*) const;
    GBool handle (const GEvent*);

    GInteractor* interactor () { return _interactor; }

    void activate ();
    void disactivate ();
    GBool active () const { return _active; }
	

    static void bindActions (const GoMetaclass*);


    /* mandatory static methods: */

    static GoSensorMetaclass* metaclass();
    static GoSensor* create();

    static int save (const GoSensor*, GoStream*);
    static int restore (GoSensor**, GoStream*);   // creates + restores;
    static int readParams (GoSensor*, GoStream*); // restores only;


    /* pre- and post- conditions: */

    void programPrecond
	(GBool (*function)(void*, const Go*, char newProgram[256]),
	 void* closure, GAddOrRemove = GADD);
    void programPostcond
	(void (*function)(void*, Go*, const char* oldProgram),
	 void* closure, GAddOrRemove = GADD);

  protected:

    /* specific protected methods */

    GoSensor ();

    /* mandatory protected virtual methods: */

    virtual ~GoSensor ();
    virtual GBool m_contains (const GPoint&, GNumber = 0) const;
    virtual void m_dup (const Go*, Go*, Go*, GBool, GBool);
    virtual GBool is_sensor () const;

    /* virtual data management */

    virtual GAssoc*	d_get_assoc () const;

    virtual Go*		d_get_parent () const;
    virtual Go*		d_get_previous () const;
    virtual Go*		d_get_next () const;
    virtual Go*		d_get_first_child () const;
    virtual GBool	d_store_children_array () const;

    virtual void	d_set_next (const Go*);
    virtual void	d_set_previous (const Go*);
    virtual void	d_remove_child (const Go*, int child_index);
    virtual void 	d_add_child (const Go*, int child_index);
    virtual void	d_set_tree (const Go* parent,
				    const Go* prev, const Go* next);

    virtual GBool	d_store_abscs () const;
    virtual GBool	d_store_relcs () const;
    virtual const GCoordSys& d_get_relcs (const GCoordSys* parabs=NULL) const;

    virtual void 	d_set_coordsys (const GCoordSys& relcs,
					const GCoordSys& abscs,
					const GCoordSys& parabs);

    virtual GRegion*	d_get_clip () const;

  private:

    /* specific private methods: */

    void remove_interactor ();
    void create_interactor ();
    
    GBool call_program_precond (const char[256]);
    GBool call_class_program_precond (const char[256]);
    void call_program_postcond (const char*);
    void call_class_program_postcond (const char*);

    /* mandatory private virtual methods */

    virtual GoMetaclass* m_metaclass() const;

    /* friends: */

    friend class GoSensorMetaclass;
    friend void parent_changed (void*, Go*, Go*, Go*);
    friend void sensor_destroyed (void*, const Go*);

  protected:
    /* protected data */
    const GEvent* start_event;
    void (*_postcond)(void*, GoSensor*);
    void* _closure;

    /* the compatibility with older GoStreams forces us to
       make these two fields protected (and not private) */
    const char* _program;
    GBool _active;

  private:
    /* private data */
    GAssoc _assoc;
    Go* _parent;
    Go* _next;

  private:
    GInteractor* _interactor;
};



/*
 *-------------------------------
 * The GoSensorMetaclass class
 *-------------------------------
 */

class GoSensorMetaclass : public GoMetaclass
{
  public:
    /* pre- and post- conditions: */

    void programPrecond
	(GBool (*function)(void*, const Go*, char newProgram[256]),
	 void* closure, GAddOrRemove = GADD);
    void programPostcond
	(void (*function)(void*, Go*, const char* oldProgram),
	 void* closure, GAddOrRemove = GADD);

  protected:

             GoSensorMetaclass ();  //to get the GoSensorMetaclass use:
    virtual ~GoSensorMetaclass ();  //        GoSensor::metaclass()

  private:

    /* mandatory private virtual methods: */

    virtual const char*       m_classname () const;
    virtual GoMetaclass*      m_superclass () const;
    virtual const struct GActionRec* m_actions () const;
    virtual Go*               m_create () const;

    /* friends */

    friend class GoSensor;

    /* private data */
};



/*
 *----------------
 * inline methods
 *----------------
 */


/* This static method must be used to create a new GoSensor. */
inline GoSensor* GoSensor::create ()
{
    return (GoSensor*) Go::Create(GoSensor::metaclass());
}

#endif /* _Sensor_h */
