/* $Id: GeomUtil.h,v 1.1.1.1 1992/03/18 09:55:57 uep Exp $ */
/*****************************************************************************
MODULE:  Tools
FILE:    GeomUtil.h
AUTHOR:  Gilles Benati
DATE:    October 18th, 1991

DESCRIPTION:

  Contains the definition of a Tool object, Line, which represent a
  mathematical line (no output, equation y = ax + b).

  Also contains some usefull functions dealing with
	o  rectangles and polylines
	o  ellipse and elliptical arcs
        o  bezier and cubic curves (approximation with a polyline)

  (approximation with a polyline, clips, bounding boxes... )


  An ellipse has an implicit coordinate system :
    o  origin = ellipse center
    o  axises = ellipse symetry axises
  In their implicit coordinate system, ellipses are defined by 
  two radii radX, radY :
    P(x, y) in ellipse(radX, radY) <==>  (x/radX)^2 + (y/radY)^2 = 1


COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#ifndef _GeomUtil_h
#define _GeomUtil_h

#include <Go/machdef.h>
#include <Go/Tools/GCdef.h>
#include <Go/Tools/CoordSys.h>
#include <Go/Tools/Region.h>


enum { GNORMAL_LINE, GY_LINE, GPOINT_LINE };

class GLine {
  private:
    char type;
    GNumber a, b;
    
  public:
    GLine(const GPoint&, const GPoint&);
    GBool setX(GPoint&);
    GBool setY(GPoint&);
};

GBool GIntersect(const GPoint&, const GPoint&,
		 const GPoint&, const GPoint&, GBool& RtoL);

/***
 * POLYLINE and POLYGON
 ***/

extern void GPolypointGetOutPoints(int nb, const GPoint *pts,
				   GPoint& pmin, GPoint& pmax);

extern void GPolypointGetAbsOutPoints(const GCoordSys &cs,
				      int nb, const GPoint *pts,
				      GPoint& pmin, GPoint& pmax);

extern GBool GPointInPolygon(int nb, const GPoint* pts, const GPoint& point,
			     GBool fill_winding);

extern GBool GPointOnPolyline(const GCoordSys &cs, int nb, const GPoint* pts,
			      const GPoint& absPoint, GNumber maxSquare,
			      GBool closed);

extern GBox GPolypointJoinMiterBox(const GCoordSys &cs, int nb,
				   const GPoint *pts, GNumber absWidth,
				   GBool closed);

/***
 * RECTANGLE
 ***/

extern void GRectangleClip(const GPoint& p1, const GPoint& p3, GBool filled,
			   GNumber maxDist, GRegion& region);

/***
 * ELLIPSE
 ***/

#define GMAX_ELLIPSE_POINTS 81

extern int GEllipseNbLines(const GCoordSys& cs, GNumber radiusX,
			   GNumber radiusY, GNumber &angle1,
			   GNumber &angle2, GQuality);

extern int GEllipseArcLimits(const GCoordSys& cs, GNumber radX, GNumber radY,
			     GNumber angle1, GNumber angle2,
			     GPoint &start, GPoint &end,
			     GPoint &near_start, GPoint &near_end,
			     GQuality quality);

extern void GEllipseComputePoints(const GCoordSys& cs,
				  GNumber radiusX, GNumber radiusY,
				  GNumber angle1, GNumber angle2,
				  GQuality quality, int& npoint,
				  GPoint* points);

extern GNumber GModulo(GNumber x, GNumber y = 2*PI);

extern void GOrientedAngles(GNumber, GNumber, GNumber*, GNumber*);

extern GNumber GEllipseGetAngle(GNumber radX, GNumber radY, GPoint&);

extern GPoint GEllipseGetTangentPoint(GNumber radX, GNumber radY,
				      const GVector& v);

extern void GEllipseGetOutPoints(GNumber radX, GNumber radY,
				 GNumber angle1, GNumber angle2, 
				 GPoint& pmin, GPoint& pmax);

extern GBox GEllipseGetBox(const GCoordSys& cs, GNumber radX, GNumber radY,
			   GNumber angle1, GNumber angle2);


/***
 * BEZIER
 ***/

#define GMAX_BEZIER_INTERVAL 40
#define GMAX_BEZIER_POINTS(nctrl) ((nctrl) * (GMAX_BEZIER_INTERVAL+1))

extern void GBezierComputePoints(int nctrl, const GPoint* ctrl,
				 GQuality quality, int& npoint,
				 GPoint* points);


/***
 * CUBIC CURVES
 ***/

#define GMAX_CUBIC_INTERVAL 20
#define GMAX_CUBIC_POINTS(nctrl, evalshift)\
    (((nctrl) - 4 + (evalshift))/(evalshift) * GMAX_CUBIC_INTERVAL + 1)

extern void GCubicCurveComputePoints(int nctrl, const GPoint* ctrl,
				     const GNumber (*basisMatrix)[4] /*[4][4]*/,
				     int evalShift, GQuality quality,
				     int& npoint, GPoint* points);

#endif /* _GeomUtil_h */


