/* $Id: ShBrush.h,v 1.4 1992/09/25 11:51:12 smadja Exp $ */
/*****************************************************************************
MODULE:  Tools/GC
FILE:    ShBrush.h
AUTHOR:  Jacques Davy
DATE:    October 1990

DESCRIPTION:
    This file describes the ShBrush shared object.
    The ShBrush object is used by the GoInterf objects when drawing filled
    objects (filled rectangles, ...). The BrushValues structure may be
    used to set ShBrush parameters. Unset parameters have default values
    (see below).

    ShBrush objects are shared objects (SharedObject subclass). See
    also Brush objects.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  license from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#ifndef _ShBrush_h
#define _ShBrush_h

#include <Go/machdef.h>
#include <Go/Tools/SharedObj.h>
#include <Go/Tools/Number.h>

/* Mask definitions for GSharedObject operations */
#include <Go/Tools/GCdef.h>

/* Value Structure for GSharedObject operations (with defaults) */
struct GBrushValues {
    ulong function;             // logical operation (Gcopy)
    ulong plane_mask;           // plane mask        (GallPlanes)
    class GShColor*  color;      // main color        (black)
    class GShColor*  alt_color;  // alternate color   (none)

    /* fill attributes */
    struct {
       unsigned style : 2;      // fill style (GfillSolid)
       unsigned rule : 1;       // fill rule  (GevenOddRule)
    } fill;
    class GShPattern* tile;      // tile pattern for tiling (none)
    class GShPattern* stipple;   // stipple 1 plane pattern for stippling (none)
    GNumber   ts_x_origin;       // x offset for tile/stipple operations (0.0)
    GNumber   ts_y_origin;       // y offset for tile/stipple operations (0.0)
};

#define GMODIFY_BRUSHVALUE(brush, maskBit, field, newValue)\
    {\
        GBrushValues values;\
        values.field = newValue;\
        GModify(brush, maskBit, &values);\
    }

#define GUPDATE_BRUSHVALUE(name, maskBit, field, newValue)\
    {\
        GBrushValues values;\
	values.field = newValue;\
        GModify (name, maskBit, &values);\
    }

/* GShBrush class definition */
class GShBrush : public GSharedObject
{
    friend class GBrush;
    
  public:
    
    friend GShBrush* GCreateShBrush ();
    friend GShBrush* GCreateShBrush (ulong mask, const GBrushValues* values,
				     const char* name = NULL);
    friend GShBrush* GCreateShBrush (const char* name);
    
    static int save(const GShBrush*, GoStream*);
    static int restore(GShBrush**, GoStream*);
    static const char* classname();
    static int readParams(GShBrush*, GoStream*);

    virtual const char* getClassname() const;
    
    inline GShBrush* dup() const;
    inline GBool match (const GShBrush* brush) const;
    inline void updatePostcond (void (*)(void*, GShBrush*, ulong),
				const void* closure, GAddOrRemove = GADD) const;
    inline void destroyPostcond (void (*)(void*, const GShBrush*),
				 const void* closure, GAddOrRemove = GADD) const;
    
    // Getting values;
    inline const GBrushValues* values() const { return &_values; }
    
    inline ulong function () const { return _values.function; }
    inline ulong planeMask () const { return _values.plane_mask; }
    inline const class GShColor* color () const { return _values.color; }
    inline const class GShColor* altColor () const { return _values.alt_color; }
    
    inline ushort fillStyle () const { return _values.fill.style; }
    inline ushort fillRule () const { return _values.fill.rule; }
    inline const class GShPattern* tile () const { return _values.tile; }
    inline const class GShPattern* stipple () const { return _values.stipple; }
    inline GNumber xtsOrigin () const { return _values.ts_x_origin; }
    inline GNumber ytsOrigin () const { return _values.ts_y_origin; }
    
    // Modifying methods;
    friend void GModify (GShBrush*& brush, ulong mask,
			 const GBrushValues* values);
    friend void GRename (GShBrush*& brush, const char* name);
    
    friend void GSetFunction (GShBrush*& brush, ulong function);
    friend void GSetPlaneMask (GShBrush*& brush, ulong plane_mask);
    friend void GSetColor (GShBrush*& brush, const class GShColor* color);
    friend void GSetAltColor (GShBrush*& brush,
			      const class GShColor* alt_color);
    
    friend void GSetFillStyle (GShBrush*& brush, ushort fill_style);
    friend void GSetFillRule (GShBrush*& brush, ushort fill_rule);
    friend void GSetTile (GShBrush*& brush, const class GShPattern* tile);
    friend void GSetStipple (GShBrush*& brush,
			     const class GShPattern* stipple);
    friend void GSetTSOrigin (GShBrush*& brush, GNumber x_origin,
			      GNumber y_origin);
    
    // updating named GShBrush;
    friend void GModify (const char* name, ulong mask,
			 const GBrushValues* values);
    
    friend void GBrushFunction (const char* name, ulong function);
    friend void GBrushPlaneMask (const char* name, ulong plane_mask);
    friend void GBrushColor (const char* name,
			     const class GShColor* color);
    friend void GBrushAltColor (const char* name,
				const class GShColor* alt_color);
    
    friend void GBrushFillStyle (const char* name, ushort fill_style);
    friend void GBrushFillRule (const char* name, ushort fill_rule);
    friend void GBrushTile (const char* name,
			    const class GShPattern* tile);
    friend void GBrushStipple (const char* name,
			       const class GShPattern* stipple);
    friend void GBrushTsOrigin (const char* name,
				GNumber x_origin, GNumber y_origin);
    
    // friend for callbacks;
    friend void ColorIsUpdated(GShBrush* This,
			       const GShColor* color_or_alt_color,
			       ulong mask);
    friend void PatternIsUpdated(GShBrush* This,
				 const GShPattern* tile_or_stipple,
				 ulong mask);
    
  protected:
    
    inline GList*& _head () { return _headOfList; }
    inline int  _valueSize () { return (sizeof(GBrushValues)); }
    virtual GBool _match (ulong mask, const void* values) const ;
    virtual void _makeNewValues (ulong mask, const void* values,
				 ulong& newMask, void* newValues) const;
    virtual ulong _update (ulong mask, const void* values);
    friend GShBrush* _newGShBrush (GFunctionList* funcList, ulong mask,
				   const void* values,
				   const char* name = NULL);
    
    GShBrush (GFunctionList* funcList, ulong mask, const GBrushValues* values,
	      const char* name = NULL);
    virtual ~GShBrush ();
    
    
  private:
    
    GBrushValues _values;
    static GList* _headOfList;
    static GFunctionList _funcList;
};

/*
 * Inline Methods
 */
inline void GModify (GShBrush*& brush, ulong mask, const GBrushValues* values)
    { brush->modify((GSharedObject*&)brush, mask, (const void*)values); }

inline void GRename (GShBrush*& brush, const char* name)
    { brush->rename ((GSharedObject*&)brush, name); }

inline void GSetFunction (GShBrush*& brush, ulong function)
    { GMODIFY_BRUSHVALUE(brush, GFUNCTION, function, function); }

inline void GSetPlaneMask (GShBrush*& brush, ulong plane_mask)
    { GMODIFY_BRUSHVALUE(brush, GPLANE_MASK, plane_mask, plane_mask); }

inline void GSetColor (GShBrush*& brush, const class GShColor* color)
    { GMODIFY_BRUSHVALUE(brush, GCOLOR, color, (class GShColor*)color); }

inline void GSetAltColor (GShBrush*& brush, const class GShColor* alt_color)
    { GMODIFY_BRUSHVALUE(brush, GALT_COLOR, alt_color, (class GShColor*)alt_color); }

inline void GSetFillStyle (GShBrush*& brush, ushort fill_style)
    { GMODIFY_BRUSHVALUE(brush, GFILL_STYLE, fill.style, fill_style); }

inline void GSetFillRule (GShBrush*& brush, ushort fill_rule)
    { GMODIFY_BRUSHVALUE(brush, GFILL_RULE, fill.rule, fill_rule); }

inline void GSetTile (GShBrush*& brush, const class GShPattern* tile)
    { GMODIFY_BRUSHVALUE(brush, GTILE, tile, (class GShPattern*)tile); }

inline void GSetStipple (GShBrush*& brush, const class GShPattern* stipple)
    { GMODIFY_BRUSHVALUE(brush, GSTIPPLE, stipple, (class GShPattern*)stipple); }

inline void GSetTSOrigin (GShBrush*& brush, GNumber x_origin, GNumber y_origin)
{
    GBrushValues values;
    values.ts_x_origin = x_origin;
    values.ts_y_origin = y_origin;
    GModify (brush, GTILE_STIP_XORIGIN | GTILE_STIP_YORIGIN, &values);    
}

/**/
inline void GBrushFunction (const char* name, ulong function)
    { GUPDATE_BRUSHVALUE (name, GFUNCTION, function, function); }

inline void GBrushPlaneMask (const char* name, ulong plane_mask)
    { GUPDATE_BRUSHVALUE (name, GPLANE_MASK, plane_mask, plane_mask); }

inline void GBrushColor (const char* name, const class GShColor* color)
    { GUPDATE_BRUSHVALUE (name, GCOLOR, color, (class GShColor*)color); }

inline void GBrushAltColor (const char* name, const class GShColor* alt_color)
    { GUPDATE_BRUSHVALUE (name, GALT_COLOR, alt_color, (class GShColor*)alt_color); }

inline void GBrushFillStyle (const char* name, ushort fill_style)
    { GUPDATE_BRUSHVALUE (name, GFILL_STYLE, fill.style, fill_style); }

inline void GBrushFillRule (const char* name, ushort fill_rule)
    { GUPDATE_BRUSHVALUE (name, GFILL_RULE, fill.rule, fill_rule); }

inline void GBrushTile (const char* name, const class GShPattern* tile)
    { GUPDATE_BRUSHVALUE (name, GTILE, tile, (class GShPattern*)tile); }

inline void GBrushStipple (const char* name, const class GShPattern* stipple)
    { GUPDATE_BRUSHVALUE (name, GSTIPPLE, stipple, (class GShPattern*)stipple); }

inline void GBrushTsOrigin (const char* name, GNumber x_origin, GNumber y_origin)
{
    GBrushValues values;
    values.ts_x_origin = x_origin;
    values.ts_y_origin = y_origin;
    GCreateShBrush (GTILE_STIP_XORIGIN | GTILE_STIP_YORIGIN, &values, name);    
}

/**/
inline GShBrush* GShBrush::dup() const
    { return (GShBrush*)((GSharedObject*)this)->dup(); }

/**/
inline GBool GShBrush::match (const GShBrush* brush) const
    { return ((GShBrush*)this)->_match (brush->mask(), brush->values()); }

/**/
inline void GShBrush::updatePostcond (void cb(void*, GShBrush*, ulong),
				     const void* closure, GAddOrRemove add) const
{
    GSharedObject::updatePostcond (
                  (void(*)(void*, GSharedObject*, ulong))cb, closure, add);
}

/**/
inline void GShBrush::destroyPostcond (void cb(void*, const GShBrush*),
				      const void* closure, GAddOrRemove add) const
{
    GSharedObject::destroyPostcond (
		          (void(*)(void*, const GSharedObject*))cb, closure, add);
}

#undef GMODIFY_BRUSHVALUE
#undef GUPDATE_BRUSHVALUE

#endif /* _ShBrush_h */
