/* $Id: ShColormap.h,v 1.4 1992/09/25 11:51:30 smadja Exp $ */
/*****************************************************************************
MODULE:  Tools
FILE:    ShColormap.h
AUTHOR:  Jacques Davy
DATE:    13 June 1990

DESCRIPTION:
    The ShColormap object is a shared object containing a color table.
    It is used when drawing images or patterns by the GoInterface. The
    result of the drawing depends on the GoInterface capabilities.

    A set of functions for modifying the color map (add, remove, replace
    colors). These functions are "just" convenient functions for the
    modify method.

    ShColormap are Shared Objects, see also Colormap objects.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#ifndef _ShColormap_h
#define _ShColormap_h

#include <Go/Tools/SharedObj.h>
#include <Go/Tools/ShColor.h>

/* Mask definitions for GSharedObject operations */
#define GCOLORMAP_COLORS        (1L<<0)

/* Value structure for GSharedObject operations */
struct GColormapValues
{
    ulong n_color;
    GShColor** colors;
};

/*
 * GShColormap new condition flags
 */
struct GShColormapCondFlags
{
    GBool _has_addColor_postcond : 1;
    GBool _has_remColor_postcond : 1;
    GBool _has_replaceColor_postcond : 1;

    GShColormapCondFlags()
    {
	_has_addColor_postcond  = GFALSE;
	_has_remColor_postcond  = GFALSE;
	_has_replaceColor_postcond  = GFALSE;
    }
};

/* GShColormap class definition */
class GShColormap : public GSharedObject
{
    friend class GColormap;

  public:

    friend void GModify (GShColormap*& sho, ulong mask,
			 const GColormapValues* values);
    friend void GModify (const char* name, ulong mask,
			 const GColormapValues* values);
    friend void GRename (GShColormap*& sho, const char* name);

    static int save(const GShColormap*, GoStream*);
    static int restore(GShColormap**, GoStream*);
    static const char* classname();
    static int readParams(GShColormap*, GoStream*);

    virtual const char* getClassname() const;

    inline GShColormap* dup () const;
    inline GBool match (const GShColormap* sho) const;
    inline void updatePostcond (void (*)(void*, GShColormap*, ulong),
				const void* closure, GAddOrRemove = GADD) const;
    inline void incrRefPostcond (void (*)(void*, GShColormap*),
				 const void* closure, GAddOrRemove = GADD) const;
    inline void decrRefPostcond (void (*)(void*, GShColormap*),
				 const void* closure, GAddOrRemove = GADD) const;
    inline void destroyPostcond (void (*)(void*, const GShColormap*),
				 const void* closure, GAddOrRemove = GADD) const;
    void addColorPostcond (void (*)(void*, GShColormap*, ulong),
			   const void* closure, GAddOrRemove = GADD) const;
    void remColorPostcond (void (*)(void*, GShColormap*, ulong),
			   const void* closure, GAddOrRemove = GADD) const;
    void replaceColorPostcond (void (*)(void*, GShColormap*, ulong),
			       const void* closure, GAddOrRemove = GADD) const;

    friend GShColormap* GCreateShColormap();
   
    friend GShColormap* GCreateShColormap (ulong mask, const GColormapValues* values,
					   const char* name = NULL);
    friend GShColormap* GCreateShColormap (const char* name);

    inline const GColormapValues* values () const { return &_values; }
    inline ulong nColor() const { return _values.n_color; }
    inline const GShColor** colors () const { return (const GShColor**) _values.colors; }

    // Modification methods;
    ulong addColor (GShColormap*& , const GShColor* color); // returns index;
    void remColor (GShColormap*& , const GShColor* color);
    void remColor (GShColormap*& , ulong index);
    ulong replaceColor (GShColormap*& , const GShColor* old_color,
			const GShColor* new_color);
    ulong replaceColor (GShColormap*& , ulong index,
			const GShColor* new_color);

    /* Getting color object, returns NULL if not found */
    const GShColor* getColor (const char* name) const;
    const GShColor* getColor (const GRgb* rgb) const;
    const GShColor* getColor (uchar gray) const;
    const GShColor* getColor (const GShPattern* halftone) const; 
    const GShColor* getColor (ulong index) const;

    /* Getting index, returns -1 if not found */
    long getIndex (const char* name) const;
    long getIndex (const GRgb* rgb) const;
    long getIndex (uchar gray) const;
    long getIndex (const GShPattern* halftone) const;
    long getIndex (const GShColor* color) const;

    /**
      friend
      **/
    friend void   ColorIsUpdated(GShColormap* This, const GShColor* color,
				 ulong mask);

  protected:

    inline GList*& _head () { return _headOfList; }
    inline int _valueSize () { return (sizeof(GColormapValues)); }
    virtual GBool _match (ulong mask, const void* values) const;
    virtual void _makeNewValues (ulong mask, const void* values,
				 ulong& newMask, void* newValues) const;
    virtual ulong _update (ulong mask, const void* values);
    friend GShColormap* _newGShColormap (GFunctionList* funcList, ulong mask,
					 const void* values,
					 const char* name = NULL);

    GShColormap (GFunctionList* funcList, ulong mask, const GColormapValues* values,
		 const char* name = NULL);
    virtual ~GShColormap ();


  private:

    ulong _array_size;
    GColormapValues _values;
    static GList* _headOfList;
    static GFunctionList _funcList;

    ulong _arraySize (const GColormapValues* values) const;
    void _resizeArray (ulong new_size);
    const GShColor* _getColor(long index) const;
    void _makeColormapValues (ulong arra_size, GColormapValues& values);

    GShColormapCondFlags _colormap_flags;

    void _call_addColor_postcond (ulong index);
    void _call_remColor_postcond (ulong index);
    void _call_replaceColor_postcond (ulong index);
};


/*
 * Inline Methods
 */
inline void GModify (GShColormap*& sho, ulong mask, const GColormapValues* values)
{
    sho->modify((GSharedObject*&)sho, mask, (const void*)values);
}

/**/
inline void GRename (GShColormap*& sho, const char* name)
{
   sho->rename((GSharedObject*&)sho, name);
}

/**/
inline GShColormap* GShColormap::dup () const
{
    return (GShColormap*)((GSharedObject*)this)->dup();
}

/**/
inline GBool GShColormap::match (const GShColormap* sho) const
{
    return ((GShColormap*)this)->_match (sho->mask(), sho->values());
}

/**/
inline void GShColormap::updatePostcond (void f(void*, GShColormap*, ulong),
					const void* closure, GAddOrRemove add) const
{
    GSharedObject::updatePostcond (
   (void(*)(void*, GSharedObject*, ulong))f, closure, add);
}

/**/
inline void GShColormap::incrRefPostcond (void f(void*, GShColormap*),
					 const void* closure, GAddOrRemove add) const
{
    GSharedObject::incrRefPostcond (
		          (void(*)(void*, GSharedObject*))f, closure, add);
}

/**/
inline void GShColormap::decrRefPostcond (void f(void*, GShColormap*),
					 const void* closure, GAddOrRemove add) const
{
    GSharedObject::decrRefPostcond (
                         (void(*)(void*, GSharedObject*))f, closure, add);
}

inline void GShColormap::destroyPostcond (void f(void*, const GShColormap*),
					 const void* closure, GAddOrRemove add) const
{
    GSharedObject::destroyPostcond ((void(*)(void*, const GSharedObject*))f, closure, add);
}


#endif /* _ShColormap_h */

