/* $Id: ShDash.h,v 1.5 1992/12/11 14:20:49 smadja Exp $ */
/*****************************************************************************
MODULE:  Tools
FILE:    ShDash.h
AUTHOR:  Jacques Davy
DATE:    October 1990

DESCRIPTION:
    This file describes the ShDash object. This object is normally used
    to draw lines.

    A ShDash is defined by:
       * A list of element, each element is a number representing the
         size in millimeters of dashes and gaps. The initial (0) and the
	 alternating elements (2, 4, ...) are the even dashes/gaps, the
	 others are the odd dashes/gaps. When the line style is
	 LineOnOffDash, the odd dashes are used as gaps. When the line style
	 is LineDoubleDash, both dashes are used.

       * A dash offset pointing on the first (even) element.

       * A dash number specifying the dash list size.

    A ShDash is a Shared Object, see also Dash objects.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#ifndef _ShDash_h
#define _ShDash_h

#include <Go/Tools/SharedObj.h>
#include <Go/Tools/Number.h>

/* Mask definitions for GSharedObject operations */
#define GDASH_LIST            (1L<<0)
#define GDASH_OFFSET          (1L<<1)

/* Value structure for GSharedObject operations */
struct GDashValues {
    ushort dash_number;
    GNumber dash_offset;
    GNumber* dash_list;
};

/* GShDash class definition */
class GShDash : public GSharedObject
{
    friend class GDash;

  public:

    friend void GModify (GShDash*& sho, ulong mask,
			 const GDashValues* values);
    friend void GModify (const char* name, ulong mask,
			 const GDashValues* values);
    friend void GRename (GShDash*& sho, const char* name);

    static int save(const GShDash*, GoStream*);
    static int restore(GShDash**, GoStream*);
    static const char* classname();
    static int readParams(GShDash*, GoStream*);

    virtual const char* getClassname() const;

    inline GShDash* dup () const;
    inline GBool match (const GShDash* sho) const;
    inline void updatePostcond (void (*)(void*, GShDash*, ulong),
				const void* closure, GAddOrRemove = GADD) const;
    inline void incrRefPostcond (void (*)(void*, GShDash*),
				 const void* closure, GAddOrRemove = GADD) const;
    inline void decrRefPostcond (void (*)(void*, GShDash*),
				 const void* closure, GAddOrRemove = GADD) const;
    inline void destroyPostcond (void (*)(void*, const GShDash*),
				 const void* closure, GAddOrRemove = GADD) const;

    friend GShDash* GCreateShDash ();
    friend GShDash* GCreateShDash (ulong mask, const GDashValues* values,
				   const char* name = NULL);
    friend GShDash* GCreateShDash (const char* name);

    inline const GDashValues* values () const { return &_values; }
    inline GNumber dash_offset () const { return _values.dash_offset; }
    inline ushort dash_number () const { return _values.dash_number; }
    inline const GNumber* dash_list () const { return _values.dash_list; }

  protected:

    inline GList*& _head () { return _headOfList; }
    inline int _valueSize () { return (sizeof(GDashValues)); }
    virtual GBool _match (ulong mask, const void* values) const;
    virtual void _makeNewValues (ulong mask, const void* values,
				 ulong& newMask, void* newValues) const;
    virtual ulong _update (ulong mask, const void* values);
    friend GShDash* _newGShDash (GFunctionList* funcList, ulong mask,
				 const void* values,
				 const char* name = NULL);

    GShDash (GFunctionList* funcList, ulong mask, const GDashValues* values,
	     const char* name = NULL);
    virtual ~GShDash ();


  private:

    GDashValues _values;
    static GList* _headOfList;
    static GFunctionList _funcList;
};

/*
 * Inline Methods
 */
inline void GModify (GShDash*& sho, ulong mask, const GDashValues* values)
{
    sho->modify((GSharedObject*&)sho, mask, (const void*)values);
}

/**/
inline void GRename (GShDash*& sho, const char* name)
{
   sho->rename((GSharedObject*&)sho, name);
}

/**/
inline GShDash* GShDash::dup () const
{
    return (GShDash*)((GSharedObject*)this)->dup();
}

/**/
inline GBool GShDash::match (const GShDash* sho) const
{
    return ((GShDash*)this)->_match (sho->mask(), sho->values());
}

/**/
inline void GShDash::updatePostcond (void cb(void*, GShDash*, ulong),
				    const void* closure, GAddOrRemove add) const
{
    GSharedObject::updatePostcond (
   (void(*)(void*, GSharedObject*, ulong))cb, closure, add);
}

/**/
inline void GShDash::incrRefPostcond (void cb(void*, GShDash*),
				     const void* closure, GAddOrRemove add) const
{
    GSharedObject::incrRefPostcond (
		          (void(*)(void*, GSharedObject*))cb, closure, add);
}

/**/
inline void GShDash::decrRefPostcond (void cb(void*, GShDash*),
				     const void* closure, GAddOrRemove add) const
{
    GSharedObject::decrRefPostcond (
                         (void(*)(void*, GSharedObject*))cb, closure, add);
}

/**/
inline void GShDash::destroyPostcond (void cb(void*, const GShDash*),
				     const void* closure, GAddOrRemove add) const
{
    GSharedObject::destroyPostcond (
		          (void(*)(void*, const GSharedObject*))cb, closure, add);
}

#endif /* _ShDash_h */
