/* $Id: ShPattern.h,v 1.4 1992/09/25 11:52:04 smadja Exp $ */
/*****************************************************************************
MODULE:  Tools
FILE:    ShPattern.h
AUTHOR:  Jacques Davy
DATE:    14 June 1990

DESCRIPTION:
    The GShPattern Object is usually used to fill Go. The GShPattern object
    points to a PixelArray (or BitArray) which is the value of the GShPattern.

    The GShPattern object may have its own size (independently of the the
    PixelArray size and expressed in mm), that means the PixelArray will be
    resized by the GoInterface at the drawing time. But the GShPattern may have an
    undefined size (=0), that means the GoInterface will use the PixelArray
    directly without any trandformation.
  
    The GShPattern object is a shared object. As most of defined GSharedObjects,
    the GShPattern object may have a name. See also Pattern objects.

COPYRIGHT :    
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  license from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#ifndef _ShPattern_h
#define _ShPattern_h

#include <Go/machdef.h>
#include <Go/Tools/SharedObj.h>
#include <Go/Tools/ShPixelArray.h>
#include <Go/Tools/ShColormap.h>

/* Mask definitions for GSharedObject operations */
#define GPTRN_PIXEL_ARRAY (1L<<0)
#define GPTRN_COLORMAP    (1L<<1)
#define GPTRN_WIDTH       (1L<<2)
#define GPTRN_HEIGHT      (1L<<3)

/* Value structure for GSharedObject operations */
struct GPatternValues
{
    GShPixelArray* pixelArray;
    GShColormap*   colormap;
    GNumber        width;    /* width in mm */
    GNumber        height;   /* height in mm */
};

/* GShPattern class definition */
class GShPattern : public GSharedObject 
{
    friend class GPattern;
	
  public:

    friend void GModify (GShPattern*& sho,
			 ulong mask, const GPatternValues* values);
    friend void GModify (const char* name, ulong mask,
			 const GPatternValues* values);
    friend void GRename (GShPattern*& sho, const char* name);

    static int save(const GShPattern*, GoStream*);
    static int restore(GShPattern**, GoStream*);
    static const char* classname();
    static int readParams(GShPattern*, GoStream*);

    virtual const char* getClassname() const;
    
    inline GShPattern* dup() const;
    inline GBool match (const GShPattern* sho) const;
    inline void updatePostcond (void (*)(void*, GShPattern*, ulong),
				const void* closure, GAddOrRemove = GADD) const;
    inline void incrRefPostcond (void (*)(void*, GShPattern*),
				 const void* closure, GAddOrRemove = GADD) const;
    inline void decrRefPostcond (void (*)(void*, GShPattern*),
				 const void* closure, GAddOrRemove = GADD) const;
    inline void destroyPostcond (void (*)(void*, const GShPattern*),
				 const void* closure, GAddOrRemove = GADD) const;

    friend GShPattern* GCreateShPattern();
    friend GShPattern* GCreateShPattern (const GShPixelArray* pixelArray,
					 const GShColormap* sh_cmap,
					 GNumber width, GNumber height,
					 const char* name);
    friend GShPattern* GCreateShPattern (ulong mask, const GPatternValues* values,
					 const char* name = NULL);
    friend GShPattern* GCreateShPattern (const char* name);

    inline const GPatternValues* values () const { return &_values; }
    inline const GShPixelArray* pixelArray () const
    {
	return _values.pixelArray;
    }
    inline const GShColormap* colormap () const { return _values.colormap; }
    inline GNumber width () const { return _values.width; }
    inline GNumber height () const { return _values.height; }

    // friend for callbacks;
    friend void ColormapIsUpdated(GShPattern* This,
				  const GShColormap* colormap,
				  ulong mask);
    friend void PixelArrayIsUpdated(GShPattern* This,
				    const GShPixelArray* pixelArray,
				    ulong mask);

  protected:

    inline GList*& _head () { return _headOfList; }
    inline int _valueSize () { return (sizeof(GPatternValues)); }
    virtual GBool _match (ulong mask, const void* values) const;
    virtual void _makeNewValues (ulong mask, const void* values,
				 ulong& newMask, void* newValues) const;
    virtual ulong _update (ulong mask, const void* values);
    friend GShPattern* _newGShPattern (GFunctionList* funcList, ulong mask,
				       const void* values,
				       const char* name = NULL);

    GShPattern (GFunctionList* funcList, ulong mask, const GPatternValues* values,
		const char* name = NULL);
    virtual ~GShPattern ();


  private:

    GPatternValues _values;
    static GList* _headOfList;
    static GFunctionList _funcList;
};

/*
 * Inline Methods
 */
inline void GModify (GShPattern*& sho, ulong mask,
		     const GPatternValues* values)
{
    sho->modify((GSharedObject*&)sho, mask, (const void*)values);
}

/**/
inline void GRename (GShPattern*& sho, const char* name)
{
    sho->rename((GSharedObject*&)sho, name);
}

/**/
inline GShPattern* GShPattern::dup() const
{
    return (GShPattern*)((GSharedObject*)this)->dup();
}

/**/
inline GBool GShPattern::match (const GShPattern* sho) const
{
    return _match (sho->mask(), sho->values());
}

/**/
inline void GShPattern::updatePostcond (void cb(void*, GShPattern*, ulong),
				       const void* closure, GAddOrRemove add) const
{
    GSharedObject::updatePostcond (
   (void(*)(void*, GSharedObject*, ulong))cb, closure, add);
}

/**/
inline void GShPattern::incrRefPostcond (void cb(void*, GShPattern*),
					const void* closure, GAddOrRemove add) const
{
    GSharedObject::incrRefPostcond (
		          (void(*)(void*, GSharedObject*))cb, closure, add);
}

/**/
inline void GShPattern::decrRefPostcond (void cb(void*, GShPattern*),
					const void* closure, GAddOrRemove add) const
{
    GSharedObject::decrRefPostcond (
                         (void(*)(void*, GSharedObject*))cb, closure, add);
}

/**/
inline void GShPattern::destroyPostcond (void cb(void*, const GShPattern*),
					const void* closure, GAddOrRemove add) const
{
    GSharedObject::destroyPostcond (
		          (void(*)(void*, const GSharedObject*))cb, closure, add);
}

#endif /* _ShPattern_h */
