/* $Id: ShPen.h,v 1.7 1992/12/11 14:26:35 smadja Exp $ */
/*****************************************************************************
MODULE:  Tools/GC
FILE:    ShPen.h
AUTHOR:  Jacques Davy
DATE:    October 1990

DESCRIPTION:
    This file describes the ShPen shared object.
    The ShPen object is used by the GoInterf objects when drawing thread
    objects (line, arc, ...). The PenValues structure may be used to
    set ShPen parameters. Unset parameters have default values (se below).

    Pen objects are shared objects (GSharedObject subclass). See also Pen
    objects

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  license from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#ifndef _ShPen_h
#define _ShPen_h

#include <Go/machdef.h>
#include <Go/Tools/SharedObj.h>
#include <Go/Tools/Number.h>


/* Mask definitions for GSharedObject operations */
#include <Go/Tools/GCdef.h>

/* Value Structure for GSharedObject operations (with defaults) */
struct GPenValues {
    ulong function;             // logical operation (Gcopy)
    ulong plane_mask;           // plane mask        (GallPlanes)
    class GShColor*  color;      // main color        (unamed black)
    class GShColor*  alt_color;  // alternate color   (none)

    /* line attributes */
    GNumber line_width;          // line width (0.3 mm)
    struct {
       unsigned style : 2;      // line style (GlineSolid)
       unsigned cap : 2;        // cap style  (GcapButt)
       unsigned join : 2;       // join style (GjoinMiter)
    } line;
    GNumber miter_limit;         // miter limit (none)
    class GShDash* dash;         // dash object (none)

    /* fill attributes */
    struct {
       unsigned style : 2;      // fill style (GfillSolid)
    } fill;
    class GShPattern* tile;      // tile pattern for tiling (none)
    class GShPattern* stipple;   // stipple 1 plane pattern for stippling (none)
    GNumber   ts_x_origin;       // x offset for tile/stipple operations (0.0)
    GNumber   ts_y_origin;       // y offset for tile/stipple operations (0.0)
};

#define GMODIFY_PENVALUE(pen, maskBit, field, newValue)\
    {\
        GPenValues values;\
        values.field = newValue;\
        GModify(pen, maskBit, &values);\
    }

#define GUPDATE_PENVALUE(name, maskBit, field, newValue)\
    {\
        GPenValues values;\
	values.field = newValue;\
        GModify (name, maskBit, &values);\
    }

/* GShPen class definition */
class GShPen : public GSharedObject
{
    friend class GPen;

  public:

    friend GShPen* GCreateShPen ();
    friend GShPen* GCreateShPen (ulong mask, const GPenValues* values,
				 const char* name = NULL);
    friend GShPen* GCreateShPen (const char* name);

    static int save(const GShPen*, GoStream*);
    static int restore(GShPen**, GoStream*);
    static const char* classname();
    static int readParams(GShPen*, GoStream*);

    virtual const char* getClassname() const;

    inline GShPen* dup() const;
    inline GBool match (const GShPen* pen) const;
    inline void updatePostcond (void (*)(void*, GShPen*, ulong),
				const void* closure, GAddOrRemove = GADD) const;
    inline void destroyPostcond (void (*)(void*, const GShPen*),
				 const void* closure, GAddOrRemove = GADD) const;

    // Getting values;
    inline const GPenValues* values() const { return &_values; }

    inline ulong function () const { return _values.function; }
    inline ulong planeMask () const { return _values.plane_mask; }
    inline const class GShColor* color () const { return _values.color; }
    inline const class GShColor* altColor () const { return _values.alt_color; }

    inline GNumber lineWidth () const { return _values.line_width; }
    inline ushort lineStyle () const { return _values.line.style; }
    inline ushort capStyle () const { return _values.line.cap; }
    inline ushort joinStyle () const { return _values.line.join; }
    inline GNumber miterLimit () const { return _values.miter_limit; }
    inline const class GShDash* dash () const { return _values.dash; }

    inline ushort fillStyle () const { return _values.fill.style; }
    inline const class GShPattern* tile () const { return _values.tile; }
    inline const class GShPattern* stipple () const { return _values.stipple; }
    inline GNumber xtsOrigin () const { return _values.ts_x_origin; }
    inline GNumber ytsOrigin () const { return _values.ts_y_origin; }

    // Modifying methods;
    friend void GModify (GShPen*& pen, ulong mask, const GPenValues* values);
    friend void GRename (GShPen*& pen, const char* name);

    friend void GSetFunction (GShPen*& pen, ulong function);
    friend void GSetPlaneMask (GShPen*& pen, ulong plane_mask);
    friend void GSetColor (GShPen*& pen, const class GShColor* color);
    friend void GSetAltColor (GShPen*& pen, const class GShColor* alt_color);

    friend void GSetLineWidth (GShPen*& pen, GNumber line_width);
    friend void GSetLineStype (GShPen*& pen, ushort line_style);
    friend void GSetCapStyle (GShPen*& pen, ushort cap_style);
    friend void GSetJoinStyle (GShPen*& pen, ushort join_style);
    friend void GSetMiterLimit (GShPen*& pen, GNumber miter_limit);
    friend void GSetDash (GShPen*& pen, const class GShDash* dash);

    friend void GSetFillStyle (GShPen*& pen, ushort fill_style);
    friend void GSetTile (GShPen*& pen, const class GShPattern* tile);
    friend void GSetStipple (GShPen*& pen, const class GShPattern* stipple);
    friend void GSetTSOrigin (GShPen*& pen, GNumber x_origin, GNumber y_origin);

    // updating named GShPen;
    friend void GModify (const char* name, ulong mask,
			 const GPenValues* values);

    friend void GPenFunction (const char* name, ulong function);
    friend void GPenPlaneMask (const char* name, ulong plane_mask);
    friend void GPenColor (const char* name, const class GShColor* color);
    friend void GPenAltColor (const char* name,
			      const class GShColor* alt_color);

    friend void GPenLineWidth (const char* name, GNumber line_width);
    friend void GPenLineStype (const char* name, ushort line_style);
    friend void GPenCapStyle (const char* name, ushort cap_style);
    friend void GPenJoinStyle (const char* name, ushort join_style);
    friend void GPenMiterLimit (const char* name, GNumber miter_limit);
    friend void GPenDash (const char* name, const class GShDash* dash);

    friend void GPenFillStyle (const char* name, ushort fill_style);
    friend void GPenTile (const char* name, const class GShPattern* tile);
    friend void GPenStipple (const char* name,
			     const class GShPattern* stipple);
    friend void GPenTsOrigin (const char* name, GNumber x_origin,
			      GNumber y_origin);
    // friend for callbacks;
    friend void   ColorIsUpdated(GShPen* This, const GShColor* color,
				 ulong mask);
    friend void   DashIsUpdated(GShPen* This, const GShDash* dash,
				ulong mask);
    friend void   PatternIsUpdated(GShPen* This, const GShPattern* pattern,
				   ulong mask);
    
  protected:

    inline GList*& _head () { return _headOfList; }
    inline int  _valueSize () { return (sizeof(GPenValues)); }
    virtual GBool _match (ulong mask, const void* values) const ;
    virtual void _makeNewValues (ulong mask, const void* values,
				 ulong& newMask, void* newValues) const;
    virtual ulong _update (ulong mask, const void* values);
    friend GShPen* _newGShPen (GFunctionList* funcList, ulong mask,
			       const void* values,
			       const char* name = NULL);

    GShPen (GFunctionList* funcList, ulong mask, const GPenValues* values,
	    const char* name = NULL);
    virtual ~GShPen ();


  private:

    GPenValues _values;
    static GList* _headOfList;
    static GFunctionList _funcList;
};

/*
 * Inline Methods
 */
inline void GModify (GShPen*& pen, ulong mask, const GPenValues* values)
    { pen->modify((GSharedObject*&)pen, mask, (const void*)values); }

inline void GRename (GShPen*& pen, const char* name)
    { pen->rename ((GSharedObject*&)pen, name); }

inline void GSetFunction (GShPen*& pen, ulong function)
    { GMODIFY_PENVALUE(pen, GFUNCTION, function, function); }

inline void GSetPlaneMask (GShPen*& pen, ulong plane_mask)
    { GMODIFY_PENVALUE(pen, GPLANE_MASK, plane_mask, plane_mask); }

inline void GSetColor (GShPen*& pen, const class GShColor* color)
    { GMODIFY_PENVALUE(pen, GCOLOR, color, (class GShColor*)color); }

inline void GSetAltColor (GShPen*& pen, const class GShColor* alt_color)
    { GMODIFY_PENVALUE(pen, GALT_COLOR, alt_color, (class GShColor*)alt_color); }

inline void GSetLineWidth (GShPen*& pen, GNumber line_width)
    { GMODIFY_PENVALUE(pen, GLINE_WIDTH, line_width, line_width); }

inline void GSetLineStyle (GShPen*& pen, ushort line_style)
    { GMODIFY_PENVALUE(pen, GLINE_STYLE, line.style, line_style); }

inline void GSetCapStyle (GShPen*& pen, ushort cap_style)
    { GMODIFY_PENVALUE(pen, GCAP_STYLE, line.cap, cap_style); }

inline void GSetJoinStyle (GShPen*& pen, ushort join_style)
    { GMODIFY_PENVALUE(pen, GJOIN_STYLE, line.join, join_style); }

inline void GSetMiterLimit (GShPen*& pen, GNumber miter_limit)
    { GMODIFY_PENVALUE(pen, GMITER_LIMIT, miter_limit, miter_limit); }

inline void GSetDash (GShPen*& pen, const class GShDash* dash)
    { GMODIFY_PENVALUE(pen, GDASH, dash, (class GShDash*)dash); }

inline void GSetFillStyle (GShPen*& pen, ushort fill_style)
    { GMODIFY_PENVALUE(pen, GFILL_STYLE, fill.style, fill_style); }

inline void GSetTile (GShPen*& pen, const class GShPattern* tile)
    { GMODIFY_PENVALUE(pen, GTILE, tile, (class GShPattern*)tile); }

inline void GSetStipple (GShPen*& pen, const class GShPattern* stipple)
    { GMODIFY_PENVALUE(pen, GSTIPPLE, stipple, (class GShPattern*)stipple); }

inline void GSetTSOrigin (GShPen*& pen, GNumber x_origin, GNumber y_origin)
{
    GPenValues values;
    values.ts_x_origin = x_origin;
    values.ts_y_origin = y_origin;
    GModify (pen, GTILE_STIP_XORIGIN | GTILE_STIP_YORIGIN, &values);    
}

/**/
inline void GPenFunction (const char* name, ulong function)
    { GUPDATE_PENVALUE (name, GFUNCTION, function, function); }

inline void GPenPlaneMask (const char* name, ulong plane_mask)
    { GUPDATE_PENVALUE (name, GPLANE_MASK, plane_mask, plane_mask); }

inline void GPenColor (const char* name, const class GShColor* color)
    { GUPDATE_PENVALUE (name, GCOLOR, color, (class GShColor*)color); }

inline void GPenAltColor (const char* name, const class GShColor* alt_color)
    { GUPDATE_PENVALUE (name, GALT_COLOR, alt_color, (class GShColor*)alt_color); }

inline void GPenLineWidth (const char* name, GNumber line_width)
    { GUPDATE_PENVALUE (name, GLINE_WIDTH, line_width, line_width); }

inline void GPenLineStype (const char* name, ushort line_style)
    { GUPDATE_PENVALUE (name, GLINE_STYLE, line.style, line_style); }

inline void GPenCapStyle (const char* name, ushort cap_style)
    { GUPDATE_PENVALUE (name, GCAP_STYLE, line.cap, cap_style); }

inline void GPenJoinStyle (const char* name, ushort join_style)
    { GUPDATE_PENVALUE (name, GJOIN_STYLE, line.join, join_style); }

inline void GPenMiterLimit (const char* name, GNumber miter_limit)
    { GUPDATE_PENVALUE (name, GMITER_LIMIT, miter_limit, miter_limit); }

inline void GPenDash (const char* name, const class GShDash* dash)
    { GUPDATE_PENVALUE (name, GDASH, dash, (class GShDash*)dash); }

inline void GPenFillStyle (const char* name, ushort fill_style)
    { GUPDATE_PENVALUE (name, GFILL_STYLE, fill.style, fill_style); }


inline void GPenTile (const char* name, const class GShPattern* tile)
    { GUPDATE_PENVALUE (name, GTILE, tile, (class GShPattern*)tile); }

inline void GPenStipple (const char* name, const class GShPattern* stipple)
    { GUPDATE_PENVALUE (name, GSTIPPLE, stipple, (class GShPattern*)stipple); }

inline void GPenTsOrigin (const char* name, GNumber x_origin, GNumber y_origin)
{
    GPenValues values;
    values.ts_x_origin = x_origin;
    values.ts_y_origin = y_origin;
    GCreateShPen (GTILE_STIP_XORIGIN | GTILE_STIP_YORIGIN, &values, name);    
}

/**/
inline GShPen* GShPen::dup() const
    { return (GShPen*)((GSharedObject*)this)->dup(); }

/**/
inline GBool GShPen::match (const GShPen* pen) const
    { return ((GShPen*)this)->_match (pen->mask(), pen->values()); }

/**/
inline void GShPen::updatePostcond (void cb(void*, GShPen*, ulong),
				 const void* closure, GAddOrRemove add) const
{
    GSharedObject::updatePostcond (
   (void(*)(void*, GSharedObject*, ulong))cb, closure, add);
}

/**/
inline void GShPen::destroyPostcond (void cb(void*, const GShPen*),
				  const void* closure, GAddOrRemove add) const
{
    GSharedObject::destroyPostcond (
		          (void(*)(void*, const GSharedObject*))cb, closure, add);
}

#undef GMODIFY_PENVALUE
#undef GUPDATE_PENVALUE

#endif /* _GShPen_h */
