/* $Id: SharedObj.h,v 1.6 1992/09/25 11:52:51 smadja Exp $ */
/*****************************************************************************
MODULE:  Shared Object Management.
FILE:    SharedObj.h
AUTHOR:  Jacques Davy, Philippe Smadja
DATE:    August 1990, (version 1.0)

DESCRIPTION:
    A SharedObject is an object which may be used simultaneously by more
    than one object. These objects are defined by a set of values. That means:

       o For the creation of a given subclass of SharedObject, the system
         tends to find a created instance matching with the desired one.
	 If it finds, the reference count is incremented and the found
	 instance is yielded. Else a new instance is created with the reference
	 count 1. The "new" C++ method cannot be called for SharedObject.

       o For the modification, the system tends to find an existing object
         matching with the modification. If it finds such an instance, it
	 decrements the current ref count and increments the found ref
	 count. (That means a modification may return a new object).

       o For the destruction, the system decrements the current ref-count.
         If it becomes null, the object is effectively destroyed. The
	 "delete" C++ cannot not be called.

    A SharedObject instance may be NAMED. That means modifying (modif())
    a named shared object will always update it. Naming a shared object
    means identifying a unique shared object. Names are managed by the
    SharedObject class.

    PostCond callback lists existent for the modification and the destruction.

    The SharedObject are organized in Lists (one list per subclass).
    Subclasses have to provide the SharedObject constructor with the ListHead.

    Writing a SharedObj subclass needs to follow a writing rule:

       o A list pointer (List*) called _headOfList must be declared
         as static of the subclass.
       o A function list (FunctionList) called _funcList must be
         declared as static of the subclass
       o A set of protected methods must written:
           List*& _head() : returns _headOfList reference
	   int  _valueSize(): returns the value structure size.
	   bool _match(): performs the match test between two (mask, values).
	   void _update(): update the values of the object
	   Constructor(FunctionList*, ulong, mask, name) :
	               SharedObject (FunctionList*, mask, name)
	   ~Destructor()
       o A set of public methods which direcly call SharedObject protected
         method. The reason is to avoid casting in the application:
           void modify() calls SharedObject::_modify()
	   Classname* dup() calls SharedObject* SharedObject::_dup()
	   void updatePostcond ()
	   void incrRefPostcond ()
	   void decrRefPostcond ()
	   void destroyPostcond ()

       o A friend for creation:
           friend Classname* CreateClassname (ulong, const ValueType*,
	                                      const char* name)
           which is written as:
              return CreateSharedObject(Classname, mask, values, name).

COPYRIGHT :	      
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.


  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#ifndef _SharedObject_h
#define _SharedObject_h

#include <Go/machdef.h>
#include <Go/Tools/Bool.h>
#include <Go/Tools/List.h>
#include <Go/Tools/Assoc.h>
#include <Go/Tools/Callback.h>

struct GFunctionList;

/*
 * GCreateSharedObject macro
 */

#if defined (__STDC__) || defined (__BORLANDC__) || defined (__MSC7__)
#define GCREATE_SHAREDOBJECT(class, mask, values, name) \
    (class*)(\
        ( (class::_funcList.create == NULL) \
	  ? \
              class::_funcList.create = (GShoConstructor)&_new##class,\
	      class::_funcList.head = (GShoHeadList)&class::_head,\
	      class::_funcList.value_size = (GShoValueSize)&class::_valueSize,\
	      0\
	  : \
	      0),\
        ((class*)ModifyOrCreateShO (&class::_funcList, mask, (const void*)values, name))\
    )
/*
 * ModifyNamedSharedObject macro
 */
#define GMODIFY_NAMED_SHAREDOBJECT(class, mask, values, name) \
    if (class::_funcList.create == NULL) \
	{\
	 class::_funcList.create = (GShoConstructor)&_new##class;\
	 class::_funcList.head = (GShoHeadList)&class::_head;\
	 class::_funcList.value_size = (GShoValueSize)&class::_valueSize;\
	}\
    ((void)ModifyNamedShO (&class::_funcList, mask, (const void*)values, name))

#else /* ! __STDC__ */

#define GCREATE_SHAREDOBJECT(class, mask, values, name) \
    (class*)(\
           ( (class::_funcList.create == NULL) \
	  ? \
              class::_funcList.create = (GShoConstructor)&_new/**/class,\
	      class::_funcList.head = (GShoHeadList)&class::_head,\
	      class::_funcList.value_size = (GShoValueSize)&class::_valueSize,\
	      0\
	  : \
	      0),\
           ((class*)ModifyOrCreateShO (&class::_funcList, mask, (const void*)values, name))\
    )
/*
 * ModifyNamedSharedObject macro
 */
#define GMODIFY_NAMED_SHAREDOBJECT(class, mask, values, name) \
    if (class::_funcList.create == NULL) \
	{\
	 class::_funcList.create = (GShoConstructor)&_new/**/class;\
	 class::_funcList.head = (GShoHeadList)&class::_head;\
	 class::_funcList.value_size = (GShoValueSize)&class::_valueSize;\
	}\
    ((void)ModifyNamedShO (&class::_funcList, mask, (const void*)values, name))

#endif /* __STDC__ */


/*
 * Postcond flags
 */
struct GShoPostcondFlags
{
    GBool _has_update_postcond :1;
    GBool _has_destroy_postcond :1;
    GBool _has_incrRef_postcond :1;
    GBool _has_decrRef_postcond :1;

    GShoPostcondFlags() {
	_has_update_postcond  = GFALSE;
	_has_destroy_postcond = GFALSE;
	_has_incrRef_postcond = GFALSE;
	_has_decrRef_postcond = GFALSE;
    }
};

class GoStream;

class GSharedObject : public GList {

  public:

    static const char* classname();
    virtual const char* getClassname() const;
    static int save(const GSharedObject*, GoStream*) { return 1; }
    //  static int restore(GSharedObject**, GoStream*);
    static int readParams(GSharedObject* sh, GoStream*){sh->_refcnt--; return 1;}

    void modify (GSharedObject*& sho, ulong mask, const void* values);
    void rename (GSharedObject*& sho, const char* name);
    GSharedObject* dup();
    void updatePostcond (void (*)(void*, GSharedObject*, ulong update_mask),
			 const void* closure, GAddOrRemove = GADD) const;
    void incrRefPostcond (void (*)(void*, GSharedObject*),
			  const void* closure, GAddOrRemove = GADD) const;
    void decrRefPostcond (void (*)(void*, GSharedObject*),
			  const void* closure, GAddOrRemove = GADD) const;
    void destroyPostcond (void (*)(void*, const GSharedObject*),
			  const void* closure, GAddOrRemove = GADD) const;
    void destroy ();

    inline long refcnt() const { return _refcnt; }
    inline ulong mask() const { return _mask; } 
    inline const char* name() const { return _name; } 
    GBool isNamed () const { return (GBool)(_name != NULL); } 

    // give an list API at SharedObj level;
    
    GSharedObject* headOfList() const;
    GSharedObject* next() const;
    
  protected:

    long  _refcnt;
    ulong _mask;
    char* _name;

    GShoPostcondFlags _flags;
    GAssoc _callbacks;

    int _addCallback (const char* type, GCBProc proc, const void* closure) const;
    int _remCallback (const char* type, GCBProc proc, const void* closure) const;

    void _call_update_postcond (ulong update_mask);
    void _call_destroy_postcond ();
    void _call_incrRef_postcond ();
    void _call_decrRef_postcond ();

    GSharedObject (GFunctionList*, ulong mask, const char* name);
    virtual ~GSharedObject ();

    // These functions call the subclass functions;
    GSharedObject* _create(ulong mask, const void* values, const char* name = NULL);

    virtual GBool _match (ulong mask, const void* values) const = 0;
    virtual void _makeNewValues (ulong mask, const void* values,
				 ulong& newMask, void* newValues) const = 0;
    virtual ulong _update (ulong mask, const void* values) = 0;
    
    GList*& _head();
    int _valueSize();

    friend GSharedObject*  ModifyOrCreateShO (GFunctionList* funcList,
					      ulong mask, const void* values, const char* name);
    friend void  ModifyNamedShO (GFunctionList* funcList,
				 ulong mask, const void* values, const char* name);


  private:
    GFunctionList* __funcList;

    friend GSharedObject*  _findAndModifyUnnamed (GList*& head, ulong mask, const void* values);
    friend GSharedObject*  _findAndModifyNamed (GList*& head, ulong mask, const void* values, const char* name);
};



/*
 * Function types
 */

typedef GSharedObject* (*GShoConstructor)(GFunctionList* funcList,
					  ulong mask,
					  const void* values,
					  const char* name);


// the 2 following typedef should be a declaration of static methods;
// but how to do  with each C++ compiler ?;
// the following method does not have to use "this" which may be NULL;
    
typedef GList*& (GSharedObject::*GShoHeadList)();
typedef int (GSharedObject::*GShoValueSize)();

/*
 * GFunctionList structure
 */
struct GFunctionList
{
    GShoConstructor create;
    GShoHeadList head;
    GShoValueSize value_size;
};

/*
 * Inline functions
 */

#endif /* _SharedObj_h */


