/* $Id: Translator.h,v 1.1 1992/12/10 15:38:18 jojic Exp $ */
/*****************************************************************************
MODULE:  Go Kernel (User Interface)
FILE  :  Translator.h  
AUTHOR:  Serge Mbikina
DATE  :  June 92

DESCRIPTION:

  A translator is used for the programing of simplest user interactions: 
  event/action translation tables.

  A GTranslator is a subclass of GInteractor. But it is intended specialy
  for sensors. Once a translator has been programed, a sensor will be able
  to retreive it only if it has a name of the form:
	"GoClassName.ActionName"
  (see Sensor(3Go)).

  The main functionality of a translator, the ``add'' method, allows
  the definition of a list of functions to call, on some user event.
  The first parameter of the ``add'' method is a string. This string is
  a mask representing a set of possible events.
  The syntax is something like:
	"key<a-z>" or "press<2>s".
  (see Translator(3Go)).
  
  The second parameter of the ``add'' method is also a string. This string
  defines the function to call and its parameters.
  For example:
	"Go::destroy(go)"
  (see Translator(3Go) for the description of the syntax and the definition
  of the set of allowed function parameters).

  If more than one function are associated to the same event mask
  (through many calls to ``add''), they will all be called (at the time
  of the execution of the interaction), if an event corresponding to
  the mask arrives.

  In fact, translators will rarely be programmed in C++. It is better
  to define translators in ascii files. Therefore the user will be able,
  without recompiling sources, to choose which event will start which action.
  These ascii files contain lines such as:
	Go.goDispatcher   key<^D>   Go::destroy(go)

  Such a line defines an entry in the event/actions table of the translator
  named "Go.goDispatcher" (See Translator(3Go)) for the exact description
  of these files).
  The ``GLoadTranslatorPgms(const char* file_name)'' function will interpret
  such files and create (or update) corresponding translators.

  During the execution, translators must retrieve function pointers which
  are associated to function names. To perform this, translators use the
  table of names/pointers managed by the ``GBind'' function (see Bind(3Go)).
  Each Go subclass provides a set of methods automatically loaded in GBind
  tables (see the documentation of each Go subclass). Translators provide the
  ``CreateSensor'' method, which is also automatically binded. This function
  allows you to start a new interaction.

  You can put the call of your own functions inside these ascii files.
  But you must make the association between the name of the function
  (inside the file) and the function pointer to be executed.
  For example:
	GBind("QuickQuit", exit);
  and inside the file:
	GoRoot.mainDispatcher   key<^C>   QuikQuit(0)


COPYRIGHT:
  Copyright (c) 1990, 1991, 1992 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.

*****************************************************************************/
#ifndef _Translator_h
#define _Translator_h

#include <Go/ItorPgm.h>
#include <Go/Tools/Queue.h>



/*
 *-------------------------------------------------------
 * Programmation of a Translator
 *-------------------------------------------------------
 */
class GPointerArray;

class GTranslatorPgm : public GInteractorPgm
{
  public:

    static GTranslatorPgm* create (const char* name);
    virtual ~GTranslatorPgm (); //all associated "GTranslator"s should be destroyed before;

    friend GBool GLoadTranslatorPgms (const char* filename);

    GBool add (const char *event_mask, const char *function_to_call);
    GBool remove (const char *event_mask, const char *function_to_call = NULL);

    //return NULL if ok. Otherwise return an error message:
    //(never free the returned string)
    static const char* checkSyntax (const char* event_mask, const char* function_to_call);
    static const char* checkFileSyntax (const char* filename, int& wrong_line_number);
    static const char* checkFileLineSyntax (const char* line);

    /* superclasses virtual methods */
    virtual GInteractor* start (const GEvent*, Go*);
    virtual GEventMask inputMask () const;

    GTranslatorPgm* superTranslatorPgm() const;
    static GPointerArray* instances();
    
  protected:

    GTranslatorPgm (const char*);
    const GQueue* GetHandlers () const { return &_handlers; }

  private:

    GQueue _handlers ;
    static GPointerArray* _instances;   // array the translator pgms;
    friend class GTranslator;
};

inline GPointerArray* GTranslatorPgm::instances()
{
    return _instances;
}

/*
  *-------------------------------------------------------
  * Execution of a Translator
  *-------------------------------------------------------
  */

class GTranslator : public GInteractor
{
  public:
    virtual ~GTranslator ();

    /* superclasses virtual methods */
    virtual GBool careAbout (const GEvent*) const;
    virtual GBool handle (const GEvent*);

  protected:
    GTranslatorPgm* Pgm () const { return (GTranslatorPgm*) pgm(); }

    GTranslator (const GTranslatorPgm*, Go*);

    friend class GTranslatorPgm;
};

#endif /* _Translator_h */
