/* $Id: XGoInterf.h,v 1.4 1992/12/11 08:35:30 jojic Exp $ */
/*****************************************************************************
MODULE:  Interfaces
FILE:    XGoInterf.h
AUTHOR:  Philippe Smadja, Jacques Davy
DATE:    August 1990 (Version 1.0)

DESCRIPTION:
    The X Window System interface for Go.

    The XGoInterf realizes the interface between Go and the X Window System.
    For performance reasons, the XGoInterf keeps lists of associations
    between Go resources and X resources. These association lists are:

       o Xgc list for graphic contexts (Pen, Brush, ... , gc)
       o Xfont list for fonts (Font, XFontStruct)
       o Xcolor list for colors (there is only one colormap)
       o Xpixmap list for PixelArray

    The Go resources are SharedObject, that means there are callback list
    available to be notified when creating, modifying or destroying such
    resources. When modifications or destructions occur, the appropriate
    callbacks are called.

COPYRIGHT :
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#ifndef _XGoInterf_h
#define _XGoInterf_h

// X includes;

#include <X11/Xlib.h>
#include <X11/Xutil.h>

// Go includes;

#include <Go/GoInterf.h>
#include <Go/GoInDisp.h>
#include <Go/Tools/Assoc.h>
#include <Go/Tools/Array.h>
#include <Go/Tools/Stack.h>
#include <Go/Tools/Alloc.h>

struct GGCClip
{
    XRectangle*    xrects;
    ushort         nxrects;
    ushort         ordering;
    Region         clip_mask;
    /**
      the Go kernel gives YXsorted in GNumber
      We are not sure it is GTRUE after pixels computation.
      So we use Ysorted.
      **/
    GGCClip() { clip_mask = NULL; xrects = NULL; nxrects = 0; ordering = YSorted;};
    ~GGCClip() { if (clip_mask) XDestroyRegion(clip_mask); GFREE(xrects); } ;
};


// private classes;

class ShXgcValues;
class ShXRegion;
class ShXFont;

// public;

#define GNO_BACKGROUND ((ulong)-1)
class GShArray;

class XGoInterf : public GoInterf
{
  public:

    XGoInterf(const char* displayName = NULL);

    XGoInterf(Display* display, Window window);

    XGoInterf(Display* display, int screen, Drawable drawable, GBool isWindow,
	      int width_in_pixels, int height_in_pixels, int depth,
	      ulong backgroundPixel, Colormap colormap, Visual* visual,
	      GNumber pixelWidth_in_mm = 0, GNumber pixelHeight_in_mm = 0);

    virtual ~XGoInterf();

    // Specific XGoInterf methods;


    inline Display*   display()       const { return _dpy; };
    inline Drawable   drawable()      const { return _drawable; };
    inline Colormap   colormap()      const { return _colormap; };
    inline uchar      screen()        const { return _screen;};
    inline Visual*    visual()        const { return _visual; };
    inline uchar      bitsPerPixel()  const { return _depth; } ;
    inline uchar      bitOrder()      const {
	return BitmapBitOrder(_dpy) == LSBFirst ? GLSB_FIRST : GMSB_FIRST;
    }
    inline uchar      byteOrder()     const {
	return ImageByteOrder(_dpy) == LSBFirst ? GLSB_FIRST : GMSB_FIRST;
    };
    
    inline uchar      bitmapUnit()    const { return BitmapUnit(_dpy); } ;
    inline uchar      bitmapPad()     const { return BitmapPad(_dpy); };
    inline uchar      bitmapCoding()  const { return GBITMAP_CODING; } ;
    inline GBool      useColorPatternMatching() { return _useColorPatternMatching;}
    int               fontHeightInPixels (const char* xfontname);

    GFont*            xFontNameToFont (const char* xfontname);
    virtual GShFont*  xFontNameToShFont (const char* xfontname);
    virtual void      handleXEvent    (const XEvent* xevent);
    virtual void      allocContinuousColormap(ulong*& pixels, int& npixels);
    virtual void      useColorPatternMatching (GBool);

    ulong             doubleClickInterval() { return _doubleClickInterval; } ;
    void              doubleClickInterval(ulong interval) {
	                      _doubleClickInterval = interval; } ;
    
    // end of specific

    //Actions on drawable:
    
    virtual void      map();
    virtual void      unmap();
    virtual void      flush();
    virtual void      synchronize(GBool discard);
    
    //Events:

    virtual void      addInputMask(GEventMask);
    virtual void      removeInputMask(GEventMask);

    virtual GBool     captureEvent(GEventMask input_mask, GBool confined_in_window,
				   GTimeStamp reference = GCURRENT_TIME);
    virtual void      unCaptureEvent(GTimeStamp reference = GCURRENT_TIME);

    // Input methods:
    
    virtual GPoint     getPointer     (const GCoordSys* = NULL);
    virtual void       setPointer     (const GPoint&,
				       const GCoordSys* = NULL);
    virtual ulong      inputFd();
    virtual GNumber    pixelWidth()   const  { return _pixelWidth; };
    virtual GNumber    pixelHeight()  const { return _pixelHeight; };


    //Output methods handled by the xinterf;

    virtual void      clear      (const GRegion&);
    virtual void      line       (const GShPen*, const GCoordSys&, const GPoint&,
				  const GPoint&, GQuality = GMAX_QUALITY);
    virtual void      polyline   (const GShPen*, const GShBrush*, const GCoordSys&, int npoint,
				  const GPoint*, GQuality =  GMAX_QUALITY);
    virtual void      polygon    (const GShPen*, const GShBrush*, const GCoordSys&, int npoint,
				  const GPoint*, GQuality = GMAX_QUALITY);
    virtual void      string     (const GShTextBrush*, const GShBrush*,
				  const GCoordSys&, const GPoint&,
				  GNumber height,
				  int nchar, const char*, GQuality = GMAX_QUALITY);
    virtual void      image      (const GShImageBrush*, const GCoordSys&, const GPoint&,
				  GNumber width, GNumber height, const GShPixelArray*,
				  GQuality = GMAX_QUALITY);

    virtual void      rectangle  (const GShPen*, const GShBrush*,
				  const GCoordSys&, const GPoint&,
				  GNumber width, GNumber height, GQuality = GMAX_QUALITY);

    virtual void      ellipse    (const GShPen*, const GShBrush*, const GCoordSys&,
				  const GPoint& center,
				  GNumber radiusX, GNumber radiusY,
				  GNumber angle1 = 0, GNumber angle2 = 2*PI,
				  GQuality = GMAX_QUALITY);

    virtual void      getImage   (GImageBrushValues& imagebrush_result,
				  GPixelArrayValues& pixelarray_result,
				  const GCoordSys& cs, const GPoint& origin,
				  GNumber width, GNumber height, ushort depth = 0,
				  GBool compress_colormap = GTRUE,
				  ushort x_resolution = 0, ushort y_resolution = 0,
				  GQuality quality = GMAX_QUALITY);

    // Setting misc methods;
    
    virtual void     resize        (GNumber width_in_mm, GNumber height_in_mm);
    virtual void     resize        (int width_in_pixel, int height_in_pixel);
    virtual void     background    (const GShColor*);
    virtual void     background    (ulong pixel);
    virtual GShColor* colorFromName (const char* color_name) const;
    
    // Query methods for font;

    virtual const  char*  fontName        (const GCoordSys&, GNumber height,
					   const GShFont* font) ;

    virtual GNumber stringLeftBearing  (const char* string, int size,
					GNumber height, const GShFont* font) ;
    virtual GNumber stringRightBearing (const char* string, int size,
					GNumber height, const GShFont* font) ;
    virtual GNumber stringAscent       (const char* string, int size,
					GNumber height, const GShFont* font) ;
    virtual GNumber stringDescent      (const char* string, int size,
					GNumber height, const GShFont* font) ;
    virtual GNumber stringWidth        (const char* string, int size,
					GNumber height, const GShFont* font) ;


    virtual GNumber fontAscent             (GNumber height,
					    const GShFont* font) ;
    virtual GNumber fontDescent            (GNumber height,
					    const GShFont* font) ;
    virtual GNumber fontUnderlineThickness (GNumber height,
					    const GShFont* font) ;
    virtual GNumber fontUnderlinePosition  (GNumber height,
					    const GShFont* font) ;
    virtual GNumber fontItalicAngle        (const GShFont* font) ;
    virtual GNumber fontStrikeoutAscent    (GNumber height,
					    const GShFont* font) ;
    virtual GNumber fontStrikeoutDescent   (GNumber height,
					    const GShFont* font) ;
    virtual GNumber fontMinSpace           (GNumber height,
					    const GShFont* font) ;
    virtual GNumber fontMaxSpace           (GNumber height,
					    const GShFont* font) ;
    virtual GNumber fontNormSpace          (GNumber height,
					    const GShFont* font) ;
    virtual GNumber fontEndSpace           (GNumber height,
					    const GShFont* font) ;

    virtual void   pushCursor             (long cursorIdent);
    virtual void   popCursor              ();

    virtual void   bell                   (int percent = 0);


  protected:
    //Conversions;

    static  ulong GEventMaskToXEventMask(GEventMask,
					 ulong xmask_to_add = StructureNotifyMask);
    static  uint ModifierMaskFromXMask (uint xmask) {
	return xmask & 0x0FF;
    }
    static  GTimeStamp XTimeToGTimeStamp(Time xtime) {
	GTimeStamp gtime;
	if (xtime == CurrentTime)
	    gtime = GCURRENT_TIME;
	else gtime = (GTimeStamp)xtime;
	return gtime;
    }
    static  Time GTimeStampToXTime(GTimeStamp gtime)   {
	Time xtime;
	if (gtime == GCURRENT_TIME)
	    xtime = CurrentTime;
	else xtime =  (Time)gtime;
	return xtime;
    }

    static GButton xbutton_to_gbutton(int xbutton) {
	switch (xbutton)
	{
	  case 1: return GBUTTON1;
	  case 2: return GBUTTON2;
	  case 3: return GBUTTON3;
	  case 4: return GBUTTON4;
	}
	return GBUTTON5;
    }


    
    GEvent*      xevent_to_gevent(const union _XEvent*,
				    const XRectangle* xrects =  NULL,
				    int nxrects = 0) const;
    GShColor*      pixel_to_color   (ulong pixel);

    inline GNumber pixx_to_mm(int x) const {
	return x * _pixelWidth;              // conform to GPositionToPixel definition;
    }
    inline GNumber pixy_to_mm(int y) const {
	return y * _pixelHeight;
    }
    inline GNumber pixw_to_mm(int w) const {
	return w * _pixelWidth;
    }
    inline GNumber pixh_to_mm(int h) const {
	return h * _pixelHeight;
    }

    inline int    mm_to_pixw(GNumber mm) const {
	return (int)(mm / _pixelWidth + (GNumber)0.5);
    }
    inline int    mm_to_pixh(GNumber mm) const {
	return (int)(mm / _pixelHeight + (GNumber)0.5);
    }
    
    // for Position conversion use the GRegion algorithm:
    
    inline int    mm_to_pixx(GNumber pos_in_mm) const {
	return GPositionToPixel(pos_in_mm, _pixelWidth);
    }

    inline int    mm_to_pixy(GNumber pos_in_mm) const {
	return GPositionToPixel(pos_in_mm, _pixelHeight);
    }

    Display*   _dpy;
    uchar      _screen;
    Drawable   _drawable;
    Visual*    _visual;
    Colormap   _colormap;
    GArray     _exposeEventRectangles;
    GGCClip    _gcclip;
    
    GAssoc*   _assocGC;
    GAssoc*   _assocColor;
    GAssoc*   _assocFont;
    GAssoc*   _assocPattern;
    GAssoc*   _assocDash;
    GAssoc*   _assocColorPattern;
    GAssoc    _assocCursor;
    GStack    _cursorStack;
    
    ushort   _isWindow :1,
    _keepDrawable:1,
    _grabPointer:1,
    _grabKeyboard:1,
    _waitingDouble:1,
    _useColorPatternMatching:1,
    _removeNextReleaseEvent:1;
    
    GNumber   _pixelWidth;
    GNumber   _pixelHeight;

    /**  for double click handling **/
    GPointerArray _unHandledEvents;
    ulong         _doubleClickInterval; // in millisecond;
    
   /** depth 1 GC **/
    GC       _clearGC; 
    GC       _setGC;
    GC       _copyGC;

    /** depth of the window GC **/
    GC       _clearWindowGC;
    
    ulong    _xinputMask;
    GEventMask    _captureInputMask;
    
    void     _commonInit();

    void              _colorToXGCValues(GShColor* color,
					GShColor* alt_color,
					ShXgcValues*, GQuality);
    
    ShXgcValues*     _shPenToShXgcValues(const GShPen* shpen,
					 ShXgcValues* shxgcvalues,
					 GQuality);
    ShXgcValues*     _shBrushToShXgcValues(const GShBrush* shbrush,
					   ShXgcValues* shxgcvalues,
					   GQuality);
    ShXgcValues*     _shTextBrushToShXgcValues(const GShTextBrush* shtextbrush,
					       const GCoordSys&,
					       GNumber font_height,
					       ShXgcValues* shxgcvalues,
					       GQuality);
    ShXgcValues*     _shImageBrushToShXgcValues(const GShImageBrush* shimagebrush,
						uchar depth, const ShXRegion*,
						int shxregion_xoffset,
						int shxregion_yoffset,
						ShXgcValues* shxgcvalues,
						GQuality);

    ulong            _colorToXPixel(const GShColor* color,
				    GBool nearest = GTRUE,
				    int min_lum = 0,
				    int max_lum = GMAX_RGB_VALUE);
    GArray*          _colormapToXPixels(const GShPixelArray*, const GShColormap* colormap);

    const  ShXFont* _fontToShXFont(const GShFont* font,
				   const GCoordSys& cs,
				   GNumber height);

    void            _shPatternToX(const GShPixelArray*,
				  const GShColormap* ,
				  GNumber width, GNumber height,
				  const GCoordSys& ,
				  GQuality quality,
				  Pixmap* pixmap_result,
				  const ShXRegion**  xregion_result,
				  ushort* pixmap_width = NULL,
				  ushort* pixmap_height = NULL,
				  ushort* pixmap_xorigin = NULL,
				  ushort* pixmap_yorigin = NULL);

    Region   _stippleToXRegion(const GShPattern* stipple, const GCoordSys& ) const;
    
    Font     _fontToXFont(const GShFont* font, const GCoordSys& cs, GNumber height);
    Pixmap   _tileToXTile(const GShPattern* tile, GQuality);
    Pixmap   _stippleToXStipple(const GShPattern* stipple, GQuality);
    ulong    _dashToXDash(const GShDash* dash, char** dash_list, int* n_dashes,
			  int* dash_offset) const;

    void   _freeXGCs();
    void   _freeXPixels();
    void   _freeXFonts();
    void   _freeXPatterns();
    void   _freeXDashes();
    void   _freeStippleXRegion() const;

    void   _setClip(GC gc);
    void   _setGCClip(GBool makeYXBanded = GTRUE);

    void   _pixelArrayValuesToXImage(const GPixelArrayValues*, XImage*) const;
    void   _ximageToPixelArrayValues(const XImage*, GPixelArrayValues*) const;
    
    GC     _shPenToGC        (const GShPen* shpen, GQuality);
    GC     _shBrushToGC      (const GShBrush* shBrush, GQuality);
    GC     _shImageBrushToGC (const GShImageBrush* shImageBrush,
			      const GCoordSys&,
			      uchar depth, const ShXRegion* xregion,
			      int xoffset, int yoffset, GQuality);
    GC     _shTextBrushToGC  (const GShTextBrush* shTextBrush, const GCoordSys&, 
			      GNumber height, const ShXFont** shxfont, GQuality);


    void   _addXInterfIfNeeded(GShArray*& sha) const ;
    void   _remXInterfIfInside(GShArray*& sha) const ;

    void   _handleUnhandledEvents();
    void   _doubleClickTimer();
    GEventMask  _currentInputMask();

    friend void  DoubleClickTimer(void* This);
};

#endif /* _XGoInterf_h */



