#ifdef REV_INFO
static char rcsid[] = "$Id: EdAsRect.cc,v 1.1 1992/12/23 12:06:37 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Go Kernel (User Interface)
FILE:    EdAsRect.cc
AUTHOR:  Philippe Smadja/Gilles Benati
DATE:    June 92

DESCRIPTION:
  See the EdAsRect.h header.

  It is not an interactor. It is use by interactor
  to compute rectangle parameter during interaction.

COPYRIGHT:
  Copyright (c) 1990, 1991, 1992 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.

*****************************************************************************/

#include <Go/EdAsRect.h>
#include <Go/Tools/Bind.h>
#include <Go/Tools/Cursor.h>

    
GEditAsRectangle::GEditAsRectangle (GNumber width, GNumber height, int gravity)
{
    _top_left.x = 0;
    _top_left.y = 0;
    
    // update the current mask with the wanted gravity;
    _mask = gravity;
    
    _initial_width = _width = width;
    _initial_height = _height = height;
}


GEditAsRectangle::~GEditAsRectangle ()
{
    // nothing to do;
}

/***
 *
 * rectangle computation
 *
 ***/

void GEditAsRectangle::compute (const GPoint& relative_point)
{
    // The following was written by gilles and adapted by philippe;


    /*
     *  closure->mask determines the behavior. valid values are :
     * 	o  G_MOVE_xxx where xxx is NORTH, SOUTH, EAST or WEST
     *	   bitwise 'or' | combination allowed are NE, NW, SE and SW.
     *	   Only the corresponding rectangle sides follow the pointer (mouse).
     * 	o  G_NO_GRAVITY ; the segment [closure->cs.O, pointer position] is
     *     always the rectangle diagonal line.
     *	o  G_OPPOSITE_GRAVITY ; the pointer (mouse) drags whichever rectangle
     *     side it crosses first. As soon as a side is dragged beyond its
     *	   opposite rectangle side, it goes back to its initial position an
     *	   the opposite side is dragged instead.
     */

    GPoint pt = relative_point;

    // So far, G_AS_SQARE only works with G_NO_GRAVITY;
    // This will hopefully be extended to other modes i.e. G_MOVE_XXXX ;
    
    if (_mask & G_AS_SQUARE) {
	// modify pt as if it was defining a perfect square;
	GNumber average = (fabs(pt.x) + fabs(pt.y)) / 2;
	pt.x = average * (int)(pt.x >= 0 ? 1 : -1);
	pt.y = average * (int)(pt.y >= 0 ? 1 : -1);
	NoGravity(pt);
    }
    else if (_mask & G_NO_GRAVITY) {
	NoGravity(pt);
    }
    else {
	if (_mask & G_OPPOSITE_GRAVITY)
	    OppositeGravity(pt);
	if (_mask & G_MOVE_SOUTH) {
	    // Keep the northern side at same y;
	    _height = (pt.y < 0.) ? _initial_height : pt.y;
	}
	else if (_mask & G_MOVE_NORTH) {
	    // Keep the southern side at same y;
	    if (pt.y > _initial_height) {
		_top_left.y = 0.;
		_height     = _initial_height;
	    }
	    else {
		_top_left.y = pt.y;
		_height     = _initial_height - pt.y;
	    }
	}
	if (_mask & G_MOVE_EAST) {
	    // Keep the western side at same x;
	    _width  = (pt.x < 0.) ? _initial_width  : pt.x;
	}
	else if (_mask & G_MOVE_WEST) {
	    // Keep the eastern side at same x;
	    if (pt.x > _initial_width) {
		_top_left.x = 0.;
		_width      = _initial_width;
	    }
	    else {
		_top_left.x = pt.x;
		_width      = _initial_width - pt.x;
	    }
	}
    }
}


void GEditAsRectangle::NoGravity(const GPoint& pt)
{
    // Updates _top_left, width & height according to pt;
    // Keeps ((0, 0), pt) as diagonal line of the rectangle;
    if (pt.x < 0.) {
	_top_left.x = pt.x;
	_width      = -pt.x;
    }
    else {
	_top_left.x = 0.;
	_width      = pt.x;
    }
    if (pt.y < 0.) {
	_top_left.y = pt.y;
	_height     = -pt.y;
    }
    else {
	_top_left.y = 0.;
	_height     = pt.y;
    }
}


void GEditAsRectangle::OppositeGravity(const GPoint& pt)
{
    /*
     * Updates top_left, width & height according to pt.
     * In fact, modifies _mask when pt goes out of the original rectangle
     * to keep the opposite side invariant
     * Other constraints (north, west etc..) must be applied afterwards.
     */
    if (pt.x < 0.)
	_mask = (_mask | G_MOVE_WEST) & ~G_MOVE_EAST;
    else if (pt.x > _initial_width) {
	_mask = (_mask | G_MOVE_EAST) & ~G_MOVE_WEST;
	_top_left.x = 0.;
    }
    if (pt.y < 0.)
	_mask = (_mask | G_MOVE_NORTH) & ~G_MOVE_SOUTH;
    else if (pt.y > _initial_height) {
	_mask = (_mask | G_MOVE_SOUTH) & ~G_MOVE_NORTH;
	_top_left.y = 0.;
    }
}


/**
FUNCTION
    GRectangleGravityToCursorIdent(int gravity)
DESCRIPTION
    returns the cursor ident associated
    to a given gravity
**/
int GRectangleGravityToCursorIdent(int gravity)
{
    int cid = GCURSOR_DEFAULT;
    gravity &=  ~(G_OPPOSITE_GRAVITY|G_NO_GRAVITY);
    switch (gravity)
    {
      default:
      case G_NO_GRAVITY:
	cid = GCURSOR_ARROW_4D;;
	break;

      case G_MOVE_NORTH:
	cid = GCURSOR_RESIZE_NORTH;
	break;

      case G_MOVE_EAST:
	cid = GCURSOR_RESIZE_EAST;
	break;
	
      case G_MOVE_SOUTH:
	cid = GCURSOR_RESIZE_SOUTH;
	break;

      case G_MOVE_WEST:
	cid = GCURSOR_RESIZE_WEST;
	break;

      case   G_MOVE_SOUTH | G_MOVE_EAST:
	cid = GCURSOR_RESIZE_SOUTH_EAST;
	break;

      case G_MOVE_SOUTH | G_MOVE_WEST:
	cid = GCURSOR_RESIZE_SOUTH_WEST;
	break;
	
      case   G_MOVE_NORTH | G_MOVE_EAST:
	cid = GCURSOR_RESIZE_NORTH_EAST;
	break;

      case G_MOVE_NORTH | G_MOVE_WEST:
	cid = GCURSOR_RESIZE_NORTH_WEST;
	break;
    }
    return cid;
}




