#ifdef REV_INFO
static char rcsid[] = "$Id: PointSensor.cc,v 1.2 1992/12/30 09:57:10 song Exp $";
#endif
/*****************************************************************************
MODULE:  GoUI
FILE:    PointSensor.cc
AUTHOR:  Gilles Benati
DATE:    September 20th, 1991

DESCRIPTION:
    < see PtSensor.h header file >
    The file is revued to be adapted to the version 1.2
    See the ::create static method.
    
COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <math.h>
#include <assert.h>
#include <Go/Tools/GeomUtil.h>
#include <Go/GoInterf.h>
#include <Go/GoStream.h>
#include <Go/PointSensor.h>
#include <Go/Polypoint.h>

/*
 *------------
 * CLASS PART
 *------------
 */


static char pointSensorMetaclassNradiusPrecond[]     = "p";
static char pointSensorMetaclassNradiusPostcond[]    = "p";

static GoPointSensorMetaclass* pointSensor_metaclass = NULL;


/**
FUNCTION
    GoPointSensorMetaclass::_pen_changed(GoPointSensorMetaclass* psm,
                                         const GPen* pen, unsigned long mask)
DESCRIPTION
    *** static member, private ***
    This function is a modify_post_condition of psm->_pen
    WARNING: it is assumed that pen **IS** psm->_pen
**/
void GoPointSensorMetaclass::_pen_changed(GoPointSensorMetaclass* psm,
					  const GPen*, unsigned long mask)
{
    Go* go;
    if (mask == GCOLOR || mask == GALT_COLOR ||
	mask == (GCOLOR & GALT_COLOR))
    {
	psm->init_sh_pen();
	for (go = psm->instances().first(); go;
	     go = psm->instances().next()) 
	    if (go->mapped())
		((GoPointSensor*)go)->ProtectedRedrawForeOnly();

    }
    else
    {
	for (go = psm->instances().first(); go;
	     go = psm->instances().next()) 
	    if (go->mapped())
		((GoPointSensor*)go)->ProtectedRedrawBackAndFore();


	psm->init_sh_pen();

	if (mask & GLINE_WIDTH)
	{
	    for (go = psm->instances().first(); go;
		 go = psm->instances().next())
	    {
		((GoPointSensor*)go)->ComputeBox();
		((GoPointSensor*)go)->ProtectedMyClipHasChanged();
	    }
	    for (go = psm->instances().first(); go;
		 go = psm->instances().next()) 
		if (go->mapped())
		    ((GoPointSensor*)go)->ProtectedRedrawForeOnly();
	}
    }
}


/**
FUNCTION
    GoPointSensorMetaclass::_brush_changed(GoPointSensorMetaclass* psm,
					   const GBrush* brush,
					   unsigned long mask)
DESCRIPTION
    *** static member, private ***
    This function is a modify_post_condition of psm->_brush
    WARNING: it is assumed that brush **IS** psm->_brush
**/
void GoPointSensorMetaclass::_brush_changed(GoPointSensorMetaclass* psm,
					    const GBrush*, unsigned long)
{
    for (Go* go = psm->instances().first(); go; go = psm->instances().next())
	if (go->mapped())
	    ((GoPointSensor*)go)->ProtectedRedrawBackAndFore();
}



GoPointSensorMetaclass* GoPointSensor::metaclass()
{
    if (pointSensor_metaclass == NULL)
    {
	pointSensor_metaclass = new GoPointSensorMetaclass;
	pointSensor_metaclass->superclass()->
	    AddSubclass(pointSensor_metaclass);
        GoStream::bind(GoPointSensor::metaclass()->classname(),
		       (int (*) (const void*, GoStream*)) GoPointSensor::save,
		       (int (*) (void**, GoStream*)) GoPointSensor::restore);

        GoSensor::bindActions(pointSensor_metaclass);
    }
    return pointSensor_metaclass;
}


GoPointSensorMetaclass::~GoPointSensorMetaclass()
{
    if (this == pointSensor_metaclass)
    {
        superclass()->RemoveSubclass(this);
	pointSensor_metaclass = NULL;
    }
    if (_old_sh_pen)
	    _old_sh_pen->destroy();
}


GoPointSensorMetaclass::GoPointSensorMetaclass()
    : _brush(0, (GBrushValues*) NULL)
    , _pen(0, (GPenValues*) NULL)
{
    _has_radius_precond = _has_radius_postcond = GFALSE;
    _radius = 1.;

    GRgb rgb_black(0, 0, 0);
    GShColor* color = GCreateShColor(&rgb_black);
    _brush.color(color);
    color->destroy();

    GRgb rgb_white(GMAX_RGB_VALUE, GMAX_RGB_VALUE, GMAX_RGB_VALUE);
    color = GCreateShColor(&rgb_white);
    _pen.color(color);
    color->destroy();

    _pen.modifyPostcond((void (*)(void*, GPen*, ulong)) _pen_changed,
			this, GADD);
    _brush.modifyPostcond((void (*)(void*, GBrush*, ulong)) _brush_changed,
			  this, GADD);
    _old_sh_pen = NULL;
    init_sh_pen();
}



Go* GoPointSensorMetaclass::m_create() const
{
    return new GoPointSensor();
}


const char* GoPointSensorMetaclass::m_classname() const
{
    return "GoPointSensor";
}


GoMetaclass* GoPointSensorMetaclass::m_superclass() const
{
    return GoSensor::metaclass();
}


void GoPointSensorMetaclass::m_check(const Go*) const {}


GBool GoPointSensorMetaclass::_call_radius_precond(GNumber *value) const
{
    // call conditions on all super-Metaclasses;
    for (const GoMetaclass* mc = this; mc; mc = mc->superclass())
    {
	if (((GoPointSensorMetaclass*) mc)->_has_radius_precond &&
	    ! mc->CallPreconditions(pointSensorMetaclassNradiusPrecond,
				    NULL, (void*) mc, value))
	    return GFALSE;
	// stop when reaching GoPointSensor Metaclass instance;
	if (mc == GoPointSensor::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

void GoPointSensorMetaclass::_call_radius_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = this; mc; mc = mc->superclass()) {
	if (((GoPointSensorMetaclass*) mc)->_has_radius_postcond)
	    mc->CallPostconditions(pointSensorMetaclassNradiusPostcond,
				   NULL, (void*) mc);
	// stop when reaching GoPointSensor Metaclass instance;
	if (mc == GoPointSensor::metaclass())
	    return;
    }
}


void GoPointSensorMetaclass::radiusPrecond(GBool
					   f(void*, const Go* /*NULL*/,
					     const GoPointSensorMetaclass*,
					     GNumber*),
					   void* closure, GAddOrRemove add)
{
    _has_radius_precond =
	AddOrRemoveCondition(pointSensorMetaclassNradiusPrecond,
			     f, closure, add);
}


void GoPointSensorMetaclass::radiusPostcond(void f(void*, Go* /*NULL*/,
						   GoPointSensorMetaclass*),
					    void* closure, GAddOrRemove add)
{
    _has_radius_postcond =
	AddOrRemoveCondition(pointSensorMetaclassNradiusPostcond,
			     f, closure, add);
}

/*
 *----------------
 * PUBLIC MEMBERS
 *----------------
 */
    


GoPointSensor* GoPointSensor::create
	(const char* name, Go* parent, const GEvent* startEvent,
	 GPointSensorPostcond postcond, void* closure) 
{
    GoPointSensor* ptsor = GoPointSensor::create ();
    if (ptsor)
    {
	ptsor->program (name);
	ptsor->start_event = startEvent;
	ptsor->_postcond = (GSensorPostcond)postcond;
	ptsor->_closure = closure;

	// reparent should be last;
	// because it may start an interactor;
	ptsor->reparent (parent);
    }
    return ptsor;
}


void create_point_sensor_action (const char* name, Go* parent,
					const GEvent* startEvent)
{
    if (name != NULL && parent != NULL)
	if (parent->getMetaclass()->isSubclass(GoPolypoint::metaclass()))
	    (void) GoPointSensor::create (name, (GoPolypoint*)parent,
					  startEvent, NULL, NULL);
}
	    

/*
 * The action table of GoPointSensor objects.
 * Functions inside this table will be accessible through
 * their name (i.e. a string). This is usePoint to implement some
 * interpreted languages dealing with graphic objects.
 */
static GActionRec PointSensor_actions[] =
{
    {"create(name, parent, event)", (GActionProc)create_point_sensor_action},
    {NULL, NULL} //this must be the last line
};


/*
 * This method returns the action table of GoPointSensor objects.
 */
const GActionRec* GoPointSensorMetaclass::m_actions () const
{
    return PointSensor_actions;
}



/**
FUNCTION
    GBool GoPointSensorMetaclass::radius(GNumber val);
DESCRIPTION
    public virtual
    If preconditions allow it, sets radius to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns
    GFALSE or bad parameters);
**/
GBool GoPointSensorMetaclass::radius(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (value >= 0. && value != _radius &&
	_call_radius_precond(&value) &&
	value >= 0. && value != _radius) {

	GNumber old_value = _radius;

	if (value < old_value) {
	    for (Go* go = instances().first(); go; go = instances().next()) {
		if (go->mapped()) {
		    ((GoPointSensor*) go)->ProtectedRedrawBackAndFore();
		}
	    }
	}
	
	_radius = value;
	for (Go* go = instances().first(); go; go = instances().next()) {
	    ((GoPointSensor*) go)->ComputeBox();
	}
	
	if (value > old_value) {
	    for (Go* go = instances().first(); go; go = instances().next()) {
		if (go->mapped()) {
		    ((GoPointSensor*) go)->ProtectedRedrawForeOnly();
		}
	    }
	}
	    
	_call_radius_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/*
 *---------------------
 * PROTECTED FUNCTIONS
 *---------------------
 */

/**
FUNCTION
    int GoPointSensor::save(const GoPointSensor* psen, GoStream* gostream);
DESCRIPTION
   **static member**
**/
int GoPointSensor::save(const GoPointSensor* psen, GoStream* gostream) {
    return GoSensor::save(psen, gostream);
}


/**
FUNCTION
    int GoPointSensor::restore(GoPointSensor** psen, GoStream* gostream);
DESCRIPTION
    
**/
int GoPointSensor::restore(GoPointSensor** psen, GoStream* gostream) {
    if (*psen = GoPointSensor::create()) {
	return GoSensor::readParams(*psen, gostream);
    }
    return 0;
}


/*
 *-----------------
 * PRIVATE MEMBERS
 *-----------------
 */

GoMetaclass* GoPointSensor::m_metaclass() const {
    return GoPointSensor::metaclass();
}


GBool GoPointSensor::m_contains(const GPoint& abspt, GNumber maxdist) const
{
    GoPolypoint *polyPt = (GoPolypoint*) parent();
    if (!polyPt)
	return GFALSE;

    if (! _bounding_box.contains(abspt, maxdist))
	return GFALSE;
    
    int nb = polyPt->nbPoint();

    if (nb < 1)
	return GFALSE;
    
    GoPointSensorMetaclass* my_metaclass =
	(GoPointSensorMetaclass*) getMetaclass();

    GNumber sqrDist = my_metaclass->_radius +
	my_metaclass->sh_pen()->lineWidth() / 2 + maxdist;
    sqrDist = sqrDist * sqrDist;
    GCoordSys cs = polyPt->Abscs();
    const GPoint *pts = polyPt->points();
    int i;

    for (i = 0; i < nb; i++) 
	if (GSquareDistance(abspt, cs.to(pts[i])) <= sqrDist)
	    return GTRUE;

    return GFALSE;
}


void GoPointSensor::m_add_clip(GRegion& region) const
{
    GoPolypoint *polyPt = (GoPolypoint*) parent();
    if (!polyPt)
	return;

    region += _bounding_box;
}


void GoPointSensor::m_draw(const GRegion& region) const
{
    GoPolypoint *polyPt = (GoPolypoint*) parent();
    if (!polyPt)
	return;

    int nb = polyPt->nbPoint();

    if (nb < 1)
	return;

    GoPointSensorMetaclass* my_metaclass =
	(GoPointSensorMetaclass*) getMetaclass();

    GoInterf         *intrf = interf();
    const GShBrush *shbrush = my_metaclass->_brush.shBrush();
    const GShPen     *shpen = my_metaclass->_pen.shPen();
    GNumber          radius = my_metaclass->_radius;
    GNumber        box_size = radius + shpen->lineWidth() / 2;
    GCoordSys            cs = polyPt->Abscs();
    const GPoint       *pts = polyPt->points();
    int i;

    for (i = 0; i < nb; i++)
    {
	GPoint center = cs.to(pts[i]);
	if (region && GBox(center.x - box_size, center.y - box_size,
			   center.x + box_size, center.y + box_size)) 
	    intrf->circle(shpen, shbrush, GUnitCoordSys(), center, radius);
    }
}

void GoPointSensor::ComputeBox()
{
    GoPointSensorMetaclass* meta_class =
	(GoPointSensorMetaclass*) getMetaclass();
    GoPolypoint* daddy = (GoPolypoint*) parent(); 
    GNumber radius = meta_class->_radius + meta_class->sh_pen()->lineWidth()/2;
    GPoint pmin, pmax;
    GPolypointGetAbsOutPoints(daddy->Abscs(), daddy->nbPoint(),
			      daddy->points(), pmin, pmax);
    _bounding_box.reset(pmin.x-radius, pmin.y-radius,
			pmax.x+radius, pmax.y+radius);
}


/**
FUNCTION
    GoPointSensor::_box_changed(void* closure, Go* go)
DESCRIPTION
    post condition on parent GoPolypoint manipulations.
    *** static member, private ***
**/
void GoPointSensor::_box_changed(void* closure, Go*)
{
    GoPointSensor* pt_s = (GoPointSensor*) closure;

    if (pt_s->mapped())
	pt_s->RedrawBackAndFore();

    pt_s->ComputeBox();
    pt_s->MyClipHasChanged();

    if (pt_s->mapped())
	pt_s->RedrawForeOnly();
}


/**
FUNCTION
    GoPointSensor::_parent_changed(void* closure, Go*, Go* old_parent, Go*)
DESCRIPTION
    post condition on reparent. GError if new parent is not a GoPolypoint.
    *** static member, private ***
**/
void GoPointSensor::_parent_changed(void* closure, Go*, Go* old_parent, Go*)
{
    GoPointSensor* pt_s = (GoPointSensor*) closure;

    GoPolypoint* ol_daddy = (GoPolypoint*) old_parent;
    
    if (ol_daddy)
    {
	ol_daddy->myClipPostcond((void (*)(void*, Go*))
			       GoPointSensor::_box_changed, pt_s, GREMOVE);
	ol_daddy->movePointPostcond((void (*)(void*, GoPolypoint*, int))
				  GoPointSensor::_box_changed, pt_s,
				    GREMOVE);
	ol_daddy->delPointPostcond((void (*)(void*, GoPolypoint*, int))
				 GoPointSensor::_box_changed, pt_s, GREMOVE);
	ol_daddy->addPointPostcond((void (*)(void*, GoPolypoint*, int))
				 GoPointSensor::_box_changed, pt_s, GREMOVE);
	ol_daddy->resetPointsPostcond((void (*)(void*, GoPolypoint*))
				    GoPointSensor::_box_changed, pt_s,
				      GREMOVE);
    }

    GoPolypoint* new_parent = (GoPolypoint*) pt_s->parent();

    if (new_parent)
    {
	// REMOVE ASSERT WHEN SAFE;
	assert(new_parent->getMetaclass()->isSubclass(GoPolypoint::metaclass()));

	new_parent->myClipPostcond(GoPointSensor::_box_changed, pt_s, GADD);
	new_parent->movePointPostcond((void (*)(void*, GoPolypoint*, int))
				    GoPointSensor::_box_changed, pt_s, GADD);
	new_parent->delPointPostcond((void (*)(void*, GoPolypoint*, int))
				   GoPointSensor::_box_changed, pt_s, GADD);
	new_parent->addPointPostcond((void (*)(void*, GoPolypoint*, int))
				   GoPointSensor::_box_changed, pt_s, GADD);
	new_parent->resetPointsPostcond((void (*)(void*, GoPolypoint*))
				      GoPointSensor::_box_changed, pt_s,
					GADD);
      GoPointSensor::_box_changed(pt_s, new_parent);
    }
}


/**
FUNCTION
    GoPointSensor()
DESCRIPTION
    GoPointSensor constructor
**/
GoPointSensor::GoPointSensor() : _bounding_box(GEmptyBox())
{
    reparentPostcond(GoPointSensor::_parent_changed, this);
}


