#ifdef REV_INFO
static char rcsid[] = "$Id: Arc.cc,v 1.3 1992/12/10 15:46:20 smadja Exp $";
#endif
/****************************************************************************
MODULE:  Graphic Objects
FILE:    Arc.cc
AUTHOR:  Gilles Benati
DATE:    Feb 1991

DESCRIPTION:
    See the Go/Tools/Arc.h header.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/

#include <assert.h>
#include <Go/Tools/Error.h>
#include <Go/Tools/GeomUtil.h>
#include <Go/GoInterf.h>
#include <Go/GoStream.h>
#include <Go/Arc.h>

static char arcNradiusXPrecond[]    = "p";
static char arcNradiusXPostcond[]   = "p";
static char arcNradiusYPrecond[]    = "p";
static char arcNradiusYPostcond[]   = "p";
static char arcNamplitudePrecond[]  = "p";
static char arcNamplitudePostcond[] = "p";
static char arcNstartPrecond[]      = "p";
static char arcNstartPostcond[]     = "p";
static char arcNresetPrecond[]      = "p";
static char arcNresetPostcond[]     = "p";
static char arcNpiePrecond[]        = "p";
static char arcNpiePostcond[]       = "p";
static char arcNclosedPrecond[]     = "p";
static char arcNclosedPostcond[]    = "p";

// to clarify pre/postcond code;
#define ARCB_PRECOND_PTR(f) GBool f(void*, const GoArc*, GBool*)
#define ARCN_PRECOND_PTR(f) GBool f(void*, const GoArc*, GNumber*)
#define ARCR_PRECOND_PTR(f) GBool f(void*, const GoArc*, GNumber*, GNumber*, GNumber*, GNumber*)
#define ARC_POSTCOND_PTR(f) void f(void*, GoArc*)


static GBool isOK(GoArc* arc)
{
    return arc && arc->getMetaclass()->isSubclass(GoArc::metaclass());
}

static void set_amplitude(GoArc *arc, GNumber value) {
    if (isOK(arc))
	arc->amplitude(value);
}

static void set_start(GoArc *arc, GNumber value) {
    if (isOK(arc))
	arc->start(value);
}

static void set_radius_x(GoArc *arc, GNumber value) {
    if (isOK(arc))
	arc->radiusX(value);
}

static void set_radius_y(GoArc *arc, GNumber value) {
    if (isOK(arc))
	arc->radiusY(value);
}

static void set_all(GoArc *arc, GNumber radX, GNumber radY,
		    GNumber ampl, GNumber strt) {
    if (isOK(arc))
	arc->reset(radX, radY, ampl, strt);
}

static void set_pie(GoArc *arc, GBool is_arc_pie) {
    if (isOK(arc))
	arc->pie(is_arc_pie);
}

static void set_closed(GoArc *arc, GBool is_closed) {
    if (isOK(arc))
	arc->closed(is_closed);
}

static GActionRec go_arc_action_array[] = {
  { "amplitude(GNumber)",  (GActionProc) set_amplitude },
  { "start(GNumber)",      (GActionProc) set_start },
  { "radiusX(GNumber)",    (GActionProc) set_radius_x },
  { "radiusY(GNumber)",    (GActionProc) set_radius_y },
  { "reset(GNumber)",      (GActionProc) set_all },
  { "pie(GNumber)",        (GActionProc) set_pie },
  { "closed(GNumber)",     (GActionProc) set_closed },
  { NULL, NULL }
};


INLINE GBool
check_values(GNumber radX, GNumber radY, GNumber amplitude, GNumber, char* f) {
    if (radX <= 0. || radY <= 0. || amplitude == 0.) {
	GWarning("**GoArc::%s : bad parameter value**", f);
	return GFALSE;
    }
    else return GTRUE;
}

INLINE void
best_values(GNumber* radX, GNumber* radY, GNumber* amplitude,GNumber*, char* f) {
    GBool message_on = GFALSE;
    if (*radX < 0.)       { message_on = GTRUE; *radX = 1; }
    if (*radY < 0.)       { message_on = GTRUE; *radY = 1; }
    if (*amplitude == 0.) { message_on = GTRUE; *amplitude = 2*PI; }
    if (message_on)
	GWarning("**GoArc::%s : bad parameter value**", f);
}


/*
 *--------------------
 * PRE/POSTCONDITIONS
 *--------------------
 */


GBool GoArc::_call_class_radiusX_precond(GNumber* radX) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_radiusX_precond &&
	    ! mc->CallPreconditions(arcNradiusXPrecond, this, radX))
	    return GFALSE;
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoArc::_call_class_radiusY_precond(GNumber* radY) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_radiusY_precond &&
	    ! mc->CallPreconditions(arcNradiusYPrecond, this, radY))
	    return GFALSE;
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoArc::_call_class_amplitude_precond(GNumber* ampli) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_amplitude_precond &&
	    ! mc->CallPreconditions(arcNamplitudePrecond, this, ampli))
	    return GFALSE;
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoArc::_call_class_start_precond(GNumber* startVal) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_start_precond &&
	    ! mc->CallPreconditions(arcNstartPrecond, this, startVal))
	    return GFALSE;
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoArc::_call_class_reset_precond(GNumber* radX, GNumber* radY,
				       GNumber* ampli, GNumber* start) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_reset_precond &&
	    ! mc->CallPreconditions(arcNresetPrecond, this,
				    radX, radY, ampli, start))
	    return GFALSE;
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoArc::_call_class_closed_precond(GBool* closedVal) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_closed_precond &&
	    ! mc->CallPreconditions(arcNclosedPrecond, this, closedVal))
	    return GFALSE;
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoArc::_call_class_pie_precond(GBool* pieVal) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_pie_precond &&
	    ! mc->CallPreconditions(arcNpiePrecond, this, pieVal))
	    return GFALSE;
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoArc::_call_radiusX_precond(GNumber* radX) const {
    return(_has_radiusX_precond ?
	   CallPreconditions(arcNradiusXPrecond, radX) : GTRUE);
}

GBool GoArc::_call_radiusY_precond(GNumber* radY) const {
    return(_has_radiusY_precond ?
	   CallPreconditions(arcNradiusYPrecond, radY) : GTRUE);
}

GBool GoArc::_call_amplitude_precond(GNumber* ampli) const {
    return(_has_amplitude_precond ?
	   CallPreconditions(arcNamplitudePrecond, ampli) : GTRUE);
}


GBool GoArc::_call_start_precond(GNumber* startVal) const {
    return(_has_start_precond ?
	   CallPreconditions(arcNstartPrecond, startVal) : GTRUE);
}

GBool GoArc::_call_reset_precond(GNumber* radX, GNumber* radY,
				       GNumber* ampli, GNumber* start) const {
    return(_has_reset_precond ?
	   CallPreconditions(arcNresetPrecond, radX, radY, ampli, start) :
	   GTRUE);
}

GBool GoArc::_call_pie_precond(GBool* pieVal) const {
    return(_has_pie_precond ?
	   CallPreconditions(arcNpiePrecond, pieVal) : GTRUE);
}

GBool GoArc::_call_closed_precond(GBool* closedVal) const {
    return(_has_closed_precond ?
	   CallPreconditions(arcNclosedPrecond, closedVal) : GTRUE);
}

void GoArc::_call_class_radiusX_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_radiusX_postcond)
	    mc->CallPostconditions(arcNradiusXPostcond, this);
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return;
    }
}

void GoArc::_call_class_radiusY_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_radiusY_postcond)
	    mc->CallPostconditions(arcNradiusYPostcond, this);
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return;
    }
}

void GoArc::_call_class_amplitude_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_amplitude_postcond)
	    mc->CallPostconditions(arcNamplitudePostcond, this);
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return;
    }
}

void GoArc::_call_class_start_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_start_postcond)
	    mc->CallPostconditions(arcNstartPostcond, this);
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return;
    }
}

void GoArc::_call_class_reset_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_reset_postcond)
	    mc->CallPostconditions(arcNresetPostcond, this);
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return;
    }
}

void GoArc::_call_class_pie_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_pie_postcond)
	    mc->CallPostconditions(arcNpiePostcond, this);
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return;
    }
}

void GoArc::_call_class_closed_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoArcMetaclass*) mc)->_has_closed_postcond)
	    mc->CallPostconditions(arcNclosedPostcond, this);
	// stop when reaching GoArc Metaclass instance;
	if (mc == GoArc::metaclass())
	    return;
    }
}

INLINE void GoArc::_call_radiusX_postcond() {
    if (_has_radiusX_postcond)
	CallPostconditions(arcNradiusXPostcond);
}

INLINE void GoArc::_call_radiusY_postcond() {
    if (_has_radiusY_postcond)
	CallPostconditions(arcNradiusYPostcond);
}

INLINE void GoArc::_call_amplitude_postcond() {
    if (_has_amplitude_postcond)
	CallPostconditions(arcNamplitudePostcond);
}

INLINE void GoArc::_call_start_postcond() {
    if (_has_start_postcond)
	CallPostconditions(arcNstartPostcond);
}

INLINE void GoArc::_call_reset_postcond() {
    if (_has_reset_postcond)
	CallPostconditions(arcNresetPostcond);
}

INLINE void GoArc::_call_pie_postcond() {
    if (_has_pie_postcond)
	CallPostconditions(arcNpiePostcond);
}

INLINE void GoArc::_call_closed_postcond() {
    if (_has_closed_postcond)
	CallPostconditions(arcNclosedPostcond);
}

void GoArcMetaclass::radiusXPrecond(ARCN_PRECOND_PTR(f),
				    void* closure, GAddOrRemove add) {
    _has_radiusX_precond =
	AddOrRemoveCondition(arcNradiusXPrecond, f, closure, add);
}

void GoArcMetaclass::amplitudePrecond(ARCN_PRECOND_PTR(f),
				      void* closure, GAddOrRemove add) {
    _has_amplitude_precond =
	AddOrRemoveCondition(arcNamplitudePrecond, f, closure, add);
}

void GoArcMetaclass::startPrecond(ARCN_PRECOND_PTR(f),
				  void* closure, GAddOrRemove add) {
    _has_start_precond =
	AddOrRemoveCondition(arcNstartPrecond, f, closure, add);
}

void GoArcMetaclass::radiusYPrecond(ARCN_PRECOND_PTR(f),
				    void* closure, GAddOrRemove add) {
    _has_radiusY_precond =
	AddOrRemoveCondition(arcNradiusYPrecond, f, closure, add);
}

void GoArcMetaclass::resetPrecond(ARCR_PRECOND_PTR(f),
				  void* closure, GAddOrRemove add) {
    _has_reset_precond =
	AddOrRemoveCondition(arcNresetPrecond, f, closure, add);
}

void GoArcMetaclass::piePrecond(ARCB_PRECOND_PTR(f),
				void* closure, GAddOrRemove add) {
    _has_pie_precond =
	AddOrRemoveCondition(arcNpiePrecond, f, closure, add);
}

void GoArcMetaclass::closedPrecond(ARCB_PRECOND_PTR(f),
				   void* closure, GAddOrRemove add) {
    _has_closed_precond =
	AddOrRemoveCondition(arcNclosedPrecond, f, closure, add);
}

void GoArcMetaclass::radiusXPostcond(ARC_POSTCOND_PTR(f),
				     void* closure, GAddOrRemove add) {
    _has_radiusX_postcond =
	AddOrRemoveCondition(arcNradiusXPostcond, f, closure, add);
}

void GoArcMetaclass::radiusYPostcond(ARC_POSTCOND_PTR(f),
				     void* closure, GAddOrRemove add) {
    _has_radiusY_postcond =
	AddOrRemoveCondition(arcNradiusYPostcond, f, closure, add);
}

void GoArcMetaclass::amplitudePostcond(ARC_POSTCOND_PTR(f),
				       void* closure, GAddOrRemove add) {
    _has_amplitude_postcond =
	AddOrRemoveCondition(arcNamplitudePostcond, f, closure, add);
}

void GoArcMetaclass::startPostcond(ARC_POSTCOND_PTR(f),
				   void* closure, GAddOrRemove add) {
    _has_start_postcond =
	AddOrRemoveCondition(arcNstartPostcond, f, closure, add);
}

void GoArcMetaclass::resetPostcond(ARC_POSTCOND_PTR(f),
				   void* closure, GAddOrRemove add) {
    _has_reset_postcond =
	AddOrRemoveCondition(arcNresetPostcond, f, closure, add);
}

void GoArcMetaclass::piePostcond(ARC_POSTCOND_PTR(f),
				 void* closure, GAddOrRemove add) {
    _has_pie_postcond =
	AddOrRemoveCondition(arcNpiePostcond, f, closure, add);
}

void GoArcMetaclass::closedPostcond(ARC_POSTCOND_PTR(f),
				    void* closure, GAddOrRemove add) {
    _has_closed_postcond =
	AddOrRemoveCondition(arcNclosedPostcond, f, closure, add);
}

void GoArc::radiusXPrecond(ARCN_PRECOND_PTR(f),
			   void* closure, GAddOrRemove add) {
    _has_radiusX_precond =
	AddOrRemoveCondition(arcNradiusXPrecond, f, closure, add);
}

void GoArc::radiusYPrecond(ARCN_PRECOND_PTR(f),
			   void* closure, GAddOrRemove add) {
    _has_radiusY_precond =
	AddOrRemoveCondition(arcNradiusYPrecond, f, closure, add);
}

void GoArc::amplitudePrecond(ARCN_PRECOND_PTR(f),
			     void* closure, GAddOrRemove add) {
    _has_amplitude_precond =
	AddOrRemoveCondition(arcNamplitudePrecond, f, closure, add);
}

void GoArc::startPrecond(ARCN_PRECOND_PTR(f), void* closure, GAddOrRemove add) {
    _has_start_precond =
	AddOrRemoveCondition(arcNstartPrecond, f, closure, add);
}

void GoArc::resetPrecond(ARCR_PRECOND_PTR(f), void* closure, GAddOrRemove add) {
    _has_reset_precond =
	AddOrRemoveCondition(arcNresetPrecond, f, closure, add);
}

void GoArc::piePrecond(ARCB_PRECOND_PTR(f), void* closure, GAddOrRemove add) {
    _has_pie_precond =
	AddOrRemoveCondition(arcNpiePrecond, f, closure, add);
}

void GoArc::closedPrecond(ARCB_PRECOND_PTR(f), void* closure, GAddOrRemove add) {
    _has_closed_precond =
	AddOrRemoveCondition(arcNclosedPrecond, f, closure, add);
}

void GoArc::radiusXPostcond(ARC_POSTCOND_PTR(f),
			    void* closure, GAddOrRemove add) {
    _has_radiusX_postcond =
	AddOrRemoveCondition(arcNradiusXPostcond, f, closure, add);
}

void GoArc::radiusYPostcond(ARC_POSTCOND_PTR(f),
			    void* closure, GAddOrRemove add) {
    _has_radiusY_postcond =
	AddOrRemoveCondition(arcNradiusYPostcond, f, closure, add);
}

void GoArc::amplitudePostcond(ARC_POSTCOND_PTR(f),
			      void* closure, GAddOrRemove add) {
    _has_amplitude_postcond =
	AddOrRemoveCondition(arcNamplitudePostcond, f, closure, add);
}

void GoArc::startPostcond(ARC_POSTCOND_PTR(f), void* closure, GAddOrRemove add) {
    _has_start_postcond =
	AddOrRemoveCondition(arcNstartPostcond, f, closure, add);
}

void GoArc::resetPostcond(ARC_POSTCOND_PTR(f), void* closure, GAddOrRemove add) {
    _has_reset_postcond =
	AddOrRemoveCondition(arcNresetPostcond, f, closure, add);
}

void GoArc::piePostcond(ARC_POSTCOND_PTR(f), void* closure, GAddOrRemove add) {
    _has_pie_postcond =
	AddOrRemoveCondition(arcNpiePostcond, f, closure, add);
}

void GoArc::closedPostcond(ARC_POSTCOND_PTR(f), void* closure, GAddOrRemove add){
    _has_closed_postcond =
	AddOrRemoveCondition(arcNclosedPostcond, f, closure, add);
}


/*
 *------------
 * CLASS PART
 *------------
 */



static GoArcMetaclass* arc_metaclass = NULL;

/**
FUNCTION
    GoArcMetaclass* GoArc::metaclass()
DESCRIPTION
    
**/
GoArcMetaclass* GoArc::metaclass() {
    if (arc_metaclass == NULL) {
	arc_metaclass = new GoArcMetaclass;
	arc_metaclass->superclass()->AddSubclass(arc_metaclass);

        GoStream::bind(GoArc::metaclass()->classname(),
		       (int (*) (const void*, GoStream*)) GoArc::save,
		       (int (*) (void**, GoStream*)) GoArc::restore);
    }
    return arc_metaclass;
}


/**
FUNCTION
    ~GoArcMetaclass()
DESCRIPTION
    
**/
GoArcMetaclass::~GoArcMetaclass() {
    if (this == arc_metaclass) {
        superclass()->RemoveSubclass(this);
	arc_metaclass = NULL;
    }
}


/**
FUNCTION
    GoArcMetaclass::GoArcMetaclass()
DESCRIPTION
    GoArcMetaclass constructor
**/
GoArcMetaclass::GoArcMetaclass() {
    _has_radiusX_precond  = _has_radiusX_postcond  = _has_radiusY_precond    =
    _has_radiusY_postcond = _has_amplitude_precond = _has_amplitude_postcond =
    _has_start_precond    = _has_start_postcond    = _has_reset_precond      =
    _has_reset_postcond   = _has_closed_postcond   = _has_pie_precond = GFALSE;
}


/**
FUNCTION
    const char* GoArcMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoArcMetaclass::m_classname() const {
    return "GoArc";
}


/**
FUNCTION
    GoMetaclass* GoArcMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoArcMetaclass::m_superclass() const {
    return GoGeometric::metaclass();
}


/**
FUNCTION
    const GActionRec* GoArcMetaclass::m_actions () const
DESCRIPTION
    
**/
const GActionRec* GoArcMetaclass::m_actions () const {
    if ((GoMetaclass*) this != GoArc::metaclass()) {
#ifdef DEBUG
	GWarning("the %s class has no \"m_actions()\" method !", classname());
#endif /* DEBUG */
	return empty_action_array;
    }
    return go_arc_action_array;
}


/**
FUNCTION
    Go* GoArcMetaclass::m_create();
DESCRIPTION
    virtual private
**/
Go* GoArcMetaclass::m_create() const {
    return (Go*) (new GoArc);
}



/*
 *-------------------
 * PRIVATE FUNCTIONS
 *-------------------
 */


/**
FUNCTION
    GoMetaclass* GoArc::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoArc::m_metaclass() const {
    return GoArc::metaclass();
}


/**
FUNCTION
    void GoArc::m_check() const
DESCRIPTION
    virtual private
**/
void GoArc::m_check() const {}


/*
 *--------------------
 * PROTECTED FUNCTIONS
 *--------------------
 */

/**
FUNCTION
    int GoArc::save(const GoArc* arc, GoStream* gostream);
DESCRIPTION
    **static member**, protected
**/
int GoArc::save(const GoArc* arc, GoStream* gostream) {
    return (gostream->writeNumber(arc->radiusX(), "radiusX") &&
	    gostream->writeNumber(arc->radiusY(), "radiusY") &&
	    gostream->writeNumber(arc->amplitude(), "amplitude") &&
	    gostream->writeNumber(arc->start(), "start") &&
	    gostream->writeBool(arc->pie(), "pie") &&
	    gostream->writeBool(arc->closed(), "closed") &&
	    GoGeometric::save(arc, gostream));
}

/**
FUNCTION
    int GoArc::restore(GoArc** arc, GoStream* gostream);
DESCRIPTION
    **static member**, protected
**/
int GoArc::restore(GoArc** arc, GoStream* gostream) {
    GNumber radx, rady, amplitude, start;
    if (! (gostream->readNumber(&radx) && gostream->readNumber(&rady) &&
	   gostream->readNumber(&amplitude) && gostream->readNumber(&start)))
	return 0;
    if (*arc = GoArc::create(radx, rady, amplitude, start)) {
	// Accept old streams (that do not contain _pie and _closed parameters;
	GBool is_pie, is_closed;
	if (gostream->readBool(&is_pie) && is_pie)
	    (*arc)->_pie = GTRUE;
	if (gostream->readBool(&is_closed) && is_closed)
	    (*arc)->_closed = GTRUE;

	return GoGeometric::readParams(*arc, gostream);
    }
    return 0;
}

/**
FUNCTION
    int GoArc::readParams(GoArc* arc, GoStream* gostream);
DESCRIPTION
    **static member**, protected
**/
int GoArc::readParams(GoArc* arc, GoStream* gostream) {
    GNumber radx, rady, amplitude, start;
    if (! (gostream->readNumber(&radx) && gostream->readNumber(&rady) &&
	   gostream->readNumber(&amplitude) && gostream->readNumber(&start)))
	return 0;
    best_values(&radx, &rady, &amplitude, &start, "readParams");

    // Accept old streams (that do not contain _pie and _closed parameters);
    GBool is_pie, is_closed;
    if (gostream->readBool(&is_pie) && is_pie)
	arc->_pie = GTRUE;
    if (gostream->readBool(&is_closed) && is_closed)
	arc->_closed = GTRUE;
    
    arc->_reset(radx, rady, amplitude, start);
    return GoGeometric::readParams(arc, gostream);
}


/**
FUNCTION
    GBool GoArc::m_contains(const GPoint& absPoint, GNumber maxDist) const
DESCRIPTION
    virtual protected
**/
GBool GoArc::m_contains(const GPoint& absPoint, GNumber maxDist) const {
    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;
    
    if (! (shbrush || shpen))
	return GFALSE;

    GNumber inf, sup;
    GCoordSys cs = Abscs();
    
    GOrientedAngles(_amplitude, _start, &inf, &sup);

    if (_pie)
	return interf()->ellipsePieContains(shpen, shbrush, cs, GPoint(0, 0),
					    _radX, _radY, inf, sup, _closed,
					    absPoint, maxDist);
    else return interf()->ellipseChordContains(shpen, shbrush, cs, GPoint(0, 0),
					       _radX, _radY, inf, sup, _closed,
					       absPoint, maxDist);
}


/**
FUNCTION
    GoArc::m_add_clip(GRegion& region) const
DESCRIPTION
    virtual protected
**/
void GoArc::m_add_clip(GRegion& region) const {
    if (! (outline() || background()))
	return;

    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;
    GNumber inf, sup;

    GOrientedAngles(_amplitude, _start, &inf, &sup);

    if (_pie)
	interf()->ellipsePieClips(shpen, shbrush, Abscs(), GPoint(0, 0), _radX,
				  _radY, inf, sup, _closed, &region, NULL);

    else interf()->ellipseChordClips(shpen, shbrush, Abscs(), GPoint(0, 0),
				     _radX, _radY, inf, sup, _closed,
				     &region, NULL);
}


/**
FUNCTION
    GoArc::m_add_opaque_clip(GRegion& region) const
DESCRIPTION
    virtual protected
**/
void GoArc::m_add_opaque_clip(GRegion& region) const {
    if (! (outline() || background()))
	return;

    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;
    GNumber inf, sup;

    GOrientedAngles(_amplitude, _start, &inf, &sup);

    if (_pie)
	interf()->ellipsePieClips(shpen, shbrush, Abscs(), GPoint(0, 0), _radX,
				  _radY, inf, sup, _closed, NULL, &region);
    else interf()->ellipseChordClips(shpen, shbrush, Abscs(), GPoint(0, 0),
				     _radX, _radY, inf, sup, _closed,
				     NULL, &region);
}


/**
FUNCTION
    void GoArc::m_draw(const GRegion&) const
DESCRIPTION
    virtual protected
**/
void GoArc::m_draw(const GRegion&) const {
    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;
    
    if (! (shbrush || shpen))
	return;

    GNumber inf, sup;
    GOrientedAngles(_amplitude, _start, &inf, &sup);

    if (_pie)
	interf()->ellipsePie(shpen, shbrush, Abscs(),
			     GPoint(0, 0), _radX, _radY, inf, sup, _closed);
    else interf()->ellipseChord(shpen, shbrush, Abscs(),
				GPoint(0, 0), _radX, _radY, inf, sup, _closed);
}



/**
FUNCTION
    void GoArc::m_wired_draw(const GCoordSys& cs, const GPen&, Quality) const
DESCRIPTION
    virtual protected
**/
void GoArc::m_wired_draw(const GCoordSys& cs, const GPen& pen,
			 GQuality quality) const
{
    GNumber inf, sup;
    GOrientedAngles(_amplitude, _start, &inf, &sup);

    if (_pie)
	interf()->ellipsePie(&pen, NULL, cs, GPoint(0, 0),
			     _radX, _radY, inf, sup, _closed, quality);
    else
	interf()->ellipseChord(&pen, NULL, cs, GPoint(0, 0),
			       _radX, _radY, inf, sup, _closed, quality);
}


/**
FUNCTION
    void GoArc::m_dup(const Go* src, Go*, Go*, GBool);
DESCRIPTION
    virtual private
**/
void GoArc::m_dup
    (const Go* src, Go* parcopy, Go* prevcopy, GBool donext, GBool dochildren)
{
    GoGeometric::m_dup(src, parcopy, prevcopy, donext, dochildren);
      
    const GoArc *arc = (const GoArc*) src;
    _reset(arc->radiusX(), arc->radiusY(), arc->amplitude(), arc->start());
    _pie    = arc->_pie;
    _closed = arc->_closed;
}

GBool GoArc::_set_pie(GBool value) {
    if ((value == 0) != (_pie == 0)) {

	GBool visible = outline() || background();
	GBool was_mapped = mapped() && visible;
	
	if (was_mapped)
	    RedrawBackAndFore();

	_pie = value != 0;

	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}


GBool GoArc::_set_closed(GBool value) {
    if ((value == 0) != (_closed == 0)) {

	GBool visible_change = outline(); // no effect if no outline;
	GBool draw_change    = mapped() && visible_change;

	if (draw_change)
	    RedrawBackAndFore();

	_closed = value != 0;
	
	if (visible_change)
	    MyClipHasChanged();
	if (draw_change)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}

/**
FUNCTION
    GoArc::Reset(GNumber, GNumber, GNumber, GNumber);
DESCRIPTION
**/
GBool GoArc::Reset(GNumber radX, GNumber radY, GNumber amplitude, GNumber start) {
    GBool redraw_back = GFALSE;
    GBool redraw_fore = GFALSE;
    GBool changed     = GFALSE;
    GBool was_mapped  = mapped() && (background() || outline());
    
    if (! check_values(radX, radY, amplitude, start, "Reset"))
	return GFALSE;
    
    if (start != _start)
	start = GModulo(start);

    if (radX != _radX || radY != _radY) {
	// Common sense would say that increasing a radius will strictly..;
	// increase the area covered by the arc and that a RedrawForeOnly..;
	// would be sufficient. However this, which is true for a circular..;
	// arc, is wrong in the general case of an elliptical arc and areas..;
	// previously hidden by the arc might be uncovered.
	changed = redraw_back = redraw_fore = GTRUE;
    }
    
    if (amplitude != _amplitude || start != _start) {
	GNumber inf, sup, _inf, _sup;
	changed = GTRUE;
	GOrientedAngles(amplitude, start, &inf, &sup);
	GOrientedAngles(_amplitude, _start, &_inf, &_sup);
	redraw_fore = (!background()) || redraw_fore ||
	    inf < _inf ||  sup > _sup;
	redraw_back = (!background()) || redraw_back ||
	    _inf <  inf || _sup >  sup;
    }
    
    if (changed) {
	if (was_mapped && redraw_back)
	    RedrawBackAndFore();
	_reset(radX, radY, amplitude, start);
	MyClipHasChanged();
	if (was_mapped && redraw_fore)
	    RedrawForeOnly();
	return GTRUE;
    }
    return GFALSE;
}

/*
 *-----------------
 * PUBLIC FUNCTIONS
 *-----------------
 */

/**
FUNCTION
    GoArc::GoArc(GNumber radiusX = 1, GNumber radiusY = 1,
                 GNumber amplitude = 2 * PI, GNumber start = 0)
DESCRIPTION
    constructor
**/
GoArc::GoArc(GNumber radiusX, GNumber radiusY,
	     GNumber amplitude, GNumber start) {

    _has_radiusX_precond  = _has_radiusX_postcond  = _has_radiusY_precond    =
    _has_radiusY_postcond = _has_amplitude_precond = _has_amplitude_postcond =
    _has_start_precond    = _has_start_postcond    = _has_reset_precond      =
    _has_reset_postcond   = _has_closed_precond    = _has_pie_postcond = GFALSE;

    best_values((GNumber*) &radiusX, (GNumber*) &radiusY,
		(GNumber*) &amplitude, (GNumber*) &start, "GoArc");
    _radX      = radiusX;
    _radY      = radiusY;
    _amplitude = amplitude;
    _start     = start;

    _pie = _closed = GFALSE;

//    if (pen().joinStyle() != GJOIN_ROUND)
//	pen().joinStyle(GJOIN_ROUND);
}


/**
FUNCTION
    GoArc* GoArc::create(GNumber radiusX, GNumber radiusY = 1.,
                         GNumber amplitude = 2*PI, GNumber start = 0.);
DESCRIPTION
    virtual private
**/
GoArc* GoArc::create(GNumber radiusX, GNumber radiusY,
		     GNumber amplitude, GNumber start) {
    GoArc *arc = (GoArc*) Go::Create(GoArc::metaclass());
    if (arc) {
	best_values((GNumber*) &radiusX, (GNumber*) &radiusY,
		    (GNumber*) &amplitude, (GNumber*) &start, "create");
	arc->_reset(radiusX, radiusY, amplitude, start);
    }
    return arc;
}


/**
FUNCTION
    GBool GoArc::start(GNumber val)
DESCRIPTION
    virtual public
    If preconditions allow it, sets start to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoArc::start(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (value != _start             &&
	_call_start_precond(&value) &&
	_call_class_start_precond(&value) &&
	(value != _start)           &&
	Reset(_radX, _radY, _amplitude, value)) {

	_call_start_postcond();
	_call_class_start_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoArc::amplitude(GNumber val)
DESCRIPTION
    virtual public
    If preconditions allow it, sets amplitude to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoArc::amplitude(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (value != _amplitude                   &&
	_call_amplitude_precond(&value)       &&
	_call_class_amplitude_precond(&value) &&
	(value != _amplitude)                 &&
	Reset(_radX, _radY, value, _start)) {

	_call_amplitude_postcond();
	_call_class_amplitude_postcond();
	return GTRUE;
    }
    return GFALSE;
}

/**
FUNCTION
    GBool GoArc::radiusX(GNumber val)
DESCRIPTION
    virtual public
    If preconditions allow it, sets radiusX to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoArc::radiusX(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (value >= 0 && value != _radX        &&
	_call_radiusX_precond(&value)       &&
	_call_class_radiusX_precond(&value) &&
	value >= 0 && value != _radX        &&
	Reset(value, _radY, _amplitude, _start)) {
	
	_call_radiusX_postcond();
	_call_class_radiusX_postcond();
	return GTRUE;
    }
    return GFALSE;
}

/**
FUNCTION
    GBool GoArc::radiusY(GNumber val)
DESCRIPTION
    virtual public
    If preconditions allow it, sets radiusY to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoArc::radiusY(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (value >= 0 && value != _radY        &&
	_call_radiusY_precond(&value)       &&
	_call_class_radiusY_precond(&value) &&
	value >= 0 && value != _radY        &&
	Reset(_radX, value, _amplitude, _start)) {

	_call_radiusY_postcond();
	_call_class_radiusY_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoArc::reset(GNumber rx_val, GNumber ry_val, GNumber a_val, GNumber s_val)
DESCRIPTION
    virtual public
    If preconditions allow it, sets radiusX, radiusX, amplitude and start to
    rx_val, ry_val, a_val and s_val (possibly modified by preconds)
    and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoArc::reset(GNumber rx_val, GNumber ry_val, GNumber a_val, GNumber s_val) {
    // do not trust parameter storage !;
    GNumber radX = rx_val, radY = ry_val, ampl = a_val, strt = s_val;

    if (radX >= 0. && radY >= 0.				&&
	(radX != _radX      || radY != _radY ||
	 ampl != _amplitude || strt != _start)			&&
	_call_reset_precond(&radX, &radY, &ampl, &strt)		&&
	_call_class_reset_precond(&radX, &radY, &ampl, &strt)	&&
	radX >= 0. && radY >= 0.				&&
	(radX != _radX      || radY != _radY ||
	 ampl != _amplitude || strt != _start)			&&
	Reset(radX, radY, ampl, strt)) {

	_call_reset_postcond();
	_call_class_reset_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoArc::pie(GBool val)
DESCRIPTION
    virtual public
**/
GBool GoArc::pie(GBool val) {
    GBool value = val; // do not trust parameter storage !;
    if ((value == 0) != (_pie == 0)     &&
	_call_pie_precond(&value)       &&
	_call_class_pie_precond(&value) &&
	_set_pie(value)) {
	
	_call_pie_postcond();
	_call_class_pie_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoArc::closed(GBool val)
DESCRIPTION
    virtual public
**/
GBool GoArc::closed(GBool val) {
    GBool value = val; // do not trust parameter storage !;
    if ((value == 0) != (_closed == 0)     &&
	_call_closed_precond(&value)       &&
	_call_class_closed_precond(&value) &&
	_set_closed(value)) {

	_call_closed_postcond();
	_call_class_closed_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GoArc* GoArc::dupAsArc() const;
DESCRIPTION
    This function creates and returns a GoArc object
    identical to "this" IN GoArc SENSE ONLY
    i.e. the returned object IS a GoArc even if "this" class is a subclass
    of GoArc.
**/
GoArc* GoArc::dupAsArc() const {
    GoArc* copy = GoArc::create();
    if (copy != NULL)
	copy->m_dup(this, NULL, NULL, GFALSE, GFALSE);
    return copy;
}
