#ifdef REV_INFO
static char rcsid[] = "$Id: Beta.cc,v 1.3 1992/12/10 15:47:27 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Geometric
FILE:    Beta.cc
AUTHOR:  Gilles Benati
DATE:    September 20th, 1991

DESCRIPTION:

    See the Go/Beta.h header.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France

  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.

  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <assert.h>
#include <Go/Beta.h>
#include <Go/GoStream.h>


static char betaSplineNbiasPrecond[]  = "p";
static char betaSplineNbiasPostcond[] = "p";

// to clarify pre/postcond code;
#define BETA_PRECOND_PTR(f)  GBool f(void*, const GoBetaSpline*, GNumber*)
#define BETA_POSTCOND_PTR(f) void f(void*, GoBetaSpline*)


/*
 *--------
 * ACTIONS
 *--------
 */

static void set_bias(GoBetaSpline *beta, GNumber value)
{
    if (beta && beta->getMetaclass()->isSubclass(GoBetaSpline::metaclass()))
	beta->bias(value);
}

static GActionRec go_beta_spline_action_array[] = {
  { "bias(GNumber)",	(GActionProc) set_bias },
  { NULL, NULL }
};



/*
 *--------------------
 * PRE/POSTCONDITIONS
 *--------------------
 */


GBool GoBetaSpline::_call_class_bias_precond(GNumber* val) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoBetaSplineMetaclass*) mc)->_has_bias_precond &&
	    ! mc->CallPreconditions(betaSplineNbiasPrecond, this, val))
	    return GFALSE;
	// stop when reaching GoBetaSpline Metaclass instance;
	if (mc == GoBetaSpline::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

void GoBetaSpline::_call_class_bias_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoBetaSplineMetaclass*) mc)->_has_bias_postcond)
	    mc->CallPostconditions(betaSplineNbiasPostcond, this);
	// stop when reaching GoBetaSpline Metaclass instance;
	if (mc == GoBetaSpline::metaclass())
	    return;
    }
}

GBool GoBetaSpline::_call_bias_precond(GNumber* val) const {
    return(_has_bias_precond ?
	   CallPreconditions(betaSplineNbiasPrecond, val) : GTRUE);
}

INLINE void GoBetaSpline::_call_bias_postcond() {
    if (_has_bias_postcond)
	CallPostconditions(betaSplineNbiasPostcond);
}


void GoBetaSplineMetaclass::biasPrecond(BETA_PRECOND_PTR(f),
					void* closure, GAddOrRemove add) {
    _has_bias_precond =
	AddOrRemoveCondition(betaSplineNbiasPrecond, f, closure, add);
}

void GoBetaSplineMetaclass::biasPostcond(BETA_POSTCOND_PTR(f),
					 void* closure, GAddOrRemove add) {
    _has_bias_postcond =
	AddOrRemoveCondition(betaSplineNbiasPostcond, f, closure, add);
}


void GoBetaSpline::biasPrecond(BETA_PRECOND_PTR(f),
			       void* closure, GAddOrRemove add) {
    _has_bias_precond =
	AddOrRemoveCondition(betaSplineNbiasPrecond, f, closure, add);
}


void GoBetaSpline::biasPostcond(BETA_POSTCOND_PTR(f),
				void* closure, GAddOrRemove add) {
    _has_bias_postcond =
	AddOrRemoveCondition(betaSplineNbiasPostcond, f, closure, add);
}


/*
 *------------
 * CLASS PART
 *------------
 */

static GoBetaSplineMetaclass* beta_spline_metaclass = NULL;

/**
FUNCTION
    GoBetaSplineMetaclass* GoBetaSpline::metaclass()
DESCRIPTION
    
**/
GoBetaSplineMetaclass* GoBetaSpline::metaclass() {
    if (beta_spline_metaclass == NULL) {
	beta_spline_metaclass = new GoBetaSplineMetaclass;
	beta_spline_metaclass->superclass()->AddSubclass(beta_spline_metaclass);

        GoStream::bind(GoBetaSpline::metaclass()->classname(),
		       (int (*) (const void*, GoStream*)) GoBetaSpline::save,
		       (int (*) (void**, GoStream*)) GoBetaSpline::restore);
    }
    return beta_spline_metaclass;
}


/**
FUNCTION
    GoBetaSplineMetaclass()
DESCRIPTION
    GoBetaSplineMetaclass constructor
**/
GoBetaSplineMetaclass::GoBetaSplineMetaclass() {
    _has_bias_precond = _has_bias_postcond = GFALSE;
}


/**
FUNCTION
    ~GoBetaSplineMetaclass()
DESCRIPTION
    
**/
GoBetaSplineMetaclass::~GoBetaSplineMetaclass() {
    if (this == beta_spline_metaclass) {
        superclass()->RemoveSubclass(this);
	beta_spline_metaclass = NULL;
    }
}


/**
FUNCTION
    const char* GoBetaSplineMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoBetaSplineMetaclass::m_classname() const {
    return "GoBetaSpline";
}


/**
FUNCTION
    GoMetaclass* GoBetaSplineMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoBetaSplineMetaclass::m_superclass() const {
    return GoCubic::metaclass();
}


/**
FUNCTION
    Go* GoBetaSplineMetaclass::m_create();
DESCRIPTION
    virtual private
**/
Go* GoBetaSplineMetaclass::m_create() const {
    return (Go*) (new GoBetaSpline);
}



/**
FUNCTION
    const GActionRec* GoBetaSplineMetaclass::m_actions () const
DESCRIPTION
    
**/
const GActionRec* GoBetaSplineMetaclass::m_actions () const {
    if ((GoMetaclass*) this != GoBetaSpline::metaclass()) {
#ifdef DEBUG
	GWarning("the %s class has no \"m_actions()\" method !", classname());
#endif /* DEBUG */
	return empty_action_array;
    }
    return go_beta_spline_action_array;
}

/*
 *---------------------
 * PROTECTED FUNCTIONS
 *---------------------
 */

/**
FUNCTION
    void GoBetaSpline::m_dup(const Go* src, Go*, Go*, GBool donext);
DESCRIPTION
    **virtual protected**
**/
void GoBetaSpline::m_dup
    (const Go* src, Go* parcopy, Go* prevcopy, GBool donext, GBool dochildren)
{
    GoCubic::m_dup(src, parcopy, prevcopy, donext, dochildren);
    
    const GoBetaSpline *beta = (const GoBetaSpline*) src;
    _bias = beta->_bias;
}


/**
FUNCTION
    int GoBetaSpline::save(const GoBetaSpline* beta, GoStream* gostream);
DESCRIPTION
    **static member**
**/
int GoBetaSpline::save(const GoBetaSpline* beta, GoStream* gostream) {
    return (gostream->writeNumber(beta->bias(), "bias") &&
	    GoCubic::save(beta, gostream));
}


/**
FUNCTION
    int GoBetaSpline::restore(GoBetaSpline** beta, GoStream* gostream); 
DESCRIPTION
    **static member**
**/
int GoBetaSpline::restore(GoBetaSpline** beta, GoStream* gostream) {
    // Might be optimized;;
    if (*beta = GoBetaSpline::create())
	return GoBetaSpline::readParams(*beta, gostream);
    else return 0;
}

/**
FUNCTION
    int GoBetaSpline::readParams(GoBetaSpline* beta, GoStream* gostream);
DESCRIPTION
    
**/
int GoBetaSpline::readParams(GoBetaSpline* beta, GoStream* gostream) {
    GNumber value;
    if (! gostream->readNumber(&value))
	return 0;
    beta->_bias   = value;
    beta->_inited = GFALSE;
    return GoCubic::readParams(beta, gostream);
}


/**
FUNCTION
    void GoBetaSpline::initCoeffMatrix();
DESCRIPTION
    virtual protected
**/
void GoBetaSpline::initCoeffMatrix() {
    GNumber b = _bias;
    GNumber b2 = b *b;
    GNumber b3 = b2* b;
    GNumber divisor = _tension + 2*b3 +4 *(b2+b) +2;
    _coeffMatrix[0][0] = -2*b3/divisor;
    _coeffMatrix[0][1] = 2*(_tension + b3 + b2 +b)/divisor;
    _coeffMatrix[0][2] = -2*(_tension+b2 +b+1)/divisor;
    _coeffMatrix[0][3] = 2/divisor;
    _coeffMatrix[1][0] = 6 *b3/divisor;
    _coeffMatrix[1][1] = -3*(_tension +2* (b3 +b2) )/divisor;
    _coeffMatrix[1][2] = (3*_tension + 6*b2)/divisor;
    _coeffMatrix[1][3] = 0;
    _coeffMatrix[2][0] = -6*b3/divisor;
    _coeffMatrix[2][1] =  6*(b3-b)/divisor;
    _coeffMatrix[2][2] =  6*b/divisor;
    _coeffMatrix[2][3] =  0;
    _coeffMatrix[3][0] =  2*b3/divisor;
    _coeffMatrix[3][1] =  (_tension + 4*(b2+b))/divisor;
    _coeffMatrix[3][2] =  ((GNumber) 2.0) /divisor;
    _coeffMatrix[3][3] = 0;
    _inited = GTRUE;
}


/*
 *-------------------
 * PRIVATE FUNCTIONS
 *-------------------
 */

/**
FUNCTION
    GoMetaclass* GoBetaSpline::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoBetaSpline::m_metaclass() const {
    return GoBetaSpline::metaclass();
}


/**
FUNCTION
    void GoBetaSpline::m_check() const
DESCRIPTION
    virtual private
**/
void GoBetaSpline::m_check() const {}


GBool GoBetaSpline::_set_bias(GNumber bias_val) {
    if ((bias_val == 0) != (_bias == 0)) {
	GBool visible = outline() || background();
	GBool was_mapped = mapped() && visible;
	
	if (was_mapped)
	    RedrawBackAndFore();
	
	_bias   = bias_val;
	_inited = GFALSE;
	
	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();
	return GTRUE;
    }
    return GFALSE; 
}


/*
 *------------------
 * PUBLIC FUNCTIONS
 *------------------
 */

#define BETA_SPLINE_INIT_VALUES() { \
    _has_bias_precond = _has_bias_postcond = GFALSE; \
    _tension = 1; \
    _bias    = 0; \
    _inited  = GFALSE; \
}

/**
FUNCTION
    GoBetaSpline()
DESCRIPTION
    GoBetaSpline constructor. It creates a beta spline of tension 1.
**/
GoBetaSpline::GoBetaSpline() : GoCubic(0, NULL) {
    BETA_SPLINE_INIT_VALUES();
}


/**
FUNCTION
    GoBetaSpline(int nb, const GPoint* pts)
DESCRIPTION
    GoBetaSpline constructor. It creates a beta spline of tension 1.
**/
GoBetaSpline::GoBetaSpline(int nb, const GPoint* pts) : GoCubic(nb, pts) {
    BETA_SPLINE_INIT_VALUES();
}


/**
FUNCTION
    GoBetaSpline* GoBetaSpline::create(int nb, const GPoint *pts = NULL);
DESCRIPTION
    The other legitimate GoBetaSpline constructor : calls pre/postcond.
**/
GoBetaSpline* GoBetaSpline::create(int nb, const GPoint *pts) {
    GoBetaSpline *beta = (GoBetaSpline*) Go::Create(GoBetaSpline::metaclass());
    if (beta)
	beta->reset(nb, pts);
    return beta;
}

/**
FUNCTION
    GBool bias(GNumber val)
DESCRIPTION
    public virtual
    This function, after calling preconditions, changes the bias of the curve
    and call postconditions if the action has been performed correctly.
    It returns GFALSE if any preconditions or the action returned GFALSE.
**/
GBool GoBetaSpline::bias(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (_bias != value                   &&
	_call_bias_precond(&value)       &&
	_call_class_bias_precond(&value) &&
	_set_bias(value)) {

	_call_bias_postcond();
	_call_class_bias_postcond();
	return GTRUE;
    }
    return GFALSE;
}

