#ifdef REV_INFO
static char rcsid[] = "$Id: BSpline.cc,v 1.1.1.1 1992/03/18 10:14:13 uep Exp $";
#endif
/*****************************************************************************
MODULE:  Geometric
FILE:    BSpline.cc
AUTHOR:  Gilles Benati
DATE:    September 20th, 1990

DESCRIPTION :

	See the Go/BSpline.h header.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France

  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.

  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <assert.h>
#include <Go/GoStream.h>
#include <Go/BSpline.h>


/*
 *------------
 * CLASS PART
 *------------
 */

static GoBSplineMetaclass* bspline_metaclass = NULL;

/**
FUNCTION
    GoBSplineMetaclass* GoBSpline::metaclass()
DESCRIPTION
    
**/
GoBSplineMetaclass* GoBSpline::metaclass() {
    if (bspline_metaclass == NULL) {
	bspline_metaclass = new GoBSplineMetaclass;
	bspline_metaclass->superclass()->AddSubclass(bspline_metaclass);

        GoStream::bind(GoBSpline::metaclass()->classname(),
		       (int (*) (const void*, GoStream*)) GoBSpline::save,
		       (int (*) (void**, GoStream*)) GoBSpline::restore);
    }
    return bspline_metaclass;
}


/**
FUNCTION
    ~GoBSplineMetaclass()
DESCRIPTION
    
**/
GoBSplineMetaclass::~GoBSplineMetaclass() {
    if (this == bspline_metaclass) {
        superclass()->RemoveSubclass(this);
	bspline_metaclass = NULL;
    }
}


/**
FUNCTION
    const char* GoBSplineMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoBSplineMetaclass::m_classname() const {
    return "GoBSpline";
}


/**
FUNCTION
    GoMetaclass* GoBSplineMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoBSplineMetaclass::m_superclass() const {
    return GoCubic::metaclass();
}


/**
FUNCTION
    Go* GoBSplineMetaclass::m_create();
DESCRIPTION
    virtual private
**/
Go* GoBSplineMetaclass::m_create() const {
    return (Go*) (new GoBSpline);
}


/*
 *---------------------
 * PROTECTED FUNCTIONS
 *---------------------
 */

/**
FUNCTION
    int GoBSpline::save(const GoBSpline* bspline, GoStream* gostream);
DESCRIPTION
    **static member**
**/
int GoBSpline::save(const GoBSpline* bspline, GoStream* gostream) {
    return GoCubic::save(bspline, gostream);
}


/**
FUNCTION
    int GoBSpline::restore(GoBSpline** bspline, GoStream* gostream); 
DESCRIPTION
    **static member**
**/
int GoBSpline::restore(GoBSpline** bspline, GoStream* gostream) {
    // Might be optimized;;
    if (*bspline = GoBSpline::create())
	return GoCubic::readParams(*bspline, gostream);
    return 0;
}

/**
FUNCTION
    void GoBSpline::initCoeffMatrix();
DESCRIPTION
    virtual protected
**/
void GoBSpline::initCoeffMatrix() {
    GNumber tsixth = _tension / 6;
    
    _coeffMatrix[0][0] = -tsixth;
    _coeffMatrix[0][1] = 2 - ((GNumber) 1.5) * _tension;
    _coeffMatrix[0][2] = -_coeffMatrix[0][1];
    _coeffMatrix[0][3] = tsixth;
    _coeffMatrix[1][0] = ((GNumber) 0.5) * _tension;
    _coeffMatrix[1][1] = 2 * _tension - 3;
    _coeffMatrix[1][2] = 3 - ((GNumber) 2.5) * _tension;
    _coeffMatrix[1][3] = 0;
    _coeffMatrix[2][0] = -_coeffMatrix[1][0];
    _coeffMatrix[2][1] = 0;
    _coeffMatrix[2][2] = _coeffMatrix[1][0];
    _coeffMatrix[2][3] = 0;
    _coeffMatrix[3][0] = tsixth;
    _coeffMatrix[3][1] = 1 - 2 * tsixth;
    _coeffMatrix[3][2] = tsixth;
    _coeffMatrix[3][3] = 0;
    _inited = GTRUE;
}


/*
 *-------------------
 * PRIVATE FUNCTIONS
 *-------------------
 */

/**
FUNCTION
    GoMetaclass* GoBSpline::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoBSpline::m_metaclass() const {
    return GoBSpline::metaclass();
}


/**
FUNCTION
    void GoBSpline::m_check() const
DESCRIPTION
    virtual private
**/
void GoBSpline::m_check() const {}



/*
 *------------------
 * PUBLIC FUNCTIONS
 *------------------
 */

/**
FUNCTION
    GoBSpline(int n = 0; const GPoint* pts = NULL)
DESCRIPTION
    GoBSpline constructor. It creates a normal B-spline (i.e. tension = 1);
**/
GoBSpline::GoBSpline(int n, const GPoint* pts) : GoCubic(n, pts) {
    _tension = 1;
}


/**
FUNCTION
    GoBSpline* GoBSpline::create(int nb, const GPoint *points = NULL);
DESCRIPTION
    The other legitimate GoBSpline constructor : calls pre/postcond.
**/
GoBSpline* GoBSpline::create(int nb, const GPoint *points) {
    GoBSpline *bspline = (GoBSpline*) Go::Create(GoBSpline::metaclass());
    if (bspline)
	bspline->reset(nb, points);
    return bspline;
}







