#ifdef REV_INFO
static char rcsid[] = "$Id: Cubic.cc,v 1.4 1992/12/10 15:54:23 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Geometric
FILE:    Cubic.cc
AUTHOR:  Gilles Benati
DATE:    September 20th, 1991

DESCRIPTION:

    See Go/Cubic.h and Go/mCubic.h for description of GoCubic and
  GoCubicMetaclass classes.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France

  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.

  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <assert.h>
#include <math.h>
#include <Go/GoInterf.h>
#include <Go/Cubic.h>
#include <Go/GoStream.h>


static char cubicNtensionPrecond[]  = "p";
static char cubicNtensionPostcond[] = "p";

// to clarify pre/postcond code;
#define CUB_PRECOND_PTR(f)  GBool f(void*, const GoCubic*, GNumber*)
#define CUB_POSTCOND_PTR(f) void f(void*, GoCubic*)

/*
 *--------
 * ACTIONS
 *--------
 */

static void set_tension(GoCubic *cubic, GNumber value)
{
    if (cubic && cubic->getMetaclass()->isSubclass(GoCubic::metaclass()))
	cubic->tension(value);
}

static GActionRec go_cubic_action_array[] = {
  { "tension(GNumber)",	(GActionProc) set_tension },
  { NULL, NULL }
};



/*
 *--------------------
 * PRE/POSTCONDITIONS
 *--------------------
 */


GBool GoCubic::_call_class_tension_precond(GNumber* val) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoCubicMetaclass*) mc)->_has_tension_precond &&
	    ! mc->CallPreconditions(cubicNtensionPrecond, this, val))
	    return GFALSE;
	// stop when reaching GoCubic Metaclass instance;
	if (mc == GoCubic::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

void GoCubic::_call_class_tension_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoCubicMetaclass*) mc)->_has_tension_postcond)
	    CallPostconditions(cubicNtensionPostcond, this);
	// stop when reaching GoCubic Metaclass instance;
	if (mc == GoCubic::metaclass())
	    return;
    }
}

GBool GoCubic::_call_tension_precond(GNumber* val) const {
    return(_has_tension_precond ?
	   CallPreconditions(cubicNtensionPrecond, val) : GTRUE);
}


INLINE void GoCubic::_call_tension_postcond() {
    if (_has_tension_postcond)
	CallPostconditions(cubicNtensionPostcond);
}


void GoCubicMetaclass::tensionPrecond(CUB_PRECOND_PTR(f),
				      void* closure, GAddOrRemove add) {
    _has_tension_precond =
	AddOrRemoveCondition(cubicNtensionPrecond, f, closure, add);
}

void GoCubicMetaclass::tensionPostcond(CUB_POSTCOND_PTR(f),
				       void* closure, GAddOrRemove add) {
    _has_tension_postcond =
	AddOrRemoveCondition(cubicNtensionPostcond, f, closure, add);
}


void GoCubic::tensionPrecond(CUB_PRECOND_PTR(f),
			     void* closure, GAddOrRemove add) {
    _has_tension_precond =
	AddOrRemoveCondition(cubicNtensionPrecond, f, closure, add);
}


void GoCubic::tensionPostcond(CUB_POSTCOND_PTR(f),
			      void* closure, GAddOrRemove add) {
    _has_tension_postcond =
	AddOrRemoveCondition(cubicNtensionPostcond, f, closure, add);
}



/*
 *------------
 * CLASS PART
 *------------
 */

static GoCubicMetaclass* cubic_metaclass = NULL;

/**
FUNCTION
    GoCubicMetaclass* GoCubic::metaclass()
DESCRIPTION
    **static member**
**/
GoCubicMetaclass* GoCubic::metaclass() {
    if (cubic_metaclass == NULL) {
	cubic_metaclass = new GoCubicMetaclass;
	cubic_metaclass->superclass()->AddSubclass(cubic_metaclass);
    }
    return cubic_metaclass;
}


/**
FUNCTION
    ~GoCubicMetaclass()
DESCRIPTION
    
**/
GoCubicMetaclass::~GoCubicMetaclass() {
    if (this == GoCubic::metaclass()) {
        superclass()->RemoveSubclass(this);
	cubic_metaclass = NULL;
    }
}


/**
FUNCTION
    GoCubicMetaclass()
DESCRIPTION
    GoCubicMetaclass constructor
**/
GoCubicMetaclass::GoCubicMetaclass() {
    _has_tension_precond = _has_tension_postcond = GFALSE;
}


/**
FUNCTION
    const char* GoCubicMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoCubicMetaclass::m_classname() const {
    return "GoCubic";
}


/**
FUNCTION
    GoMetaclass* GoCubicMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoCubicMetaclass::m_superclass() const {
    return GoCurve::metaclass();
}


/**
FUNCTION
    const GActionRec* GoCubicMetaclass::m_actions () const
DESCRIPTION
    
**/
const GActionRec* GoCubicMetaclass::m_actions () const {
    if ((GoMetaclass*) this != GoCubic::metaclass()) {
#ifdef DEBUG
	GWarning("the %s class has no \"m_actions()\" method !", classname());
#endif /* DEBUG */
	return empty_action_array;
    }
    return go_cubic_action_array;
}



/*
 *--------------------
 * PROTECTED FUNCTIONS
 *--------------------
 */

/**
FUNCTION
    int GoCubic::evalShift() const;
DESCRIPTION
    virtual protected.
**/
int GoCubic::evalShift() const {
    return 1;
}


/**
FUNCTION
    GPoint* GoCubic::getCurvePoints()
DESCRIPTION
    virtual protected
**/
void GoCubic::initCoeffMatrix() {
    _inited = GTRUE;
}


void GoCubic::m_draw(const GRegion&) const {
    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;

    if (nbPoint() <= 3 || ! (shbrush || shpen))
	return;
    
    if (!_inited)
	((GoCubic*)this)->initCoeffMatrix();
    // dirty cast but, for efficiency sake, matrix is only updated when needed;
    // initCoeffMatrix() is not a const function even though m_draw is;

#if defined(__BORLANDC__) || defined(__MSC7__)
    interf()->cubicCurve(shpen, shbrush, Abscs(), nbPoint(), points(),
			 (const float(*)[4])_coeffMatrix, evalShift(), GMAX_QUALITY);
#else
    interf()->cubicCurve(shpen, shbrush, Abscs(), nbPoint(), points(),
			 _coeffMatrix, evalShift(), GMAX_QUALITY);
#endif
}


GPoint* GoCubic::getCurvePoints(int& np, int nb_ctrl,
				const GPoint* ctrl) const {
    if (!_inited)
	((GoCubic*)this)->initCoeffMatrix();
    // dirty cast but, for efficiency sake, matrix is only updated when needed;
    // initCoeffMatrix() is not a const function even though getCurvePoints is;
    
    int numseg = (nb_ctrl - 4 + evalShift()) / evalShift();
    
    if (numseg <= 0)
        return NULL;

    int j = 20;	 // default values for number of points per segment;
    if (np >0)
	j = np / numseg;

    np = j * numseg + 1;
    GPoint *drawPoints = new GPoint[np];
    
    int k = 0;
    GNumber dt = (GNumber) 1 / (GNumber) j;
    GNumber dt2 = dt * dt;
    GNumber dt3 = dt2 * dt;

    GNumber dfx = 0, dfy = 0;  // initialize to avoid compilation warning;
    GNumber df2x, df2y, df3x, df3y;

    int start_index = 0;
    for (int sn = 0; sn < numseg; sn++, start_index += evalShift()) {
	GNumber ax = 0, ay = 0, bx = 0, by = 0;
	GNumber cx = 0, cy = 0, dx = 0, dy = 0;

	const GPoint* pt = ctrl + start_index;
	for (int i = 0; i < 4; i++, pt++) {
	    ax += pt->x * _coeffMatrix[0][i];
	    ay += pt->y * _coeffMatrix[0][i];
	    bx += pt->x * _coeffMatrix[1][i];
	    by += pt->y * _coeffMatrix[1][i];
	    cx += pt->x * _coeffMatrix[2][i];
	    cy += pt->y * _coeffMatrix[2][i];
	    dx += pt->x * _coeffMatrix[3][i];
	    dy += pt->y * _coeffMatrix[3][i];
	}

	drawPoints[k] = GPoint(dx,dy);
	k++;
	dfx  = ax * dt3 + bx * dt2 + cx * dt;
        dfy  = ay * dt3 + by * dt2 + cy * dt;
	df3x = 6 * ax * dt3 ;
	df3y = 6 * ay * dt3 ;
	df2x = df3x + 2 * bx * dt2;
	df2y = df3y + 2 * by* dt2;

	for (i = 1; i < j; i++, k++) {
	    drawPoints[k] = drawPoints[k-1] + GPoint(dfx, dfy);
	    dfx  += df2x;
     	    dfy  += df2y;
	    df2x += df3x;
	    df2y += df3y;
	}
    }
    drawPoints[k] = drawPoints[k-1] + GPoint(dfx, dfy);
    return drawPoints;
}


/**
FUNCTION
    void GoCubic::m_dup(const Go* src, Go* parcopy, Go* prevcopy, GBool donext, GBool dochildren) {
DESCRIPTION
    **virtual protected**
**/
void GoCubic::m_dup
    (const Go* src, Go* parcopy, Go* prevcopy, GBool donext, GBool dochildren)
{
    GoCurve::m_dup(src, parcopy, prevcopy, donext, dochildren);
    
    const GoCubic *cubic = (const GoCubic*) src;
    _tension = cubic->_tension;
    _inited  = cubic->_inited;
    memcpy(_coeffMatrix, cubic->_coeffMatrix, 16 * sizeof(GNumber));
}


/*
 *-----------------
 * PUBLIC FUNCTIONS
 *-----------------
 */


/**
FUNCTION
    GBool tension(GNumber val)
DESCRIPTION
    public virtual
    This function changes the tension of the curve. 
    Returns GFALSE if one of the preconditions or the action returned GFALSE.
    Postconditions are only called if the action succeeded.
**/
GBool GoCubic::tension(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (_tension != value                   &&
	_call_tension_precond(&value)       &&
	_call_class_tension_precond(&value) &&
	_set_tension(value)) {
	
	_call_tension_postcond();
	_call_class_tension_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool delPoint(int index) 
DESCRIPTION
    virtual public.
    This function deletes the point at the specified index from the polypoint 
    array.
**/
GBool GoCubic::delPoint(int index) {
    // Cubic curves need at least 4 points;
    return (_nbPoint > 4) ? GoCurve::delPoint(index) : GFALSE;
}


/*
 *--------------------
 * PROTECTED FUNCTIONS
 *--------------------
 */

/**
FUNCTION
    GoCubic(int nb = 0, const GPoint* pts = NULL);
DESCRIPTION
    
**/
GoCubic::GoCubic(int nb, const GPoint* pts) : GoCurve(nb, pts) {
    _tension = 1;
    _inited  = GFALSE;
    _has_tension_precond = _has_tension_postcond = GFALSE;
}


/**
FUNCTION
    int GoCubic::save(const GoCubic* cubic, GoStream* gostream);
DESCRIPTION
    
**/
int GoCubic::save(const GoCubic* cubic, GoStream* gostream) {
    return (gostream->writeNumber(cubic->tension(), "tension") &&
	    GoCurve::save(cubic, gostream));
}


/**
FUNCTION
    int GoCubic::readParams(GoCubic* cubic, GoStream* gostream);
DESCRIPTION
    
**/
int GoCubic::readParams(GoCubic* cubic, GoStream* gostream) {
    GNumber tens;
    if (! gostream->readNumber(&tens))
	return 0;
    cubic->_inited  = GFALSE;
    cubic->_tension = tens;
    return GoCurve::readParams(cubic, gostream);
}

/*
 *-------------------
 * PRIVATE FUNCTIONS
 *-------------------
 */


/**
FUNCTION
    GoMetaclass* GoCubic::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoCubic::m_metaclass() const {
    return GoCubic::metaclass();
}


/**
FUNCTION
    void GoCubic::m_check() const
DESCRIPTION
    virtual private
**/
void GoCubic::m_check() const {}


GBool GoCubic::_set_tension(GNumber value) {
    if (_tension != value) {

	GBool visible    = outline() || background();
	GBool was_mapped = mapped()  && visible;

	if (was_mapped)
	    RedrawBackAndFore();

	_tension = value;
	_inited  = GFALSE;

	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}



