#ifdef REV_INFO
static char rcsid[] = "$Id: Curve.cc,v 1.2 1992/12/10 15:56:09 smadja Exp $";
#endif
/*****************************************************************************
MODULE: Graphic Objects
FILE:   Curve.cc
AUTHOR: Aboobacker
DATE:   Feb 1991

DESCRIPTION :
    See the Go/Tools/Curve.h header.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/

#include <assert.h>
#include <Go/Tools/Error.h>
#include <Go/GoInterf.h>
#include <Go/GoStream.h>
#include <Go/Curve.h>


/*
 *------------
 * CLASS PART
 *------------
 */

static GoCurveMetaclass* curve_metaclass = NULL;

/**
FUNCTION
    GoCurveMetaclass* GoCurve::metaclass()
DESCRIPTION
    ***static member***
**/
GoCurveMetaclass* GoCurve::metaclass() {
    if (curve_metaclass == NULL) {
	curve_metaclass = new GoCurveMetaclass;
	curve_metaclass->superclass()->AddSubclass(curve_metaclass);
    }
    return curve_metaclass;
}


/**
FUNCTION
    ~GoCurveMetaclass()
DESCRIPTION
    
**/
GoCurveMetaclass::~GoCurveMetaclass() {
    if (this == curve_metaclass) {
        superclass()->RemoveSubclass(this);
	curve_metaclass = NULL;
    }
}


/**
FUNCTION
    const char* GoCurveMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoCurveMetaclass::m_classname() const {
    return "GoCurve";
}


/**
FUNCTION
    GoMetaclass* GoCurveMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoCurveMetaclass::m_superclass() const {
    return GoPolypoint::metaclass();
}


/*
 *-------------------
 * PRIVATE FUNCTIONS
 *-------------------
 */

/**
FUNCTION
    GoMetaclass* GoCurve::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoCurve::m_metaclass() const {
    return GoCurve::metaclass();
}


/**
FUNCTION
    void GoCurve::m_check() const
DESCRIPTION
    virtual private
**/
void GoCurve::m_check() const {
}


/**
FUNCTION
    GBool GoCurve::m_contains(const GPoint& absPoint, GNumber maxDist) const
DESCRIPTION
    virtual private
    **** The curve is seen as a polyline.
    **** This is the DEFAULT BEHAVIOUR for all GoCurve subclasses.
**/
GBool GoCurve::m_contains(const GPoint& absPoint, GNumber maxDist) const {
    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;

    if (! (shbrush || shpen))
	return GFALSE;

    int nb = nbPoint() * 3;
    GPoint* pts = getCurvePoints(nb, nbPoint(), points());

    if (! pts) 
	return GFALSE;

    GBool answer = interf()->polylineContains(shpen, shbrush, Abscs(), nb, pts,
					      absPoint, maxDist);
    delete pts;
    return answer;
}


/**
FUNCTION
    GoCurve::m_draw(const GRegion& region) const
DESCRIPTION
    virtual private    
    **** The curve is seen as a polyline.
    **** This is the DEFAULT BEHAVIOUR for all GoCurve subclasses.
**/
void GoCurve::m_draw(const GRegion&) const {
    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;

    if (! (shbrush || shpen))
	return;

    int nb = -1;
    GPoint *pts = getCurvePoints(nb, nbPoint(), points());
	
    if (!pts)
	return;
	
    interf()->polyline(shpen, shbrush, Abscs(), nb, pts, GMAX_QUALITY);
    delete pts;
}


/**
FUNCTION
    GoCurve::m_wired_draw(const GCoordSys& cs, const GPen& pen, GQuality quality) const
DESCRIPTION
    **** The curve is seen as a polyline.
    **** This is the DEFAULT BEHAVIOUR for all GoCurve subclasses.
**/
void GoCurve::m_wired_draw(const GCoordSys& cs, const GPen& pen,
			   GQuality quality) const
{
    int nb = nbPoint() * 3;
    GPoint *pts = getCurvePoints(nb, nbPoint(), points());

    if (! pts)
	return;
    interf()->polyline(&pen, NULL, cs, nb, pts, quality);
    delete pts;
}



/*
 *--------------------
 * PROTECTED FUNCTIONS
 *--------------------
 */


/**
FUNCTION
    GoCurve::m_add_clip(GRegion& region) const
DESCRIPTION
    virtual protected
**/
void GoCurve::m_add_clip(GRegion& region) const {    
    if (! (outline() || background()))
	return;

    int nb = -1;
    GPoint *pts = getCurvePoints(nb, nbPoint(), points());

    if (!pts)
	return;

    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;

    interf()->polylineClips(shpen, shbrush, Abscs(), nb, pts, &region, NULL);
    delete pts;
}


/**
FUNCTION
    GoCurve::GoCurve(int n = 0, const GPoint* pts = NULL);
DESCRIPTION
    GoCurve constructor (protected)
**/
GoCurve::GoCurve(int n, const GPoint* pts) : GoPolypoint(n , pts) {
    if (pen().joinStyle() != GJOIN_ROUND)
	pen().joinStyle(GJOIN_ROUND);
}



/*
 *---------------
 * PUBLIC MEMBERS
 *---------------
 */

/**
FUNCTION
    GBool delPoint(int index) 
DESCRIPTION
    virtual public.
    This function deletes the point at the specified index from the polypoint 
    array.
**/
GBool GoCurve::delPoint(int index) {
    return (_nbPoint > 2) ? GoPolypoint::delPoint(index) : GFALSE;
}

