#ifdef REV_INFO
static char rcsid[] = "$Id: Geometric.cc,v 1.4 1992/12/10 15:58:55 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Graphic Objects
FILE:    Geometric.cc
AUTHOR:  Gilles Benati
DATE:    Feb 1991

DESCRIPTION:
    See the Geometric.h header.
    
COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/

#include <assert.h>
#include <Go/Tools/Error.h>
#include <Go/GoStream.h>
#include <Go/Geometric.h>


static char geometricNoutlinePrecond[]     = "p";
static char geometricNoutlinePostcond[]    = "p";

static char geometricNbackgroundPrecond[]  = "p";
static char geometricNbackgroundPostcond[] = "p";

// to clarify pre/postcond code;
#define GEOM_PRECOND_PTR(f)  GBool f(void*, const GoGeometric*, GBool*)
#define GEOM_POSTCOND_PTR(f) void f(void*, GoGeometric*)

/*
 *--------
 * ACTIONS
 *--------
 */

GActionRec empty_action_array[] = { { NULL, NULL } };

static void set_outline(GoGeometric *geom, GBool yes) {
    geom->outline(yes);
}

static void set_background(GoGeometric *geom, GBool yes) {
    geom->background(yes);
}

static GActionRec go_geometric_action_array[] = {
  { "outline(GBool)",     (GActionProc) set_outline },
  { "background(GBool)",  (GActionProc) set_background },
  { NULL, NULL }
};



/*
 *-------------------
 * PRE/POSTCONDITIONS
 *-------------------
 */


GBool GoGeometric::_call_class_outline_precond(GBool *value) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoGeometricMetaclass*) mc)->_has_outline_precond && 
	    ! mc->CallPreconditions(geometricNoutlinePrecond, this, value))
	    return GFALSE;
	// stop when reaching GoGeometric Metaclass instance;
	if (mc == GoGeometric::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}


void GoGeometric::_call_class_outline_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoGeometricMetaclass*) mc)->_has_outline_postcond)
	    mc->CallPostconditions(geometricNoutlinePostcond, this);
	// stop when reaching GoGeometric Metaclass instance;
	if (mc == GoGeometric::metaclass())
	    return;
    }
}


void GoGeometricMetaclass::outlinePrecond(GEOM_PRECOND_PTR(f),
					  void* closure, GAddOrRemove add) {
    _has_outline_precond =
	AddOrRemoveCondition(geometricNoutlinePrecond, f, closure, add);
}


void GoGeometricMetaclass::outlinePostcond(GEOM_POSTCOND_PTR(f),
					   void* closure, GAddOrRemove add) {
    _has_outline_postcond =
	AddOrRemoveCondition(geometricNoutlinePostcond, f, closure, add);
}


GBool GoGeometric::_call_class_background_precond(GBool *value) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoGeometricMetaclass*) mc)->_has_background_precond &&
	    ! mc->CallPreconditions(geometricNbackgroundPrecond, this, value))
	    return GFALSE;
	// stop when reaching GoGeometric Metaclass instance;
	if (mc == GoGeometric::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}


void GoGeometric::_call_class_background_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoGeometricMetaclass*) mc)->_has_background_postcond)
	    mc->CallPostconditions(geometricNbackgroundPostcond, this);
	// stop when reaching GoGeometric Metaclass instance;
	if (mc == GoGeometric::metaclass())
	    return;
    }
}


void
  GoGeometricMetaclass::backgroundPrecond(GEOM_PRECOND_PTR(f),
					  void* closure, GAddOrRemove add) {
    _has_background_precond =
	AddOrRemoveCondition(geometricNbackgroundPrecond, f, closure, add);
}


void GoGeometricMetaclass::backgroundPostcond(GEOM_POSTCOND_PTR(f),
					      void* closure, GAddOrRemove add) {
    _has_background_postcond =
	AddOrRemoveCondition(geometricNbackgroundPostcond, f, closure, add);
}


GBool GoGeometric::_call_outline_precond(GBool *value) const {
    return(_has_outline_precond	?
	   CallPreconditions(geometricNoutlinePrecond, value) : GTRUE);
}

INLINE void GoGeometric::_call_outline_postcond() {
    if (_has_outline_postcond)
	CallPostconditions(geometricNoutlinePostcond);
}


GBool GoGeometric::_call_background_precond(GBool *value) const {
    return(_has_background_precond ?
	   CallPreconditions(geometricNbackgroundPrecond, value) : GTRUE);
}

INLINE void GoGeometric::_call_background_postcond() {
    if (_has_background_postcond)
	CallPostconditions(geometricNbackgroundPostcond);
}



/*
 *------------
 * CLASS PART
 *------------
 */


static GoGeometricMetaclass* geometric_metaclass = NULL;

/**
FUNCTION
    GoGeometricMetaclass* GoGeometric::metaclass()
DESCRIPTION
    
**/
GoGeometricMetaclass* GoGeometric::metaclass() {
    if (geometric_metaclass == NULL) {
	geometric_metaclass = new GoGeometricMetaclass;
	geometric_metaclass->superclass()->AddSubclass(geometric_metaclass);
    }
    return geometric_metaclass;
}


/**
FUNCTION
    GoGeometricMetaclass::~GoGeometricMetaclass()
DESCRIPTION
    
**/
GoGeometricMetaclass::~GoGeometricMetaclass() {
    if (this == geometric_metaclass) {
        superclass()->RemoveSubclass(this);
	geometric_metaclass = NULL;
    }
}


/**
FUNCTION
    GoGeometricMetaclass
DESCRIPTION
    GoGeometricMetaclass
**/
GoGeometricMetaclass::GoGeometricMetaclass() {
    _has_outline_precond    = _has_outline_postcond    =
    _has_background_precond = _has_background_postcond = GFALSE;
}


/**
FUNCTION
    const char* GoGeometricMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoGeometricMetaclass::m_classname() const {
    return "GoGeometric";
}


/**
FUNCTION
    GoMetaclass* GoGeometricMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoGeometricMetaclass::m_superclass() const {
    return GoStdDS::metaclass();
}


/**
FUNCTION
    const GActionRec* GoGeometricMetaclass::m_actions () const
DESCRIPTION
    
**/
const GActionRec* GoGeometricMetaclass::m_actions () const {
    if ((GoMetaclass*) this != GoGeometric::metaclass()) {
#ifdef DEBUG
	GWarning("the %s class has no \"m_actions()\" method !", classname());
#endif /* DEBUG */
	return empty_action_array;
    }
    return go_geometric_action_array;
}


/*
 *-------------------
 * PRIVATE FUNCTIONS
 *-------------------
 */

/**
FUNCTION
    GoMetaclass* GoGeometric::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoGeometric::m_metaclass() const {
    return GoGeometric::metaclass();
}


/**
FUNCTION
    void GoGeometric::m_dup(const Go* src, Go* dst, Go*, Go*,GBool)
DESCRIPTION
    virtual private
**/
void GoGeometric::m_dup
    (const Go* src, Go* parcopy, Go* prevcopy, GBool donext, GBool dochildren)
{
    GoStdDS::m_dup(src, parcopy, prevcopy, donext, dochildren);

    const GoGeometric *geo_src = (const GoGeometric *) src;

    _outline    = geo_src->outline();
    _background = geo_src->background();
    pen().modify(geo_src->pen().shPen());
    brush().modify(geo_src->brush().shBrush());
}


/**
FUNCTION
    void GoGeometric::m_check() const
DESCRIPTION
    virtual private
**/
void GoGeometric::m_check() const {
}


GBool GoGeometric::_set_outline(GBool value) {
    if ((value != 0) != (_outline != 0)) {
	GBool was_mapped = mapped();

	if (was_mapped && ! value)
	    RedrawBackAndFore();
	
	_outline = (value != 0);
	MyClipHasChanged();

	if (was_mapped && value) {
	    RedrawForeOnly();
	}

	return GTRUE;
    }
    return GFALSE;
}

GBool GoGeometric::_set_background(GBool value) {
    if ((value != 0) != (_background != 0)) {
	GBool was_mapped = mapped();

	if (was_mapped && ! value)
	    RedrawBackAndFore();

	_background = (value != 0);
	MyClipHasChanged();

	if (was_mapped && value)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}
    

/**
FUNCTION
    GoGeometric::_pen_changed(GoGeometric* gogeom,
                              const GPen* pen, unsigned long mask)
DESCRIPTION
    *** static member, private ***
    This function is a modify_post_condition of gogeom->_pen
    WARNING: it is assumed that pen **IS** gogeom->_pen
**/

void GoGeometric::_pen_changed(GoGeometric* gogeom,
			       const GPen*, unsigned long mask) {
    if (! gogeom->_outline) {
	gogeom->init_sh_pen();
	return;
    }

    if (mask == GCOLOR || mask == GALT_COLOR || mask == (GCOLOR & GALT_COLOR)) {
	gogeom->init_sh_pen();
	gogeom->RedrawForeOnly();
    }
    else if (mask & (GLINE_WIDTH | GJOIN_STYLE | GMITER_LIMIT)) {
	gogeom->RedrawBackAndFore();
	gogeom->init_sh_pen();
	gogeom->MyClipHasChanged();
	gogeom->RedrawForeOnly();
    }
    else {
	gogeom->RedrawBackAndFore();
	gogeom->init_sh_pen();
    }
}


/**
FUNCTION
    GoGeometric::_brush_changed(GoGeometric* gogeom,
                                const GBrush* brush, unsigned long mask)
DESCRIPTION
    *** static member, private ***
    This function is a modify_post_condition of gogeom->_brush
    WARNING: it is assumed that brush **IS** gogeom->_brush
**/

void GoGeometric::_brush_changed(GoGeometric* gogeom,
				 const GBrush*, unsigned long) {
    if (gogeom->_background && gogeom->mapped())
	gogeom->RedrawBackAndFore();
}


/*
 *---------------------
 * PROTECTED FUNCTIONS
 *---------------------
 */

/**
FUNCTION
    int GoGeometric::save(const GoGeometric* geom, GoStream* gostream);
DESCRIPTION
    ** static member, protected **
**/
int GoGeometric::save(const GoGeometric* geom, GoStream* gostream) {
    assert(geom && gostream); // *** TO BE REMOVED WHEN SAFE ***;

    const GShPen* shpen = geom->pen().shPen();
    const GShBrush* shbrush = geom->brush().shBrush();
    return (gostream->writeBool(geom->outline(), "outline") &&
	    gostream->writeBool(geom->background(), "background") &&
	    gostream->saveObj(shpen, shpen->getClassname(), "pen") &&
	    gostream->saveObj(shbrush, shbrush->getClassname(), "brush") &&
	    GoStdDS::save(geom, gostream));
}


/**
FUNCTION
    int GoGeometric::readParams(GoGeometric* geom, GoStream* gostream);
DESCRIPTION
    
**/
int GoGeometric::readParams(GoGeometric* geom, GoStream* gostream) {
    GBool outline = GTRUE, background = GFALSE;
    GShPen* shpen = NULL;
    GShBrush* shbrush = NULL;
    const char* class_name;

    if (! (gostream->readBool(&outline) && gostream->readBool(&background)))
	return 0;
    if (! gostream->restoreObj((void**) &shpen, &class_name))
	return 0;
    if (strcmp(class_name, GShPen::classname()))
	GWarning("GoGeometric::restore: class %s where GShPen expected");
    
    if (! gostream->restoreObj((void**) &shbrush, &class_name))
	return 0;
    if (strcmp(class_name, GShBrush::classname()))
	GWarning("GoGeometric::restore: class %s where GShBrush expected");

    geom->_outline = outline;
    geom->_background = background;
    geom->pen().modify(shpen);
    geom->brush().modify(shbrush);
    return GoStdDS::readParams(geom, gostream);
}

/**
FUNCTION
    GoGeometric::GoGeometric()
DESCRIPTION
    
**/
GoGeometric::GoGeometric() {
    _has_outline_precond    = _has_outline_postcond    =
    _has_background_precond = _has_background_postcond = GFALSE;

    _outline    = GTRUE;
    _background = GFALSE;
    
    _pen.modifyPostcond((void (*)(void*, GPen*, ulong)) _pen_changed,
			this, GADD);
    _brush.modifyPostcond((void (*)(void*, GBrush*, ulong)) _brush_changed,
			  this, GADD);
    _old_sh_pen = NULL;
    init_sh_pen();
}

GoGeometric::~GoGeometric() {
    if (_old_sh_pen)
	    _old_sh_pen->destroy();
}


/*
 *------------------
 * PUBLIC FUNCTIONS
 *------------------
 */


/**
FUNCTION
    GoGeometric::outlinePrecond(GEOM_PRECOND_PTR(f),
                                void* closure, GAddOrRemove add)
DESCRIPTION
    
**/
void GoGeometric::outlinePrecond(GEOM_PRECOND_PTR(f),
				 void* closure, GAddOrRemove add) {
    _has_outline_precond =
	AddOrRemoveCondition(geometricNoutlinePrecond, f, closure, add);
}

/**
FUNCTION
    GoGeometric::outlinePostcond(GEOM_POSTCOND_PTR(f),
                                 void* closure, GAddOrRemove add)
DESCRIPTION
    
**/
void GoGeometric::outlinePostcond(GEOM_POSTCOND_PTR(f),
				  void* closure, GAddOrRemove add) {
    _has_outline_postcond =
	AddOrRemoveCondition(geometricNoutlinePostcond, f, closure, add);
}


/**
FUNCTION
    GoGeometric::backgroundPrecond(GEOM_PRECOND_PTR(f),
                                   void* closure, GAddOrRemove add)
DESCRIPTION
    
**/
void GoGeometric::backgroundPrecond(GEOM_PRECOND_PTR(f),
				    void* closure, GAddOrRemove add) {
    _has_background_precond =
	AddOrRemoveCondition(geometricNbackgroundPrecond, f, closure, add);
}

/**
FUNCTION
    GoGeometric::backgroundPostcond(GEOM_POSTCOND_PTR(f),
                                    void* closure, GAddOrRemove add)
DESCRIPTION
    
**/
void GoGeometric::backgroundPostcond(GEOM_POSTCOND_PTR(f),
				     void* closure, GAddOrRemove add) {
    _has_background_postcond =
	AddOrRemoveCondition(geometricNbackgroundPostcond, f, closure, add);
}


/**
FUNCTION
    void GoGeometric::outline(GBool val)
DESCRIPTION
    public virtual
    If preconditions allow it, sets outline to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns
    GFALSE or bad parameters);
**/
GBool GoGeometric::outline(GBool val) {
    GBool value = val; // do not trust parameter storage !;
    if ((value != 0) != (_outline != 0)     &&
	_call_outline_precond(&value)       &&
	_call_class_outline_precond(&value) &&
	_set_outline(value)) {

	_call_outline_postcond();
	_call_class_outline_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GoGeometric::background(GBool val)
DESCRIPTION
    public virtual
    If preconditions allow it, sets background to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoGeometric::background(GBool val) {
    GBool value = val; // do not trust parameter storage !;
    if ((value != 0) != (_background != 0)     &&
	_call_background_precond(&value)       &&
	_call_class_background_precond(&value) &&
	_set_background(value)) {

	_call_background_postcond();
	_call_class_background_postcond();
	return GTRUE;
    }
    return GFALSE;
}

