#ifdef REV_INFO
static char rcsid[] = "$Id: Polyline.cc,v 1.3 1992/12/10 16:02:22 smadja Exp $";
#endif
/*****************************************************************************
MODULE: Geometric Objects
FILE:   Polyline.cc
AUTHOR: Gilles Benati
DATE:   Feb 1991

DESCRIPTION :
    See the Go/Tools/Polyline.h header file

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/

#include <assert.h>
#include <Go/Tools/Error.h>
#include <Go/GoInterf.h>
#include <Go/GoStream.h>
#include <Go/Polyline.h>


#define SIN_11_DEG .1908089954

static char polylineNroundPrecond[]        = "p";
static char polylineNroundPostcond[]       = "p";
static char polylineNroundOffsetPrecond[]  = "p";
static char polylineNroundOffsetPostcond[] = "p";

// to clarify pre/postcond code;
#define PLR_PRECOND_PTR(f)   GBool f(void*, const GoPolyline*, GBool*)
#define PLRO_PRECOND_PTR(f)  GBool f(void*, const GoPolyline*, GNumber*)
#define PL_POSTCOND_PTR(f)   void f(void*, GoPolyline*)

/*
 *--------
 * ACTIONS
 *--------
 */

static void set_round(GoPolyline *polyline, GBool yes)
{
    if (polyline &&
	polyline->getMetaclass()->isSubclass(GoPolypoint::metaclass()))
	polyline->round(yes);
}

static void set_round_offset(GoPolyline *polyline, GNumber round_offset)
{
    if (polyline &&
	polyline->getMetaclass()->isSubclass(GoPolypoint::metaclass()))
	polyline->roundOffset(round_offset);
}

static GActionRec polyline_action_array[] = {
  { "round(GBool)",         (GActionProc) set_round },
  { "roundOffset(GNumber)", (GActionProc) set_round_offset },
  { NULL, NULL }
};



/*
 *--------------------
 * PRE/POSTCONDITIONS
 *--------------------
 */

GBool GoPolyline::_call_class_round_precond(GBool* yes) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoPolylineMetaclass*) mc)->_has_round_precond &&
	    ! mc->CallPreconditions(polylineNroundPrecond, this, yes))
	    return GFALSE;
	// stop when reaching GoPolyline Metaclass instance;
	if (mc == GoPolyline::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoPolyline::_call_class_round_offset_precond(GNumber* val) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoPolylineMetaclass*) mc)->_has_round_offset_precond &&
	    ! mc->CallPreconditions(polylineNroundOffsetPrecond, this, val))
	    return GFALSE;
	// stop when reaching GoPolyline Metaclass instance;
	if (mc == GoPolyline::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoPolyline::_call_round_precond(GBool *yes) const {
    return(_has_round_precond ?
	   CallPreconditions(polylineNroundPrecond, yes) : GTRUE);
}

GBool GoPolyline::_call_round_offset_precond(GNumber *val) const {
    return(_has_round_offset_precond ?
	   CallPreconditions(polylineNroundOffsetPrecond, val) : GTRUE);
}

void GoPolyline::_call_class_round_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoPolylineMetaclass*) mc)->_has_round_postcond)
	    mc->CallPostconditions(polylineNroundPostcond, this);
	// stop when reaching GoPolyline Metaclass instance;
	if (mc == GoPolyline::metaclass())
	    return;
    }
}

void GoPolyline::_call_class_round_offset_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoPolylineMetaclass*) mc)->_has_round_offset_postcond)
	    mc->CallPostconditions(polylineNroundOffsetPostcond, this);
	// stop when reaching GoPolyline Metaclass instance;
	if (mc == GoPolyline::metaclass())
	    return;
    }
}

INLINE void GoPolyline::_call_round_postcond() {
    if (_has_round_postcond)
	CallPostconditions(polylineNroundPostcond);
}

INLINE void GoPolyline::_call_round_offset_postcond() {
    if (_has_round_offset_postcond)
	CallPostconditions(polylineNroundOffsetPostcond);
}

void GoPolylineMetaclass::roundPrecond(PLR_PRECOND_PTR(f),
				       void* closure, GAddOrRemove add) {
    _has_round_precond =
	AddOrRemoveCondition(polylineNroundPrecond, f, closure, add);
}

void GoPolylineMetaclass::roundOffsetPrecond(PLRO_PRECOND_PTR(f),
					     void* closure, GAddOrRemove add) {
    _has_round_offset_precond =
	AddOrRemoveCondition(polylineNroundOffsetPrecond, f, closure, add);
}

void GoPolylineMetaclass::roundPostcond(PL_POSTCOND_PTR(f),
					void* closure, GAddOrRemove add) {
    _has_round_postcond =
	AddOrRemoveCondition(polylineNroundPostcond, f, closure, add);
}

void GoPolylineMetaclass::roundOffsetPostcond(PL_POSTCOND_PTR(f),
					      void* closure, GAddOrRemove add) {
    _has_round_offset_postcond =
	AddOrRemoveCondition(polylineNroundOffsetPostcond, f, closure, add);
}

void GoPolyline::roundPrecond(PLR_PRECOND_PTR(f),
			      void* closure, GAddOrRemove add) {
    _has_round_precond =
	AddOrRemoveCondition(polylineNroundPrecond, f, closure, add);
}

void GoPolyline::roundOffsetPrecond(PLRO_PRECOND_PTR(f),
				    void* closure, GAddOrRemove add) {
    _has_round_offset_precond =
	AddOrRemoveCondition(polylineNroundOffsetPrecond, f, closure, add);
}

void GoPolyline::roundPostcond(PL_POSTCOND_PTR(f),
			       void* closure, GAddOrRemove add) {
    _has_round_postcond =
	AddOrRemoveCondition(polylineNroundPostcond, f, closure, add);
}

void GoPolyline::roundOffsetPostcond(PL_POSTCOND_PTR(f),
				     void* closure, GAddOrRemove add) {
    _has_round_offset_postcond =
	AddOrRemoveCondition(polylineNroundOffsetPostcond, f, closure, add);
}



/*
 *------------
 * CLASS PART
 *------------
 */

static GoPolylineMetaclass* polyline_metaclass = NULL;

/**
FUNCTION
    GoPolylineMetaclass* GoPolyline::metaclass()
DESCRIPTION
    
**/
GoPolylineMetaclass* GoPolyline::metaclass() {
    if (polyline_metaclass == NULL) {
	polyline_metaclass = new GoPolylineMetaclass;
	polyline_metaclass->superclass()->AddSubclass(polyline_metaclass);

        GoStream::bind(GoPolyline::metaclass()->classname(),
		       (int (*) (const void*, GoStream*)) GoPolyline::save,
		       (int (*) (void**, GoStream*)) GoPolyline::restore);
    }
    return polyline_metaclass;
}


/**
FUNCTION
    ~GoPolylineMetaclass()
DESCRIPTION
    
**/
GoPolylineMetaclass::~GoPolylineMetaclass() {
    if (this == polyline_metaclass) {
        superclass()->RemoveSubclass(this);
	polyline_metaclass = NULL;
    }
}


/**
FUNCTION
    GoPolylineMetaclass::GoPolylineMetaclass();
DESCRIPTION
    
**/
GoPolylineMetaclass::GoPolylineMetaclass() {
    _has_round_precond        = _has_round_postcond        =
    _has_round_offset_precond = _has_round_offset_postcond = GFALSE;
}


/**
FUNCTION
    const char* GoPolylineMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoPolylineMetaclass::m_classname() const {
    return "GoPolyline";
}


/**
FUNCTION
    GoMetaclass* GoPolylineMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoPolylineMetaclass::m_superclass() const {
    return GoPolypoint::metaclass();
}


/**
FUNCTION
    const GActionRec* GoPolylineMetaclass::m_actions () const
DESCRIPTION
    
**/
const GActionRec* GoPolylineMetaclass::m_actions () const {
    if ((GoMetaclass*) this != GoPolyline::metaclass()) {
#ifdef DEBUG
	GWarning("the %s class has no \"m_actions()\" method !", classname());
#endif /* DEBUG */
	return empty_action_array;
    }
    return polyline_action_array;
}


/**
FUNCTION
    Go* GoPolylineMetaclass::m_create();
DESCRIPTION
    virtual private
**/
Go* GoPolylineMetaclass::m_create() const {
    return (Go*) (new GoPolyline);
}


/*
 *-------------------
 * PRIVATE FUNCTIONS
 *-------------------
 */

/**
FUNCTION
    GoMetaclass* GoPolyline::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoPolyline::m_metaclass() const {
    return GoPolyline::metaclass();
}


/**
FUNCTION
    void GoPolyline::m_check() const
DESCRIPTION
    virtual private
**/
void GoPolyline::m_check() const {
}


/**
FUNCTION
    int GoPolyline::get_round_points(GPoint**) const;
DESCRIPTION
    virtual private
**/
int GoPolyline::get_round_points(GPoint** roundPoints) const {
    GPoint center, *newPoints = new GPoint[12 * nbPoint()];
    GVector v1, v2;
    int i, j, k = 0;
    GNumber da = ((GNumber) PI) / 20;

    newPoints[k++] = points()[0] ;

    for (i = 1; i < nbPoint() - 1; i++) {
	GNumber a = da;
	DeltaVect(points()[i - 1], points()[i], points()[i + 1], v1, v2);
	newPoints[k++] = points()[i] - v1;
	center = points()[i] - v1 - v2;
	for (j = 0; j < 9; ++j, a += da)
	    newPoints[k++] = center + v1 * sin(a) + v2 * cos(a);
	newPoints[k++] = points()[i] - v2;
    }
    newPoints[k++] = points()[i];
    *roundPoints = newPoints;
    return k;
}



GBool GoPolyline::_set_round(GBool value) {
    if ((value != 0) != (_round != 0)) {
	GBool visible = outline() || background();
	GBool was_mapped = mapped() && visible;

	if (was_mapped)
	    RedrawBackAndFore();
	_round = value != 0;
	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}

#define GOOD_OFFSET_VALUE(value)  (0 <= (value) && (value) <= 0.5)

GBool GoPolyline::_set_round_offset(GNumber value) {
    if (_round_offset != value && GOOD_OFFSET_VALUE(value)) {
	GBool visible = (outline() || background()) && round();
	GBool was_mapped = mapped() && visible;

	if (was_mapped)
	    RedrawBackAndFore();
	_round_offset = value;
	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}


/*
 *---------------------
 * PROTECTED FUNCTIONS
 *---------------------
 */

/**
FUNCTION
    int GoPolyline::save(const GoPolyline* poll, GoStream* gostream);
DESCRIPTION
    
**/
int GoPolyline::save(const GoPolyline* poll, GoStream* gostream) {
    return (gostream->writeBool(poll->round(), "round") &&
	    gostream->writeNumber(poll->roundOffset(), "round offset") &&
	    GoPolypoint::save(poll, gostream));
}

/**
FUNCTION
    int GoPolyline::restore(GoPolyline** poll, GoStream* gostream);
DESCRIPTION
    
**/
int GoPolyline::restore(GoPolyline** poll, GoStream* gostream) {
    GBool round;
    GNumber roffset;
    if (! (gostream->readBool(&round) && gostream->readNumber(&roffset)))
	return 0;

    if (*poll = GoPolyline::create()) {
	(*poll)->_round = round;
	(*poll)->_round_offset = roffset;
	return GoPolypoint::readParams(*poll, gostream);
    }
    return 0;
}

/**
FUNCTION
    int GoPolyline::readParams(GoPolyline* poll, GoStream* gostream);
DESCRIPTION
    
**/
int GoPolyline::readParams(GoPolyline* poll, GoStream* gostream) {
    GBool round;
    GNumber roffset;
    if (! (gostream->readBool(&round) && gostream->readNumber(&roffset)))
	return 0;
    poll->_round = round;
    poll->_round_offset = roffset;
    return GoPolypoint::readParams(poll, gostream);
}

/**
FUNCTION
    void GoPolyline::m_dup(const Go* src, Go*, Go* ,GBool, GBool)
DESCRIPTION
    virtual protected
**/
void GoPolyline::m_dup
    (const Go* src, Go *parcopy, Go *prevcopy, GBool donext, GBool dochildren)
{
    GoPolypoint::m_dup(src, parcopy, prevcopy, donext, dochildren);
    _round = ((const GoPolyline*) src)->_round;
    _round_offset = ((const GoPolyline*) src)->_round_offset;
}


/**
FUNCTION
    GBool GoPolyline::m_contains(const GPoint& absPoint, GNumber maxDist = 0) const
DESCRIPTION
    virtual protected
**/
GBool GoPolyline::m_contains(const GPoint& absPoint, GNumber maxDist) const {
    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;
    
    if (nbPoint() <= 0 || ! (shbrush || shpen))
	return GFALSE;
    
    GPoint *pts;
    int nb;
	
    if (round() && nbPoint() > 2)
	nb = get_round_points(&pts);
    else {
	nb = nbPoint();
	pts = (GPoint*) points();
    }
    GBool answer = interf()->polylineContains(shpen, shbrush, Abscs(), nb, pts,
					      absPoint, maxDist);
    if (pts != points())
	delete(pts);

    return answer;
}


/**
FUNCTION
    GoPolyline::m_add_clip(GRegion& region) const
DESCRIPTION
    virtual protected
**/
void GoPolyline::m_add_clip(GRegion& region) const {
    if (nbPoint() < 2 || ! (background() || outline()))
	return;

    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;
    
    if (nbPoint() >= 3 && round()) {
	GPoint *pts;
	int nb = get_round_points(&pts);

	if (shpen &&
	    (shpen->lineWidth() != 0) && (shpen->joinStyle() == GJOIN_MITER)) {
	    GShPen* tmp_shpen = shpen->dup();
	    GSetJoinStyle(tmp_shpen, GJOIN_ROUND);
	    interf()->polylineClips(tmp_shpen, shbrush, Abscs(), nb, pts,
				    &region, NULL);
	    tmp_shpen->destroy();
	}
	else interf()->polylineClips(shpen, shbrush, Abscs(), nb, pts,
				     &region, NULL);
	delete pts;
    }
    else interf()->polylineClips(shpen, shbrush, Abscs(), nbPoint(), points(),
				 &region, NULL);
}


/**
FUNCTION
    void GoPolyline::m_draw(const GRegion&) const
DESCRIPTION
    virtual protected
**/
void GoPolyline::m_draw(const GRegion&) const
{
    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;

    if (nbPoint() < 2 || ! (shbrush || shpen))
	return;

    GPoint *pts;
    int nb;
	
    if (round() && nbPoint() > 2)
	nb = get_round_points(&pts);
    else {
	nb = nbPoint();
	pts = (GPoint*) points();
    }
    interf()->polyline(shpen, shbrush, Abscs(), nb, pts, GMAX_QUALITY);
	
    if (pts != points())
	delete(pts);
}


/**
FUNCTION
    void GoPolyline::m_wired_draw(const GCoordSys& cs, const GPen&, Quality) const
DESCRIPTION
    virtual protected
**/
void GoPolyline::m_wired_draw(const GCoordSys& cs, const GPen& pen,
			      GQuality quality) const
{
    if (nbPoint() < 2)
	return;
    interf()->polyline(&pen, NULL, cs, nbPoint(), points(), quality);
}


/**
FUNCTION
    GoPolyline::DeltaVect(const GPoint &p1, &p2, &p3, GVector &v1, &v2) const;
DESCRIPTION
    ***************???????????????**********
**/
void GoPolyline::DeltaVect(const GPoint &p1, const GPoint &p2, const GPoint &p3,
			   GVector &v1, GVector &v2) const {
    GCoordSys cs = Abscs();
    GVector av1 = cs.to(p2 - p1);
    GVector av2 = cs.to(p2 - p3);
    GNumber sz = roundOffset() *
	min(max((GNumber) fabs(av1.x), (GNumber) fabs(av1.y)),
	    max((GNumber) fabs(av2.x), (GNumber) fabs(av2.y)));
    av1.resize(sz);
    av2.resize(sz);
    v1 = cs.from(av1);
    v2 = cs.from(av2);
}



/*
 *------------------
 * PUBLIC FUNCTIONS
 *------------------
 */

#define POLYLINE_INIT_VALUES() { \
    _has_round_precond        = _has_round_postcond        = \
    _has_round_offset_precond = _has_round_offset_postcond = GFALSE; \
    _round        = GFALSE; \
    _round_offset = 0.2; \
}
    
/**
FUNCTION
    GoPolyline()
DESCRIPTION
    GoPolyline constructor.
**/
GoPolyline::GoPolyline() {
    POLYLINE_INIT_VALUES();
}


/**
FUNCTION
    GoPolyline(int n, const GPoint* pts) : (n, pts) 
DESCRIPTION
    GoPolyline constructor.
**/
GoPolyline::GoPolyline(int n, const GPoint* pts) : GoPolypoint(n, pts) {
    POLYLINE_INIT_VALUES();
}


/**
FUNCTION
    GoPolyline* GoPolyline::create(int nb, const GPoint *points = NULL)
DESCRIPTION
    public legitimate constructor of GoPolyline (calls pre/postconditions)
    **static member**
**/
GoPolyline* GoPolyline::create(int nb, const GPoint *points) {
    GoPolyline *polyline = (GoPolyline*) Go::Create(GoPolyline::metaclass());
    if (polyline)
	polyline->reset(nb, points);
    return polyline;
}

	
/**
FUNCTION
    GBool GoPolyline::round(GBool val)
DESCRIPTION
    public virtual
**/
GBool GoPolyline::round(GBool val) {
    GBool value = val; // do not trust parameter storage !;
    if ((value != 0) != (_round != 0)     &&
	_call_round_precond(&value)       &&
	_call_class_round_precond(&value) &&
	_set_round(value)) {

	_call_round_postcond();
	_call_class_round_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoPolyline::roundOffset(GNumber val)
DESCRIPTION
    public virtual
**/
GBool GoPolyline::roundOffset(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (_round_offset != value && GOOD_OFFSET_VALUE(value) &&
	_call_round_offset_precond(&value)                 &&
	_call_class_round_offset_precond(&value)           &&
	_set_round_offset(value)) {

	_call_round_offset_postcond();
	_call_class_round_offset_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool delPoint(int index) 
DESCRIPTION
    This function deletes the point at the specified index from the polypoint 
    array.
**/
GBool GoPolyline::delPoint(int index) {
    return (_nbPoint > 2) ? GoPolypoint::delPoint(index) : GFALSE;
}

