#ifdef REV_INFO
static char rcsid[] = "$Id: Polypoint.cc,v 1.5 1992/12/10 16:03:29 smadja Exp $";
#endif
/*****************************************************************************
MODULE: Geometric Objects
FILE:   Polypoint.cc
AUTHOR: Aboobacker, Gilles Benati
DATE:   Feb 1991

DESCRIPTION :
    See the Go/Polypoint.h header.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/

#include <assert.h>
#include <Go/Tools/GeomUtil.h>
#include <Go/Tools/Error.h>
#include <Go/GoStream.h>
#include <Go/Polypoint.h>

static char polypointNaddPointPrecond[]     = "p";
static char polypointNaddPointPostcond[]    = "p";
static char polypointNdelPointPrecond[]     = "p";
static char polypointNdelPointPostcond[]    = "p";
static char polypointNmovePointPrecond[]    = "p";
static char polypointNmovePointPostcond[]   = "p";
static char polypointNresetPointsPrecond[]  = "p";
static char polypointNresetPointsPostcond[] = "p";

// to clarify pre/postcond code;
// (A)dd&move, (D)el and (R)eset;
#define PPA_PRECOND_PTR(f)  GBool f(void*, const GoPolypoint*, int*, GPoint*)
#define PPD_PRECOND_PTR(f)  GBool f(void*, const GoPolypoint*, int*)
#define PPR_PRECOND_PTR(f)  GBool f(void*, const GoPolypoint*, int, const GPoint*)
#define PPR_POSTCOND_PTR(f) void f(void*, GoPolypoint*)
#define PP_POSTCOND_PTR(f)  void f(void*, GoPolypoint*, int)

/*
 *--------
 * ACTIONS
 *--------
 */

void _polypoint_add_previous_middle_point(GoPolypoint *polypt, GPoint* pt)
{
    if (!polypt ||
	!polypt->getMetaclass()->isSubclass(GoPolypoint::metaclass()))
	return;
    int index = polypt->posToIndex(*pt);
    const GPoint &cur  = polypt->points()[index];
    const GPoint &prev = index ?
	polypt->points()[index - 1] :
    polypt->points()[polypt->nbPoint() - 1];
    polypt->addPoint(index,
		     GPoint((cur.x + prev.x)/2, (cur.y + prev.y)/2), GFALSE);
}

void _polypoint_add_next_middle_point(GoPolypoint *polypt, GPoint* pt)
{
    if (!polypt ||
	!polypt->getMetaclass()->isSubclass(GoPolypoint::metaclass()))
	return;
    int nb = polypt->nbPoint();
    int index = polypt->posToIndex(*pt);
    const GPoint &cur  = polypt->points()[index];
    const GPoint &next = (index == nb - 1) ?
	polypt->points()[0] :
    polypt->points()[index + 1];
    polypt->addPoint(index + 1,
		     GPoint((cur.x + next.x)/2, (cur.y + next.y)/2), GFALSE);
}

void _polypoint_delete_closest_point(GoPolypoint *polypt, GPoint* pt)
{
    if (!polypt ||
	!polypt->getMetaclass()->isSubclass(GoPolypoint::metaclass()))
	return;
    int index = polypt->posToIndex(*pt);
    polypt->delPoint(index);

}

static GActionRec go_polypoint_action_array[] = {
  { "addPointForward(GPoint*)",
	(GActionProc) _polypoint_add_next_middle_point },
  { "addPointBackward(GPoint*)",
	(GActionProc) _polypoint_add_previous_middle_point },
  { "deleteClosestPoint(GPoint*)",
	(GActionProc) _polypoint_delete_closest_point },
  { NULL, NULL }
};



/*
 *--------------------
 * PRE/POSTCONDITIONS
 *--------------------
 */


GBool GoPolypoint::_call_class_add_point_precond(int* index,
						 GPoint* point) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
    	if (((GoPolypointMetaclass*) mc)->_has_add_point_precond &&
	    ! mc->CallPreconditions(polypointNaddPointPrecond, this,
				    index, point))
	    return GFALSE;
	// stop when reaching GoPolypoint Metaclass instance;
	if (mc == GoPolypoint::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoPolypoint::_call_class_del_point_precond(int* index) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
    	if (((GoPolypointMetaclass*) mc)->_has_del_point_precond &&
	    ! mc->CallPreconditions(polypointNdelPointPrecond, this, index))
	    return GFALSE;
	// stop when reaching GoPolypoint Metaclass instance;
	if (mc == GoPolypoint::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoPolypoint::_call_class_move_point_precond(int* index,
						  GPoint* point) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
    if (((GoPolypointMetaclass*) mc)->_has_move_point_precond &&
	! mc->CallPreconditions(polypointNmovePointPrecond, this,
				index, point))
	    return GFALSE;
	// stop when reaching GoPolypoint Metaclass instance;
	if (mc == GoPolypoint::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoPolypoint::_call_class_reset_points_precond(int nb,
						    const GPoint* points)const{
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoPolypointMetaclass*) mc)->_has_reset_points_precond &&
	    ! mc->CallPreconditions(polypointNresetPointsPrecond, this,
				    (void*) nb, (void*) points))
	    return GFALSE;
	// stop when reaching GoPolypoint Metaclass instance;
	if (mc == GoPolypoint::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool  GoPolypoint::_call_add_point_precond(int* index, GPoint* point) const {
    return(_has_add_point_precond ?
	   CallPreconditions(polypointNaddPointPrecond, index, point) : GTRUE);
}

GBool GoPolypoint::_call_del_point_precond(int* index) const {
    return(_has_del_point_precond ?
	   CallPreconditions(polypointNdelPointPrecond, index) : GTRUE);
}

GBool  GoPolypoint::_call_move_point_precond(int* index, GPoint* point) const {
    return(_has_move_point_precond ?
	   CallPreconditions(polypointNmovePointPrecond, index, point) : GTRUE);
}

GBool  GoPolypoint::_call_reset_points_precond(int nb, const GPoint* points) const {
    return(_has_reset_points_precond ?
	   CallPreconditions(polypointNresetPointsPrecond,
			     (void*) nb, (void*) points) : GTRUE);
}

void GoPolypoint::_call_class_add_point_postcond(int index){
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoPolypointMetaclass*) mc)->_has_add_point_postcond)
	    mc->CallPostconditions(polypointNaddPointPostcond, this,
				   (void*) index);
	// stop when reaching GoPolypoint Metaclass instance;
	if (mc == GoPolypoint::metaclass())
	    return;
    }
}

void GoPolypoint::_call_class_del_point_postcond(int index){
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoPolypointMetaclass*) mc)->_has_del_point_postcond)
	    mc->CallPostconditions(polypointNdelPointPostcond, this,
				   (void*) index);
	// stop when reaching GoPolypoint Metaclass instance;
	if (mc == GoPolypoint::metaclass())
	    return;
    }
}

void GoPolypoint::_call_class_move_point_postcond(int index) {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoPolypointMetaclass*) mc)->_has_move_point_postcond)
	    mc->CallPostconditions(polypointNmovePointPostcond, this,
				   (void*) index);
	// stop when reaching GoPolypoint Metaclass instance;
	if (mc == GoPolypoint::metaclass())
	    return;
    }
}

void GoPolypoint::_call_class_reset_points_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoPolypointMetaclass*) mc)->_has_reset_points_postcond)
	    mc->CallPostconditions(polypointNresetPointsPostcond, this);
	// stop when reaching GoPolypoint Metaclass instance;
	if (mc == GoPolypoint::metaclass())
	    return;
    }
}

INLINE void GoPolypoint::_call_add_point_postcond(int index) {
    if (_has_add_point_postcond)
	CallPostconditions(polypointNaddPointPostcond, (void*) index);
}

INLINE void GoPolypoint::_call_del_point_postcond(int index) {
    if (_has_del_point_postcond)
	CallPostconditions(polypointNdelPointPostcond, (void*) index);
}

INLINE void GoPolypoint::_call_move_point_postcond(int index) {
    if (_has_move_point_postcond)
	CallPostconditions(polypointNmovePointPostcond, (void*) index);
}

INLINE void GoPolypoint::_call_reset_points_postcond() {
    if (_has_reset_points_postcond)
	CallPostconditions(polypointNresetPointsPostcond);
}

void GoPolypointMetaclass::addPointPrecond(PPA_PRECOND_PTR(f),
					   void* closure, GAddOrRemove add) {
    _has_add_point_precond =
	AddOrRemoveCondition(polypointNaddPointPrecond, f, closure, add);
}

void GoPolypointMetaclass::delPointPrecond(PPD_PRECOND_PTR(f),
					   void* closure, GAddOrRemove add) {
    _has_del_point_precond =
	AddOrRemoveCondition(polypointNdelPointPrecond, f, closure, add);
}

void GoPolypointMetaclass::addPointPostcond(PP_POSTCOND_PTR(f),
					    void* closure, GAddOrRemove add) {
    _has_add_point_postcond =
	AddOrRemoveCondition(polypointNaddPointPostcond, f, closure, add);
}

void GoPolypointMetaclass::delPointPostcond(PP_POSTCOND_PTR(f),
					    void* closure, GAddOrRemove add) {
    _has_del_point_postcond =
	AddOrRemoveCondition(polypointNdelPointPostcond, f, closure, add);
}

void GoPolypointMetaclass::movePointPrecond(PPA_PRECOND_PTR(f),
					    void* closure, GAddOrRemove add) {
    _has_move_point_precond =
	AddOrRemoveCondition(polypointNmovePointPrecond, f, closure, add);
}

void GoPolypointMetaclass::resetPointsPrecond(PPR_PRECOND_PTR(f),
					      void* closure, GAddOrRemove add) {
    _has_reset_points_precond =
	AddOrRemoveCondition(polypointNresetPointsPrecond, f, closure, add);
}

void GoPolypointMetaclass::movePointPostcond(PP_POSTCOND_PTR(f),
					     void* closure, GAddOrRemove add) {
    _has_move_point_postcond =
	AddOrRemoveCondition(polypointNmovePointPostcond, f, closure, add);
}

void GoPolypointMetaclass::resetPointsPostcond(PPR_POSTCOND_PTR(f),
					       void* closure, GAddOrRemove add) {
    _has_reset_points_postcond =
	AddOrRemoveCondition(polypointNresetPointsPostcond, f, closure, add);
}

void GoPolypoint::addPointPrecond(PPA_PRECOND_PTR(f),
				  void* closure, GAddOrRemove add) {
    _has_add_point_precond =
	AddOrRemoveCondition(polypointNaddPointPrecond, f, closure, add);
}

void GoPolypoint::delPointPrecond(PPD_PRECOND_PTR(f),
				  void* closure, GAddOrRemove add) {
    _has_del_point_precond =
	AddOrRemoveCondition(polypointNdelPointPrecond, f, closure, add);
}

void GoPolypoint::addPointPostcond(PP_POSTCOND_PTR(f),
				   void* closure, GAddOrRemove add) {
    _has_add_point_postcond =
	AddOrRemoveCondition(polypointNaddPointPostcond, f, closure, add);
}

void GoPolypoint::delPointPostcond(PP_POSTCOND_PTR(f),
				   void* closure, GAddOrRemove add) {
    _has_del_point_postcond =
	AddOrRemoveCondition(polypointNdelPointPostcond, f, closure, add);
}

void GoPolypoint::movePointPrecond(PPA_PRECOND_PTR(f),
		                   void* closure, GAddOrRemove add) {
    _has_move_point_precond =
	AddOrRemoveCondition(polypointNmovePointPrecond, f, closure, add);
}

void GoPolypoint::resetPointsPrecond(PPR_PRECOND_PTR(f),
				     void* closure, GAddOrRemove add) {
    _has_reset_points_precond =
	AddOrRemoveCondition(polypointNresetPointsPrecond, f, closure, add);
}

void GoPolypoint::movePointPostcond(PP_POSTCOND_PTR(f),
				    void* closure, GAddOrRemove add) {
    _has_move_point_postcond =
	AddOrRemoveCondition(polypointNmovePointPostcond, f, closure, add);
}

void GoPolypoint::resetPointsPostcond(PPR_POSTCOND_PTR(f),
				      void* closure, GAddOrRemove add) {
    _has_reset_points_postcond =
	AddOrRemoveCondition(polypointNresetPointsPostcond, f, closure, add);
}

/*
 *------------
 * CLASS PART
 *------------
 */

static GoPolypointMetaclass* polypoint_metaclass = NULL;

/**
FUNCTION
    GoPolypointMetaclass* GoPolypoint::metaclass()
DESCRIPTION
    **static member**
**/
GoPolypointMetaclass* GoPolypoint::metaclass() {
    if (polypoint_metaclass == NULL) {
	polypoint_metaclass = new GoPolypointMetaclass;
	polypoint_metaclass->superclass()->AddSubclass(polypoint_metaclass);
    }
    return polypoint_metaclass;
}


/**
FUNCTION
    ~GoPolypointMetaclass()
DESCRIPTION
    
**/
GoPolypointMetaclass::~GoPolypointMetaclass() {
    if (this == GoPolypoint::metaclass()) {
        superclass()->RemoveSubclass(this);
	polypoint_metaclass = NULL;
    }
}


/**
FUNCTION
    GoPolypointMetaclass()
DESCRIPTION
    GoPolypointMetaclass constructor
**/
GoPolypointMetaclass::GoPolypointMetaclass() {
    _has_add_point_precond    = _has_add_point_postcond    =
    _has_del_point_precond    = _has_del_point_postcond    =
    _has_move_point_precond   = _has_move_point_postcond   =
    _has_reset_points_precond = _has_reset_points_postcond = GFALSE;
}


/**
FUNCTION
    const char* GoPolypointMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoPolypointMetaclass::m_classname() const {
    return "GoPolypoint";
}


/**
FUNCTION
    GoMetaclass* GoPolypointMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoPolypointMetaclass::m_superclass() const {
    return GoGeometric::metaclass();
}


/**
FUNCTION
    const GActionRec* GoPolypointMetaclass::m_actions () const
DESCRIPTION
    
**/
const GActionRec* GoPolypointMetaclass::m_actions () const {
    if ((GoMetaclass*) this != GoPolypoint::metaclass()) {
#ifdef DEBUG
	GWarning("the %s class has no \"m_actions()\" method !", classname());
#endif /* DEBUG */
	return empty_action_array;
    }
    return go_polypoint_action_array;
}


/*
 *-------------------
 * PRIVATE FUNCTIONS
 *-------------------
 */

/**
FUNCTION
    GoMetaclass* GoPolypoint::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoPolypoint::m_metaclass() const {
    return GoPolypoint::metaclass();
}


/**
FUNCTION
    GoPolypoint::m_dup(const Go* src, Go*, Go*, GBool, GBool)
DESCRIPTION
    virtual private
**/
void GoPolypoint::m_dup
    (const Go* src, Go *parcopy, Go *prevcopy, GBool donext, GBool dochildren)
{
    GoGeometric::m_dup(src, parcopy, prevcopy, donext, dochildren);
    new_points(((GoPolypoint*) src)->nbPoint(), ((GoPolypoint*) src)->points());
}


/**
FUNCTION
    void GoPolypoint::m_check() const
DESCRIPTION
    virtual private
**/
void GoPolypoint::m_check() const {}

/**
FUNCTION
    GBool GoPolypoint::_reset_points(int n, const GPoint* pts);
DESCRIPTION
    
**/
GBool GoPolypoint::_reset_points(int n, const GPoint* pts) {
    if (pts && n > 0) {
	GBool visible    = outline() || background();
	GBool was_mapped = mapped() && visible;
	
	if (was_mapped)
	    RedrawBackAndFore();
	new_points(n, pts);
	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoPolypoint::_add_point(int index, GPoint& point);
DESCRIPTION
    
**/
GBool GoPolypoint::_add_point(int index, const GPoint& rel_pt) {
    if (index >= 0 && index <= _nbPoint) {
	GBool visible = outline() || background();
	GBool was_mapped = mapped() && visible;

	if (was_mapped)
	    RedrawBackAndFore();

	GPoint* tmp = _points;
	
	++_nbPoint;
	_points = (GPoint*) GMALLOC(_nbPoint * sizeof(GPoint));

	if (_points) {
	    memcpy(_points, tmp, index * sizeof(GPoint));
	    _points[index] = rel_pt;
	    memcpy(_points + index + 1, tmp + index,
		   (_nbPoint - index - 1) * sizeof(GPoint));
	    delete(tmp);

	    if (visible)
		MyClipHasChanged();
	    if (was_mapped)
		RedrawForeOnly();
	    
	    return GTRUE;
	}
	else {
	    _nbPoint--;
	    _points = tmp;
	    GWarning("GoPolypoint::addPoint : GMALLOC failed");
	}
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoPolypoint::_del_point(int index);
DESCRIPTION
    
**/
GBool GoPolypoint::_del_point(int index) {
    if (index >= 0 && index < _nbPoint) {
	GBool visible = outline() || background();
	GBool was_mapped = mapped() && visible;

	if (was_mapped)
	    RedrawBackAndFore();

	--_nbPoint;
	if (_nbPoint)
	    memcpy(_points + index, _points + index + 1,
		   (_nbPoint - index) * sizeof(GPoint));
	else {
	    delete _points;
	    _points = NULL;
	}

	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoPolypoint::_move_point(int index, const GPoint& rel_pt);
DESCRIPTION
    
**/
GBool GoPolypoint::_move_point(int index, const GPoint& rel_pt) {
    if (index >= 0 && index <= _nbPoint && (rel_pt != _points[index])) {
	GBool visible = outline() || background();
	GBool was_mapped = mapped() && visible;

	if (was_mapped)
	    RedrawBackAndFore();
	_points[index] = rel_pt;
	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}

/*
 *--------------------
 * PROTECTED FUNCTIONS
 *--------------------
 */

/**
FUNCTION
    int GoPolypoint::save(const GoPolypoint* polp, GoStream* gostream);
DESCRIPTION
    
**/
int GoPolypoint::save(const GoPolypoint* polp, GoStream* gostream) {
    const GPoint* pts = polp->points();
    int nb = polp->nbPoint();
    if (! gostream->writeInt(nb, "nb of points"))
	return 0;
    if (nb) {
	if (! gostream->writePoint(pts[0], "Points array"))
	    return 0;
	for (int i = 1; i < nb; i++) {
	    if (! gostream->writePoint(pts[i]))
		return 0;
	}
    }
    return GoGeometric::save(polp, gostream);
}

/**
FUNCTION
    int GoPolypoint::readParams(GoPolypoint* polp, GoStream* gostream);
DESCRIPTION
    
**/
int GoPolypoint::readParams(GoPolypoint* polp, GoStream* gostream) {
    int nb = 0;
    int ret = gostream->readInt(&nb);
    if (nb > 0) {
#ifdef NO_ALLOCA
	GPoint *pts = (GPoint*) GMALLOC(nb*sizeof(GPoint));
#else
	GPoint *pts = (GPoint*) alloca(nb*sizeof(GPoint));
#endif
	
	for (int i = 0; i < nb && ret; i++)
	    ret = gostream->readPoint(&(pts[i]));
	if (ret)
	    polp->new_points(nb, pts);
#ifdef NO_ALLOCA
	GFREE(pts);
#endif
    }
    
    return (ret ? GoGeometric::readParams(polp, gostream) : 0);
}


/**
FUNCTION
    GBool GoPolypoint::m_contains(const GPoint& absPoint, GNumber maxDist) const
DESCRIPTION
    virtual protected
**/
GBool GoPolypoint::m_contains(const GPoint& absPoint, GNumber maxDist) const {
    if (_nbPoint <= 0)
	return GFALSE;
    
    GCoordSys cs = Abscs();

    if (outline())
	maxDist += sh_pen()->lineWidth() / 2;
    GNumber maxSquare = maxDist * maxDist;

    for (int i = 0; i < _nbPoint; i++) {
	GPoint crt = cs.to(_points[i]);
	if (GSquareDistance(absPoint, crt) < maxSquare)
	    return GTRUE;
    }
    return GFALSE;
}



/**
FUNCTION
    GoPolypoint::new_points(int n, const GPoint* pts, GBool remove_old_pts = GTRUE)
DESCRIPTION
    
**/
void GoPolypoint::new_points(int n, const GPoint* pts, GBool remove_old_pts) {
    if (n <= 0) {
	if (_points && remove_old_pts)
	    delete _points;
	_nbPoint = 0;
	_points = NULL;
    }
    else {
	int i;
	if (n > _nbPoint) {
	    if (_points && remove_old_pts)
		delete _points;
	    _points = new GPoint[n];
	}
	_nbPoint = n;
	for (i = 0; i < n; i++)
	    _points[i] = pts[i];
    }
}


/**
FUNCTION
    GoPolypoint(int n = 0, const GPoint* pts = NULL) {
DESCRIPTION
    This is  the constructor for GoPolypoint class. As this class is never
    instantiated on its own , this function is protected. It creates a point
    array of 'n' points and copies the points specified by 'pts' into it.
**/
GoPolypoint::GoPolypoint(int n, const GPoint* pts) {
    _has_add_point_precond    = _has_add_point_postcond    =
    _has_del_point_precond    = _has_del_point_postcond    =
    _has_move_point_precond   = _has_move_point_postcond   =
    _has_reset_points_precond = _has_reset_points_postcond = GFALSE;

    _nbPoint = 0;
    _points  = NULL;
    new_points(n, pts);
}


/**
FUNCTION
    ~GoPolypoint()
DESCRIPTION
    This is the destructor for GoPolypoint calss.
    It deletes the polypoint array.
**/
GoPolypoint::~GoPolypoint() { 
    if (_nbPoint)
	delete _points;
}
    

/*
 *------------------
 * PUBLIC FUNCTIONS
 *------------------
 */


/**
FUNCTION
    GBool reset(int n, const GPoint* pts) 
DESCRIPTION
    This function allows to change the geometric definition of the object.
    It updates the polypoint array with new points specified by 'pts' and 'n'.
**/
GBool GoPolypoint::reset(int n, const GPoint* pts) {
    if (pts && n > 0 &&
	_call_reset_points_precond(n, pts)       &&
	_call_class_reset_points_precond(n, pts) &&
        _reset_points(n, pts)) {

	_call_reset_points_postcond();
	_call_class_reset_points_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    int posToIndex(const GPoint& pos) const
DESCRIPTION
    This function returns the index of the point nearest to 'pos' in the point
    array.
**/
int GoPolypoint::posToIndex(const GPoint& pos) const {
    if (_nbPoint <= 0)
	return(-1);
    
    int i, minIndex = 0;
    GCoordSys cs = Abscs();
    GNumber curSquare, minSquare = GSquareDistance(cs.to(_points[0]), pos);

    for (i = 1; i < _nbPoint; i++) {
	if ((curSquare = GSquareDistance(cs.to(_points[i]), pos)) < minSquare) {
	    minSquare = curSquare; 
	    minIndex = i;
	}
    }
    return minIndex;
}


/**
FUNCTION
    GBool addPoint(int index, const GPoint& point, GBool isAbsPoint) 
DESCRIPTION
    This function adds the point 'rel_pt' at the 'index'-th position in the
    polypoint array. if isAbsPoint is GTRUE the point is considered to be an 
    absolute point else a relative point.
**/
GBool GoPolypoint::addPoint(int index, const GPoint& point, GBool isAbsPoint) {
    GPoint rel_pt = point;
    if (isAbsPoint)
	rel_pt = Abscs().from(point);
    
    if (index >= 0 && index <= _nbPoint                &&
	_call_add_point_precond(&index, &rel_pt)       &&
	_call_class_add_point_precond(&index, &rel_pt) &&
	_add_point(index, rel_pt)) {

	_call_add_point_postcond(index);
	_call_class_add_point_postcond(index);
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool delPoint(int index) 
DESCRIPTION
    This function deletes the point at the specified index from the polypoint 
    array.
**/
GBool GoPolypoint::delPoint(int index) {
    if (index >= 0 && index < _nbPoint        &&
	_call_del_point_precond(&index)       &&
	_call_class_del_point_precond(&index) &&
	_del_point(index)) {

	_call_del_point_postcond(index);
	_call_class_del_point_postcond(index);
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool movePoint(int index, const GPoint& rel_pt, GBool isAbsPoint) 
DESCRIPTION
    This function moves the point specified by 'index' to the new position
    specified by point. 
**/
GBool GoPolypoint::movePoint(int index, const GPoint& point, GBool isAbsPoint){
    GPoint rel_pt = point;
    if (isAbsPoint)
	rel_pt = Abscs().from(point);

    if (index >= 0 && index <= _nbPoint && (rel_pt != _points[index]) &&
	_call_move_point_precond(&index, &rel_pt)                     &&
	_call_class_move_point_precond(&index, &rel_pt)               &&
	_move_point(index, rel_pt)) {

	_call_move_point_postcond(index);
	_call_class_move_point_postcond(index);
	return GTRUE;
    }
    return GFALSE;
}





