#ifdef REV_INFO
static char rcsid[] = "$Id: Rectangle.cc,v 1.4 1992/12/29 15:16:13 davy Exp $";
#endif
/***************************************************************************
MODULE: Geometric Objects
FILE:   Rectangle.cc
AUTHOR: Oliver Jojic
DATE:   Feb 1991

DESCRIPTION:
    See the Go/Rectangle.h header.
    
COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/

#include <assert.h>
#include <Go/Tools/Error.h>
#include <Go/GoInterf.h>
#include <Go/GoStream.h>
#include <Go/Rectangle.h>



static char rectangleNwidthPrecond[]   = "p";
static char rectangleNwidthPostcond[]  = "p";
static char rectangleNheightPrecond[]  = "p";
static char rectangleNheightPostcond[] = "p";
static char rectangleNresetPrecond[]   = "p";
static char rectangleNresetPostcond[]  = "p";

// to clarify pre/postcond code;
#define RECT_PRECOND_PTR(f)  GBool f(void*, const GoRectangle*, GNumber*)
#define RECTR_PRECOND_PTR(f) GBool f(void*, const GoRectangle*, GNumber*, GNumber*)
#define RECT_POSTCOND_PTR(f) void f(void*, GoRectangle*)

/*
 *--------
 * ACTIONS
 *--------
 */

static void set_width(GoRectangle *rect, GNumber width)
{
    if (rect &&
	rect->getMetaclass()->isSubclass(GoRectangle::metaclass()))
	rect->width(width);
}

static void set_height(GoRectangle *rect, GNumber height)
{
    if (rect &&
	rect->getMetaclass()->isSubclass(GoRectangle::metaclass()))
	rect->height(height);
}

static void set_width_height(GoRectangle *rect, GNumber width, GNumber height) {
    if (rect &&
	rect->getMetaclass()->isSubclass(GoRectangle::metaclass()))
	rect->reset(width, height);
}


static GActionRec go_rectangle_action_array[] = {
  { "width(GNumber)",         (GActionProc) set_width },
  { "height(GNumber)",        (GActionProc) set_height },
  { "reset(GNumber, GNumber)", (GActionProc) set_width_height },
  { NULL, NULL }
};



/*
 *--------------------
 * PRE/POSTCONDITIONS
 *--------------------
 */

GBool GoRectangle::_call_class_width_precond(GNumber* w) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRectangleMetaclass*) mc)->_has_width_precond &&
	    ! mc->CallPreconditions(rectangleNwidthPrecond, this, w))
	    return GFALSE;
	// stop when reaching GoRectangle Metaclass instance;
	if (mc == GoRectangle::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoRectangle::_call_class_height_precond(GNumber* h) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRectangleMetaclass*) mc)->_has_height_precond &&
	    ! mc->CallPreconditions(rectangleNheightPrecond, this, h))
	    return GFALSE;
	// stop when reaching GoRectangle Metaclass instance;
	if (mc == GoRectangle::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoRectangle::_call_class_reset_precond(GNumber* w, GNumber* h) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRectangleMetaclass*) mc)->_has_reset_precond &&
	    ! mc->CallPreconditions(rectangleNresetPrecond, this, w, h))
	    return GFALSE;
	// stop when reaching GoRectangle Metaclass instance;
	if (mc == GoRectangle::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoRectangle::_call_width_precond(GNumber* w) const {
    return(_has_width_precond ?
	   CallPreconditions(rectangleNwidthPrecond, w) : GTRUE);
}

GBool GoRectangle::_call_height_precond(GNumber* h) const {
    return(_has_height_precond ?
	   CallPreconditions(rectangleNheightPrecond, h) : GTRUE);
}

GBool GoRectangle::_call_reset_precond(GNumber* w, GNumber* h) const {
    return(_has_reset_precond ?
	   CallPreconditions(rectangleNresetPrecond, w, h) : GTRUE);
}

void GoRectangle::_call_class_width_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRectangleMetaclass*) mc)->_has_width_postcond)
	    CallPostconditions(rectangleNwidthPostcond, this);
	// stop when reaching GoRectangle Metaclass instance;
	if (mc == GoRectangle::metaclass())
	    return;
    }
}

void GoRectangle::_call_class_height_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRectangleMetaclass*) mc)->_has_height_postcond)
	    CallPostconditions(rectangleNheightPostcond, this);
	// stop when reaching GoRectangle Metaclass instance;
	if (mc == GoRectangle::metaclass())
	    return;
    }
}

void GoRectangle::_call_class_reset_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRectangleMetaclass*) mc)->_has_reset_postcond)
	    CallPostconditions(rectangleNresetPostcond, this);
	// stop when reaching GoRectangle Metaclass instance;
	if (mc == GoRectangle::metaclass())
	    return;
    }
}

void GoRectangle::_call_width_postcond() {
    if (_has_width_postcond)
	CallPostconditions(rectangleNwidthPostcond);
}

void GoRectangle::_call_height_postcond() {
    if (_has_height_postcond)
	CallPostconditions(rectangleNheightPostcond);
}

void GoRectangle::_call_reset_postcond() {
    if (_has_reset_postcond)
	CallPostconditions(rectangleNresetPostcond);
}

void GoRectangleMetaclass::widthPrecond(RECT_PRECOND_PTR(f),
					void* closure, GAddOrRemove add) {
    _has_width_precond =
	AddOrRemoveCondition(rectangleNwidthPrecond, f, closure, add);
}

void GoRectangleMetaclass::heightPrecond(RECT_PRECOND_PTR(f),
					 void* closure, GAddOrRemove add) {
    _has_height_precond =
	AddOrRemoveCondition(rectangleNheightPrecond, f, closure, add);
}

void GoRectangleMetaclass::widthPostcond(RECT_POSTCOND_PTR(f),
					 void* closure, GAddOrRemove add) {
    _has_width_postcond =
	AddOrRemoveCondition(rectangleNwidthPostcond, f, closure, add);
}

void GoRectangleMetaclass::heightPostcond(RECT_POSTCOND_PTR(f),
					  void* closure, GAddOrRemove add) {
    _has_height_postcond =
	AddOrRemoveCondition(rectangleNheightPostcond, f, closure, add);
}

void GoRectangleMetaclass::resetPrecond(RECTR_PRECOND_PTR(f),
					void* closure, GAddOrRemove add) {
    _has_reset_precond =
	AddOrRemoveCondition(rectangleNresetPrecond, f, closure, add);
}

void GoRectangleMetaclass::resetPostcond(RECT_POSTCOND_PTR(f),
					 void* closure, GAddOrRemove add) {
    _has_reset_postcond =
	AddOrRemoveCondition(rectangleNresetPostcond, f, closure, add);
}

void GoRectangle::widthPrecond(RECT_PRECOND_PTR(f),
			       void* closure, GAddOrRemove add) {
    _has_width_precond =
	AddOrRemoveCondition(rectangleNwidthPrecond, f, closure, add);
}

void GoRectangle::heightPrecond(RECT_PRECOND_PTR(f),
				void* closure, GAddOrRemove add) {
    _has_height_precond =
	AddOrRemoveCondition(rectangleNheightPrecond, f, closure, add);
}

void GoRectangle::widthPostcond(RECT_POSTCOND_PTR(f),
				void* closure, GAddOrRemove add) {
    _has_width_postcond =
	AddOrRemoveCondition(rectangleNwidthPostcond, f, closure, add);
}

void GoRectangle::heightPostcond(RECT_POSTCOND_PTR(f),
				 void* closure, GAddOrRemove add) {
    _has_height_postcond =
	AddOrRemoveCondition(rectangleNheightPostcond, f, closure, add);
}

void GoRectangle::resetPrecond(RECTR_PRECOND_PTR(f),
			       void* closure, GAddOrRemove add) {
    _has_reset_precond =
	AddOrRemoveCondition(rectangleNresetPrecond, f, closure, add);
}

void GoRectangle::resetPostcond(RECT_POSTCOND_PTR(f),
				void* closure, GAddOrRemove add) {
    _has_reset_postcond =
	AddOrRemoveCondition(rectangleNresetPostcond, f, closure, add);
}



/*
 *------------
 * CLASS PART
 *------------
 */

static GoRectangleMetaclass* rectangle_metaclass = NULL;

/**
FUNCTION
    GoRectangleMetaclass* GoRectangle::metaclass()
DESCRIPTION
    
**/
GoRectangleMetaclass* GoRectangle::metaclass() {
    if (rectangle_metaclass == NULL) {
	rectangle_metaclass = new GoRectangleMetaclass;
	rectangle_metaclass->superclass()->AddSubclass(rectangle_metaclass);

        GoStream::bind(GoRectangle::metaclass()->classname(),
		       (int (*) (const void*, GoStream*)) GoRectangle::save,
		       (int (*) (void**, GoStream*)) GoRectangle::restore);
    }
    return rectangle_metaclass;
}


/**
FUNCTION
    ~GoRectangleMetaclass()
DESCRIPTION
    
**/
GoRectangleMetaclass::~GoRectangleMetaclass() {
    if (this == rectangle_metaclass) {
        superclass()->RemoveSubclass(this);
	rectangle_metaclass = NULL;
    }
}


/**
FUNCTION
    GoRectangleMetaclass::GoRectangleMetaclass()
DESCRIPTION
    
**/
 GoRectangleMetaclass::GoRectangleMetaclass() {
    _has_width_precond  = _has_width_postcond  =
    _has_height_precond = _has_height_postcond =
    _has_reset_precond  = _has_reset_postcond  = GFALSE;
}


/**
FUNCTION
    const char* GoRectangleMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoRectangleMetaclass::m_classname() const {
    return "GoRectangle";
}


/**
FUNCTION
    GoMetaclass* GoRectangleMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoRectangleMetaclass::m_superclass() const {
    return GoPolygon::metaclass();
}


/**
FUNCTION
    const GActionRec* GoRectangleMetaclass::m_actions () const
DESCRIPTION
    
**/
const GActionRec* GoRectangleMetaclass::m_actions () const {
    if ((GoMetaclass*) this != GoRectangle::metaclass()) {
#ifdef DEBUG
	GWarning("the %s class has no \"m_actions()\" method !", classname());
#endif /* DEBUG */
	return empty_action_array;
    }
    return go_rectangle_action_array;
}


/**
FUNCTION
    Go* GoRectangleMetaclass::m_create();
DESCRIPTION
    virtual private
**/
Go* GoRectangleMetaclass::m_create() const {
    return (Go*) (new GoRectangle);
}


/*
 *----------------
 * PRIVATE MEMBERS
 *----------------
 */

/**
FUNCTION
    GoMetaclass* GoRectangle::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoRectangle::m_metaclass() const {
    return GoRectangle::metaclass();
}


/**
FUNCTION
    void GoRectangle::m_check() const
DESCRIPTION
    virtual private
**/
void GoRectangle::m_check() const {}


/**
FUNCTION
    _set_wh(GNumber w, GNumber h)
DESCRIPTION
    
**/
void GoRectangle::_set_wh(GNumber w, GNumber h) {
    GPoint pts[4];

    if (w < 0)
	w = 1;
    if (h < 0)
	h = 1;
    pts[0] = GPoint(0, 0);
    pts[1] = GPoint(w, 0);
    pts[2] = GPoint(w, h);
    pts[3] = GPoint(0, h);
    new_points(4, pts);
}


/*
 *------------------
 * PROTECTED MEMBERS
 *------------------
 */


/**
FUNCTION
    int GoRectangle::save(const GoRectangle* rect, GoStream* gostream);
DESCRIPTION
    
**/
int GoRectangle::save(const GoRectangle* rect, GoStream* gostream) {
    return (gostream->writeNumber(rect->width(), "width") &&
	    gostream->writeNumber(rect->height(), "height") &&
	    gostream->writeBool(rect->round(), "round") &&
	    gostream->writeNumber(rect->roundOffset(), "round offset") &&
	    GoGeometric::save(rect, gostream));
}


/**
FUNCTION
    int GoRectangle::restore(GoRectangle** rect, GoStream* gostream);
DESCRIPTION
    
**/
int GoRectangle::restore(GoRectangle** rect, GoStream* gostream) {
    GNumber w, h, roffset;
    GBool round;
    if (! (gostream->readNumber(&w) && gostream->readNumber(&h) &&
	   gostream->readBool(&round) && gostream->readNumber(&roffset)))
	return 0;
    if (*rect = GoRectangle::create(w, h)) {
	(*rect)->_round = round;
	if (0. <= roffset && roffset <= 0.5)
	    (*rect)->_round_offset = roffset;
	return GoGeometric::readParams(*rect, gostream);
    }
    return 0;
}


/**
FUNCTION
    int GoRectangle::readParams(GoRectangle* rect, GoStream* gostream);
DESCRIPTION
    
**/
int GoRectangle::readParams(GoRectangle* rect, GoStream* gostream) {
    GNumber w, h, roffset;
    GBool round;
    if (! (gostream->readNumber(&w) && gostream->readNumber(&h) &&
	   gostream->readBool(&round) && gostream->readNumber(&roffset)))
	return 0;
    rect->reset(w, h);
    rect->_round = round;
    if (0. <= roffset && roffset <= 0.5)
	rect->_round_offset = roffset;
    return GoGeometric::readParams(rect, gostream);
}

/**
FUNCTION
    GBool GoRectangle::m_contains(const GPoint& absPoint, GNumber maxDist=0) const
DESCRIPTION
    virtual protected
**/
GBool GoRectangle::m_contains(const GPoint& absPoint, GNumber maxDist) const {
    const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;
    
    if (! (shbrush || shpen))
	return GFALSE;

    if (round())
	return GoPolygon::m_contains(absPoint, maxDist);
    else return interf()->rectangleContains(shpen, shbrush, Abscs(),
					    points()[0], width(), height(),
					    absPoint, maxDist);
}


/**
FUNCTION
    GoRectangle::m_add_clip(GRegion& region) const
DESCRIPTION
    virtual protected
**/
void GoRectangle::m_add_clip(GRegion& region) const {
    if (! (outline() || background()))
	return;

    if (round())
        GoPolygon::m_add_clip(region);
    else {
	const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
	const GShBrush *shbrush = background() ? brush().shBrush() : NULL;
    
	interf()->rectangleClips(shpen, shbrush, Abscs(), points()[0],
				 width(), height(), &region, NULL);
    }
}


/**
FUNCTION
    GoRectangle::m_add_opaque_clip(GRegion& region) const
DESCRIPTION
    virtual protected
**/
void GoRectangle::m_add_opaque_clip(GRegion& region) const {
    if (! (outline() || background()))
	return;

    if (round())
        this->GoPolygon::m_add_opaque_clip(region);
    else  {
	const GShPen   *shpen   = outline()    ? sh_pen()          : NULL;
	const GShBrush *shbrush = background() ? brush().shBrush() : NULL;
    
	interf()->rectangleClips(shpen, shbrush, Abscs(), points()[0],
				 width(), height(), NULL, &region);
    }
}


/**
FUNCTION
    void GoRectangle::m_draw(const GRegion& region) const
DESCRIPTION
    virtual protected
**/
void GoRectangle::m_draw(const GRegion& region) const {
    const GShPen   *shpen   = outline()    ? sh_pen()   : NULL;
    const GShBrush *shbrush = background() ? brush().shBrush() : NULL;
    
    if (! (shbrush || shpen))
	return;

    if (round())
        GoPolygon::m_draw(region);
    else interf()->rectangle(shpen, shbrush, Abscs(), points()[0],
			     width(), height());
}


/**
FUNCTION
    void GoRectangle::m_wired_draw(const GCoordSys&, const GPen&,
                                   GQuality) const
DESCRIPTION
    virtual protected
**/
void GoRectangle::m_wired_draw(const GCoordSys& cs,
			       const GPen& pen, GQuality quality) const
{
    interf()->rectangle(&pen, NULL, cs, points()[0],
			width(), height(), quality);
}



/**
FUNCTION
    GBool GoRectangle::Reset(GNumber w, GNumber h)
DESCRIPTION
    
**/
GBool GoRectangle::Reset(GNumber w, GNumber h) {
    if (w < 0 && h < 0)
	return GFALSE;

    GNumber old_w = width();
    GNumber old_h = height();
    
    if (w < 0)
	w = old_w;
    if (h < 0)
	h = old_h;

    if (w == old_w && h == old_h)
	return GFALSE;
    
    GBool visible = outline() || background();
    GBool was_mapped = mapped() && visible;
    
    if (was_mapped && (old_w > w || old_h > h || ! background()))
	RedrawBackAndFore();
    _set_wh(w, h);
    
    if (visible)
	MyClipHasChanged();
    if (was_mapped && (old_w < w || old_h < h))
	RedrawForeOnly();

    return GTRUE;
}


/*
 *------------------
 * PUBLIC FUNCTIONS
 *------------------
 */


/**
FUNCTION
    GoRectangle::GoRectangle(GNumber w = 1, GNumber h = 1)
DESCRIPTION
    
**/
GoRectangle::GoRectangle(GNumber w, GNumber h) {
    _has_width_precond  = _has_width_postcond  =
    _has_height_precond = _has_height_postcond =
    _has_reset_precond  = _has_reset_postcond  = GFALSE;

    _set_wh(w, h);
}


/**
FUNCTION
    GoRectangle* GoRectangle::create(GNumber w, GNumber h);
DESCRIPTION
    Another legitimate GoRectangle constructor (calls pre/postcondition).
    **static member**
**/
GoRectangle* GoRectangle::create(GNumber w, GNumber h) {
    GoRectangle* rect = (GoRectangle*) Go::Create(GoRectangle::metaclass());
    if (rect)
	rect->_set_wh(w, h);
    return rect;
}


/**
FUNCTION
    GBool GoRectangle::reset(GNumber w_val, GNumber h_val);
DESCRIPTION
    If preconditions allow it, sets width and height to w_val and h_val
    (possibly modified by preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoRectangle::reset(GNumber w_val, GNumber h_val) {
    GNumber w = w_val, h = h_val; // do not trust parameter storage !;
    if ((w >= 0 || h >= 0) && (w != width() || h != height()) &&
	_call_reset_precond(&w, &h)			      &&
	_call_class_reset_precond(&w, &h)                     &&
	Reset(w, h)) {

	_call_reset_postcond();
	_call_class_reset_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoRectangle::width(GNumber val)
DESCRIPTION
    If preconditions allow it, sets width to val (possibly modified by preconds)
    and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoRectangle::width(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (value >= 0 && value != width()    &&
	_call_width_precond(&value)       &&
	_call_class_width_precond(&value) &&
	Reset(value, -1)) {

	_call_width_postcond();
	_call_class_width_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoRectangle::height(GNumber val)
DESCRIPTION
    If preconditions allow it, sets height to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoRectangle::height(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (value >= 0 && value != height()    &&
	_call_height_precond(&value)       &&
	_call_class_height_precond(&value) &&
	Reset(-1, value)) {

	_call_height_postcond();
	_call_class_height_postcond();
	return GTRUE;
    }
    return GFALSE;
}

/**
FUNCTION
    GBool GoRectangle::reset(int, const GPoint*);
    GBool GoRectangle::addPoint(int, const GPoint&, GBool);
    GBool GoRectangle::movePoint(int, const GPoint&, GBool);
    GBool GoRectangle::delPoint(int);
DESCRIPTION
    virtual private : just disable superclass methods;
**/
GBool GoRectangle::reset(int, const GPoint*)            { return GFALSE; }
GBool GoRectangle::addPoint(int, const GPoint&, GBool)  { return GFALSE; }
GBool GoRectangle::movePoint(int, const GPoint&, GBool) { return GFALSE; }
GBool GoRectangle::delPoint(int)                        { return GFALSE; }
