#ifdef REV_INFO
static char rcsid[] = "$Id: RegPoly.cc,v 1.4 1992/12/14 09:10:16 smadja Exp $";
#endif
/*****************************************************************************
MODULE: Geometric Objects
FILE:   RegPoly.cc
AUTHOR: Gilles Benati
DATE:   Feb 1991

DESCRIPTION:
    See the Go/RegPoly.h header file
    
COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/

#include <assert.h>
#include <Go/Tools/Error.h>
#include <Go/GoStream.h>
#include <Go/RegPoly.h>

/*
 * To use "static" as in C (static to a file), use STATIC_C
 * If the compiler understands it as "static member of class ...", you need
 * to add -DSTATIC_CONFUSION to CC defines.
 */
#ifndef STATIC_CONFUSION
#define STATIC_C static
#else
#define STATIC_C
#endif /* STATIC_CONFUSION */

static char regpolygonNradiusPrecond[]   = "p";
static char regpolygonNradiusPostcond[]  = "p";
static char regpolygonNanglePrecond[]    = "p";
static char regpolygonNanglePostcond[]   = "p";
static char regpolygonNnbPointPrecond[]  = "p";
static char regpolygonNnbPointPostcond[] = "p";
static char regpolygonNresetPrecond[]    = "p";
static char regpolygonNresetPostcond[]   = "p";


// to clarify pre/postcond code;
#define RPN_PRECOND_PTR(f) GBool f(void*, const GoRegPolygon*, GNumber*)
#define RPI_PRECOND_PTR(f) GBool f(void*, const GoRegPolygon*, int*)
#define RPR_PRECOND_PTR(f) GBool f(void*,const GoRegPolygon*,int*,GNumber*,GNumber*)
#define RP_POSTCOND_PTR(f) void f(void*, GoRegPolygon*)

/*
 *--------
 * ACTIONS
 *--------
 */

static GBool isOK(GoRegPolygon* reg_poly)
{
    return reg_poly &&
	reg_poly->getMetaclass()->isSubclass(GoRegPolygon::metaclass());
}

static void add_edge(GoRegPolygon *reg_poly) {
    if (isOK(reg_poly))
	reg_poly->nbPoint(reg_poly->GoPolypoint::nbPoint() + 1);
}

static void del_edge(GoRegPolygon *reg_poly) {
    if (isOK(reg_poly))
	reg_poly->nbPoint(reg_poly->GoPolypoint::nbPoint() - 1);
}

static void set_angle(GoRegPolygon *reg_poly, GNumber angle) {
    if (isOK(reg_poly))
	reg_poly->angle(angle);
}

static void set_radius(GoRegPolygon *reg_poly, GNumber rad) {
    if (isOK(reg_poly))
	reg_poly->radius(rad);
}

static void set_nb_point(GoRegPolygon *reg_poly, int nb) {
    if (isOK(reg_poly))
	reg_poly->nbPoint(nb);
}

static void set_all(GoRegPolygon *reg_poly, int nb, GNumber rad, GNumber angle) {
    if (isOK(reg_poly))
	reg_poly->reset(nb, rad, angle);
}

static GActionRec go_regpolygon_action_array[] =
{ { "addEdge()",                           (GActionProc) add_edge },
  { "delEdge()",                           (GActionProc) del_edge },
  { "angle(GNumber)",                       (GActionProc) set_angle },
  { "radius(GNumber)",                      (GActionProc) set_radius },
  { "nbPoint(unsigned int)",               (GActionProc) set_nb_point },
  { "reset(unsigned int, GNumber, GNumber)", (GActionProc) set_all },
  { NULL, NULL }
};

/*
 *--------------------
 * pre/postconditions
 *--------------------
 */

STATIC_C
GBool GoRegPolygon::_call_class_radius_precond(GNumber* rad) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRegPolygonMetaclass*) mc)->_has_radius_precond &&
	    ! mc->CallPreconditions(regpolygonNradiusPrecond, this, rad))
	    return GFALSE;
	// stop when reaching GoRegPolygon Metaclass instance;
	if (mc == GoRegPolygon::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

STATIC_C
GBool GoRegPolygon::_call_class_angle_precond(GNumber* angle) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRegPolygonMetaclass*) mc)->_has_angle_precond &&
	    ! mc->CallPreconditions(regpolygonNanglePrecond, this, angle))
	    return GFALSE;
	// stop when reaching GoRegPolygon Metaclass instance;
	if (mc == GoRegPolygon::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

STATIC_C
GBool GoRegPolygon::_call_class_nbPoint_precond(int* nb) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRegPolygonMetaclass*) mc)->_has_nbPoint_precond &&
	    ! mc->CallPreconditions(regpolygonNnbPointPrecond, this, nb))
	    return GFALSE;
	// stop when reaching GoRegPolygon Metaclass instance;
	if (mc == GoRegPolygon::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

STATIC_C
GBool GoRegPolygon::_call_class_reset_precond(int* nb, GNumber* rad,
					      GNumber* angle) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRegPolygonMetaclass*) mc)->_has_reset_precond &&
	    ! mc->CallPreconditions(regpolygonNresetPrecond, this,
				    nb, rad, angle))
	    return GFALSE;
	// stop when reaching GoRegPolygon Metaclass instance;
	if (mc == GoRegPolygon::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoRegPolygon::_call_radius_precond(GNumber* rad) const {
    return(_has_radius_precond ?
	   CallPreconditions(regpolygonNradiusPrecond, rad) : GTRUE);
}

GBool GoRegPolygon::_call_angle_precond(GNumber* angle) const {
    return(_has_angle_precond ?
	   CallPreconditions(regpolygonNanglePrecond, angle) : GTRUE);
}

GBool GoRegPolygon::_call_nbPoint_precond(int* nb) const {
    return(_has_nbPoint_precond ?
	   CallPreconditions(regpolygonNnbPointPrecond, nb) : GTRUE);
}

GBool GoRegPolygon::_call_reset_precond(int* nb, GNumber* rad,
					      GNumber* angle) const {
    return(_has_reset_precond ?
	   CallPreconditions(regpolygonNresetPrecond, nb, rad, angle) : GTRUE);
}

void GoRegPolygon::_call_class_radius_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRegPolygonMetaclass*) mc)->_has_radius_postcond)
	    mc->CallPostconditions(regpolygonNradiusPostcond, this);
	// stop when reaching GoRegPolygon Metaclass instance;
	if (mc == GoRegPolygon::metaclass())
	    return;
    }
}

void GoRegPolygon::_call_class_angle_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRegPolygonMetaclass*) mc)->_has_angle_postcond)
	    mc->CallPostconditions(regpolygonNanglePostcond, this);
	// stop when reaching GoRegPolygon Metaclass instance;
	if (mc == GoRegPolygon::metaclass())
	    return;
    }
}

void GoRegPolygon::_call_class_nbPoint_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRegPolygonMetaclass*) mc)->_has_nbPoint_postcond)
	    mc->CallPostconditions(regpolygonNnbPointPostcond, this);
	// stop when reaching GoRegPolygon Metaclass instance;
	if (mc == GoRegPolygon::metaclass())
	    return;
    }
}

void GoRegPolygon::_call_class_reset_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoRegPolygonMetaclass*) mc)->_has_reset_postcond)
	    mc->CallPostconditions(regpolygonNresetPostcond, this);
	// stop when reaching GoRegPolygon Metaclass instance;
	if (mc == GoRegPolygon::metaclass())
	    return;
    }
}

INLINE void GoRegPolygon::_call_radius_postcond() {
    if (_has_radius_postcond)
	CallPostconditions(regpolygonNradiusPostcond);
}

INLINE void GoRegPolygon::_call_angle_postcond() {
    if (_has_angle_postcond)
	CallPostconditions(regpolygonNanglePostcond);
}

INLINE void GoRegPolygon::_call_nbPoint_postcond() {
    if (_has_nbPoint_postcond)
	CallPostconditions(regpolygonNnbPointPostcond);
}

INLINE void GoRegPolygon::_call_reset_postcond() {
    if (_has_reset_postcond)
	CallPostconditions(regpolygonNresetPostcond);
}

void GoRegPolygonMetaclass::radiusPrecond(RPN_PRECOND_PTR(f),
					  void* closure, GAddOrRemove add) {
    _has_radius_precond =
	AddOrRemoveCondition(regpolygonNradiusPrecond, f, closure, add);
}

void GoRegPolygonMetaclass::anglePrecond(RPN_PRECOND_PTR(f),
					 void* closure, GAddOrRemove add) {
    _has_angle_precond =
	AddOrRemoveCondition(regpolygonNanglePrecond, f, closure, add);
}

void GoRegPolygonMetaclass::radiusPostcond(RP_POSTCOND_PTR(f),
					   void* closure, GAddOrRemove add) {
    _has_radius_postcond =
	AddOrRemoveCondition(regpolygonNradiusPostcond, f, closure, add);
}

void GoRegPolygonMetaclass::anglePostcond(RP_POSTCOND_PTR(f),
				          void* closure, GAddOrRemove add) {
    _has_angle_postcond =
	AddOrRemoveCondition(regpolygonNanglePostcond, f, closure, add);
}

void GoRegPolygonMetaclass::nbPointPrecond(RPI_PRECOND_PTR(f),
					   void* closure, GAddOrRemove add) {
    _has_nbPoint_precond =
	AddOrRemoveCondition(regpolygonNnbPointPrecond, f, closure, add);
}

void GoRegPolygonMetaclass::resetPrecond(RPR_PRECOND_PTR(f),
					 void* closure, GAddOrRemove add) {
    _has_reset_precond =
	AddOrRemoveCondition(regpolygonNresetPrecond, f, closure, add);
}

void GoRegPolygonMetaclass::nbPointPostcond(RP_POSTCOND_PTR(f),
					    void* closure, GAddOrRemove add) {
    _has_nbPoint_postcond =
	AddOrRemoveCondition(regpolygonNnbPointPostcond, f, closure, add);
}

void GoRegPolygonMetaclass::resetPostcond(RP_POSTCOND_PTR(f),
					  void* closure, GAddOrRemove add) {
    _has_reset_postcond =
	AddOrRemoveCondition(regpolygonNresetPostcond, f, closure, add);
}

void GoRegPolygon::radiusPrecond(RPN_PRECOND_PTR(f),
				 void* closure, GAddOrRemove add) {
    _has_radius_precond =
	AddOrRemoveCondition(regpolygonNradiusPrecond, f, closure, add);
}

void GoRegPolygon::anglePrecond(RPN_PRECOND_PTR(f),
				void* closure, GAddOrRemove add) {
    _has_angle_precond =
	AddOrRemoveCondition(regpolygonNanglePrecond, f, closure, add);
}

void GoRegPolygon::radiusPostcond(RP_POSTCOND_PTR(f),
				  void* closure, GAddOrRemove add) {
    _has_radius_postcond =
	AddOrRemoveCondition(regpolygonNradiusPostcond, f, closure, add);
}

void GoRegPolygon::anglePostcond(RP_POSTCOND_PTR(f),
				 void* closure, GAddOrRemove add) {
    _has_angle_postcond =
	AddOrRemoveCondition(regpolygonNanglePostcond, f, closure, add);
}

void GoRegPolygon::nbPointPrecond(RPI_PRECOND_PTR(f),
				  void* closure, GAddOrRemove add) {
    _has_nbPoint_precond =
	AddOrRemoveCondition(regpolygonNnbPointPrecond, f, closure, add);
}

void GoRegPolygon::resetPrecond(RPR_PRECOND_PTR(f),
				void* closure, GAddOrRemove add) {
    _has_reset_precond =
	AddOrRemoveCondition(regpolygonNresetPrecond, f, closure, add);
}

void GoRegPolygon::nbPointPostcond(RP_POSTCOND_PTR(f),
				   void* closure, GAddOrRemove add) {
    _has_nbPoint_postcond =
	AddOrRemoveCondition(regpolygonNnbPointPostcond, f, closure, add);
}

void GoRegPolygon::resetPostcond(RP_POSTCOND_PTR(f),
				 void* closure, GAddOrRemove add) {
    _has_reset_postcond =
	AddOrRemoveCondition(regpolygonNresetPostcond, f, closure, add);
}


/*
 *------------
 * CLASS PART
 *------------
 */


static GoRegPolygonMetaclass* regPolygon_metaclass = NULL;

/**
FUNCTION
    GoRegPolygonMetaclass* GoRegPolygon::metaclass()
DESCRIPTION
    
**/
GoRegPolygonMetaclass* GoRegPolygon::metaclass() {
    if (regPolygon_metaclass == NULL) {
	regPolygon_metaclass = new GoRegPolygonMetaclass;
	regPolygon_metaclass->superclass()->
	    AddSubclass(regPolygon_metaclass);
        GoStream::bind(GoRegPolygon::metaclass()->classname(),
		       (int (*) (const void*, GoStream*)) GoRegPolygon::save,
		       (int (*) (void**, GoStream*)) GoRegPolygon::restore);
    }
    return regPolygon_metaclass;
}


/**
FUNCTION
    ~GoRegPolygonMetaclass()
DESCRIPTION
    
**/
GoRegPolygonMetaclass::~GoRegPolygonMetaclass() {
    if (this == regPolygon_metaclass) {
        superclass()->RemoveSubclass(this);
	regPolygon_metaclass = NULL;
    }
}


/**
FUNCTION
    GoRegPolygonMetaclass::GoRegPolygonMetaclass();
DESCRIPTION
    
**/
GoRegPolygonMetaclass::GoRegPolygonMetaclass() {
    _has_radius_precond  = _has_radius_postcond  =
    _has_angle_precond   = _has_angle_postcond   =
    _has_nbPoint_precond = _has_nbPoint_postcond =
    _has_reset_precond   = _has_reset_postcond   = GFALSE;
}


/**
FUNCTION
    const char* GoRegPolygonMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoRegPolygonMetaclass::m_classname() const {
    return "GoRegPolygon";
}


/**
FUNCTION
    GoMetaclass* GoRegPolygonMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoRegPolygonMetaclass::m_superclass() const {
    return GoPolygon::metaclass();
}

/**
FUNCTION
    const GActionRec* GoRegPolygonMetaclass::m_actions () const
DESCRIPTION
    
**/
const GActionRec* GoRegPolygonMetaclass::m_actions () const {
    if ((GoMetaclass*) this != GoRegPolygon::metaclass()) {
#ifdef DEBUG
	GWarning("the %s class has no \"m_actions()\" method !", classname());
#endif /* DEBUG */
	return empty_action_array;
    }
    return go_regpolygon_action_array;
}


/**
FUNCTION
    Go* GoRegPolygonMetaclass::m_create();
DESCRIPTION
    virtual private
**/
Go* GoRegPolygonMetaclass::m_create() const {
    return (Go*) (new GoRegPolygon);
}


/**
FUNCTION
    GoRegPolygon* GoRegPolygon::create(int n, GNumber radius = 1.0,
                                       GNumber angle = 0);
DESCRIPTION
    Another legitimate GoRegPolygon constructor (calls pre/postconditions)
**/
GoRegPolygon* GoRegPolygon::create(int n, GNumber radius, GNumber angle) {
    GoRegPolygon *regpoly =
	(GoRegPolygon*) Go::Create(GoRegPolygon::metaclass());
    if (regpoly)
	regpoly->Reset(n, radius, angle, GTRUE);
    return regpoly;
}


/*
 *-------------------
 * PRIVATE FUNCTIONS
 *-------------------
 */

static void compute_points(int n, GNumber rad, GNumber ang, GPoint* pts) {
    GNumber Nangle    = ((GNumber) 2 * PI) / n;
    GNumber cosa      = cos(ang);
    GNumber sina      = sin(ang);
    GNumber cosNangle = cos(Nangle);
    GNumber sinNangle = sin(Nangle);

    for (int i = 0; i < n; i++) {
	GNumber tcos = cosa;
	pts[i] = rad * GPoint(cosa, sina);
	cosa = cosa * cosNangle - sina * sinNangle;
	sina = sina * cosNangle + tcos * sinNangle;
    }
}


/**
FUNCTION
    GoMetaclass* GoRegPolygon::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoRegPolygon::m_metaclass() const {
    return GoRegPolygon::metaclass();
}


/**
FUNCTION
    void GoRegPolygon::m_check() const
DESCRIPTION
    virtual private
**/
void GoRegPolygon::m_check() const {}


STATIC_C
GBool GoRegPolygon::_radius(GNumber rad) {
    if (_nbPoint > 1 && rad >= 0. && rad != radius()) {
	GBool visible = outline() || background();
	GBool was_mapped = mapped() && visible;

	if (was_mapped)
	    RedrawBackAndFore();
	
	GNumber ratio = rad / radius();
	for (int i = 0; i < _nbPoint; i++)
	    _points[i] *= ratio;

	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}
    

STATIC_C
GBool GoRegPolygon::_angle(GNumber ang) {
    if (_nbPoint > 1 && ang != angle()) {
	
	GBool visible = outline() || background();
	GBool was_mapped = mapped() && visible;

	if (was_mapped)
	    RedrawBackAndFore();

	compute_points(_nbPoint, radius(), ang, _points);

	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();
	
	return GTRUE;
    }
    return GFALSE;
}


/*
 *--------------------
 * PROTECTED FUNCTIONS
 *--------------------
 */

/**
FUNCTION
    int GoRegPolygon::save(const GoRegPolygon* regp, GoStream* gostream);
DESCRIPTION
    
**/
int GoRegPolygon::save(const GoRegPolygon* regp, GoStream* gostream) {
    return (gostream->writeInt(regp->_nbPoint, "nb of points") &&
	    gostream->writeNumber(regp->radius(), "radius") &&
	    gostream->writeNumber(regp->angle(), "angle") &&
	    gostream->writeBool(regp->round(), "round") &&
	    gostream->writeNumber(regp->roundOffset(), "round offset") &&
	    GoGeometric::save(regp, gostream));
}

/**
FUNCTION
    int GoRegPolygon::restore(GoRegPolygon** regp, GoStream* gostream);
DESCRIPTION
    
**/
int GoRegPolygon::restore(GoRegPolygon** regp, GoStream* gostream) {
    int nb;
    GNumber radius, angle, roffset;
    GBool round;
    if (! (gostream->readInt(&nb) && gostream->readNumber(&radius) &&
	   gostream->readNumber(&angle) && gostream->readBool(&round) &&
	   gostream->readNumber(&roffset)))
	return 0;
    if (*regp = GoRegPolygon::create(nb, radius, angle)) {
	(*regp)->_round = round;
	if (0. <= roffset && roffset <= 0.5)
	    (*regp)->_round_offset = roffset;
	return GoGeometric::readParams(*regp, gostream);
    }
    return 0;
}

/**
FUNCTION
    int GoRegPolygon::readParams(GoRegPolygon* regp,  GoStream* gostream);
DESCRIPTION
    
**/
int GoRegPolygon::readParams(GoRegPolygon* regp,  GoStream* gostream) {
    int nb;
    GNumber radius, angle, roffset;
    GBool round;
    if (! (gostream->readInt(&nb) && gostream->readNumber(&radius) &&
	   gostream->readNumber(&angle) && gostream->readBool(&round) &&
	   gostream->readNumber(&roffset)))
	return 0;
    regp->_round = round;
    if (0. <= roffset && roffset <= 0.5)
	regp->_round_offset = roffset;
    regp->Reset(nb, radius, angle, GTRUE);
    return GoGeometric::readParams(regp, gostream);
}


/**
FUNCTION
    GBool GoRegPolygon::Reset(int n, GNumber rad, GNumber ang,
                             GBool basic = GFALSE)
DESCRIPTION
    if basic == GTRUE, do not call through GoPolypoint public methods to
    avoid useless work (pre/postcondition on reset, myClipHasChange, etc..)
    Thus, basic should only be GTRUE at INITIALIZATION.
**/
GBool GoRegPolygon::Reset(int n, GNumber rad, GNumber ang, GBool basic){
    if (n < 2) {
	if (basic)
	    new_points(0, NULL);
	else GoPolygon::reset(0, NULL);
	return GTRUE;
    }

    if (rad <= 0.)
	return GFALSE;

#define STACK_ARRAY_SIZE 10
    GPoint pts_array[STACK_ARRAY_SIZE]; 
    GPoint *pts = NULL;
    
    if (n > STACK_ARRAY_SIZE)
	pts = new GPoint[n];
    else pts = (GPoint*) pts_array;

    compute_points(n, rad, ang, pts);

    if (basic)
	new_points(n, pts);
    else GoPolygon::reset(n, pts);

    if (n > STACK_ARRAY_SIZE)
	delete pts;
#undef STACK_ARRAY_SIZE

    return GTRUE;
}



/*
 *-------------------
 * PUBLIC FUNCTIONS
 *-------------------
 */

/**
FUNCTION
    GoRegPolygon(int n, GNumber rad = 1.0, GNumber ang = 0) : (0, NULL)
DESCRIPTION
    
**/
GoRegPolygon::GoRegPolygon(int n, GNumber rad, GNumber ang) : GoPolygon(0, NULL) {
    _has_radius_precond  = _has_radius_postcond  =
    _has_angle_precond   = _has_angle_postcond   =
    _has_nbPoint_precond = _has_nbPoint_postcond =
    _has_reset_precond   = _has_reset_postcond   = GFALSE;

    Reset(n, rad, ang, GTRUE);
}


/**
FUNCTION
    GNumber GoRegPolygon::radius() const;
DESCRIPTION
    
**/
GNumber GoRegPolygon::radius() const {
    return(_nbPoint < 1 ? 0. : _points[0].size());
}


/**
FUNCTION
    GoRegPolygon::angle(GNumber val)
DESCRIPTION
    If preconditions allow it, sets angle to val (possibly modified by preconds)
    and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoRegPolygon::angle(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (_nbPoint > 1 && value != angle()  &&
	_call_angle_precond(&value)       &&
	_call_class_angle_precond(&value) &&
	_angle(value)) {
	
	_call_angle_postcond();
	_call_class_angle_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GoRegPolygon::nbPoint(int nb)
DESCRIPTION
    If preconditions allow it, sets number of GPoint to nb (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
    
**/
GBool GoRegPolygon::nbPoint(int nb) {
    if (nb > 1 && nb != _nbPoint         &&
	_call_nbPoint_precond(&nb)       &&
	_call_class_nbPoint_precond(&nb) &&
	Reset(nb, radius(), angle(), GFALSE)) {

	_call_nbPoint_postcond();
	_call_class_nbPoint_postcond();
	return GTRUE;
    }
    return GFALSE;
}



/**
FUNCTION
    GoRegPolygon::radius(GNumber val) 
DESCRIPTION
    If preconditions allow it, sets radius to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoRegPolygon::radius(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (_nbPoint > 1 && value >= 0. && value != radius() &&
	_call_radius_precond(&value)                     &&
	_call_class_radius_precond(&value)               &&
	_radius(value)) {

	_call_radius_postcond();
	_call_class_radius_postcond();
	return GTRUE;
    }
    return GFALSE;
}

/**
FUNCTION
    GBool GoRegPolygon::reset(int n, GNumber r_val = 1.0, GNumber a_val = 0)
DESCRIPTION
    If preconditions allow it, sets number of points, radius, first angle
    to the given values (all may be modified by preconds) and call
    postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoRegPolygon::reset(int n, GNumber r_val, GNumber a_val) {
    GNumber rad = r_val, ang = a_val; // do not trust parameter storage !;
    if (rad >= 0 && n >= 0                        &&
	_call_reset_precond(&n, &rad, &ang)       &&
	_call_class_reset_precond(&n, &rad, &ang) &&
	Reset(n, rad, ang, GFALSE)) {

	_call_reset_postcond();
	_call_class_reset_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoRegPolygon::reset(int, const GPoint*);
    GBool GoRegPolygon::addPoint(int, const GPoint&, GBool);
    GBool GoRegPolygon::movePoint(int, const GPoint&, GBool);
    GBool GoRegPolygon::delPoint(int);
DESCRIPTION
    Just disable superclass virtual methods;
**/
GBool GoRegPolygon::reset(int, const GPoint*)		{ return GFALSE; }
GBool GoRegPolygon::addPoint(int, const GPoint&, GBool)	{ return GFALSE; }
GBool GoRegPolygon::movePoint(int, const GPoint&, GBool)	{ return GFALSE; }
GBool GoRegPolygon::delPoint(int)			{ return GFALSE; }


