#ifdef REV_INFO
static char rcsid[] = "$Id: Tau.cc,v 1.3 1992/12/14 09:59:04 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Geometric
FILE:    Tau.cc
AUTHOR:  Gilles Benati
DATE:    September 20th, 1991

DESCRIPTION:

    See Go/Tau.h header file

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France

  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.

  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <assert.h>
#include <Go/Tau.h>
#include <Go/GoStream.h>


static char tauSplineNbiasPrecond[]  = "p";
static char tauSplineNbiasPostcond[] = "p";

// to clarify pre/postcond code;
#define TAU_PRECOND_PTR(f)  GBool f(void*, const GoTauSpline*, GNumber*)
#define TAU_POSTCOND_PTR(f) void f(void*, GoTauSpline*)

/*
 *--------
 * ACTIONS
 *--------
 */

static void set_bias(GoTauSpline *tau, GNumber value)
{
    if (tau && tau->getMetaclass()->isSubclass(GoTauSpline::metaclass()))
	tau->bias(value);
}

static GActionRec go_tau_spline_action_array[] = {
  { "bias(GNumber)",	(GActionProc) set_bias },
  { NULL, NULL }
};



/*
 *--------------------
 * PRE/POSTCONDITIONS
 *--------------------
 */


GBool GoTauSpline::_call_class_bias_precond(GNumber* val) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoTauSplineMetaclass*) mc)->_has_bias_precond &&
	    ! mc->CallPreconditions(tauSplineNbiasPrecond, this, val))
	    return GFALSE;
	// stop when reaching GoTauSpline Metaclass instance;
	if (mc == GoTauSpline::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

void GoTauSpline::_call_class_bias_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoTauSplineMetaclass*) mc)->_has_bias_postcond)
	    mc->CallPostconditions(tauSplineNbiasPostcond, this);
	// stop when reaching GoTauSpline Metaclass instance;
	if (mc == GoTauSpline::metaclass())
	    return;
    }
}

GBool GoTauSpline::_call_bias_precond(GNumber* val) const {
    return(_has_bias_precond ?
	   CallPreconditions(tauSplineNbiasPrecond, val) : GTRUE);
}


INLINE void GoTauSpline::_call_bias_postcond() {
    if (_has_bias_postcond)
	CallPostconditions(tauSplineNbiasPostcond);
}


void GoTauSplineMetaclass::biasPrecond(TAU_PRECOND_PTR(f),
				       void* closure, GAddOrRemove add) {
    _has_bias_precond =
	AddOrRemoveCondition(tauSplineNbiasPrecond, f, closure, add);
}

void GoTauSplineMetaclass::biasPostcond(TAU_POSTCOND_PTR(f),
					void* closure, GAddOrRemove add) {
    _has_bias_postcond =
	AddOrRemoveCondition(tauSplineNbiasPostcond, f, closure, add);
}


void GoTauSpline::biasPrecond(TAU_PRECOND_PTR(f),
			      void* closure, GAddOrRemove add) {
    _has_bias_precond =
	AddOrRemoveCondition(tauSplineNbiasPrecond, f, closure, add);
}


void GoTauSpline::biasPostcond(TAU_POSTCOND_PTR(f),
			       void* closure, GAddOrRemove add) {
    _has_bias_postcond =
	AddOrRemoveCondition(tauSplineNbiasPostcond, f, closure, add);
}


/*
 *------------
 * CLASS PART
 *------------
 */

static GoTauSplineMetaclass* tau_spline_metaclass = NULL;

/**
FUNCTION
    GoTauSplineMetaclass* GoTauSpline::metaclass()
DESCRIPTION
    
**/
GoTauSplineMetaclass* GoTauSpline::metaclass() {
    if (tau_spline_metaclass == NULL) {
	tau_spline_metaclass = new GoTauSplineMetaclass;
	tau_spline_metaclass->superclass()->AddSubclass(tau_spline_metaclass);

        GoStream::bind(GoTauSpline::metaclass()->classname(),
		       (int (*) (const void*, GoStream*)) GoTauSpline::save,
		       (int (*) (void**, GoStream*)) GoTauSpline::restore);
    }
    return tau_spline_metaclass;
}


/**
FUNCTION
    GoTauSplineMetaclass()
DESCRIPTION
    GoTauSplineMetaclass constructor
**/
GoTauSplineMetaclass::GoTauSplineMetaclass() {
    _has_bias_precond = _has_bias_postcond = GFALSE;
}


/**
FUNCTION
    ~GoTauSplineMetaclass()
DESCRIPTION
    
**/
GoTauSplineMetaclass::~GoTauSplineMetaclass() {
    if (this == tau_spline_metaclass) {
        superclass()->RemoveSubclass(this);
	tau_spline_metaclass = NULL;
    }
}


/**
FUNCTION
    const char* GoTauSplineMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoTauSplineMetaclass::m_classname() const {
    return "GoTauSpline";
}


/**
FUNCTION
    GoMetaclass* GoTauSplineMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoTauSplineMetaclass::m_superclass() const {
    return GoCubic::metaclass();
}


/**
FUNCTION
    Go* GoTauSplineMetaclass::m_create();
DESCRIPTION
    virtual private
**/
Go* GoTauSplineMetaclass::m_create() const {
    return (Go*) (new GoTauSpline);
}



/**
FUNCTION
    const GActionRec* GoTauSplineMetaclass::m_actions () const
DESCRIPTION
    
**/
const GActionRec* GoTauSplineMetaclass::m_actions () const {
    if ((GoMetaclass*) this != GoTauSpline::metaclass()) {
#ifdef DEBUG
	GWarning("the %s class has no \"m_actions()\" method !", classname());
#endif /* DEBUG */
	return empty_action_array;
    }
    return go_tau_spline_action_array;
}

/*
 *---------------------
 * PROTECTED FUNCTIONS
 *---------------------
 */

/**
FUNCTION
    void GoTauSpline::m_dup(const Go* src, Go*, Go*, GBool donext, GBool);
DESCRIPTION
    **virtual protected**
**/
void GoTauSpline::m_dup
    (const Go* src, Go* parcopy, Go* prevcopy, GBool donext, GBool dochildren)
{
    GoCubic::m_dup(src, parcopy, prevcopy, donext, dochildren);
    
    const GoTauSpline *tau = (const GoTauSpline*) src;
    _bias = tau->_bias;
}


/**
FUNCTION
    int GoTauSpline::save(const GoTauSpline* tau, GoStream* gostream);
DESCRIPTION
    **static member**
**/
int GoTauSpline::save(const GoTauSpline* tau, GoStream* gostream) {
    return (gostream->writeNumber(tau->bias(), "bias") &&
	    GoCubic::save(tau, gostream));
}


/**
FUNCTION
    int GoTauSpline::restore(GoTauSpline** tau, GoStream* gostream); 
DESCRIPTION
    **static member**
**/
int GoTauSpline::restore(GoTauSpline** tau, GoStream* gostream) {
    // Might be optimized;;
    if (*tau = GoTauSpline::create())
	return GoTauSpline::readParams(*tau, gostream);
    else return 0;
}

/**
FUNCTION
    int GoTauSpline::readParams(GoTauSpline* tau, GoStream* gostream);
DESCRIPTION
    
**/
int GoTauSpline::readParams(GoTauSpline* tau, GoStream* gostream) {
    GNumber value;
    if (! gostream->readNumber(&value))
	return 0;
    tau->_bias   = value;
    tau->_inited = GFALSE;
    return GoCubic::readParams(tau, gostream);
}


/**
FUNCTION
    void GoTauSpline::initCoeffMatrix();
DESCRIPTION
    virtual protected
**/
void GoTauSpline::initCoeffMatrix() {
    _coeffMatrix[0][0] = (_bias - 1) * _tension;
    _coeffMatrix[0][1] = 2 - _bias * _tension;
    _coeffMatrix[0][2] = (1 - _bias) * _tension - 2;
    _coeffMatrix[0][3] = _bias * _tension;
    _coeffMatrix[1][0] = 2 * (1 - _bias) * _tension;
    _coeffMatrix[1][1] = (3 * _bias - 1) * _tension - 3;
    _coeffMatrix[1][2] = 3 - _tension;
    _coeffMatrix[1][3] = - _bias * _tension;
    _coeffMatrix[2][0] = (_bias - 1) * _tension;
    _coeffMatrix[2][1] = (1 - 2 * _bias) * _tension;
    _coeffMatrix[2][2] = _bias * _tension;
    _coeffMatrix[2][3] = 0;
    _coeffMatrix[3][0] = 0;
    _coeffMatrix[3][1] = 1;
    _coeffMatrix[3][2] = 0;
    _coeffMatrix[3][3] = 0;
    _inited = GTRUE;
}


/*
 *-------------------
 * PRIVATE FUNCTIONS
 *-------------------
 */

/**
FUNCTION
    GoMetaclass* GoTauSpline::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoTauSpline::m_metaclass() const {
    return GoTauSpline::metaclass();
}


/**
FUNCTION
    void GoTauSpline::m_check() const
DESCRIPTION
    virtual private
**/
void GoTauSpline::m_check() const {}


GBool GoTauSpline::_set_bias(GNumber value) {
    if (_bias != value) {
	
	GBool visible = outline() || background();
	GBool was_mapped = mapped() && visible;

	if (was_mapped)
	    RedrawBackAndFore();

	_bias   = value;
	_inited = GFALSE;

	if (visible)
	    MyClipHasChanged();
	if (was_mapped)
	    RedrawForeOnly();

	return GTRUE;
    }
    return GFALSE;
}


/*
 *------------------
 * PUBLIC FUNCTIONS
 *------------------
 */

#define TAU_SPLINE_INIT_VALUES() { \
    _has_bias_precond = _has_bias_postcond = GFALSE; \
    _tension = 1; \
    _bias    = 0; \
    _inited  = GFALSE; \
}

/**
FUNCTION
    GoTauSpline()
DESCRIPTION
    GoTauSpline constructor. It creates a tau spline of tension 1.
**/
GoTauSpline::GoTauSpline() : GoCubic(0, NULL) {
    TAU_SPLINE_INIT_VALUES();
}


/**
FUNCTION
    GoTauSpline(int nb, const GPoint* pts)
DESCRIPTION
    GoTauSpline constructor. It creates a tau spline of tension 1.
**/
GoTauSpline::GoTauSpline(int nb, const GPoint* pts) : GoCubic(nb, pts) {
    TAU_SPLINE_INIT_VALUES();
}


/**
FUNCTION
    GoTauSpline* GoTauSpline::create(int nb, const GPoint *pts = NULL);
DESCRIPTION
    The other legitimate GoTauSpline constructor : calls pre/postcond.
**/
GoTauSpline* GoTauSpline::create(int nb, const GPoint *pts) {
    GoTauSpline *tau = (GoTauSpline*) Go::Create(GoTauSpline::metaclass());
    if (tau)
	tau->reset(nb, pts);
    return tau;
}

/**
FUNCTION
    GBool bias(GNumber val)
DESCRIPTION
    public virtual.
    If preconditions allow it, sets bias to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoTauSpline::bias(GNumber val) {
    GNumber value = val; // do not trust parameter storage !;
    if (_bias != value                   &&
	_call_bias_precond(&value)       &&
	_call_class_bias_precond(&value) &&
	_set_bias(value)) {

	_call_bias_postcond();
	_call_class_bias_postcond();
	return GTRUE;
    }
    return GFALSE;
}

