#ifdef REV_INFO
static char rcsid[] = "$Id: GraphElt.cc,v 1.5 1992/12/10 16:11:52 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  BULL Graphic Objects
FILE:    GraphElt.cc
AUTHOR:  Aline Jacomet
DATE:    January 11th 1991

DESCRIPTION:
         Definition of the graph element object abstract class.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <Go/machdef.h>
#include <Go/GraphElt.h>
#include <Go/mGraphElt.h>
#include <Go/Tools/Error.h>
#include <Go/Tree.h>
#include <Go/TreeNode.h>
#include <Go/Graph.h>
#include <Go/GraphNode.h>
#include <Go/GoStream.h>




                           /*********************
			    * PRIVATE FUNCTIONS *
			    *********************/



static GoGraphElementMetaclass* graphElt_metaclass = NULL;

/**
FUNCTION
    GoGraphElement:metaclass()
DESCRIPTION
    
**/
GoGraphElementMetaclass* GoGraphElement::metaclass() {
    if (graphElt_metaclass == NULL) {
	graphElt_metaclass = new GoGraphElementMetaclass;
	graphElt_metaclass->superclass()->AddSubclass(graphElt_metaclass);
    }
    return graphElt_metaclass;
}


/**
FUNCTION
    ~GoGraphElementMetaclass()
DESCRIPTION
    
**/
GoGraphElementMetaclass::~GoGraphElementMetaclass() {
    if (this == graphElt_metaclass) {
        superclass()->RemoveSubclass(this);
	graphElt_metaclass = NULL;
    }
    reparentPrecond(accept_reparent, NULL, GREMOVE);
    relcsPrecond(change_relcs, NULL, GREMOVE);
}


/**
FUNCTION
    GoGraphElementMetaclass::GoGraphElementMetaclass()
DESCRIPTION
    
**/
 GoGraphElementMetaclass::GoGraphElementMetaclass() {
    reparentPrecond(accept_reparent, NULL, GADD);
    relcsPrecond(change_relcs, NULL, GADD);
}


/**
FUNCTION
    m_classname()
DESCRIPTION
    Returns the ClassName.
**/
const char* GoGraphElementMetaclass::m_classname () const
{
    return "GoGraphElement";
}


/**
FUNCTION
    m_superclass()
DESCRIPTION
    Returns the SuperClass.
**/
GoMetaclass* GoGraphElementMetaclass::m_superclass () const
{
    return Go::metaclass();
}


/**
FUNCTION
    d_get_assoc()
DESCRIPTION
    Returns the assoc.
**/
GAssoc* GoGraphElement::d_get_assoc() const
{
    const GAssoc* a = &_assoc;
    return (GAssoc*)a;
}


/**
FUNCTION
    d_get_parent()
DESCRIPTION
    Returns the parent id.
**/
Go* GoGraphElement::d_get_parent() const
{
    return _parent;
}


/**
FUNCTION
    d_get_next()
DESCRIPTION
    Returns the next id.
**/
Go* GoGraphElement::d_get_next() const
{
    return _next;
}


/**
FUNCTION
    d_get_first_child()
DESCRIPTION
    Returns the first_child id.
**/
Go* GoGraphElement::d_get_first_child() const
{
    return _child;
}


/**
FUNCTION
    d_store_children_array()
DESCRIPTION
    
**/
GBool GoGraphElement::d_store_children_array () const
{
    return GFALSE;
}


/**
FUNCTION
    d_set_next()
DESCRIPTION
    Sets the next id.
**/
void GoGraphElement::d_set_next (const Go* go)
{
    _next = (Go*)go;
}


/**
FUNCTION
    d_set_tree()
DESCRIPTION
    Sets the next and parent id.
**/
void GoGraphElement::d_set_tree (const Go* parent, const Go* , const Go* next)
{
    _parent = (Go*)parent;
    _next = (Go*)next;
}


/**
FUNCTION
    d_store_relcs()
DESCRIPTION
    
**/
GBool GoGraphElement::d_store_relcs () const
{
    return GFALSE;
}


/**
FUNCTION
    d_store_abscs()
DESCRIPTION
    
**/
GBool GoGraphElement::d_store_abscs () const
{
    return GTRUE;
}


/**
FUNCTION
    d_get_abscs()
DESCRIPTION
    Returns the absolute coordinate system.
**/
const GCoordSys& GoGraphElement::d_get_abscs(const GCoordSys* parabs) const
{
//    static GCoordSys tmp_cs;
    static GCoordSys* ptmp_cs;
    if(ptmp_cs == NULL)
	ptmp_cs = new GCoordSys;
    GCoordSys& tmp_cs = *ptmp_cs;   
    tmp_cs.O = _origin;
    if (parabs) {
	tmp_cs.I = parabs->I;
	tmp_cs.J = parabs->J;
    } else
	if (parent()) {
	    GCoordSys cs = parent()->Abscs();
	    tmp_cs.I = cs.I;
	    tmp_cs.J = cs.J;
	}
    return tmp_cs;
}


/**
FUNCTION
    d_set_coordsys()
DESCRIPTION
    Sets the coordinate systems.
**/
void GoGraphElement::d_set_coordsys(const GCoordSys& ,
				    const GCoordSys& abscs, const GCoordSys&)
{
    _origin = abscs.O;
}


/**
FUNCTION
    d_get_clip()
DESCRIPTION
    Returns the clip.
**/
GRegion* GoGraphElement::d_get_clip() const
{
    const GRegion* r = &_clip;
    return (GRegion*)r;
}


/**
FUNCTION
    m_dup()
DESCRIPTION
    Dups a Go.
**/
void GoGraphElement::m_dup
    (const Go* src, Go* copy_par, Go* prv, GBool donext, GBool dochildren)
{
    Go::m_dup(src, copy_par, prv, donext, dochildren);
    
    const GoGraphElement* orig = (const GoGraphElement*) src;
    
    _parent = copy_par;
    _next = NULL;
    // cast to avoid C++ATT (unlegitimate) warning;
    _clip = ((GoGraphElement*)orig)->_clip;
    _origin = orig->_origin;
    in_destroy = orig->in_destroy;
    
    if (dochildren)
	for (Go* tmp = orig->_child; tmp; tmp = tmp->Next()) {
	    _child = tmp->Dup(this, NULL, GTRUE, GTRUE);
	    if (_child) 
		break;
	}

    if (donext)
	for (Go* orig_nxt = orig->_next; orig_nxt; orig_nxt = orig_nxt->Next())
	    if (_next = orig_nxt->Dup(copy_par, this, GTRUE, dochildren))
		break;
}



/**
FUNCTION
    m_metaclass()
DESCRIPTION
    Returns the metaclass.
**/
GoMetaclass* GoGraphElement::m_metaclass () const
{
    return GoGraphElement::metaclass();
}



static GActionRec empty_action_array[] = { { NULL, NULL } };

/**
FUNCTION
    m_actions()
DESCRIPTION
    
**/
const GActionRec* GoGraphElementMetaclass::m_actions() const
{
    return empty_action_array;
}


/**
FUNCTION
    d_add_child()
DESCRIPTION
    Adds a child.
**/
void GoGraphElement::d_add_child (const Go* ch, int index)
{
    if (index == 0)
	_child = (Go*)ch;
}


/**
FUNCTION
    d_get_previous()
DESCRIPTION
    
**/
Go* GoGraphElement::d_get_previous () const
{
    if (!_parent)
	return NULL;
    
    return _parent->Child(_parent->childIndex(this) - 1);
}


/**
FUNCTION
    d_set_previous()
DESCRIPTION
    
**/
void GoGraphElement::d_set_previous (const Go*)
{
}


/**
FUNCTION
    d_remove_child()
DESCRIPTION
    Removes a child.
**/
void GoGraphElement::d_remove_child (const Go* child, int )
{
    if (child == _child)
	_child = _child->Next(0);
}


/**
FUNCTION
    change_relcs()
DESCRIPTION
    
**/
GBool change_relcs(void*, const Go*, GCoordSys* cs, const GCoordSys*,
		  const GCoordSys*)
{
    GVector I1(1,0);
    GVector J1(0,1);
    if (cs->I != I1)
	cs->I = I1;
    if (cs->J != J1)
	cs->J = J1;
    return GTRUE;
}


/**
FUNCTION
    accept_reparent()
DESCRIPTION
    This method is a reparent precondition.
    A GoGraphElement may only be reparented on a GoGraphStruct object or
    on NULL.
**/
GBool accept_reparent(void* , const Go* go,
		     const Go* new_parent,
		     const Go** )
{
    if (go->getMetaclass()->isSubclass(GoLink::metaclass())) {
	if (new_parent)
	    if (new_parent->getMetaclass()->isSubclass(GoGraphStruct::metaclass())) {
		return GTRUE;
	    } else {
		GWarning("GoGraphElement::reparent: parent has to be an instance of GoTree or GoGraph classes.");
		return GFALSE;
	    }
	else {
	    if (((GoGraphElement*)go)->in_destroy)
		return GTRUE;
	    else
		return GFALSE;
	}
    }
    
    if (go->getMetaclass()->isSubclass(GoTreeNode::metaclass())) {
	if (new_parent)
	    if (new_parent->getMetaclass()->isSubclass(GoTree::metaclass())) {
		return GTRUE;
	    } else {
		GWarning("GoGraphElement::reparent: parent has to be an instance of GoTree class.");
		return GFALSE;
	    }
	else {
	    if (((GoGraphElement*)go)->in_destroy)
		return GTRUE;
	    else
		return GFALSE;
	}
    }

    if (go->getMetaclass()->isSubclass(GoGraphNode::metaclass())) {
	if (new_parent)
	    if (new_parent->getMetaclass()->isSubclass(GoGraph::metaclass())) {
		return GTRUE;
	    } else {
		GWarning("GoGraphElement::reparent: parent has to be an instance of GoGraph class.");
		return GFALSE;
	    }
	else {
	    if (((GoGraphElement*)go)->in_destroy)
		return GTRUE;
	    else
		return GFALSE;
	}
    }
    return GTRUE;
}



                           /********************
			    * PUBLIC FUNCTIONS *
			    ********************/

/**
FUNCTION
    GoGraphElement()
DESCRIPTION
    constructor.
**/
GoGraphElement::GoGraphElement () : _origin(0,0)
{
    _parent = NULL;
    _child = NULL;
    _next = NULL;
    in_destroy = GFALSE;
}

/**
FUNCTION
    ~GoGraphElement()
DESCRIPTION
    destructor.
**/
 GoGraphElement::~GoGraphElement ()
{
}



/**
FUNCTION
    int GoGraphElement::save();
DESCRIPTION
    ** static member, protected **
**/
int GoGraphElement::save(const GoGraphElement* gr, GoStream* gostream) {
    for (Go* ch = (Go*)gr->Child(); ch; ch = ch->Next())
	if (ch->callSavePrecond()) {
	    if (!gostream->saveObj(ch, ch->getMetaclass()->classname(),
				   "child"))
		return GFALSE;
	}
    if (!gostream->saveObj(NULL, NULL, "end of children"))
	return GFALSE;
	    
    if (!gostream->writePoint(gr->origin(), "origin"))
	return GFALSE;
    
    return(Go::save(gr, gostream));
}


/**
FUNCTION
    int GoGraphElement::readParams();
DESCRIPTION
    
**/
int GoGraphElement::readParams(GoGraphElement* gr, GoStream* gostream) {
    const char* class_name;
    for (;;)
    {
	void* next_child;
	if (!gostream->restoreObj(&next_child, &class_name))
	    return GFALSE;
	if (next_child)
	    ((Go*)next_child)->DoReparent(gr);
	else
	    break;
    }

    GPoint p;
    if (!gostream->readPoint(&p))
	return GFALSE;
    gr->_origin = p;

    return Go::readParams(gr, gostream);
}


