#ifdef REV_INFO
static char rcsid[] = "$Id: GraphStr.cc,v 1.9 1992/12/21 16:22:51 jacomet Exp $";
#endif
/*****************************************************************************
MODULE:  BULL Graphic Objects
FILE:    GraphStr.cc
AUTHOR:  Aline Jacomet
DATE:    January 11th 1991

DESCRIPTION:
         Definition of the graph structure object abstract class.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <assert.h>
#include <Go/GraphStr.h>
#include <Go/mGraphStr.h>
#include <Go/Tools/Error.h>
#include <Go/Link.h>
#include <Go/Node.h>
#include <Go/TreeNode.h>
#include <Go/Graph.h>
#include <Go/Tree.h>
#include <Go/GoStream.h>
#include <Go/GoInterf.h>


GAssoc save_precond;


                           /*********************
			    * PRIVATE FUNCTIONS *
			    *********************/


static GoGraphStructMetaclass* gs_metaclass = NULL;

GoGraphStructMetaclass* GoGraphStruct::metaclass ()
{
    if (gs_metaclass == NULL)
    {
	gs_metaclass = new GoGraphStructMetaclass;
	gs_metaclass->superclass()->AddSubclass(gs_metaclass);
    }

    return gs_metaclass;
}

GoGraphStructMetaclass::~GoGraphStructMetaclass ()
{
    if (this == gs_metaclass)
    {
	superclass()->RemoveSubclass(this);
	gs_metaclass = NULL;
    }
}


/**
FUNCTION
    m_classname()
DESCRIPTION
    Returns the ClassName.
**/
const char* GoGraphStructMetaclass::m_classname () const
{
    return "GoGraphStruct";
}


/**
FUNCTION
    m_superclass()
DESCRIPTION
    Returns the SuperClass.
**/
GoMetaclass* GoGraphStructMetaclass::m_superclass () const
{
    return Go::metaclass();
}


/**
FUNCTION
    m_dup()
DESCRIPTION
    Dups a Go.  
**/
void GoGraphStruct::m_dup (const Go* src, Go* copy_par, Go* copy_prev,
			   GBool do_next, GBool do_children)
{
  Go::m_dup(src, copy_par, copy_prev, do_next, do_children);
    
    const GoGraphStruct* orig = (const GoGraphStruct*)src;
    
    _parent = copy_par;
    _prev = copy_prev;
    _next = NULL;
    
    _abscs = orig->_abscs;
    _relcs = orig->_relcs;
    // cast to avoid C++ATT (unlegitimate) warning;
    _clip = ((GoGraphStruct*)orig)->_clip;
    _background = orig->_background;
    brush().modify(orig->brush().shBrush());
    // cast to avoid C++ATT (unlegitimate) warning;
    my_clip = ((GoGraphStruct*)orig)->my_clip; 
    app_function.proc = orig->app_function.proc;
    app_function.closure = orig->app_function.closure;
    
    nb_node = orig->nb_node;
    nb_link = orig->nb_link;
    link_type = orig->link_type;
    anchor_type = orig->anchor_type;
    position_type = orig->position_type;

    Go* child = NULL;
    for (Go* orig_ch = orig->Child(); orig_ch; orig_ch = orig_ch->Next())
	if (do_children &&
	    (child = orig_ch->Dup(this, NULL, GTRUE, do_children))) {
	    break;
	} else
	    _children.add(NULL);
    
    for (Go* next = child; next; next = next->Next()) {
	if (next->getMetaclass()->isSubclass(GoNode::metaclass())) 
	{
	    nb_node--;
	    _children.add(next);
	}
	else if (next->getMetaclass()->isSubclass(GoLink::metaclass()))
	{
	    nb_link--;
	    _children.add(next);
	}
	else
	    _children.add(do_children ? next : NULL);
    }
    
    if (do_next) 
	for (Go* tmp = orig->_next; tmp; tmp = tmp->Next()) {
	    _next = tmp->Dup(copy_par, this, GTRUE, do_children);
	    if (_next)
		break;
	}

    /*
     * All gos are duplicated, so set references
     */
    first_node =
	(GoNode*)Child(orig->childIndex(orig->first_node));
    
    int index_max = orig->_children.nelems();
    Go** o_data = (Go**)orig->_children.data();
    Go** c_data = (Go**)_children.data();
    for (int index = 0; index < index_max; index++) {
	if (o_data[index]->getMetaclass()->isSubclass(GoLink::metaclass()) &&
	    c_data[index] != NULL) {
	    ((GoLink*)c_data[index])->prev_node = (GoNode*)
		Child(orig->childIndex(((GoLink*)o_data[index])->prev_node));
	    ((GoLink*)c_data[index])->next_node = (GoNode*)
		Child(orig->childIndex(((GoLink*)o_data[index])->next_node));
	} else
	if (o_data[index]->getMetaclass()->isSubclass(GoTreeNode::metaclass()) &&
	    c_data[index] != NULL) {
	    if (((GoTreeNode*)o_data[index])->parent_node)
		((GoTreeNode*)c_data[index])->parent_node = (GoTreeNode*)
		    Child(orig->childIndex(((GoTreeNode*)o_data[index])->parent_node));
	    if (((GoTreeNode*)o_data[index])->child_node)
		((GoTreeNode*)c_data[index])->child_node = (GoTreeNode*)
		    Child(orig->childIndex(((GoTreeNode*)o_data[index])->child_node));
	    if (((GoTreeNode*)o_data[index])->next_node)
		((GoTreeNode*)c_data[index])->next_node = (GoTreeNode*)
		    Child(orig->childIndex(((GoTreeNode*)o_data[index])->next_node));
	    if (((GoTreeNode*)o_data[index])->prev_node)
		((GoTreeNode*)c_data[index])->prev_node = (GoTreeNode*)
		    Child(orig->childIndex(((GoTreeNode*)o_data[index])->prev_node));
	    if (((GoTreeNode*)o_data[index])->parent_link)
		((GoTreeNode*)c_data[index])->parent_link = (GoLink*)
		    Child(orig->childIndex(((GoTreeNode*)o_data[index])->parent_link));
	} else
	if (o_data[index]->getMetaclass()->isSubclass(GoGraphNode::metaclass()) &&
	    c_data[index] != NULL) {
	    int prev_max = ((GoGraphNode*)o_data[index])->prev_link.nelems();
	    if (prev_max) {
		GoLink** orig_data =
		    (GoLink**)(((GoGraphNode*)o_data[index])->prev_link.data());
		for (int i = 0; i < prev_max; i++) {
		    GoLink* l =
			(GoLink*) Child(orig->childIndex(orig_data[i]));
		    if (l)
			((GoGraphNode*)c_data[index])->prev_link.add(l);
		}
	    }
	    int next_max = ((GoGraphNode*)o_data[index])->next_link.nelems();
	    if (next_max) {
		GoLink** orig_data =
		    (GoLink**)(((GoGraphNode*)o_data[index])->next_link.data());
		for (int i = 0; i < next_max; i++) {
		    GoLink* l =
			(GoLink*) Child(orig->childIndex(orig_data[i]));
		    if (l)
			((GoGraphNode*)c_data[index])->next_link.add(l);
		}
	    }
	}
    }

    /*
     * If some children are NULL, delete them
     */
    GPointerArray tab;
    int nb = _children.nelems();
    for (int i = 0; i < nb; i++) {
	if (c_data[i] == NULL) {
	    if (o_data[i]->getMetaclass()->isSubclass(GoLink::metaclass())) {
		if (orig->getMetaclass()->isSubclass(GoTree::metaclass())) {
		      tab.add(Child(orig->childIndex(((GoLink*)o_data[i])->next_node)));
		} 
	    } else
		if (o_data[i]->getMetaclass()->isSubclass(GoNode::metaclass())) {
		    if (orig->getMetaclass()->isSubclass(GoTree::metaclass())) {
			if (((GoTreeNode*)o_data[i])->parent_link)
			    tab.add(Child(orig->childIndex(((GoTreeNode*)o_data[i])->parent_link)));
			if (((GoTreeNode*)o_data[i])->child_node) {
			    GoTreeNode *nc, *vc;
			    for (vc = ((GoTreeNode*)o_data[i])->child_node,
				 nc = (GoTreeNode*)Child(orig->childIndex(vc));
				 vc;
				 vc = vc->nextNode(),
				 nc = (GoTreeNode*)Child(orig->childIndex(vc)))
				tab.add(nc);
			}
			// GSet references;
			GoTreeNode* n = ((GoTreeNode*)o_data[i])->next_node;
			GoTreeNode* new_next_node = NULL;
			while (new_next_node == NULL) {
			    if (n) {
				new_next_node = (GoTreeNode*)
				    Child(orig->childIndex(n));
				n = n->nextNode();
			    } else
				break;
			}
			n = ((GoTreeNode*)o_data[i])->prev_node;
			GoTreeNode* new_prev_node = NULL;
			while (new_prev_node == NULL) {
			    if (n) {
				new_prev_node = (GoTreeNode*)
				    Child(orig->childIndex(n));
				n = n->previousNode();
			    } else
				break;
			}
			if (new_prev_node)
			    new_prev_node->next_node = new_next_node;
			if (new_next_node)
			    new_next_node->prev_node = new_prev_node;
			if (((GoTreeNode*)o_data[i])->parent_node)
			    if (((GoTreeNode*)o_data[i])->parent_node->child_node == o_data[i]) {
				n = (GoTreeNode*)Child(orig->childIndex(((GoTreeNode*)o_data[i])->parent_node));
				if (n)
				    n->child_node = new_next_node;
			    }
		    } else {
			int prev_max = ((GoGraphNode*)o_data[i])->prev_link.nelems();
			if (prev_max) {
			    GoLink** orig_data =
				(GoLink**)(((GoGraphNode*)o_data[i])->prev_link.data());
			    for (int j = 0; j < prev_max; j++) {
				GoLink* l =
				    (GoLink*) Child(orig->childIndex(orig_data[j]));
				if (l)
				    tab.add(l);
			    }
			}
			int next_max = ((GoGraphNode*)o_data[i])->next_link.nelems();
			if (next_max) {
			    GoLink** orig_data =
				(GoLink**)(((GoGraphNode*)o_data[i])->next_link.data());
			    for (int j = 0; j < next_max; j++) {
				GoLink* l =
				    (GoLink*) Child(orig->childIndex(orig_data[j]));
				if (l)
				    tab.add(l);
			    }
			}
		    }
		}
	}
    }

    nb = tab.nelems();
    Go** data = (Go**)tab.data();
    for (i = 0; i < nb; i++)
	if (data[i])
	    if (data[i]->getMetaclass()->isSubclass(GoLink::metaclass()) &&
		getMetaclass()->isSubclass(GoGraph::metaclass())) {
		((GoGraph*)this)->removeLink((GoLink*)data[i]);
	    } else
		if (data[i]->getMetaclass()->isSubclass(GoLink::metaclass()) &&
		    getMetaclass()->isSubclass(GoTree::metaclass())) {
		    ((GoLink*)data[i])->in_destroy = GTRUE;
		    data[i]->destroy();
		} else
		    if (data[i]->getMetaclass()->isSubclass(GoTreeNode::metaclass())){
			((GoTree*)this)->removeNode((GoTreeNode*)data[i]);
		    }
    

    _children.rem((void*)NULL, GFALSE);

    if (getMetaclass()->isSubclass(GoGraph::metaclass()))
	if (! first_node) {
	    int ind;
	    Go* c = Child();
	    for(ind = 0;
		c && !c->getMetaclass()->isSubclass(GoGraphNode::metaclass());
		ind++)
		c = Child(ind);
	    first_node = (GoNode*)c;
	}
    if (first_node)
	first_node->clipPostcond(graphstr_set_coordsys, this, GADD);
}


/**
FUNCTION
    GoGraphStructMetaclass()
DESCRIPTION
    class constructor
**/
GoGraphStructMetaclass::GoGraphStructMetaclass()
{
}


/**
FUNCTION
    d_get_assoc()
DESCRIPTION
    Returns the assoc.
**/
GAssoc* GoGraphStruct::d_get_assoc() const
{
    const GAssoc* a = &_assoc;
    return (GAssoc*)a;
}


/**
FUNCTION
    d_get_parent()
DESCRIPTION
    Returns the parent id.
**/
Go* GoGraphStruct::d_get_parent() const
{
    return _parent;
}


/**
FUNCTION
    d_get_next()
DESCRIPTION
    Returns the next id.
**/
Go* GoGraphStruct::d_get_next() const
{
    return _next;
}


/**
FUNCTION
    d_get_first_child()
DESCRIPTION
    Returns the first_child id.
**/
Go* GoGraphStruct::d_get_first_child() const
{
    if (_children.nelems() > 0) {
	return ((Go*)(_children.data(0)));
    } else
	return NULL;
}


/**
FUNCTION
    d_store_children_array()
DESCRIPTION
    
**/
GBool GoGraphStruct::d_store_children_array () const
{
    return GTRUE;
}


/**
FUNCTION
    d_get_children()
DESCRIPTION
    Returns the children array
**/
Go** GoGraphStruct::d_get_children (int& count) const
{
    count = _children.nelems();
    return (Go**)_children.data();
}


/**
FUNCTION
    d_set_next()
DESCRIPTION
    Sets the next id.
**/
void GoGraphStruct::d_set_next (const Go* go)
{
    _next = (Go*)go;
}


/**
FUNCTION
    d_set_tree()
DESCRIPTION
    Sets the next, previous and parent id.
**/
void GoGraphStruct::d_set_tree (const Go* parent, const Go* previous,
				const Go* next)
{
    _parent = (Go*)parent;
    _prev = (Go*)previous;
    _next = (Go*)next;
}


/**
FUNCTION
    d_store_relcs()
DESCRIPTION
    
**/
GBool GoGraphStruct::d_store_relcs () const
{
    return GTRUE;
}


/**
FUNCTION
    d_get_relcs()
DESCRIPTION
    Returns the relative coordinate system.
**/
const GCoordSys& GoGraphStruct::d_get_relcs(const GCoordSys*) const
{
    return _relcs;
}




/**
FUNCTION
    d_store_abscs()
DESCRIPTION
    
**/
GBool GoGraphStruct::d_store_abscs () const
{
    return GTRUE;
}


/**
FUNCTION
    d_get_abscs()
DESCRIPTION
    Returns the absolute coordinate system.
**/
const GCoordSys& GoGraphStruct::d_get_abscs(const GCoordSys*) const
{
    return _abscs;
}


/**
FUNCTION
    d_set_coordsys()
DESCRIPTION
    Sets the coordinate systems.
**/
void GoGraphStruct::d_set_coordsys(const GCoordSys& relcs,
				   const GCoordSys& abscs,
				   const GCoordSys&)
{
    _relcs = relcs;
    _abscs = abscs;
}


/**
FUNCTION
    d_get_clip()
DESCRIPTION
    Returns the clip.
**/
GRegion* GoGraphStruct::d_get_clip() const
{
    const GRegion* c = &_clip;
    return (GRegion*)c;
}



/**
FUNCTION
    m_metaclass()
DESCRIPTION
    Returns the metaclass.
**/
GoMetaclass* GoGraphStruct::m_metaclass () const
{
    return GoGraphStruct::metaclass();
}


static void graphstr_set_background (GoGraphStruct* g, GBool b)
{
    g->setBackground(b);
}

static void graphstr_set_position_type (GoGraphStruct* g, ushort pos)
{
    g->setPositionType(pos);
}

static void graphstr_set_anchoring_type (GoGraphStruct* g, ushort pos)
{
    g->setAnchoringType(pos);
}

static void graphstr_make_position
    (GoGraphStruct* g, GoNode* node, const GPoint* p)
{
    g->makePosition(node, *p);
}


static GActionRec action_array[] =
{ 
  { "setBackground(GBool)",
	(GActionProc) graphstr_set_background },
  { "setPositionType(ushort)",
	(GActionProc) graphstr_set_position_type },
  { "setAnchoringType(ushort)",
	(GActionProc) graphstr_set_anchoring_type },
  { "makePosition(GoNode*=NULL, const GPoint&=*(GPoint*)NULL)",
	(GActionProc) graphstr_make_position },
  { NULL, NULL } };


/**
FUNCTION
    m_actions()
DESCRIPTION
    Actions for user interface.
**/
const GActionRec* GoGraphStructMetaclass::m_actions() const
{
    return action_array;
}


/**
FUNCTION
    m_draw()
DESCRIPTION
    Draws the graphic object.
**/
void GoGraphStruct::m_draw(const GRegion&) const
{
    if (_background) {
	GBox box = GEmptyBox();
	GRegion tmp_reg;
	Go** data = (Go**)_children.data();
	int index_max = _children.nelems();
	for (int i = 0; i < index_max; i++)
	    if (data[i]->mapped() &&
		data[i]->getMetaclass()->isSubclass(GoNode::metaclass())) {
		data[i]->addClip(tmp_reg);
		box += tmp_reg.bigbox();
	    }
	const GBrush* b = &_brush;
	interf()->rectangle(NULL, b, GUnitCoordSys(),
			    GPoint(box.xmin, box.ymin),
			    box.width(), box.height());
    }
}


/**
FUNCTION
    m_wired_draw()
DESCRIPTION
    
**/
void GoGraphStruct::m_wired_draw(const GCoordSys& cs, const GPen& pen,
				 GQuality quality) const
{
    GRegion tmp_reg;
    addClip(tmp_reg);
    GBox clip = tmp_reg.bigbox();
    GCoordSys c;
    c = cs * (~Abscs());
    interf()->rectangle(&pen, NULL, c, GPoint(clip.xmin, clip.ymin),
			clip.width(), clip.height(), quality);
}


/**
FUNCTION
    m_add_clip()
DESCRIPTION
    Adds to region the own clip of the object.
**/
void GoGraphStruct::m_add_clip(GRegion& region) const
{
    if (_background) {
	Go** data = (Go**)_children.data();
	int index_max = _children.nelems();
	GBox box(GEmptyBox());
	GRegion r;
	
	for (int i = 0; i < index_max; i++)
	    if (data[i]->mappable() &&
		data[i]->getMetaclass()->isSubclass(GoNode::metaclass())) {
		r = GEmptyRegion();
		data[i]->addClip(r);
		box += r.bigbox();
	    }
	region += box;
    }
}


/**
FUNCTION
    m_add_opaque_clip()
DESCRIPTION
    Adds to region the own opaque clip of the object.
**/
void GoGraphStruct::m_add_opaque_clip(GRegion& region) const
{
    m_add_clip(region);
}


/**
FUNCTION
    m_contains()
DESCRIPTION
    Returns GTRUE if absPoint is contained in the object, otherwise it returns
    GFALSE
**/
GBool GoGraphStruct::m_contains(const GPoint& absPoint, GNumber maxDist) const
{
    if (_background) {
	GRegion r;
	addMyClip(r);
	return r.contains(absPoint, maxDist);
    } else
	return GFALSE;
}


/**
FUNCTION
    d_add_child()      
DESCRIPTION
    Adds a child.
**/
void GoGraphStruct::d_add_child (const Go* ch, int index)
{
    if (index == GUNKNOWN_INDEX)
	if (ch->getMetaclass()->isSubclass(GoLink::metaclass()))
	    _children.add(ch, 0);
	else
	    _children.add(ch);
    else
	_children.add(ch, index);
    if (ch->getMetaclass()->isSubclass(GoNode::metaclass()))
	nb_node++;
    if (ch->getMetaclass()->isSubclass(GoLink::metaclass()))
	nb_link++;
}


/**
FUNCTION
    d_get_previous()      
DESCRIPTION
    Returns the previous object.
**/
Go* GoGraphStruct::d_get_previous () const
{
    return _prev;
}



/**
FUNCTION
    d_set_previous()
DESCRIPTION
    Sets the previous object.
**/
void GoGraphStruct::d_set_previous (const Go* go)
{
    _prev = (Go*)go;
}

/**
FUNCTION
    d_remove_child()
DESCRIPTION
    Removes a child.
**/
void GoGraphStruct::d_remove_child (const Go* child, int child_index)
{
    if (child_index != GUNKNOWN_INDEX)
	_children.rem(child_index);
    else
	_children.rem(child);
    if (child->getMetaclass()->isSubclass(GoNode::metaclass()))
	nb_node--;
    if (child->getMetaclass()->isSubclass(GoLink::metaclass()))
	nb_link--;
}


/**
FUNCTION
    graphstr_brush_changed(void* graphstr, GBrush* brush, ulong mask)
DESCRIPTION
    virtual private
    This function is a modify_post_condition of str->_brush
    WARNING: it is assumed that brush **IS** str->_brush
**/
void graphstr_brush_changed(void* closure, GBrush* brush, ulong )
{
    GoGraphStruct* str = (GoGraphStruct*)closure;
    assert(brush == &str->_brush);
    if (str->background() && str->mapped()) {
	str->MyClipHasChanged();
	str->RedrawForeOnly();
    }
}


void graphstr_set_coordsys (void* closure, Go* go)
{
    GoGraphStruct* graphstr = (GoGraphStruct*) closure;
    graphstr->set_coordsys(go);
}


/**
FUNCTION
    set_coordsys()
DESCRIPTION
    virtual private
    This function is a clip_post_condition called when the clip of the
    first_node is modified. Ensures that coordinate system always is the center
    of the first node.
**/
void GoGraphStruct::set_coordsys(Go* go)
{
    GRegion tmp_reg;
    go->addClip(tmp_reg);
    GBox b = tmp_reg.bigbox();
    if (!b.empty()) {
	GCoordSys a = Abscs();
	GCoordSys cs(GPoint((b.xmin + b.xmax) / 2, (b.ymin + b.ymax) /2),
		    a.I, a.J);
	if (cs != a)
	    SetThisAbscs(cs);
    }
}


/**
FUNCTION
    left_right_placement()
DESCRIPTION
    
**/
void GoGraphStruct::left_right_placement(GoNode*, const GPoint&)
{
}


/**
FUNCTION
    up_bottom_placement()
DESCRIPTION
    
**/
void GoGraphStruct::up_bottom_placement(GoNode*, const GPoint&)
{
}


/**
FUNCTION
    bottom_up_placement()
DESCRIPTION
    
**/
void GoGraphStruct::bottom_up_placement(GoNode*, const GPoint&)
{
}


/**
FUNCTION
    right_left_placement()
DESCRIPTION
    
**/
void GoGraphStruct::right_left_placement(GoNode*, const GPoint&)
{
}



                           /***********************
			    * PROTECTED FUNCTIONS *
			    ***********************/



/**
FUNCTION
    GoGraphStruct()
DESCRIPTION
    constructor.
**/
GoGraphStruct::GoGraphStruct() : _children(64)
{
    _parent = NULL;
    _next = NULL;
    _prev = NULL;
    _background = GFALSE;
    my_clip = GEmptyBox();
    app_function.proc = NULL;
    app_function.closure = NULL;
    first_node = NULL;
    nb_node = 0;
    nb_link = 0;
    link_type = GoLink::metaclass();
    anchor_type = GCENTER;
    position_type = GLEFT_TO_RIGHT;
    _brush.modifyPostcond(graphstr_brush_changed, this, GADD);
}


/**
FUNCTION
    ~GoGraphStruct()
DESCRIPTION
    destructor.
**/
 GoGraphStruct::~GoGraphStruct ()
{
}



                           /********************
			    * PUBLIC FUNCTIONS *
			    ********************/


/**
FUNCTION
    makePosition()
DESCRIPTION
    Calls node's placement algorythm.
**/
void GoGraphStruct::makePosition(GoNode* node, const GPoint& point)
{
    if (!node) 
	node = first_node;

    if (getMetaclass()->isSubclass(GoGraph::metaclass())) {
	if (app_function.proc)
	    app_function.proc(app_function.closure, this, node, point);
	else
	    GWarning("GoGraphStruct::makePosition: there is no application's function");
    } else
	if (position_type == GLEFT_TO_RIGHT)
	    left_right_placement(node, point);
	else
	    if (position_type == GRIGHT_TO_LEFT)
		right_left_placement(node, point);
	else
	    if (position_type == GUP_TO_BOTTOM)
		up_bottom_placement(node, point);
	    else
		if (position_type == GBOTTOM_TO_UP)
		    bottom_up_placement(node, point);
		else
		    if (position_type == GAPPLICATION_DEFINE)
			if (app_function.proc)
			    app_function.proc(app_function.closure, this, node,
					      point);
			else
			    
			    GWarning("GoGraphStruct::makePosition: there is no application's function");
}


/**
FUNCTION
    applicationFunction()
DESCRIPTION
    Sets the application placement algorythm.
**/
void GoGraphStruct::applicationFunction(void cbproc(void*, Go*, GoNode*,
						    const GPoint&),
					void* closure)
{
    app_function.proc = cbproc;
    app_function.closure = closure;
}


/**
FUNCTION
    defaultLinkType()
DESCRIPTION
    Sets the default link type.
**/
GBool GoGraphStruct::defaultLinkType(GoMetaclass* type)
{
    if (type->isSubclass(GoLink::metaclass())) {
	link_type = type;
	return GTRUE;
    } else
	return GFALSE;
}


/**
FUNCTION
    background()
DESCRIPTION
    Sets the background.
**/
void GoGraphStruct::background(GBool back)
{
    if (back == _background)
	return;

    GBool is_mapped = mapped();
    if (is_mapped && !back)
	RedrawBackAndFore();
    _background = back;
    my_clip = GEmptyBox();
    if (_background) {
	// Calculate new clip;
	GRegion r;
	for (Go* ch = child(); ch; ch = ch->next())
	    if (ch->mappable() &&
		ch->getMetaclass()->isSubclass(GoNode::metaclass())) {
		r = GEmptyRegion();
		ch->addClip(r);
		my_clip += r.bigbox();
	    }
    }
    MyClipHasChanged();
    if (is_mapped && back)
	RedrawForeOnly();
}



/**
FUNCTION
    positionType()
DESCRIPTION
    
**/
GBool GoGraphStruct::positionType(ushort pos)
{
    if (/* pos >= GLEFT_TO_RIGHT && */
	pos <= GAPPLICATION_DEFINE) {
	position_type = pos;
	return GTRUE;
    } else
	return GFALSE;
}


/**
FUNCTION
    anchoringType()
DESCRIPTION
    
**/
GBool GoGraphStruct::anchoringType(ushort pos)
{
    if (/* pos >= GCENTER && */
	pos <= GMIDDLE_HEIGHT) {
	anchor_type = pos;
	return GTRUE;
    } else
	return GFALSE;
}


/**
FUNCTION
    int GoGraphStruct::save();
DESCRIPTION
    ** static member, protected **
**/
int GoGraphStruct::save(const GoGraphStruct* gr, GoStream* gostream) {
    // first node and children are saved by the subclass;

    if (! (gostream->writeShort(gr->numberNodes(), "nb nodes") &&
	   gostream->writeShort(gr->numberLinks(), "nb links") &&
	   gostream->writeString(gr->defaultLinkType()->classname(),
				 "default link type") &&
	   gostream->writeShort(gr->anchoringType(), "anchor type") &&
	   gostream->writeShort(gr->positionType(), "position type") &&
	   gostream->writeCoordSys(gr->Abscs(), "abscs") &&
	   gostream->writeBool(gr->background(), "background") &&
	   gostream->saveObj(gr->brush().shBrush(), "GShBrush", "brush")))
	return GFALSE;
	    
    return(Go::save(gr, gostream));
}


/**
FUNCTION
    int GoGraphStruct::readParams();
DESCRIPTION
    
**/
int GoGraphStruct::readParams(GoGraphStruct* gr, GoStream* gostream) {
    char str[30];
    const char* class_name = str;
    short sh;
    if (!gostream->readShort(&sh))
	return GFALSE;
    gr->nb_node = sh;
    
    if (!gostream->readShort(&sh))
	return GFALSE;
    gr->nb_link = sh;
    
    if (!gostream->readString(30, (char**)&class_name))
	return GFALSE;
    gr->link_type = Go::metaclass(class_name);
    if (class_name != str)
	GFREE(class_name);
    
    if (!gostream->readShort(&sh))
	return GFALSE;
    gr->anchor_type = sh;
    
    if (!gostream->readShort(&sh))
	return GFALSE;
    gr->position_type = sh;
    
    GCoordSys cs;
    if (!gostream->readCoordSys(&cs))
	return GFALSE;
    gr->_abscs = cs;
    gr->_relcs = gr->parentCoordsys().from(cs);
    
    GBool b;
    if (!gostream->readBool(&b))
	return GFALSE;
    gr->_background = b;
    
    GShBrush* shbrush = NULL;
    if (! gostream->restoreObj((void**) &shbrush, &class_name))
	return GFALSE;
    if (strcmp(class_name, "GShBrush") != 0)
	GWarning("GraphStruct::readParams: can't restore brush.");
    else
	gr->brush().modify(shbrush);
    
    return Go::readParams(gr, gostream);
}
