#ifdef REV_INFO
static char rcsid[] = "$Id: Node.cc,v 1.8 1992/12/10 16:20:23 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  BULL Graphic Objects
FILE:    Node.cc
AUTHOR:  Aline Jacomet
DATE:    January 11th 1991

DESCRIPTION:
         Definition of the node object abstract class.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <assert.h>
#include <Go/Node.h>
#include <Go/mNode.h>
#include <Go/GoStream.h>
#include <Go/GoInterf.h>


/*
 *------------
 * CLASS PART
 *------------
 */



static GoNodeMetaclass* node_metaclass = NULL;

/**
FUNCTION
    GoNode::metaclass()
DESCRIPTION
    
**/
GoNodeMetaclass* GoNode::metaclass() {
    if (node_metaclass == NULL) {
	node_metaclass = new GoNodeMetaclass;
	node_metaclass->superclass()->AddSubclass(node_metaclass);
    }
    return node_metaclass;
}


/**
FUNCTION
    ~GoNodeMetaclass()
DESCRIPTION
    class destructor
**/
GoNodeMetaclass::~GoNodeMetaclass() {
    if (this == node_metaclass) {
        superclass()->RemoveSubclass(this);
	node_metaclass = NULL;
    }
    clipPostcond(nodemc_call_modif_clip, this, GREMOVE);
}


/**
FUNCTION
    GoNodeMetaclass::GoNodeMetaclass()
DESCRIPTION
    class constructor
**/
GoNodeMetaclass::GoNodeMetaclass() {
    clipPostcond(nodemc_call_modif_clip, this, GADD);
}


/**
FUNCTION
    const char* GoNodeMetaclass::m_classname() const
DESCRIPTION

**/
const char* GoNodeMetaclass::m_classname() const
{
    return "GoNode";
}


/**
FUNCTION
    GoMetaclass* GoNodeMetaclass::m_superclass() const
DESCRIPTION

**/
GoMetaclass* GoNodeMetaclass::m_superclass() const
{
    return GoGraphElement::metaclass();
}


/**
FUNCTION
    Go* GoNodeMetaclass::m_create();
DESCRIPTION

**/
Go* GoNodeMetaclass::m_create () const
{
    return NULL;
}


/**
FUNCTION
    GoNode::m_dup()
DESCRIPTION

**/
void GoNode::m_dup(const Go* src, Go* par, Go* prv, GBool b, GBool dochildren)
{
    GoGraphElement::m_dup(src, par, prv, b, dochildren);

    const GoNode* s = (const GoNode*) src;

    _background = s->_background;
    if (s->p_brush) {
	p_brush = new GBrush;
	brush().modify(s->brush().shBrush());
	p_brush->modifyPostcond(node_brush_modified, this, GADD);
    }
    width_2 = s->width_2;
    height_2 = s->height_2;
}


/**
FUNCTION
    m_metaclass()
DESCRIPTION
    Returns the metaclass.
**/
GoMetaclass* GoNode::m_metaclass () const
{
    return GoNode::metaclass();
}



static void node_set_background (GoNode* n, GBool b) { n->background(b); }

static GActionRec goNode_action_array[] =
{ 
  { "setBackground(GBool)", (GActionProc) node_set_background },
  { NULL, NULL }
};

/**
FUNCTION
    m_actions()
DESCRIPTION
    
**/
const GActionRec* GoNodeMetaclass::m_actions() const
{
    return goNode_action_array;
}


/*
 *--------------
 * PRIVATE DATA
 *--------------
 */



/**
FUNCTION
    m_draw()
DESCRIPTION
    Draws the graphic object.
**/
void GoNode::m_draw(const GRegion&) const
{
    if (_background) {
	GBox box = GEmptyBox();
	GRegion tmp_reg;
	for (Go* c = child(); c; c = c->next())
	    if (c->mapped()) {
		tmp_reg = GEmptyRegion();
		c->addClip(tmp_reg);
		box += tmp_reg.bigbox();
	    }
	interf()->rectangle(NULL, p_brush, GUnitCoordSys(),
			    GPoint(box.xmin, box.ymin),
			    box.width(), box.height());
    }
}

/**
FUNCTION
    m_wired_draw()
DESCRIPTION
    
**/
void GoNode::m_wired_draw(const GCoordSys&, const GPen&, GQuality) const
{
}


/**
FUNCTION
    m_add_clip()
DESCRIPTION
    Adds to region the own clip of the object.
**/
void GoNode::m_add_clip(GRegion& region) const
{
    if (_background) {
	GBox box(GEmptyBox());
	GRegion r;
	
	for (Go* c = child(); c; c = c->next())
	    if (c->mappable()) {
		r = GEmptyRegion();
		c->addClip(r);
		box += r.bigbox();
	    }
	region += box;
    }
}


/**
FUNCTION
    m_add_opaque_clip()
DESCRIPTION
    Adds to region the own opaque clip of the object.
**/
void GoNode::m_add_opaque_clip(GRegion& region) const
{
    m_add_clip(region);
}


/**
FUNCTION
    m_contains()
DESCRIPTION
    Returns GTRUE if absPoint is contained in the object, otherwise it returns
    GFALSE
**/
GBool GoNode::m_contains(const GPoint& absPoint, GNumber maxDist) const
{
    if (_background) {
	GRegion r;
	addMyClip(r);
	return r.contains(absPoint, maxDist);
    } else
	return GFALSE;
}


/**
FUNCTION
    node_brush_modified(void* node, GBrush* brush, ulong mask)
DESCRIPTION
    virtual private
    This function is a modify_post_condition of n->_brush
    WARNING: it is assumed that brush **IS** n->_brush
**/
void node_brush_modified(void* closure, GBrush* brush, ulong )
{
    GoNode* n = (GoNode*)closure;
    assert(brush == n->p_brush);
    if (n->background() && n->mapped()) {
	n->MyClipHasChanged();
	n->RedrawForeOnly();
    }
}


/**
FUNCTION
    nodemc_call_modif_clip()
DESCRIPTION
    This function is a clip postcondition.
**/
void nodemc_call_modif_clip(void* closure, Go* go)
{
    GoNodeMetaclass* mc = (GoNodeMetaclass*) closure;
    mc->modif_clip(go);
}


/**
FUNCTION
    modif_clip()
DESCRIPTION
    Sets the origin of the coordinate system at the center of node's clip, and
    redraw the tree or graph if necessary.
**/
void GoNodeMetaclass::modif_clip(Go* node)
{
    /*
     * Sets the origin of the coordinate system at the center of the clip;
     */
    GoNode* n = (GoNode*)node;
    // clip without sensor;
    GRegion tmp_reg;
    GBox big(GEmptyBox());
    for (Go* c = n->child(); c; c = c->next())
	if (c->mappable()) {
	    tmp_reg = GEmptyRegion();
	    c->addClip(tmp_reg);
	    big += tmp_reg.bigbox();
	}
    if (! big.empty()) {
	GCoordSys cs(GPoint((big.xmin+big.xmax)/2, (big.ymin+big.ymax)/2),
		    n->Abscs().I, n->Abscs().J);
	if (n->_background && n->mapped()) {
	    if (n->width_2 != big.width() / 2 ||
		n->height_2 != big.height() / 2)
		n->RedrawBackAndFore(GBox(GPoint(n->_origin.x - n->width_2,
					       n->_origin.y - n->height_2),
					 GPoint(n->_origin.x + n->width_2,
					       n->_origin.y + n->height_2)));
	}
	if (cs != n->Abscs())
	    n->SetThisAbscs(cs);
	if (n->width_2 != big.width() / 2 ||
	    n->height_2 != big.height() / 2) {
	    n->width_2 = big.width() / 2;
	    n->height_2 = big.height() / 2;
	    if (n->_background && n->mapped()) {
		n->MyClipHasChanged();
		n->RedrawForeOnly();
	    }
	}
	GoGraphStruct* str = (GoGraphStruct*)n->parent();
	if (str->background() && str->mapped()) {
            // Calculate new clip;
	    GBox box(GEmptyBox());
	    GRegion r;
	    for (Go* ch = str->child(); ch; ch = ch->next())
		if (ch->mappable() &&
		    ch->getMetaclass()->isSubclass(GoNode::metaclass())) {
		    r = GEmptyRegion();
		    ch->addClip(r);
		    box += r.bigbox();
		}
	    if (box != str->my_clip) {
		str->RedrawBackAndFore(str->my_clip);
		str->my_clip = box;
		str->MyClipHasChanged();
		str->RedrawForeOnly();
	    }
	}
    }
}
    


/*
 *--------------
 * PUBLIC DATA
 *--------------
 */


/**
FUNCTION
    GoNode()
DESCRIPTION
    constructor
**/
 GoNode::GoNode()
{
    p_brush = NULL;
    _background = GFALSE;
    width_2 = 0;
    height_2 = 0;
}


/**
FUNCTION
    ~GoNode()
DESCRIPTION
    destructor
**/
 GoNode::~GoNode()
{
    if (p_brush)
	delete p_brush;
}


/**
FUNCTION
    background()
DESCRIPTION
    Sets the background.
**/
void GoNode::background(GBool back)
{
    if (back == _background)
	return;

    if (back && !p_brush) {
	p_brush = new GBrush;
	p_brush->modifyPostcond(node_brush_modified, this, GADD);
    }
    GBool is_mapped = mapped();
    if (is_mapped && !back)
	RedrawBackAndFore();
    _background = back;
    MyClipHasChanged();
    if (is_mapped && back)
	RedrawForeOnly();
}



/**
FUNCTION
    int GoNode::save();
DESCRIPTION
    ** static member, protected **
**/
int GoNode::save(const GoNode* node, GoStream* gostream) {
    if(! (gostream->writeNumber(node->width_2, "width-2") &&
	   gostream->writeNumber(node->height_2, "height-2") &&
	   gostream->writeBool(node->_background, "background")))
	return GFALSE;
    if (node->p_brush) {
	if (!gostream->saveObj(node->brush().shBrush(), "GShBrush", "brush"))
	    return GFALSE;
    } else
	if (!gostream->saveObj(NULL, NULL, "brush"))
	    return GFALSE;
    return (GoGraphElement::save(node, gostream));
}


/**
FUNCTION
    int GoNode::readParams();
DESCRIPTION
    
**/
int GoNode::readParams(GoNode* node, GoStream* gostream) {
    GNumber n;
    if (!gostream->readNumber(&n))
	return GFALSE;
    node->width_2 = n;
    if (!gostream->readNumber(&n))
	return GFALSE;
    node->height_2 = n;

    GBool b;
    if (!gostream->readBool(&b))
	return GFALSE;
    node->_background = b;

    GShBrush* shbrush = NULL;
    const char* class_name;
    if (! gostream->restoreObj((void**) &shbrush, &class_name))
	return GFALSE;

    if (shbrush)
	if (strcmp(class_name, "GShBrush") != 0)
	    GWarning("GoNode::readParams: can't restore brush.");
	else {
	    node->p_brush = new GBrush;
	    node->p_brush->modifyPostcond(node_brush_modified, node, GADD);
	    node->brush().modify(shbrush);
	}
    
    return GoGraphElement::readParams(node, gostream);
}




