#ifdef REV_INFO
static char rcsid[] = "$Id: Line.cc,v 1.4 1992/12/10 16:30:08 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Graphic Objects
FILE:    Line.cc
AUTHOR:  Gilles Benati
DATE:    March 5th, 1992

DESCRIPTION:
    see Line.h header file

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France

  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.

  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <assert.h>
#include <Go/Tools/Error.h>
#include <Go/GoInterf.h>
#include <Go/GoStream.h>
#include <Go/Line.h>


static char lineNfirstPrecond[]   = "p";
static char lineNfirstPostcond[]  = "p";
static char lineNsecondPrecond[]  = "p";
static char lineNsecondPostcond[] = "p";
static char lineNresetPrecond[]   = "p";
static char lineNresetPostcond[]  = "p";

// to clarify pre/postcond code;
#define LINE_PRECOND_PTR(f)  GBool f(void*, const GoLine*, GPoint*)
#define LINER_PRECOND_PTR(f) GBool f(void*, const GoLine*, GPoint*, GPoint*)
#define LINE_POSTCOND_PTR(f) void f(void*, GoLine*)

/*
 *--------
 * ACTIONS
 *--------
 */

// TO BE DEFINED AS A GLOBAL VARIABLE;
static GActionRec empty_action_array[] = { { NULL, NULL } };

static void set_first(GoLine *line, GPoint frst) {
    line->first(frst);
}

static void set_second(GoLine *line, GPoint scnd) {
    line->second(scnd);
}

static void set_first_second(GoLine *line, GPoint frst, GPoint scnd) {
    line->reset(frst, scnd);
}


static GActionRec go_line_action_array[] = {
  { "first(GPoint)",         (GActionProc) set_first },
  { "second(GPoint)",        (GActionProc) set_second },
  { "reset(GPoint, GPoint)", (GActionProc) set_first_second },
  { NULL, NULL }
};



/*
 *--------------------
 * PRE/POSTCONDITIONS
 *--------------------
 */

GBool GoLine::_call_class_first_precond(GPoint* frst) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoLineMetaclass*) mc)->_has_first_precond &&
	    ! mc->CallPreconditions(lineNfirstPrecond, this, frst))
	    return GFALSE;
	// stop when reaching GoLine Metaclass instance;
	if (mc == GoLine::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoLine::_call_class_second_precond(GPoint* scnd) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoLineMetaclass*) mc)->_has_second_precond &&
	    ! mc->CallPreconditions(lineNsecondPrecond, this, scnd))
	    return GFALSE;
	// stop when reaching GoLine Metaclass instance;
	if (mc == GoLine::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}


GBool GoLine::_call_class_reset_precond(GPoint* frst, GPoint* scnd) const {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoLineMetaclass*) mc)->_has_reset_precond &&
	    ! mc->CallPreconditions(lineNresetPrecond, this, frst, scnd))
	    return GFALSE;
	// stop when reaching GoLine Metaclass instance;
	if (mc == GoLine::metaclass())
	    return GTRUE;
    }
    return GTRUE;
}

GBool GoLine::_call_first_precond(GPoint* frst) const {
    return(_has_first_precond ?
	   CallPreconditions(lineNfirstPrecond, frst) : GTRUE);
}

GBool GoLine::_call_second_precond(GPoint* scnd) const {
    return(_has_second_precond ?
	   CallPreconditions(lineNsecondPrecond, scnd) : GTRUE);
}

GBool GoLine::_call_reset_precond(GPoint* frst, GPoint* scnd) const {
    return(_has_reset_precond ?
	   CallPreconditions(lineNresetPrecond, frst, scnd) : GTRUE);
}

void GoLine::_call_class_first_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoLineMetaclass*) mc)->_has_first_postcond)
	    CallPostconditions(lineNfirstPostcond, this);
	// stop when reaching GoLine Metaclass instance;
	if (mc == GoLine::metaclass())
	    return;
    }
}

void GoLine::_call_class_second_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoLineMetaclass*) mc)->_has_second_postcond)
	    CallPostconditions(lineNsecondPostcond, this);
	// stop when reaching GoLine Metaclass instance;
	if (mc == GoLine::metaclass())
	    return;
    }
}

void GoLine::_call_class_reset_postcond() {
    // call conditions on all super-Metaclasses;
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass()) {
	if (((GoLineMetaclass*) mc)->_has_reset_postcond)
	    CallPostconditions(lineNresetPostcond, this);
	// stop when reaching GoLine Metaclass instance;
	if (mc == GoLine::metaclass())
	    return;
    }
}

void GoLine::_call_first_postcond() {
    if (_has_first_postcond)
	CallPostconditions(lineNfirstPostcond);
}

void GoLine::_call_second_postcond() {
    if (_has_second_postcond)
	CallPostconditions(lineNsecondPostcond);
}

void GoLine::_call_reset_postcond() {
    if (_has_reset_postcond)
	CallPostconditions(lineNresetPostcond);
}

void GoLineMetaclass::firstPrecond(LINE_PRECOND_PTR(f),
				   void* closure, GAddOrRemove add) {
    _has_first_precond =
	AddOrRemoveCondition(lineNfirstPrecond, f, closure, add);
}

void GoLineMetaclass::secondPrecond(LINE_PRECOND_PTR(f),
				    void* closure, GAddOrRemove add) {
    _has_second_precond =
	AddOrRemoveCondition(lineNsecondPrecond, f, closure, add);
}

void GoLineMetaclass::firstPostcond(LINE_POSTCOND_PTR(f),
				    void* closure, GAddOrRemove add) {
    _has_first_postcond =
	AddOrRemoveCondition(lineNfirstPostcond, f, closure, add);
}

void GoLineMetaclass::secondPostcond(LINE_POSTCOND_PTR(f),
				     void* closure, GAddOrRemove add) {
    _has_second_postcond =
	AddOrRemoveCondition(lineNsecondPostcond, f, closure, add);
}

void GoLineMetaclass::resetPrecond(LINER_PRECOND_PTR(f),
				   void* closure, GAddOrRemove add) {
    _has_reset_precond =
	AddOrRemoveCondition(lineNresetPrecond, f, closure, add);
}

void GoLineMetaclass::resetPostcond(LINE_POSTCOND_PTR(f),
				    void* closure, GAddOrRemove add) {
    _has_reset_postcond =
	AddOrRemoveCondition(lineNresetPostcond, f, closure, add);
}

void GoLine::firstPrecond(LINE_PRECOND_PTR(f),
			  void* closure, GAddOrRemove add) {
    _has_first_precond =
	AddOrRemoveCondition(lineNfirstPrecond, f, closure, add);
}

void GoLine::secondPrecond(LINE_PRECOND_PTR(f),
			   void* closure, GAddOrRemove add) {
    _has_second_precond =
	AddOrRemoveCondition(lineNsecondPrecond, f, closure, add);
}

void GoLine::firstPostcond(LINE_POSTCOND_PTR(f),
			   void* closure, GAddOrRemove add) {
    _has_first_postcond =
	AddOrRemoveCondition(lineNfirstPostcond, f, closure, add);
}

void GoLine::secondPostcond(LINE_POSTCOND_PTR(f),
			    void* closure, GAddOrRemove add) {
    _has_second_postcond =
	AddOrRemoveCondition(lineNsecondPostcond, f, closure, add);
}

void GoLine::resetPrecond(LINER_PRECOND_PTR(f),
			  void* closure, GAddOrRemove add) {
    _has_reset_precond =
	AddOrRemoveCondition(lineNresetPrecond, f, closure, add);
}

void GoLine::resetPostcond(LINE_POSTCOND_PTR(f),
			   void* closure, GAddOrRemove add) {
    _has_reset_postcond =
	AddOrRemoveCondition(lineNresetPostcond, f, closure, add);
}



/*
 *------------
 * CLASS PART
 *------------
 */

static GoLineMetaclass* line_metaclass = NULL;

/**
FUNCTION
    GoLineMetaclass* GoLine::metaclass()
DESCRIPTION
    
**/
GoLineMetaclass* GoLine::metaclass() {
    if (line_metaclass == NULL) {
	line_metaclass = new GoLineMetaclass;
	line_metaclass->superclass()->AddSubclass(line_metaclass);

        GoStream::bind(GoLine::metaclass()->classname(),
		       (int (*) (const void*, GoStream*)) GoLine::save,
		       (int (*) (void**, GoStream*)) GoLine::restore);
    }
    return line_metaclass;
}


/**
FUNCTION
    ~GoLineMetaclass()
DESCRIPTION
    
**/
GoLineMetaclass::~GoLineMetaclass() {
    if (this == line_metaclass) {
        superclass()->RemoveSubclass(this);
	line_metaclass = NULL;
    }
}


/**
FUNCTION
    GoLineMetaclass::GoLineMetaclass()
DESCRIPTION
    
**/
 GoLineMetaclass::GoLineMetaclass() {
    _has_first_precond  = _has_first_postcond  =
    _has_second_precond = _has_second_postcond =
    _has_reset_precond  = _has_reset_postcond  = GFALSE;
}


/**
FUNCTION
    const char* GoLineMetaclass::m_classname() const
DESCRIPTION
    virtual private
**/
const char* GoLineMetaclass::m_classname() const {
    return "GoLine";
}


/**
FUNCTION
    GoMetaclass* GoLineMetaclass::m_superclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoLineMetaclass::m_superclass() const {
    return GoStdDS::metaclass();
}


/**
FUNCTION
    const GActionRec* GoLineMetaclass::m_actions () const
DESCRIPTION
    
**/
const GActionRec* GoLineMetaclass::m_actions () const {
    if ((GoMetaclass*) this != GoLine::metaclass()) {
#ifdef DEBUG
	GWarning("the %s class has no \"m_actions()\" method !", classname());
#endif /* DEBUG */
	return empty_action_array;
    }
    return go_line_action_array;
}


/**
FUNCTION
    Go* GoLineMetaclass::m_create();
DESCRIPTION
    virtual private
**/
Go* GoLineMetaclass::m_create() const {
    return (Go*) new GoLine(GPoint(0, 0), GPoint(0, 0));
}


/*
 *----------------
 * PRIVATE MEMBERS
 *----------------
 */

/**
FUNCTION
    GoMetaclass* GoLine::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoLine::m_metaclass() const {
    return GoLine::metaclass();
}


/**
FUNCTION
    void GoLine::m_check() const
DESCRIPTION
    virtual private
**/
void GoLine::m_check() const {}


/**
FUNCTION
    void GoLine::m_dup(const Go* src, Go* dst, Go*, Go*, GBool, GBool)
DESCRIPTION
    virtual private
**/
void GoLine::m_dup
    (const Go* src,Go* parcopy,Go* prevcopy, GBool donext, GBool dochildren)
{
    GoStdDS::m_dup(src, parcopy, prevcopy, donext, dochildren);

    const GoLine *line_src = (const GoLine *) src;

    _first  = line_src->first();
    _second = line_src->second();
    pen().modify(line_src->pen().shPen());
}

/*
 *------------------
 * PROTECTED MEMBERS
 *------------------
 */


/**
FUNCTION
    GoLine::_pen_changed(GoLine* line, const GPen* pen, unsigned long mask)
DESCRIPTION
    *** static member, private ***
    This function is a modify_post_condition of line->_pen
    WARNING: it is assumed that pen **IS** line->_pen
**/

void GoLine::_pen_changed(GoLine* line, const GPen*, unsigned long mask) {
    if (mask == GCOLOR || mask == GALT_COLOR || mask == (GCOLOR & GALT_COLOR)) {
	line->init_sh_pen();
	line->RedrawForeOnly();
    }
    else if (mask & (GLINE_WIDTH | GJOIN_STYLE | GMITER_LIMIT)) {
	line->RedrawBackAndFore();
	line->init_sh_pen();
	line->MyClipHasChanged();
	line->RedrawForeOnly();
    }
    else {
	line->RedrawBackAndFore();
	line->init_sh_pen();
    }
}


/**
FUNCTION
    int GoLine::save(const GoLine* line, GoStream* gostream);
DESCRIPTION
    
**/
int GoLine::save(const GoLine* line, GoStream* gostream) {
    const GShPen* shpen = line->pen().shPen();
    return (gostream->writePoint(line->first(), "first") &&
	    gostream->writePoint(line->second(), "second") &&
	    gostream->saveObj(shpen, shpen->getClassname(), "pen") &&
	    GoStdDS::save(line, gostream));
}


/**
FUNCTION
    int GoLine::restore(GoLine** line, GoStream* gostream);
DESCRIPTION
    
**/
int GoLine::restore(GoLine** line, GoStream* gostream) {
    GPoint frst, scnd;
    const char* class_name = NULL;

    if (! (gostream->readPoint(&frst) && gostream->readPoint(&scnd)))
	return 0;

    if (*line = GoLine::create(frst, scnd)) {
	GShPen* shpen = NULL;
	if (! gostream->restoreObj((void**) &shpen, &class_name))
	    return 0;
	if (strcmp(class_name, GShPen::classname()))
	    GWarning("GoGeometric::restore: class %s where GShPen expected");
	(*line)->pen().modify(shpen);

	return GoStdDS::readParams(*line, gostream);
    }
    return 0;
}


/**
FUNCTION
    int GoLine::readParams(GoLine* line, GoStream* gostream);
DESCRIPTION
    
**/
int GoLine::readParams(GoLine* line, GoStream* gostream) {
    GPoint frst, scnd;
    const char* class_name = NULL;

    if (! (gostream->readPoint(&frst) && gostream->readPoint(&scnd)))
	return 0;
    line->reset(frst, scnd);

    GShPen* shpen = NULL;
    if (! gostream->restoreObj((void**) &shpen, &class_name))
	return 0;
    if (strcmp(class_name, GShPen::classname()))
	GWarning("GoGeometric::restore: class %s where GShPen expected");
    line->pen().modify(shpen);

    return GoStdDS::readParams(line, gostream);
}

/**
FUNCTION
    GBool GoLine::m_contains(const GPoint& absPoint, GNumber maxDist=0) const
DESCRIPTION
    virtual protected
**/
GBool GoLine::m_contains(const GPoint& absPoint, GNumber maxDist) const {
    return interf()->lineContains(sh_pen(), Abscs(), first(), second(),
				  absPoint, maxDist);
}


/**
FUNCTION
    GoLine::m_add_clip(GRegion& region) const
DESCRIPTION
    virtual protected
**/
void GoLine::m_add_clip(GRegion& region) const {
    interf()->lineClips(sh_pen(), Abscs(), first(), second(), &region, NULL);
}


/**
FUNCTION
    GoLine::m_add_opaque_clip(GRegion& region) const
DESCRIPTION
    virtual protected
**/
void GoLine::m_add_opaque_clip(GRegion& region) const {
    interf()->lineClips(sh_pen(), Abscs(), first(), second(),
			NULL, &region);
}


/**
FUNCTION
    void GoLine::m_draw(const GRegion&) const
DESCRIPTION
    virtual protected
**/
void GoLine::m_draw(const GRegion&) const {
    interf()->line(sh_pen(), Abscs(), first(), second());
}


/**
FUNCTION
    void GoLine::m_wired_draw(const GCoordSys& cs, const GPen& pen, Quality quality) const
DESCRIPTION
    virtual protected
**/
void GoLine::m_wired_draw
    (const GCoordSys& cs, const GPen& pen, GQuality quality) const
{
    interf()->line(&pen, cs, first(), second(), quality);
}


/**
FUNCTION
    GBool GoLine::Reset(consr GPoint& frst, const GPoint& scnd)
DESCRIPTION
    
**/
GBool GoLine::Reset(const GPoint& frst, const GPoint& scnd) {

    if (frst == _first && scnd == _second)
	return GFALSE;
    
    GBool was_mapped = mapped();
    
    if (was_mapped)
	RedrawBackAndFore();

    _first  = frst;
    _second = scnd;
    
    MyClipHasChanged();

    if (was_mapped)
	RedrawForeOnly();

    return GTRUE;
}


/*
 *------------------
 * PUBLIC FUNCTIONS
 *------------------
 */


/**
FUNCTION
    GoLine::GoLine(const GPoint& frst, const GPoint& scnd)
DESCRIPTION
    
**/
GoLine::GoLine(const GPoint& frst, const GPoint& scnd) {
    _has_first_precond  = _has_first_postcond  =
    _has_second_precond = _has_second_postcond =
    _has_reset_precond  = _has_reset_postcond  = GFALSE;

    _first  = frst;
    _second = scnd;

    _pen.modifyPostcond((void (*)(void*, GPen*, ulong)) _pen_changed,
			this, GADD);
    _old_sh_pen = NULL;
    init_sh_pen();
}

GoLine::~GoLine() {
    if (_old_sh_pen)
	    _old_sh_pen->destroy();
}

/**
FUNCTION
    GoLine* GoLine::create(const GPoint& frst, const GPoint& scnd);
DESCRIPTION
    Another legitimate GoLine constructor (calls pre/postcondition).
    **static member**
**/
GoLine* GoLine::create(const GPoint& frst, const GPoint& scnd) {
    GoLine* line = (GoLine*) Go::Create(GoLine::metaclass());
    if (line) {
	line->_first  = frst;
	line->_second = scnd;
    }
    return line;
}


/**
FUNCTION
    GBool GoLine::reset(const GPoint& frst_val, const GPoint& scnd_val);
DESCRIPTION
    If preconditions allow it, sets first and second to frst_val and scnd_val
    (possibly modified by preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoLine::reset(const GPoint& frst_val, const GPoint& scnd_val) {
    GPoint frst = frst_val, scnd = scnd_val;
    if ((frst != first() || scnd != second())	&&
	_call_reset_precond(&frst, &scnd)	&&
	_call_class_reset_precond(&frst, &scnd)	&&
	Reset(frst, scnd)) {

	_call_reset_postcond();
	_call_class_reset_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoLine::first(const GPoint& val)
DESCRIPTION
    If preconditions allow it, sets first to val (possibly modified by preconds)
    and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoLine::first(const GPoint& val) {
    GPoint value = val;
    if (value != first()		  &&
	_call_first_precond(&value)	  &&
	_call_class_first_precond(&value) &&
	Reset(value, second())) {

	_call_first_postcond();
	_call_class_first_postcond();
	return GTRUE;
    }
    return GFALSE;
}


/**
FUNCTION
    GBool GoLine::second(const GPoint& val)
DESCRIPTION
    If preconditions allow it, sets second to val (possibly modified by
    preconds) and call postconditions.
    Returns GFALSE if the action does not succeed (a precondition returns GFALSE
    or bad parameters);
**/
GBool GoLine::second(const GPoint& val) {
    GPoint value = val;
    if (value != second()		   &&
	_call_second_precond(&value)	   &&
	_call_class_second_precond(&value) &&
	Reset(first(), value)) {

	_call_second_postcond();
	_call_class_second_postcond();
	return GTRUE;
    }
    return GFALSE;
}
