#ifdef REV_INFO
static char rcsid[] = "$Id: Label.cc,v 1.7 1992/12/10 16:34:15 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  	GoLabel
FILE:    	GoLabel.cc
AUTHOR:		Andre Michel
DATE:    	March 1991

DESCRIPTION:
  GoLabel management (See GoTextual.h and GoLabel.h)

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/GoInterf.h>
#include <Go/GoStream.h>
#include <Go/GoRoot.h>
#include <Go/Label.h>


static GoLabelMetaclass* label_metaclass = NULL;


/**
FUNCTION
    _modifyValuePrecond (void*, const GoTextual*, GModifyValueStruct*)
DESCRIPTION
    This function is used to avoid the introduction
    of a \n character.
**/

static GBool
_modifyValuePrecond (void*, const GoTextual*, GModifyValueStruct* str)
{
GBool ret = GTRUE;

    switch (str->type){
	case GINSERT_CHAR_NOTIFY:
	case GREPLACE_CHAR_NOTIFY:
	    if (str->charValue.Char == '\n') ret = GFALSE;
	    break;

	case GINSERT_STRING_NOTIFY:
	case GSET_STRING_NOTIFY:
	case GREPLACE_STRING_NOTIFY: {
	    const char* ptr = str->stringValue.str;
	    GPos len = str->stringValue.len;
	    for (GPos nb = 0; nb < len; nb++, ptr++)
		if (*ptr == '\n' || *ptr == '\r') ret = GFALSE;
	    }	
	    break;

	default:
	    break;
    }

    return ret;
}


/**
FUNCTION
    _noInterlineJustificationPrecond (void*, const GoTextual*, GModifyAttributeStruct*)
DESCRIPTION
**/

static GBool
_noInterlineJustificationPrecond (void*, const GoTextual*, GModifyAttributeStruct* str)
{
GBool ret = GTRUE;

    if (str->type == GINTERLINE_CHANGE_NOTIFY ||
        str->type == GJUSTIFICATION_CHANGE_NOTIFY)
        ret = GFALSE;

    return ret;
}


/**
FUNCTION
    _modifyUnderlinePen (GoLabel*, GPen*, ulong);
DESCRIPTION
    This function is called when _underlinePen is modified.
**/

static
void _modifyUnderlinePen (void* closure, GPen*, ulong)
{
    GoLabel* lab = (GoLabel*)closure;

    /* only if goroot() != NULL */
    GoRoot* root = lab->root();
    if (root != NULL && lab->underlineMapped()){
	if (lab->backgroundMapped())
	    root->redrawForeOnly (lab, lab->myClip());
	else
	    root->redrawBackAndFore (lab->myClip());
    }
}


/**
FUNCTION
    _modifyStringTextBrush (GoLabel*, GTextBrush*, ulong);
DESCRIPTION
    This function is called when _stringTextBrush is modified.
**/

static
void _modifyStringTextBrush (void* closure, GTextBrush* tbr, ulong mask)
{
    GoLabel* lab = (GoLabel*)closure;

    /* only if goroot() != NULL */
    GoRoot* root = lab->root();
    if (root != NULL){
	if (mask & GFONT || tbr->isNamed())
	    lab->resize();
	else{
	    if (lab->backgroundMapped())
	        root->redrawForeOnly (lab, lab->myClip());
	    else
	        root->redrawBackAndFore (lab->myClip());
	}
    }
}


/**
FUNCTION
    metaclass();
DESCRIPTION
    The GoLabel metaclass function.
**/

GoLabelMetaclass* GoLabel::metaclass ()
{
    if (label_metaclass == NULL)
    {
	label_metaclass = new GoLabelMetaclass;
	label_metaclass->superclass()->AddSubclass(label_metaclass);
	GoStream::bind(label_metaclass->classname(),
	    (int (*) (const void*, GoStream*)) GoLabel::save,
	    (int (*) (void**, GoStream*)) GoLabel::restore);

	/* no '\n' */
	label_metaclass->modifyValuePrecond (
	    _modifyValuePrecond, NULL, GADD);

	/* no interline and justification */
	label_metaclass->modifyAttributePrecond (
	    _noInterlineJustificationPrecond, NULL, GADD);
    }

    return label_metaclass;
}


/**
FUNCTION
    m_classname();
DESCRIPTION
    The GoLabel m_classname function.
**/

const char*
GoLabelMetaclass::m_classname() const
{
    return "GoLabel";
}


/**
FUNCTION
    ~GoLabelMetaclass();
DESCRIPTION
    The GoLabel destructor.
**/

GoLabelMetaclass::~GoLabelMetaclass ()
{
    if (this == label_metaclass)
    {
	superclass()->RemoveSubclass(this);
	label_metaclass = NULL;
    }
}


/**
FUNCTION
    GoMetaclass* GoLabel::m_metaclass() const
DESCRIPTION
    virtual private
**/
GoMetaclass* GoLabel::m_metaclass() const {
    return GoLabel::metaclass();
}


/**
FUNCTION
    GoMetaclass* GoLabelMetaclass::m_superclass() const
DESCRIPTION
    The virtual private m_superclass method.
**/
GoMetaclass* GoLabelMetaclass::m_superclass() const {
    return GoTextual::metaclass();
}


/**
FUNCTION
    m_create();
DESCRIPTION
    The GoLabel m_create function.
**/

Go*
GoLabelMetaclass::m_create() const
{
    return new GoLabel;
}


/**
FUNCTION
    GoLabel();
DESCRIPTION
    The GoLabel constructor.
**/

GoLabel::GoLabel()
{
    _drawUnderline = GFALSE;

    /* postconds on graphical attributes */
    _stringTextBrush.modifyPostcond (_modifyStringTextBrush, this, GADD);
    _underlinePen.modifyPostcond (_modifyUnderlinePen, this, GADD);
}


/**
FUNCTION
    GoLabel::m_dup (const Go*, Go*, Go*, Go*, GBool)
DESCRIPTION
    The GoLabel m_dup function.
**/

void
GoLabel::m_dup
    (const Go* orig, Go* parcopy, Go* prevcopy, GBool donext, GBool dochildren)
{
    GoTextual::m_dup (orig, parcopy, prevcopy, donext, dochildren);
    const GoLabel* lab_orig = (const GoLabel*) orig;

    /* boolean of underlining */
    _drawUnderline = lab_orig->_drawUnderline;
}


/**
FUNCTION
    m_draw (const GRegion&)
DESCRIPTION
    The GoLabel m_draw function.
**/

void
GoLabel::m_draw (const GRegion&) const
{
GNumber w1, w2;

    GCoordSys cs = Abscs();
    GTextBrush* tb = &stringTextBrush();
    GBrush* sb = &selectionBrush();
    GPen* lp = &underlinePen();
    GPen* cp = &cursorPen();
    GBrush* bb = &backgroundBrush();
    GoInterf* interf = this->interf();

    if (backgroundMapped())
	interf->rectangle (NULL, bb, cs,
	    GPoint (-_leftMargin, -_topMargin),
		_width + _leftMargin + _rightMargin,
		    _lineHeightMax + _topMargin + _bottomMargin);

    /* selection list */
    for (GTextSeg* ts = _selectionList.first();
	ts; ts=ts->next())
    {
	w1 = interf->stringWidth (_string, ts->first(),
	    _lineHeight, tb->font());
	w2 = interf->stringWidth (_string, ts->last(),
	    _lineHeight, tb->font());
	interf->rectangle (NULL, sb, cs,
	    GPoint (w1, 0), w2-w1, _height);
    }

    if (cursorMapped()){
	w1 = interf->stringWidth (_string, _currentPos,
	    _lineHeight, tb->font());
	interf->line (cp, cs, GPoint (w1, 0),
	    GPoint (w1, _height - interf->pixelHeight() / cs.J.size()));
    }

    if (stringMapped())
	interf->string (tb, NULL, cs, GPoint(0, 0), _lineHeight,
	    length(), _string);

    if (underlineMapped() && _drawUnderline){
	GNumber h = (GNumber)_underlinePosition * _height;
	interf->line (lp, cs, GPoint (0, h), GPoint (_width, h));
    }
}


/**
FUNCTION
    _lineH (GNumber newLineHeight, GPos, GPos)
DESCRIPTION
    This function is used to set the height of the GoLabel.
**/

void
GoLabel::_lineH (GNumber newLineHeight, GPos, GPos)
{
    GBool was_mapped = mapped();
    if (was_mapped) TempUnmap();

    /* update of height and width */
    _lineHeight = newLineHeight;
    _width = _Width();
    _lineHeightMax = interf()->fontStrikeoutAscent (newLineHeight,
	_stringTextBrush.font()) + interf()->fontStrikeoutDescent
	    (newLineHeight, _stringTextBrush.font());
    _height = _lineHeightMax;

    MyClipHasChanged();
    if (was_mapped) TempMap();
}


/**
FUNCTION
    _underlinePos (GNumber, GPos, GPos);
DESCRIPTION
    Sets the underlinePosition.
**/

void
GoLabel::_underlinePos (GNumber newUnderlinePosition, GPos, GPos)
{
    if (underlineMapped()){
	/* undraw the old line */
         mapUnderline (GFALSE);

	_underlinePosition = newUnderlinePosition;

	/* draw the new line */
         mapUnderline (GTRUE);
    }
    else
	_underlinePosition = newUnderlinePosition;
}


/**
FUNCTION
    _underline (GBool, GPos, GPos);
DESCRIPTION
    Sets the underline state.
**/

void
GoLabel::_underline (GBool set, GPos, GPos)
{
    if (set != _drawUnderline && underlineMapped()){
	if (!set)
	    mapUnderline (GFALSE);

	mapUnderline (GTRUE);
    }

    _drawUnderline = set;
}


/**
FUNCTION
    int GoLabel::save (const GoLabel*, GoStream*);
DESCRIPTION
    The GoLabel save method.
**/

int
GoLabel::save (const GoLabel* label, GoStream* gostream)
{
    if (!(gostream->writeBool (label->_drawUnderline,
	"drawUnderline"))) return (0);
    return (GoTextual::save (label, gostream));
}


/**
FUNCTION
    int GoLabel::restore (GoLabel** label, GoStream* gostream);
DESCRIPTION
    The GoLabel restore method.
**/

int
GoLabel::restore (GoLabel** label, GoStream* gostream)
{
    if (*label = GoLabel::create()){
	if (!(gostream->readBool (&((*label)->_drawUnderline))))
	    return 0;
        return GoTextual::readParams (*label, gostream);
    }

    return 0;
}


/**
FUNCTION
    int GoLabel::readParams (GoLabel*, GoStream*);
DESCRIPTION
    The GoLabel readParams method.
**/

int
GoLabel::readParams (GoLabel* label, GoStream* gostream)
{
    if (!(gostream->readBool (&(label->_drawUnderline))))
	return 0;
    return GoTextual::readParams (label, gostream);
}
