#ifdef REV_INFO
static char rcsid[] = "$Id: TextList.cc,v 1.1 1992/08/06 09:35:55 uep Exp $";
#endif
/*****************************************************************************
MODULE:  Textual
FILE:    TextList.cc
AUTHOR:  Michel Andre
DATE:    1/7/92

DESCRIPTION: See TextList.h
    

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France

  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.

  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <assert.h>
#include <Go/machdef.h>
#include <Go/TextList.h>


/**
FUNCTION
    GTextSeg();
DESCRIPTION
    The GTextSeg constructor.
**/

GTextSeg::GTextSeg()
{
    _first = -1;
    _last = -1;
    _next = NULL;
}


/**
FUNCTION
    shift (GPos pos, GPos step);
DESCRIPTION
    This function is used to update the values of a GTextSeg
    after a text deletion or insertion of step charaters.
**/

void
GTextSeg::shift (GPos pos, GPos step)
{
    /* suppression */
    if (step < 0){
	if (_last > pos)
	    _last += step;
	if (_first > pos)
	    _first += step;

	/* GTextSeg to be deleted later in list::shift */
	if (_first < 0) _first = 0;
	if (_last < 0) _last = 0;
    }

    /* insertion */
    else{
	if (_last >= pos)
	    _last += step;
	if (_first >= pos)
	    _first += step;
    }
}


/**
FUNCTION
    deletePortion (GPos pos1, GPos pos2);
DESCRIPTION
**/

void
GTextSeg::deletePortion (GPos pos1, GPos pos2)
{
    if (pos1 > _first || pos2 < _last)
	if (_first >= pos1 && _first < pos2)
	    _first = pos2;
	else
	    if (_last > pos1 && _last <= pos2)
		_last = pos1;
}


/**
FUNCTION
    contains (GPos pos) const;
DESCRIPTION
    Returns GTRUE if pos is inside the segment.
**/

GBool
GTextSeg::contains (GPos pos) const
{
    return pos >= _first && pos < _last;
}


/**
FUNCTION
    first (GPos pos);
DESCRIPTION
**/

void
GTextSeg::first (GPos pos)
{
    assert (pos >= -1);
    _first = pos;
}


/**
FUNCTION
    last (GPos pos);
DESCRIPTION
**/

void
GTextSeg::last (GPos pos)
{
    assert (pos >= -1);
    _last = pos;
}


/**
FUNCTION
    GTextList();
DESCRIPTION
    The GTextList constructor.
**/

GTextList::GTextList()
{
    _first = NULL;
    _size = 0;
}


/**
FUNCTION
    ~GTextList();
DESCRIPTION
    The GTextList destructor.
**/

GTextList::~GTextList()
{
    GTextSeg* s1 = _first;
    while (s1){
	GTextSeg* next = s1->next();
	delete s1;
	s1 = next;
    }
}


/**
FUNCTION
    insertSeg (GTextSeg*);
DESCRIPTION
    This function is used to insert a segment in the GTextList.
**/

void
GTextList::insertSeg (GTextSeg* Seg)
{
    GPos segFirst = Seg->first();
    GPos segLast = Seg->last();

    /* destruction des segments intermediaires */
    this->deletePortion (segFirst, segLast);

    /* recherche du segment immediatement superieur */
    GTextSeg* s1 = _first;
    GTextSeg* prec = s1;  // to avoid Warning
    char nb = 0;
    while (s1 && s1->last() <= segFirst){
	prec = s1;
	s1 = s1->next();
	nb++;
    }

    /* the segment to insert is not final */
    if (s1)
	Seg->_next = s1;

    if (!nb)
	_first = Seg;
    else
	prec->_next = Seg;
    _size++;
}


/**
FUNCTION
    shift (GPos pos, GPos step);
DESCRIPTION
    Shift beginning at position pos with a step.
**/

void
GTextList::shift (GPos pos, GPos step)
{
    GTextSeg* s1 = _first;
    GTextSeg *prec = NULL;

    while (s1){
	s1->shift (pos, step);
	GTextSeg* next = s1->next();

	/* null segment to destroy */
	if (s1->first() == s1->last()){
	    if (prec)
		prec->_next = next;
	    else
		_first = next;
	    delete s1;
	    _size--;
	}
	prec = s1;
	s1 = next;
    }

    /* all segments have been destroyed */
    if (_size == 0)
	_first = NULL;
}


/**
FUNCTION
    deletePortion (GPos pos1, GPos pos2);
DESCRIPTION
**/

void
GTextList::deletePortion (GPos pos1, GPos pos2)
{
GPos first, last;

    GTextSeg* s1 = _first;
    GTextSeg* prec = NULL;
    while (s1 && ((first = s1->first()) <= pos2)){
	GTextSeg* next = s1->next();
	last = s1->last();

	/* segment included in the portion */
	/* must be destroyed */
	if (first >= pos1 && last <= pos2){
	    if (prec)
		prec->_next = next;
	    else
		_first = next;
	    delete s1;
	    _size--;
	}

	else{
	    /* segment included in the portion */
	    /* must be cutted in 2 parts */
	    if (first <= pos1 && last >= pos2){
		GTextSeg *new_s = new GTextSeg;
		new_s->first (pos2);
		new_s->last (last);
		new_s->_next = next;
		s1->last (pos1);
		s1->_next = new_s;
		_size++;
	    }

	    else{
		/* segment partially included or outside */
		s1->deletePortion (pos1, pos2);
		prec = s1;
	    }
	}
	s1 = next;
    }
}


/**
FUNCTION
    deleteSeg (GTextSeg* Seg);
DESCRIPTION
    Description dans la liste du segment pointe par Seg.
**/

void
GTextList::deleteSeg (GTextSeg* Seg)
{
    GTextSeg* s1 = _first;
    GTextSeg *prec = NULL;
    char nb = 0;
    while (s1 && s1 != Seg){
	prec = s1;
	s1 = s1->next();
	nb++;
    }
    if (s1 == Seg){
	if (nb)
	    prec->_next = s1->next();
	else
	    _first = s1->next();
	delete s1;
	_size--;
    }
}


/**
FUNCTION
    reinit()
DESCRIPTION
    Reinitialization of a text list.
**/

void
GTextList::reinit()
{
    GTextSeg* s1 = _first;
    while (s1){
	GTextSeg* next = s1->next();
	delete s1;
	s1 = next;
    }

    /* reinit _first */
    _first = NULL;
    _size = 0;
}


/**
FUNCTION
    contains (GPos pos) const;
DESCRIPTION
    Returns GTRUE if pos is inside the list.
**/

GBool
GTextList::contains (GPos pos) const
{
    GBool ret = GFALSE;
    GTextSeg* seg = _first;
    while (seg && !ret){
	ret = seg->contains (pos);
	seg = seg->next();
    }

    return ret;
}


/**
FUNCTION
    posToSeg (GPos pos) const;
DESCRIPTION
    Returns the segment containing pos.
**/

GTextSeg*
GTextList::posToSeg (GPos pos) const
{
    GTextSeg* seg = _first;
    while (seg){
	if (seg->contains (pos)) return seg;
	seg = seg->next();
    }
    return NULL;
}
