#ifdef REV_INFO
static char rcsid[] = "$Id: Textual.cc,v 1.14 1992/12/21 16:29:45 andre Exp $";
#endif
/*****************************************************************************
MODULE:      GoTextual
FILE:        Textual.cc
AUTHOR:      Andre Michel
DATE:        March 1991

DESCRIPTION:
  Text management for all the predefined text Gos. (See Textual.h)

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <stdio.h>
#include <Go/Tools/Alloc.h>
#include <Go/GoInterf.h>
#include <Go/GoStream.h>
#include <Go/GoRoot.h>
#include <Go/Textual.h>
#include <Go/mTextual.h>

static GoTextualMetaclass* textual_metaclass = NULL;
extern GRegion oldClip;

long streamMask;
#define	VERSION_1_2	(1 << 15)
#define	VERSION_1_2_BIS	(1 << 14)


/**
FUNCTION
    _modifyValuePrecond();
DESCRIPTION
    Modify value precond used to avoid bad modifications.
**/

static
GBool _modifyValuePrecond (void*, const GoTextual* text, GModifyValueStruct* str)
{
GBool ret = GTRUE;

    switch (str->type){
	case GREPLACE_CHAR_NOTIFY:
	    /* replaced char must not be '\n' */
	    if (str->charValue.Char == '\n' ||
		(*text)[str->charValue.pos] == '\n')
		    ret = GFALSE;
	    break;

	case GREPLACE_STRING_NOTIFY: {
	    /* the old sub-string must not contain '\n' */
	    const char* adpos = text->getString() +
		str->stringValue.pos;
	    int i = 0;
	    while (ret && i++ < str->stringValue.oldStringSize) 
		if (*adpos == '\n') ret = GFALSE;

	    /* the new sub-string must not contain '\n' */
	    adpos = str->stringValue.str;
	    i = 0;
	    while (ret && i++ < str->stringValue.len) 
		if (*adpos == '\n') ret = GFALSE;
	    }

	    break;

	default:
	    break;
    }

    return ret;
}


/**
FUNCTION
    _stringTextBrushPrecond (void*, const GTextBrush*, ulong&, GTextBrushValues*);
DESCRIPTION
    The string text brush precond.
**/

static GBool
_stringTextBrushPrecond (void* closure, const GTextBrush*,
    ulong&, GTextBrushValues*)
{
    GoTextual* text = (GoTextual*)closure;
    if (text->root() != NULL)
	oldClip = text->myClip();
    return GTRUE;
}


/**
FUNCTION
    _stringTextBrushPostcond (void*, GTextBrush*, ulong);
DESCRIPTION
    The string text brush postcond.
**/

static void
_stringTextBrushPostcond (void* closure, GTextBrush*, ulong)
{
    GoTextual* text = (GoTextual*)closure;

    /* only if goroot() != NULL */
    GoRoot* root = text->root();
    if (root == NULL) return;

    /* redraw the text clip differences */
    GRegion deltaClip;
    GRegion myclip = ((Go*)text)->myClip();

    /* the clip may have increased */
    GSub (myclip, oldClip, deltaClip);
    if (!deltaClip.empty())
	root->redrawForeOnly (text, deltaClip);

    /* the clip may have decreased */
    GSub (oldClip, myclip, deltaClip);
    if (!deltaClip.empty())
	root->redrawBackAndFore (deltaClip);

    if (text->backgroundMapped())
	root->redrawForeOnly (text, myclip);
    else
	root->redrawBackAndFore (myclip);
}


/**
FUNCTION
    _backgroundBrushPostcond (void*, GBrush*, ulong);
DESCRIPTION
    The  background Brush postcond.
**/

static
void _backgroundBrushPostcond (void* closure, GBrush*, ulong)
{
    GoTextual* text = (GoTextual*)closure;
    if (text->root() != NULL && text->backgroundMapped())
	text->mapBackground (GTRUE);
}


/**
FUNCTION
    _cursorPenPrecond (void*, const GPen*, ulong&, GPenValues*)
DESCRIPTION
    Function called before a cursor pen modification occurs.
    The currentPos() function calls the redrawing of it.
**/

static
GBool _cursorPenPrecond (void* closure, const GPen*, ulong&, GPenValues*)
{
    GoTextual* text = (GoTextual*)closure;
    if (text->root() != NULL && text->cursorMapped())
	text->currentPos (text->getCurrentPos());
    return GTRUE;
}


/**
FUNCTION
    _cursorPenPostcond (void*, GPen*, ulong)
DESCRIPTION
    Function called when a cursor pen modification occurs.
    The currentPos() function calls the redrawing of it.
    Useful when the cursor thickness decreases.
**/

static
void _cursorPenPostcond (void* closure, GPen*, ulong)
{
    GoTextual* text = (GoTextual*)closure;
    if (text->root() != NULL && text->cursorMapped())
	text->currentPos (text->getCurrentPos());
}


/**
FUNCTION
    _selectionBrushPostcond (GoTextual*, GBrush*, ulong);
DESCRIPTION
    The selection brush change postcondition.
**/

static
void _selectionBrushPostcond (void* closure, GBrush*, ulong)
{
    GoTextual* text = (GoTextual*)closure;

    /* called when restore */
    GoRoot* root = text->root();
    if (root == NULL) return;

    GRegion reg;
    for (GTextSeg* s = text->selectionList().first();
	s; s=s->next())
	    GAdd (reg, text->boundariesToRegion (
		s->first(), s->last()), reg);
    root->redrawForeOnly (text, reg);
}


/**
FUNCTION
    _interfPostcond (void*, Go*, GoInterf*)
DESCRIPTION
    Function to reframe a Label after a restore.
**/

static
void _interfPostcond (void*, Go* go, GoInterf*)
{
    if (go->root()) ((GoTextual*)go)->resize();
}


/**
FUNCTION
    metaclass();
DESCRIPTION
    Returns the GoTextual GoMetaclass*.
**/

GoTextualMetaclass* GoTextual::metaclass()
{
    if (textual_metaclass == NULL){
	textual_metaclass = new GoTextualMetaclass;
	textual_metaclass->superclass()->AddSubclass (
	    textual_metaclass);

	/* select postcond */
	textual_metaclass->selectPostcond (&GoTextual::_select,
	    NULL, GADD);

	/* modify value precond */
	textual_metaclass->modifyValuePrecond (
            _modifyValuePrecond, NULL, GADD);

        /* postcond on gointerf modification */
        textual_metaclass->interfPostcond (
	    _interfPostcond, NULL, GADD);
    }

    return textual_metaclass;
}


/**
FUNCTION
    ~GoTextualMetaclass()
DESCRIPTION
    The GoTextual metaclass destructor.
**/

GoTextualMetaclass::~GoTextualMetaclass()
{
    if (this == textual_metaclass){
	superclass()->RemoveSubclass(this);
	textual_metaclass = NULL;
    }
}


/**
FUNCTION
    m_classname() const
DESCRIPTION
    Returns the m_classname string.
**/

const char*
GoTextualMetaclass::m_classname() const
{
    return "GoTextual";
}


/**
FUNCTION
    GoMetaclass* GoTextualMetaclass::m_superclass() const
DESCRIPTION
    Returns the GoTextual metaclass superclass.
**/

GoMetaclass* GoTextualMetaclass::m_superclass() const {
    return GoStdDS::metaclass();
}


/**
FUNCTION
    GoMetaclass* GoTextual::m_metaclass() const
DESCRIPTION
**/
GoMetaclass* GoTextual::m_metaclass() const {
    return GoTextual::metaclass();
}


/**
FUNCTION
    GoTextualMetaclass::GoTextualMetaclass()
DESCRIPTION
    The GoTextualMetaclass constructor.
**/

GoTextualMetaclass::GoTextualMetaclass()
{
    /* flags initialisation */
    _flags._has_modify_value_precond = GFALSE;
    _flags._has_modify_value_postcond = GFALSE;
    _flags._has_current_pos_precond = GFALSE;
    _flags._has_current_pos_postcond = GFALSE;
    _flags._has_modify_attribute_precond = GFALSE;
    _flags._has_modify_attribute_postcond = GFALSE;
    _flags._has_selection_precond = GFALSE;
    _flags._has_selection_postcond = GFALSE;
}


/**
FUNCTION
    GoTextual();
DESCRIPTION
    The GoTextual constructor.
**/

GoTextual::GoTextual()
{
    _size = 0;
    _allocString (8);
    *_string = '\0';
    _currentPos = 0;
    _present = GNO_PRESENTATION;
    _mask = VERSION_1_2 | VERSION_1_2_BIS |
	GSTRING | GUNDERLINE;
    _width = 0;
    _justification = GLEFT_JUSTIFICATION;
    _interline = 0;
    _lineHeight = 4;
    _height = 4;
    _underlinePosition = 0.8;
    _leftMargin = 1;
    _rightMargin = 1;
    _topMargin = 1;
    _bottomMargin = 1;

    /* string precond / postcond */
    _stringTextBrush.modifyPrecond(
	_stringTextBrushPrecond, this, GADD);
    _stringTextBrush.modifyPostcond(
	_stringTextBrushPostcond, this, GADD);

    /* cursor pen precond / postcond */
    _cursorPen.modifyPostcond(
	_cursorPenPostcond, this, GADD);
    _cursorPen.modifyPrecond(
	_cursorPenPrecond, this, GADD);

     /* selection brush postcond */
    _selectionBrush.modifyPostcond(
	_selectionBrushPostcond, this, GADD);

    /* background brush postcond */
    _backgroundBrush.modifyPostcond(
        _backgroundBrushPostcond, this, GADD);

    /* flags initialisation */
    _flags._has_modify_value_precond = GFALSE;
    _flags._has_modify_value_postcond = GFALSE;
    _flags._has_current_pos_precond = GFALSE;
    _flags._has_current_pos_postcond = GFALSE;
    _flags._has_modify_attribute_precond = GFALSE;
    _flags._has_modify_attribute_postcond = GFALSE;
    _flags._has_selection_precond = GFALSE;
    _flags._has_selection_postcond = GFALSE;
}


/**
FUNCTION
    ~GoTextual();
DESCRIPTION
    The GoTextual destructor.
**/

GoTextual::~GoTextual()
{
    GFREE (_string);
}


/**
FUNCTION
    m_wired_draw (const GCoordSys&, const GPen& pen, GQuality quality) const;
DESCRIPTION
    The GoTextual wireddrawing function.
    Just draw the outline of the bounding box.
**/

void
GoTextual::m_wired_draw (const GCoordSys& cs, const GPen& pen,
			 GQuality quality) const
{
    GRegion region = Abscs().from (myClip());
    GBox box = region.bigbox();
    interf()->rectangle (&pen, NULL, cs,
			 GPoint (box.xmin, box.ymin),
			 box.width(), box.height(), quality);
}


/**
FUNCTION
    m_add_clip (GRegion&);
DESCRIPTION
    The GoTextual m_add_clip function.
**/

void
GoTextual::m_add_clip (GRegion& reg) const
{
    /* clip of the frame */
    interf()->rectangleClips(NULL, &backgroundBrush(),
	Abscs(), GPoint (-_leftMargin, -_topMargin),
	    _width + _leftMargin + _rightMargin,
		_height + _topMargin + _bottomMargin, &reg, NULL);
}


/**
FUNCTION
    m_add_opaque_clip (GRegion&);
DESCRIPTION
    The GoTextual m_opaque_clip function.
    Returns gEmptyGRegion when there is no background.
**/

void
GoTextual::m_add_opaque_clip (GRegion& reg) const
{
    if (backgroundMapped())
	m_add_clip (reg);
}


/**
FUNCTION
    m_contains (const GPoint &pt, GNumber maxDistance);
DESCRIPTION
    The GoTextual m_contains function.
**/

GBool
GoTextual::m_contains (const GPoint &abs_pt, GNumber) const
{
    return myClip().contains (abs_pt);
}


/**
FUNCTION
    GoTextual::m_dup (const Go*, Go*, Go*, Go*, GBool)
DESCRIPTION
    The GoTextual m_dup function.
**/

void
GoTextual::m_dup
    (const Go* orig, Go* parcopy, Go* prevcopy, GBool donext, GBool dochildren)
{
    GoStdDS::m_dup(orig, parcopy, prevcopy, donext, dochildren);

    const GoTextual* text_orig = (const GoTextual*)orig;

    /* string copy */
    _allocString (text_orig->_size);
    strcpy (_string, text_orig->_string);

    /* internal variables copy */
    _mask = text_orig->_mask;
    _present = text_orig->_present;
    _underlinePosition = text_orig->_underlinePosition;
    _currentPos = text_orig->_currentPos;
    _width = text_orig->_width;
    _height = text_orig->_height;
    _lineHeight = text_orig->_lineHeight;
    _lineHeightMax = text_orig->_lineHeightMax;
    _flags = text_orig->_flags;
    _justification = text_orig->_justification;
    _interline = text_orig->_interline;

    /* selection list */
    for (GTextSeg* ts = text_orig->_selectionList.first();
	ts; ts=ts->next())
    {
	GTextSeg* new_seg = new GTextSeg;
	new_seg->first (ts->first());
	new_seg->last (ts->last());
	_selectionList.insertSeg (new_seg);
    }

    if (_mask & VERSION_1_2){
	_leftMargin = text_orig->_leftMargin;
	_rightMargin = text_orig->_rightMargin;
	_topMargin = text_orig->_topMargin;
	_bottomMargin = text_orig->_bottomMargin;
    }

    /* gc copy */
    _stringTextBrush.modify (text_orig->
	_stringTextBrush.shTextBrush());
    _backgroundBrush.modify (text_orig->
	_backgroundBrush.shBrush());
    _selectionBrush.modify (text_orig->
	_selectionBrush.shBrush());
    _underlinePen.modify (text_orig->
	_underlinePen.shPen());
    _cursorPen.modify (text_orig->_cursorPen.shPen());
}


/**
FUNCTION
    currentPos (GPos)
DESCRIPTION
    This function is used to set a new point of insertion.
**/

GBool
GoTextual::currentPos (GPos newPos)
{
    if (newPos < 0 || newPos > length()) return GFALSE;

    /* preconditions call */
    GBool ret;
    if (ret = _call_current_pos_precond (newPos) &&
	((GoTextualMetaclass*)getMetaclass())->
	    _call_current_pos_precond (this, newPos) &&
		GoTextual::metaclass()->
		    _call_current_pos_precond (this, newPos))
    {
	GPos oldPos = _currentPos;
	if (cursorMapped()){
	    mapCursor (GFALSE);
	    _currentPos = newPos;
	    mapCursor (GTRUE);
	}
	else
	    _currentPos = newPos;

	/* postconditions call */
	_call_current_pos_postcond (oldPos);
	((GoTextualMetaclass*)getMetaclass())->
	    _call_current_pos_postcond (this, oldPos);
	GoTextual::metaclass()->
	    _call_current_pos_postcond (this, oldPos);
    }

    return ret;
}


/**
FUNCTION
    lineHeight (GNumber, GPos, GPos);
DESCRIPTION
    This function sets the lineHeight.
**/

GBool
GoTextual::lineHeight (GNumber newLineHeight, GPos firstPos, GPos lastPos)
{
    GPos len = length();
    if (firstPos < -1 || lastPos > len || lastPos < firstPos)
	return GFALSE;

    GModifyAttributeStruct *str = new GModifyAttributeStruct;
    str->type = GLINEHEIGHT_CHANGE_NOTIFY;
    str->lineHeight = newLineHeight;
    str->firstPosition = firstPos;
    str->lastPosition = lastPos;

    /* lineHeight preconditions call */
    GBool ret;
    if (ret = _call_modify_attribute_precond (str) &&
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_precond (this, str) &&
		GoTextual::metaclass()->
		    _call_modify_attribute_precond (this, str))
    {
	oldClip = myClip();
	str->lineHeight = _lineHeight;

	/* call the private method */
	_lineH (newLineHeight, firstPos, lastPos);

        /* redraw the text clip differences */
        GRegion deltaClip;
        GRegion myclip = myClip();

        /* the clip may have increased */
        GSub (myclip, oldClip, deltaClip);
        if (!deltaClip.empty()){
	    MyClipHasChanged();
	    RedrawForeOnly (deltaClip);
        }

        /* the clip may have decreased */
        GSub (oldClip, myclip, deltaClip);
        if (!deltaClip.empty()){
            MyClipHasChanged();
            RedrawBackAndFore (deltaClip);
        }

	if (backgroundMapped())
	    RedrawForeOnly (myclip);
	else
	    RedrawBackAndFore (myclip);

	/* lineHeight postconditions call */
	_call_modify_attribute_postcond (str);
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_postcond (this, str);
	GoTextual::metaclass()->
	    _call_modify_attribute_postcond (this, str);

    }

    delete str;
    return ret;
}


/**
FUNCTION
    underlinePosition (GNumber, GPos, GPos);
DESCRIPTION
    Places the Y position of the underlining.
**/

GBool
GoTextual::underlinePosition (GNumber newUnderlinePosition, GPos firstPos, GPos lastPos)
{
    GPos len = length();
    if (firstPos < -1 || lastPos > len || lastPos < firstPos ||
	newUnderlinePosition < 0 || newUnderlinePosition > 1)
	    return GFALSE;

    GModifyAttributeStruct *str = new GModifyAttributeStruct;
    str->type = GUNDERLINE_POSITION_CHANGE_NOTIFY;
    str->underlinePosition = newUnderlinePosition;
    str->firstPosition = firstPos;
    str->lastPosition = lastPos;

    /* underlinePosition preconditions call */
    GBool ret;
    if (ret = _call_modify_attribute_precond (str) &&
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_precond (this, str) &&
		GoTextual::metaclass()->
		    _call_modify_attribute_precond (this, str))
   {
	str->underlinePosition = _underlinePosition;

	/* call the private method */
	_underlinePos (newUnderlinePosition, firstPos, lastPos);

	/* underlinePosition postconditions call */
	_call_modify_attribute_postcond (str);
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_postcond (this, str);
	GoTextual::metaclass()->
	    _call_modify_attribute_postcond (this, str);
    }

    delete str;
    return ret;
}


/**
FUNCTION
    underline (GBool, GPos, GPos);
DESCRIPTION
    Un/underlines a text portion.
**/

GBool
GoTextual::underline (GBool set, GPos firstPos, GPos lastPos)
{
    GPos len = length();
    if (firstPos < -1 || lastPos > len || lastPos < firstPos)
	return GFALSE;

    GModifyAttributeStruct *str = new GModifyAttributeStruct;
    str->type = GUNDERLINE_SET_NOTIFY;
    str->underlineSet = set;
    str->firstPosition = firstPos;
    str->lastPosition = lastPos;

    /* underline preconditions call */
    GBool ret;
    if (ret = _call_modify_attribute_precond (str) &&
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_precond (this, str) &&
		GoTextual::metaclass()->
		    _call_modify_attribute_precond (this, str))
    {
	/* call the private method */
	_underline (set, firstPos, lastPos);

	/* underline postconditions call */
	_call_modify_attribute_postcond (str);
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_postcond (this, str);
	GoTextual::metaclass()->
	    _call_modify_attribute_postcond (this, str);
    }

    delete str;
    return ret;
}


/**
FUNCTION
    cursorLeft();
DESCRIPTION
    Places the cursor one position before.
**/

void
GoTextual::cursorLeft()
{
    if (_currentPos > 0)
	currentPos (_currentPos-1);
}


/**
FUNCTION
    cursorRight();
DESCRIPTION
    Places the cursor one position after.
**/

void
GoTextual::cursorRight()
{
    if (_currentPos < length())
	currentPos (_currentPos+1);
}


/**
FUNCTION
    cursorUp();
DESCRIPTION
    Places the cursor one line upper.
**/

void
GoTextual::cursorUp()
{
    int currentLine = lineNumber (_currentPos);
    if (currentLine){
        GPoint currentPosOrig = posToXy (_currentPos);
        GPoint currentLineOrig = posToXy (
	    begLinePos (currentLine));
	currentPos (xToPos (currentLine - 1,
	    currentPosOrig.x - currentLineOrig.x));
    }
}


/**
FUNCTION
    cursorDown();
DESCRIPTION
    Places the cursor one line downer.
**/

void
GoTextual::cursorDown()
{
    int currentLine = lineNumber (_currentPos);
    if (currentLine < nbLines() - 1){
        GPoint currentPosOrig = posToXy (_currentPos);
        GPoint currentLineOrig = posToXy (
	    begLinePos (currentLine));
	currentPos (xToPos (currentLine + 1,
	    currentPosOrig.x - currentLineOrig.x));
    }
}


/**
FUNCTION
    CursorBegText();
DESCRIPTION
    Places the cursor at the beginning of the text.
**/

void
GoTextual::cursorBegText()
{
    currentPos (0);
}


/**
FUNCTION
    cursorEndText();
DESCRIPTION
    Places the cursor at the end of the text.
**/

void
GoTextual::cursorEndText()
{
    currentPos (length());
}


/**
FUNCTION
    cursorBegLine();
DESCRIPTION
    Places the cursor at the beginning of the line.
**/

void
GoTextual::cursorBegLine()
{
    currentPos (begLinePos());
}


/**
FUNCTION
    CursorEndline();
DESCRIPTION
    Places the cursor at the end of the line.
**/

void
GoTextual::cursorEndLine()
{
    currentPos (endLinePos());
}


/**
FUNCTION
    _select (void*, GoTextual*)
DESCRIPTION
    Selection postcondition.
    Used to update the GoTextual width.
**/

#define SIZE_W 10
#define _update_W(cs,w) if ((cs).I.size()*(w) < SIZE_W) (w) = SIZE_W / (cs).I.size()

#ifndef STATIC_CONFUSION
static
#endif
void GoTextual::_select (void*, Go* go)
{
    GoTextual* text = (GoTextual*)go;

    GNumber updatedWidth = text->_width;
    GCoordSys cs = text->Abscs();

    if (text->selected()){
	_update_W (cs, updatedWidth);
	if (updatedWidth != text->_width){
	    text->_width = updatedWidth;
	    text->MyClipHasChanged();
	    text->RedrawForeOnly();
	}
    }

    else{
	updatedWidth = text->_Width();
	if (updatedWidth < text->_width){
	    text->RedrawBackAndFore();
	    text->_width = updatedWidth;
	    text->MyClipHasChanged();
	}
    }
}


/**
FUNCTION
    save (const GoTextual*, GoStream*);
DESCRIPTION
    The GoTextual save function.
**/

int
GoTextual::save (const GoTextual* text, GoStream* gostream)
{
    if (!gostream->writeLong (text->_mask,
	"mask")) return 0;
    if (!gostream->writeInt (text->_present,
	"presentation")) return 0;
    if (!gostream->writeNumber (text->_underlinePosition,
	"underlinePosition")) return 0;
    if (!gostream->writeNumber (text->_lineHeight,
	"lineHeight")) return 0;
    if (!gostream->writeString (text->_string,
	"string")) return 0;
    if (!gostream->writeInt (text->_currentPos,
	"currentPos")) return 0;
    if (!(gostream->writeInt ((int)text->_justification,
	"justification"))) return 0;
    if (!(gostream->writeNumber (text->_interline,
	"interline"))) return 0;

    if (!gostream->saveObj(text->stringTextBrush().shTextBrush(),
	GShTextBrush::classname(), "stringTextBrush")) return 0;
    if (!gostream->saveObj(text->backgroundBrush().shBrush(),
	GShBrush::classname(), "backgroundBrush")) return 0;
    if (!gostream->saveObj(text->selectionBrush().shBrush(),
	GShBrush::classname(), "selectionBrush")) return 0;
    if (!gostream->saveObj(text->underlinePen().shPen(),
	GShPen::classname(), "underlinePen")) return 0;
    if (!gostream->saveObj(text->cursorPen().shPen(),
	GShPen::classname(), "cursorPen")) return 0;

    /* margins */
    if (!gostream->writeNumber (text->_leftMargin,
	"leftMargin")) return 0;
    if (!gostream->writeNumber (text->_rightMargin,
	"rightMargin")) return 0;
    if (!gostream->writeNumber (text->_topMargin,
	"topMargin")) return 0;
    if (!gostream->writeNumber (text->_bottomMargin,
	"bottomMargin")) return 0;

    /* multiple selection segments */
    GPos size = text->_selectionList.size();
    if (!(gostream->writeInt (size,
	"selectionList size"))) return 0;
    GTextSeg *s1 = (text->_selectionList).first();
    GPos i = 0;
    char chaine[16];
    while (s1){
        GPos first = s1->first();
        sprintf (chaine, "Seg_%d_first", i);
        if (!gostream->writeInt (first, chaine)) return 0;
        GPos last = s1->last();
        sprintf (chaine, "Seg_%d_last", i);
        if (!gostream->writeInt (last, chaine)) return 0;
        s1 = s1->next();
        i++;
    }

    return (GoStdDS::save(text, gostream));
}


/**
FUNCTION
    readParams(GoGeometric*, GoStream*);
DESCRIPTION
    The GoTextual readParams function.
**/

int
GoTextual::readParams (GoTextual* text, GoStream* gostream)
{
GShPen* shpen = NULL;
GShTextBrush* shtbrush = NULL;
GShBrush* shbrush = NULL;
const char* class_name = NULL;

    if (!gostream->readLong ((long*)&text->_mask)) return 0;

    /* stream mask used by the GoTextSegments and the GoTextLines */
    streamMask = text->_mask;

    if (!gostream->readInt ((int*)&text->_present)) return 0;
    if (!gostream->readNumber (&text->
	_underlinePosition)) return 0;
    if (!gostream->readNumber (&text->
	_lineHeight)) return 0;

    /* string load */
    if (text->_size != 0)
    GFREE (text->_string);
    if (!gostream->readString (0, &(text->_string))) return 0;
    text->_size = text->length();

    if (!gostream->readInt (&text->_currentPos)) return 0;

    /* equivalent to _firstSelected and _lastSelected */
    if (!(text->_mask & VERSION_1_2)){
	int first, last;
	if (!gostream->readInt (&first)) return 0;
	if (!gostream->readInt (&last)) return 0;
	if (first != -1){
	    GTextSeg* ts = new GTextSeg;
	    ts->first (first);
	    ts->last (last);
	    text->selectionList().insertSeg (ts);
	}
    }

    if (!(gostream->readInt ((int*)&text->
	_justification))) return 0;
    if (!(gostream->readNumber (&text->
	_interline))) return 0;

    /* stringTextBrush */
    if (!gostream->restoreObj ((void**) &shtbrush, &class_name))
	return 0;
    text->stringTextBrush().modify(shtbrush);

    /* backgroundBrush */
    if (!gostream->restoreObj ((void**) &shbrush, &class_name))
	return 0;
    text->backgroundBrush().modify(shbrush);

    /* selectionBrush */
    if (!gostream->restoreObj ((void**) &shbrush, &class_name))
	return 0;
    text->selectionBrush().modify(shbrush);

    /* underlinePen */
    if (!gostream->restoreObj ((void**) &shpen, &class_name))
	return 0;
    text->underlinePen().modify(shpen);

    /* cursorPen */
    if (!gostream->restoreObj ((void**) &shpen, &class_name))
	 return 0;
    text->cursorPen().modify(shpen);

    /* margins */
    if (text->_mask & VERSION_1_2){
	if (!(gostream->readNumber (&text->_leftMargin)))
	    return 0;
	if (!(gostream->readNumber (&text->_rightMargin)))
	    return 0;
	if (!(gostream->readNumber (&text->_topMargin)))
	    return 0;
	if (!(gostream->readNumber (&text->_bottomMargin)))
	    return 0;

	/* selection list */
        int size;
        if (!(gostream->readInt (&size))) return 0;

        while (size--){
            GTextSeg* new_s = new GTextSeg;
	    int aux;
            if (!gostream->readInt (&aux)) return 0;
	    new_s->first (aux);
            if (!gostream->readInt (&aux)) return 0;
	    new_s->last (aux);
            text->_selectionList.insertSeg (new_s);
        }
    }

    text->_mask |= VERSION_1_2;
    text->_mask |= VERSION_1_2_BIS;
    return (GoStdDS::readParams (text, gostream));
}


/**
FUNCTION
    justification (justification);
DESCRIPTION
    Changes the justification.
**/

GBool
GoTextual::justification (GJustification newJustification)
{
    GBool ret = GTRUE;
    GModifyAttributeStruct* str = new GModifyAttributeStruct;
    str->type = GJUSTIFICATION_CHANGE_NOTIFY;
    str->justification = newJustification;

    /* justification preconditions call */
    if (ret = _call_modify_attribute_precond (str) &&
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_precond (this, str) &&
		GoTextual::metaclass()->
		    _call_modify_attribute_precond (this, str))
    {
	/* update _justification */
	str->justification = _justification;
        _justification = newJustification;

        /* justification preconditions call */
	_call_modify_attribute_postcond (str);
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_postcond (this, str);
	GoTextual::metaclass()->
	    _call_modify_attribute_postcond (this, str);
    }

    delete str;
    return ret;
}


/**
FUNCTION
    interline (GNumber);
DESCRIPTION
    Changes the interline.
**/

GBool
GoTextual::interline (GNumber newInterline)
{
    GBool ret = GTRUE;
    GModifyAttributeStruct* str = new GModifyAttributeStruct;
    str->type = GINTERLINE_CHANGE_NOTIFY;
    str->interline = newInterline;

    /* call interline postconditions */
    if (ret = _call_modify_attribute_precond (str) &&
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_precond (this, str) &&
		GoTextual::metaclass()
		    ->_call_modify_attribute_precond (this, str))
    {
	/* update _interline */
	oldClip = myClip();
        str->interline = _interline;
	_interline = newInterline;

        /* call presentation postconditions */
	_call_modify_attribute_postcond (str);
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_postcond (this, str);
	GoTextual::metaclass()->
	    _call_modify_attribute_postcond (this, str);
    }

    delete str;
    return ret;
}


/**
FUNCTION
    margin (GNumber, GMarginSide);
DESCRIPTION
    Function used to set the text margins.
**/

GBool
GoTextual::margin (GNumber newMargin, GMarginSide marginSide)
{
GBool ret;

    if (newMargin < 0) return GFALSE;
    GModifyAttributeStruct *str = new GModifyAttributeStruct;
    str->type = GMARGIN_SIDE_CHANGE_NOTIFY;
    str->marginValue.marginSide = marginSide;
    str->marginValue.margin = newMargin;
    str->firstPosition = -1;
    str->lastPosition = -1;

    /* new margin preconditions call */
    if (ret = _call_modify_attribute_precond (str) &&
	((GoTextualMetaclass*)getMetaclass())->
	   _call_modify_attribute_precond (this, str) &&
		GoTextual::metaclass()->
		    _call_modify_attribute_precond (this, str))
    {
	oldClip = myClip();
	GNumber oldMargin = 0;
	switch (marginSide){
	    case GLEFT_SIDE_MARGIN:
		str->marginValue.margin = _leftMargin;
		oldMargin = _leftMargin;
		_leftMargin = newMargin;
		break;

	    case GRIGHT_SIDE_MARGIN:
		str->marginValue.margin = _rightMargin;
		oldMargin = _rightMargin;
		_rightMargin = newMargin;
		break;

	    case GTOP_SIDE_MARGIN:
		str->marginValue.margin = _topMargin;
		oldMargin = _topMargin;
		_topMargin = newMargin;
		break;

	    case GBOTTOM_SIDE_MARGIN:
		str->marginValue.margin = _bottomMargin;
		oldMargin = _bottomMargin;
		_bottomMargin = newMargin;
		break;
	}

	/* update */
	MyClipHasChanged();
	if (backgroundMapped()){
	    GRegion deltaClip;
	    if (newMargin > oldMargin){
		GSub (myClip(), oldClip, deltaClip);
		RedrawForeOnly (deltaClip);
	    }
	    else{
		GSub (oldClip, myClip(), deltaClip);
		RedrawBackAndFore (deltaClip);
	    }
	}

	/* new margin postconditions call */
	_call_modify_attribute_postcond (str);
	((GoTextualMetaclass*)getMetaclass())->
	    _call_modify_attribute_postcond (this, str);
	GoTextual::metaclass()->
	    _call_modify_attribute_postcond (this, str);
    }

    delete str;
    return ret;
}


/**
FUNCTION
    margin (GMarginSide);
DESCRIPTION
    Function used to get the margin.
**/

GNumber
GoTextual::margin (GMarginSide marginSide) const
{
    switch (marginSide){
	case GLEFT_SIDE_MARGIN:
	    return (_leftMargin);
	case GRIGHT_SIDE_MARGIN:
	    return (_rightMargin);
	case GTOP_SIDE_MARGIN:
	    return (_topMargin);
	case GBOTTOM_SIDE_MARGIN:
	    return (_bottomMargin);
	default:
	    GError("GoTextual::margin ERROR!");
	    return 0.0;	
    }
}
