#ifdef REV_INFO
static char rcsid[] = "$Id: GifLoad.cc,v 1.2 1992/12/03 09:04:18 jacomet Exp $";
#endif
/*****************************************************************************
MODULE:  TOOLS
FILE:    gifload.cc
AUTHOR:  Philippe Smadja
DATE:    June 91

DESCRIPTION:
  Load a gif file to make an PixelArray and an ImageBrush
  This code uses the gif description (a C program) found in the "xloadimage"
  free software.
  The gif soft nis written in C++ to make
  the pixelarray and the imagebrush.
  The GifLoad() Function take a Medium in parameter or a filename.
  
COPYRIGHT:
  Copyright (c) 1990, 199n1 Bull S.A., Paris, France

  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.

  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.

  
  Copyright 1989, 1990 Kirk L. Johnson.   For GIF


  Permission to use, copy, modify, distribute, and sell this
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in
  all copies and that both that copyright notice and this
  permission notice appear in supporting documentation. The
  author makes no representations about the suitability of this
  software for any purpose. It is provided "as is" without express
  or implied warranty.

  THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
  INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS,
  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT
  OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
  LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
  NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
  CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

******************************************************************************/

#include <Go/Ext/Gif.h>
#include <Go/Tools/Alloc.h>
#include <Go/Tools/FileMedium.h>
#include <Go/Tools/Error.h>
#include <assert.h>

/****
****/


/****
 **
 ** local #defines
 **
 ****/

/*
 * gifin return codes
 */
#define GIFIN_SUCCESS       0   /* success */
#define GIFIN_DONE          1   /* no more images */

#define GIFIN_ERR_BAD_SD   -1   /* bad screen descriptor */
#define GIFIN_ERR_BAD_SEP  -2   /* bad image separator */
#define GIFIN_ERR_BAD_SIG  -3   /* bad signature */
#define GIFIN_ERR_EOD      -4   /* unexpected end of raster data */
#define GIFIN_ERR_EOF      -5   /* unexpected end of input stream */
#define GIFIN_ERR_FAO      -6   /* file already open */
#define GIFIN_ERR_IAO      -7   /* image already open */
#define GIFIN_ERR_NFO      -8   /* no file open */
#define GIFIN_ERR_NIO      -9   /* no image open */

/*
 * colormap indices 
 */

#define GIF_RED  0
#define GIF_GRN  1
#define GIF_BLU  2

/*
 *
 */
#define GIF_SIG      "GIF87a"
#define GIF_SIG_LEN  6          /* GIF signature length */
#define GIF_SD_SIZE  7          /* GIF screen descriptor size */
#define GIF_ID_SIZE  9          /* GIF image descriptor size */

#define GIF_SEPARATOR   ','     /* GIF image separator */
#define GIF_EXTENSION   '!'     /* GIF extension block marker */
#define GIF_TERMINATOR  ';'     /* GIF terminator */

#define STAB_SIZE  4096         /* string table size */
#define PSTK_SIZE  1024         /* pixel stack size */
#define LOCAL_BUFSZ 256
#define NULL_CODE  -1           /* string table null code */

#define valToMem(VAL,PTR,LEN)    ((LEN) == 1 ? \
				  (unsigned long)(*(PTR) = (uchar)(VAL)) : \
				  doValToMem(VAL,PTR,LEN))


#define PUSH_PIXEL(p)                                     {  \
  if (pstk_idx == PSTK_SIZE)                                \
    GError("GGifLoad(): pixel stack overflow in PUSH_PIXEL()");    \
  else                                                      \
    pstk[pstk_idx++] = (char)(p);                            \
}


/****
 **
 ** local variables for stream decoder
 **
 ****/

static int interlace_start[4]= { /* start line for interlacing */
  0, 4, 2, 1
};

static int interlace_rate[4]= { /* rate at which we accelerate vertically */
  8, 8, 4, 2
};



static int  root_size;          /* root code size */
static int  clr_code;           /* clear code */
static int  eoi_code;           /* end of information code */
static int  code_size;          /* current code size */
static int  code_mask;          /* current code mask */
static int  prev_code;          /* previous code */

/*
 * NOTE: a long is assumed to be at least 32 bits wide
 */
static long work_data;          /* working bit buffer */
static int  work_bits;          /* working bit count */

static uchar* buf = NULL;       /* byte buffer */
static int  buf_cnt;            /* byte count */
static int  buf_idx;            /* buffer index */

static int table_size;          /* string table size */
static int* prefix = NULL;      /* string table : prefixes */
static int* extnsn = NULL;      /* string table : extensions */

static uchar* pstk = NULL;      /* pixel stack */
static int  pstk_idx;           /* pixel stack pointer */

/**
  local functions
  **/
static int   gifin_get_pixel(GMedium& medium, int* pel);
static int   gifin_skip_extension(GMedium& medium);
static int   gifin_read_data_block(GMedium& medium);
static int   gifin_push_string(int code);
static void  gifin_add_string(int p, int e);
static ulong doValToMem(ulong val, uchar* p, uint len);


/*
 * load a colormap from the input stream
 */

static int gifin_load_cmap(GMedium& medium, GImageBrushValues& ib_values, int ncolors)
{
#ifdef NO_ALLOCA
    GShColor** colors  = (GShColor**)GMALLOC(ncolors * sizeof (GShColor*));
#else
    GShColor** colors = (GShColor**)alloca(ncolors * sizeof (GShColor*));
#endif
    
    int i;
    
    for (i=0; i< ncolors; i++)  {

	char buf[3];
	if (medium.read( (char*)buf, 3) != 3) {
	    /**
	      ERROR:
	      free the previous loaded color
	      **/
	    for (int j = 0; j < i; j++)
		colors[j]->destroy();
	    return GIFIN_ERR_EOF;
	}

	GRgb rgb;
	rgb.red   = buf[GIF_RED] * (GMAX_RGB_VALUE / 256);
	rgb.green = buf[GIF_GRN] * (GMAX_RGB_VALUE / 256);
	rgb.blue  = buf[GIF_BLU] * (GMAX_RGB_VALUE / 256);
	colors[i] = GCreateShColor(&rgb);
    }

    /**
      make the colormap with the set of colors.
      **/
    GColormapValues cmap_values;
    cmap_values.colors =  colors;
    cmap_values.n_color = ncolors;
    ib_values.colormap = GCreateShColormap(GCOLORMAP_COLORS, &cmap_values);

    /**
      we does need the colorsr
      for the function so destroy.
      **/
    for (i = 0; i < ncolors;)
	colors[i++]->destroy();
    
#ifdef NO_ALLOCA
    GFREE(colors);
#endif
    return GIFIN_SUCCESS;
}

static int gifin_open_file(GMedium& medium,
			   GImageBrushValues& ib_values, GPixelArrayValues& pa_values)
{


  /* check GIF signature */
  if (medium.read( (char*)buf, GIF_SIG_LEN) != GIF_SIG_LEN)
    return GIFIN_ERR_EOF;

  buf[GIF_SIG_LEN] = '\0';
  if (strcmp((char *) buf, GIF_SIG) != 0)
    return GIFIN_ERR_BAD_SIG;

  /* read screen descriptor */
  if (medium.read( (char*)buf, GIF_SD_SIZE) != GIF_SD_SIZE)
    return GIFIN_ERR_EOF;

  /* decode screen descriptor */
  
  pa_values.width   = (buf[1] << 8) + buf[0];
  pa_values.height  = (buf[3] << 8) + buf[2];
  pa_values.depth   = (buf[4] & 0x07) + 1;
  GBool cmap_flag  = (buf[4] & 0x80) ? 1 : 0;

  /**
    gifin_color_bits   = ((buf[4] & 0x70) >> 4) + 1;  bits for color resolution ??
    gifin_bg_color     = buf[5];                      background ??
    **/

  if (buf[6] != 0)
    return GIFIN_ERR_BAD_SD;

  /* load global colormap */
  if (cmap_flag)  {
    if (gifin_load_cmap(medium, ib_values, 1 << pa_values.depth) != GIFIN_SUCCESS)
      return GIFIN_ERR_EOF;
  } else 
      ib_values.colormap = NULL;

  /* done! */
  return GIFIN_SUCCESS;
}


/*
 * open next GIF image in the input stream; returns GIFIN_SUCCESS if
 * successful. if there are no more images, returns GIFIN_DONE. (might
 * also return various GIFIN_ERR codes.)
 */

static int gifin_open_image(GMedium& medium, GImageBrushValues& ib_values,
			    GPixelArrayValues& pa_values,
			    GBool& interlace_flag)
{
  int i;
  int separator;


  /* skip over any extension blocks */
  do
  {
    separator = medium.mgetc();
    if (separator == GIF_EXTENSION)
    {
      if (gifin_skip_extension(medium) != GIFIN_SUCCESS)
        return GIFIN_ERR_EOF;
    }
  }
  while (separator == GIF_EXTENSION);

  /* check for end of file marker */
  if (separator == GIF_TERMINATOR)
    return GIFIN_DONE;

  /* make sure we've got an image separator */
  if (separator != GIF_SEPARATOR)
    return GIFIN_ERR_BAD_SEP;

  /* read image descriptor */
  if (medium.read( (char*)buf, GIF_ID_SIZE) != GIF_ID_SIZE)
    return GIFIN_ERR_EOF;

  /* decode image descriptor */
  /***************************/

  /**
    gifin_img_left       = (buf[1] << 8) + buf[0];  unused
    gifin_img_top        = (buf[3] << 8) + buf[2];  unused
    **/
#define MIN(x,y)   ((ushort)min((long)(x), (long)(y)))
  
  pa_values.width            = MIN (pa_values.width,  (buf[5] << 8) + buf[4]);
  pa_values.height           = MIN (pa_values.height,  (buf[7] << 8) + buf[6]);
  /**
    sometime error:
    pa_values.depth                = (buf[8] & 0x07) + 1;
    **/
  pa_values.xoffset          = 0;
  pa_values.bitmap_unit      = 8;
  pa_values.padding          = 8;
  pa_values.bits_per_pixel   =  (pa_values.depth + 7) / 8 * 8;
  pa_values.format.coding    = GBITMAP_CODING;
  pa_values.format.bit       = GMSB_FIRST;      /* cf memToVal... */
  pa_values.format.byte      = GMSB_FIRST;
  pa_values.data             = (uchar*)GMALLOC(((ulong)pa_values.bits_per_pixel) / 8 *
					       pa_values.width * pa_values.height);


  GBool cmap_flag       = (buf[8] & 0x80) ? GTRUE : GFALSE;
  interlace_flag       = (buf[8] & 0x40) ? GTRUE : GFALSE;

  /* load local colormap */
  if (cmap_flag) {
      if (ib_values.colormap)
	  ib_values.colormap->destroy();
      if (gifin_load_cmap(medium, ib_values, 1 << pa_values.depth) != GIFIN_SUCCESS)
	  return GIFIN_ERR_EOF;
  }


  /* initialize raster data stream decoder */
  root_size = medium.mgetc();
  clr_code  = 1 << root_size;
  eoi_code  = clr_code + 1;
  code_size = root_size + 1;
  code_mask = (1 << code_size) - 1;
  work_bits = 0;
  work_data = 0;
  buf_cnt   = 0;
  buf_idx   = 0;

  /* initialize string table */
  for (i=0; i<STAB_SIZE; i++)
  {
    prefix[i] = NULL_CODE;
    extnsn[i] = i;
  }

  /* initialize pixel stack */
  pstk_idx = 0;

  /* done! */
  return GIFIN_SUCCESS;
}



static int gifin_get_pixel(GMedium& medium, int* pel)
{
  int  code;
  int  first;
  int  place;

  /* decode until there are some pixels on the pixel stack */
  while (pstk_idx == 0)
  {
    /* load bytes until we have enough bits for another code */
    while (work_bits < code_size)
    {
      if (buf_idx == buf_cnt)
      {
        /* read a new data block */
        if (gifin_read_data_block(medium) != GIFIN_SUCCESS)
          return GIFIN_ERR_EOF;

        if (buf_cnt == 0)
          return GIFIN_ERR_EOD;
      }

      work_data |= ((long) buf[buf_idx++]) << work_bits;
      work_bits += 8;
    }

    /* get the next code */
    code        = (int) (work_data & code_mask);
    work_data >>= code_size;
    work_bits  -= code_size;

    /* interpret the code */
    if (code == clr_code)
    {
      /* reset decoder stream */
      code_size  = root_size + 1;
      code_mask  = (1 << code_size) - 1;
      prev_code  = NULL_CODE;
      table_size = eoi_code + 1;
    }
    else if (code == eoi_code)
    {
      /* Ooops! no more pixels */
      return GIFIN_ERR_EOF;
    }
    else if (prev_code == NULL_CODE)
    {
      gifin_push_string(code);
      prev_code = code;
    }
    else
    {
      if (code < table_size)
      {
        first = gifin_push_string(code);
      }
      else
      {
        place = pstk_idx;
        PUSH_PIXEL(NULL_CODE);
        first = gifin_push_string(prev_code);
        pstk[place] = first;
      }

      gifin_add_string(prev_code, first);
      prev_code = code;
    }
  }

  /* pop a pixel off the pixel stack */
  *pel = (int) pstk[--pstk_idx];

  /* done! */
  return GIFIN_SUCCESS;
}

static ulong doValToMem(ulong val, uchar* p, uint len)
{
    int a;
    for (a= len - 1; a >= 0; a--) {
	*(p + a)= (uchar)(val & 0xff);
	val >>= 8;
    }
    return val;
}




 
/*
 * skip an extension block in the input stream
 */

static int gifin_skip_extension(GMedium& medium)
{
  int function;

  /* get the extension function byte */
  function = medium.mgetc();

  /* skip any remaining raster data */
  do
  {
    if (gifin_read_data_block(medium) != GIFIN_SUCCESS)
      return GIFIN_ERR_EOF;
  }
  while (buf_cnt > 0);

  /* done! */
  return GIFIN_SUCCESS;
}

/*
 * read a new data block from the input stream
 */

static int gifin_read_data_block(GMedium& medium)
{
  /* read the data block header */
  buf_cnt = medium.mgetc();

  /* read the data block body */
  if (medium.read( (char*)buf, buf_cnt) != buf_cnt)
    return GIFIN_ERR_EOF;

  buf_idx = 0;

  /* done! */
  return GIFIN_SUCCESS;
}

/*
 * push a string (denoted by a code) onto the pixel stack
 * (returns the code of the first pixel in the string)
 */

static int gifin_push_string(int code)
{
  int rslt;

  while (prefix[code] != NULL_CODE)
  {
    PUSH_PIXEL(extnsn[code]);
    code = prefix[code];
  }

  PUSH_PIXEL(extnsn[code]);
  rslt = extnsn[code];

  return rslt;
}

/*
 * add a new string to the string table
 */

static void gifin_add_string(int p, int e)
{
  prefix[table_size] = p;
  extnsn[table_size] = e;

  if ((table_size == code_mask) && (code_size < 12))
  {
    code_size += 1;
    code_mask  = (1 << code_size) - 1;
  }

  table_size += 1;
}


/**
FUNCTION
    Alloc/Free GifData
DESCRIPTION
    alloc/free the gif data. these data were declared int global
    in the source C version.
**/
static void AllocGifData()
{
    pstk =   (uchar*)GMALLOC(PSTK_SIZE *  sizeof(uchar));
    buf    = (uchar*)GMALLOC(LOCAL_BUFSZ * sizeof(uchar));
    prefix =  (int*) GMALLOC(STAB_SIZE *  sizeof(int));
    extnsn =  (int*) GMALLOC(STAB_SIZE *  sizeof(int));
}

static void FreeGifData()
{
    GFREE(pstk); assert (pstk == NULL);
    GFREE(buf);
    GFREE(prefix);
    GFREE(extnsn);
}


/**
FUNCTION
    GGifLoad()
DESCRIPTION
    Read the fullname file to make a GPixelArray and a Imagebrush.
    for Gif the GImageBrushValues will initialize:
            the function:  copy,
            the planemask: all,
            the fillStyle: GfillSolid,
            the colormap:  available if non NULL.
	    the stipple is initialize to NULL

    All GPixelArrayValues will be initialized to 0 on error.

    The function returns GTRUE if ALL the above fields are available
    else it returns GFALSE (the file is not at the GIF format).
**/
GBool GGifLoad(const char* fullname, GImageBrushValues& ib_values,
	    GPixelArrayValues& pa_values)
{
    GFileMedium* medium = new GFileMedium;
    GBool ret;
    
#if defined (__BORLANDC__) || defined (__MSC7__)
    if (medium->open(fullname, "r+b") == 0)
#else
    if (medium->open(fullname, "r") == 0)
#endif
	ret = GGifLoad (medium, ib_values, pa_values);
    else
	ret = GFALSE;

    delete medium;
    return ret;
}

GBool GGifLoad(GMedium* medium, GImageBrushValues& ib_values,
	     	    GPixelArrayValues& pa_values)
{
    int    x, y, pixel, pass, scanlen;
#if defined (__BORLANDC__) || defined (__MSC7__)
    uchar huge * pixptr;
    uchar huge * pixline;
    uchar huge * datatemp;
#else
    uchar* pixptr;
    uchar* pixline;
#endif
    GBool   interlace_flag;
    GBool   ret;
    int    pixlen;

    ib_values.function   = GCOPY;
    ib_values.plane_mask = GALL_PLANES;
    ib_values.fill.style = GFILL_SOLID;
    ib_values.colormap   = NULL;
    ib_values.stipple = NULL ;
    
    bzero(&pa_values, sizeof (GPixelArrayValues));
    
    AllocGifData();
    
    if ((gifin_open_file(*medium, ib_values, pa_values) != GIFIN_SUCCESS) ||
	(gifin_open_image(*medium, ib_values, pa_values, interlace_flag) != GIFIN_SUCCESS)) {

	ret  = GFALSE;
	goto out;
    }

    pixlen =  pa_values.bits_per_pixel / 8;
#if defined (__BORLANDC__) || defined (__MSC7__)
    datatemp = pa_values.data;
#endif

    if (interlace_flag) {
	/**
	  interlaced image -- futz with the vertical trace.  i wish i knew what
	  kind of drugs the GIF people were on when they decided that they
	  needed to support interlacing.
	  **/
	scanlen= pa_values.height *  pixlen;

	/* interlacing takes four passes to read, each starting at a different
	 * vertical point.
	 */

	for (pass= 0; pass < 4; pass++) {
	    y= interlace_start[pass];
	    scanlen= pa_values.width * pixlen * interlace_rate[pass];
#if defined (__BORLANDC__) || defined (__MSC7__)
	    pixline= datatemp + (y * ((ulong)pa_values.width & 0x0000ffffL) * pixlen);
#else
	    pixline= pa_values.data + (y * pa_values.width * pixlen);
#endif
	    while (y < pa_values.height) {
		pixptr= pixline;
		for (x= 0; x < pa_values.width; x++) {
		    if (gifin_get_pixel(*medium, &pixel) != GIFIN_SUCCESS) {
			ret = GFALSE;
			goto out;
		    }
		    valToMem(pixel, pixptr, pixlen);
#if defined (__BORLANDC__) || defined (__MSC7__)
		    pixptr += (((ulong)pixlen) & 0x0ffffL);
#else
		    pixptr += pixlen;
#endif
		}
		y += interlace_rate[pass];
#if defined (__BORLANDC__) || defined (__MSC7__)
		pixline += ((ulong)(scanlen) & 0x0ffffL);
#else
		pixline += scanlen;
#endif
	    }
	}
    }

    /**
      not an interlaced image, just read in sequentially
      **/
    else {
	pixptr= pa_values.data;
	for (y= 0; y < pa_values.height; y++)
	    for (x= 0; x < pa_values.width; x++) {
		if (gifin_get_pixel(*medium, &pixel) != GIFIN_SUCCESS) {
		    /**
		      error on file: stop with the read data
		      the valid height becomes  the current y.
		      width does not change.
		      **/
		    pa_values.height = y;  
		    break;  
		}
		valToMem(pixel, pixptr, pixlen);
#if defined (__BORLANDC__) || defined (__MSC7__)
		pixptr += ((ulong)(pixlen) & 0x0FFFFL);
#else
		pixptr += pixlen;
#endif
	    }
    }

    ret = GTRUE;
    
  out:
    FreeGifData();
    if (ret == GFALSE)
	GFREE(pa_values.data);
    return ret;
}
