#ifdef REV_INFO
static char rcsid[] = "$Id: GoInDisp.cc,v 1.2 1992/11/25 15:21:45 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Tools
FILE:    InDisp.cc
AUTHOR:  Olivier Jojic, Philippe Smadja
DATE:    29 May 1990

DESCRIPTION:
    See Go/GoInDisp.h

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/GoInDisp.h>

/*
 * declaration of the global input dispatcher;
 * It MUST be instancied by application needing input.
 */
GoInputDispatcher*& GInputDispatcher ()
{
static GoInputDispatcher* _gInputDispatcher = NULL;
	return _gInputDispatcher;
}


struct Filter {
    GInDispFilterProc procedure;
    void*             closure;
    GBool             is_to_be_deleted;

    Filter(GInDispFilterProc p, void* c) {
	procedure = p;
	closure = c;
	is_to_be_deleted = GFALSE;
    };
};


struct Idle {
    GInDispIdleProc procedure;
    void*           closure;
    GBool           first_call;
    GIdle           request;

    inline Idle(GInDispIdleProc p, void*  c) {
	procedure = p;
	closure = c;
	first_call = GTRUE;
	// request not significant is first_call is true;
    };
};

GoInputDispatcher::GoInputDispatcher() :
	_filters(sizeof(Filter), 2), _idles(sizeof(Idle),2)
{
    _stop = GFALSE;
    _value = NULL;
    _reset_idle = GFALSE;
}

GoInputDispatcher::~GoInputDispatcher()
{
}

void GoInputDispatcher::stopEventLoop (void* value)
{
    _stop = GTRUE;
    _value = value;
}

/**
FUNCTION
    add/rem filterProc.
DESCRIPTION
   put the filter in a private list.
   This list will be use by "CallFilters" to retreive filters.
**/
void GoInputDispatcher::filterProc(GInDispFilterProc procedure,
				   void* closure, GAddOrRemove add)
{
    if (add == GADD)
    {
	Filter filter (procedure, closure);
    	_filters.add(&filter);
    }
    else
    {
	Filter* filter = (Filter*)_filters.data();
	for (int i = 0 ; i < _filters.nelems(); i++, filter++)
	    if (filter->procedure == procedure && filter->closure == closure)
	    {
		_filters.rem(i);
		break;
	    }
    }
}


/**
FUNCTION
    protected CallFilters()
DESCRIPTION
    This method calles all the filter. If one returns GFALSE we stop the loop
    and we returns GFALSE: the caller (a subclass generaly the methiod runEventLoop)
    does have to handle event.
    The function duplicates the list because a filter can remove filter from the list.
**/
GBool GoInputDispatcher::CallFilters(void* event)
{
    GBool handle_event = GTRUE;
	
    if (_filters.nelems())
    {
	// dup the filters list in local GArray;

	GArray dup_filters(_filters.data(), _filters.nelems(), sizeof(Filter));
	Filter* filter = (Filter*)dup_filters.data();
	
	for (int i = 0; i < dup_filters.nelems(); i++, filter++)
	    if ((*filter->procedure)(filter->closure, event) == GFALSE)
	    {
		handle_event = GFALSE;
		break;
	    }
    }
    return handle_event;
}




/**
FUNCTION
    protected methods to handle the idle.
DESCRIPTION
    
**/

GBool GoInputDispatcher::CallIdles()
{
    GBool again = GFALSE;

    _reset_idle = GFALSE;

    if (_idles.nelems())
    {
	// dup the list in local variable;
	GArray dup_idles(_idles.data(), _idles.nelems(), sizeof(Idle));
	
	Idle* idle = (Idle*)dup_idles.data();
	
	for (int i = 0; i < dup_idles.nelems(); i++, idle++)
	{
	    if (idle->first_call) {
		idle->first_call = GFALSE;
		idle->request = (*idle->procedure)(idle->closure);
	    }
	    else  if (idle->request == GIDLE_AGAIN)
		idle->request = (*idle->procedure)(idle->closure);
	
	    switch (idle->request) {
	      case GIDLE_REMOVE:
		// remove it from the NON DUPLICATED list;
		idleProc(idle->procedure, idle->closure, GREMOVE);
		break;
	      case GIDLE_AGAIN:
		again = GTRUE;
		break;
	      default:
		break;
	    }
	}
	_reset_idle = GTRUE;
    }
    return again;
}

void GoInputDispatcher::idleProc(GInDispIdleProc procedure, void* closure,
				 GAddOrRemove add)
{
    if (add == GADD)
    {
	Idle idle (procedure, closure);
    	_idles.add(&idle, 1, 0);         // the most recent has the highest priority;
    }
    else
    {
	Idle* idle = (Idle*)_idles.data();
	for (int i = 0 ; i < _idles.nelems(); i++, idle++)
	    if (idle->procedure == procedure && idle->closure == closure)
	    {
		_idles.rem(i);
		break;
	    }
    }
}


void  GoInputDispatcher::ResetIdles()
{
    if (_reset_idle)
    {
	_reset_idle = GFALSE;
	
	Idle* idle = (Idle*)_idles.data();
	for (int i = 0 ; i < _idles.nelems(); i++, idle++)
	    idle->first_call = GTRUE;
    }
}

