#ifdef REV_INFO
static char rcsid[] = "$Id: XtInDisp.cc,v 1.5 1992/11/25 16:09:24 kuta Exp $";
#endif
/*****************************************************************************
MODULE:  Tools
FILE:    XtInDisp.cc
AUTHOR:  Philippe Smadja
DATE:    29 May 1990

DESCRIPTION:
    See Go/XtInDisp.h

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/XtInDisp.h>
#include <Go/Tools/Error.h>
#include <assert.h>

#define UNDEF_WORKPROC  (XtWorkProcId)0
    
/**
FUNCTION
    add/rem timer;
DESCRIPTION
   we add/remove a StaticTimerProc in the Xt Toolkit . One Timer by
   application request.
   The StaticTimerProc retreives the corresponding timer, calls it
   and removes it  from the list.
**/
    
void StaticTimerProc(XtPointer indisp, XtIntervalId* xtid)
{
    ((XtInputDispatcher*)indisp)->HandleTimerProc(*xtid);
}

struct Timer {
    GInDispTimerProc procedure;
    void*            closure;
    ulong            interval;
    XtIntervalId     xtid;      // the corresponding ident in the toolkit;

    Timer(XtInputDispatcher* indisp, XtAppContext app, GInDispTimerProc p, void* c, ulong i)
    {
	procedure = p;
	closure = c;
	interval = i;
	xtid = XtAppAddTimeOut(app, interval, StaticTimerProc, (XtPointer)indisp);
    };
    ~Timer()
    {
	if (xtid)
	    XtRemoveTimeOut(xtid);
    }
};


void XtInputDispatcher::HandleTimerProc(XtIntervalId xtid)
{
    Timer** ptimer = (Timer**)_timers.data();
    if (ptimer)
    {
	Timer* timer = *ptimer;
	for (int i = 0; i < _timers.nelems(); i++, timer = *ptimer++)
	    if (timer->xtid == xtid)
	    {
		// remove the timer from the list;
		// so the field is now free. It will be reused;
		// if the timer itself add a new timer in the list;
		_timers.rem(i);
	    
		// call it;
		(*timer->procedure)(timer->closure);

		// delete it: In this case the remove is done;
		// by the toolkit itself;
		timer->xtid = 0;
		delete timer;

		// In Go: timer is an event so we have to update;
		// The Idle management;
		ResetIdles();
		UpdateWorkProc();
		    
		break;
	    }
    }
}

/**
FUNCTION
    add/rem timer
DESCRIPTION
     use the XtApp...Timeout().
**/
void XtInputDispatcher::timerProc(GInDispTimerProc procedure,
				  void* closure, ulong interval,
				  GAddOrRemove add)
{
    if (add == GADD)
    {
	Timer* timer = new Timer(this, _app, procedure, closure, interval);
	_timers.add(timer);
    }
    else
    {
	// GREMOVE;
	Timer** ptimer = (Timer**)_timers.data();
	if (ptimer)
	{
	    Timer* timer = *ptimer;
	    for (int i = 0 ; i < _timers.nelems(); i++, timer = *ptimer++)
		if (timer->procedure == procedure && timer->closure == closure &&
		    timer->interval == interval)
		{
		    delete timer;
		    _timers.rem(i);
		    break;
		}
	}
    }
}

Boolean StaticWorkProc(XtPointer indisp)
{
    return (Boolean)((XtInputDispatcher*)indisp)->HandleWorkProc();
}

GBool XtInputDispatcher::HandleWorkProc()
{
    GBool again = CallIdles();
    if (again)
    {
	// do not remove the XtWorkProc;
	return GFALSE;
    }
    // return True: the toolkit will removed it;
    // I will be added by  ::UpdateWorkProc();
    _workproc = UNDEF_WORKPROC;
    return GTRUE;
}

/**
FUNCTION
    add/rem idleProc
DESCRIPTION
    append ore remove an idleProc is the idles list;
    if the list is non empty ask Xt to use the  StaticWorkProc as a workproc.
    by be carefull. The specification of an IdleProc are not
    the same as an  XtWorkProc.
**/
void XtInputDispatcher::idleProc(GInDispIdleProc procedure, void* closure,
				 GAddOrRemove add)
{

    GoInputDispatcher::idleProc(procedure, closure, add);
    UpdateWorkProc();
}

void XtInputDispatcher::UpdateWorkProc()
{
    if (_workproc == UNDEF_WORKPROC && _idles.nelems())
	_workproc = XtAppAddWorkProc(_app, StaticWorkProc, (XtPointer)this);
    else if (_workproc && _idles.nelems() == 0)
    {
	XtRemoveWorkProc(_workproc);
	_workproc = UNDEF_WORKPROC;
    }
}

/**
FUNCTION
    runEventLoop()
DESCRIPTION
    it is like an XtMainLoop(). But we can return from it and we call filters.
    Note that idle is handled by XtAppNextEvent using XtWorkProc.
    All added handler are used + all event handler added to by XtAddEventHandler().
**/
void* XtInputDispatcher::runEventLoop()
{
    _run_event_loop_is_running++;
    _stop = GFALSE;
    while (! _stop)
    {
	XEvent xevent;
	XtAppNextEvent(_app, &xevent);
	if (CallFilters(&xevent))
	    XtDispatchEvent(&xevent);

	ResetIdles();
	UpdateWorkProc();
    }
    
    _stop = GFALSE;
    _run_event_loop_is_running--;
    return _value;
}

/**
FUNCTION
    eventReady();
DESCRIPTION
    return true if an event (or  timer) is ready.
**/
GBool XtInputDispatcher::eventReady()
{
    return (GBool)(XtAppPending(_app) != 0);
}

/**
FUNCTION
    handleNextEvent()
DESCRIPTION
    read (block if no event) the next event (possibily timer etc...)
    and handle it;
**/

void XtInputDispatcher::handleNextEvent()
{
    XtInputMask  mask = XtAppPending(_app);
    if (!mask || mask & XtIMXEvent)
    {
	XEvent xevent;
	XtAppNextEvent(_app, &xevent);
	if (CallFilters(&xevent))
	    XtDispatchEvent(&xevent);

	ResetIdles();
	UpdateWorkProc();
    }
    else if (mask & XtIMTimer)
	XtAppProcessEvent(_app, XtIMTimer);
    else
    {
	assert (mask & XtIMAlternateInput);
	XtAppProcessEvent(_app, XtIMAlternateInput);
    }
}

/**
FUNCTION
    constructor/destructor
DESCRIPTION
    
**/
 XtInputDispatcher::XtInputDispatcher (XtAppContext app)
{
    _app = app;
    _run_event_loop_is_running = 0;
    _workproc = UNDEF_WORKPROC;
}

XtInputDispatcher::~XtInputDispatcher()
{
    Timer** timer = (Timer**)_timers.data();
    for (int i = 0 ; i < _timers.nelems(); i++, timer++)
	delete *timer;

    if (_workproc)
	XtRemoveWorkProc(_workproc);
}

