#ifdef REV_INFO
static char rcsid[] = "$Id: Event.cc,v 1.3 1992/12/10 14:05:33 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  Go Kernel
FILE:    Event.cc
AUTHOR:  Olivier Jojic
DATE:    Jan 90

DESCRIPTION:
    See the Event.h header.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/machdef.h>
#include <Go/Event.h>
#include <Go/Tools/Error.h>
#include <string.h>

/*
 * GEvent methods
 */
char  GEvent::character () const
{
    GError("event %d (from %x) has no character !", _type, _origin);
    return 0;
}

const char* GEvent::message () const
{
    GError("event %d (from %x) has no message !", _type, _origin);
    return NULL;
}

GButton GEvent::button () const
{
    GError("event %d (from %x) has no button !", _type, _origin);
    return GBUTTON1;
}

GFunctionKey GEvent::functionKey () const
{
    GError("event %d (from %x) has no function key !", _type, _origin);
    return GKEY_LEFT;
}

ulong GEvent::modifiers () const
{
    GError("event %d (from %x) has no modifiers !", _type, _origin);
    return 0;
}

GPoint GEvent::position () const
{
    GError("event %d (from %x) has no position !", _type, _origin);
    return GPoint(0,0);
}

const GRegion& GEvent::region () const
{
    GError("event %d (from %x) has no region !", _type, _origin);
    return *(new GRegion());
}

const GoSensor* GEvent::sensor () const
{
    GError("event %d (from %x) has no sensor !", _type, _origin);
    return NULL;
}

/**
FUNCTION
    GButtonEvent and subclass methods
DESCRIPTION
    
**/
GButtonEvent::GButtonEvent (GEventType tp, GButton b, ulong m,
			    const GPoint& p, const GoInterf* gointerf, GTimeStamp time)
    : GEvent(tp, gointerf, time)
{
    _button = b;
    _modifiers = m;
    _pos = p;
}

GButton GButtonEvent::button() const
{
    return _button;
}

ulong GButtonEvent::modifiers() const
{
    return _modifiers;
}

GPoint GButtonEvent::position() const
{
    return _pos;
}

GButtonPressEvent::GButtonPressEvent (GButton b, ulong m,
				      const GPoint& p,
				      const GoInterf* gointerf, GTimeStamp time)
    : GButtonEvent(GBUTTON_PRESS_EVENT, b, m, p, gointerf, time)
{
}


GButtonReleaseEvent::GButtonReleaseEvent (GButton b, ulong m,
					  const GPoint& p,
					  const GoInterf* gointerf, GTimeStamp time)
    : GButtonEvent(GBUTTON_RELEASE_EVENT, b, m, p, gointerf, time)
{
}

GDoubleClickEvent::GDoubleClickEvent (GButton b, ulong m,
					const GPoint& p,
					const GoInterf* gointerf, GTimeStamp time)
    : GButtonEvent(GDOUBLE_CLICK_EVENT, b, m, p, gointerf, time)
{
}

GTripleClickEvent::GTripleClickEvent (GButton b, ulong m,
				      const GPoint& p,
				      const GoInterf* gointerf, GTimeStamp time)
    : GButtonEvent(GTRIPLE_CLICK_EVENT, b, m, p, gointerf, time)
{
}

/**
FUNCTION
    GCharEvent methods
DESCRIPTION
    
**/
GCharEvent::GCharEvent (char c, ulong modmask,
			const GPoint& pos, const GoInterf* gointerf, GTimeStamp time)
                        : GEvent(GCHAR_EVENT, gointerf, time)
{
    _char = c;
    _modifiers = modmask;
    _pos = pos;
}

char GCharEvent::character() const
{
    return _char;
}

ulong GCharEvent::modifiers() const
{
    return _modifiers;
}

GPoint GCharEvent::position() const
{
    return _pos;
}

/**
FUNCTION
    GStringEvent methods
DESCRIPTION
    
**/
GMessageEvent::GMessageEvent (const char* s, ulong modmask,
			    const GPoint& pos, const GoInterf* gointerf, GTimeStamp time)
                           : GEvent(GMESSAGE_EVENT, gointerf, time)
{
    _string = new char[strlen(s) + 1];
    strcpy(_string, s);
    _modifiers = modmask;
    _pos = pos;
}

GMessageEvent::~GMessageEvent ()
{
    delete _string;
}

const char* GMessageEvent::message() const
{
    return _string;
}

ulong GMessageEvent::modifiers() const
{
    return _modifiers;
}

GPoint GMessageEvent::position() const
{
    return _pos;
}


/**
FUNCTION
    GFunctionKeyEvent methods
DESCRIPTION
    
**/
GFunctionKeyEvent::GFunctionKeyEvent (GFunctionKey funkey, ulong modmask,
				      const GPoint& pos,
				      const GoInterf* gointerf, GTimeStamp time)
    : GEvent(GFUNCTION_KEY_EVENT, gointerf, time)
{
    _functionKey = funkey;
    _modifiers = modmask;
    _pos = pos;
}

GFunctionKey GFunctionKeyEvent::functionKey() const
{
    return _functionKey;
}

ulong GFunctionKeyEvent::modifiers() const
{
    return _modifiers;
}

GPoint GFunctionKeyEvent::position() const
{
    return _pos;
}

/**
FUNCTION
    GExposeEvent methods
DESCRIPTION
    
**/
GExposeEvent::GExposeEvent (const GRegion& region, const GoInterf* gointerf, GTimeStamp time)
                           : GEvent(GEXPOSE_EVENT, gointerf, time)
{
    _region = new GRegion (region);
}

GExposeEvent::~GExposeEvent ()
{
    delete _region;
}

const GRegion&  GExposeEvent::region() const
{
    return *_region;
}

/**
FUNCTION
    GMotionEvent methods
DESCRIPTION
    
**/
GMotionEvent::GMotionEvent (const GPoint& p, const GoInterf* gointerf, GTimeStamp time)
                           : GEvent(GMOTION_EVENT, gointerf, time)
{
    _pos = p;
}

GPoint GMotionEvent::position() const
{
    return _pos;
}

/**
FUNCTION
    GEnter/LeaveSensorEvent methods
DESCRIPTION
    
**/
GEnterSensorEvent::GEnterSensorEvent
    (const GoSensor* sensor, const GoInterf* gointerf, GTimeStamp time) :
    GEvent(GENTER_SENSOR_EVENT, gointerf, time)
{
    _sensor = sensor;
}

const GoSensor* GEnterSensorEvent::sensor() const
{
    return _sensor;
}

GLeaveSensorEvent::GLeaveSensorEvent
    (const GoSensor* sensor, const GoInterf* gointerf, GTimeStamp time)
	: GEvent(GENTER_SENSOR_EVENT, gointerf, time)
{
    _sensor = sensor;
}

const GoSensor* GLeaveSensorEvent::sensor() const
{
    return _sensor;
}


/**
FUNCTION
    GFocusIn/OutEvent methods
DESCRIPTION
    
**/
GFocusInEvent::GFocusInEvent
    (const GoSensor* sensor, const GoInterf* gointerf, GTimeStamp time) :
    GEvent(GFOCUS_IN_EVENT, gointerf, time)
{
    _sensor = sensor;
}

const GoSensor* GFocusInEvent::sensor() const
{
    return _sensor;
}

GFocusOutEvent::GFocusOutEvent
    (const GoSensor* sensor, const GoInterf* gointerf, GTimeStamp time) :
    GEvent(GFOCUS_OUT_EVENT, gointerf, time)
{
    _sensor = sensor;
}

const GoSensor* GFocusOutEvent::sensor() const
{
    return _sensor;
}


/* A RETIRER JUSQU'A LA FIN DU FICHIER */
class GStringEvent : public GEvent
{
  public:
    GStringEvent(const char* string, ulong modmask, const GPoint& pos,
		 const GoInterf* origin =  NULL, GTimeStamp time = 0);
    ~GStringEvent();

    virtual const char* string() const;
    virtual ulong modifiers() const;
    virtual GPoint position() const;

  private:
    char* _string;
    ulong _modifiers;
    GPoint _pos;
};

const char* GEvent::string () const
{
    GError("event %d (from %x) has no string !", _type, _origin);
    return NULL;
}
GStringEvent::GStringEvent (const char* s, ulong modmask,
			    const GPoint& pos, const GoInterf* gointerf, GTimeStamp time)
                           : GEvent(GMESSAGE_EVENT, gointerf, time)
{
    _string = new char[strlen(s) + 1];
    strcpy(_string, s);
    _modifiers = modmask;
    _pos = pos;
}

GStringEvent::~GStringEvent ()
{
    delete _string;
}

const char* GStringEvent::string() const
{
    return _string;
}

ulong GStringEvent::modifiers() const
{
    return _modifiers;
}

GPoint GStringEvent::position() const
{
    return _pos;
}
