#ifdef REV_INFO
static char rcsid[] = "$Id: GoImgInterf.cc,v 1.3 1992/09/25 15:22:32 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Go Interface
FILE:    GoInterfImg.cc
AUTHOR:  Philippe Smadja
DATE:    November 1990 (Version 1.0)

DESCRIPTION:
  Implementation of image function like deformation, resize ...

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/GoInterf.h>
#include <Go/Tools/Alloc.h>
#include <Go/Tools/Error.h>

/**
FUNCTION
    _transformPixelArrayValues.
DESCRIPTION
    General function to transform an pixel array to another one.
    the function handles rotation, resize, deformation, pixel translation etc...
**/
#if defined( __BORLANDC__) || defined(__MSC7__)
uchar huge * dstData;
const uchar huge * srcData;
#endif
void GoInterf::_transformPixelArrayValues(GPixelArrayValues* result,
					  ushort* xOrigin_result,
					  ushort* yOrigin_result,
					  const GShPixelArray* src,
					  const GCoordSys&   cs,
					  const GArray*      pixel_translation,
					  uchar bits_per_pixel, uchar depth,
					  uchar bit_order,      uchar byte_order,
					  uchar quality) const

{
    *result = *src->values();
    *xOrigin_result = *yOrigin_result = 0;

#define NtoI(n) ((int)((n)+0.5))
    int cs_I_x = NtoI(cs.I.x);
    int cs_I_y = NtoI(cs.I.y);
    int cs_J_x = NtoI(cs.J.x);
    int cs_J_y = NtoI(cs.J.y);
#undef NtoI
    
    if (cs_I_y == 0 && cs_I_x == src->values()->width &&
	cs_J_x == 0 && cs_J_y == src->values()->height) {
	/**
	  no resize and deformation
	  **/
	if (pixel_translation &&  src->values()->bits_per_pixel == 8 &&
	    bits_per_pixel == 8) {

	    result->depth   = depth;
	    result->padding = 8;

#if defined( __BORLANDC__) || defined( __MSC7__)
	    dstData =  (uchar*)GMALLOC(src->values()->width*
						src->values()->height);
#else
	    uchar*  dstData =  (uchar*)GMALLOC(src->values()->width*
						src->values()->height);
#endif
	    result->data =(uchar*)  dstData;
	    ulong* pixels = (ulong*)pixel_translation->data();
#if defined( __BORLANDC__) || defined( __MSC7__)
	     srcData = src->values()->data;
#else
	    const uchar*  srcData = src->values()->data;
#endif
	    if (src->values()->padding == 8) 
		for (ulong i=0; i < src->dataSize(); i++)
		    *dstData++ = (uchar)pixels[*srcData++];
	    else
		for (int y=0; y < src->values()->height; y++,
		     srcData = src->values()->data + y * src->bytesPerLine()) 
		    for (int x = 0; x < src->values()->width; x++)
			*dstData++ = (uchar)pixels[*srcData++];
	    return;
	}
    }


    /**
      resize and/or deformation
      *************************
      **/

    /**
      1ER: LOOK FOR FAST CASE
      ***********************
      **/
    if (cs_I_y == 0 && cs_I_x > 0 &&
	cs_J_x == 0 && cs_J_y > 0) {
	/**
	  normal resize
	  **/
	GBool ok  = GFALSE;
	switch (src->values()->bits_per_pixel) {
	  case 1:
	    if (pixel_translation == NULL) {
		/**
		  try to use fast function for DEPTH 1 image.
		  **/
		int src_width = src->values()->width;
		int x_reduce = (int)((float)src_width / cs_I_x + (GNumber)0.5) ;
		switch (x_reduce) {
		  case 2:
		    ok = _reduceImage_R2_bitsPerPixel1(result,src, cs_I_x, cs_I_y,
						       quality);
		    break;
		  case 3:
		    ok = _reduceImage_R3_bitsPerPixel1(result,src, cs_I_x, cs_I_y,
						       quality);
		    break;
		  case 4:
		    ok = _reduceImage_R4_bitsPerPixel1(result,src, cs_I_x, cs_I_y,
						       quality);
		    break;
		  default: {
		      int x_extend = (int) ((float)cs_I_x / src_width + (GNumber)0.5) ;
		      switch (x_extend) {
			case 2:
			  ok = _extendImage_E2_bitsPerPixel1(result,src, cs_I_x, cs_I_y,
							     quality);
			  break;
			case 3:
			  ok = _extendImage_E3_bitsPerPixel1(result,src, cs_I_x, cs_I_y,
							     quality);
			  break;
			case 4:
			  ok = _extendImage_E4_bitsPerPixel1(result,src, cs_I_x, cs_I_y,
							     quality);
			  break;
		      }
		  }
		}

		if (!ok) 
		    ok = _resizeImage_bitsPerPixel1(result, src,
						    cs_I_x, cs_J_y,
						    bits_per_pixel, depth,
						    bit_order, byte_order, quality);
	    }  /** end pixel_translation == NULL **/
	    if (!ok)
		goto general_Resize;
	    break;
	  case /*bits_per_pixel*/ 4:
	    ok = _resizeImage_bitsPerPixel4(result, src, pixel_translation,
					    cs_I_x, cs_J_y,
					    bits_per_pixel, depth,
					    bit_order, byte_order, quality);
	    if (!ok)
		goto general_Resize;
	    break;
	  case /*bits_per_pixel*/ 8:
	    ok = _resizeImage_bitsPerPixel8(result, src, pixel_translation,
					    cs_I_x, cs_J_y,
					    bits_per_pixel, depth,
					    bit_order, byte_order, quality);
	    if (!ok)
		goto general_Resize;
	    break;

	  general_Resize:
	  default:
	    ok = _resizeImage(result, src, pixel_translation,
			      cs_I_x, cs_J_y,
			      bits_per_pixel, depth,
			      bit_order, byte_order, quality);
	    break;
	}
	if (ok)
	    return;
	goto _GENERAL_DEFORMATION;
    }

    
    if (cs_I_y == 0 && cs_I_x == - src->values()->width &&
	cs_J_x == 0 && cs_J_y == src->values()->height) {
	/**
	  X miror
	  **/
	GBool ok ;
	switch (src->values()->bits_per_pixel) {
	  case 1:
	    ok = _mirorXaxe_bitsPerPixel1(result, src,
					  bits_per_pixel, depth,
					  bit_order, byte_order);
	    if (!ok)
		goto generalXaxe;
	    break;
	  case 4:
	    ok = _mirorXaxe_bitsPerPixel4(result, src, pixel_translation,
					  bits_per_pixel, depth,
					  bit_order, byte_order);
	    if (!ok)
		goto generalXaxe;
	    break;
	  case 8:
	    ok = _mirorXaxe_bitsPerPixel8(result, src, pixel_translation,
					  bits_per_pixel, depth,
					  bit_order, byte_order);
	    if (!ok)
		goto generalXaxe;
	    break;

	  generalXaxe:
	  default:
	    ok = _mirorXaxe(result, src, pixel_translation,
			    bits_per_pixel, depth,
			    bit_order, byte_order);
	    break;
	}
	if (ok) {
	    *xOrigin_result = - cs_I_x;
	    return;
	}
	goto _GENERAL_DEFORMATION;
    }

    if (cs_I_y == 0 && cs_I_x == src->values()->width &&
	cs_J_x == 0 && cs_J_y == - src->values()->height) {
	/**
	  Y miror
	  **/
	GBool ok ;
	switch (src->values()->bits_per_pixel) {
	  case 1:
	    ok = _mirorYaxe_bitsPerPixel1(result, src,
					  bits_per_pixel, depth,
					  bit_order, byte_order);
	    if (!ok)
		goto generalYaxe;
	    break;
	  case 4:
	    ok = _mirorYaxe_bitsPerPixel4(result, src, pixel_translation,
					  bits_per_pixel, depth,
					  bit_order, byte_order);
	    if (!ok)
		goto generalYaxe;
	    break;
	  case 8:
	    ok = _mirorYaxe_bitsPerPixel8(result, src, pixel_translation,
					  bits_per_pixel, depth,
					  bit_order, byte_order);
	    if (!ok)
		goto generalYaxe;
	    break;
	    
	  generalYaxe:
	  default:
	    ok = _mirorYaxe(result, src, pixel_translation,
			    bits_per_pixel, depth,
			    bit_order, byte_order);
	    break;
	}
	if (ok) {
	    *yOrigin_result = - cs_I_y;
	    return;
	}
	goto _GENERAL_DEFORMATION;
    }

    if (cs_I_x == 0 && cs_I_y == src->values()->width &&
	cs_J_y == 0 && cs_J_x ==  - src->values()->height) {
	/**
	  rotation 90 degrees
	  **/
	GBool ok ;
	switch (src->values()->bits_per_pixel) {
	  case 1:
	    ok = _rotation90_bitsPerPixel1(result, src,
					   bits_per_pixel, depth,
					   bit_order, byte_order);
	    if (!ok)
		goto general_Rotation90;
	    break;
	  case 4:
	    ok = _rotation90_bitsPerPixel4(result, src, pixel_translation,
					   bits_per_pixel, depth,
					   bit_order, byte_order);
	    if (!ok)
		goto general_Rotation90;
	    break;
	  case 8:
	    ok = _rotation90_bitsPerPixel8(result, src, pixel_translation,
					   bits_per_pixel, depth,
					   bit_order, byte_order);
	    if (!ok)
		goto general_Rotation90;
	    break;
	  general_Rotation90:
	  default:
	    ok = _rotation90(result, src, pixel_translation,
			    bits_per_pixel, depth,
			    bit_order, byte_order);
	    break;
	}

	if (ok) {
	    *xOrigin_result = - cs_I_y;
	    *yOrigin_result = - cs_J_x;
	    return;
	}
	goto _GENERAL_DEFORMATION;
    }

    if (cs_I_y == 0 && cs_I_x == - src->values()->width &&
	cs_J_x == 0 && cs_J_y == - src->values()->height) {
	/**
	  rotation 180 degrees
	  **/
	GBool ok ;
	switch (src->values()->bits_per_pixel) {
	  case 1:
	    ok = _rotation180_bitsPerPixel1(result, src,
					    bits_per_pixel, depth,
					    bit_order, byte_order);
	    if (!ok)
		goto general_Rotation180;
	    break;
	  case 4:
	    ok = _rotation180_bitsPerPixel4(result, src, pixel_translation,
					    bits_per_pixel, depth,
					    bit_order, byte_order);
	    if (!ok)
		goto general_Rotation180;
	    break;
	  case 8:
	    ok = _rotation180_bitsPerPixel8(result, src, pixel_translation,
					    bits_per_pixel, depth,
					    bit_order, byte_order);
	    if (!ok)
		goto general_Rotation180;
	    break;
	    
          general_Rotation180:
	  default:
	    ok = _rotation180(result, src, pixel_translation,
			      bits_per_pixel, depth,
			      bit_order, byte_order);
	    break;
	}

	if (ok) {
	    *xOrigin_result = - cs_I_x;
	    *xOrigin_result = - cs_J_y;
	    return;
	}
	goto _GENERAL_DEFORMATION;
    }




    _GENERAL_DEFORMATION:
    /*******************/
    _generalImageDeformation(result, xOrigin_result, yOrigin_result,
			     src,
			     cs,
			     pixel_translation,
			     bits_per_pixel, depth,
			     bit_order, byte_order, quality);

}


/**
FUNCTION
    _generalImageDeformation().
DESCRIPTION
    the function is called by _transformPixelArray when any other
    virtual functions are unable to implemente the image deformations.
    Use the tool.
**/
void GoInterf::_generalImageDeformation(GPixelArrayValues* result,
					ushort* xOrigin_result,
					ushort* yOrigin_result,
					const GShPixelArray* src,
					const GCoordSys&   cs,
					const GArray*      pixel_translation,
					uchar bits_per_pixel, uchar depth,
					uchar bit_order,      uchar byte_order,
					uchar quality) const
{
    src->deformation(result, xOrigin_result, yOrigin_result,
		     cs,
		     pixel_translation,
		     bits_per_pixel, depth,
		     bit_order, byte_order, quality);
}

/**
FUNCTION
    GBool _resizeImage().
DESCRIPTION
    the function is a virtual one. If it returns GFALSE
    the GENERAL_DEFORMATION is used by "transformPixelArray"
    An application that think the _resizeImage function which is in TOOLS
    is ok has to make an INTERFACE which defines the "_resizeImage()"
    with the TOOL function or another one.
**/
GBool GoInterf::_resizeImage_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					  int      /* wanted_width */,
					  int      /* wanted_height */,
					  uchar    /* bits_per_pixel */,
					  uchar    /* depth */,
					  uchar    /* bit_order */,
					  uchar    /* byte_order */,
					  GQuality  /* quality */) const
{
    return GFALSE;
}
GBool GoInterf::_resizeImage_bitsPerPixel4(GPixelArrayValues*, const GShPixelArray*,
					  const GArray* /* pixels_translation */,
					  int          /* wanted_width */,
					  int          /* wanted_height */,
					  uchar        /* bits_per_pixel */,
					  uchar        /* depth */,
					  uchar        /* bit_order */,
					  uchar        /* byte_order */,
					  GQuality      /* quality */) const
{
    return GFALSE;
}
GBool GoInterf::_resizeImage_bitsPerPixel8(GPixelArrayValues*, const GShPixelArray*,
					  const GArray* /* pixels_translation */,
					  int          /* wanted_width */,
					  int          /* wanted_height */,
					  uchar        /* bits_per_pixel */,
					  uchar        /* depth */,
					  uchar        /* bit_order */,
					  uchar        /* byte_order */,
					  GQuality      /* quality */) const
{
    return GFALSE;
}

GBool GoInterf::_resizeImage(GPixelArrayValues*, const GShPixelArray*,
			    const GArray* /* pixels_translation */,
			    int          /* wanted_width */,
			    int          /* wanted_height */,
			    uchar        /* bits_per_pixel */,
			    uchar        /* depth */,
			    uchar        /* bit_order */,
			    uchar        /* byte_order */,
			    GQuality      /* quality */) const
{
    return GFALSE;
}


/**
FUNCTION
    GBool _rotation180().
DESCRIPTION
    the function is a virtual one. If it returns GFALSE
    the GENERAL_DEFORMATION is used by "transformPixelArray"
    An application that think the _rotation180 function which is in TOOLS
    is ok has to make an INTERFACE which defines the "_rotation180()"
    with the TOOL function or another one.
**/
GBool GoInterf::_rotation180_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					  uchar      /* bits_per_pixel */,
					  uchar      /* depth */,
					  uchar      /* bit_order */,
					  uchar      /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_rotation180_bitsPerPixel4(GPixelArrayValues*, const GShPixelArray*,
					  const GArray* /* pixels_translation */,
					  uchar        /* bits_per_pixel */,
					  uchar        /* depth */,
					  uchar        /* bit_order */,
					  uchar        /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_rotation180_bitsPerPixel8(GPixelArrayValues*, const GShPixelArray*,
					  const GArray* /* pixels_translation */,
					  uchar        /* bits_per_pixel */,
					  uchar        /* depth */,
					  uchar        /* bit_order */,
					  uchar        /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_rotation180(GPixelArrayValues*, const GShPixelArray*,
			    const GArray* /* pixels_translation */,
			    uchar        /* bits_per_pixel */,
			    uchar        /* depth */,
			    uchar        /* bit_order */,
			    uchar        /* byte_order */) const
{
    return GFALSE;
}

/**
FUNCTION
    GBool _rotation90().
DESCRIPTION
    the function is a virtual one. If it returns GFALSE
    the GENERAL_DEFORMATION is used by "transformPixelArray"
    An application that think the _rotation90 function which is in TOOLS
    is ok has to make an INTERFACE which defines the "_rotation90()"
    with the TOOL function or another one.
**/
GBool GoInterf::_rotation90_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					 uchar /* bits_per_pixel */, uchar /* depth */,
					 uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_rotation90_bitsPerPixel4(GPixelArrayValues*, const GShPixelArray*,
					 const GArray* /* pixels_translation */,
					 uchar /* bits_per_pixel */, uchar /* depth */,
					 uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_rotation90_bitsPerPixel8(GPixelArrayValues*, const GShPixelArray*,
					 const GArray* /* pixels_translation */,
					 uchar /* bits_per_pixel */, uchar /* depth */,
					 uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_rotation90(GPixelArrayValues*, const GShPixelArray*,
			  const GArray* /* pixels_translation */,
			  uchar /* bits_per_pixel */, uchar /* depth */,
			  uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}

/**
FUNCTION
    GBool _mirorYaxe().
DESCRIPTION
    the function is a virtual one. If it returns GFALSE
    the GENERAL_DEFORMATION is used by "transformPixelArray"
    An application that think the _mirotYaxe function which is in TOOLS
    is ok has to make an INTERFACE which defines the "_mirorYaxe()"
    with the TOOL function or another one.
**/
GBool GoInterf::_mirorYaxe_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					uchar /* bits_per_pixel */, uchar /* depth */,
					uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_mirorYaxe_bitsPerPixel4(GPixelArrayValues*, const GShPixelArray*,
					const GArray* /* pixels_translation */,
					uchar /* bits_per_pixel */, uchar /* depth */,
					uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_mirorYaxe_bitsPerPixel8(GPixelArrayValues*, const GShPixelArray*,
					const GArray* /* pixels_translation */,
					uchar /* bits_per_pixel */, uchar /* depth */,
					uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_mirorYaxe(GPixelArrayValues*, const GShPixelArray*,
			  const GArray* /* pixels_translation */,
			  uchar /* bits_per_pixel */, uchar /* depth */,
			  uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}


/**
FUNCTION
    GBool _mirorXaxe().
DESCRIPTION
    the function is a virtual one. If it returns GFALSE
    the GENERAL_DEFORMATION is used by "transformPixelArray"
    An application that think the _mirorXaxe function which is in TOOLS
    is ok has to make an INTERFACE which defines the "_mirorXaxe()"
    with the TOOL function or another one.
**/
GBool GoInterf::_mirorXaxe_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					uchar /* bits_per_pixel */, uchar /* depth */,
					uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_mirorXaxe_bitsPerPixel4(GPixelArrayValues*, const GShPixelArray*,
					const GArray* /* pixels_translation */,
					uchar /* bits_per_pixel */, uchar /* depth */,
					uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_mirorXaxe_bitsPerPixel8(GPixelArrayValues*, const GShPixelArray*,
					const GArray* /* pixels_translation */,
					uchar /* bits_per_pixel */, uchar /* depth */,
					uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}
GBool GoInterf::_mirorXaxe(GPixelArrayValues*, const GShPixelArray*,
			  const GArray* /* pixels_translation */,
			  uchar /* bits_per_pixel */, uchar /* depth */,
			  uchar /* bit_order */, uchar /* byte_order */) const
{
    return GFALSE;
}



/**
FUNCTION
    GBool _extendImage_E234_bitsPerPixel1().
    GBool _reduceImage_R234_bitsPerPixel1().
DESCRIPTION
    the function is a virtual one. If it returns GFALSE
    the _resizeImage_bitsPerPixel1() is used by "transformPixelArray"
    An application that think another function is ok, has to make an INTERFACE
    which defines the virtual function
**/
GBool GoInterf::_reduceImage_R2_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					     int /* wanted_width */,
					     int /* wanted_height */,
					     GQuality /* quality */) const
{
    return GFALSE;
}
GBool GoInterf::_reduceImage_R3_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					     int /* wanted_width */,
					     int /* wanted_height */,
					     GQuality /* quality */) const
{
    return GFALSE;
}
GBool GoInterf::_reduceImage_R4_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					     int /* wanted_width */,
					     int /* wanted_height */,
					     GQuality /* quality */) const
{
    return GFALSE;
}

GBool GoInterf::_extendImage_E2_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					     int /* wanted_width */,
					     int /* wanted_height */,
					     GQuality /* quality */) const
{
    return GFALSE;
}
GBool GoInterf::_extendImage_E3_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					     int /* wanted_width */,
					     int /* wanted_height */,
					     GQuality /* quality */) const
{
    return GFALSE;
}
GBool GoInterf::_extendImage_E4_bitsPerPixel1(GPixelArrayValues*, const GShPixelArray*,
					     int /* wanted_width */,
					     int /* wanted_height */,
					     GQuality /* quality */) const
{
    return GFALSE;
}



