#ifdef REV_INFO
static char rcsid[] = "$Id: GoInterf.cc,v 1.10 1992/12/21 09:33:13 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Go Interface
FILE:    GoInterf.cc
AUTHOR:  Olivier Jojic, Philippe Smadja, Jacques Davy, Gilles Benati
DATE:    November 1990 (Version 1.0)

DESCRIPTION:
    See <Go/GoInterf.h>

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/GoInterf.h>


#include <Go/Tools/Cons.h>
#include <Go/Tools/GeomUtil.h>
#include <Go/Tools/Alloc.h>

/*
 * GCallback types
 */
static char gointerfNexposePostcond[1];
static char gointerfNeventPostcond[1];
static char gointerfNdestroyPostcond[1];
static char gointerfNresizePostcond[1];

/*
 * Protected Methods
 * *****************
 */
/**/
GoInterf::GoInterf ()
{
    for (int i = 0; i < GN_EVENT_MASK; i++)
	_eventMaskCount[i] = 0;
    _eventMaskCount[GEXPOSE_MASK_INDEX] = 1;
    _background = NULL;
    _clipCount = 0;
}

/**/
int GoInterf::_addCallback (const char* type, GCBProc proc,
			    const void* closure) const
{
    const GAssoc* _cb = &_callbacks;
    GAssoc* cb = (GAssoc*)_cb;
    
    GCallbackRec cbrec (proc, closure);
    return cb->add(type, sizeof cbrec, (char*)&cbrec);
}

/**/
int GoInterf::_remCallback (const char* type, GCBProc proc,
			    const void* closure) const
{
    const GAssoc* _cb = &_callbacks;
    GAssoc* cb = (GAssoc*)_cb;

    GCallbackRec cbrec (proc, closure);
    return cb->rem(type, sizeof cbrec, (char*)&cbrec);
}

/**/
void GoInterf::_call_expose_postcond (const GExposeEvent* event)
{
    if (_cond._has_expose_postcond)
    {
        int count;
        GCallbackRec* cbrec = (GCallbackRec*)_callbacks.get(gointerfNexposePostcond, count);
	GCallCallbacks(cbrec, count, this, (void*)event);
    }
}

/**/
void GoInterf::_call_event_postcond (const GEvent* event)
{
    if (_cond._has_event_postcond)
    {
        int count;
        GCallbackRec* cbrec = (GCallbackRec*)_callbacks.get (gointerfNeventPostcond, count);
	GCallCallbacks(cbrec, count, this, (void*)event);
    }
}

/**/
void GoInterf::_call_destroy_postcond ()
{
    if (_cond._has_destroy_postcond)
    {
        int count;
        GCallbackRec* cbrec = (GCallbackRec*)_callbacks.get (gointerfNdestroyPostcond, count);
	GCallCallbacks(cbrec, count, this);
    }
}

/**/
void GoInterf::_call_resize_postcond ()
{
    if (_cond._has_resize_postcond)
    {
        int count;
        GCallbackRec* cbrec = (GCallbackRec*)_callbacks.get (gointerfNresizePostcond, count);
	GCallCallbacks(cbrec, count, this);
    }
}


/**/
void GoInterf::_handle(const GEvent* event)
{
    if (event->type() == GEXPOSE_EVENT)
        _call_expose_postcond ((const GExposeEvent*)event);
    _call_event_postcond (event);
}

void GoInterf::handle(GEvent* gevent)
{
    _handle(gevent);
}

/*
 * Public Methods
 * **************
 */
GoInterf::~GoInterf ()
{
    _call_destroy_postcond ();
    delete _background;
}

/**/
void GoInterf::map () {}
void GoInterf::unmap () {}
void GoInterf::flush () {}
void GoInterf::synchronize (GBool /* sync not used */) {}

    
/**
FUNCTION
  Output Fonctions
DESCRIPTION
  Each function allows to draw something on the output taking
  care of the coordinate system, the pen/brush, the quality.
  **/

/* lines */
void GoInterf::line (const GShPen* shpen, const GCoordSys& cs,
		     const GPoint& p1, const GPoint& p2, GQuality quality)
{
    GPoint pts[2];
    pts[0] = p1;
    pts[1] = p2;
    polyline (shpen, NULL, cs, 2, pts, quality);
}


/* rectangles */
void GoInterf::rectangle (const GShPen* shpen, const GShBrush* shbrush,
			  const GCoordSys& cs, const GPoint& p,
			  GNumber w, GNumber h, GQuality quality)
{
    // GSet point list:
    GPoint pts[4];
    pts[0] = p;
    pts[1].x = p.x + w; 	pts[1].y = p.y;
    pts[2].x = pts[1].x; 	pts[2].y = p.y + h;
    pts[3].x = p.x;		pts[3].y = pts[2].y;
    polygon(shpen, shbrush, cs, 4, pts, quality);
}

/* Ellipses */
void GoInterf::ellipse (const GShPen* shpen, const GShBrush* shbrush,
			const GCoordSys& cs, const GPoint& origin,
			GNumber radiusX, GNumber radiusY,
			GNumber angle1, GNumber angle2, GQuality quality)
{
    int npoint;
    GPoint pts[GMAX_ELLIPSE_POINTS];

    GCoordSys CS = cs;
    CS.O = cs.to(origin);
    GEllipseComputePoints (CS,  radiusX, radiusY, angle1, angle2,
			  quality, npoint, pts);
    polyline (shpen, shbrush, CS, npoint, pts, quality); 
}

void GoInterf::ellipseChord (const GShPen* shpen, const GShBrush* shbrush,
			     const GCoordSys& cs, const GPoint& origin,
			     GNumber radiusX, GNumber radiusY,
			     GNumber angle1, GNumber angle2, GBool closed,
			     GQuality quality)
{
    int npoint;
    GPoint pts[GMAX_ELLIPSE_POINTS];

    GCoordSys CS = cs;
    CS.O = cs.to(origin);
    GEllipseComputePoints (CS, radiusX, radiusY, angle1, angle2,
			  quality, npoint, pts);
    if (closed) {
	// fill_chord and draw arc+chord;
	polygon (shpen, shbrush, CS, npoint, pts, quality);
    }
    else {
	// fill_chord and draw arc;
	polyline(shpen, shbrush, CS, npoint, pts, quality);
    }
}

void GoInterf::ellipsePie (const GShPen* shpen, const GShBrush* shbrush,
			   const GCoordSys& cs, const GPoint& origin,
			   GNumber radiusX, GNumber radiusY,
			   GNumber angle1, GNumber angle2, GBool closed,
			   GQuality quality)
{
    if (angle2 - angle1 >= 2*PI)
	ellipse(shpen, shbrush, cs,  origin, radiusX, radiusY,
		angle1, angle2, quality);
    else {
	/**
	  add the first point and the center to the end of the list
	  **/
	int npoint;
	GPoint pts[GMAX_ELLIPSE_POINTS+1];

	GCoordSys CS = cs;
	CS.O = cs.to(origin);
	GEllipseComputePoints (CS,  radiusX, radiusY, angle1, angle2,
			      quality, npoint, pts);
	pts[npoint] = GPoint(0,0);

	if (closed) {
	    // fill_pie and draw arc+pie
	    polygon (shpen, shbrush, CS, npoint+1, pts, quality);
	}
	else {
	    // fill_pie and draw arc;
	    if (shbrush)
		polygon(NULL, shbrush, CS, npoint+1, pts, quality);
	    if (shpen)
		polyline(shpen, NULL, CS, npoint, pts, quality);
	}
    }
}


/* circles */
void GoInterf::circle (const GShPen* shpen, const GShBrush* shbrush,
		       const GCoordSys& cs, const GPoint& origin,
		       GNumber radius,
		       GNumber angle1, GNumber angle2, GQuality quality)
{
    ellipse(shpen, shbrush, cs,  origin, radius, radius,
	    angle1, angle2, quality);
}

void GoInterf::circleChord (const GShPen* shpen, const GShBrush* shbrush,
			    const GCoordSys& cs, const GPoint& origin,
			    GNumber radius, GNumber angle1, GNumber angle2,
			    GBool closed, GQuality quality)
{
    ellipseChord (shpen, shbrush, cs, origin,
		  radius,  radius, angle1, angle2, closed, quality);
}

void GoInterf::circlePie (const GShPen* shpen, const GShBrush* shbrush,
			  const GCoordSys& cs, const GPoint& origin,
			  GNumber radius, GNumber angle1, GNumber angle2,
			  GBool closed, GQuality quality)
{
    ellipsePie (shpen, shbrush, cs, origin,
		radius, radius, angle1,	angle2, closed, quality);
}

/* bezier curves */
void GoInterf::bezier (const GShPen* shpen, const GShBrush* shbrush,
		       const GCoordSys& cs,
		       int nctrl, const GPoint* ctrl, GQuality quality)
{
    int npoint;
#ifdef NO_ALLOCA
    GPoint* pts = (GPoint*)GMALLOC(GMAX_BEZIER_POINTS(nctrl)* sizeof(GPoint));
#else
    GPoint* pts = (GPoint*)alloca(GMAX_BEZIER_POINTS(nctrl)* sizeof(GPoint));
#endif

    GBezierComputePoints (nctrl, ctrl, quality, npoint, pts);

    polyline (shpen, shbrush, cs, npoint, pts, quality);
    
#ifdef NO_ALLOCA
    GFREE (pts);
#endif
}

/* cubic curves */
void GoInterf::cubicCurve (const GShPen* shpen, const GShBrush* shbrush,
			   const GCoordSys& cs,
			   int nctrl, const GPoint* ctrl,
			   const GNumber (*basisMatrix)[4] /* [4][4] */,
			   int evalShift, GQuality quality)
{
    int npoint;
#ifdef NO_ALLOCA
    GPoint* pts = (GPoint*)GMALLOC(GMAX_CUBIC_POINTS(nctrl, evalShift) *
				   sizeof(GPoint));
#else
    GPoint* pts = (GPoint*)alloca(GMAX_CUBIC_POINTS(nctrl, evalShift) *
				  sizeof(GPoint));
#endif
 
    GCubicCurveComputePoints (nctrl, ctrl, basisMatrix, evalShift,
			      quality, npoint, pts);
    
    polyline(shpen, shbrush, cs, npoint, pts, quality);
    
#ifdef NO_ALLOCA
    GFREE (pts);
#endif
}


/**
FUNCTION
    Clips management
DESCRIPTION
    compute the best clip for all the drawing methods
**/
/**
  base functions for clips management
  **/
static void _PolylineClips(uint line_width, GBool joinmiter,
			   GBool /* filled */, GBool closed,
			   const GCoordSys& cs,
			   int npoint, const GPoint* points,
			   GRegion* clip, GRegion* /* opaque_clip */,
			   GNumber pw, GNumber ph)
{
    if (clip) {
	
	// maximise the clip;
	int up_left_width = line_width / 2;
	int down_right_width = (line_width + 1)/ 2;

	if (down_right_width != 0 && npoint > 2 && joinmiter)
	{
	    // do not compute the effective clip (with pixel);
	    // just major the theoric one;
	    
	    *clip += GPolypointJoinMiterBox(cs, npoint, points,
					    max(pw,ph) * (line_width + 1),
					    closed);
	}
	else
	{
	    GPoint pmin, pmax;
	    
	    GPolypointGetAbsOutPoints(cs, npoint, points, pmin, pmax);
	    
	    pmin.x = (GPositionToPixel(pmin.x, pw) - up_left_width) * pw;
	    pmax.x = (GPositionToPixel(pmax.x, pw) + down_right_width) * pw;

	    pmin.y = (GPositionToPixel(pmin.y, ph) - up_left_width) * ph;
	    pmax.y = (GPositionToPixel(pmax.y, ph) + down_right_width) * ph;

	    *clip += GBox(pmin.x, pmin.y,  pmax.x, pmax.y);
	}
    }
}
    
		     
void GoInterf::polylineClips (const GShPen* shpen, const GShBrush* shbrush,
			      const GCoordSys& cs,
			      int npoint, const GPoint* points,
			      GRegion* clip, GRegion*  opaque_clip,
			      GQuality )
{
    uint i_lw = shpen ? (uint)mm_to_linew(shpen->lineWidth()) : 0;

    _PolylineClips(i_lw,
		   shpen ? (GBool)(shpen->joinStyle() == GJOIN_MITER): GFALSE,
		   (GBool)(shbrush && shbrush->fillStyle() != GFILL_STIPPLED),
		   GFALSE, cs, npoint, points, clip, opaque_clip,
		   pixelWidth(), pixelHeight());

}



void GoInterf::polygonClips (const GShPen* shpen, const GShBrush* shbrush,
			     const GCoordSys& cs,
			     int npoint, const GPoint* points,
			     GRegion* clip,
			     GRegion* opaque_clip,
			     GQuality quality)
{
    if (clip && shpen && npoint > 2 && shpen->joinStyle() == GJOIN_MITER)
    {
	uint i_lw = (uint)mm_to_linew(shpen->lineWidth());
	
	*clip += GPolypointJoinMiterBox(cs, npoint, points,
					(i_lw+1)/2*2 * max(pixelWidth(),pixelHeight()),
					GTRUE);
	// opaque_clip is not handled yet;
    }
    else
	polylineClips(shpen, shbrush, cs, npoint, points, clip, opaque_clip,
		      quality);
}

void GoInterf::rectangleClips (const GShPen* shpen, const GShBrush* shbrush,
			       const GCoordSys& cs, const GPoint& p,
			       GNumber w, GNumber h,
			       GRegion* clip, GRegion* opaque_clip,
			       GQuality quality)
{
    if (!shbrush && !shpen)
	return;
    if (!clip && !opaque_clip)
	return;
    
    if (cs.isOrthogonal())
    {
	GPoint a = cs.to(p);
	GPoint b = cs.to(GPoint(p.x + w, p.y + h));
	GBool filled = (GBool)(shbrush != NULL);
	GPoint p1;
	GPoint p2;
	GNumber pw = pixelWidth();
	GNumber ph = pixelHeight();
	
	p1.x = GPositionToPixel(min(a.x, b.x), pw) * pw;
	p2.x = GPositionToPixel(max(a.x, b.x), pw) * pw;
	p1.y = GPositionToPixel(min(a.y, b.y), ph) * ph;
	p2.y = GPositionToPixel(max(a.y, b.y), ph) * ph;

	if (shpen == NULL)
	{
	    if (clip)
		GRectangleClip(p1, p2, filled, 0, *clip);
	    if (opaque_clip && filled && shbrush->fillStyle() != GFILL_STIPPLED)
		GRectangleClip(p1, p2, GTRUE, 0, *opaque_clip);
	}
	else
	{
	    // shpen is available:
	    // for clip it is important to use the effective line_width;
	    
	    uint  i_lw = (uint)mm_to_linew(shpen->lineWidth());
	    if (i_lw & 1)
	    {
		// non symetric area;
		p2.x += pw;
		p2.y += ph;
	    }
	    
	    GNumber lw_div_2 = (i_lw/2) * pw;
	    GNumber lh_div_2 = (i_lw/2) * ph;
	    if (filled)
	    {
		GPoint delta(lw_div_2, lh_div_2);
		if (clip)
		    GRectangleClip(p1-delta, p2+delta, GTRUE, 0, *clip);
		
		
		if (opaque_clip && shbrush->fillStyle() != GFILL_STIPPLED)
		{
		    // verify that every pixel will be effectively drawn;

		    if (shpen->lineStyle() != GLINE_ON_OFF_DASH &&
			shpen->joinStyle() == GJOIN_MITER)
		    {
			// the line is a full one;
			GRectangleClip(p1-delta, p2+delta, GTRUE,
				       0, *opaque_clip);
		    }
		    else 
		    {
			// the line is NOT a full one;
			p1 += delta;
			p2 -= delta;
			if (p1.x < p2.x && p1.y < p2.y)
			    GRectangleClip(p1, p2, GTRUE,
					   0, *opaque_clip);
		    }
		}
	    }
	    else  // compute only lines clip
	    {
		GNumber xmin = p1.x - lw_div_2;
		GNumber ymin = p1.y - lh_div_2;
		GNumber xmax = p2.x + lw_div_2;
		GNumber ymax = p2.y + lh_div_2;

		GNumber lw = i_lw * pw;
		GNumber lh = i_lw * ph;

		if (clip)
		{
		    *clip += GBox(xmin, ymin, xmax, ymin + lh);
		    *clip += GBox(xmin, ymin, xmin + lw, ymax);
		    *clip += GBox(xmin, ymax - lh, xmax, ymax);
		    *clip += GBox(xmax - lw, ymin, xmax, ymax);
		}
		
		/*********************************************
		  for performance reason
		  without background
		  do not add the lines opaque_clip
		  -------------------------------
		  

		  if (opaque_clip && shpen->lineStyle() != GLINE_ON_OFF_DASH)
		  {
	               *opaque_clip += GBox(xmin, ymin, xmax, ymin + lh);
	               *opaque_clip += GBox(xmin, ymin, xmin + lw, ymax);
	               *opaque_clip += GBox(xmin, ymax - lh, xmax, ymax);
	               *opaque_clip += GBox(xmax - lw, ymin, xmax, ymax);

		  }
		  *******************************************/
	    }
	}
    }
    else
    {
	// not orthogonal cs:
	// GSet point list:
	GPoint pts[4];
	pts[0] = p;
	pts[1].x = p.x + w; 	pts[1].y = p.y;
	pts[2].x = pts[1].x; 	pts[2].y = p.y + h;
	pts[3].x = p.x;		pts[3].y = pts[2].y;

	polygonClips(shpen, shbrush, cs, 4, pts, clip, opaque_clip, quality);
    }
}

/**
FUNCTION
    static filled_brush()
DESCRIPTION
    return  a filled bush (SOLID) used  bt stringClips and stringContains
    to simulated a filled rectangle.
**/
static const GShBrush* filled_brush()
{
    static const GShBrush* p = NULL;
    if (p == NULL) 
	p = GCreateShBrush();
    return p;
}

void GoInterf::stringClips    (const GShTextBrush* shtextbrush, const GShBrush* ,
			       const GCoordSys& cs, const GPoint& origin,
			       GNumber height, int size, const char* string,
			       GRegion* clip, GRegion* opaque_clip,
			       GQuality quality)
{
    if (size == 0)
	return;
    
    const GShFont* shfont = shtextbrush->font();

    GNumber string_width = stringRightBearing(string, size, height, shfont);

    GNumber lbearing = stringLeftBearing(string, size, height, shfont);
    string_width += lbearing;
    
    GNumber string_height = stringAscent(string, size, height, shfont);
    string_height += stringDescent(string, size, height, shfont);

    GPoint  position = origin;
    position.x -= lbearing;
    
    rectangleClips(NULL, filled_brush(), cs, position,
		   string_width, string_height,
		   clip, opaque_clip, quality);
}

void GoInterf::imageClips     (const GShImageBrush* shibrush,
			       const GCoordSys& cs, const GPoint& p,
			       GNumber width, GNumber height,
			       GRegion* clip, GRegion* opaque_clip,
			       GQuality  /* quality */)
{
    if (!shibrush)
	return;

    if (cs.isOrthogonal())
    {
	GPoint p1 = cs.to(p);
	GPoint p2 = cs.to(GPoint(p.x + width, p.y + height));

	GNumber pw = pixelWidth();
	GNumber ph = pixelHeight();
	p1.x = GPositionToPixel(p1.x, pw) * pw;
	p2.x = GPositionToPixel(p2.x, pw) * pw;

	p1.y = GPositionToPixel(p1.y, ph) * ph;
	p2.y = GPositionToPixel(p2.y, ph) * ph;
	
	/**
	  do not forget to handle the outline
	  when it will be available in the imagebrush.
	  **/
	
	if (clip)
	    GRectangleClip(p1, p2, GTRUE, 0, *clip);

	if (opaque_clip && shibrush->fillStyle() != GFILL_STIPPLED)
	    GRectangleClip(p1, p2, GTRUE, 0, *opaque_clip);

    }
    else
    {
	// GSet point list:
	GPoint pts[4];
	pts[0] = p;
	pts[1].x = p.x + width;	pts[1].y = p.y;
	pts[2].x = pts[1].x; 	pts[2].y = p.y + height;
	pts[3].x = p.x;		pts[3].y = pts[2].y;
	
	_PolylineClips (0, GFALSE,
			(GBool)(shibrush->fillStyle() != GFILL_STIPPLED),GTRUE,
			cs, 4, pts, clip, opaque_clip,
			pixelWidth(), pixelHeight());
    }
}

void GoInterf::ellipseClips (const GShPen* shpen, const GShBrush* shbrush,
			     const GCoordSys& cs, const GPoint& p,
			     GNumber radiusX, GNumber radiusY,
			     GRegion* clip,
			     GRegion* /* opaque_clip not handled yet */,
			     GNumber angle1, GNumber angle2,
			     GQuality /* quality not used */)
{
    
    if (clip && (shbrush || shpen)) 
    {
	if (angle1 > angle2 )
	{
	    GNumber tmp = angle2;
	    angle2 = angle1;
	    angle1 = tmp;
	}

	GCoordSys CS = cs;
	CS.O = cs.to(p);

	GBox absBox;
	if (CS.isOrthogonal())
	{
	    GPoint pmin, pmax;
	    GEllipseGetOutPoints(radiusX, radiusY, angle1, angle2, pmin, pmax);
	    pmin = CS.to(pmin);
	    pmax = CS.to(pmax);
	    absBox = GBox(min(pmin.x, pmax.x), min(pmin.y, pmax.y),
			  max(pmin.x, pmax.x), max(pmin.y, pmax.y));
	}
	else 
	    absBox = GEllipseGetBox(CS, radiusX, radiusY, angle1, angle2);

	// map the Number to pixel.
	    
	GNumber pw = pixelWidth();
	GNumber ph = pixelHeight();

	absBox.xmin = GPositionToPixel(absBox.xmin, pw) * pw;
	absBox.xmax = GPositionToPixel(absBox.xmax, pw) * pw;	
	absBox.ymin = GPositionToPixel(absBox.ymin, ph) * ph;
	absBox.ymax = GPositionToPixel(absBox.ymax, ph) * ph;	

	if (shpen)
	{
	    int i_lw = mm_to_linew(shpen->lineWidth());
	    
	    GNumber extension = (i_lw+1)/2 * max(pw, ph);
	    absBox.reset(absBox.xmin - extension, absBox.ymin - extension,
			 absBox.xmax + extension, absBox.ymax + extension);
	}
	
	*clip += absBox;
    }
}

/**
  the others function uses the base functions for clips management
  **/

/* lines */
void GoInterf::lineClips (const GShPen* shpen, const GCoordSys& cs,
			  const GPoint& p1, const GPoint& p2,
			  GRegion* clip, GRegion* opaque_clip,
			  GQuality quality)
{
    GPoint pts[2];
    pts[0] = p1;
    pts[1] = p2;
    polylineClips (shpen, NULL, cs, 2, pts, clip, opaque_clip, quality);
}



void GoInterf::ellipseChordClips (const GShPen* shpen, const GShBrush* shbrush,
				  const GCoordSys& cs,  const GPoint& origin,
				  GNumber radiusX, GNumber radiusY,
				  GNumber angle1, GNumber angle2, GBool closed,
				  GRegion* clip, GRegion* opaque_clip,
				  GQuality quality)
{
    ellipseClips(shpen , shbrush, cs, origin,  radiusX, radiusY,
		 clip, opaque_clip, angle1, angle2, quality);

    // add clip of the chord segment, if need be;
    if (shpen && closed && (angle2 - angle1 < 2*PI))
    {
	GCoordSys CS = cs;
	CS.O = cs.to(origin);
	GPoint pts[4];
	/*
	 * get the chord segment plus one point close to each extremity
	 * on the arc to take care of join style clip extensions.
	 * pts[0] and pts[3] will be these extra points, pts[1] and pts[2]
	 * will be the chord extremities.
	 * pts[0] and pts[3] will only be use if there is an arc outline.
	 */
	if (GEllipseArcLimits(CS, radiusX, radiusY, angle1, angle2,
			      pts[1], pts[2], pts[0], pts[3], quality))
	{
	    // we need the 2 extra points pts[0] and pts[3];
	    polylineClips(shpen, shbrush, CS, 4, pts,
			  clip, opaque_clip, quality);
	}
	else
	{
	    // we cannot use pts[0] and pts[3];
	    polylineClips(shpen, shbrush, CS, 2, pts+1,
			  clip, opaque_clip, quality);
	}
    }
}


void GoInterf::ellipsePieClips (const GShPen* shpen, const GShBrush* shbrush,
				const GCoordSys& cs, const GPoint& origin,
				GNumber radiusX, GNumber radiusY,
				GNumber angle1, GNumber angle2, GBool closed,
				GRegion* clip, GRegion* opaque_clip,
				GQuality quality)
{
    ellipseClips(shpen , shbrush, cs, origin,  radiusX, radiusY,
		 clip, opaque_clip, angle1, angle2, quality);

    // add clip of the triangle [chord_segment+center], if need be;
    if (((shpen && closed) || shbrush) && (angle2 - angle1 < 2*PI))
    {
	GCoordSys CS = cs;
	CS.O = cs.to(origin);
	GPoint pts[5];
	/*
	 * get the chord points plus one point close to each extremity on the
	 * arc to take care of join style clip extensions at arc limits,
	 * pts[0] and pts[4] will be these extra points, pts[1] and pts[3]
	 * will be the chord extremities and pts[2] the center of the ellipse.
	 * pts[0] and pts[4] will only be use if there is an arc outline.
	 */
	pts[2].x = pts[2].y = 0;
	int extra = GEllipseArcLimits(CS, radiusX, radiusY, angle1, angle2,
				      pts[1], pts[3], pts[0], pts[4], quality);
	if (shpen && closed)
	{
	    // there is an outline;
	    if (extra) {
		// use extra points;
		polylineClips(shpen, shbrush, CS, 5, pts,
			      clip, opaque_clip, quality);
	    }
	    else {
		polylineClips(shpen, shbrush, CS, 3, pts+1,
			      clip, opaque_clip, quality);
	    }
	}
	else polylineClips(NULL, shbrush, CS, 3, pts+1, clip,opaque_clip,quality);
    }
}


/* circles */
void GoInterf::circleClips (const GShPen* shpen, const GShBrush* shbrush,
			    const GCoordSys& cs, const GPoint& origin,
			    GNumber radius,
			    GRegion* clip, GRegion* opaque_clip,
			    GNumber angle1, GNumber angle2,
			    GQuality quality)
{
    ellipseClips(shpen, shbrush, cs, origin,
		 radius, radius,
		 clip, opaque_clip,
		 angle1, angle2, quality);

}

void GoInterf::circleChordClips (const GShPen* shpen, const GShBrush* shbrush,
				 const GCoordSys& cs, const GPoint& origin,
				 GNumber radius, GNumber angle1, GNumber angle2,
				 GBool closed, GRegion* clip, GRegion* opaque_clip,
				 GQuality quality)
{
    ellipseChordClips (shpen, shbrush, cs, origin, radius, radius,
		       angle1, angle2, 
		       closed, clip, opaque_clip, quality);

}

void GoInterf::circlePieClips (const GShPen* shpen, const GShBrush* shbrush,
			       const GCoordSys& cs,  const GPoint& origin,
			       GNumber radius, GNumber angle1, GNumber angle2,
			       GBool closed, GRegion* clip, GRegion* opaque_clip,
			       GQuality quality)
{
    ellipsePieClips (shpen, shbrush, cs, origin, radius, radius,
		     angle1, angle2,
		     closed, clip, opaque_clip, quality);
}

/* bezier curves */
void GoInterf::bezierClips (const GShPen* shpen, const GShBrush* shbrush,
			    const GCoordSys& cs, int nctrl, const GPoint* ctrl,
			    GRegion* clip, GRegion* opaque_clip, GQuality quality)
{
    int npoint;
#ifdef NO_ALLOCA
    GPoint* pts = (GPoint*)GMALLOC(GMAX_BEZIER_POINTS(nctrl)* sizeof(GPoint));
#else
    GPoint* pts = (GPoint*)alloca(GMAX_BEZIER_POINTS(nctrl)* sizeof(GPoint));
#endif

    GBezierComputePoints (nctrl, ctrl, quality, npoint, pts);
    polylineClips (shpen, shbrush, cs,  npoint, pts, clip, opaque_clip,
		   quality);
    
#ifdef NO_ALLOCA
    GFREE (pts);
#endif
}

/* cubic curves */
void GoInterf::cubicCurveClips (const GShPen* shpen, const GShBrush* shbrush,
				const GCoordSys& cs,
				int nctrl, const GPoint* ctrl,
				const GNumber (*basisMatrix)[4] /* [4][4] */,
				GRegion* clip, GRegion* opaque_clip,
				int evalShift, GQuality quality)
{
    int npoint;
#ifdef NO_ALLOCA
    GPoint* pts = (GPoint*)GMALLOC(GMAX_CUBIC_POINTS(nctrl, evalShift) *
				   sizeof(GPoint));
#else
    GPoint* pts = (GPoint*)alloca(GMAX_CUBIC_POINTS(nctrl, evalShift) *
				  sizeof(GPoint));
#endif
 
    GCubicCurveComputePoints (nctrl, ctrl, basisMatrix, evalShift,
			      quality, npoint, pts);

    polylineClips (shpen, shbrush, cs, npoint, pts, clip, opaque_clip,
		   quality);
    
#ifdef NO_ALLOCA
    GFREE (pts);
#endif
}

#define NOT_FILLED      0
#define FILLED_WINDING  1
#define FILLED_ODD_EVEN 2
/**
FUNCTION
    Contains management.
DESCRIPTION
    Compute if an absolute GPoint is inside or not the area drawn by the
    interface.
**/

/**
  base functions for contains management
  **/
static GBool _PolylineContains(int fill_rule, GBool closed,
			       const GCoordSys& cs,
			       int npoint, const GPoint* points,
			       const GPoint& absolute, GNumber maxDistance)
{
    if (fill_rule != NOT_FILLED && npoint > 2 &&
	GPointInPolygon(npoint, points, cs.from(absolute),
		       fill_rule == FILLED_WINDING))
	return GTRUE;

 
    return GPointOnPolyline(cs, npoint, points, absolute,
			    maxDistance*maxDistance, closed);
}

GBool GoInterf::polylineContains (const GShPen* shpen, const GShBrush* shbrush,
				  const GCoordSys& cs,
				  int npoint, const GPoint* points,
				  const GPoint& absolute, GNumber maxDistance,
				  GQuality quality)
{
    /**
      we consider that a fill with a stipple does not change the result
      **/
    return _PolylineContains((shbrush == NULL) ? NOT_FILLED :
			     ((shbrush->fillRule() == GWINDING_RULE) ?
			      FILLED_WINDING : FILLED_ODD_EVEN),
			     GFALSE, cs,
			     npoint, points, absolute,
			     quality != GMIN_QUALITY && shpen ?
			     maxDistance+shpen->lineWidth()/2 : maxDistance);
}
    

GBool GoInterf::polygonContains (const GShPen* shpen, const GShBrush* shbrush,
				const GCoordSys& cs,
				int npoint, const GPoint* points,
				const GPoint& absolute, GNumber maxDistance,
				GQuality quality)
{
    GBool ret = polylineContains(shpen, shbrush, cs, npoint, points,
				 absolute, maxDistance, quality);

    /**
      handle the last line
      **/
    if (ret || shpen == NULL || points[npoint - 1] == points[0])
	return ret;

    maxDistance += shpen->lineWidth() / 2;
    
    GPoint pts[2];
    pts[0] = points[0];
    pts[1] = points[npoint - 1];
    return GPointOnPolyline(cs, 2, pts, absolute, maxDistance*maxDistance,
			    GFALSE);

}

GBool GoInterf::rectangleContains (const GShPen* shpen,
				   const GShBrush* shbrush,
				   const GCoordSys& cs, const GPoint& p,
				   GNumber w, GNumber h,
				   const GPoint& absolute, GNumber maxDistance,
				   GQuality quality)
{
    if (cs.isOrthogonal()) {
	if (shpen)
	    maxDistance += shpen->lineWidth() / 2;

	GRegion clip;

	GPoint p1 = cs.to(p);
	GPoint p2 = cs.to(GPoint(p.x + w, p.y + h));

	GRectangleClip(p1, p2, (GBool)(shbrush != NULL), maxDistance, clip);
	return clip.contains(absolute);
    }

    // GSet point list:
    GPoint pts[4];
    pts[0] = p;
    pts[1].x = p.x + w; 	pts[1].y = p.y;
    pts[2].x = pts[1].x; 	pts[2].y = p.y + h;
    pts[3].x = p.x;		pts[3].y = pts[2].y;
    
    return polygonContains(shpen, shbrush, cs, 4, pts, absolute, maxDistance,
			   quality);
}



GBool GoInterf::stringContains    (const GShTextBrush* shtextbrush,  const GShBrush*,
				   const GCoordSys& cs, const GPoint& origin,
				   GNumber height, int size, const char* string,
				   const GPoint& absolute, GNumber maxDistance,
				   GQuality quality)
{
    if (size == 0)
	return GFALSE;
    
    const GShFont* shfont = shtextbrush->font();
    
    GNumber string_width = stringRightBearing(string, size, height, shfont);

    GNumber lbearing = stringLeftBearing(string, size, height, shfont);
    string_width += lbearing;

    GNumber string_height = stringAscent(string, size, height, shfont);
    string_height += stringDescent(string, size, height, shfont);

    GPoint  position = origin;
    position.x -= lbearing;

    return rectangleContains(NULL, filled_brush(), cs, position, string_width,
			     string_height,
			     absolute, maxDistance, quality);
}

GBool GoInterf::imageContains (const GShImageBrush* shibrush,
			       const GCoordSys& cs,
			       const GPoint& p, GNumber width, GNumber height,
			       const GPoint& absolute, GNumber maxDistance,
			       GQuality /* quality not used */)
{
    if (!shibrush)
	return GFALSE;

    /**
      do not forget to handle the outline
      when it will be available in the imagebrush.
      **/

    if (cs.isOrthogonal()) {
	GRegion clip;

	GPoint p1 = cs.to(p);
	GPoint p2 = cs.to(GPoint(p.x + width, p.y + height));

	GRectangleClip(p1, p2, GTRUE, maxDistance, clip);
	return clip.contains(absolute);
    }
	
    // GSet point list:
    GPoint pts[4];
    pts[0] = p;
    pts[1].x = p.x + width;	pts[1].y = p.y;
    pts[2].x = pts[1].x; 	pts[2].y = p.y + height;
    pts[3].x = p.x;		pts[3].y = pts[2].y;
	
    return _PolylineContains (FILLED_ODD_EVEN, GTRUE, cs, 4, pts,
			      absolute, maxDistance);

}

GBool GoInterf::ellipseContains(const GShPen* shpen, const GShBrush* shbrush,
				const GCoordSys& cs, const GPoint& origin,
				GNumber radiusX, GNumber radiusY,
				const GPoint& absolute, GNumber maxDistance,
				GNumber angle1, GNumber angle2,
				GQuality quality) {
    if (!(shpen || shbrush))
     return GFALSE;

    if (shpen && quality != GMIN_QUALITY)
	maxDistance += shpen->lineWidth() / 2;

    GCoordSys CS = cs;
    CS.O = cs.to(origin);

    GCoordSys inversed_cs = ~CS; // do not compute the determinent twice;
    GPoint pt = inversed_cs.to(absolute);

    /*
     * convert pt in the coordinate system where ellipse is a circle;
     * and get the angle associated with the resulting point;
     * then, test if that angle is within limits;
     */
    GPoint abstract_pt(pt.x / radiusX, pt.y / radiusY);
    GNumber abstract_angle = abstract_pt.angle();

    if (abstract_angle < angle1)
	abstract_angle += 2*PI;
    else if (abstract_angle > angle2)
	abstract_angle -= 2*PI;

    if (abstract_angle >= angle1 && abstract_angle <= angle2) {

	/*
	 * with a point (x,y),
	 * let A = (radiusY*x)^2 + (radiusX*y)^2 - (radiusX*radiusY)^2;
	 * A < 0 ==> (x,y) is in the filled ellipse;
	 * A = 0 ==> (x,y) is on the filled outline;
	 * A > 0 ==> (x,y) is out of the filled ellipse;
	 */
	
	GNumber rx2 = radiusX*radiusX;
	GNumber ry2 = radiusY*radiusY;
	GNumber A   = rx2*pt.y*pt.y + ry2*pt.x*pt.x - rx2*ry2;

	GVector ellipse_normal(CS.to(GVector(-abstract_pt.y * radiusX,
					     abstract_pt.x * radiusY)));
	// So far, ellipse_normal is only tangent to the ellipse;
	// Now, take an orthogonal direction and ..;
	// resize it to maxDistance (in absolute cs sense);
	GNumber tmp = ellipse_normal.x;
	ellipse_normal.x = ellipse_normal.y;
	ellipse_normal.y = -tmp;
	ellipse_normal.resize(maxDistance);

	// go back to local coordsys ;
	ellipse_normal = inversed_cs.to(ellipse_normal);

	// let us work in the first quarter of relative GCoordSys space;
	pt.x = fabs(pt.x);
	pt.y = fabs(pt.y);
	ellipse_normal.x = fabs(ellipse_normal.x);
	ellipse_normal.y = fabs(ellipse_normal.y);

	if (A <= 0.) {
	    // pt is in;
	    if (shbrush && shbrush->fillStyle() != GFILL_STIPPLED)
		return GTRUE;
	    
	    // return GTRUE if "pt + ellipse_normal" is out;
	    pt += ellipse_normal;
	    return (GBool)(rx2*pt.y*pt.y + ry2*pt.x*pt.x - rx2*ry2 >= 0);
	}
	// pt is out;
	// return GTRUE if "pt - ellipse_normal" is in;
	pt -= ellipse_normal;
	return (GBool)(rx2*pt.y*pt.y + ry2*pt.x*pt.x - rx2*ry2 <= 0);
    }
    return GFALSE;
}


GBool GoInterf::lineContains (const GShPen* shpen, const GCoordSys& cs,
			     const GPoint& p1, const GPoint& p2,
			     const GPoint& absolute, GNumber maxDist,
			     GQuality quality)
{
    GPoint pts[2];
    pts[0] = p1;
    pts[1] = p2;
    return polylineContains (shpen, NULL, cs, 2, pts, absolute, maxDist, quality);
}


GBool GoInterf::ellipseChordContains(const GShPen* shpen,
				     const GShBrush* shbrush,
				     const GCoordSys& cs, const GPoint& origin,
				     GNumber radiusX, GNumber radiusY,
				     GNumber angle1, GNumber angle2, GBool closed,
				     const GPoint& absolute, GNumber maxDist,
				     GQuality quality)
{
    if (!(shpen || shbrush))
	return GFALSE;

    GBool in_pie = ellipseContains(shpen, shbrush, cs, origin, radiusX, radiusY,
				  absolute, maxDist, angle1, angle2, quality);
    // in_pie would be the answer for a pie style elliptical arc, not closed;
    
    GNumber amplitude = fabs(angle2 - angle1);

    // return quickly on simple cases;

    // full ellipse ;
    if (amplitude >= 2*PI)
	return in_pie;

    // if outline only ;
    if (in_pie && ! shbrush)
	return GTRUE;

    /*
     * if amplitude >= 2*PI
     * then arc_chord filled area contains arc_pie filled area;
     * else arc_pie filled area contains arc_chord filled area;
     */
    if (amplitude >= PI) {
	if (in_pie)
	    return GTRUE;
    }
    else if (shbrush && ! in_pie) {
	return GFALSE;
    }

    GCoordSys CS = cs;
    CS.O = cs.to(origin);
    GPoint relative = CS.from(absolute); // in the ellipse intrinsic coord sys;

    GNumber rx2 = radiusX*radiusX;
    GNumber ry2 = radiusY*radiusY;
    if (rx2*relative.y*relative.y + ry2*relative.x*relative.x - rx2*ry2 > 0.) {
	// the point is out of the entire ellipse;
	return GFALSE;
    }
    
    GPoint pts[4];
    /*
     * get the chord segment plus one point close to each extremity
     * on the arc to take care of join style clip extensions.
     * pts[0] and pts[3] will be these extra points, pts[1] and pts[2]
     * will be the chord extremities.
     * pts[0] and pts[3] will only be use if there is an arc outline.
     */
    int extra = GEllipseArcLimits(CS, radiusX, radiusY, angle1, angle2,
				 pts[1], pts[2], pts[0], pts[3], quality);
    // extra != 0 means pts[0] and pts[3] are set;
    
    if (shpen && quality != GMIN_QUALITY)
	maxDist += shpen->lineWidth() / 2;

    // let us see if point is on the chord;
    if (closed && shpen &&
	((extra && GPointOnPolyline(CS, 4, pts, absolute,
				    maxDist * maxDist, GFALSE)) ||
	 (((!extra) && GPointOnPolyline(CS, 2, pts+1, absolute,
					maxDist * maxDist, GFALSE)))))
	return GTRUE;
    else if (! shbrush)
	return GFALSE;
    
    GBool dummy;
    GBool radius_cuts_chord = GIntersect(GPoint(0, 0), relative,
					 pts[1], pts[2], dummy);
    return ((amplitude >= PI) ? !radius_cuts_chord : radius_cuts_chord);
}

GBool GoInterf::ellipsePieContains (const GShPen* shpen, const GShBrush* shbrush,
				   const GCoordSys& cs, const GPoint& origin,
				   GNumber radiusX, GNumber radiusY,
				   GNumber angle1, GNumber angle2, GBool closed,
				   const GPoint& absolute, GNumber maxDist,
				   GQuality quality)
{
    if (!(shpen || shbrush))
	return GFALSE;

    GBool in_pie = ellipseContains(shpen, shbrush, cs, origin, radiusX, radiusY,
				  absolute, maxDist, angle1, angle2, quality);
    if ((angle1 - angle2 >= 2*PI) || ! closed)
	return in_pie;
    else if (in_pie)
	return GTRUE;
    else if (shpen && closed) {

	if (quality != GMIN_QUALITY)
	    maxDist += shpen->lineWidth() / 2;
	
	GCoordSys CS = cs;
	CS.O = cs.to(origin);
	GPoint pts[5];
	/*
	 * get the chord points plus one point close to each extremity on the
	 * arc to take care of join style extensions at arc limits,
	 * pts[0] and pts[4] will be these extra points, pts[1] and pts[3]
	 * will be the chord extremities and pts[2] the center of the ellipse.
	 * pts[0] and pts[4] will only be use if there is an arc outline.
	 */
	pts[2].x = pts[2].y = 0;
	if (GEllipseArcLimits(CS, radiusX, radiusY, angle1, angle2,
			     pts[1], pts[3], pts[0], pts[4], quality)) {
	    return GPointOnPolyline(CS, 5, pts, absolute,
				   maxDist * maxDist, GFALSE);
	}
	else {
	    return GPointOnPolyline(CS, 3, pts+1, absolute,
				   maxDist * maxDist, GFALSE);
	}
    }
    else return GFALSE;
}


GBool GoInterf::circleContains (const GShPen* shpen, const GShBrush* shbrush,
			       const GCoordSys& cs, const GPoint& center,
			       GNumber radius,
			       const GPoint& absolute, GNumber maxDist,
			       GNumber angle1, GNumber angle2,
			       GQuality quality)
{
    return ellipseContains (shpen, shbrush, cs,  center,
			    radius, radius,
			    absolute, maxDist, angle1, angle2, quality);

}

GBool GoInterf::circleChordContains (const GShPen* shpen, const GShBrush* shbrush,
				    const GCoordSys& cs,  const GPoint& center,
				    GNumber radius, GNumber angle1, GNumber angle2,
				    GBool closed, const GPoint& absolute,
				    GNumber maxDist, GQuality quality)
{
    return ellipseChordContains (shpen, shbrush, cs, center,
				 radius, radius, angle1, angle2, closed,
				 absolute, maxDist, quality);

}

GBool GoInterf::circlePieContains (const GShPen* shpen, const GShBrush* shbrush,
				  const GCoordSys& cs, const GPoint& center,
				  GNumber radius, GNumber angle1, GNumber angle2,
				  GBool closed, const GPoint& absolute,
				  GNumber maxDist, GQuality quality)
{
    return ellipsePieContains (shpen, shbrush, cs, center,
			       radius, radius, angle1, angle2, closed,
			       absolute, maxDist, quality);

}

GBool GoInterf::bezierContains (const GShPen* shpen, const GShBrush* shbrush,
				const GCoordSys& cs,
				int nctrl, const GPoint* ctrl,
				const GPoint& absolute, GNumber maxDist,
				GQuality quality)
{
    int npoint;
    GBool ret;
#ifdef NO_ALLOCA
    GPoint* pts = (GPoint*)GMALLOC(GMAX_BEZIER_POINTS(nctrl)* sizeof(GPoint));
#else
    GPoint* pts = (GPoint*)alloca(GMAX_BEZIER_POINTS(nctrl)* sizeof(GPoint));
#endif

    GBezierComputePoints (nctrl, ctrl, quality, npoint, pts);

    ret =  polylineContains (shpen, shbrush, cs, npoint, pts,
			     absolute, maxDist, quality);
#ifdef NO_ALLOCA
    GFREE (pts);
#endif
    return ret;
}

GBool GoInterf::cubicCurveContains (const GShPen* shpen,
				    const GShBrush* shbrush,
				    const GCoordSys& cs, int nctrl,
				    const GPoint* ctrl,
				    const GNumber (*basisMatrix)[4] /* [4][4] */,
				    const GPoint& absolute, GNumber maxDist,
				    int evalShift, GQuality quality)
{
    int npoint;
    GBool ret;
#ifdef NO_ALLOCA
    GPoint* pts = (GPoint*)GMALLOC(GMAX_CUBIC_POINTS(nctrl, evalShift) *
				   sizeof(GPoint));
#else
    GPoint* pts = (GPoint*)alloca(GMAX_CUBIC_POINTS(nctrl, evalShift) *
				  sizeof(GPoint));
#endif
 
    GCubicCurveComputePoints (nctrl, ctrl, basisMatrix, evalShift,
			      quality, npoint, pts);

    ret = polylineContains (shpen, shbrush, cs, npoint, pts,
			    absolute, maxDist, quality);
#ifdef NO_ALLOCA
    GFREE (pts);
#endif
    return ret;
}


  
/**
FUNCTION
    GCallback management
DESCRIPTION
    Add or remove postconditions for expose, event and destroy
**/
#define _add_rem_callback(type, flag) \
    ((GoInterfCondFlags*)&_cond)->flag =\
         0 < ((add == GADD)\
	      ? _addCallback(type, (GCBProc)cb, closure)\
	      : _remCallback(type, (GCBProc)cb, closure))

void GoInterf::exposePostcond (void cb(void*, GoInterf*, const GExposeEvent*),
			       const void* closure, GAddOrRemove add)
{
    _add_rem_callback (gointerfNexposePostcond, _has_expose_postcond);
    if (_cond._has_expose_postcond)
	addInputMask(GEXPOSE_MASK);
    else
	removeInputMask(GEXPOSE_MASK);
}

void GoInterf::eventPostcond (void cb(void*, GoInterf*, const GEvent*),
			      const void* closure, GAddOrRemove add)
{
    _add_rem_callback (gointerfNeventPostcond, _has_event_postcond);
}

void GoInterf::destroyPostcond (void cb (void*, GoInterf*),
				const void* closure, GAddOrRemove add)
{
    _add_rem_callback (gointerfNdestroyPostcond, _has_destroy_postcond);
}

void GoInterf::resizePostcond (void cb (void*, GoInterf*),
			       const void* closure, GAddOrRemove add)
{
    _add_rem_callback (gointerfNresizePostcond, _has_resize_postcond);
}
#undef _add_rem_callback


/**
FUNCTION
    Input Methods
DESCRIPTION
    Allow to manage pointer, input mask.s
**/

void GoInterf::addInputMask (GEventMask mask)
{
    int m;
    int i;
    for (i=0, m=1; mask && i < GN_EVENT_MASK ; m <<= 1, i++)
	if (mask & m) {
	    _eventMaskCount[i] += 1;
	    mask &= ~m;
	}
}

void GoInterf::removeInputMask (GEventMask mask)
{
    int m;
    int i;
    for (i=0, m=1; mask && i < GN_EVENT_MASK ; m <<= 1, i++)
	if (mask & m) {
	    _eventMaskCount[i] -= 1;
	    mask &= ~m;
	}
}

GEventMask GoInterf::inputMask () const
{
    ulong mask = 0;
    if (_eventMaskCount[GBUTTON_PRESS_MASK_INDEX])
	mask |= GBUTTON_PRESS_MASK;
    if (_eventMaskCount[GBUTTON_RELEASE_MASK_INDEX])
	mask |= GBUTTON_RELEASE_MASK;
    if (_eventMaskCount[GDOUBLE_CLICK_MASK_INDEX])
	mask |= GDOUBLE_CLICK_MASK;
    if (_eventMaskCount[GTRIPLE_CLICK_MASK_INDEX])
	mask |= GTRIPLE_CLICK_MASK;
    if (_eventMaskCount[GCHAR_MASK_INDEX])
	mask |= GCHAR_MASK;
    if (_eventMaskCount[GMESSAGE_MASK_INDEX])
	mask |= GMESSAGE_MASK;
    if (_eventMaskCount[GFUNCTION_KEY_MASK_INDEX])
	mask |= GFUNCTION_KEY_MASK;
    if (_eventMaskCount[GEXPOSE_MASK_INDEX])
	mask |= GEXPOSE_MASK;
    if (_eventMaskCount[GMOTION_MASK_INDEX])
	mask |= GMOTION_MASK;
    if (_eventMaskCount[GMESSAGE_MASK_INDEX])
	mask |= GMESSAGE_MASK;
    if (_eventMaskCount[GCOMPRESS_MOTION_MASK_INDEX])
	mask |= GCOMPRESS_MOTION_MASK;

    return mask;
}

/**
FUNCTION
    Query misc methods
DESCRIPTION
    pixel sizes, toPixel conversions
**/
int GoInterf::toPixelWidth(GNumber mm) const
{
    return int(mm/pixelWidth());
}

int GoInterf::toPixelHeight(GNumber mm) const
{
    return int(mm/pixelHeight());
}

/**
FUNCTION
    Setting misc methods
DESCRIPTION
    resize and set background
**/
void GoInterf::resize (GNumber width, GNumber height)
{
    if (width != _width || height != _height)
    {
	_width = width;
	_height = height;
	_call_resize_postcond();
    }
}

void GoInterf::background(const class GShColor* bg)
{
    if (bg) 
	if (_background)
	    _background->modify(bg);
	else _background = new GColor(bg);
    else {
	delete _background;
	_background = NULL;
    }
}


void GoInterf::pushCursor(long)
{
}

void GoInterf::pushCursor(const class GCursor*)
{
}

void GoInterf::pushCursor(const class GShCursor*)
{
}

void GoInterf::popCursor()
{
}

void GoInterf::bell(int)
{
}
    
