#ifdef REV_INFO
static char rcsid[] = "$Id: XColor.cc,v 1.8 1992/06/26 08:36:29 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  X INTERFACE
FILE:    XColor.cc
AUTHOR:  Philippe Smadja
DATE:    Nov 1990

DESCRIPTION:
         This file constains
	 1: the global association function for color
	 2: the XGoInterf methods relative to color
	 3: some static functions used inside.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/


#include <Go/XGoInterf.h>
#include "XColorP.h"
#include <assert.h>

#include <Go/Tools/Cons.h>
#include <Go/Tools/Alloc.h>
#include <Go/Tools/Array.h>
#include <Go/Tools/Colormap.h>
#include <Go/Tools/Error.h>

#include <X11/Xatom.h>
#include <math.h>

static Pixel mapRgbToXColor (GBool gray, GRgbValue red, GRgbValue green, GRgbValue blue,
			     XColor* xcolors, int nxcolors , int min_br, int max_br);

typedef uchar VisualClass;

/**
  COLOR ASSOC TOOLS
  *****************
  **/

struct _dColor {
    const Display*     dpy;
    Colormap           colormap;
    GAssoc*            assoc;
    int                nref;
};

static GArray& DColors() {
    static GArray* d_colors = NULL;

    if (! d_colors)
	d_colors = new GArray(sizeof(struct _dColor));

    return *d_colors;
}

#define MAX_BRIGHTNESS (65536*2/3)
#define L_OFFSET      (65536-MAX_BRIGHTNESS)
#define RED(v,m) ((GRgbValue)((MAX_BRIGHTNESS * (v) / (m)) +   L_OFFSET))
#define GREEN(v,m) ((GRgbValue)((MAX_BRIGHTNESS * (v) / (m)) + L_OFFSET))
#define BLUE(v,m) ((GRgbValue)((MAX_BRIGHTNESS * (v) / (m)) +  L_OFFSET))

/**/
void   freeAssocColor(GAssoc* assoc)
{
    struct _dColor* dColor = (struct _dColor*)DColors().data();
    int    nbdColors = DColors().nelems();
    
    for (int i = 0 ; i  < nbdColors; i++, dColor++)
	if (dColor->assoc == assoc) {
	    if( --dColor->nref == 0)  {
		delete assoc;
		DColors().rem(i);
	    }
	    return ; /* one assoc is found */
	}
}

/**/
GAssoc* getAssocColor(const Display* dpy, Colormap colormap, const Visual*)
{
    struct _dColor* dColor = (struct _dColor*)DColors().data();
    int    nbdColors = DColors().nelems();
    
    
    for (int i = 0 ; i  < nbdColors; i++, dColor++)
	if (dColor->colormap == colormap &&
	    dColor->dpy == dpy)
	{
	    dColor->nref++;
	    return dColor->assoc;
	}


    struct _dColor local;

    local.dpy = dpy;
    local.colormap = colormap;
    local.nref = 1;
    local.assoc = new GAssoc();
    DColors().add(&local);
    return local.assoc;
}


/**
FUNCTION
    HSBToRGB
DESCRIPTION
   h: [0..6.0[
   s: [0..1]
   v:  the base of the red,green,blue result.
**/
inline void HSV_to_RGB(double h, double s, double v,  GRgbValue& red, GRgbValue& green, GRgbValue& blue)
{
    double vs;
    int i, iv, ivs, ivsf;
    iv = (int) v;
    if (s < 1.0/100 || h < 0.)
    {
	red = green = blue = iv;
	return;
    }

    i = (int) h; 
    vs = v * s;
    ivsf = (int) (vs * (h - (double) i));
    ivs = (int) vs; 
    
    if (i == 0)
    {
	red = iv;
	blue = iv - ivs;
	green = blue + ivsf;
	return;
    }
    if (i == 1)
    {
	green = iv;
	blue = iv - ivs;
	red = iv - ivsf;
	return;
    } 
    if (i == 2)
    {
	green = iv;
	red = iv - ivs;
	blue = red + ivsf;
	return;
    } 
    if (i == 3)
    {
	blue = iv;
	red = iv - ivs;
	green = iv   - ivsf;
	return;
    } 
    if (i == 4) 
    {
	blue = iv;
	green = iv - ivs;
	red = green + ivsf;
	return;
    } 
    red = iv;
    green = iv - ivs;
    blue = iv - ivsf;
    return;
    
} /* HSVtoRGB */

inline double deltaH(int h, int nbh)
{
    double b = 6.0 / nbh;
    double r = b * h;
    if (r > 0.5 && r < 1.5)
	return  b * 0.5;
    else if (r > 3.2 && r < 4.5)
	return b * 1.65;
    else
	return b * 0.90;
}

static double sTable[3] = {
    0.25,
    0.45,
    0.90,
};

static double vTable[4] = {
    30000.0,
    40000.0,
    50000.0,
    65535.0,
};

/**
FUNCTION
    allocContinuousColormap()
DESCRIPTION
    make continuous colormap for visual of 256 map_entries
    make 6 gray, the rgb cuble corners.
    continue by variation of HUE and Value in the HSV model
    4 loop en value, 3 loop on saturation.
    the remaining on HUE;
**/
void XGoInterf::allocContinuousColormap(Pixel*& pixels, int& npixels)
{
    pixels = NULL;
    npixels = 0;

    switch (_visual->c_class)
    {
      default:
	break;
      case PseudoColor: 
	/**
	  make a continuous colormap
	  **/
	if (_visual->map_entries < 256)
	    return;
	
	XColor xcolor;
	/**
	  To do not have an unjustify warning: "use before set" with
	  some compilo
	  **/
	bzero(&xcolor, sizeof(xcolor));

	int nbH = 0;
	pixels = (Pixel*) GMALLOC(sizeof(Pixel) * _visual->map_entries);
	
	for (npixels = 1; npixels < 7;)
	{
	    xcolor.red = xcolor.green = xcolor.blue = GMAX_RGB_VALUE * npixels / 8;
	    if (XAllocColor((Display*)_dpy, _colormap, &xcolor) == 0)
		goto out;
	    pixels[npixels++] = xcolor.pixel;
	}

	/**
	  alloc the rgb cube corners
	  **/
	// black.;
	xcolor.red = 0, xcolor.green = 0, xcolor.blue = 0;
	if (XAllocColor((Display*)_dpy, _colormap, &xcolor) == 0)
	    goto out;
	pixels[npixels++] = xcolor.pixel;
	
	// red:
	xcolor.red = GMAX_RGB_VALUE;
	if (XAllocColor((Display*)_dpy, _colormap, &xcolor) == 0)
	    goto out;
	pixels[npixels++] = xcolor.pixel;

	// green:
	xcolor.red = 0, xcolor.green = GMAX_RGB_VALUE;
	if (XAllocColor((Display*)_dpy, _colormap, &xcolor) == 0)
	    goto out;
	pixels[npixels++] = xcolor.pixel;

	// yellow:
	xcolor.red = GMAX_RGB_VALUE;
	if (XAllocColor((Display*)_dpy, _colormap, &xcolor) == 0)
	    goto out;
	pixels[npixels++] = xcolor.pixel;

	// blue:
	xcolor.red = 0, xcolor.green = 0, xcolor.blue = GMAX_RGB_VALUE;
	if (XAllocColor((Display*)_dpy, _colormap, &xcolor) == 0)
	    goto out;
	pixels[npixels++] = xcolor.pixel;

	// magenta:
	xcolor.red = GMAX_RGB_VALUE;
	if (XAllocColor((Display*)_dpy, _colormap, &xcolor) == 0)
	    goto out;
	pixels[npixels++] = xcolor.pixel;

	// cyan:
	xcolor.red = 0; xcolor.green = GMAX_RGB_VALUE; xcolor.blue = GMAX_RGB_VALUE;
	if (XAllocColor((Display*)_dpy, _colormap, &xcolor) == 0)
	    goto out;
	pixels[npixels++] = xcolor.pixel;

	// white:
	xcolor.red = GMAX_RGB_VALUE;
	if (XAllocColor((Display*)_dpy, _colormap, &xcolor) == 0)
	    goto out;
	pixels[npixels++] = xcolor.pixel;


	
	// continue with  HSV variation:
	
	nbH = (_visual->map_entries-(8+6)) / (3*4);
	int h,s,v;
	
	for (s = 2; s >= 0; s--)
	{
	    double _s = sTable[s];
	    for (v = 0; v < 4; v++) 
	    {
		double _v = vTable[v];
		double _h = 0;
		for(h = 0; h < nbH;  h++, _h += deltaH(h, nbH))
		{
		    HSV_to_RGB(_h, _s, _v,
			       xcolor.red, xcolor.green, xcolor.blue);
		    if (XAllocColor((Display*)_dpy, _colormap, &xcolor) == 0) 
			goto out;
		    pixels[npixels++] = xcolor.pixel;
		}
	    }
	}

      out:
	GREALLOC(pixels, sizeof(Pixel)*npixels);
	break;
    }
}

/**
  XGOINTERF COLOR FUNCTIONS
  *************************
  **/


/**
FUNCTION
    shColorPostcond (update and destroy)
DESCRIPTION
    These postcond updates (destroys) the associations between
    the shared color and associated XColor
**/

static void ShColorIsDestroyed(void* closure, const GShColor* color)
{
    GAssoc* assoc = (GAssoc*)closure;
    int num;
    PixelAssoc* pf = (PixelAssoc*)assoc->get(color, num);
    assert (pf && num == 1);
    /**
      any interface in the list is available to get dpy, colormap etc...
      **/
    XGoInterf** xi = (XGoInterf**)(pf->xinterfs->data());
    VisualClass vclass = (*xi)->visual()->c_class;
    switch (vclass)
    {
      case DirectColor:
      case PseudoColor:
      case GrayScale:
	/**
	  Programmable colormap
	  free the pixel.
	  **/
	if (pf->pixel != BAD_PIXEL && pf->wasAllocated) 
	    XFreeColors((*xi)->display(), (*xi)->colormap(), &pf->pixel, 1, 0L);

	break;
      default:
	/**
	  NON programmable colormap
	  **/
	break;
    }

    /**
      remove: from the used list
      the association
      the callbacks
      **/
    pf->xinterfs->destroy();
    assoc->rem(color);
    
}

static void ShColorIsUpdated(void* assoc,
			     GShColor* color /* should be declare const */,
			     ulong /* mask */)
{
    ShColorIsDestroyed(assoc, color);
    /**
      remove the callbacks
      **/
    color->updatePostcond(ShColorIsUpdated, assoc, GREMOVE);
    color->destroyPostcond(ShColorIsDestroyed, assoc, GREMOVE);
}



/**
FUNCTION
    colorToXPixel
DESCRIPTION
    returns the nearest(if GTRUE) for the wanted color.
    if GFALSE and the color cannot be mapped it returns BAD_PIXEL
**/

ulong XGoInterf::_colorToXPixel(const GShColor* color,
				GBool nearest, int min_br, int max_br)
{
    PixelAssoc* pf;
    PixelAssoc _pf;
    XColor     xcolor;
    int num;
    pf = (PixelAssoc*)_assocColor->get(color, num);

    if (pf) 
    {
	assert (num == 1);
	/**
	  put the interface in used list if not yet
	  **/
	_addXInterfIfNeeded(pf->xinterfs);
	return PIXEL_TO_ULONG(pf->pixel);
    }

    /**
      COMPUTE the xcolor
      **/
	
    VisualClass vclass = _visual->c_class;
    switch(vclass)
    {
      case DirectColor:
      case PseudoColor:
      case GrayScale:
	/**
	  Programmable colormap
	  **/
	if (color->mask() & GCOLOR_RGB)
	{
	    xcolor.red =   color->rgb()->red & 0xFF00;
	    xcolor.green = color->rgb()->green & 0xFF00;
	    xcolor.blue =  color->rgb()->blue & 0xFF00;
	}
	else if (color->mask() & GCOLOR_GRAY)
	{
	    xcolor.red = color->gray() * 0xffff / 256;
	    xcolor.green = xcolor.blue = xcolor.red;
	}
	else
	{
	    GWarning("XGoInterf::_colorToXPixel() no RGB or GRAY. white is assumed");
	    xcolor.red = GMAX_RGB_VALUE;
	    xcolor.green = GMAX_RGB_VALUE;
	    xcolor.blue = GMAX_RGB_VALUE;
	}
	    
	/**
	  Now alloc the pixel inside the colormap
	  **/
	if (_pf.wasAllocated = XAllocColor(_dpy, _colormap, &xcolor))
	    _pf.pixel = xcolor.pixel;
	else if (!nearest) 
	    _pf.pixel = BAD_PIXEL;
	else
	{
	    // get the nearest color for the wanted color;

	    XColor* xcolors = NULL;
	    int nxcolors;
	    GBool free_xcolors;
	    getXScreenColor(_dpy, _screen, _colormap, _visual,
			    &xcolors, &nxcolors, &free_xcolors);

	    _pf.pixel =mapRgbToXColor(vclass == GrayScale, xcolor.red, xcolor.green,
				      xcolor.blue,xcolors, nxcolors, min_br, max_br);
	    /**
	      The returns od XAllocColor may be FALSE for colormat allocated with
	      read/write cells. In This case the color will follows the rgb of the cell.
	      **/
	    _pf.wasAllocated = XAllocColor(_dpy, _colormap,  &xcolors[_pf.pixel]);
		
	    if (free_xcolors)
		GFREE(xcolors);
	}

	break;
	      
      default:
	/**
	  NON programmable colormap
	  **/
	XColor* xcolors;
	int nxcolors;
	GBool free_xcolors;
	GRgbValue red;
	GRgbValue green;
	GRgbValue blue;
	getXScreenColor(_dpy, _screen, _colormap, _visual,
			&xcolors, &nxcolors, &free_xcolors);

	if (color->mask() & GCOLOR_RGB) 
	{
	    red =   color->rgb()->red & 0xFF00;
	    green = color->rgb()->green & 0xFF00;
	    blue =  color->rgb()->blue & 0xFF00;
	}
	else if (color->mask() & GCOLOR_GRAY)
	{
	    red = color->gray() * 0xffff / 256;
	    green = blue = red;
	}
	else
	{
	    GWarning("XGoInterf::_colorToXPixel() no RGB or GRAY. white is assumed");
	    red = GMAX_RGB_VALUE;
	    green = GMAX_RGB_VALUE;
	    blue = GMAX_RGB_VALUE;
	}
	    
	_pf.pixel = BAD_PIXEL;
	for (int px=0; px < nxcolors; px++)
	    if (xcolors[px].red == red &&
		xcolors[px].green == green &&			
		xcolors[px].blue == blue) {
		_pf.pixel = px;
		break;
	    }

	if (_pf.pixel == BAD_PIXEL &&
	    (color->mask() & (GCOLOR_RGB|GCOLOR_GRAY)) == (GCOLOR_RGB|GCOLOR_GRAY))
	{
	    red = color->gray() * 0xffff / 256;
	    green = blue = red;
	    for (int px=0; px < nxcolors; px++)
		if (xcolors[px].red == red &&
		    xcolors[px].green == green &&			
		    xcolors[px].blue == blue)
		{
		    _pf.pixel = px;
		    break;
		}
	}
	
	if (nearest && _pf.pixel == BAD_PIXEL)
	    _pf.pixel = mapRgbToXColor(vclass == StaticGray, red, green, blue,
				       xcolors, nxcolors, min_br, max_br);
	if (free_xcolors)
	    GFREE(xcolors);
	break;
    }


    if (nearest || _pf.pixel != BAD_PIXEL)
    {
	/**
	  make the association between color and pixel
	  **/
	void* t  = this;
	_pf.xinterfs =  GCreateShArray(&t);
	_assocColor->set(color, sizeof _pf, &_pf);

	color->updatePostcond (ShColorIsUpdated, _assocColor, GADD);
	color->destroyPostcond(ShColorIsDestroyed, _assocColor, GADD);
    }
    return PIXEL_TO_ULONG(_pf.pixel);
}


void XGoInterf::_freeXPixels()
{
    int nkeys;
    void** keys = _assocColor->getKeys(nkeys);
    void** _keys = keys;
    int npx = 0;
#define MAXPX 64
    Pixel pixels[MAXPX] ;

    for (int i = 0; i < nkeys; i++, keys++)
    {
	int num;
	PixelAssoc* pf = (PixelAssoc*)_assocColor->get(*keys, num);
	assert (pf && num == 1);
	/**
	  remove this from xinterfs list
	  **/
	_remXInterfIfInside(pf->xinterfs);
	if (pf->xinterfs == NULL)
	{
	    /**
	      remove the association and free the pixel
	      **/
	    ((GShColor*)*keys)->updatePostcond(ShColorIsUpdated,
					       _assocColor, GREMOVE);
	    ((GShColor*)*keys)->destroyPostcond(ShColorIsDestroyed,
						_assocColor, GREMOVE);
	    switch(_visual->c_class)
	    {
	      default:
		break;
	      case DirectColor:
	      case PseudoColor:
	      case GrayScale:
		/**
		  free the pixels
		  **/
		if (pf->wasAllocated &&  (pixels[npx] = pf->pixel) != BAD_PIXEL) 
		{
		    if (++npx == MAXPX) 
		    {
			XFreeColors(_dpy, _colormap, pixels, MAXPX, 0L);
			npx = 0 ;
		    }
		}
		break;
	    }
	    _assocColor->rem(*keys);
	}
    }
#undef MAXPX
    if (npx)
	XFreeColors(_dpy, _colormap, pixels, npx, 0L);
    GFREE(_keys);
}




/**
  STATIC FUNCTIONS  FOR THIS FILE
  *******************************
  **/
/**
FUNCTION
    double RGB_to_HSV(r,g,b, h&, s&, v&)
DESCRIPTION
    compute the HUE of the (r,g,b) in the HSV model
    and return the value in the range [0, 6[, -1: hue is undefined
**/
static void
RGB_to_HSV(GRgbValue red, GRgbValue green, GRgbValue blue, double& h, double& s, double& v)
{
  double delta;
  long _min, _max;

  _max = max(max( (long)red, (long)green), (long)blue);
  v = (double)_max;

  if (_max == 0) 
  {
      h = -1;     /* undefined */
      s = 0;
      return;
  }

  _min = min(min( (long)red, (long)green), (long)blue);

  if (_min == _max) 
  {
     h = -1;  /* undedined */
     s = 0;
     return;
 }

  delta = (double)(_max - _min);
  s = delta / v;
  
  double incr = 0.0;
  if (red == _max)
  {
      h = (double) (green - blue) / delta ;
      if (h < 0.)
	  incr = 6;
  } 
  else  if (green == _max) 
  {
      h = (double) (blue - red) / delta ;
      incr = 2.0;
  } else 
  {
      h = (double)(red - green) / delta ;
      incr = 4.0;
  }
  h += incr;
}

/**
FUNCTION
    mapRgbToXColor
DESCRIPTION
    These function returns the nearest  existing color  for
    the wanted RGB value.
    the result brightness will be in the range [min_br ... max_br]
**/
static Pixel
mapRgbToXColor (GBool grayscreen, GRgbValue red, GRgbValue green, GRgbValue blue,
		XColor* xcolors, int nxcolors,
		int min_br, int max_br)
{
    int index;
    int i;
    
    if (grayscreen)
    {
	/*
	  compute the brightness.
	  and take the min in the xcolors table.
	  red=green=blue in the xolors table
	  */
	uint Y = GBrightness(red, green, blue);
	uint Min_dist = GMAX_RGB_VALUE;
	index = 0;
	for (i=0; i < nxcolors; i++)
	{
	    int Dist;
	    Dist = xcolors[i].red;
	    // max_br and min_br unused;
	    Dist = Y - Dist;
	    if (Dist < 0) Dist = -Dist;
	    if ((uint)Dist < Min_dist)
	    {
		Min_dist = Dist;
		index = i;
	    }
	}

    }
    else		      /* !grayscreen */
    {
	/*
	  GColor.
	  calculate the Hue of each xcolors (if needed).
	  The Hue is computed is the HSV model.
	  After compute the Hue of the GRgb and take the min
	  in the xcolors table.
	  */
	double h;
	double s;
	double v;
	RGB_to_HSV(red, green, blue, h, s, v);
	if (h == -1) 
	{
	    /**
	      gray case: compute the brightness
	      to get the Dist
	      **/
	    uint Y = GBrightness(red, green, blue);
	    int Min_dist_Y = 3*GMAX_RGB_VALUE;
	    index = 0;
	    for (i=0; i < nxcolors; i++)  
	    {
		int dist_Y;
		dist_Y = GBrightness(xcolors[i].red,
				     xcolors[i].green,
				     xcolors[i].blue);
		if (dist_Y > max_br || dist_Y < min_br)
		    continue;

		dist_Y = Y - dist_Y;
		if (dist_Y < 0)
		    dist_Y = -dist_Y;

		if (dist_Y < Min_dist_Y) {
		    Min_dist_Y = dist_Y;
		    index = i;
		}
	    }
	}
	else		      /* HUE != -1 */ 
	{
	    /**
	      COLOR CASE
	      compute the closest RGB
	      **/
#define MAXRGB__ ((float)((uint)GMAX_RGB_VALUE))
	    float minDist = 3*MAXRGB__*MAXRGB__ + 1;
#undef MAXRGB__
	    index = 0;
	    for (i=0; i < nxcolors; i++) 
	    {
		float dr = red   - xcolors[i].red;
		float dg = green - xcolors[i].green;
		float db = blue  - xcolors[i].blue;
		float dist = dr*dr + dg*dg + db*db;
	
		if (dist < minDist) 
		{
		    minDist = dist;
		    index = i;
		}
	    }
	}
    }

    return (Pixel)xcolors[index].pixel ;
}

struct _screenDesc {
	const Display* dpy;
	uchar   screen;
	uchar   vclass;
	Colormap colormap;
	XColor* xcolors;
	int     nxcolors;
};

static _screenDesc* screenDescList = NULL ;
static maxScreenDesc = 0 ;
static XColor* XColorToFree = NULL;
static int     NxcolorsToFree = 0;
static GBool    ColormapToXPixelsRunning = GFALSE;
    
void getXScreenColor(const Display* dpy, uchar screen, Colormap colormap,
		     const Visual* visual,
		     XColor** xcolors, int* nxcolors,
		     GBool*   free_xcolors)
{
    XColor* tmp;
    
    switch(visual->c_class)
    {
      case DirectColor:
      case PseudoColor:
      case GrayScale:
	/**
	  programmable colormap
	  ask the server for the current colormap
	  allocation
	  **/
	if (ColormapToXPixelsRunning && XColorToFree) 
	{
	    assert ( NxcolorsToFree == visual->map_entries);
	    *xcolors = XColorToFree;
	    *nxcolors = NxcolorsToFree;
	    *free_xcolors = GFALSE;
	    return;
	}

	*nxcolors = visual->map_entries;
	tmp = *xcolors = (XColor*)GMALLOC(sizeof(XColor)* *nxcolors);
	for (int i = 0; i < *nxcolors; i++, tmp++) 
	{
	    tmp->pixel = i ;
	    tmp->pad = 0;     /* hue undefined see rgbToXColor */
	}
	XQueryColors((Display*)dpy, colormap, *xcolors, *nxcolors);

	/**
	  find the last unused color
	  for programmable visual
	  **/
	for (Pixel px = *nxcolors-1 ; px != 0; px--)
	{
	    XColor result ;
	    result.red = (*xcolors)[px].red;
	    result.green = (*xcolors)[px].green;
	    result.blue = (*xcolors)[px].blue;
	    int status = XAllocColor((Display*)dpy, colormap, &result);
	    if (!status)
		break ;

	    XFreeColors((Display*)dpy, colormap, &result.pixel, 1, 0L);

	    if (result.pixel == (*xcolors)[px].pixel)
		break ;
	}
    
	*nxcolors = (int)px + 1;

	*free_xcolors = (ColormapToXPixelsRunning == GFALSE || *nxcolors != visual->map_entries);

	if (*free_xcolors == GFALSE && ColormapToXPixelsRunning)
	{
	    XColorToFree = *xcolors;
	    NxcolorsToFree =  visual->map_entries;
	}
	else
	{
	    XColorToFree = NULL;
	    NxcolorsToFree = 0;
	}

	break;

      default:
	/**
	  NON programmable colormap
	  get it and save it inside the application
	  **/

	*free_xcolors = GFALSE;

	for(i=0; i <maxScreenDesc; i++)
	    if (dpy ==  screenDescList[i].dpy &&
		screen == screenDescList[i].screen &&
		visual->c_class == screenDescList[i].vclass &&
		colormap == screenDescList[i].colormap)
	    {
		/**
		  color is NULL: keep the current one: no change on colormap
		  allocation
		  **/
		*xcolors = screenDescList[i].xcolors;
		*nxcolors = screenDescList[i].nxcolors;
		return ;
	    }

	/**
	  no place: alloc a new one and set it
	  **/
	GREALLOC(screenDescList, (maxScreenDesc+1)* sizeof(_screenDesc));

	_screenDesc* s = &screenDescList[maxScreenDesc++] ;

	s->dpy = dpy;
	s->screen = screen ;
	s->vclass = visual->c_class ;
	s->colormap = colormap;
	*xcolors = s->xcolors = (XColor*)GMALLOC(sizeof(XColor) * visual->map_entries);
	*nxcolors = s->nxcolors = visual->map_entries;
	
	for (i = 0, tmp = s->xcolors ; i < s->nxcolors; i++, tmp++)
	{
	    tmp->pixel = i ;
	    tmp->pad = 0;     /* hue undefined see rgbToXColor */
	}
	XQueryColors((Display*)dpy, colormap, s->xcolors, s->nxcolors);
	break;
    }
}

struct Ref  {
    Pixel  pixel;
    int    br;
} ;


GArray* XGoInterf::_colormapToXPixels(const GShPixelArray* pxlarray,
				      const GShColormap* colormap)
{
    GArray* array = new GArray(NULL, (int)colormap->nColor(), sizeof (Pixel));
    Pixel* pixels = (Pixel*)array->data();
    const  GShColor** colors = colormap->colors();
    const  GPixelArrayValues* pav = pxlarray->values();
    ulong  bytesPerLine = pxlarray->bytesPerLine();
    ulong  i;

    /**
      the colormap will not be changed during the loop
      even by another application (normally we have to grag the server)
      **/
    ColormapToXPixelsRunning = GTRUE;
    
    if (pav->bits_per_pixel != 8) 
    {
	/**
	  do not lost time to compute the used color in the pixelarray
	  **/

	for(i=0; i < colormap->nColor(); i++, colors++, pixels++)
	    if (*colors == NULL)
		*pixels = BAD_PIXEL;
	    else
		*pixels = _colorToXPixel(*colors, GTRUE);

    }
    else		      /* bits_per_pixel = 8 */
    {
	/**
	  compute the used colors in the colormap
	  **/
#ifdef NO_ALLOCA
	Ref* ref = (Ref*)GMALLOC(colormap->nColor() * sizeof (Ref));
#else
	Ref* ref = (Ref*)alloca((int) (colormap->nColor() * sizeof (Ref)));
#endif
	for (i = 0 ; i < colormap->nColor(); i++) 
	{
	    ref[i].pixel = i;
	    ref[i].br = -1;
	}
	    
	const uchar* data = pav->data;
	ulong  delta  = bytesPerLine  - pav->width;
	int min_br = GMAX_RGB_VALUE;
	int max_br = 0;
	int nbUsed = 0;

	for (int y=0 ; y < pav->height; y++, data += delta)
	    for (int x = 0; x < pav->width; x++) 
	    {
		uchar p = *data++;
		if (ref[p].br == -1)
		{
		    int br = ref[p].br  = colors[p]->brightness();
		    if (br > max_br)
			max_br = br;
		    if (br < min_br)
			min_br = br;
		    if (++nbUsed == colormap->nColor())
			goto  compute;
		}
	    }

      compute:
#ifdef NO_ALLOCA
	Ref* saveRef = ref;
#endif

	if (min_br == max_br)
	{
	    min_br = 0;
	    max_br = GMAX_RGB_VALUE;
	}
	/**
	  try to alloc all needed pixels
	  **/
	for (i = 0; i < colormap->nColor();  i++, ref++)
	{
	    const GShColor* color = colors[ref->pixel] ;
	    if (ref->br == -1 || color == NULL)
		pixels[ref->pixel] = BAD_PIXEL;
	    else 
		pixels[ref->pixel] = _colorToXPixel(color, GTRUE, min_br, max_br);
	}
#ifdef NO_ALLOCA
	GFREE(saveRef);
#endif
    }

    ColormapToXPixelsRunning = GFALSE;
    if (XColorToFree)
    {
	GFREE(XColorToFree);
	assert (XColorToFree == NULL);
	NxcolorsToFree = 0;
    }
    
    return array;
}


GShColor* XGoInterf::colorFromName(const char* color_name) const
{
    XColor xcolor;

    int status = XParseColor(_dpy, _colormap, color_name, &xcolor);
    if (!status)
    {
	GWarning("XGoInterf:colorFromName: unknown color name %s\n black is assumed\n", color_name);
	xcolor.red = 0;
	xcolor.green = 0;
	xcolor.blue = 0;
    }

    GColorValues values;

    values.rgb.red = xcolor.red;
    values.rgb.green = xcolor.green;
    values.rgb.blue = xcolor.blue;

    return GCreateShColor(GCOLOR_RGB, &values);
}
