#ifdef REV_INFO
static char rcsid[] = "$Id: XDash.cc,v 1.6 1992/06/26 08:36:33 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  XGoInterface
FILE:    XDash.cc
AUTHOR:  Philippe Smadja
DATE:    Nov 90

DESCRIPTION:
    Dash management in the X-Go interface
    
COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/XGoInterf.h>

#include "XDashP.h"
#include <assert.h>

#include <Go/Tools/Cons.h>
#include <Go/Tools/Alloc.h>
#include <Go/Tools/ShArray.h>
#include <Go/Tools/ShDash.h>

/**
  DASH ASSOC TOOLS
  ****************
  **/
struct _dDash {
    GNumber pixelSize;
    GAssoc* assoc;
    int    nref;
};

static GArray& DDashes() {
    static GArray *d_dashes = NULL;

    if (! d_dashes)
	d_dashes = new GArray(sizeof(struct _dDash));

    return *d_dashes;
}


void   freeAssocDash(GAssoc* assoc)
{
    struct _dDash* dDash  = (struct _dDash*)DDashes().data();
    int nbdDashes = DDashes().nelems();
    
    for (int i = 0 ; i  < nbdDashes; i++, dDash++)
	if (dDash->assoc == assoc)
	{
	    if( --dDash->nref == 0)
	    {
		delete  assoc;
		DDashes().rem(i);
	    }
	    return ;
	}
}


GAssoc* getAssocDash(GNumber pixelWidth, GNumber  pixelHeight)
{
    struct _dDash* dDash  = (struct _dDash*)DDashes().data();
    int nbdDashes = DDashes().nelems();

    
    for (int i = 0 ; i  < nbdDashes; i++, dDash++)
	if (dDash->pixelSize == (pixelWidth + pixelHeight)/2)
	{
	    dDash->nref++;
	    return dDash->assoc;
	}

    struct _dDash local ;
    local.pixelSize = (pixelWidth + pixelHeight)/2;
    local.nref = 1;
    local.assoc = new GAssoc();

    DDashes().add(&local);
    return local.assoc;
}


struct DashAssoc {
    GShArray* xinterfs;
    ulong dash_ident;
    char* dash_list;
    int   n_dashes;
    int   dash_offset;
};


/**
FUNCTION
    shDashPostcond (update and destroy)
DESCRIPTION
    These postcond updates (destroys) the associations between
    the shared dash and associated XDash
**/

static void ShDashIsDestroyed(void* closure, const GShDash* dash)
{
    GAssoc* assoc = (GAssoc*)closure;
    int num;
    DashAssoc* pf = (DashAssoc*)assoc->get(dash, num);
    assert (pf && num == 1);

    /**
      delete the  dash list
      **/
    GFREE(pf->dash_list);
    /**
      remove: from the used list
      the association
      **/
    pf->xinterfs->destroy();
    assoc->rem(dash);
    
}

static void ShDashIsUpdated(void* assoc,
			    GShDash* dash /* should be declare const */,
			    ulong /* mask not used */)
{
    ShDashIsDestroyed(assoc, dash);
    /**
      remove the callbacks
      **/
    dash->updatePostcond(ShDashIsUpdated, assoc, GREMOVE);
    dash->destroyPostcond(ShDashIsDestroyed, assoc, GREMOVE);
}

static ulong DASH_IDENT = (ulong)(((long)NO_DASH)+1); /** first valid dash_ident **/

ulong XGoInterf::_dashToXDash(const GShDash* dash, char** dash_list,
			      int* n_dashes, int* dash_offset) const
{
    int num;
    DashAssoc *pf = (DashAssoc*)_assocDash->get(dash, num);
    if (pf)
    {
#ifdef DEBUG
	assert (num == 1);
#endif
	/**
	  put interfaces in used list
	  **/
	_addXInterfIfNeeded(pf->xinterfs);
	*dash_list = pf->dash_list;
	*n_dashes = pf->n_dashes;
	*dash_offset = pf->dash_offset;
	return  pf->dash_ident;
    }

    /**
      compute dash_list
      **/
    DashAssoc _pf;
    void* t = (void*)this;
    _pf.xinterfs = GCreateShArray(&t);
    _pf.n_dashes = dash->dash_number();
    _pf.dash_offset =  mm_to_linew(dash->dash_offset());
    _pf.dash_list = (char*)GMALLOC(_pf.n_dashes);
    _pf.dash_ident = DASH_IDENT++;
    for (int i = 0 ; i < _pf.n_dashes; i++)
	_pf.dash_list[i] = mm_to_linew(dash->dash_list()[i]);
    
    _assocDash->set(dash, sizeof _pf, &_pf);

    dash->updatePostcond(ShDashIsUpdated, _assocDash, GADD);
    dash->destroyPostcond(ShDashIsDestroyed, _assocDash, GADD);

    *dash_list = _pf.dash_list;
    *n_dashes = _pf.n_dashes;
    *dash_offset = _pf.dash_offset;
    return _pf.dash_ident;
}


void XGoInterf::_freeXDashes()
{
    int nkeys;
    void** keys = _assocDash->getKeys(nkeys);
    void** _keys = keys;

    for (int i = 0; i < nkeys; i++, keys++) 
    {
	int num;
	DashAssoc* pf = (DashAssoc*)_assocDash->get(*keys, num);
	assert (pf && num == 1);
	/**
	  remove this from xinterfs list
	  **/
	_remXInterfIfInside(pf->xinterfs);
	if (pf->xinterfs == NULL)
	{
	    /**
	      remove the association and free the dash list.
	      **/
	    ((GShDash*)*keys)->updatePostcond(ShDashIsUpdated,
					     _assocDash, GREMOVE);
	    ((GShDash*)*keys)->destroyPostcond(ShDashIsDestroyed,
					      _assocDash, GREMOVE);
	    GFREE(pf->dash_list);
	    _assocDash->rem(*keys);
	}
    }

    /**
      free all the keys
      **/
    GFREE(_keys);
}


