#ifdef REV_INFO
static char rcsid[] = "$Id: XGC.cc,v 1.10 1992/09/25 15:27:30 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  X INTERFACE
FILE:    XGC.cc
AUTHOR:  Philippe Smadja
DATE:    Nov 1990

DESCRIPTION:
         This file contains:
	 1: the global functions for association.
	 2: the XGoInterf methods relatives to gc
	 3: the ShXgc methods  of the ShXgc object

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/

#include <Go/XGoInterf.h>
#include <assert.h>

#include "XGCP.h"
#include "XColorP.h"
#include "XFontP.h"
#include "XRegionP.h"
#include "XDashP.h"

#include <Go/Tools/ShColor.h>
#include <Go/Tools/ShColormap.h>
#include <Go/Tools/ShPattern.h>


#include <Go/Tools/Cons.h>
#include <Go/Tools/Atom.h>
#include <Go/Tools/Alloc.h>
#include <Go/Tools/Array.h>
#include <Go/Tools/ShArray.h>
#include <Go/Tools/Error.h>

#define functionToXfunction(f)    (int)((f))
#define lineStyleToXLineStyle(s)  (int)((s))
#define capStyleToXCapStyle(s)    (int)((s))
#define joinStyleToXJointStyle(s) (int)((s))
#define fillStyleToXFillStyle(s)  (int)((s))
#define fillRuleToXFillRule(r)    (int)((r))

/**
  DEFAULT for
  min GQuality management
  **/

static GPen&        DefaultPen() {
    static GPen *default_pen = NULL;

    if (! default_pen)
	default_pen = new GPen();

    return *default_pen;
}

static GBrush&      DefaultBrush() {
    static GBrush *default_brush = NULL;

    if (! default_brush)
	default_brush = new GBrush();

    return *default_brush;
}

static GTextBrush&  DefaultTextBrush() {
    static GTextBrush *default_text_brush = NULL;

    if (! default_text_brush)
	default_text_brush = new GTextBrush();

    return *default_text_brush;
}


/**
  GC ASSOC TOOLS
  *****************
  **/

struct _dGC {
    Display*    dpy;
    uchar       screen;
    uchar       depth;
    GAssoc*      assoc;
    GC          copyGC;
    GC          clearGC;
    GC          setGC;
    int         nref;
};

static GArray& DGCs() {
    static GArray *d_gcs = NULL;

    if (! d_gcs)
	d_gcs = new GArray(sizeof(struct _dGC));

    return *d_gcs;
}


void   freeAssocGC(GAssoc* assoc)
{
    struct _dGC* dGC = (struct _dGC*)DGCs().data();
    int nbdGCs = DGCs().nelems();
    
    for (int i = 0 ; i  < nbdGCs; i++, dGC++)
	if (dGC->assoc == assoc) {
	    if (--dGC->nref == 0)  {
		delete assoc;
		XFreeGC(dGC->dpy, dGC->copyGC);
		XFreeGC(dGC->dpy, dGC->clearGC);
		XFreeGC(dGC->dpy, dGC->setGC);
		DGCs().rem(i);
	    }
	    return ;
	}
}


GAssoc* getAssocGC(Display* dpy, uchar screen, uchar depth,
		  GC* copyGC, GC* clearGC, GC* setGC)
{
    struct _dGC* dGC = (struct _dGC*)DGCs().data();
    int nbdGCs = DGCs().nelems();
    
    for (int i = 0 ; i  < nbdGCs; i++, dGC++)
	if (dGC->screen == screen && dGC->depth == depth &&
	    dGC->dpy == dpy)  {
	    dGC->nref++;
	    *copyGC   = dGC->copyGC;
	    *clearGC  = dGC->clearGC;
	    *setGC    = dGC->setGC;
	    return dGC->assoc;
	}

    struct _dGC local;
    local.dpy =  dpy;
    local.screen  = screen ;
    local.depth = depth;
    local.nref = 1;
    local.assoc = new GAssoc();

    /**
      make 2 GCs for depth 1 drawables
      **/
    XGCValues values;
    Pixmap depth1 = XCreatePixmap(dpy, RootWindow(dpy,screen),1,1,1);
    values.function = GXcopy;
    local.copyGC = XCreateGC(dpy, depth1,GCFunction, &values);
    values.function = GXclear;
    local.clearGC = XCreateGC(dpy, depth1,GCFunction, &values);
    values.function = GXset;
    local.setGC   = XCreateGC(dpy, depth1,GCFunction, &values);
    XFreePixmap(dpy,depth1);
 
    DGCs().add(&local);
    *copyGC   = local.copyGC;
    *clearGC  = local.clearGC;
    *setGC    = local.setGC;
    return local.assoc;
}


/**
  ColorPatternAssoc TOOL
  **/
struct _dColorPattern {
    Display*    dpy;
    Colormap    colormap;
    GAssoc*     assoc;
    int         nref;
};

static GArray& DColorPatterns() {
    static GArray *d_color_patterns = NULL;

    if (! d_color_patterns)
	d_color_patterns = new GArray(sizeof(struct _dColorPattern));

    return *d_color_patterns;
}


void   freeAssocColorPattern(GAssoc* assoc)
{
    struct _dColorPattern* dColorPattern = (struct _dColorPattern*)DColorPatterns().data();
    int nbdColorPatterns = DColorPatterns().nelems();
    
    for (int i = 0 ; i  < nbdColorPatterns; i++, dColorPattern++)
	if (dColorPattern->assoc == assoc) {
	    if (--dColorPattern->nref == 0)  {
		delete assoc;
		DColorPatterns().rem(i);
	    }
	    return ;
	}
}


GAssoc* getAssocColorPattern(Display* dpy, Colormap colormap)
{
    struct _dColorPattern* dColorPattern = (struct _dColorPattern*)DColorPatterns().data();
    int nbdColorPatterns = DColorPatterns().nelems();
    
    for (int i = 0 ; i  < nbdColorPatterns; i++, dColorPattern++)
	if (dColorPattern->dpy == dpy && dColorPattern->colormap == colormap)
	{
	    dColorPattern->nref++;
	    return dColorPattern->assoc;
	}

    struct _dColorPattern local;
    local.dpy =  dpy;
    local.colormap = colormap;
    local.nref = 1;
    local.assoc = new GAssoc();

    DColorPatterns().add(&local);
    return local.assoc;
}

/*
 * static variable of the shxgc class
 */
GList*        ShXgc::_headOfList;
GFunctionList ShXgc::_funcList;

/**
  PROTECTED SHXGC METHODS
  **/

ShXgc* _newShXgc (GFunctionList* funcList, ulong mask,
		  const void* values, const char* name)
{
   return new ShXgc(funcList, mask, (const ShXgcValues*)values, name);
}

/**/
ShXgc::ShXgc (GFunctionList* funcList, ulong mask, const ShXgcValues* values, const char* name) : GSharedObject(funcList, mask, name)
{
    assert (mask == values->xgcmask);
    assert (name == NULL);

    _dpy = values->dpy;
    _depth = values->depth;
    _screen = values->screen;
    _colormap  = values->colormap;    
    _visual = values->visual;

    if (values->shxregion)
	_shxregion = values->shxregion->dup();
    else
	_shxregion = NULL;

    _xoffset = values->xoffset;
    _yoffset = values->yoffset;

    
    _xgcvalues = *values->xgcvalues;
    _xgcmask   = values->xgcmask;

    /**
      CREATE GC
      **/
    _xgcvalues.arc_mode = ArcChord;      /** the X default ArcPieSlice
					   and the Go mode is ArcChord
					   **/
    
    _gc  = XCreateGC(_dpy, values->drawable,
		     (_xgcmask & ~GCDashList) | GCArcMode, &_xgcvalues);

    if (_xgcmask & GCDashList) {
	_dash_ident = values->dash_ident;
	_dash_list = values->dash_list;
	_n_dashes = values->n_dashes;
	_dash_offset = values->dash_offset;
	XSetDashes(_dpy, _gc, _dash_offset, _dash_list, _n_dashes);
    } else
	_dash_ident = (ulong)NO_DASH;
	
}

/**/
GList*& ShXgc::_head ()
{
    return _headOfList;
}

/**/
int ShXgc::_valueSize ()
{
    return (sizeof(ShXgcValues));
}

/**/
GBool ShXgc::_match (ulong mask, const void* __values) const
{
    const ShXgcValues* values = (const ShXgcValues*)__values;
    assert (mask == values->xgcmask);

#define _matchValue(maskBit, field) \
    if (mask == 0)  \
	goto _true ; \
    if (mask & maskBit) {\
	mask &= ~maskBit ;\
	if (_xgcmask & maskBit &&\
	    values->xgcvalues->field != _xgcvalues.field) goto _false ;\
    }

    _matchValue (GCFont, font);
    _matchValue (GCForeground, foreground);
    _matchValue (GCBackground, background);
    _matchValue (GCLineWidth, line_width);
    _matchValue (GCFunction, function);
    _matchValue (GCPlaneMask, plane_mask);
    _matchValue (GCLineStyle, line_style);
    _matchValue (GCCapStyle, cap_style);
    _matchValue (GCJoinStyle, join_style);
    _matchValue (GCFillStyle, fill_style);
    _matchValue (GCFillRule, fill_rule);
    _matchValue (GCTile, tile);
    _matchValue (GCStipple, stipple);
    _matchValue (GCTileStipXOrigin, ts_x_origin);
    _matchValue (GCTileStipYOrigin, ts_y_origin);
#undef _matchValue

    /**
      DASH
      **/
    if (mask & GCDashList) {
	mask &= ~GCDashList ;
	if (_xgcmask & GCDashList && values->dash_ident != _dash_ident)
	    goto _false ;
    }

_true:
    if (_dpy == values->dpy &&
	_colormap == values->colormap &&
	_depth == values->depth &&
	_screen == values->screen &&
	_shxregion == values->shxregion && _xoffset == values->xoffset &&
	_yoffset == values->yoffset)
	return GTRUE;
_false:
    return GFALSE;
}

/**/
void ShXgc::_makeNewValues (ulong mask, const void* values,
			    ulong& newMask, void* newValues) const
{
    *(ShXgcValues*)newValues = *(const ShXgcValues*)values;
    newMask = mask;
}

/**/
ulong ShXgc::_update (ulong mask, const void* __values)
{
    const ShXgcValues* values = (const ShXgcValues*)__values;
    int create = 0 ;
    
    assert (mask == values->xgcmask);

    if (_dpy !=  values->dpy)
	create++;
    else if (_depth != values->depth)
	create++ ;
    else if (_screen != values->screen)
	create++;
    else if (_colormap != values->colormap)
	create++;
    else if (_visual->c_class != values->visual->c_class)
	create++;
    
    if (create)  {
	XFreeGC(_dpy, _gc);
	/**
	  save the new values
	  **/
	_dpy = values->dpy;
	_depth = values->depth;
	_screen = values->screen;
	_colormap  = values->colormap;    
	_visual = values->visual;
    }

    
    ShXRegion* tmp = NULL;
    if (values->shxregion)
	tmp = values->shxregion->dup();
    if (_shxregion) {
	assert (_shxregion->refcnt() >= 1);
	_shxregion->destroy();
    }
    _shxregion = tmp;
    _xoffset = values->xoffset;
    _yoffset = values->yoffset;


    ulong old_mask = _xgcmask;
    ulong update_mask_return = 0;
    XGCValues* xgcvalues = values->xgcvalues;
    
    if (_refcnt == 1)
	_xgcmask = mask;
    else _xgcmask |= mask;
    
#define _updateValue(maskBit, field)\
    if (!mask) goto _execute;\
    if (mask & maskBit) {\
	if  (!(old_mask & maskBit) || (_xgcvalues.field != xgcvalues->field)){\
           update_mask_return |= maskBit;\
	   _xgcvalues.field = xgcvalues->field;\
        }\
        mask &= ~maskBit;\
    }

    _updateValue (GCFont, font);
    _updateValue (GCForeground, foreground);
    _updateValue (GCBackground, background);
    _updateValue (GCLineWidth, line_width);
    _updateValue (GCFunction, function);
    _updateValue (GCPlaneMask, plane_mask);
    _updateValue (GCLineStyle, line_style);
    _updateValue (GCCapStyle, cap_style);
    _updateValue (GCJoinStyle, join_style);
    _updateValue (GCFillStyle, fill_style);
    _updateValue (GCFillRule, fill_rule);
    _updateValue (GCTile, tile);
    _updateValue (GCStipple, stipple);
    _updateValue (GCTileStipXOrigin, ts_x_origin);
    _updateValue (GCTileStipYOrigin, ts_y_origin);

#undef _updateValue

    if (mask & GCDashList) {
	assert (values->dash_ident != NO_DASH);
	if ( !(old_mask & GCDashList) || _dash_ident != values->dash_ident) {
	    _dash_ident = values->dash_ident;
	    _dash_list = values->dash_list;
	    _n_dashes = values->n_dashes;
	    _dash_offset = values->dash_offset;
	    update_mask_return |= GCDashList;
	}
    }


_execute:
    /**
      NOW create or update the GC
      **/
    if (create)  {
	assert (_xgcvalues.arc_mode == ArcChord);
	_gc = XCreateGC(_dpy, values->drawable,
			(_xgcmask & ~GCDashList) | GCArcMode, &_xgcvalues);
    } else
	XChangeGC(_dpy, _gc,
		  _xgcmask & ~GCDashList, &_xgcvalues);

    if (_xgcmask & GCDashList)
	XSetDashes(_dpy, _gc, _dash_offset, _dash_list, _n_dashes);

    return update_mask_return;
}

/**/
ShXgc::~ShXgc ()
{
    XFreeGC(_dpy, _gc);
    if (_shxregion) {
	assert (_shxregion->refcnt() >= 1);
	_shxregion->destroy();
    }
}

/**
  PUBLIC SHXGC METHODS
  **/
void GModify (ShXgc*& sh, const ShXgcValues* values)
{
   sh->modify((GSharedObject*&)sh, values->xgcmask, values);
}

/**/
ShXgc* ShXgc::dup() const
{
    return (ShXgc*)((GSharedObject*)this)->dup();
}


/**/
ShXgc* GCreateShXgc (const ShXgcValues* values)
{
   return GCREATE_SHAREDOBJECT(ShXgc, values->xgcmask, values, NULL);
}



/**
  XGOINTERF XGC FUNCTIONS
  ***********************
  **/

struct ColorPatternAssoc {
    GShPattern* shpattern;
    GShArray*   xinterfs;
};

static void  DestroyAssociatedShPattern(void* a, const GShColor* color)
{
    GAssoc* assoc = (GAssoc*)a;
    int npf;
    ColorPatternAssoc* pf = (ColorPatternAssoc*)assoc->get(color, npf);
    assert (pf && npf == 1) ;

    pf->xinterfs->destroy();
    pf->shpattern->destroy();

    assoc->rem(color);
}


/**
FUNCTION
    _colorToXGCValues
DESCRIPTION
    map the wanted color and alternate color to the X screen.
    If color are not availables and half_tone are used we try
    to used it with a tile.
    if color is not available and halftone is NULL we try to
    make a pattern if the colormap size is less or equal to 16
    else we take the nearest color available in the colormap.
    
**/
void
XGoInterf::_colorToXGCValues (GShColor* color, GShColor* alt_color,
			      ShXgcValues* shxgcvalues, GQuality quality)
{
    Pixel pxcolor;
    
    if (color) {
	pxcolor = _colorToXPixel(color, quality == GMIN_QUALITY ||
				 _useColorPatternMatching == GFALSE || 
				 (shxgcvalues->xgcmask & (GCStipple|GCTile)) != 0);

        if (pxcolor == BAD_PIXEL)  {
	    const GShPattern* shpattern = color->halftone();
	    if (shpattern == NULL) {
		/**
		  test if the pattern is not computed yet
		  **/
		int npf;
		ColorPatternAssoc* pf = (ColorPatternAssoc*)_assocColorPattern->get(color,npf);
		if (pf) {
		    assert (npf == 1);
		    shpattern = pf->shpattern;
		} else {
		    /**
		      compute it if possible
		      **/
		    shpattern = color->makeHalftone(_visual->c_class == GrayScale ||
						    _visual->c_class == StaticGray);
		    if (shpattern)  {
			void* t = this;
			ColorPatternAssoc _pf;
			/**
			  the cast disable the const. here it is normal
			  because the above affectation non const to const
			  is possible but not the reverse affectation
			  **/
			_pf.shpattern = (GShPattern*)shpattern;
			_pf.xinterfs = GCreateShArray(&t);
			_assocColorPattern->set(color, sizeof(_pf), &_pf);
			/**
			  post cond on color destruction
			  to destroy the associated pattern
			  **/
			color->destroyPostcond(DestroyAssociatedShPattern,
					       _assocColorPattern, GADD);
		    } else {
			pxcolor = _colorToXPixel(color, GTRUE);
			shxgcvalues->xgcmask |= GCForeground;
			shxgcvalues->xgcvalues->foreground = pxcolor;
		    }
		}
	    }

	    if (shpattern) {
		shxgcvalues->xgcmask |= GCTile;
		shxgcvalues->xgcvalues->fill_style = FillTiled;
		shxgcvalues->xgcvalues->tile = _tileToXTile(shpattern, quality);
	    }
	} else {
	    shxgcvalues->xgcmask |= GCForeground;
	    shxgcvalues->xgcvalues->foreground = pxcolor;
	}
    }

    if (alt_color) {
	pxcolor = _colorToXPixel( alt_color, GTRUE);
	if (pxcolor != BAD_PIXEL) {
	    shxgcvalues->xgcmask |= GCBackground;
	    shxgcvalues->xgcvalues->background = pxcolor;
	}
    }
}





/**
  Tools for XGCValues creation
  **/
ShXgcValues* XGoInterf::_shPenToShXgcValues(const GShPen* shpen,
					    ShXgcValues* shxgcvalues,
					    GQuality quality)
{
    const GPenValues* pv = shpen->values();

    shxgcvalues->dpy =   _dpy;
    shxgcvalues->screen = _screen;
    shxgcvalues->depth =    _depth;
    shxgcvalues->drawable = _drawable;
    shxgcvalues->colormap = _colormap;
    shxgcvalues->visual = _visual;
    shxgcvalues->shxregion = NULL;
    shxgcvalues->xoffset = 0;
    shxgcvalues->yoffset = 0;
    
    
    ulong* mask = &shxgcvalues->xgcmask;
    XGCValues* xgcvalues = shxgcvalues->xgcvalues;
    
    *mask = GCFunction|GCPlaneMask|GCLineWidth|GCLineStyle|
	GCCapStyle|GCJoinStyle|GCFillStyle;

    
    xgcvalues->function = functionToXfunction(pv->function);
    xgcvalues->plane_mask = pv->plane_mask;
    /**
      line
      **/
    xgcvalues->line_width = mm_to_linew(pv->line_width);
    if (xgcvalues->line_width == 1)
	xgcvalues->line_width = 0;
    
    xgcvalues->line_style = lineStyleToXLineStyle(pv->line.style);
    xgcvalues->cap_style =  capStyleToXCapStyle(pv->line.cap);
    xgcvalues->join_style = joinStyleToXJointStyle(pv->line.join);

    if (shpen->mask() & GDASH && xgcvalues->line_style != LineSolid) {

	 *mask |= GCDashList;
	 shxgcvalues->dash_ident = _dashToXDash(pv->dash,
						&shxgcvalues->dash_list,
						&shxgcvalues->n_dashes,
						&shxgcvalues->dash_offset);

    }


/**
    How to miter limit in X ?????
    *****************************
 **/
    if (shpen->mask() & GMITER_LIMIT && pv->miter_limit != GNO_MITER_LIMIT)
	GWarning("Sorry in XGoInterf::_shPenToGC() %s not implemented yet", "miterlimit");

    
    /**
      fill
      **/
    xgcvalues->fill_style = fillStyleToXFillStyle(pv->fill.style);

    if (shpen->mask() & GTILE && xgcvalues->fill_style == FillTiled) {

	// TILE: the color is inside the tile.

	assert (pv->tile);
	*mask |= GCTile;
	xgcvalues->tile = _tileToXTile(pv->tile, quality);

	if (shpen->mask() & GTILE_STIP_XORIGIN) {
	    *mask |= GCTileStipXOrigin;
	    xgcvalues->ts_x_origin = mm_to_pixx(pv->ts_x_origin);
	}

	if (shpen->mask() & GTILE_STIP_YORIGIN) {
	    *mask |= GCTileStipYOrigin;
	    xgcvalues->ts_y_origin = mm_to_pixy(pv->ts_y_origin);
	}

    } else {
	if (shpen->mask() & GSTIPPLE &&
	    (xgcvalues->fill_style == FillStippled
	     || xgcvalues->fill_style == FillOpaqueStippled)) {

	    // STIPPLE:
	  
	    assert (pv->stipple);
	    *mask |= GCStipple;
	    xgcvalues->stipple = _stippleToXStipple(pv->stipple, quality);

	    if (shpen->mask() & GTILE_STIP_XORIGIN) {
		*mask |= GCTileStipXOrigin;
		xgcvalues->ts_x_origin = mm_to_pixx(pv->ts_x_origin);
	    }

	    if (shpen->mask() & GTILE_STIP_YORIGIN) {
		*mask |= GCTileStipYOrigin;
		xgcvalues->ts_y_origin = mm_to_pixy(pv->ts_y_origin);
	    }

	} 

	/**
	  manage color and alt_color: Tile may be used for half-tone.
	  **/
	_colorToXGCValues (pv->color, pv->alt_color,
			   shxgcvalues, quality);

	if (xgcvalues->function == GXxor &&  xgcvalues->fill_style == FillSolid &&
	    (*mask & (GCForeground|GCBackground)) == (GCForeground|GCBackground))
	    xgcvalues->foreground ^=  xgcvalues->background;
    }
    
    return shxgcvalues;
}



ShXgcValues* XGoInterf::_shBrushToShXgcValues(const GShBrush* shbrush,
					      ShXgcValues* shxgcvalues,
					      GQuality quality)
{
    const GBrushValues* pv = shbrush->values();

    shxgcvalues->dpy =    _dpy;
    shxgcvalues->screen = _screen;
    shxgcvalues->depth =    _depth;
    shxgcvalues->drawable = _drawable;
    shxgcvalues->colormap = _colormap;
    shxgcvalues->visual = _visual;
    shxgcvalues->shxregion = NULL;
    shxgcvalues->xoffset = 0;
    shxgcvalues->yoffset = 0;
    
    ulong* mask = &shxgcvalues->xgcmask;
    XGCValues* xgcvalues = shxgcvalues->xgcvalues;
    
    *mask = GCFunction|GCPlaneMask|GCFillStyle|GCFillRule;
    
    xgcvalues->function = functionToXfunction(pv->function);
    xgcvalues->plane_mask = pv->plane_mask;
    xgcvalues->fill_style = fillStyleToXFillStyle(pv->fill.style);
    xgcvalues->fill_rule =  fillRuleToXFillRule(pv->fill.rule);


    if (shbrush->mask() & GTILE && xgcvalues->fill_style == FillTiled) {

	// TILE: the color is inside the tile.

	assert (pv->tile);
	*mask |= GCTile;
	xgcvalues->tile = _tileToXTile(pv->tile, quality);

	if (shbrush->mask() & GTILE_STIP_XORIGIN) {
	    *mask |= GCTileStipXOrigin;
	    xgcvalues->ts_x_origin = mm_to_pixx(pv->ts_x_origin);
	}

	if (shbrush->mask() & GTILE_STIP_YORIGIN) {
	    *mask |= GCTileStipYOrigin;
	    xgcvalues->ts_y_origin = mm_to_pixy(pv->ts_y_origin);
	}

    } else {
	if (shbrush->mask() & GSTIPPLE &&
	    (xgcvalues->fill_style == FillStippled
	     || xgcvalues->fill_style == FillOpaqueStippled)) {

	    // STIPPLE:
	  
	    assert (pv->stipple);
	    *mask |= GCStipple;
	    xgcvalues->stipple = _stippleToXStipple(pv->stipple, quality);

	    if (shbrush->mask() & GTILE_STIP_XORIGIN) {
		*mask |= GCTileStipXOrigin;
		xgcvalues->ts_x_origin = mm_to_pixx(pv->ts_x_origin);
	    }

	    if (shbrush->mask() & GTILE_STIP_YORIGIN) {
		*mask |= GCTileStipYOrigin;
		xgcvalues->ts_y_origin = mm_to_pixy(pv->ts_y_origin);
	    }

	} 

	/**
	  manage color and alt_color: Tile may be used for half-tone.
	  **/
	_colorToXGCValues (pv->color, pv->alt_color,
			   shxgcvalues, quality);

	if (xgcvalues->function == GXxor &&  xgcvalues->fill_style == FillSolid &&
	    (*mask & (GCForeground|GCBackground)) == (GCForeground|GCBackground))
	    xgcvalues->foreground ^=  xgcvalues->background;
    }
    
    return shxgcvalues;
}



ShXgcValues* XGoInterf::_shTextBrushToShXgcValues(const GShTextBrush* shtextbrush,
						  const GCoordSys& cs, GNumber font_height,
						  ShXgcValues* shxgcvalues,
						  GQuality quality)
{
    const GTextBrushValues* pv = shtextbrush->values();

    shxgcvalues->dpy =    _dpy;
    shxgcvalues->screen = _screen;
    shxgcvalues->depth =  _depth;
    shxgcvalues->drawable = _drawable;
    shxgcvalues->colormap = _colormap;
    shxgcvalues->visual = _visual;
    shxgcvalues->shxregion = NULL;
    shxgcvalues->xoffset = 0;
    shxgcvalues->yoffset = 0;

    ulong* mask = &shxgcvalues->xgcmask;
    XGCValues* xgcvalues = shxgcvalues->xgcvalues;
    
    *mask = GCFunction|GCPlaneMask|GCFillStyle;
    
    xgcvalues->function = functionToXfunction(pv->function);
    xgcvalues->plane_mask = pv->plane_mask;
    xgcvalues->fill_style = fillStyleToXFillStyle(pv->fill.style);

    if (shtextbrush->mask() & GTILE && xgcvalues->fill_style == FillTiled) {

	// TILE: the color is inside the tile.

	assert (pv->tile);
	*mask |= GCTile;
	xgcvalues->tile = _tileToXTile(pv->tile, quality);

	if (shtextbrush->mask() & GTILE_STIP_XORIGIN) {
	    *mask |= GCTileStipXOrigin;
	    xgcvalues->ts_x_origin = mm_to_pixx(pv->ts_x_origin);
	}

	if (shtextbrush->mask() & GTILE_STIP_YORIGIN) {
	    *mask |= GCTileStipYOrigin;
	    xgcvalues->ts_y_origin = mm_to_pixy(pv->ts_y_origin);
	}

    } else {
	if (shtextbrush->mask() & GSTIPPLE &&
	    (xgcvalues->fill_style == FillStippled
	     || xgcvalues->fill_style == FillOpaqueStippled)) {

	    // STIPPLE:
	  
	    assert (pv->stipple);
	    *mask |= GCStipple;
	    xgcvalues->stipple = _stippleToXStipple(pv->stipple, quality);

	    if (shtextbrush->mask() & GTILE_STIP_XORIGIN) {
		*mask |= GCTileStipXOrigin;
		xgcvalues->ts_x_origin = mm_to_pixx(pv->ts_x_origin);
	    }

	    if (shtextbrush->mask() & GTILE_STIP_YORIGIN) {
		*mask |= GCTileStipYOrigin;
		xgcvalues->ts_y_origin = mm_to_pixy(pv->ts_y_origin);
	    }

	} 

	/**
	  manage color. NO alternate color(background) in XDrawString().
	  **/
	_colorToXGCValues (pv->color,
			   xgcvalues->function == GXxor &&
			   xgcvalues->fill_style == FillSolid ?  pv->alt_color : NULL,
			   shxgcvalues, quality);

	if (xgcvalues->function == GXxor &&  xgcvalues->fill_style == FillSolid &&
	    (*mask & (GCForeground|GCBackground)) == (GCForeground|GCBackground)) {
	    xgcvalues->foreground ^=  xgcvalues->background;
	    *mask &= ~GCBackground;
	}
    }

    if (shtextbrush->mask() & GFONT) {
	*mask |= GCFont;
	xgcvalues->font = _fontToXFont(pv->font, cs, font_height);
    }
	
    return shxgcvalues;
}



ShXgcValues*
XGoInterf::_shImageBrushToShXgcValues(const GShImageBrush* shimagebrush,
				      uchar depth,
				      const ShXRegion* shxregion, int xoffset, int yoffset,
				      ShXgcValues* shxgcvalues, GQuality)
{
    shxgcvalues->dpy = _dpy;
    shxgcvalues->screen = _screen;
    shxgcvalues->depth =   _depth;
    shxgcvalues->drawable = _drawable;
    shxgcvalues->colormap = _colormap;
    shxgcvalues->visual = _visual;

    shxgcvalues->shxregion = shxregion;
    if (shxregion == NULL) {
	xoffset = 0;
	yoffset = 0;
    }
    shxgcvalues->xoffset = xoffset;
    shxgcvalues->yoffset = yoffset;

    XGCValues* xgcvalues = shxgcvalues->xgcvalues;
    ulong* mask = &shxgcvalues->xgcmask;
    
    *mask = GCPlaneMask|GCFunction;
    xgcvalues->function = functionToXfunction(shimagebrush->function());
    xgcvalues->plane_mask = shimagebrush->planeMask();

    if (depth == 1) {
	const GShColormap* cmap = shimagebrush->colormap();
	xgcvalues->foreground = _colorToXPixel(cmap->colors()[1]);
	if (xgcvalues->foreground == BAD_PIXEL) {
	    GWarning("XGoInterf::_shImageBrushToGC: %s", "bad colormap");
	    return shxgcvalues;
	}
	*mask |= GCForeground;
	
	xgcvalues->background = _colorToXPixel(cmap->colors()[0]);
	if (xgcvalues->background == BAD_PIXEL) {
	    GWarning("XGoInterf::_shImageBrushToGC: %s", "bad colormap");
	    return shxgcvalues;
	}
	*mask |= GCBackground;
    }
    
    return   shxgcvalues;
}


/**
  PEN
  **/
struct GCAssoc {
    ShXgc* shxgc;
    ushort modified:1,
           clipok:1;
    XGoInterf* clipOfXi;
    GShArray*   xinterfs;
    struct {
	GNumber height;       // GTextBrush;
	GVector I;
	GVector J;
    } tb;

};


static void ShPBIBisUpdated(void* assoc,
			    GSharedObject* sho /* should be declare const */,
			    ulong /* mask */)
{
    int npf;
    GCAssoc* pf = (GCAssoc*)((GAssoc*)assoc)->get(sho, npf);
    for (; npf; npf--, pf++) 
	pf->modified = GTRUE;
}

static void ShPBIBisDestroyed(void* a, const GSharedObject* sho)
{
    GAssoc* assoc = (GAssoc*)a;
    int npf;
    GCAssoc* pf = (GCAssoc*)assoc->get(sho, npf);

    for (; npf; npf--, pf++) {
	pf->shxgc->destroy();
	pf->xinterfs->destroy();
    }

    assoc->rem(sho);
}

void XGoInterf::_setClip(GC gc)
{
    if (_gcclip.clip_mask)
	XSetRegion(_dpy, gc, _gcclip.clip_mask);
    else if (_gcclip.xrects)
	XSetClipRectangles(_dpy, gc, 0, 0,
			   _gcclip.xrects, _gcclip.nxrects,
			   _gcclip.ordering);
    else
	XSetClipMask(_dpy, gc, None);
}

void XGoInterf::_setGCClip(GBool /* makeYXBanded is always done: do not worry */)
{
    /**
      remove the clipok flag in the gcs
      **/
    int nkeys;
    void** keys = _assocGC->getKeys(nkeys);
    void** _keys = keys;
    for(int i = 0 ; i < nkeys; i++, keys++) {
	int npf;
	GCAssoc* pf = (GCAssoc*)_assocGC->get(*keys, npf);
	for (; npf; npf--, pf++)
	    if (pf->clipOfXi == this)
		pf->clipok = GFALSE;
    }

    GFREE(_keys);

    if (_clip == GBiggestRegion()) {
	GFREE(_gcclip.xrects);
	_gcclip.nxrects = 0;
	goto _out;
    }
    
    /**
      make the new clip with clip goregion
      **/

    GFREE(_gcclip.xrects);
    int num_boxes;
    _gcclip.xrects = (XRectangle*)
	_clip.pixelBoxes(_pixelWidth, _pixelHeight, num_boxes);
    _gcclip.nxrects = num_boxes;
    
_out:
    /**
      free the clip_mask and return.
      **/
    if (_gcclip.clip_mask)
	XDestroyRegion(_gcclip.clip_mask);
    _gcclip.clip_mask = NULL;
    _gcclip.ordering = YXBanded;
} 


GC XGoInterf::_shPenToGC (const GShPen* shPen, GQuality quality)
{

    int num;
    if (quality == GMIN_QUALITY) {
	/**
	  conserve only  function, color
	  and altColor if function is GXOR
	  note that the values.color (and alt_color) cast are available because
	  it is a temporary copy.
	  **/
	ulong mask = GFUNCTION|GCOLOR;
	GPenValues values;
	values.function = shPen->function();
	values.color    = (GShColor*)shPen->color();
	if (values.function == GXor && shPen->mask() & GALT_COLOR) {
	    mask |= GALT_COLOR;
	    values.alt_color = (GShColor*)shPen->altColor();
	}
	DefaultPen().modify(mask, &values);
	shPen = DefaultPen().shPen();
    }
    
    GCAssoc _pf;
    GCAssoc* pf = (GCAssoc*)_assocGC->get(shPen, num) ;
#ifdef DEBUG
    if (pf) assert (num == 1);
#endif
    if (pf && pf->modified == GFALSE) {
	/**
	  put the interface in usedlist if not yet
	  **/
	_addXInterfIfNeeded(pf->xinterfs);
	goto _return ;
    }

    ShXgcValues shxgcvalues;
    XGCValues xgcvalues;
    shxgcvalues.xgcvalues = &xgcvalues;

    _shPenToShXgcValues(shPen,  &shxgcvalues, quality);

    if (pf)
    {
	GModify(pf->shxgc, &shxgcvalues);
	pf->modified = GFALSE;
	pf->clipok = GFALSE;
	/**
	  put the interface in usedlist if not yet
	  **/
	_addXInterfIfNeeded(pf->xinterfs);
	goto _return ;
	
    }
    else
    {
	void* t = this;
	_pf.shxgc = GCreateShXgc(&shxgcvalues);
	_pf.modified = GFALSE;
	_pf.clipok = GFALSE;
	_pf.xinterfs = GCreateShArray(&t);
	
	/**
	  add postcond
	  **/
	((GSharedObject*)shPen)->updatePostcond(ShPBIBisUpdated, _assocGC, GADD);
	((GSharedObject*)shPen)->destroyPostcond(ShPBIBisDestroyed, _assocGC, GADD);

	pf = &_pf;
    }

_return:
    /**
      handle the clip and return the GC
      **/
    if (_clipCount) {
	_setGCClip();
	_clipCount = 0;
    }
    if (!pf->clipok || pf->clipOfXi != this) {
	_setClip(pf->shxgc->gc());
	pf->clipok = GTRUE;
	pf->clipOfXi = this;
    }
    if (pf == &_pf)
	_assocGC->set(shPen, sizeof(*pf), pf);

    return pf->shxgc->gc();
}

GC XGoInterf::_shBrushToGC (const GShBrush* shBrush, GQuality quality)
{
    int num;
    if (quality == GMIN_QUALITY) {
	/**
	  conserve only  function, color
	  and altColor if function is GXOR
	  note that the values.color (and alt_color) cast are available because
	  it is a temporary copy.
	  **/
	ulong mask = GFUNCTION|GCOLOR;
	GBrushValues values;
	values.function = shBrush->function();
	values.color    = (GShColor*)shBrush->color();
	if (values.function == GXor && shBrush->mask() & GALT_COLOR) {
	    mask |= GALT_COLOR;
	    values.alt_color = (GShColor*)shBrush->altColor();
	}
	DefaultBrush().modify(mask, &values);
	shBrush = DefaultBrush().shBrush();
    }
    
    GCAssoc _pf;
    GCAssoc* pf = (GCAssoc*)_assocGC->get(shBrush, num) ;
#ifdef DEBUG
    if (pf) assert (num == 1);
#endif
    if (pf && pf->modified == GFALSE) {
	/**
	  put the interface in usedlist if not yet
	  **/
	_addXInterfIfNeeded(pf->xinterfs);
	goto _return;
    }
    
    ShXgcValues shxgcvalues;
    XGCValues xgcvalues;
    shxgcvalues.xgcvalues = &xgcvalues;

    _shBrushToShXgcValues(shBrush,  &shxgcvalues, quality);

    if (pf)
    {
	GModify(pf->shxgc, &shxgcvalues);
	pf->modified = GFALSE;
	pf->clipok = GFALSE;
	/**
	  put the interface in usedlist if not yet
	  **/
	_addXInterfIfNeeded(pf->xinterfs);
	goto _return;

    }
    else
    {
	void* t= this;
	_pf.shxgc = GCreateShXgc(&shxgcvalues);
	_pf.modified = GFALSE;
	_pf.clipok = GFALSE;
	_pf.xinterfs = GCreateShArray(&t);
	
	((GSharedObject*)shBrush)->updatePostcond( ShPBIBisUpdated, _assocGC, GADD);
	((GSharedObject*)shBrush)->destroyPostcond( ShPBIBisDestroyed, _assocGC, GADD);

	pf = &_pf;
    }

_return:
    /**
      handle the clip and return the GC
      **/
    if (_clipCount) {
	_setGCClip();
	_clipCount = 0;
    }
    if (!pf->clipok || pf->clipOfXi != this) {
	_setClip(pf->shxgc->gc());
	pf->clipok = GTRUE;
	pf->clipOfXi = this;
    }
    if (pf == &_pf)
	_assocGC->set(shBrush, sizeof(*pf), pf);

    return pf->shxgc->gc();
}

/**
FUNCTION
    shImageBrushToGC.
DESCRIPTION
    Note that we always modify the GC associated to the shImageBrush
    in particular to handle the different triple( shxregion, xoffset, yoffset).
**/
GC XGoInterf::_shImageBrushToGC (const GShImageBrush* shImageBrush,
				 const GCoordSys& cs,
				 uchar depth, const ShXRegion* shxregion,
				 int xoffset, int yoffset, GQuality quality)
{
    int npf;
    GCAssoc* pf = (GCAssoc*)_assocGC->get(shImageBrush, npf) ;
#ifdef DEBUG
    if (pf) assert(npf == 1);
#endif
    
    
    GCAssoc _pf;
    ShXgcValues shxgcvalues;
    XGCValues xgcvalues;
    shxgcvalues.xgcvalues = &xgcvalues;

    const ShXRegion* lastShXRegion = NULL;
    int lastXoffset;
    int lastYoffset;
    Region stipple_xregion = NULL;
    
    if (shImageBrush->mask() & GSTIPPLE &&
	shImageBrush->fillStyle() == GFILL_STIPPLED)
	/**
	  make a region with the stipple
	  **/
	stipple_xregion = _stippleToXRegion(shImageBrush->stipple(), cs);


    /**
      the shxregion is represent the deformation
      which is also done for the stipple if any.
      so the shxregion may ne used for the match criterium the ShXgc

      stipple_xregion and shxregion->_region has the SAME x(y)offset.
      which  are 0.
      **/
    _shImageBrushToShXgcValues(shImageBrush,  depth,
			       shxregion, xoffset, yoffset,
			       &shxgcvalues, quality);

    if (pf)
    {
	// handle any modification: in particular the (xregion,xoffset, yoffset);
	// NOTE that the triple is used only for match criterium;

	lastShXRegion = pf->shxgc->shXRegion();
	lastXoffset = pf->shxgc->xoffset();
	lastYoffset = pf->shxgc->yoffset();

	if (pf->modified || lastShXRegion != shxregion ||
	    lastXoffset != xoffset || lastYoffset != yoffset)
	    pf->clipok = GFALSE;

	GModify(pf->shxgc, &shxgcvalues);
	pf->modified  = GFALSE;
    }
    else
    {
	void* t = this;
	_pf.shxgc = GCreateShXgc(&shxgcvalues);
	_pf.modified = GFALSE;
	_pf.clipok = GFALSE;
	_pf.xinterfs = GCreateShArray(&t);
    
	((GSharedObject*)shImageBrush)->updatePostcond( ShPBIBisUpdated, _assocGC, GADD);
	((GSharedObject*)shImageBrush)->destroyPostcond( ShPBIBisDestroyed, _assocGC, GADD);
	pf = &_pf;
    }
    
    /**
      handle the clip and return the GC
      first handle the region to draw.
      **/
    
    if (_clipCount) 
    {
	// it is for the go _clip;

	_setGCClip();
	_clipCount = 0;
    }
    
    if (!pf->clipok || pf->clipOfXi != this) 
    {
	if (shxregion->region() || stipple_xregion) 
	{
	    // use the clip_mask to handle the region;
	    // BUT ONLY FOR THE CURRENT GC;

	    // 1er: ask X to make union with the xrects and the xregion;
	    // ********************************************************;

	    Region go_xregion = NULL;
	    if (_clip != GBiggestRegion())
	    {
		// add the xrects;
		go_xregion =  XCreateRegion();
		XRectangle* xrect  = _gcclip.xrects + _gcclip.nxrects - 1;
		for (int i = _gcclip.nxrects-1 ; i>=0 ; i--, xrect--)
		    XUnionRectWithRegion(xrect, go_xregion, go_xregion);
	    }
	
	    // add the src_xregion;

	    if (_gcclip.clip_mask)
		XDestroyRegion(_gcclip.clip_mask);
	    _gcclip.clip_mask = XCreateRegion();

	    // init clip_mask : the (Region) cast is needed;
	    // because the Xprototype does not handle const functionnality;

	    if (shxregion->region() && stipple_xregion) 
		XIntersectRegion((Region)shxregion->region(),
				 stipple_xregion, _gcclip.clip_mask);
	    else
		if (stipple_xregion)
		    XUnionRegion(stipple_xregion, stipple_xregion,
				 _gcclip.clip_mask);
		else
		    XUnionRegion((Region)shxregion->region(),
				 (Region)shxregion->region(),
				 _gcclip.clip_mask);


	    // displace the clip_mask;

	    XOffsetRegion(_gcclip.clip_mask, xoffset, yoffset);

	    // intersection with the goregion;

	    if (go_xregion)
	    {
		XIntersectRegion(go_xregion, _gcclip.clip_mask, _gcclip.clip_mask);
		/**
		  free go_xregion
		  **/
		XDestroyRegion(go_xregion);
	    }
	}
    
	_setClip(pf->shxgc->gc());
	/**
	  free the clip_mask if needed
	  **/
	if (_gcclip.clip_mask)
	{
	    XDestroyRegion(_gcclip.clip_mask);
	    _gcclip.clip_mask = NULL;
	}
	pf->clipok = GTRUE;
	pf->clipOfXi = this;
    }

    
    if (pf == &_pf)
	_assocGC->set(shImageBrush, sizeof(*pf), pf);

    return pf->shxgc->gc();
}


GC XGoInterf::_shTextBrushToGC (const GShTextBrush* shTextBrush,
				const GCoordSys& cs, GNumber height,
				const ShXFont** shxfont,
				GQuality quality)
{
    int npf;
    if (quality == GMIN_QUALITY) 
    {
	/**
	  conserve only  function, color
	  and altColor if function is GXOR
	  note that the values.color (and alt_color) cast are available because
	  it is a temporary copy.
	  **/
	ulong mask = GFUNCTION|GCOLOR|GFONT;
	GTextBrushValues values;
	values.function = shTextBrush->function();
	values.color    = (GShColor*)shTextBrush->color();
	if (values.function == GXor && shTextBrush->mask() & GALT_COLOR)
	{
	    mask |= GALT_COLOR;
	    values.alt_color = (GShColor*)shTextBrush->altColor();
	}
	values.font = (GShFont*)shTextBrush->font();
	DefaultTextBrush().modify(mask, &values);
	shTextBrush = DefaultTextBrush().shTextBrush();
    }

    GCAssoc  _pf;
    GCAssoc* pf = (GCAssoc*)_assocGC->get(shTextBrush, npf);
#ifdef DEBUG
    if (pf) assert(npf >= 1);
#endif
    int remove = 0;
    int NPF = npf;
    ShXgc* saveShXgc = NULL;
    for (; npf; npf--, pf++)
	if (pf->modified)  
	{
	    pf->xinterfs->destroy();

	    // remove all the associated GC;

	    if (remove++ == 0)
		saveShXgc = pf->shxgc;
	    else pf->shxgc->destroy();

	} else if (pf->tb.height == height &&
		   pf->tb.I == cs.I && pf->tb.J == cs.J)
	{
	    // put the interface in usedlist if not yet;

	    _addXInterfIfNeeded(pf->xinterfs);
	    goto _return;
	}
    if (NPF && remove)
    {
	assert (remove == NPF);
	_assocGC->rem(shTextBrush);
    }
    
    ShXgcValues shxgcvalues;
    XGCValues xgcvalues;
    shxgcvalues.xgcvalues = &xgcvalues;

    _shTextBrushToShXgcValues(shTextBrush, cs, height, &shxgcvalues, quality);

    if (saveShXgc)
    {
	_pf.shxgc = saveShXgc;
	GModify(_pf.shxgc, &shxgcvalues);
    } else
	_pf.shxgc = GCreateShXgc(&shxgcvalues);

    // we need a block;
{
    void* t = this;
    _pf.xinterfs = GCreateShArray(&t);
}
    _pf.modified = GFALSE;
    _pf.clipok = GFALSE;
    _pf.tb.height = height;
    _pf.tb.I = cs.I;
    _pf.tb.J = cs.J;
    

    if (NPF == 0) 
    {
	((GSharedObject*)shTextBrush)->updatePostcond( ShPBIBisUpdated, _assocGC, GADD);
	((GSharedObject*)shTextBrush)->destroyPostcond( ShPBIBisDestroyed, _assocGC, GADD);
    }
    pf = &_pf;
    
    _return:
    
    /**
      handle the clip and return the GC
      **/
    if (_clipCount) 
    {
	_setGCClip();
	_clipCount = 0;
    }
    if (!pf->clipok || pf->clipOfXi != this) 
    {
	_setClip(pf->shxgc->gc());
	pf->clipok = GTRUE;
	pf->clipOfXi = this;
    }

    if (pf == &_pf)
	_assocGC->add(shTextBrush, sizeof(_pf), &_pf);

    if (shTextBrush->mask() & GFONT) 
    {
	*shxfont = _fontToShXFont(shTextBrush->font(), cs, height);
	assert (*shxfont);
    }
    else *shxfont = NULL ;

    return pf->shxgc->gc();
}


void XGoInterf::_freeXGCs()
{
    int nkeys;
    void** keys = _assocGC->getKeys(nkeys);
    void** _keys = keys;

    for(int i = 0 ; i < nkeys; i++, keys++)
    {
	int npf;
	GCAssoc* pf = (GCAssoc*)_assocGC->get(*keys, npf);
	assert (pf);
	int _npf = 0;
	int NPF = npf;
#ifdef NO_ALLOCA
	GCAssoc* _pf = (GCAssoc*)GMALLOC(npf * sizeof(GCAssoc));
#else
	GCAssoc* _pf = (GCAssoc*)alloca(npf * sizeof(GCAssoc));
#endif
	for (; npf; npf--, pf++)
	{
	    /**
	      remove the interface from the used list
	      if the list becomes NULL remove all this associations
	      **/
	    _remXInterfIfInside(pf->xinterfs);
	    if  (pf->xinterfs == NULL)
	    {
		/**
		  remove the callbacks
		  **/
		((GSharedObject*)*keys)->updatePostcond(ShPBIBisUpdated,
							_assocGC, GREMOVE);
		((GSharedObject*)*keys)->destroyPostcond(ShPBIBisDestroyed,
							 _assocGC, GREMOVE);
		/**
		  destroy the associated xgcs
		  **/
		pf->shxgc->destroy();
		/**
		  save the structure for destroy association step.
		  **/
		_pf[_npf++] = *pf;
	    }
	}
	if (_npf == NPF )
	    _assocGC->rem(*keys);
	else
	    for (NPF=0; NPF < _npf; NPF++, _pf++)
		_assocGC->rem(*keys, sizeof(GCAssoc), _pf);
#ifdef NO_ALLOCA
	GFREE (_pf);
#endif
    }
    /**
      free the keys list
      **/
    GFREE(_keys);

}
	

/**
  global assoc for stipple to xregion
  XRegion... functions do not need Display
  and so on.
**/
static GAssoc& AssocStipple() {
    static GAssoc* assoc_stipple = NULL;

    if (! assoc_stipple)
	assoc_stipple = new GAssoc();

    return *assoc_stipple;
}

struct StippleAssoc{
    Region  xregion;
    GCoordSys  cs;
    GShArray*  xinterfs;
};


static Region MakeStippleXRegion(const GPixelArrayValues* pav)
{
    XRectangle xrect;
    xrect.height = 1;
    int x_begin ;
    int maxX = (pav->width + 7)/8 * 8;
    int bpl = (pav->width+pav->xoffset + pav->padding -1) /pav->padding *pav->padding /8;
    Region xregion =  XCreateRegion();
    int x,y;
    
    if (pav->format.bit == GMSB_FIRST) 
	for (y = 0; y < pav->height; y++) 
	{
	    uchar* srcdata = pav->data + y*bpl;
	    x_begin = -1;
	    for (x = 0; x < maxX; x += 8) 
	    {
		uchar byte = *(srcdata + ((uint)(x+pav->xoffset))/8);
		int k = 0;
		if (x_begin == -1)
		    if (byte) {
			for (; k < 8; k++)
			    if (byte & (0x80 >> k)) 
			    {
				if (x + k < pav->width) 
				{
				    x_begin = x + k;
				    xrect.x = x_begin;
				}
				break;
			    }
		    }
		if (x_begin != -1)
		    if (byte != 0xFF)
		    {
			for (; k < 8; k++)
			    if ((byte & (0x80 >> k)) == 0)
			    {
				xrect.width = x + k - x_begin;
				xrect.y = y;
				XUnionRectWithRegion(&xrect, xregion,
						     xregion);
				x_begin = -1;
				break;
			    }
		    }
		    else if ( x == maxX - 8)
		    {
			xrect.width = maxX - x_begin;
			xrect.y = y;
			XUnionRectWithRegion(&xrect, xregion,
					     xregion);
		    }
	    }
	}
    else
	/** GLSB_FIRST **/
	for (y = 0; y < pav->height; y++)
	{
	    uchar* srcdata = pav->data + y * bpl;
	    x_begin = -1;
	    for (x = 0; x < maxX; x += 8)
	    {
		uchar byte = *(srcdata + ((uint)(x+pav->xoffset))/8);
		int k = 0;
		if (x_begin == -1)
		    if (byte)
		    {
			for (; k < 8; k++)
			    if (byte & (0x01 << k))
			    {
				if (x + k < pav->width)
				{
				    x_begin = x + k;
				    xrect.x = x_begin;
				}
				break;
			    }
		    }
		if (x_begin != -1)
		    if (byte != 0xFF)
		    {
			for (; k < 8; k++)
			    if ((byte & (0x01 << k)) == 0)
			    {
				xrect.width = x + k - x_begin;
				xrect.y = y;
				XUnionRectWithRegion(&xrect, xregion,
						     xregion);
				x_begin = -1;
				break;
			    }
		    }
		    else if ( x == maxX - 8)
		    {
			xrect.width = maxX - x_begin;
			xrect.y = y;
			XUnionRectWithRegion(&xrect, xregion,
					     xregion);
		    }
	    }
	}
    return xregion;
}


static void ShStippleIsDestroyed(void* _null, const GShPattern* stipple)
{
    int num ;
    assert (_null == NULL);
    StippleAssoc* pf = (StippleAssoc*)AssocStipple().get(stipple, num);

    XDestroyRegion(pf->xregion);
    
    /**
      remove: from the used list
      the association
      the callbacks
      **/
    pf->xinterfs->destroy();
    AssocStipple().rem(stipple);

}

static void ShStippleIsUpdated(void* _null,
			       GShPattern* stipple /* should be declare const */,
			       ulong /* mask unused */)
{
    ShStippleIsDestroyed(_null, stipple);
    /**
      remove the callbacks
      **/
    stipple->updatePostcond(ShStippleIsUpdated, _null, GREMOVE);
    stipple->destroyPostcond(ShStippleIsDestroyed, _null, GREMOVE);
}



Region   XGoInterf::_stippleToXRegion(const GShPattern* stipple,
				      const GCoordSys& cs) const
{
    const GPixelArrayValues* pav = stipple->pixelArray()->values();
    
    if (pav->depth != 1 || pav->bits_per_pixel != 1)
    {
	GWarning("XGoInterf::_shImageBrushToGC: %s", "bad depth for the stipple");
	return NULL;

    }
    else
    {
	int num;
	StippleAssoc* pf = (StippleAssoc*)AssocStipple().get(stipple, num);
	if (pf )
	{
	    assert (num == 1);
	    if (pf->cs == cs)
	    {
		_addXInterfIfNeeded(pf->xinterfs);
		return pf->xregion;
	    }
	    XDestroyRegion (pf->xregion);
	}

	StippleAssoc _pf;
	const void* t = this;


	/**
	  handle the stipple size
	  **/
	GPixelArrayValues PAV;
	ushort xorigin;
	ushort yorigin;
	
	_transformPixelArrayValues(&PAV, &xorigin, &yorigin, stipple->pixelArray(),
				   cs, /* pixel_translation */NULL,
				   1, 1, GMSB_FIRST, GMSB_FIRST, GMAX_QUALITY);
	_pf.cs = cs;
	
	/**
	  make the xregion
	  **/
	_pf.xregion = MakeStippleXRegion(&PAV);
	
	if (PAV.data != pav->data)
	    /**
	      free the PAV.data
	      **/
	    GFREE (PAV.data);
	
	_pf.xinterfs = GCreateShArray(&t);

	/**
	  make the association: It association ease the old one if any
	  **/
	AssocStipple().set(stipple, sizeof _pf, &_pf);
	/**
	  add postcond
	  **/

	if (pf == NULL)
	{
	    // 1er association;
	    stipple->updatePostcond( ShStippleIsUpdated, NULL, GADD);
	    stipple->destroyPostcond( ShStippleIsDestroyed, NULL, GADD);
	}
	return _pf.xregion;
    }
}




void XGoInterf::_freeStippleXRegion() const
{
    int nkeys;
    void** keys = AssocStipple().getKeys(nkeys);
    void** _keys = keys;

    for(int i = 0 ; i < nkeys; i++, keys++)
    {
	int npf;
	StippleAssoc* pf = (StippleAssoc*)AssocStipple().get(*keys, npf);
	assert (pf && npf == 1);

	_remXInterfIfInside(pf->xinterfs);
	if  (pf->xinterfs == NULL)
	{
	    // remove the callbacks;

	    ((GShPattern*)*keys)->updatePostcond(ShStippleIsUpdated,
						NULL, GREMOVE);
	    ((GShPattern*)*keys)->destroyPostcond(ShStippleIsDestroyed,
						 NULL, GREMOVE);
	    /**
	      destroy the associated xregion
	      **/
	    XDestroyRegion(pf->xregion);
	    AssocStipple().rem(*keys);
	}
    }
    /**
      free the keys list
      **/
    GFREE(_keys);
}


