#ifdef REV_INFO
static char rcsid[] = "$Id: XGoInterf.cc,v 1.9 1992/12/17 14:52:08 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  X INTERFACE
FILE:    XGoInterf.cc
AUTHOR:  Philippe Smadja, Jacques Davy
DATE:    August 90 (Version 1.0)

DESCRIPTION:
    The X Window System interface for Go.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/XGoInterf.h>
#include <assert.h>

#include "XGCP.h"
#include "XColorP.h"
#include "XFontP.h"
#include "XPatternP.h"
#include "XDashP.h"

#include <Go/Tools/Error.h>
#include <Go/Tools/StringAssoc.h>
#include <Go/Tools/Alloc.h>
#include <Go/Tools/Cursor.h>

#define XK_MISCELLANY
#define XK_LATIN1
#include <X11/keysymdef.h>
#include <X11/cursorfont.h>


#define EXPOSE_MODULO 3
#define UNHANDLED_MODULO 5
#define MAX_DISPLAY_NAME 64

static GStringAssoc& DpyAssoc() {
    static GStringAssoc* dpy_assoc = NULL;

    if (! dpy_assoc)
	dpy_assoc = new GStringAssoc();

    return *dpy_assoc;
}

struct DpyRef {
    Display* dpy;
    ushort   nref;
    GBool     dpyWasOpenedByMe;
};



void DisplayName(const char* dn, char* effective_dn)
{
    // remove the screen number from the display name (dn)
    // effective_dn must be large enougt to receive the result;
    
    const char* _dn = dn + strlen(dn) -1;
    while (_dn >= dn && *_dn != '.') _dn--;
    int len = _dn - dn;
    
    if (len <= 0)
    {
	// append ".0" to the end of the source name;
	char tmp[256];
	strcpy(tmp, dn), strcat(tmp, ".0");

	// call me again;
	DisplayName(tmp, effective_dn);
	return;
    }
    else
    {
	strncpy(effective_dn, dn, len);
	*(effective_dn+len)= '\0';
    }
}

static void addDisplay(Display* dpy)
{
    int num;

    char effective_display_name[MAX_DISPLAY_NAME];
    DisplayName(dpy->display_name, effective_display_name);
    
    DpyRef* dpyref = (DpyRef*)DpyAssoc().get(effective_display_name, num);

    if (dpyref && dpyref->dpy == dpy) 
    {
	dpyref->nref++;
	return ;
    }

    DpyRef _dpyref;
    _dpyref.nref = 1;
    _dpyref.dpy = dpy;
    _dpyref.dpyWasOpenedByMe = GFALSE;

    DpyAssoc().add(effective_display_name, sizeof _dpyref,  &_dpyref);
}    

static Display* getDisplay(const char* displayname)
{
    int num;
    DpyRef* dpyref;
    if (displayname == NULL)
    {
	displayname = getenv("DISPLAY");
	if (displayname == NULL)
	    GError("XGoInterf::%s", "no display\n");
    }

    char effective_display_name[MAX_DISPLAY_NAME];
    DisplayName(displayname, effective_display_name);
    
    if ((dpyref = (DpyRef*)DpyAssoc().get(effective_display_name, num)) != NULL) 
    {
	dpyref->nref++;
	return dpyref->dpy;
    }

    DpyRef _dpyref;

    _dpyref.nref = 1;
    _dpyref.dpy = XOpenDisplay(displayname);
    _dpyref.dpyWasOpenedByMe = GTRUE;
    
    if (_dpyref.dpy == NULL)
	GError("cannot open display \"%s\"", displayname);

    DpyAssoc().add(effective_display_name, sizeof _dpyref,  &_dpyref);
    return _dpyref.dpy;
}

static void freeDisplay(char* displayname)
{
    int num;

    char effective_display_name[MAX_DISPLAY_NAME];
    DisplayName(displayname, effective_display_name);
    
    DpyRef* dpyref = (DpyRef*)DpyAssoc().get(effective_display_name, num);
    assert (dpyref);

    for (int i = 0; i < num; i++, dpyref++)
	if (--dpyref->nref == 0)
	{
	    if (dpyref->dpyWasOpenedByMe)
		XCloseDisplay(dpyref->dpy);
	    DpyAssoc().rem(effective_display_name);
	}
}


ulong XGoInterf::GEventMaskToXEventMask (GEventMask gmask, ulong xmask_to_add)
{
    ulong xmask = xmask_to_add;

    // we get ButtonPress AND ButtonRelease  even user select only ;
    // ButtonPress or ButtonRelease;
    // because changing the X input mask after a press and before;
    // a release is impossible with X.
	
    if (gmask & (GDOUBLE_CLICK_MASK|GBUTTON_PRESS_MASK|GBUTTON_RELEASE_MASK))
	xmask |= (ButtonPressMask|ButtonReleaseMask);
    if (gmask & (GCHAR_MASK | GFUNCTION_KEY_MASK)) xmask |= KeyPressMask;
    if (gmask & GEXPOSE_MASK) xmask |= ExposureMask;

    /**
      GMOTION_MASK and GCOMPRESS_MOTION_MASK are exclusive
      **/
    if (gmask & GMOTION_MASK)  xmask |= PointerMotionMask;
    else if (gmask & GCOMPRESS_MOTION_MASK)
	xmask |= PointerMotionMask;

    return xmask;
}

XGoInterf::XGoInterf (const char* displayName) :
	_exposeEventRectangles(sizeof(XRectangle), EXPOSE_MODULO),
	_unHandledEvents(UNHANDLED_MODULO)
{
    _dpy = getDisplay(displayName);
    _keepDrawable = GFALSE;
    _isWindow = GTRUE;

    _screen = DefaultScreen(_dpy);
    _depth = DefaultDepth(_dpy, _screen);
    _colormap = DefaultColormap(_dpy, _screen);
    _visual = DefaultVisual(_dpy, _screen);
    

    _width = 150;
    _height = 100;
    /**
      DEFAULT: background white
      **/
    GRgb rgb(GMAX_RGB_VALUE, GMAX_RGB_VALUE,GMAX_RGB_VALUE);
    _background = new GColor (GCreateShColor(&rgb));

    Screen* screen = DefaultScreenOfDisplay(_dpy);
    _pixelWidth = ((GNumber)screen->mwidth) / screen->width;
    _pixelHeight = ((GNumber)screen->mheight) / screen->height;

    _drawable = XCreateSimpleWindow(_dpy, DefaultRootWindow(_dpy),
				    mm_to_pixx(100), mm_to_pixy(100),
				    mm_to_pixw(_width), mm_to_pixh(_height),
				    _depth, WhitePixelOfScreen(screen),
				    WhitePixelOfScreen(screen));

    _commonInit();
}

XGoInterf::XGoInterf (Display* display, Window window) :
	_exposeEventRectangles(sizeof(XRectangle), EXPOSE_MODULO),
	_unHandledEvents(UNHANDLED_MODULO)
{
    XWindowAttributes wa;
    XGetWindowAttributes(display, window, &wa);

    addDisplay (display);
    
    _dpy      = display;
    _keepDrawable = GTRUE;
    _screen   = wa.screen - _dpy->screens;
    _drawable = window;
    _isWindow = GTRUE;
    _depth    = wa.depth;
    _colormap = wa.colormap;
    if (_colormap == None)
	_colormap = DefaultColormap(_dpy,_screen);
    _visual   = wa.visual;

    Screen* screen = wa.screen;
    _pixelWidth = ((GNumber)screen->mwidth) / screen->width;
    _pixelHeight = ((GNumber)screen->mheight) / screen->height;

    _width = pixw_to_mm(wa.width);
    _height = pixh_to_mm(wa.height);

    _background = new GColor(pixel_to_color(wa.backing_pixel));

    _commonInit();
}

XGoInterf::XGoInterf (Display* display, int screen, Drawable drawable,
		      GBool isWindow,
		      int width_in_pixels, int height_in_pixels, int depth,
		      ulong backgroundPixel, Colormap colormap,
		      Visual* visual,
		      GNumber pixelWidth_in_mm                       /* = 0 */,
		      GNumber pixelHeight_in_mm                      /* = 0 */) :
	_exposeEventRectangles(sizeof(XRectangle), EXPOSE_MODULO),
	_unHandledEvents(UNHANDLED_MODULO)
{
    addDisplay (display);
    
    _dpy      = display;
    _keepDrawable = GTRUE;
    _screen   = screen;
    _drawable = drawable;
    _depth    = depth;
    _isWindow = isWindow;
    _colormap = colormap;
    _visual   = visual;

    if (pixelWidth_in_mm == 0 || pixelHeight_in_mm == 0)
    {
	Screen* screen = ScreenOfDisplay(_dpy, _screen);
	_pixelWidth = ((GNumber)screen->mwidth) / screen->width;
	_pixelHeight = ((GNumber)screen->mheight) / screen->height;
    }
    else
    {
	_pixelWidth = pixelWidth_in_mm;
	_pixelHeight = pixelHeight_in_mm;
    }


    _width = pixw_to_mm(width_in_pixels);
    _height = pixh_to_mm(height_in_pixels);

    if (backgroundPixel != GNO_BACKGROUND) 
	_background = new GColor(pixel_to_color(backgroundPixel));

    _commonInit();
}


void XGoInterf::_commonInit ()
{
    resetClip();
    _assocGC =      getAssocGC(_dpy, _screen, _depth, &_copyGC, &_clearGC, &_setGC);
    _assocColor =   getAssocColor(_dpy, _colormap, _visual);
    _assocFont  =   getAssocFont (_dpy);
    _assocPattern = getAssocPattern (_dpy, _screen, _colormap);
    _assocDash    = getAssocDash(_pixelWidth, _pixelHeight);
    _assocColorPattern = getAssocColorPattern(_dpy, _colormap);
    

    _xinputMask = 0;

    if (_background)
    {
	XGCValues values;
	values.foreground = _colorToXPixel(_background->shColor());
	_clearWindowGC = XCreateGC(_dpy, _drawable, GCForeground, &values);
    }

    _useColorPatternMatching = (_visual->map_entries <= 16);
    _waitingDouble = GFALSE;
    _grabPointer = GFALSE;
    _grabKeyboard = GFALSE;
    _removeNextReleaseEvent = GFALSE;
    _doubleClickInterval = 180; // ms;
    _captureInputMask = 0;
}

XGoInterf::~XGoInterf ()
{
    _freeXGCs();
    _freeXPixels();
    _freeXFonts();
    _freeXPatterns();
    _freeXDashes();
    _freeStippleXRegion();
    
    freeAssocGC     (_assocGC);
    freeAssocColor  (_assocColor);
    freeAssocFont   (_assocFont);
    freeAssocPattern(_assocPattern);
    freeAssocDash   (_assocDash);
    freeAssocColorPattern(_assocColorPattern);
    
    XFreeGC(_dpy, _clearWindowGC);

    if (_keepDrawable == GFALSE)
	XDestroyWindow(_dpy, _drawable);

    
    freeDisplay(_dpy->display_name);
}


void XGoInterf::_addXInterfIfNeeded(GShArray*& sha) const
{
    XGoInterf** data = (XGoInterf**)sha->data();
    for (int i = 0 ; i < sha->nelems(); i++, data++)
	if (*data == this)
	    return;
    const void* t = this;
    GAddElem(sha, &t);
}

void XGoInterf::_remXInterfIfInside(GShArray*& sha) const
{
    XGoInterf** data = (XGoInterf**)sha->data();
    for (int i = 0 ; i < sha->nelems(); i++, data++)
	if (*data == this)
	{
	    GRemElem(sha, i);
	    return;
	}
}

void XGoInterf::map ()
{
    if (_isWindow)
	XMapWindow(_dpy, _drawable);
}

void XGoInterf::unmap ()
{
    if (_isWindow)
	XUnmapWindow(_dpy, _drawable);
}

void XGoInterf::flush ()
{
    XFlush(_dpy);
}

void XGoInterf::synchronize(GBool discard)
{
    XSync(_dpy, discard);
}

GShColor*   XGoInterf::pixel_to_color(ulong pixel)
{
    XColor xcolor ;
    xcolor.pixel = pixel;
    XQueryColors(_dpy, _colormap, &xcolor, 1);

    GRgb rgb;
    rgb.red = xcolor.red;
    rgb.green = xcolor.green;
    rgb.blue = xcolor.blue;

    return GCreateShColor(&rgb);
}

GEvent* XGoInterf::xevent_to_gevent (const XEvent* xevent,
				     const XRectangle* xrects,
				     int  nxrects) const
{
    ulong mask = ModifierMaskFromXMask(xevent->xkey.state);

    switch (xevent->type)
    {
      case KeyPress: 
      {
	  GPoint pos(pixx_to_mm(xevent->xkey.x), pixy_to_mm(xevent->xkey.y));
	  GBool functionKey = GTRUE;
	  char buf[32];
	  int detail = 0; //inited to avoid a compilation warning;
	  KeySym keysym;
	  static XComposeStatus compose;
	  const XKeyEvent* xkey = &xevent->xkey;
	  int count = XLookupString((XKeyEvent*)xkey, buf,
				    sizeof buf, &keysym, &compose);

	  switch (keysym)
	  {
	    case XK_Left:   detail = GKEY_LEFT; break;
	    case XK_Right:  detail = GKEY_RIGHT; break;
	    case XK_Up:     detail = GKEY_UP; break;
	    case XK_Down:   detail = GKEY_DOWN; break;
	    case XK_Insert: detail = GKEY_INSERT; break;
	    case XK_Home:   detail = GKEY_HOME; break;
	    case XK_End:    detail = GKEY_END; break;
	    case XK_Prior:  detail = GKEY_PAGE_UP; break;
	    case XK_Next:   detail = GKEY_PAGE_DOWN; break;
	    case XK_F1:	    detail = GKEY_F1; break;
	    case XK_F2:	    detail = GKEY_F2; break;
	    case XK_F3:	    detail = GKEY_F3; break;
	    case XK_F4:     detail = GKEY_F4; break;
	    case XK_F5:     detail = GKEY_F5; break;
	    case XK_F6:     detail = GKEY_F6; break;
	    case XK_F7:     detail = GKEY_F7; break;
	    case XK_F8:     detail = GKEY_F8; break;
	    case XK_F9:     detail = GKEY_F9; break;
	    case XK_F10:    detail = GKEY_F10; break;
	    default:
	      functionKey = GFALSE;
	      if (count <= 0)
		  return NULL;
	      else
	      {
		  if (count > 1)
		      GError
		       ("XGoInterf.cc: internal error in xevent_to_goevent");

		  if (mask & GCONTROL)
		  {
		      buf[0] &= 0x1F;
		      mask &= ~GCONTROL;
		  }
		  mask &= ~GSHIFT;
		  return new GCharEvent(buf[0], mask, pos, this,
					XTimeToGTimeStamp(xkey->time));
	      }
	  }

	  if (functionKey)
	      return new GFunctionKeyEvent((GFunctionKey)detail, mask,
					   pos, this,
					   XTimeToGTimeStamp(xkey->time));
	  else
	      return NULL;
      }
      case ButtonPress:
      {
	  mask = ModifierMaskFromXMask(xevent->xbutton.state);
	  GPoint p(pixx_to_mm(xevent->xbutton.x), pixy_to_mm(xevent->xbutton.y));
	  return new GButtonPressEvent(xbutton_to_gbutton(xevent->xbutton.button),
				       mask, p, this,
				       XTimeToGTimeStamp(((XButtonEvent*)xevent)->time));
      }
      case ButtonRelease: {
	  mask = ModifierMaskFromXMask(xevent->xbutton.state);
	  GPoint p2(pixx_to_mm(xevent->xbutton.x), pixy_to_mm(xevent->xbutton.y));
	  return new GButtonReleaseEvent(xbutton_to_gbutton(xevent->xbutton.button),
					 mask, p2, this,
					 XTimeToGTimeStamp(((XButtonEvent*)xevent)->time));
      }
	
      case Expose: {
	  // GTimeStamp is unknown for expose event;
	  GNumber x, y;
	  if (xrects && nxrects)
	  {
	      GRegion region(nxrects);
	      for (int i = 0; i < nxrects; i++, xrects++)
	      {
		  x = pixx_to_mm(xrects->x);
		  y = pixy_to_mm(xrects->y);
		  region += GBox(x, y,
				x + pixw_to_mm(xrects->width),
				y + pixh_to_mm(xrects->height));
	      }
	      return new GExposeEvent(region, this);
	  }
	  x = pixx_to_mm(xevent->xexpose.x);
	  y = pixy_to_mm(xevent->xexpose.y);
	  GBox box(x, y,
		   x + pixw_to_mm(xevent->xexpose.width),
		   y + pixh_to_mm(xevent->xexpose.height));
	  return new GExposeEvent(box, this);
      }
      case MotionNotify: {
	  return new GMotionEvent(GPoint(pixx_to_mm(xevent->xmotion.x),
					 pixy_to_mm(xevent->xmotion.y)),
				  this,
				  XTimeToGTimeStamp(((XMotionEvent*)xevent)->time));
      }
      default:
	return NULL;
    }
}

void XGoInterf::clear (const GRegion& region)
{
    int nxrect;
    // here it is better to handle the clip in the client;

    GRegion reg;
    GIntersect(region, _clip, reg);

    XRectangle* xrect = (XRectangle*)reg.pixelBoxes(_pixelWidth,
						    _pixelHeight,
						    nxrect);
    if (_background) 
    {
	// ok with pixmap and window;
	XFillRectangles(_dpy, _drawable, _clearWindowGC,  xrect, nxrect);

    } else if (_isWindow)
    {
	// not that XClearArea will do nothing if the backgound is None;
	XRectangle* _xrect = xrect;
	for (int i = 0; i < nxrect; i++, _xrect++)
	    XClearArea(_dpy, _drawable,
		       _xrect->x, _xrect->y, _xrect->width, _xrect->height,
		       GFALSE);   // no exposure event;
    }
    
    GFREE(xrect);
}


void XGoInterf::addInputMask (GEventMask mask)
{
  GoInterf::addInputMask(mask);
    _xinputMask =  GEventMaskToXEventMask(inputMask());
    if (_isWindow)
    {
	XSelectInput(_dpy, _drawable, (long)_xinputMask);
	XFlush(_dpy);
    }
}

void XGoInterf::removeInputMask (GEventMask mask)
{
  GoInterf::removeInputMask(mask);
    _xinputMask =  GEventMaskToXEventMask(inputMask());
    if (_isWindow)
    {
	XSelectInput(_dpy, _drawable, (long)_xinputMask);
	XFlush(_dpy);
    }
}


void XGoInterf::resize(GNumber width_in_mm, GNumber height_in_mm)
{
    if (!_isWindow)
    {
	GWarning("XGoInterf::%s", "resize of pixmap unavailable");
	return;
    }
    
    int old_width_in_pixel = mm_to_pixw(_width);
    int old_height_in_pixel = mm_to_pixh(_height);

    GoInterf::resize(width_in_mm, height_in_mm);

    int width_in_pixel = mm_to_pixw(_width);
    int height_in_pixel = mm_to_pixh(_height);

    if (width_in_pixel != old_width_in_pixel ||
	height_in_pixel != old_height_in_pixel)
	    XResizeWindow(_dpy, _drawable,
			  width_in_pixel, height_in_pixel);
}
	    
void XGoInterf::resize(int width_in_pixel, int height_in_pixel)
{
    resize(pixw_to_mm(width_in_pixel), pixh_to_mm(height_in_pixel));
}

void XGoInterf::background(const GShColor* shcolor)
{
    if (!_isWindow)
    {
	GWarning("XGoInterf::%s", "background of pixmap unavailable");
	return;
    }
    
    if (shcolor == NULL)
    {
	if (_background)
	{
	    delete _background;
	    _background = NULL;
	    XSetWindowBackgroundPixmap(_dpy, _drawable, None);
	}
	return ;
    }
    
    if (_background && shcolor == _background->shColor())
	return;
    
    GoInterf::background(shcolor);

    Pixel pixel = _colorToXPixel(shcolor);

    if (pixel == BAD_PIXEL)
    {
	GWarning("XGoInterf::%s", "background() bad color");
	return;
    }
    XSetWindowBackground(_dpy, _drawable, pixel);

    /**
      change GC for clear function
      **/
    XGCValues values;
    values.foreground = pixel;
    XChangeGC(_dpy, _clearWindowGC, GCForeground, &values);
    /**
      exec the clear to show the new color
      **/
    XClearArea(_dpy, _drawable, 0, 0, 32000, 32000, GTRUE);

}


void XGoInterf::background(ulong pixel)
{
    background(pixel_to_color(pixel));
}


static uint CursorToXCursorShape(long cursorIdent)
{
    switch (cursorIdent)
    {
      default:
	return (uint)-1;

      case GCURSOR_DEFAULT:
	return XC_top_left_arrow;  // cf motif style guide;
	    
      case GCURSOR_CROSS:
	return XC_X_cursor;
	
      case GCURSOR_CROSSHAIR:
	return XC_crosshair;
	
      case GCURSOR_ARROW_4D:
	  return XC_fleur;
	
      case GCURSOR_RESIZE_SOUTH:
	return XC_bottom_side;

      case GCURSOR_RESIZE_NORTH:
	return XC_top_side;

      case GCURSOR_RESIZE_EAST:
	  return XC_right_side;

      case GCURSOR_RESIZE_WEST:
	  return XC_left_side;

      case GCURSOR_RESIZE_NORTH_EAST:
	return XC_top_right_corner;
	
      case GCURSOR_RESIZE_SOUTH_EAST:
	return XC_bottom_right_corner;
	
      case GCURSOR_RESIZE_SOUTH_WEST:
	return XC_bottom_left_corner;
	
      case GCURSOR_RESIZE_NORTH_WEST:
	return XC_top_left_corner;

      case GCURSOR_WATCH:
	  return XC_watch;

      case GCURSOR_HOURGLASS:
	return XC_clock;

      case GCURSOR_QUESTION:
	return XC_question_arrow;

      case GCURSOR_IBEAM:
	return XC_xterm;

      case GCURSOR_UPPER_LEFT_ARROW:
	return XC_top_left_arrow;
	
      case GCURSOR_UPPER_RIGHT_ARROW:
	return XC_draft_small;

      case GCURSOR_EAST_WEST_ARROW:
	return XC_sb_h_double_arrow;
	
      case GCURSOR_NORTH_SOUTH_ARROW:
	return XC_sb_v_double_arrow;
    }
}

struct AssocCursor {
    Cursor xcursor;
    uint nref;
};


void XGoInterf::pushCursor(long cursorIdent)
{
    if (!_isWindow)
	return;
    
    int num;
    AssocCursor* pf = (AssocCursor*)_assocCursor.get((void*)cursorIdent, num);
    if (pf )
	pf->nref++;
    else
    {
	AssocCursor _pf;
	uint shape = CursorToXCursorShape(cursorIdent);
	if (shape == (uint)-1)
	    _pf.xcursor = None;
	else
	    _pf.xcursor = XCreateFontCursor(_dpy, shape);
	_pf.nref = 1;
	_assocCursor.add((void*)cursorIdent, sizeof _pf, &_pf);
	pf = &_pf;
    }


    _cursorStack.push((void*)cursorIdent);
    XDefineCursor(_dpy, _drawable, pf->xcursor);
    XFlush(_dpy);
}    



void XGoInterf::popCursor()
{
    if (!_isWindow || _cursorStack.empty())
	return;
    
    int num;

    /**
      pop the current and free it
      **/
    long  cursorIdent         = (long)_cursorStack.pop();
    AssocCursor* pf = (AssocCursor*)_assocCursor.get((void*)cursorIdent, num);
    assert (pf && num == 1);
    if (--pf->nref == 0)
    {
	if (pf->xcursor != None)
	    XFreeCursor(_dpy, pf->xcursor);
	_assocCursor.rem ((void*)cursorIdent);
    }

    Cursor xcursor;
    
    if (_cursorStack.empty()) 
	xcursor = None;
    else
    {
	// get the top of stack;

	cursorIdent = (long)_cursorStack.top();
	pf = (AssocCursor*)_assocCursor.get((void*)cursorIdent, num);
	assert (pf && num == 1);
	xcursor = pf->xcursor;
    }

    XDefineCursor(_dpy, _drawable, xcursor);
    XFlush(_dpy);
}


void XGoInterf::bell(int percent)
{
    XBell(_dpy, percent);
    XFlush(_dpy);
}


/**
FUNCTION
    void useColorPatternMatching(GBool)
DESCRIPTION
    sets the useColorPatternMatching flag to the parameter value.
**/
void   XGoInterf::useColorPatternMatching (GBool flag)
{
    _useColorPatternMatching = flag;
}


