#ifdef REV_INFO
static char rcsid[] = "$Id: XImage.cc,v 1.6 1992/06/29 08:06:57 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  X INTERFACE
FILE:    XImage.cc
AUTHOR:  Philippe Smadja
DATE:    August 1990

DESCRIPTION:
    Image management in the X Window environment

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/XGoInterf.h>
#include <assert.h>
#include "XPrivate.h"
#include "XColorP.h"

#include <Go/Tools/ShPattern.h>
#include <Go/Tools/Error.h>
#include <Go/Tools/Alloc.h>

static GPen& MinQualityPen() {
    static GPen* min_quality_pen = NULL;

    if (! min_quality_pen)
	min_quality_pen = new GPen;

    return *min_quality_pen;
}


void XGoInterf::image (const GShImageBrush* shimagebrush, const GCoordSys& cs,
		       const GPoint& pos, GNumber w, GNumber h, const GShPixelArray* pa,
		       GQuality quality)
{
    if (quality == GMIN_QUALITY)
    {
	MinQualityPen().function(shimagebrush->function());
	GoInterf::rectangle (&MinQualityPen(), NULL, cs, pos, w, h, quality);
	return;
    }

    
    Pixmap pixmap;
    ushort pixmap_width;
    ushort pixmap_height;
    ushort pixmap_xorigin;
    ushort pixmap_yorigin;
    const  ShXRegion* shxregion;
    uchar  depth = pa->values()->depth;

    GCoordSys imgCS;
    imgCS.I = cs.I;
    imgCS.J = cs.J;

    GPoint p1 = cs.to(pos);

    if (cs.isOrthogonal())
    {
	GPoint p2 = cs.to(GPoint(pos.x + w, pos.y + h));

	GNumber pw = pixelWidth();
	GNumber ph = pixelHeight();
	p1.x = GPositionToPixel(p1.x, pw) * pw;
	p2.x = GPositionToPixel(p2.x, pw) * pw;

	p1.y = GPositionToPixel(p1.y, ph) * ph;
	p2.y = GPositionToPixel(p2.y, ph) * ph;
	

	GNumber _W = fabs(p2.x - p1.x);
	GNumber _H = fabs(p2.y - p1.y);

	if (fabs(imgCS.I.y) <= 0.001)      // see orthogonality criterum for a GCoordSys;
	{
	    imgCS.J.x = imgCS.I.y = 0;

	    w = _W, h = _H;
	    imgCS.I.x = imgCS.I.x > 0 ? 1 : -1;
	    imgCS.J.y = imgCS.J.y > 0 ? 1 : -1;
	}
	else
	{
	    imgCS.I.x = imgCS.J.y = 0;

	    w = _H, h = _W;
	    imgCS.I.y = imgCS.I.y > 0 ? 1 : -1;
	    imgCS.J.x = imgCS.J.x > 0 ? 1 : -1;
	}
    }
    
    _shPatternToX(pa, (depth == 1) ? NULL : shimagebrush->colormap(), w, h,
		  imgCS, quality,
		  &pixmap, &shxregion, &pixmap_width, &pixmap_height,
		  &pixmap_xorigin, &pixmap_yorigin);
    
    if (pixmap) 
    {

	int xpos = mm_to_pixx(p1.x);
	int ypos = mm_to_pixy(p1.y);

	xpos -= pixmap_xorigin;
	ypos -= pixmap_yorigin;

	imgCS.I.resize(mm_to_pixw (imgCS.I.size() * w));
	imgCS.J.resize(mm_to_pixh (imgCS.J.size() * h));
    
	GC gc = _shImageBrushToGC(shimagebrush, imgCS,
				  depth, shxregion,
				  xpos, ypos, quality);

	if (depth == 1)
	    XCopyPlane(_dpy, pixmap, _drawable, gc, 0, 0,  pixmap_width, pixmap_height,
		       xpos, ypos, 1);
	else
	    XCopyArea(_dpy, pixmap, _drawable, gc, 0, 0,  pixmap_width, pixmap_height,
		      xpos, ypos);
    }
}


/**
FUNCTION
    getImage()
DESCRIPTION
    The function reads the image inside the _output according the
    CS, width and height parameters.

    If the cs is not orthogonal, the result is NOT DEFORMED, the function
    just initializes the GImageBrushValues.stipple with a computed stipple mask.
    
    Note that all the fields of the "values" are set to  NULL  for unused pointer.

    The pixelarrayvalues_result.width, height, data may be NULL if the wanted image
    is not inside the drawable.
    
    It initializes the  GImageBrushValues with the stipple needed to display
    the image. If the GCoordSys is orthogonal the stipple of the GImageBrushValues
    will be NULL.

    It also reads the colormap  of the _output to initialize the colormap field
    of the  GImageBrushValues. If compress_colormap is GTRUE, the function will
    make the colormap so small as possible and will update the pixelarray data
    if needed.

    It also handle the x,y resolution (in DOT/GINCH). If the resolutions are 0
    so resize of the image is done else resize() is done according the wanted resolution
    and the current device one.
    The GQuality is only used here for the resize function.
**/

void  XGoInterf::getImage (GImageBrushValues& imagebrushvalues_result,
			   GPixelArrayValues& pixelarrayvalues_result,
			   const GCoordSys& cs, const GPoint& origin,
			   GNumber width, GNumber height, ushort depth,
			   GBool compress_colormap /* = GTRUE */, 
			   ushort x_resolution    /* = 0: drawable x_resol */,
			   ushort y_resolution    /* = 0: drawable y_resol */,
			   GQuality quality        /* = GmaxQuality */)
{
    
    /**
      PLANE
      **/
    imagebrushvalues_result.function = GCOPY;
    imagebrushvalues_result.plane_mask = GALL_PLANES;
    imagebrushvalues_result.fill.style = GFILL_SOLID;

    if (depth == 0) depth = _depth;

    /**
      CS for STIPPLE and XGetImage
      ****************************
      **/
    GCoordSys CS = GCoordSys(GPoint(0,0), cs.I * mm_to_pixw(width),
			   cs.J * mm_to_pixh(height));
    CS_TO_INT(CS);
    
    GBox box = CS.to(GBox(0,0,1,1));
    
    int box_width = (int)box.width();
    int box_height = (int)box.height();

    if (box_width == 0 || box_height == 0) 
    {
	imagebrushvalues_result.stipple = NULL;
	imagebrushvalues_result.colormap  = NULL;
 	pixelarrayvalues_result.width = 0;
	pixelarrayvalues_result.height = 0;
 	pixelarrayvalues_result.data = NULL;
	return;
    }
    
    /**
      IMAGEBRUSH.STIPPLE
      ******************
      **/
    if (CS.isOrthogonal()) 
	imagebrushvalues_result.stipple = NULL;
    else 
    {
	Pixmap stipple = XCreatePixmap(_dpy, _drawable, box_width, box_height, 1);

	XPoint xpts[5];
	for (int i=0, k=0; i < 2; i++) 
	    for (int j=0; j < 2; j++) 
	    {
		GPoint absp = CS.to(GPoint(i,j));
		xpts[k].x = (int)(absp.x - box.xmin);
		xpts[k++].y = (int)(absp.y - box.ymin);
	    }
	
	XPoint tmp = xpts[2];
	xpts[2]= xpts[3];
	xpts[3] = tmp;
	xpts[4] = xpts[0];

	XFillRectangle(_dpy, stipple, _clearGC, 0, 0, box_width, box_height);
	XFillPolygon(_dpy, stipple, _setGC, xpts, 5, Complex, CoordModeOrigin);
	XImage* ximage = XGetImage(_dpy, stipple, 0, 0, box_width, box_height,
				   1, XYPixmap);
	XFreePixmap(_dpy, stipple);
	
	GPatternValues patternValues;
	GPixelArrayValues pav;

	_ximageToPixelArrayValues(ximage, &pav);

	/**
	  do not worry about the colormap for depth 1
	  **/
	patternValues.pixelArray = GCreateShPixelArray(GPXL_VALUE|GPXL_SHARED, &pav);

	/**
	  set pattern width and height in mm.
	  **/
	patternValues.width = pixw_to_mm( box_width);   
	patternValues.height = pixh_to_mm(box_height);
	/**
	  free management
	  **/
	if (patternValues.pixelArray->refcnt() > 1) /** free the data **/
	    XFree(ximage->data);

	ximage->data = NULL;  /* in all case free the ximage struct **/
	XDestroyImage(ximage);
	/**
	  create stipple for the ImageBrush_result.
	  **/
	imagebrushvalues_result.fill.style = FillStippled;
	imagebrushvalues_result.stipple =
	    GCreateShPattern(GPTRN_PIXEL_ARRAY| GPTRN_WIDTH|
				GPTRN_HEIGHT, &patternValues);

	/** the pixel_aray is got by the pattern so destroy it **/
	assert (patternValues.pixelArray->refcnt() > 1);
	patternValues.pixelArray->destroy();
    }


    /**
      PIXEL_ARRAY_VALUES ( XGetImage)
      ******************************
      **/
    GPoint absp(cs.to(origin));
    int pixmap_xorigin = (int)-box.xmin;
    int pixmap_yorigin = (int)-box.ymin;
    int xpos = mm_to_pixx(absp.x) - pixmap_xorigin;
    int ypos = mm_to_pixy(absp.y) - pixmap_yorigin;

    /**
      confine the rectangle inside the drawable
      **/
    if (xpos < 0 )
    {
	GWarning ("XGoInterf::getImage(): cannot get image outside the drawable %c=0 is assumed", 'x');
	box_width += xpos, xpos = 0;
    }
    if (ypos < 0 )
    {
	GWarning ("XGoInterf::getImage(): cannot get image outside the drawable %c=0 is assumed", 'y');
	box_height += ypos, ypos = 0;
    }
    
    int size = mm_to_pixw(_width);
    if (xpos + box_width > size) box_width = size - xpos;

    size = mm_to_pixh(_height);
    if (ypos + box_height > size) box_height = size - ypos;

    if (box_width <= 0 || box_height <= 0)
    {
	if (imagebrushvalues_result.stipple)
	{
	    imagebrushvalues_result.stipple->destroy();    
	    imagebrushvalues_result.stipple = NULL;
	}
	imagebrushvalues_result.colormap = NULL;
	pixelarrayvalues_result.width =0;
	pixelarrayvalues_result.height =0;
	pixelarrayvalues_result.data = NULL;
	return;
    }
    
    /**
      read the image
      **/
    XImage* ximage = XGetImage(_dpy, _drawable, xpos, ypos, box_width, box_height,
			       imagebrushvalues_result.plane_mask, ZPixmap);

    _ximageToPixelArrayValues(ximage, &pixelarrayvalues_result);

    /**
      MAKE COLORMAP
      *************
      **/
    XColor* xcolors;
    int nxcolors;
    GBool free_xcolors;

    if (ximage->depth != 1  || _depth == 1) 
	getXScreenColor(_dpy, _screen, _colormap, _visual,
			&xcolors, &nxcolors,
			&free_xcolors);
    else
    {
	xcolors = NULL;
	nxcolors = 0;
	free_xcolors = GFALSE;
    }

	
    if (compress_colormap && ximage->depth != 1) 
    {
	if (ximage->bits_per_pixel == 8)
	{
	    int x, y;
#ifdef NO_ALLOCA
	    Pixel* translation = (Pixel*)GMALLOC(nxcolors* sizeof(Pixel));
	    Pixel* pixel_number = (Pixel*)GMALLOC(nxcolors* sizeof(Pixel));
#else
	    Pixel* translation = (Pixel*)alloca(nxcolors* sizeof(Pixel));
	    Pixel* pixel_number = (Pixel*)alloca(nxcolors* sizeof(Pixel));
#endif
	    for (int i = 0; i < nxcolors; i++, translation++) 
		*translation = BAD_PIXEL;
	    translation -= nxcolors;
	    int nb_used = 0;
	
	    uchar *srcdata = (uchar*)ximage->data + ximage->xoffset;
	    for (y = 0 ; y < ximage->height; y++,
		 srcdata = (uchar*)ximage->data +
		 y * ximage->bytes_per_line + ximage->xoffset)
		for (x = 0 ; x < ximage->width; x++, srcdata++)
		    /**
		      translation the current pixel (it is used)
		      **/
		    if (translation[*srcdata] == BAD_PIXEL) 
		    {
			translation[*srcdata] =  nb_used;
			pixel_number[nb_used++] = *srcdata;

			if (nb_used == nxcolors)
			    goto _makeColormapWithXcolors;
		    }
	    /**
	      make the pixel translation inside the data
	      **/
	    srcdata = (uchar*)ximage->data + ximage->xoffset;
	    for (y = 0; y < ximage->height; y++,
		 srcdata = (uchar*)ximage->data +
		 y * ximage->bytes_per_line + ximage->xoffset)
		for (x=0; x < ximage->width; x++, srcdata++)
		    *srcdata = (uchar)translation[*srcdata] ;
	    /**
	      compress the xcolors array
	      **/
#ifdef NO_ALLOCA
	    XColor* used_xcolors = (XColor*)GMALLOC(nb_used * sizeof(XColor));
#else
	    XColor* used_xcolors = (XColor*)alloca(nb_used * sizeof(XColor));
#endif
	    for (y = 0 ; y < nb_used; y++)
		   used_xcolors[y] = xcolors[pixel_number[y]];

	    if (free_xcolors)
	    {
		free_xcolors = GFALSE;
		GFREE (xcolors);
	    }
	    /**
	      update xcolors for makeColormapWithXcolors
	      **/
	    xcolors = used_xcolors;
	    nxcolors = nb_used;
	    
#ifdef NO_ALLOCA
	    free_xcolors = GTRUE;
	    GFREE(translation);
	    GFREE(pixel_number);
#else
	    free_xcolors = GFALSE;
#endif
	}
	else
	    GWarning("Sorry: XGoInterf::getImage() compress colormap for bits_per_pixel != 1and 8 not implemented yet");
	
    }
    
    _makeColormapWithXcolors:
    /***********************/
    if (nxcolors == 0)
	imagebrushvalues_result.colormap = NULL;
    else 
    {
	GColormapValues values;
	values.n_color = nxcolors;
#ifdef NO_ALLOCA
	values.colors = (GShColor**)GMALLOC(nxcolors* sizeof(GShColor*));
#else
	values.colors = (GShColor**)alloca(nxcolors* sizeof(GShColor*));
#endif
	XColor* tmp = xcolors;
	for (int i = 0; i < nxcolors; i++, tmp++)
	{
	    GRgb rgb;
	    rgb.red = tmp->red;
	    rgb.green = tmp->green;
	    rgb.blue  = tmp->blue;
	    values.colors[i] = GCreateShColor(&rgb);
	}

	imagebrushvalues_result.colormap = GCreateShColormap(GCOLORMAP_COLORS, &values);
	/**
	  I do not need the color for this function so destroy the colors
	  **/
	for (i = 0 ; i < nxcolors; i++)
	    values.colors[i]->destroy();

	/**
	  free management: handled the NO_ALLOCA and
	  the normal free  color management
	  **/
	if (free_xcolors)
	    GFREE(xcolors);

#ifdef NO_ALLOCA
	GFREE(values.colors);
#endif
    }

    /**
      HANDLE THE RESOLUTION
      *********************/
    if (x_resolution == 0 && y_resolution == 0 )
	/**
	  nothing to do
	  **/
	;
    else
    {
	ushort xorigin;
	ushort yorigin;
	GNumber i_factor = x_resolution == 0 ? 1 : x_resolution / (1.0/pixelWidth())*25.4;
	GNumber j_factor = y_resolution == 0 ? 1 : y_resolution / (1.0/pixelHeight())*25.4;

	CS = GCoordSys(GPoint(0,0), GPoint(box_width * i_factor, box_height * j_factor));
	CS_TO_INT(CS);

	GShPixelArray* pxarray = GCreateShPixelArray(GPXL_VALUE, &pixelarrayvalues_result);
	_transformPixelArrayValues(&pixelarrayvalues_result, &xorigin, &yorigin,
				   pxarray, CS, NULL,
				   pixelarrayvalues_result.bits_per_pixel,
				   pixelarrayvalues_result.depth,
				   pixelarrayvalues_result.format.bit, 
				   pixelarrayvalues_result.format.byte, quality);
	assert (xorigin == 0 && yorigin == 0);
	pxarray->destroy();
    }
    
    /**
      free the ximage struct : not the data
      **/
    ximage->data = NULL;
    XDestroyImage(ximage);
}


void XGoInterf::_ximageToPixelArrayValues(const XImage* ximage, GPixelArrayValues* pav) const
{
    
    pav->format.coding = bitmapCoding();
    pav->width         = ximage->width;
    pav->height        = ximage->height;
    pav->xoffset       = ximage->xoffset;
    pav->data          = (uchar*)ximage->data;
    pav->format.byte   = (ximage->byte_order == LSBFirst) ? GLSB_FIRST : GMSB_FIRST;
    pav->format.bit    = (ximage->bitmap_bit_order == LSBFirst) ? GLSB_FIRST : GMSB_FIRST;
    pav->padding       = ximage->bitmap_pad;
    pav->bitmap_unit   = ximage->bitmap_unit;
    pav->depth         = ximage->depth;
    pav->bits_per_pixel = ximage->bits_per_pixel;
}
