#ifdef REV_INFO
static char rcsid[] = "$Id: XInput.cc,v 1.13 1992/12/17 14:52:19 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  X INTERFACE
FILE:    XInput.cc
AUTHOR:  Philippe Smadja
DATE:    August 90 (Version 1.0)

DESCRIPTION:
    The X Window System interface for Go.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 

  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/XGoInterf.h>
#include <Go/GoInDisp.h>
#include <Go/Tools/Error.h>
#include <assert.h>

#define XK_MISCELLANY
#define XK_LATIN1
#include <X11/keysymdef.h>

/**/
GPoint XGoInterf::getPointer (const GCoordSys* cs)
{
    if (!_isWindow)
	GError("XGoInterf::%getPointer() on a pixmap", 'g');

    Window root, child;
    int root_x, root_y, win_x, win_y;
    uint mask;
    if (!XQueryPointer(_dpy, _drawable, &root, &child,
		      &root_x, &root_y, &win_x, &win_y, &mask))
	win_x = win_y = 0;

    GPoint pos(pixx_to_mm(win_x), pixy_to_mm(win_y));
    return (cs == NULL ? pos : cs->from(pos));
}

/**/
void XGoInterf::setPointer (const GPoint& relPos, const GCoordSys* cs)
{
    if (!_isWindow)
	GError("XGoInterf::%cetPointer() on a pixmap", 's');
    

    GPoint pos;
    if (cs == NULL) pos = relPos;
    else 	    pos = cs->to(relPos);
    XWarpPointer(_dpy, None, _drawable, 0, 0, 0, 0,
		 mm_to_pixw(pos.x), mm_to_pixh(pos.y));
}

/**/
ulong XGoInterf::inputFd() { return _dpy->fd;}

/**/
/**/
GBool XGoInterf::captureEvent(GEventMask capture_input_mask,
			      GBool confined, GTimeStamp gtime)
{
    if (!_isWindow)
	return GFALSE;

    GBool grabbed_yet = GFALSE;
    Time time = GTimeStampToXTime(gtime);
    ulong xmask = GEventMaskToXEventMask(capture_input_mask, 0);
    if (xmask & KeyPressMask)
	if (XGrabKeyboard(_dpy, _drawable, False,
			  GrabModeAsync, GrabModeAsync, time) != GrabSuccess)
	    grabbed_yet = GTRUE;
	else
	{
	    xmask &= ~KeyPressMask;
	    _grabKeyboard = GTRUE;
	}

    if (!grabbed_yet && (xmask & PointerMotionMask))
	if (XGrabPointer(_dpy, _drawable, False,  (uint)xmask,
			 GrabModeAsync,  GrabModeAsync,
			 (confined) ? _drawable : None,
			 None, time) != GrabSuccess)
	    grabbed_yet = GTRUE;
	else
	    _grabPointer = GTRUE;

    if (!grabbed_yet)
    {
	_captureInputMask = capture_input_mask;
	return GTRUE;
    }
    return GFALSE;
    
}

void XGoInterf::unCaptureEvent(GTimeStamp gtime)
{
    if (_grabPointer || _grabKeyboard)
    {
	Time time = GTimeStampToXTime(gtime);
	_captureInputMask = 0;
	if (_grabPointer)
	{
	    _grabPointer = GFALSE;
	    XUngrabPointer(_dpy, time);
	}
	if (_grabKeyboard)
	{
	    _grabKeyboard = GFALSE;
	    XUngrabKeyboard(_dpy, time);
	}
	XFlush(_dpy);
    }
}


/**
FUNCTION
    _currentInputMask()
DESCRIPTION
    returns the current input mask (handle capture or not)
**/
GEventMask XGoInterf::_currentInputMask()
{
    if (_captureInputMask == 0)
	return inputMask();
    else
	return _captureInputMask;
}

/**
FUNCTION
    HandleUnhandedEvents()
DESCRIPTION

**/

void DoubleClickTimer(void* xinterf)
{
    ((XGoInterf*)xinterf)->_doubleClickTimer();
}

void XGoInterf::_doubleClickTimer()
{
   _waitingDouble = GFALSE;
   _handleUnhandledEvents();
}

void XGoInterf::_handleUnhandledEvents()
{
    while (_unHandledEvents.nelems())
    {
	GEvent* ev = (GEvent*)_unHandledEvents.data(0);
	_unHandledEvents.rem(0);
	if (!GInputDispatcher()->stopped())
	    handle(ev);
	delete ev;
    }
}

/**/
void XGoInterf::handleXEvent(const XEvent* xevent)
{
    if (!_waitingDouble)
    {
	// handle remainding GEvent(s) (if any);
	(void)_handleUnhandledEvents();
	
	// and continue by handling the current XEvent;
    }
    
    GEvent* gevent = NULL;
    switch (xevent->type)
    {
      case ConfigureNotify:
	/**
	  update my width and height and call the postconds;
	  **/
      GoInterf::resize(pixw_to_mm(xevent->xconfigure.width),
		       pixh_to_mm(xevent->xconfigure.height));
	break;
	      
      case Expose:
	/**
	  compress expose event
	  **/
	XRectangle xrect;

	xrect.x = xevent->xexpose.x;
	xrect.y = xevent->xexpose.y;
	xrect.width = xevent->xexpose.width;
	xrect.height = xevent->xexpose.height;
	    
	_exposeEventRectangles.add(&xrect,1);

	if (xevent->xexpose.count > 0)
	    return;
	/**
	  it is last Expose
	  **/
	gevent = xevent_to_gevent(xevent,
				  (const XRectangle*)_exposeEventRectangles.data(),
				  _exposeEventRectangles.nelems());
	_exposeEventRectangles.reset();
	break;

      case MotionNotify:
	if (_isWindow && _currentInputMask() &  GCOMPRESS_MOTION_MASK)
	{
	    XEvent tmp = *xevent;
	    xevent = &tmp;
	    // compress motion event;
	    while (XCheckTypedWindowEvent(_dpy, _drawable, MotionNotify,
					  &tmp));
	}
	// no break;
    
      default:
	    
	gevent = xevent_to_gevent(xevent);
	if (gevent)
	{
	    if (_removeNextReleaseEvent &&
		gevent->type() == GBUTTON_RELEASE_EVENT)
	    {
		assert (_waitingDouble == GFALSE);
		_removeNextReleaseEvent = GFALSE;
		delete gevent;
		return;
	    }
	    GEventMask mask = _currentInputMask();

	    if (mask & GDOUBLE_CLICK_MASK)
		if (!_waitingDouble)
		{
		    if (gevent->type() == GBUTTON_PRESS_EVENT)
		    {
			_unHandledEvents.add(gevent);
			_waitingDouble = GTRUE;	
			GInputDispatcher()->timerProc(DoubleClickTimer, this,
						      doubleClickInterval());
			return;
		    }
		}
		else
		{
		    // we are waiting for a double click;
		    // and the timer was not called;
		    
		    if (gevent->type() == GBUTTON_PRESS_EVENT &&
			((GButtonEvent*)gevent)->button() ==
			((GButtonEvent*)_unHandledEvents.data(0))->button())
		    {
			// we have a double click;

			delete gevent;
			ulong modifier_mask = ModifierMaskFromXMask(xevent->xbutton.state);
			GPoint p(pixx_to_mm(xevent->xbutton.x),
				 pixy_to_mm(xevent->xbutton.y));
			gevent = new
			    GDoubleClickEvent(xbutton_to_gbutton(xevent->xbutton.button),
					      modifier_mask, p, this,
					      XTimeToGTimeStamp(xevent->xbutton.time));

			// reset timer and remove (ignore) unHandledEvents;

			GInputDispatcher()->timerProc(DoubleClickTimer, this,
						      doubleClickInterval(),
						      GREMOVE);
			_waitingDouble = GFALSE;
			    
			for (int i = 0; i < _unHandledEvents.nelems(); i++)
			    delete (GEvent*)_unHandledEvents.data(i);
			_unHandledEvents.reset();

			_removeNextReleaseEvent = GTRUE;
		    }
		    else
		    {
			// append the current event;
			// in the un handled list;
			_unHandledEvents.add(gevent);
			return;
		    }
		}
	}
	break;
    }   

    if (!gevent)
    {
	// The event is unknown by Go;
	// CallXEventHandlers(xevent);
    }
    else
    {
	// The event is a Go one;
	if (!GInputDispatcher()->stopped())
	    handle(gevent);
	delete gevent;
    }
}

