#ifdef REV_INFO
static char rcsid[] = "$Id: XLine.cc,v 1.2 1992/12/10 14:04:37 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  X INTERFACE
FILE:    XLine.cc
AUTHOR:  Jacques Davy, Philippe Smadja
DATE:    August 1990

DESCRIPTION:
    Line or Fill drawing:

      1) Check the Xgc
      2) Apply the coordinate system transformation
      3) Draw the line, polyline, ellipse, rectangle or the filled respective
         function.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/XGoInterf.h>
#include "XPrivate.h"
#include "XGCP.h"

#ifdef NO_ALLOCA
#include <Go/Tools/Alloc.h>
#endif

/**
FUNCTION
    XGoInterf::line()
DESCRIPTION
    Draws a line between two points
**/

void XGoInterf::line (const GShPen* shpen, const GCoordSys& cs, const GPoint& a, const GPoint& b,
		      GQuality quality)
{
    GC gc = _shPenToGC(shpen, quality);

    GPoint absA = cs.to(a);
    GPoint absB = cs.to(b);

    int ax = mm_to_pixx(absA.x);
    int ay = mm_to_pixy(absA.y);
    int bx = mm_to_pixx(absB.x);
    int by = mm_to_pixy(absB.y);

    XDrawLine(_dpy, _drawable, gc, ax, ay, bx, by);
}

/**
FUNCTION
    XGoInterf::polyline()
DESCRIPTION
    Draws a polyline describe with the points list.
**/
void XGoInterf::polyline (const GShPen* shpen, const GShBrush* shbrush,
			  const GCoordSys& cs, int npoint, const GPoint* points,
			  GQuality quality)
{
#ifdef NO_ALLOCA
    XPoint* xpts;
    XPoint _xpts[32];
    if (npoint > 31)
	xpts = (XPoint*)GMALLOC(sizeof(XPoint) * (npoint+1));
    else xpts = _xpts;
#else
    XPoint* xpts  = (XPoint*)alloca(sizeof(XPoint) * (npoint+1));
#endif

    int np = 0;
    for (int i = 0; i < npoint; i++, np++)
    {
	GPoint absp(cs.to(points[i]));
	xpts[np].x = mm_to_pixx(absp.x);
	xpts[np].y = mm_to_pixy(absp.y);

	// handle identical points;
	if (np && xpts[np-1].x == xpts[np].x
	    && xpts[np-1].y == xpts[np].y)
	    np--;
    }

    if (shbrush)
    {
	GC gc = _shBrushToGC(shbrush, quality);
	XFillPolygon(_dpy, _drawable, gc, xpts, np, Complex, CoordModeOrigin);	
    }
    if (shpen)
    {
	GC gc = _shPenToGC(shpen, quality);
	XDrawLines(_dpy, _drawable, gc, xpts, np, CoordModeOrigin);
    }
#ifdef NO_ALLOCA
    if (npoint > 31)
	GFREE(xpts);
#endif
}

/**
FUNCTION
    XGoInterf::polygon().
DESCRIPTION
    Draws the polyline describe by the points list. The function closes the polyline to
    make the polygon.
**/
void XGoInterf::polygon (const GShPen* shpen, const GShBrush* shbrush,
			 const GCoordSys& cs, int npoint, const GPoint* points,
			 GQuality quality)
{
#ifdef NO_ALLOCA
    XPoint* xpts;
    XPoint _xpts[32];
    if (npoint > 31)
	xpts = (XPoint*)GMALLOC(sizeof(XPoint) * (npoint+1));
    else xpts = _xpts;
#else
    XPoint* xpts  = (XPoint*)alloca(sizeof(XPoint) * (npoint+1));
#endif

    int np = 0;
    for (int i = 0; i < npoint; i++, np++)
    {
	GPoint absp(cs.to(points[i]));
	xpts[np].x = mm_to_pixx(absp.x);
	xpts[np].y = mm_to_pixy(absp.y);

	// handle identical points;
	if (np && xpts[np-1].x == xpts[np].x
	    && xpts[np-1].y == xpts[np].y)
	    np--;
    }

    if (shbrush)
    {
	GC gc = _shBrushToGC(shbrush, quality);
	XFillPolygon(_dpy, _drawable, gc, xpts, np, Complex, CoordModeOrigin);	
    }
    if (shpen)
    {
	GC gc = _shPenToGC(shpen, quality);
	xpts[np] = xpts[0];
	XDrawLines(_dpy, _drawable, gc, xpts, np + 1, CoordModeOrigin);
    }
#ifdef NO_ALLOCA
    if (npoint > 31)
	GFREE(xpts);
#endif
}



/**
FUNCTION
    XGoInterf::rectangle()
DESCRIPTION
    Draws the rectangle with "X" only if the cs is "X" oriented else
    It asks the general case.
**/
void XGoInterf::rectangle (const GShPen* shpen, const GShBrush* shbrush,
			   const GCoordSys& cs,
			   const GPoint& origin,
			   GNumber width, GNumber height, GQuality quality)
{
    if (DEGREE(cs.I) != 0 || DEGREE(cs.J) != 90)
    {
	// general case;
      GoInterf::rectangle(shpen, shbrush, cs,  origin, width, height, quality);
	return;
    }

    GPoint absp(cs.to(origin));

    // X and W;
    // *******;
    
    GNumber _X = absp.x + cs.I.size() * width;
    int xmin = mm_to_pixx(absp.x);
    int xmax = mm_to_pixx(_X);
    int w = xmax - xmin;       // no last point;
    
    // Y and H;
    // *******;
    
    GNumber _Y = absp.y + cs.J.size() * height;
    int ymin = mm_to_pixy(absp.y);
    int ymax = mm_to_pixy(_Y);
    int h = ymax - ymin;      // no last point;

    // DRAW;
    // ****;
    
    if (shbrush)
    {
	GC gc = _shBrushToGC (shbrush, quality);
	if (shpen)
	    XFillRectangle(_dpy, _drawable, gc, xmin+1, ymin+1, w-1, h-1);
	else
	    XFillRectangle(_dpy, _drawable, gc, xmin, ymin, w, h);
    }
    if (shpen)
    {
	GC gc = _shPenToGC (shpen, quality);
	XDrawRectangle(_dpy, _drawable, gc, xmin, ymin, w, h);
    }
}	    

#define RadianToXArcAngle(r) ((int)((r)*180*64/PI))

/**
FUNCTION
    XGoInterf::ellipse()
DESCRIPTION
    Draws the ellipse arc with "X" only if the cs
    is "X" oriented else it asks the general case.
**/
void XGoInterf::ellipse (const GShPen* shpen, const GShBrush* shbrush,
			 const GCoordSys& cs,
			 const GPoint& center,
			 GNumber radiusX, GNumber radiusY,
			 GNumber angle1, GNumber angle2, GQuality quality)
{
    if (DEGREE(cs.I) != 0 || DEGREE(cs.J) != 90)
    {
      GoInterf::ellipse(shpen, shbrush, cs, center, radiusX, radiusY,
			angle1, angle2, quality);
	return;
    }

    GPoint absp = cs.to(center);

    GNumber _width = cs.I.size() * radiusX;
    int xmin = mm_to_pixx(absp.x - _width);
    int xmax = mm_to_pixx(absp.x + _width);

    int width =  xmax - xmin;

    GNumber _height = cs.J.size() * radiusY;
    int ymin = mm_to_pixy(absp.y - _height);
    int ymax = mm_to_pixy(absp.y + _height);

    int height = ymax - ymin;

    
    while (angle1 < 0 )
	angle1 += 2*PI;
    while (angle2 < 0)
	angle2 += 2*PI;

    while (angle2 < angle1) 
	angle2 += 2*PI;
    
    int start = -RadianToXArcAngle(angle1);
    int delta = -RadianToXArcAngle(angle2-angle1);

    if (shbrush)
    {
	GC gc = _shBrushToGC(shbrush, quality);
	XFillArc(_dpy, _drawable, gc, xmin, ymin, width, height, start, delta);
    }
    if (shpen)
    {
	GC gc = _shPenToGC(shpen, quality);
	XDrawArc(_dpy, _drawable, gc, xmin, ymin, width, height, start, delta);
    }
}
