#ifdef REV_INFO
static char rcsid[] = "$Id: XPattern.cc,v 1.9 1992/07/21 12:50:46 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  X INTERFACE
FILE:    XPattern.cc
AUTHOR:  Philippe Smadja
DATE:    Nov 90

DESCRIPTION:
    Pattern management in the X environment

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/XGoInterf.h>
#include <assert.h>
#include "XPrivate.h"
#include "XRegionP.h"

#include <Go/Tools/ShPattern.h>
#include <Go/Tools/ShArray.h>
#include <Go/Tools/Error.h>
#include <Go/Tools/Alloc.h>


/**
  PATTERN ASSOC TOOLS
  ****************
  **/
struct _dPattern {
    Display*   dpy;
    Colormap colormap;
    uchar      screen;
    GAssoc*     assoc;
    int        nref;
};

static GArray& DPatterns() {
    static GArray* d_patterns = NULL;

    if (! d_patterns)
	d_patterns = new GArray(sizeof(struct _dPattern));

    return *d_patterns;
}

void  freeAssocPattern(GAssoc* assoc)
{
   struct _dPattern* dPattern  = (struct _dPattern*)DPatterns().data();
   int nbdPatterns = DPatterns().nelems();
   
   for (int i = 0 ; i  < nbdPatterns; i++, dPattern++)
	if (dPattern->assoc == assoc)
	{
	    if ( --dPattern->nref == 0)
	    {
		delete assoc;
		DPatterns().rem(i);
	    }
	    return ;
	}
}


GAssoc* getAssocPattern(Display* dpy, uchar screen, Colormap colormap)
{
    struct _dPattern* dPattern  = (struct _dPattern*)DPatterns().data();
    int nbdPatterns = DPatterns().nelems();
    
    for (int i = 0 ; i  < nbdPatterns; i++, dPattern++)
	if (dPattern->colormap == colormap  && dPattern->screen == screen &&
	    dPattern->dpy == dpy)
	{
	    dPattern->nref++;
	    return dPattern->assoc;
	}

    struct _dPattern local;
    local.dpy = dpy;
    local.colormap  = colormap ;
    local.screen = screen;
    local.nref = 1;
    local.assoc = new GAssoc();

    DPatterns().add(&local);
    return local.assoc;
}


Pixmap XGoInterf::_stippleToXStipple(const GShPattern* stipple, GQuality quality)
{
    Pixmap pixmap;
    const ShXRegion* mask;

    if (stipple->pixelArray()->values()->depth != 1) 
	GError("XGoInterf::_stippleToXStipple  depth of the stipple != 1");

    _shPatternToX(stipple->pixelArray(), NULL,
		  stipple->width(), stipple->height(),
		  GCoordSys(), /* coordsys unit  */
		  quality,
		  &pixmap, &mask);
    assert (mask->region() == NULL);
    return pixmap;
}


Pixmap XGoInterf::_tileToXTile(const GShPattern* tile, GQuality quality)
{
    Pixmap pixmap;
    const ShXRegion* mask;

    _shPatternToX(tile->pixelArray(),
		  tile->colormap(),
		  tile->width(), tile->height(),
		  GCoordSys(), /* coordsys unit  */
		  quality,
		  &pixmap, &mask);
    assert (mask->region() == NULL);
    return pixmap;
}


struct PatternAssoc {
    Pixmap pixmap;
    ushort pixmap_width;
    ushort pixmap_height;
    ushort pixmap_xorigin;
    ushort pixmap_yorigin;
    ShXRegion* shXRegion;
    GShArray* xinterfs;
    GArray*   translation;
    const GShColormap* colormap;
};


/**
FUNCTION
    FreeAssociatedResources()
DESCRIPTION
   Note that const is here because *pf must not be modified.
**/
static void FreeAssociatedResources(Display* dpy, const PatternAssoc* pf)
{
    XFreePixmap(dpy, pf->pixmap);
    assert (pf->shXRegion->refcnt() >= 1);
    pf->shXRegion->destroy();
    delete pf->translation;		
}


/**
FUNCTION
    ShPixelArrayIsDestroyed
DESCRIPTION
    remove the association.  and free the xResource.
**/
static void ShPixelArrayIsDestroyed(void* a, const GShPixelArray* pixelArray)
{
    GAssoc* assoc = (GAssoc*)a;
    int npf;
    PatternAssoc* pf = (PatternAssoc*)assoc->get(pixelArray, npf);
    /**
      pf may be NULL because it may have been deleted
      by ShColormapIsDestroyed().
      **/
    if (npf)
    {
	assert (pf);
	XGoInterf** xi = (XGoInterf**)(pf->xinterfs->data());

	for (; npf; npf--, pf++)
	    FreeAssociatedResources((*xi)->display(), pf);

	// remove the association;
	assoc->rem(pixelArray);
    }
}

/**
FUNCTION
    ShPixelArrayIsUpdated
DESCRIPTION
    delete the association and remove the postconds.
**/

static void ShPixelArrayIsUpdated(void* assoc,
				  GShPixelArray* pixelArray /* should be declare const*/,
				  ulong /* mask unused */)
{
    ShPixelArrayIsDestroyed(assoc, pixelArray);
    /**
      remove the callbacks
      **/
    pixelArray->updatePostcond(ShPixelArrayIsUpdated, assoc, GREMOVE);
    pixelArray->destroyPostcond(ShPixelArrayIsDestroyed, assoc, GREMOVE);
}

/**
FUNCTION
    ShColormapIsDestroyed
DESCRIPTION
    remove the XPattern ssociation using the colormap.  and free the xResource.
**/
static void ShColormapIsDestroyed(void* a, const GShColormap* colormap)
{
    GAssoc* assoc = (GAssoc*)a;
    int nkeys;
    void** keys = assoc->getKeys(nkeys);
    void** _keys = keys;

    for (int i = 0; i < nkeys; i++, keys++)
    {
	int npf;
	PatternAssoc* pf = (PatternAssoc*)assoc->get(*keys, npf);
	assert (pf);
	XGoInterf** xi = (XGoInterf**)(pf->xinterfs->data());

	int _npf = 0;
	int NPF = npf;
#ifdef NO_ALLOCA
	PatternAssoc* _pf = (PatternAssoc*)GMALLOC(npf* sizeof(PatternAssoc));	
#else
	PatternAssoc* _pf = (PatternAssoc*)alloca(npf* sizeof(PatternAssoc));
#endif
	for (; npf; npf--, pf++)
	    if (pf->colormap == colormap) 
	    {
		FreeAssociatedResources((*xi)->display(), pf);
		/**
		  save the structure for destroy association step.
		  **/
		_pf[_npf++] = *pf;
	    }

	/**
	  remove the association
	  **/
    	if (_npf == NPF )
	    assoc->rem(*keys);
	else
	    for (NPF=0; NPF < _npf; NPF++, _pf++)
		assoc->rem(*keys, sizeof(PatternAssoc), _pf);
#ifdef NO_ALLOCA
	GFREE(_pf);
#endif
    }
    /**
      free the keys list
      **/
    GFREE(_keys);
}

/**
FUNCTION
    ShColormapIsUpdated
DESCRIPTION
    delete the association and remove the postconds.
**/

static void ShColormapIsUpdated(void* assoc,
				GShColormap* colormap /* should be declare const */,
				ulong /* mask not used */)
{
    ShColormapIsDestroyed(assoc, colormap);
    /**
      remove the callbacks
      **/
    colormap->updatePostcond(ShColormapIsUpdated, assoc, GREMOVE);
    colormap->destroyPostcond(ShColormapIsDestroyed, assoc, GREMOVE);
}


/**
FUNCTION
    ShPatternToX:
DESCRIPTION
    common function for _tileToXTile and _stippleToXStipple and image().
**/

void XGoInterf::_shPatternToX(const GShPixelArray* pixelArray,
			      const GShColormap*   colormap,
			      GNumber width, GNumber height,
			      const GCoordSys& cs,
			      GQuality quality,
			      Pixmap* pixmap_result, const ShXRegion** shxregion_result,
			      ushort* pixmap_width_result,
			      ushort* pixmap_height_result,
			      ushort* pixmap_xorigin_result,
			      ushort* pixmap_yorigin_result)
{
    int npf;
    int NPF ;
    PatternAssoc* pf =  (PatternAssoc*)_assocPattern->get(pixelArray, npf);
    NPF = npf;
    const GShColormap* localColormap = pixelArray->values()->depth != 1 ? colormap : NULL;

    GVector I = cs.I;
    if (width)
    {
	I.x = mm_to_pixx (I.x * width);
	I.y = mm_to_pixx (I.y * width);
    }
    else I.resize(pixelArray->values()->width);

    GVector J = cs.J;
    if (height)
    {
	J.x =  mm_to_pixy (J.x * height);
	J.y =  mm_to_pixy (J.y * height);
    }
    else
	J.resize(pixelArray->values()->height);


    GCoordSys lcs = GCoordSys(GPoint(0,0), I, J);
    CS_TO_INT(lcs);
    
    ShXRegion* shXRegion = GCreateShXRegion(lcs);

    if (shXRegion->width() == 0 || shXRegion->height() == 0)
    {
	/**
	  empty pattern: generaly empty coordsys
	  do not save anything in association and return NULL values
	  **/
	shXRegion->destroy();
	*pixmap_result = 0;
	*shxregion_result = NULL;
	return;
    }
	

    for(; npf; npf--, pf++) 
    {
	/**
	  we conserve  only  pixelArray->refcnt()  pixmap for a given pattern
	  **/
	if (pf->shXRegion != shXRegion)
	{
	    if (npf == NPF - pixelArray->refcnt() + 1)
	    {
		/**
		  free the associated resources
		  **/
		FreeAssociatedResources(_dpy, pf);
		/**
		  pf will be reused
		  **/
		break;
	    }
	    continue;	      /* verif is GFALSE ==> next pf */
	}

	    
	if (localColormap != pf->colormap)
	{
	    if (npf == NPF - pixelArray->refcnt() + 1)
	    {
		/**
		  free the associated resources
		  **/
		FreeAssociatedResources(_dpy, pf);
		/**
		  pf will be reused
		  **/
		break;
	    }
	    continue;	      /* verif is GFALSE ==> next pf */
	}
	    
	/**
	  the verif is OK
	  **/
	_addXInterfIfNeeded(pf->xinterfs);
	shXRegion->destroy();

	/**
	  RETURN the RESULTS
	  **/
	*pixmap_result = pf->pixmap ;
	*shxregion_result   = pf->shXRegion;
	if (pixmap_width_result)
	    *pixmap_width_result =  pf->pixmap_width;
	if (pixmap_height_result)
	    *pixmap_height_result = pf->pixmap_height;
	if (pixmap_xorigin_result)
	    *pixmap_xorigin_result = pf->pixmap_xorigin;
	if (pixmap_yorigin_result)
	    *pixmap_yorigin_result = pf->pixmap_yorigin;
	
	return ;
    }

    /**
      compute the pattern
      *******************
      **/
    
    GPixelArrayValues PAV;
    ushort pav_xorigin;
    ushort pav_yorigin;
    GArray* translation    = localColormap ?
	_colormapToXPixels(pixelArray, localColormap) : NULL;

    int wanted_Depth = (pixelArray->values()->depth == 1) ? 1 :  depth();
    int wanted_BitsPerPixel = (wanted_Depth == 1) ? 1 : (wanted_Depth + 7) / 8 * 8;

    /**
      now  get the PixelArrayValue.
      the resulting data is allocated (if needed) by the function.
      **/
    _transformPixelArrayValues(&PAV, &pav_xorigin, &pav_yorigin,
			       pixelArray, lcs, translation,
			       wanted_BitsPerPixel, wanted_Depth,  bitOrder(),
			       byteOrder(), quality);
    /**
      initialize a pixmap
      **/
    XImage ximage;
    _pixelArrayValuesToXImage(&PAV, &ximage);
    Pixmap pixmap ;
	
    pixmap = XCreatePixmap(_dpy, _drawable,
			   ximage.width, ximage.height,
			   ximage.depth);
    XPutImage(_dpy, pixmap,
	      ximage.format == XYPixmap ? _copyGC : DefaultGC(_dpy, _screen),
	      &ximage, 0, 0, 0, 0, ximage.width, ximage.height);

    if (ximage.data != (char*)pixelArray->values()->data)
    {
	/**
	  we do not use the image data yet.
	  so free it.
	  **/
	delete ximage.data;
    }
    
    /**
      Update the associations
      **/

    if (npf == 0)
    {
	PatternAssoc _pf;
	pf = &_pf;
    }

    /**
      else: (npf > 0) the last pf does not match that we want
      we will update it.
      **/

    pf->pixmap = pixmap;
    pf->pixmap_width =  ximage.width;
    pf->pixmap_height = ximage.height;
    pf->pixmap_xorigin = pav_xorigin;
    pf->pixmap_yorigin = pav_yorigin;
    pf->shXRegion = shXRegion;
    pf->translation = translation;
    pf->colormap = localColormap;

    void* t = this;
    pf->xinterfs = GCreateShArray(&t);
	
    if (npf == 0) 
    {
	//add the new association;
	_assocPattern->add(pixelArray, sizeof (PatternAssoc), pf);
    }
    else
    {
	/**
	  the current pf was upadted
	  free the pf from pf+1 to the last one.
	  **/
	pf++; npf--;
	if (npf)
	{
#ifdef NO_ALLOCA
	    PatternAssoc* tmp = (PatternAssoc*)GMALLOC(npf * sizeof(PatternAssoc));
#else
	    PatternAssoc* tmp = (PatternAssoc*)alloca(npf * sizeof(PatternAssoc));
#endif
	    for (int i = 0; npf; npf--, pf++)
	    {
		// free the associated resources;

		FreeAssociatedResources(_dpy, pf);
		/**
		  save the structure for destroy association step.
		  **/
		tmp[i++] = *pf;

	    }
	    while (i--)
		_assocPattern->rem(pixelArray, sizeof(PatternAssoc), &tmp[i]);
#ifdef NO_ALLOCA
	    GFREE(tmp);
#endif
	}
    }

    if (NPF == 0) 
    {
	// add postconds for the first association;

	pixelArray->updatePostcond(ShPixelArrayIsUpdated,
				   _assocPattern, GADD);
	pixelArray->destroyPostcond(ShPixelArrayIsDestroyed,
				    _assocPattern, GADD);
	if (localColormap)
	{
	    localColormap->updatePostcond(ShColormapIsUpdated,
					  _assocPattern, GADD);
	    localColormap->destroyPostcond(ShColormapIsDestroyed,
					   _assocPattern, GADD);
	}
    }

    /**
      return the results
      **/
    *pixmap_result = pixmap;
    *shxregion_result = shXRegion;
    
    if (pixmap_width_result)
	*pixmap_width_result =  ximage.width;
    if (pixmap_height_result)
	*pixmap_height_result =  ximage.height;
    if (pixmap_xorigin_result)
	*pixmap_xorigin_result = pav_xorigin;
    if (pixmap_yorigin_result)
	*pixmap_yorigin_result = pav_yorigin;
}


/**
FUNCTION
    freeXPatterns ()
DESCRIPTION
    Free all the XPatterns.
    The function is used with NULL at XGoInterf destruction.
**/

void XGoInterf::_freeXPatterns()
{
    int nkeys;
    void** keys = _assocPattern->getKeys(nkeys);
    void** _keys = keys;

    for (int i = 0; i < nkeys; i++, keys++) 
    {
	int npf;
	PatternAssoc* pf = (PatternAssoc*)_assocPattern->get(*keys, npf);
	assert (pf);
	int _npf = 0;
	int NPF = npf;
#ifdef NO_ALLOCA
	PatternAssoc* _pf = (PatternAssoc*)GMALLOC(npf* sizeof(PatternAssoc));	
#else
	PatternAssoc* _pf = (PatternAssoc*)alloca(npf* sizeof(PatternAssoc));
#endif
	for (; npf; npf--, pf++ )
	{
	    // remove this from xinterfs list;

	    _remXInterfIfInside(pf->xinterfs);
	    if (pf->xinterfs == NULL)
	    {
		// remove the association and free the Xresources;

		((GShPixelArray*)*keys)->updatePostcond(ShPixelArrayIsUpdated,
							_assocPattern, GREMOVE);
		((GShPixelArray*)*keys)->destroyPostcond(ShPixelArrayIsDestroyed,
							 _assocPattern, GREMOVE);

		if (pf->colormap)
		{
		    pf->colormap->updatePostcond(ShColormapIsUpdated,
						 _assocPattern, GREMOVE);
		    pf->colormap->destroyPostcond(ShColormapIsDestroyed,
						  _assocPattern, GREMOVE);
		}
		FreeAssociatedResources(_dpy, pf);

		// save the structure for destroy association step;

		_pf[_npf++] = *pf;
	    }
	}
	/**
	  remove the association
	  **/
	if (_npf == NPF )
	    _assocPattern->rem(*keys);
	else
	    for (NPF=0; NPF < _npf; NPF++, _pf++)
		_assocPattern->rem(*keys, sizeof(PatternAssoc), _pf);
#ifdef NO_ALLOCA
	GFREE (_pf);
#endif
    }
    /**
      free the keys list
      **/
    GFREE(_keys);
}
	    

/**
FUNCTION
    _pixelArrayValuesToXImage()
DESCRIPTION
    The function is just a translator: It does not compute any data.
**/
void XGoInterf::_pixelArrayValuesToXImage(const GPixelArrayValues* pav,
					  XImage* ximage) const
{
    assert (pav->format.coding == bitmapCoding());
    ximage->width = (int)pav->width;
    ximage->height = (int)pav->height;
    ximage->xoffset = pav->xoffset;
    ximage->format = (pav->depth == 1 && pav->bits_per_pixel == 1) ? XYPixmap : ZPixmap;
    ximage->data = (char*)pav->data;
    ximage->byte_order = pav->format.byte == GLSB_FIRST ? LSBFirst : MSBFirst;
    ximage->bitmap_unit = pav->bitmap_unit;
    ximage->bitmap_bit_order = pav->format.bit == GLSB_FIRST ? LSBFirst : MSBFirst;
    ximage->bitmap_pad = pav->padding;
    ximage->depth = pav->depth;
    ximage->bytes_per_line = (((pav->width + pav->xoffset) * pav->bits_per_pixel) +
			      pav->padding -1) / pav->padding * pav->padding / 8;
    ximage->bits_per_pixel = pav->bits_per_pixel;
    ximage->red_mask = 0;
    ximage->green_mask = 0;
    ximage->blue_mask = 0;
    ximage->obdata = NULL;
    bzero(&ximage->f, sizeof (ximage->f));
}

