#ifdef REV_INFO
static char rcsid[] = "$Id: GoCoordSys.cc,v 1.2 1992/12/10 13:58:58 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  Go Kernel
FILE:    GoCoordSys.cc
AUTHOR:  Oliver Jojic
DATE:    April 91

DESCRIPTION:
  This file contains methods that manage go coordinate system.         

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <Go/Go.h>
#include <Go/mGo.h>


/*
 * pre and post condition types:
 */
char goNrelcsPrecond[] = "relcsPrecond";
char goNabscsPrecond[] = "abscsPrecond";
char goNrelcsPostcond[] = "relcsPostcond";
char goNabscsPostcond[] = "abscsPostcond";


void Go::relcsPrecond
    (GBool proc(void*, const Go*, GCoordSys*, const GCoordSys*, const GCoordSys*),
     void* closure, GAddOrRemove add)
{
    has_relcs_precond =
	AddOrRemoveCondition(goNrelcsPrecond, proc, closure, add);
}

void Go::relcsPostcond
    (void cb(void*, Go*, const GCoordSys*), void* closure, GAddOrRemove add)
{
    has_relcs_postcond =
	AddOrRemoveCondition(goNrelcsPostcond, cb, closure, add);
}

void Go::abscsPrecond
    (GBool cb(void*, const Go*, GCoordSys*, const GCoordSys*, const GCoordSys*),
     void* closure, GAddOrRemove add)
{
    has_abscs_precond =
	AddOrRemoveCondition(goNabscsPrecond, cb, closure, add);
}

void Go::abscsPostcond
    (void cb(void*, Go*, const GCoordSys*), void* closure, GAddOrRemove add)
{
    has_abscs_postcond =
	AddOrRemoveCondition(goNabscsPostcond, cb, closure, add);
}

void GoMetaclass::relcsPrecond
    (GBool cb(void*, const Go*, GCoordSys*, const GCoordSys*, const GCoordSys*),
     void* closure, GAddOrRemove add)
{
    has_relcs_precond =
	AddOrRemoveCondition(goNrelcsPrecond, cb, closure, add);
}

void GoMetaclass::relcsPostcond
    (void cb(void*, Go*, const GCoordSys*),
     void* closure, GAddOrRemove add)
{
    has_relcs_postcond =
	AddOrRemoveCondition(goNrelcsPostcond, cb, closure, add);
}

void GoMetaclass::abscsPrecond
    (GBool cb(void*, const Go*, GCoordSys*, const GCoordSys*, const GCoordSys*),
     void* closure, GAddOrRemove add)
{
    has_abscs_precond =
	AddOrRemoveCondition(goNabscsPrecond, cb, closure, add);
}

void GoMetaclass::abscsPostcond
    (void cb(void*, Go*, const GCoordSys*),
     void* closure, GAddOrRemove add)
{
    has_abscs_postcond =
	AddOrRemoveCondition(goNabscsPostcond, cb, closure, add);
}



GCoordSys Go::coordsys (const Go* relto) const
{
    if (relto == NULL)
	return Abscs();
    else if (relto == parent())
	return Relcs();
    else
	return relto->Abscs().from(Abscs());
}


GBool Go::rotate (GNumber angle, const GPoint& center, Go* relto)
{
    GCoordSys cs(Abscs());

    if (relto == this)
	cs.rotate(center, angle);
    else
    {
	GPoint abscenter;
	if (relto == NULL)
	    abscenter = center;
	else
	    abscenter = relto->Abscs().to(center);

	cs.rotate(cs.from(abscenter), angle);
    }
    
    return setCoordsys(cs);
}


GCoordSys Go::Abscs (const GCoordSys& parabs) const
{
    if (d_store_abscs()) return d_get_abscs(&parabs);
    if (&parabs) return parabs.to(d_get_relcs(&parabs));

    GCoordSys pabs = parentCoordsys();
    return pabs.to(d_get_relcs(&pabs));
}

GCoordSys Go::Relcs (const GCoordSys& parabs) const
{
    if (d_store_relcs()) return d_get_relcs(&parabs);
    if (&parabs) return parabs.from(d_get_abscs(&parabs));
    GCoordSys pabs = parentCoordsys();
    return pabs.from(d_get_abscs(&pabs));
}



INLINE GBool Go::call_relcs_precond
    (GCoordSys& newrel, const GCoordSys& oldpabs, const GCoordSys& newpabs) const
{
    return (has_relcs_precond)
	? CallPreconditions(goNrelcsPrecond, &newrel,
			    (void*)&oldpabs, (void*)&newpabs)
	    : GTRUE;
}

GBool Go::call_class_relcs_precond
    (GCoordSys& newrel, const GCoordSys& oldpabs, const GCoordSys& newpabs) const
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_relcs_precond &&
	    !mc->CallPreconditions(goNrelcsPrecond, this, &newrel,
				   (void*)&oldpabs, (void*)&newpabs))
	    return GFALSE;

    return GTRUE;
}


/*
 * precondition on absolute coordinate system
 * is the only one that is void (and not bool).
 */
INLINE GBool Go::call_abscs_precond
    (GCoordSys& newabs, const GCoordSys& oldpabs, const GCoordSys& newpabs) const
{
    return (has_abscs_precond)
	? CallPreconditions(goNabscsPrecond, &newabs,
			    (void*)&oldpabs, (void*)&newpabs)
	    : GTRUE;
}

GBool Go::call_class_abscs_precond
    (GCoordSys& newabs, const GCoordSys& oldpabs, const GCoordSys& newpabs) const
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_abscs_precond &&
	    !mc->CallPreconditions(goNabscsPrecond, this, &newabs,
				   (void*)&oldpabs, (void*)&newpabs))
	    return GFALSE;

    return GTRUE;
}

INLINE void Go::call_relcs_postcond (const GCoordSys& parabs)
{
    if (has_relcs_postcond)
	CallPostconditions(goNrelcsPostcond, (void*)&parabs);
}

void Go::call_class_relcs_postcond (const GCoordSys& parabs)
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_relcs_postcond)
	    mc->CallPostconditions(goNrelcsPostcond, this, (void*)&parabs);
}

INLINE void Go::call_abscs_postcond (const GCoordSys& parabs)
{
    if (has_abscs_postcond)
	CallPostconditions(goNabscsPostcond, (void*)&parabs);
}

void Go::call_class_abscs_postcond (const GCoordSys& parabs)
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_abscs_postcond)
	    mc->CallPostconditions(goNabscsPostcond, this, (void*)&parabs);
}


/*
 * TODO:
 *
 *	A new SetCSRec is build on each go of this go subtree.
 *	But a SetCSRec is mandatory only on goes that:
 *	   - store a abscs, or
 *	   - have an abscs precond or postcond.
 *
 *	Therefore some optimization may be done.
 */
struct SetCSRec
{
    SetCSRec* next;
    Go* go;
    GCoordSys newabs;
    GCoordSys newrel;
    GCoordSys parabs;
    GBool call_relcs_postcond :1;
    GBool call_abscs_postcond :1;
    GBool TempMap_me :1;
};


GBool Go::set_abscs (const GCoordSys& newabs, GBool this_only)
{
    GCoordSys parabs(parentCoordsys());
    GCoordSys oldabs(Abscs(parabs));

    SetCSRec* list = new SetCSRec;
    list->next = NULL;
    list->go = this;
    list->newabs = newabs;
    list->newrel = parabs.from(newabs);
    list->parabs = parabs;
    list->call_relcs_postcond = GTRUE;
    list->call_abscs_postcond = GTRUE;
    
    GCoordSys ini_newrel(list->newrel);
    if (!call_relcs_precond(list->newrel, parabs, parabs) ||
	!call_class_relcs_precond(list->newrel, parabs, parabs))
    {
	delete list;
	return GFALSE;
    }
    if (list->newrel != ini_newrel)
	list->newabs = parabs.to(list->newrel);

    GCoordSys ini_newabs(list->newabs);
    if (!call_abscs_precond(list->newabs, parabs, parabs) ||
	!call_class_abscs_precond(list->newabs, parabs, parabs))
    {
	delete list;
	return GFALSE;
    }

    if (list->newabs != ini_newabs)
	list->newrel = parabs.from(list->newabs);

    // The Unmap/Map sequence will automatically set up clip regions.

    GBool was_mapped = mapped();

    
    if (was_mapped)
	if (!this_only)
	    TempUnmap();
	else
	{
	    GRegion region; m_add_clip(region);
	    RedrawBackAndFore(region);
	    for (Go* par = parent(); par; par = par->parent())
	    {
		if (par->clip_was_updated && !par->clip_is_ready)
		    break;
	 	par->clip_not_ready();
	    }
	}

    // must be after the _unmap:
    if (this_only)
	build_setthiscs_list_on_subtree(list, oldabs, newabs);
    else
	build_setcs_list_on_subtree(list, oldabs, newabs);

    
    for (SetCSRec* l = list; l; l = l->next)
	l->go->d_set_coordsys(l->newrel, l->newabs, l->parabs);

    // preserve lib for postconds;
    DifferDestroy();
    
    SetCSRec* nextl;
    for (l = list; l; l = nextl)
    {
	nextl = l->next;

	if (l->call_relcs_postcond)
	{
	    l->go->call_relcs_postcond(l->parabs);
	    l->go->call_class_relcs_postcond(l->parabs);
	}
	if (l->call_abscs_postcond)
	{
	    l->go->call_abscs_postcond(l->parabs);
	    l->go->call_class_abscs_postcond(l->parabs);
	}
	
	if (this_only && l->TempMap_me)
	    l->go->TempMap();

	delete l;
    }

    
    if (was_mapped)
	if (!this_only)
	    TempMap();
	else
	{
	    GRegion region; m_add_clip(region);
	    RedrawForeOnly(region);
	    for (Go* par = parent(); par; par = par->parent())
	    {
		if (par->clip_was_updated && !par->clip_is_ready)
		    break;
	 	par->clip_not_ready();
	    }
	}

    HandleDestroy();
    return GTRUE;
}


void Go::build_setthiscs_list_on_subtree
    (SetCSRec*& list, const GCoordSys& oldabs, const GCoordSys& newabs)
{
    my_clip_was_updated = GTRUE;
    clip_not_ready();
    GBool this_mapped = mapped();
    list->TempMap_me = GFALSE;

    Go* child = Child();
    for (int i = 0; child; child = child->Next(i++))
    {
	SetCSRec* childrec = new SetCSRec;
	childrec->next = list;
	list = childrec;
	childrec->go = child;
	childrec->parabs = newabs;
	childrec->newabs = child->Abscs(oldabs);
	childrec->newrel = newabs.from(childrec->newabs);
	childrec->call_relcs_postcond = GTRUE;
	childrec->call_abscs_postcond = GFALSE;
	childrec->TempMap_me = GFALSE;

	GCoordSys ini_newrel(childrec->newrel);
	if (child->call_relcs_precond(childrec->newrel, oldabs, newabs) &&
	    child->call_class_relcs_precond
	    	(childrec->newrel, oldabs, newabs) &&
	    childrec->newrel != ini_newrel)
	{
	    if (this_mapped && child->_mappable)
	    {
		child->TempUnmap();
		childrec->TempMap_me = GTRUE;
	    }

	    childrec->call_abscs_postcond = GTRUE;
	    GCoordSys child_oldabs(childrec->newabs);
	    childrec->newabs = newabs.to(childrec->newrel);
	    build_setcs_list_on_subtree
		(list, child_oldabs, childrec->newabs);
	}
    }
}



void Go::build_setcs_list_on_subtree
    (SetCSRec*& list, const GCoordSys& oldabs, const GCoordSys& newabs)
{
    my_clip_was_updated = GTRUE;
    clip_not_ready();

    Go* child = Child();
    for (int i = 0; child; child = child->Next(i++))
    {
	SetCSRec* childrec = new SetCSRec;
	childrec->next = list;
	childrec->go = child;
	childrec->parabs = newabs;
	GCoordSys oldchildabs = child->Abscs(oldabs);
	childrec->newrel = child->Relcs(oldabs);
	childrec->newabs = newabs.to(childrec->newrel);
	GCoordSys ini_childnewabs(childrec->newabs);
	childrec->call_relcs_postcond = GFALSE;
	childrec->call_abscs_postcond = oldchildabs != ini_childnewabs;
	
	if (childrec->call_abscs_postcond == GFALSE)
	{
	    delete childrec;
	    continue;
	}

	if (!child->call_abscs_precond(childrec->newabs, oldabs, newabs) ||
	    !child->call_class_abscs_precond(childrec->newabs,
					     oldabs, newabs))
	{
	    childrec->newabs = oldabs.to(childrec->newrel);
	    childrec->call_abscs_postcond = GFALSE;
	}

	if (childrec->newabs != ini_childnewabs)
	{
	    childrec->newrel = newabs.from(childrec->newabs);
	    childrec->call_relcs_postcond = GTRUE;
	}
	
	list = childrec;

	child->build_setcs_list_on_subtree(list,
					   oldchildabs, childrec->newabs);
    }
}


GBool Go::callCSPrecond (GCoordSys& newabs) const
{
    GCoordSys parabs(parentCoordsys());
    GCoordSys newrel(parabs.from(newabs));

    GCoordSys ini_newrel(newrel);

    if (!call_relcs_precond(newrel, parabs, parabs) ||
	!call_class_relcs_precond(newrel, parabs, parabs))
	return GFALSE;

    if (newrel != ini_newrel)
	newabs = parabs.to(newrel);

    if (!call_abscs_precond(newabs, parabs, parabs) ||
	!call_class_abscs_precond(newabs, parabs, parabs))
	return GFALSE;

    return GTRUE;
}
