#ifdef REV_INFO
static char rcsid[] = "$Id: GoRoot.cc,v 1.14 1992/12/21 13:24:42 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  Go Kernel
FILE:    GoRoot.cc
AUTHOR:  Oliver Jojic
DATE:    April 91

DESCRIPTION:
  See the GoRoot.h header.

  flush mechanism:
    Drawings of Gos are done only at flush time. The four methods
    "redraw_from_unmap", "redraw_from_map", "redrawBackAndFore" and
    "redrawForeOnly" only store informations inside goroot to be able
    to retrieve what must be drawn at flush time.

    The flush algorithm must:
      1) draw only what was asked for,
      2) do not call drawings that are hidden by opaque clips of upper Gos.

    An exact algorithm should set the clip region of the interface almost
    for each call to "go->m_draw()". But the setting of the clip region is
    expensive. We adopted a less precise algorithm (sometime gos are drawn
    even if they could not be drawn) but much more efficient:
    we set the clip region of the interface only one time per flush.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <Go/GoRoot.h>
#include <Go/mGoRoot.h>
#include <Go/GoInterf.h>
#include <Go/Tools/Error.h>


/*
 * pre and post condition types:
 */
char goNclipPostcond[] = "clipPostcond";
char goNmyClipPostcond[] = "myClipPostcond";


GAssoc* GoRoot::d_get_assoc () const
{
    return &((GoRoot*)this)->_assoc;
}

Go* GoRoot::d_get_parent () const
{
    return NULL;
}

Go* GoRoot::d_get_previous () const
{
    return NULL;
}

Go* GoRoot::d_get_next () const
{
    return NULL;
}

Go* GoRoot::d_get_first_child () const
{
    return _child;
}

GBool GoRoot::d_store_children_array () const
{
    return GFALSE;
}

void GoRoot::d_set_previous (const Go* previous)
{
    if (previous != NULL)
	GError("GoRoot::d_set_previous(): my previous brother must be NULL");
}

void GoRoot::d_set_next (const Go* next)
{
    if (next != NULL)
	GError("GoRoot::d_set_next(): my next brother must be NULL");
}

void GoRoot::d_add_child (const Go* ch, int index)
{
    if (index == 0 || _child == NULL)
	_child = (Go*)ch;
}

void GoRoot::d_remove_child (const Go* ch, int)
{
    if (ch == _child)
	_child = _child->Next(0);
}

GBool GoRoot::d_store_abscs () const
{
    return GTRUE;
}

void GoRoot::d_set_tree (const Go* parent, const Go* previous, const Go* next)
{
    if (parent != NULL || previous != NULL || next != NULL)
	GError(
	  "GoRoot::d_set_tree(): my parent, previous and next must be NULL");
}

GBool GoRoot::d_store_relcs () const
{
    return GFALSE;
}

const GCoordSys& GoRoot::d_get_abscs (const GCoordSys*) const
{
    return _cs;
}

void GoRoot::d_set_coordsys
    (const GCoordSys&, const GCoordSys& abscs, const GCoordSys&)
{
    _cs = abscs;
}

GRegion* GoRoot::d_get_clip () const
{
    return &((GoRoot*)this)->_clip;
}

GBool GoRoot::is_root () const
{
    return GTRUE;
}


GoMetaclass* GoRoot::m_metaclass () const
{
    return GoRoot::metaclass();
}

void GoRoot::m_draw (const GRegion& region) const
{
    _interf->clear(region);
}

void GoRoot::m_add_clip (GRegion&) const
{
}
	

void GoRoot::m_dup
    (const Go* src, Go* parcopy, Go* prevcopy, GBool do_next, GBool do_children)
{
    if (parcopy != NULL || prevcopy != NULL)
	GError("GoRoot::m_dup(): I cannot have a parent");

    Go::m_dup(src, parcopy, prevcopy, do_next, do_children);

    GoRoot* orig = (GoRoot*)src;

    if (_interf != orig->_interf)
    {
	_interf = orig->_interf;
	_interf->exposePostcond(root_expose_handler, this);
    }

    _child = NULL;
    if (do_children)
	for (Go* ch = orig->_child; ch; ch = ch->Next())
	    if (_child = ch->Dup(this, NULL, GTRUE, GTRUE))
		break;
    
    _cs = orig->_cs;
}


char goNflushPrecond[] = "flushPrecond";
char goNflushPostcond[] = "flushPostcond";

inline void GoRoot::call_flush_preconditions ()
{
    if (has_flush_precond)
	CallPostconditions (goNflushPrecond, &full_drawing_region);
}

inline void GoRoot::call_flush_postconditions ()
{
    if (has_flush_postcond)
	CallPostconditions (goNflushPostcond, &full_drawing_region);
}


GoRoot* GoRoot::create (GoInterf* interf)
{
    GoRoot* goroot = GoRoot::create();
    if (goroot)
	goroot->interf(interf);
    return goroot;
}


GoInterf* Go::interf () const
{
    GoRoot* goroot = root();
    return (goroot) ? goroot->interf() : GoDefaultGoInterf();
}


GoRoot::GoRoot ()
{
    use_drawing_box = GFALSE;
    _interf = GoDefaultGoInterf();
    _child = NULL;
    min_opaque_square_area = 250;
    max_boxes_count = 128;
    max_opaque_search = 256;
    lower_opaque_search_count = max_opaque_search;
    damage_enabled = GTRUE;
}

GoRoot::~GoRoot ()
{
    // TODO
}

GBool GoRoot::interf (GoInterf* interf)
{
    if (interf == NULL)
	interf = GoDefaultGoInterf();

    if (interf == _interf)
	return GTRUE;

    //TODO: manage the GoInterf destruction;
    
    if (!call_interf_preconditions_on_subtree(interf))
	return GFALSE;

    if (_interf != GoDefaultGoInterf())
    {
	_interf->exposePostcond(root_expose_handler, this, GREMOVE);
	_interf->clear (GBiggestRegion());
    }

    GoInterf* oldinterf = _interf;
    _interf = interf;
    if (_interf != GoDefaultGoInterf())
	_interf->exposePostcond(root_expose_handler, this);

    // preserve delete;
    DifferDestroy();
    
    call_interf_postconditions_on_subtree(oldinterf);

    if (!isToBeDestroyed() && _interf != GoDefaultGoInterf())
    {
	redraw_from_map(this);
	flush();
    }
    
    HandleDestroy();
    return GTRUE;
}

void root_expose_handler
    (void* closure, GoInterf* i, const GExposeEvent* ev)
{
    GoRoot* root = (GoRoot*) closure;
    root->expose_handler(i, ev);
}

void GoRoot::expose_handler (GoInterf*, const GExposeEvent* event)
{
    if (!_mappable)
	return;

    redrawBackAndFore(event->region());
/*OJ:30/01/92: next line removed
    flush();
*/
}


void IntersectWithInterf (GRegion& r, GoInterf* i)
{
    GBox winBox(0, 0, i->width(), i->height());
    if (r.bigbox().xmin < 0 || r.bigbox().ymin < 0 ||
	r.bigbox().xmax > winBox.xmax || r.bigbox().ymax > winBox.ymax)
    {
	GIntersect(r, GRegion(winBox), r);
    }
}

// may be modified by application if needed;
int warning_clip_postcond_count = 100;
int max_clip_postcond_count = 1000;

void GoRoot::setup_clips (int call_count)
{
    if (call_count > warning_clip_postcond_count)
    {
	if (call_count == warning_clip_postcond_count)
GWarning("It seems that you have too recursive clip postconditions ! (%d)",
		call_count);
	if (call_count > max_clip_postcond_count)
	{
	    GWarning("I stop clip postconditions (after %d calls)",
		    max_clip_postcond_count);
	    return;
	}
    }

    push_clip_postconditions(this);
    if (!_stack.empty())
    {
	call_clip_postconditions();
	setup_clips(call_count + 1);
    }

}

void GoRoot::push_clip_postconditions (Go* go)
{
    if (go == NULL)
	return;

    push_clip_postconditions(go->Next());

    if (go->clip_was_updated && go->_mappable)
    {
	push_clip_postconditions(go->Child());

	if (go->my_clip_was_updated || go->clip_was_updated)
	{
	    GBool push_go = GFALSE;

	    if (go->clip_was_updated && go->has_clip_postcond)
		push_go = GTRUE;
	    else if (go->my_clip_was_updated &&
		     (go->has_clip_postcond || go->has_my_clip_postcond))
		push_go = GTRUE;
	    else
	    {
		GBool has_class_clp = GFALSE;
		GBool has_class_myclp = GFALSE;
		for (GoMetaclass* mc = go->getMetaclass();
		     mc;
		     mc = mc->superclass())
		{
		    if (mc->has_clip_postcond)
		    {
			has_class_clp = GTRUE;
			break;
		    }
		    if (mc->has_my_clip_postcond)
		    {
			has_class_myclp = GTRUE;
			if (go->my_clip_was_updated)
			    break;
		    }
		}

		if (has_class_clp ||
		    has_class_myclp && go->my_clip_was_updated)
		    push_go = GTRUE;
	    }

	    if (push_go)
	    {
		_stack.push(go);
		_stack.push((void*)((go->my_clip_was_updated) ? 1 : 0));
	    }

	    go->clip_was_updated = go->my_clip_was_updated = GFALSE;
	}
    }
}


INLINE void Go::call_clip_postcond ()
{
    if (has_clip_postcond)
	CallPostconditions(goNclipPostcond);
}

INLINE void Go::call_my_clip_postcond ()
{
    if (has_my_clip_postcond)
	CallPostconditions(goNmyClipPostcond);
}

void Go::call_class_clip_postcond ()
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_clip_postcond)
	    mc->CallPostconditions(goNclipPostcond, this);
}

void Go::call_class_my_clip_postcond ()
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_my_clip_postcond)
	    mc->CallPostconditions(goNmyClipPostcond, this);
}


void GoRoot::call_clip_postconditions ()
{
    // preserve lib;
    DifferDestroy();
    
    while (!_stack.empty())
    {
	int do_my_clip = (int)_stack.pop();
	Go* go = (Go*)_stack.pop();
	if (go->_mappable && !go->WillBeDestroyed())
	{
	    if (do_my_clip)
	    {
		go->call_my_clip_postcond();
		go->call_class_my_clip_postcond();
	    }
	    go->call_clip_postcond();
	    go->call_class_clip_postcond();
	}
    }

    HandleDestroy();
}


void Go::clipPostcond (void cb(void*, Go*), void* closure, GAddOrRemove add)
{
    has_clip_postcond =
	AddOrRemoveCondition(goNclipPostcond, cb, closure, add);
}

void Go::myClipPostcond
    (void cb(void*, Go*), void* closure, GAddOrRemove add)
{
    has_my_clip_postcond =
	AddOrRemoveCondition(goNmyClipPostcond, cb, closure, add);
}

GoRoot* GoRoot::create ()
{
    return (GoRoot*)Go::Create(GoRoot::metaclass());
}

static GoRootMetaclass* root_metaclass = NULL;

GoMetaclass* GoRoot::metaclass ()
{
    if (root_metaclass == NULL)
    {
	root_metaclass = new GoRootMetaclass;
	root_metaclass->superclass()->AddSubclass(root_metaclass);
    }

    return root_metaclass;
}

GoRootMetaclass::GoRootMetaclass ()
{
}

GoRootMetaclass::~GoRootMetaclass ()
{
    if (this == root_metaclass)
    {
	superclass()->RemoveSubclass(this);
	root_metaclass = NULL;
    }
}

const char* GoRootMetaclass::m_classname () const
{
    return "GoRoot";
}

GoMetaclass* GoRootMetaclass::m_superclass () const
{
    return Go::metaclass();
}


static GActionRec no_actions[] = { {NULL, NULL} };

const GActionRec* GoRootMetaclass::m_actions () const
{
    return no_actions;
}


Go* GoRootMetaclass::m_create () const
{
    return new GoRoot;
}

void GoMetaclass::myClipPostcond
    (void cb(void*, Go*), void* closure, GAddOrRemove add)
{
    has_my_clip_postcond =
	AddOrRemoveCondition(goNmyClipPostcond, cb, closure, add);
}


void GoMetaclass::clipPostcond
    (void cb(void*, Go*), void* closure, GAddOrRemove add)
{
    has_clip_postcond =
	AddOrRemoveCondition(goNclipPostcond, cb, closure, add);
}

    
Go* GoRoot::upper_go_not_child (Go* go)
{
    for (Go* p = go; p != NULL; p = p->parent())
    {
	Go* nxt = p->Next();
	if (nxt) return nxt;
    }

    return NULL;
}


Go* GoRoot::upper_go (Go* go)
{
    Go* ch = go->Child();
    if (ch)
	return ch;

    return upper_go_not_child(go);
}


void GoRoot::push_all_children (Go* go)
{
    go->draw_this = go->draw_subtree = GFALSE;
    for (Go* ch = go->Child(); ch != NULL; ch = ch->Next())
	if (ch->_mappable)
	{
	    _stack.push(ch);
	    push_all_children(ch);
	}
}


void GoRoot::push_drawings (Go* go)
{
    if (go == NULL)
	return;

    if (!go->_mappable)
    {
	remove_draw_flags_on_subtree(go);
	push_drawings(upper_go_not_child(go));
	return;
    }

    GBool draw_this_go = go->draw_this;

    if (draw_this_go)
    {
	_stack.push(go);
	go->draw_this = GFALSE;

	/* optimization: if the full go subtree must be drawn it is
	   faster to add the full clip at one time than each myClip */
	if (go->draw_subtree)
	{
	    GBool full_subtree = GTRUE;
	    for (Go* ch = go->Child(); ch != NULL; ch = ch->Next())
		if (!ch->draw_this)
		{
		    full_subtree = GFALSE;
		    break;
		}

	    if (full_subtree)
	    {
		add_to_region(go, drawing_region, GTRUE);
		GIntersect(drawing_region, full_drawing_region,drawing_region);
		push_all_children(go);
		push_drawings(upper_go_not_child(go));
		return;
	    }
	}
 	
	add_to_region(go, drawing_region);
	GIntersect(drawing_region, full_drawing_region, drawing_region);
    }

    GBool full_inside = GFALSE;
    if (go->draw_subtree ||
	subtree_intersect_with_drawing_region(go, full_inside))
    {
	if (full_inside)
	{
	    if (!draw_this_go)
		_stack.push(go);
	    push_all_children(go);
	    push_drawings(upper_go_not_child(go));
	}
	else
	{
	    go->draw_subtree = GFALSE;
	    if (!draw_this_go)
	    {
		STATIC_OBJECT(GRegion, region);
		region.setEmpty();
		go->addMyClip(region);

		GIntersect(region, full_drawing_region, region);

		if (region && drawing_region)
		{
		    _stack.push(go);
		    add_to_region(region, drawing_region);
		}
	    }
	
	    push_drawings(upper_go(go));
	}
    }
    else
	push_drawings(upper_go_not_child(go));
}


GBool GoRoot::subtree_intersect_with_drawing_region
    (Go* go, GBool& full_inside)
{
    if (drawing_region.empty())
	return GFALSE;

    GRegion* stored_clip;
    if (go->clip_is_ready && (stored_clip = go->d_get_clip()))
    {
	if (stored_clip->empty())
	    return GFALSE;
	full_inside = drawing_region >= *stored_clip;
	return full_inside ? GTRUE : (*stored_clip && drawing_region);
    }
    
    STATIC_OBJECT(GRegion, clip);
    clip.setEmpty();
    go->addClip(clip);
    if (clip.empty())
	return GFALSE;
    full_inside = drawing_region >= clip;
    return full_inside ? GTRUE : (clip && drawing_region);
}


void GoRoot::flush ()
{
    if (clip_was_updated)
	setup_clips();

    if (!draw_this && !draw_subtree)
	return;

    static GRegion *clear_region = NULL;
    if (! clear_region)
	clear_region = new GRegion();

    *clear_region = drawing_region;
    use_drawing_box = GFALSE;

    IntersectWithInterf(full_drawing_region, _interf);
    if (!_mappable)
	remove_draw_flags_on_subtree(this);
    else
    {
	draw_this = draw_subtree = GFALSE;
	push_drawings(Child());
    }
    IntersectWithInterf(drawing_region, _interf);

    _interf->setClip(full_drawing_region);
    call_flush_preconditions ();
    _interf->clear(*clear_region);
    pop_drawings();
    call_flush_postconditions ();
    
    _interf->resetClip();
    _interf->flush();

    full_drawing_region.setEmpty();
    use_drawing_box = GFALSE;
    lower_opaque_search_count = max_opaque_search;
}


void GoRoot::pop_drawings ()
{
    GStack drawing_stack;
    GRegion copy_drawing_region(drawing_region);

    GBool opaque_already_used = GFALSE;
    while (!_stack.empty())
    {
	Go* go = (Go*)_stack.pop();

	GBool push_me;
	if (!opaque_already_used)
	    push_me = GTRUE;
	else
	{
            STATIC_OBJECT(GRegion, region);
	    region.setEmpty();
	    go->addMyClip(region);

	    push_me = region && copy_drawing_region;
	}
	
	if (push_me)
	{
	    drawing_stack.push(go);
	    STATIC_OBJECT(GRegion, opaqueClip);
	    opaqueClip.setEmpty();
	    go->addMyOpaqueClip(opaqueClip);
	    if (opaqueClip.squareAreaBiggerThan(min_opaque_square_area))
	    {
		GSub(copy_drawing_region, opaqueClip, copy_drawing_region);
		opaque_already_used = GTRUE;
	    }
	}
    }

    while (!drawing_stack.empty())
    {
	Go* go = (Go*)drawing_stack.pop();
	go->m_draw(drawing_region);
    }

    drawing_region.setEmpty();
}


Go* GoRoot::opaque_go_containing_region (const GRegion& region, Go* go)
{
    if (--lower_opaque_search_count < 0)
	return NULL;

    if (go == NULL)
	return NULL;

    Go* result = opaque_go_containing_region(region, go->Next());
    if (result)
	return result;

    if (!go->_mappable)
	return NULL;

    GRegion* clip;
    if (!(go->clip_is_ready && (clip = go->d_get_clip())))
    {
	STATIC_OBJECT(GRegion, local_clip);
	local_clip.setEmpty();
	go->addClip(local_clip);
	clip = &local_clip;
    }

    if (!(*clip >= region))
	return NULL;

    result = opaque_go_containing_region(region, go->Child());
    if (result)
	return result;

    STATIC_OBJECT(GRegion, opaque_clip);

    opaque_clip.setEmpty();
    go->addMyOpaqueClip(opaque_clip);
    return (opaque_clip >= region) ? go : NULL;
}


void GoRoot::redraw_from_unmap (Go* go)
{
    if (!damage_enabled)
	return;

    GRegion* region;
    if (!(go->clip_is_ready && (region = go->d_get_clip())))
    {
	STATIC_OBJECT(GRegion, local_region);
	local_region.setEmpty();
	go->addClip(local_region);
	region = &local_region;
    }
    Go* opaqueGo = opaque_go_containing_region(*region, this);
    
    if (opaqueGo) /* (opaqueGo != go) because go is unmapped */
    {
	if (go->isUpper(opaqueGo))
	    redrawForeOnly(opaqueGo, *region);
	return;
    }

    set_draw_this(this);
    add_to_region(go, drawing_region, GTRUE);
    add_to_region(go, full_drawing_region, GTRUE);
}


void GoRoot::remove_draw_flags_on_subtree (Go* go)
{
    go->draw_this = GFALSE;

    if (go->draw_subtree)
    {
	go->draw_subtree = GFALSE;
	for (Go* ch = go->Child(); ch != NULL; ch = ch->Next())
	    remove_draw_flags_on_subtree(ch);
    }
}

void GoRoot::set_draw_this_recur (Go* go)
{
    if (!go->_mappable)
	return;
    
    if (!go->draw_this)
	go->draw_this = GTRUE;
    
    Go* ch = go->Child();
    if (ch) {
	go->draw_subtree = GTRUE;
	do {
	    set_draw_this_recur(ch);
	    ch = ch->Next();
	} while (ch != NULL);
    }
}


void GoRoot::redraw_from_map (Go* go)
{
    if (!damage_enabled)
	return;

    set_draw_this(go);
    go->draw_subtree = GTRUE;
    for (Go* ch = go->Child(); ch != NULL; ch = ch->Next())
	set_draw_this_recur(ch);

    add_to_region(go, full_drawing_region, GTRUE);
}



void GoRoot::redrawForeOnly (Go* go, const GRegion& region)
{
    if (!damage_enabled)
	return;

    if (!go->_mappable)
	return;

    set_draw_this(go);
    if (&region != NULL)
	add_to_region(region, full_drawing_region);
    else
	add_to_region(go, full_drawing_region);
}

	    
void GoRoot::set_draw_this (Go* go)
{
    go->draw_this = GTRUE;

    for (Go* par = go->parent(); par != NULL; par = par->parent())
    {
	if (par->draw_subtree == GTRUE)
	    break;
	par->draw_subtree = GTRUE;
    }
}


void Go::RedrawBackAndFore (const GRegion& region)
{
    if (!_mappable)
	return;

    GoRoot* goroot = root();
    if (goroot && goroot->damageEnabled())
    {
	const GRegion* preg = &region;
	if (preg == NULL)
	{
	    STATIC_OBJECT(GRegion, clip);
	    clip.setEmpty();
	    addMyClip(clip);
	    preg = &clip;
	}
	
	Go* opaqueGo =
	    goroot->opaque_go_containing_region(*preg, goroot);
	if (opaqueGo && opaqueGo != this)
	{
	    if (this->isUpper(opaqueGo))
		goroot->redrawForeOnly(opaqueGo, *preg);
	    return;
	}

	goroot->redrawBackAndFore(*preg);
    }
}


void GoRoot::redrawBackAndFore (const GRegion& givenRegion)
{
    set_draw_this(this);
    add_to_region(givenRegion, drawing_region);
    add_to_region(givenRegion, full_drawing_region);
}


void GoRoot::add_to_region (const GRegion& src, GRegion& dst)
{
    if (use_drawing_box)
	dst = src.bigbox() + dst.bigbox();
    else
    {
	GAdd(dst, src, dst);
	if (dst.numboxes() > max_boxes_count)
	{
	    use_drawing_box = GTRUE;
	    dst = dst.bigbox();
	}
    }
}


void GoRoot::add_to_region (Go* go, GRegion& region, GBool subtree)
{
    if (subtree)
	go->addClip(region);
    else
	go->AddMyClip(region);

    if (!use_drawing_box && region.numboxes() > max_boxes_count)
	use_drawing_box = GTRUE;

    if (use_drawing_box)
	region = region.bigbox();
}


void GoRoot::flushPrecond
    (void (*precond)(void*, GoRoot*, const GRegion*), void* closure, GAddOrRemove add)
{
    has_flush_precond =
	AddOrRemoveCondition (goNflushPrecond, precond, closure, add);
}

void GoRoot::flushPostcond
    (void (*postcond)(void*, GoRoot*, const GRegion*), void* closure, GAddOrRemove add)
{
    has_flush_postcond =
	AddOrRemoveCondition (goNflushPostcond, postcond, closure, add);
}

