#ifdef REV_INFO
static char rcsid[] = "$Id: GoTree.cc,v 1.10 1992/12/21 11:39:04 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  Go Kernel
FILE:    GoTree.cc
AUTHOR:  Oliver Jojic
DATE:    April 90

DESCRIPTION:
  This file contains methods dealing with the tree structure of gos.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <Go/GoRoot.h>
#include <Go/mGo.h>
#include <Go/Tools/Stack.h>
#include <Go/Tools/Bool.h>
#include <assert.h>

// we do not need any other declaration of <Go/GoInterf.h>;
extern GoInterf* GoDefaultGoInterf();

/*
 * pre and postcondition types:
 */
char goNshiftPrecond[] = "shiftPrecond";
char goNreparentPrecond[] = "reparentPrecond";
char goNremoveChildPrecond[] = "removeChildPrecond";
char goNaddChildPrecond[] = "addChildPrecond";
char goNrootPrecond[] = "rootPrecond";
char goNinterfPrecond[] = "interfPrecond";
char goNcrossTypePrecond[] = "crossTypePrecond";

char goNshiftPostcond[] = "shiftPostcond";
char goNreparentPostcond[] = "reparentPostcond";
char goNremoveChildPostcond[] = "removeChildPostcond";
char goNaddChildPostcond[] = "addChildPostcond";
char goNrootPostcond[] = "rootPostcond";
char goNinterfPostcond[] = "interfPostcond";
char goNcrossTypePostcond[] = "crossTypePostcond";


void Go::shiftPrecond
    (GBool cb(void*, const Go*, const Go*(*)),
     void* closure, GAddOrRemove add)
{
    has_shift_precond =
	AddOrRemoveCondition(goNshiftPrecond, cb, closure, add);
}

void Go::shiftPostcond
    (void cb(void*, Go*, Go*), void* closure, GAddOrRemove add)
{
    has_shift_postcond =
	AddOrRemoveCondition(goNshiftPostcond, cb, closure, add);
}

void Go::removeChildPrecond
    (GBool cb(void*, const Go*, const Go*), void* closure, GAddOrRemove add)
{
    has_remove_child_precond =
	AddOrRemoveCondition(goNremoveChildPrecond, cb, closure, add);
}

void Go::removeChildPostcond
    (void cb(void*, Go*, Go*, Go*), void* closure, GAddOrRemove add)
{
    has_remove_child_postcond =
	AddOrRemoveCondition(goNremoveChildPostcond, cb, closure, add);
}

void Go::addChildPrecond
    (GBool cb(void*, const Go*, const Go*, const Go*(*)),
     void* closure, GAddOrRemove add)
{
    has_add_child_precond =
	AddOrRemoveCondition(goNaddChildPrecond, cb, closure, add);
}

void Go::addChildPostcond
    (void cb(void*, Go*, Go*), void* closure, GAddOrRemove add)
{
    has_add_child_postcond =
	AddOrRemoveCondition(goNaddChildPostcond, cb, closure, add);
}

void Go::reparentPrecond
    (GBool cb(void*, const Go*, const Go*, const Go*(*)),
     void* closure, GAddOrRemove add)
{
    has_reparent_precond =
	AddOrRemoveCondition(goNreparentPrecond, cb, closure, add);
}

void Go::reparentPostcond
    (void cb(void*, Go*, Go*, Go*), void* closure, GAddOrRemove add)
{
    has_reparent_postcond =
	AddOrRemoveCondition(goNreparentPostcond, cb, closure, add);
}

void Go::rootPrecond
    (GBool cb(void*, const Go*, const Go*), void* closure, GAddOrRemove add)
{
    has_root_precond =
	AddOrRemoveCondition(goNrootPrecond, cb, closure, add);
}

void Go::rootPostcond
    (void cb(void*, Go*, Go*), void* closure, GAddOrRemove add)
{
    has_root_postcond =
	AddOrRemoveCondition(goNrootPostcond, cb, closure, add);
}

void Go::interfPrecond
    (GBool cb(void*, const Go*, const GoInterf*),
     void* closure, GAddOrRemove add)
{
    has_interf_precond =
	AddOrRemoveCondition(goNinterfPrecond, cb, closure, add);
}

void Go::interfPostcond
    (void cb(void*, Go*, GoInterf*), void* closure, GAddOrRemove add)
{
    has_interf_postcond =
	AddOrRemoveCondition(goNinterfPostcond, cb, closure, add);
}

void Go::crossTypePrecond
    (GBool cb(void*, const Go*, GCrossType*), void* closure, GAddOrRemove add)
{
    has_cross_type_precond =
	AddOrRemoveCondition(goNcrossTypePrecond, cb, closure, add);
}

void Go::crossTypePostcond
    (void cb(void*, Go*, const GCrossType*), void* closure, GAddOrRemove add)
{
    has_cross_type_postcond =
	AddOrRemoveCondition(goNcrossTypePostcond, cb, closure, add);
}

void GoMetaclass::shiftPrecond
    (GBool cb(void*, const Go*, const Go*(*)),
     void* closure, GAddOrRemove add)
{
    has_shift_precond =
	AddOrRemoveCondition(goNshiftPrecond, cb, closure, add);
}

void GoMetaclass::shiftPostcond
    (void cb(void*, Go*, Go*), void* closure, GAddOrRemove add)
{
    has_shift_postcond =
	AddOrRemoveCondition(goNshiftPostcond, cb, closure, add);
}

void GoMetaclass::removeChildPrecond
    (GBool cb(void*, const Go*, const Go*),
     void* closure, GAddOrRemove add)
{
    has_remove_child_precond =
	AddOrRemoveCondition(goNremoveChildPrecond, cb, closure, add);
}

void GoMetaclass::removeChildPostcond
    (void cb(void*, Go*, Go*, Go*), void* closure, GAddOrRemove add)
{
    has_remove_child_postcond =
	AddOrRemoveCondition(goNremoveChildPostcond, cb, closure, add);
}

void GoMetaclass::addChildPrecond
    (GBool cb(void*, const Go*, const Go*, const Go*(*)),
     void* closure, GAddOrRemove add)
{
    has_add_child_precond =
	AddOrRemoveCondition(goNaddChildPrecond, cb, closure, add);
}

void GoMetaclass::addChildPostcond
    (void cb(void*, Go*, Go*), void* closure, GAddOrRemove add)
{
    has_add_child_postcond =
	AddOrRemoveCondition(goNaddChildPostcond, cb, closure, add);
}

void GoMetaclass::reparentPrecond
    (GBool cb(void*, const Go*, const Go*, const Go*(*)),
     void* closure, GAddOrRemove add)
{
    has_reparent_precond =
	AddOrRemoveCondition(goNreparentPrecond, cb, closure, add);
}

void GoMetaclass::reparentPostcond
    (void cb(void*, Go*, Go*, Go*), void* closure, GAddOrRemove add)
{
    has_reparent_postcond =
	AddOrRemoveCondition(goNreparentPostcond, cb, closure, add);
}

void GoMetaclass::rootPrecond
    (GBool cb(void*, const Go*, const Go*), void* closure, GAddOrRemove add)
{
    has_root_precond =
	AddOrRemoveCondition(goNrootPrecond, cb, closure, add);
}

void GoMetaclass::rootPostcond
    (void cb(void*, Go*, Go*), void* closure, GAddOrRemove add)
{
    has_root_postcond =
	AddOrRemoveCondition(goNrootPostcond, cb, closure, add);
}

void GoMetaclass::interfPrecond
    (GBool cb(void*, const Go*, const GoInterf*),
     void* closure, GAddOrRemove add)
{
    has_interf_precond =
	AddOrRemoveCondition(goNinterfPrecond, cb, closure, add);
}

void GoMetaclass::interfPostcond
    (void cb(void*, Go*, GoInterf*), void* closure, GAddOrRemove add)
{
    has_interf_postcond =
	AddOrRemoveCondition(goNinterfPostcond, cb, closure, add);
}

void GoMetaclass::crossTypePrecond
    (GBool cb(void*, const Go*, GCrossType*), void* closure, GAddOrRemove add)
{
    has_cross_type_precond =
	AddOrRemoveCondition(goNcrossTypePrecond, cb, closure, add);
}

void GoMetaclass::crossTypePostcond
    (void cb(void*, Go*, const GCrossType*), void* closure, GAddOrRemove add)
{
    has_cross_type_postcond =
	AddOrRemoveCondition(goNcrossTypePostcond, cb, closure, add);
}



Go* Go::previous () const
{
    for (Go* prv = Previous(); prv; prv = prv->Previous())
	if (prv->crossType()==GCROSS_AS_NODE || prv->crossType()==GCROSS_AS_LEAF)
	    return prv;

    return NULL;
}

Go* Go::next () const
{
    for (Go* nxt = Next(); nxt; nxt = nxt->Next())
	if (nxt->crossType()==GCROSS_AS_NODE || nxt->crossType()==GCROSS_AS_LEAF)
	    return nxt;

    return NULL;
}

Go* Go::child () const
{
    if (crossType() == GCROSS_AS_LEAF)
	return NULL;

    Go* ch = Child();
    if (ch)
	return (ch->crossType()==GCROSS_AS_NODE || ch->crossType()==GCROSS_AS_LEAF)
	    ? ch : ch->next();
    return NULL;
}

Go* Go::lastChild () const
{
    if (crossType() == GCROSS_AS_LEAF)
	return NULL;

    Go* last_child = NULL;
    for (Go* ch = Child(); ch; ch = ch->Next())
	if (ch->crossType()==GCROSS_AS_NODE || ch->crossType()==GCROSS_AS_LEAF)
	    last_child = ch;

    return last_child;
}

void go_at_proc (void*, Go* go)
{
    go->stopRecursion(go);
}

GBool Go::isAncestor (const Go* go) const
{
    for (; go; go = go->parent())
	if (go == this)
	    return GTRUE;

    return GFALSE;
}

Go* GoCommonAncestor (const Go* a, const Go* b)
{
    for (Go* p = (Go*)a; p; p = p->parent())
	p->common_ancestor_mark = GTRUE;

    for (Go* common = (Go*)b; common; common = common->parent())
	if (common->common_ancestor_mark)
	    break;

    for (p = (Go*)a; p; p = p->parent())
	p->common_ancestor_mark = GFALSE;

    return common;
}

GBool Go::isUpper (const Go* go) const //if (this == go) it is not upper
{
    GBool is_upper = GFALSE;

    for (Go* thispar = (Go*)this; thispar; thispar = thispar->parent())
	thispar->common_ancestor_mark = GTRUE;

    if (go->common_ancestor_mark)
	is_upper = GTRUE;
    else
    {
	Go* goparpar;
	for (Go* gopar = (Go*)go;
	     (goparpar = gopar->parent());
	     gopar = goparpar)
	{
	    if (goparpar->common_ancestor_mark)
	    {
		if (this != goparpar)
		{
		    Go* thisparpar;
		    for (thispar = (Go*)this;
			 (thisparpar = thispar->parent()) != goparpar;
			 thispar = thisparpar);
		    //now gopar and thispar are brothers !;
		    for (gopar=gopar->Next(); gopar; gopar=gopar->Next())
			if (gopar == thispar)
			{
			    is_upper = GTRUE;
			    break;
			}
		}

		break;
	    }
	}
    }

    for (thispar = (Go*)this; thispar; thispar = thispar->parent())
	thispar->common_ancestor_mark = GFALSE;

    return is_upper;
}

Go* Go::Previous (int myindex) const
{
    Go* par = parent();
    if (par == NULL)
	return NULL;

    if (myindex != GUNKNOWN_INDEX && par->d_store_children_array())
    {
	int count;
	Go** pch = par->d_get_children(count);
	int previndex = myindex - 1;
	return (previndex < 0 || previndex >= count)
	    ? NULL : pch[previndex];
    }

    return d_get_previous();
}

Go* Go::Next (int myindex) const
{
    Go* par = parent();
    if (par == NULL)
	return NULL;

    if (myindex != GUNKNOWN_INDEX && par->d_store_children_array())
    {
	int count;
	Go** pch = par->d_get_children(count);
	int nextindex = myindex + 1;
	return (nextindex < 0 || nextindex >= count)
	    ? NULL : pch[nextindex];
    }

    return d_get_next();
}

Go* Go::Child (int chindex) const
{
    if (chindex < 0)
	return NULL;

    if (d_store_children_array())
    {
	int count;
	Go** pch = d_get_children(count);
	return (chindex >= count)
	    ? NULL : pch[chindex];
    }

    Go* ch = d_get_first_child();
    for (int i = 0; i < chindex; i++)
    {
	if (ch == NULL)
	    return NULL;
	ch = ch->d_get_next();
    }

    return ch;
}

Go* Go::LastChild () const
{
    if (d_store_children_array())
    {
	int count;
	Go** pch = d_get_children(count);
	return (count <= 0) ? NULL : pch[count - 1];
    }

    Go* prevch = NULL;
    for (Go* ch = d_get_first_child(); ch; ch = ch->d_get_next())
	prevch = ch;

    return prevch;
}

Go* Go::Root () const
{
    const Go* prevpar = this;
    for (Go* par = parent(); par; par = par->parent())
	prevpar = par;

    return (Go*)prevpar;
}

int Go::childIndex (const Go* child) const
{
    int index = 0;
    for (Go* ch = Child(); ch; ch = ch->Next(index++))
	if (ch == child)
	    return index;

    if (child)
	GWarning("Go::childIndex() called but not for a child");

    return index;
}


INLINE GBool Go::call_shift_precond
    (const Go*& newnext, int oldindex, int newindex) const
{
    return (has_shift_precond)
	? CallPreconditions(goNshiftPrecond,(void*)&newnext, &oldindex, &newindex)
	    : GTRUE;
}

GBool Go::call_class_shift_precond (const Go*& newnext, int oi, int ni) const
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_shift_precond &&
	    !mc->CallPreconditions(goNshiftPrecond, this, (void*)&newnext, &oi, &ni))
	    return GFALSE;

    return GTRUE;
}


INLINE GBool Go::call_reparent_precond
    (const Go* newpar, const Go*& newnext) const
{
    return (has_reparent_precond)
	? CallPreconditions(goNreparentPrecond, (void*)newpar, (void*)&newnext)
	    : GTRUE;
}

GBool Go::call_class_reparent_precond
    (const Go* newpar, const Go*& newnext) const
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_reparent_precond &&
	    !mc->CallPreconditions(goNreparentPrecond, this,
				   (void*)newpar, (void*)&newnext))
	    return GFALSE;

    return GTRUE;
}


INLINE GBool Go::call_remove_child_precond
    (const Go* chtorem, int chindex) const
{
    return (has_remove_child_precond)
	? CallPreconditions(goNremoveChildPrecond, (void*)chtorem, &chindex)
	    : GTRUE;
}

GBool Go::call_class_remove_child_precond
    (const Go* chtorem, int chindex) const
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_remove_child_precond &&
	    !mc->CallPreconditions(goNremoveChildPrecond, this,
				   (void*)chtorem, &chindex))
	    return GFALSE;

    return GTRUE;
}


INLINE GBool Go::call_add_child_precond
    (const Go* chtoadd, const Go*& newnextbr, int chindex) const
{
    return (has_add_child_precond)
	? CallPreconditions(goNaddChildPrecond,
			    (void*)chtoadd, (void*)&newnextbr, &chindex)
	    : GTRUE;
}

GBool Go::call_class_add_child_precond
    (const Go* chtoadd, const Go*& newnextbr, int chindex) const
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_add_child_precond &&
	    !mc->CallPreconditions(goNaddChildPrecond, this,
				   (void*)chtoadd, (void*)&newnextbr, &chindex))
	    return GFALSE;

    return GTRUE;
}


INLINE GBool Go::call_root_precond (const Go* newroot) const
{
    return (has_root_precond)
	? CallPreconditions(goNrootPrecond, (void*)newroot)
	    : GTRUE;
}


GBool Go::call_class_root_precond (const Go* newroot) const
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_root_precond &&
	    !mc->CallPreconditions(goNrootPrecond, this, (void*)newroot))
	    return GFALSE;

    return GTRUE;
}


INLINE GBool Go::call_interf_precond (const GoInterf* newinterf) const
{
    return (has_interf_precond)
	? CallPreconditions(goNinterfPrecond, (void*)newinterf)
	    : GTRUE;
}


GBool Go::call_class_interf_precond
    (const GoInterf* newinterf) const
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_interf_precond &&
	    !mc->CallPreconditions(goNinterfPrecond, this, (void*)newinterf))
	    return GFALSE;

    return GTRUE;
}


GBool Go::call_root_and_interf_precond_on_subtree
    (const Go* newroot, const GoInterf* newinterf) const
{
    if (!call_root_precond(newroot) ||
	!call_class_root_precond(newroot))
	return GFALSE;

    if (newinterf &&
	(!call_interf_precond(newinterf) ||
	 !call_class_interf_precond(newinterf)))
	return GFALSE;

    Go* child = Child();
    for (int i = 0; child; child = child->Next(i++))
	if (!child->call_root_and_interf_precond_on_subtree
	    		(newroot, newinterf))
	    return GFALSE;

    return GTRUE;
}

GBool Go::callReparentPrecond
    (const Go* newParent, const Go* (&nextBrother)) const
{
    if (nextBrother && nextBrother->parent() != newParent)
    {
	GWarning("reparent with newnext not a child of newparent (not done)");
	return GFALSE;
    }

    if (newParent && isAncestor(newParent))
    {
	GWarning("reparent on a go of this go subtree (not done)");
	return GFALSE;
    }

    if (newParent == parent() && nextBrother == this)
	return GTRUE;

    int new_ix = GUNKNOWN_INDEX;
    return
	((Go*)this)->reparent_call_preconditions
	    (GUNKNOWN_INDEX, parent(),
	     new_ix, newParent, nextBrother);
}
    

GBool Go::reparent (Go* newpar, Go* newnext)
{
    if (newnext && newnext->parent() != newpar)
    {
	GWarning("reparent with newnext not a child of newparent (not done)");
	return GFALSE;
    }

    if (newpar && isAncestor(newpar))
    {
	GWarning("reparent on a go of this go subtree (not done)");
	return GFALSE;
    }

    Go* oldpar = parent();
    int oldix = (oldpar && oldpar->d_store_children_array())
	? oldpar->childIndex(this) : GUNKNOWN_INDEX;
    int newix = (newpar) ? newpar->childIndex(newnext) : GUNKNOWN_INDEX;
    Go* oldnext = Next(oldix);

    if (newpar == oldpar && newnext == this)
	return GTRUE;

#if defined(__BORLANDC__) || defined(__MSC7__)
    const Go* newnext_ref = newnext;
    if (!reparent_call_preconditions(oldix, oldpar, newix, newpar, newnext_ref))
	return GFALSE;
    newnext =  (Go*) newnext_ref;
#else
    if (!reparent_call_preconditions(oldix, oldpar, newix, newpar, newnext))
	return GFALSE;
#endif

    if (newpar == oldpar && newnext == oldnext)
	return GFALSE; //same result as if it were refused

    GCoordSys oldabs(Abscs());
    GCoordSys oldrel(Relcs());

    GBool was_mapped = mapped();
    if (was_mapped)
	TempUnmap();

    GoRoot* goroot = root();
    if (goroot == NULL && newpar != NULL)
	goroot = newpar->root();

    // do it;
    do_reparent(oldix, oldpar, oldnext, newix, newpar, newnext);
    abs_reparent_setup_relcs(oldpar, newpar, oldabs, oldrel);


    // preserve destroy this by postconds;
    DifferDestroy();
	
    // postconds;
    reparent_call_postconditions(oldix, oldpar, newix, newpar, newnext);
    abs_reparent_call_cs_postconditions();

    if (was_mapped)
	TempMap();

    HandleDestroy();
    return GTRUE;
}


GBool Go::parentRelativeReparent (Go* newpar, Go* newnext)
{
    if (newnext && newnext->parent() != newpar)
    {
	GWarning("rel reparent with next br. not a ch. of new par. (not done)");
	return GFALSE;
    }

    if (newpar && isAncestor(newpar))
    {
	GWarning("rel reparent on a go of this subtree (not done)");
	return GFALSE;
    }

    Go* oldpar = parent();
    int oldix = (oldpar && oldpar->d_store_children_array())
	? oldpar->childIndex(this) : GUNKNOWN_INDEX;
    int newix = (newpar) ? newpar->childIndex(newnext) : GUNKNOWN_INDEX;
    Go* oldnext = Next(oldix);

    if (newpar == oldpar && newnext == this)
	return GTRUE;

#if defined(__BORLANDC__) || defined(__MSC7__)
    const Go* newnext_ref = newnext;
    if (!reparent_call_preconditions(oldix, oldpar, newix, newpar, newnext_ref))
	return GFALSE;
    newnext = (Go*) newnext_ref;
#else
    if (!reparent_call_preconditions(oldix, oldpar, newix, newpar, newnext))
	return GFALSE;
#endif

    if (newpar == oldpar && newnext == oldnext)
	return GFALSE; //same result as if it were refused

    GCoordSys oldabs(Abscs());
    GCoordSys oldrel(Relcs());

    GBool was_mapped = mapped();
    if (was_mapped)
	TempUnmap();

    do_reparent(oldix, oldpar, oldnext, newix, newpar, newnext);
    rel_reparent_setup_abscs(oldpar, newpar, oldabs, oldrel);


    // preserve destroy this by postconds;
    DifferDestroy();

    // postconds;
    reparent_call_postconditions(oldix, oldpar, newix, newpar, newnext);
    rel_reparent_call_cs_postconditions();
    
    if (was_mapped)
	TempMap();

    HandleDestroy();
    return GTRUE;
}


void Go::rel_reparent_call_cs_postconditions ()
{
    assert (differ_destroy);
    /*
     * after a relative reparent, abscs change.
     * We have to call abscs postconditions.
     */
    set_cs_asis();
}


GBool Go::reparent_call_preconditions
    (int oldix, const Go* oldpar,
     int& newix, const Go* newpar, const Go*& newnext)
{
    const Go* ini_newnext = newnext;

    if (newpar == oldpar)
    {
	if (newnext == Next(oldix))
	    return GFALSE;

	if (!call_shift_precond(newnext, oldix, newix))
	    return GFALSE;
	if (ini_newnext != newnext)
	{ ini_newnext = newnext; newix = newpar->childIndex(newnext); }

	if (!call_class_shift_precond(newnext, oldix, newix))
	    return GFALSE;
	if (ini_newnext != newnext)
	{ ini_newnext = newnext; newix = newpar->childIndex(newnext); }
    }

    if (!call_reparent_precond(newpar, newnext) ||
	!call_class_reparent_precond(newpar, newnext))
	return GFALSE;
    if (newnext != ini_newnext)
    { ini_newnext = newnext; newix = newpar->childIndex(newnext); }

    if (oldpar &&
	(!oldpar->call_remove_child_precond(this, oldix) ||
	 !oldpar->call_class_remove_child_precond(this, oldix)))
	return GFALSE;

    if (newpar)
    {
	if (!newpar->call_add_child_precond(this, newnext, newix))
	    return GFALSE;
	if (newnext != ini_newnext)
	{ ini_newnext = newnext; newix = newpar->childIndex(newnext); }
	
	if (!newpar->call_class_add_child_precond(this, newnext, newix))
	    return GFALSE;
	if (newnext != ini_newnext)
	{ ini_newnext = newnext; newix = newpar->childIndex(newnext); }
    }

    Go* oldroot = Root();
    Go* newroot = (newpar) ? newpar->Root() : this;
    GoInterf* newinterf = (newpar) ? newpar->interf() : GoDefaultGoInterf();
    if (oldroot != newroot &&
	!call_root_and_interf_precond_on_subtree
	    (newroot,
	     (newroot->interf() == newinterf) ? NULL : newinterf))
	return GFALSE;

    return GTRUE;
}


void Go::DoReparent (Go* newParent, Go* newNext)
{
    Go* oldParent = parent();
    int oldIndex = (oldParent && oldParent->d_store_children_array())
	? oldParent->childIndex(this) : GUNKNOWN_INDEX;
    int newIndex = (newParent)
	? newParent->childIndex(newNext) : GUNKNOWN_INDEX;

    do_reparent(oldIndex, oldParent, Next(oldIndex),
		newIndex, newParent, newNext);
}

void Go::do_reparent
    (int oldix, Go* oldpar, Go* oldnext,
     int newix, Go* newpar, Go* newnext)
{
    Go* oldprev = Previous(oldix);
    Go* newprev =
	(newnext) ? newnext->Previous(newix)
	    : (newpar) ? newpar->LastChild()
		: NULL;

    if (oldpar)	oldpar->d_remove_child(this, oldix);
    if (oldprev) oldprev->d_set_next(oldnext);
    if (oldnext) oldnext->d_set_previous(oldprev);

    d_set_tree(newpar, newprev, newnext);

    if (newprev) newprev->d_set_next(this);
    if (newnext) newnext->d_set_previous(this);
    if (newpar) newpar->d_add_child(this, newix);
}


void Go::abs_reparent_setup_relcs
    (Go* oldpar, Go* newpar, GCoordSys& oldabs, GCoordSys& oldrel)
{
    /*
     * (newabs = oldabs <==> oldparabs * oldrel = newparabs * newrel)
     * ==> newrel = ~newparabs * oldparabs * oldrel
     */

    GCoordSys newparabs;
    GCoordSys oldparabs ;
    if (newpar)  newparabs =  newpar->Abscs();
    if (oldpar)  oldparabs =  oldpar->Abscs();

    if (newparabs != oldparabs)
	d_set_coordsys(~newparabs * oldparabs * oldrel, oldabs, newparabs);
}



INLINE void Go::call_shift_postcond ()
{
    if (has_shift_postcond)
	CallPostconditions(goNshiftPostcond);
}

void Go::call_class_shift_postcond ()
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_shift_postcond)
	    mc->CallPostconditions(goNshiftPostcond, this);
}

INLINE void Go::call_reparent_postcond (Go* oldpar, int oldindex)
{
    if (has_reparent_postcond)
	CallPostconditions(goNreparentPostcond, oldpar, &oldindex);
}

void Go::call_class_reparent_postcond (Go* oldpar, int oldindex)
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_reparent_postcond)
	    mc->CallPostconditions(goNreparentPostcond, this,
				   oldpar, &oldindex);
}

INLINE void Go::call_remove_child_postcond (Go* oldch, int oldindex)
{
    if (has_remove_child_postcond) 
	CallPostconditions(goNremoveChildPostcond, oldch, &oldindex);
}

void Go::call_class_remove_child_postcond (Go* oldch, int oldindex)
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_remove_child_postcond)
	    mc->CallPostconditions(goNremoveChildPostcond, this,
				   oldch, &oldindex);
}

INLINE void Go::call_add_child_postcond (Go* ch, int index)
{
    if (has_add_child_postcond) 
	CallPostconditions(goNaddChildPostcond, ch, &index);
}

void Go::call_class_add_child_postcond (Go* ch, int index)
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_add_child_postcond)
	    mc->CallPostconditions(goNaddChildPostcond, this, ch, &index);
}


void Go::reparent_call_postconditions
    (int oldix, Go* oldpar,
     int newix, Go* newpar, Go*)
{
    assert (differ_destroy);
    
    Go* oldroot;
    if (oldpar) 
	oldroot = oldpar->Root();
    else oldroot = this;

    Go* newroot;
    if (newpar) 
	newroot = newpar->Root();
    else newroot = this;

    if (newpar == oldpar)
    {
	call_shift_postcond();
	call_class_shift_postcond();
	return;
    }

    call_reparent_postcond(oldpar, oldix);
    call_class_reparent_postcond(oldpar, oldix);

    if (oldpar)
    {
	oldpar->call_remove_child_postcond(this, oldix);
	oldpar->call_class_remove_child_postcond(this, oldix);
    }

    if (newpar)
    {
	newpar->call_add_child_postcond(this, newix);
	newpar->call_class_add_child_postcond(this, newix);
    }

    if (newroot != oldroot) 
    {
	GoInterf* oldinterf = (oldpar) ? oldpar->interf() : GoDefaultGoInterf();
	call_root_and_interf_postcond_on_subtree
	    (oldroot,
	     (oldinterf == newroot->interf()) ? NULL : oldinterf);
    }
}


void Go::abs_reparent_call_cs_postconditions ()
{
    assert (differ_destroy);
    /*
     * after the absolute reparent the relcs has changed.
     * We have to call  relcs postconditions.
     */
    set_cs_asis();
}


INLINE void Go::call_root_postcond (Go* oldroot)
{
    if (has_root_postcond)
	CallPostconditions(goNrootPostcond, oldroot);
}

void Go::call_class_root_postcond (Go* oldroot)
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_root_postcond)
	    mc->CallPostconditions(goNrootPostcond, this, oldroot);
}

INLINE void Go::call_interf_postcond (GoInterf* oldinterf)
{
    if (has_interf_postcond)
	CallPostconditions(goNinterfPostcond, oldinterf);
}

void Go::call_class_interf_postcond (GoInterf* oldinterf)
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_interf_postcond)
	    mc->CallPostconditions(goNinterfPostcond, this, oldinterf);
}


void Go::call_root_and_interf_postcond_on_subtree
    (Go* oldroot, GoInterf* oldinterf)
{
    call_root_postcond(oldroot);
    call_class_root_postcond(oldroot);
    if (oldinterf)
    {
	call_interf_postcond(oldinterf);
	call_class_interf_postcond(oldinterf);
    }
    
    Go* child = Child();
    for (int i = 0; child; child = child->Next(i++))
	child->call_root_and_interf_postcond_on_subtree(oldroot, oldinterf);
}



INLINE GBool Go::call_cross_type_precond (GCrossType& ctyp) const
{
    return (has_cross_type_precond)
	? CallPreconditions(goNcrossTypePrecond, &ctyp)
	    : GTRUE;
}

GBool Go::call_class_cross_type_precond (GCrossType& ctyp) const
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_cross_type_precond &&
	    !mc->CallPreconditions(goNcrossTypePrecond, this, &ctyp))
	    return GFALSE;

    return GTRUE;
}

INLINE void Go::call_cross_type_postcond (GCrossType oldctyp)
{
    if (has_cross_type_postcond)
	CallPostconditions(goNcrossTypePostcond, &oldctyp);
}

void Go::call_class_cross_type_postcond (GCrossType oldctyp)
{
    for (GoMetaclass* mc = getMetaclass(); mc; mc = mc->superclass())
	if (mc->has_cross_type_postcond)
	    mc->CallPostconditions(goNcrossTypePostcond, this, &oldctyp);
}


inline GBool Go::callCrossTypePrecond (GCrossType ctyp) const
{
    return
	ctyp == cross_type
	    || call_cross_type_precond (ctyp)
		&& call_class_cross_type_precond (ctyp);
}


GBool Go::crossType (GCrossType ctyp)
{
    if (!callCrossTypePrecond (ctyp))
	return GTRUE;

    GCrossType oldctyp = (GCrossType)cross_type;
    cross_type = ctyp;

    // preserve delete;
    DifferDestroy();
    
    call_cross_type_postcond(oldctyp);
    call_class_cross_type_postcond(oldctyp);

    HandleDestroy();
    return GTRUE;
}


/*
 * The stack used by `forEvery...()' methods to store
 * `recursion value' and `stop recursion ?'.
 */
static GStack& RecursionStack() {
    static GStack *recursion_stack = NULL;

    if (! recursion_stack)
	recursion_stack = new GStack();

    return *recursion_stack;
}


void* execute_go_stack
    (GStack* go_stack, void cb(void*, Go*), void* closure)
{
    RecursionStack().push(NULL); //push the return value;
    RecursionStack().push((void*)GFALSE); //push the "stop recursion ?" flag;
	
    while (!go_stack->empty())
    {
	cb(closure, (Go*)go_stack->pop());
	if (RecursionStack().top() != (void*)GFALSE) //stop recursion ?
	    break;
    }

    RecursionStack().pop(); //pop the "stop recursion?" flag;
    return RecursionStack().pop(); //pop the return value;
}

    
/*
 * Call proc on every go of this go subtree.
 * Proc is called with two parameters:
 *	proc(closure, go)
 * The Go tree is parsed from child to parent and from
 * next brother to previous brother.
 *
 * Nota:
 * 	While we are executing a recursion we want that another
 *	recursion can be started. This explains the
 *	use of the `RecursionStack()'.
 */
void* Go::ForEveryGo (void cb(void*, Go*), void* closure)
{
    void* v;
{ //due to a compiler bug which not call stack destructor.
    GStack go_stack;
    push_every_go(go_stack,
		  /*public*/ GFALSE,
		  /*mapped*/ GFALSE,
		  /*selected*/ GFALSE,
		  /*subtree*/ GTRUE,
		  /*sensor*/ GFALSE);

    v = execute_go_stack(&go_stack, cb, closure);
}
    return v;
}

void* Go::forEveryGo (void cb(void*, Go*), void* closure)
{
    void* v;
{ //due to a compiler bug which not call stack destructor.
    GStack go_stack;
    push_every_go(go_stack,
		  /*public*/ GTRUE,
		  /*mapped*/ GFALSE,
		  /*selected*/ GFALSE,
		  /*subtree*/ GTRUE,
		  /*sensor*/ GFALSE);

    v = execute_go_stack(&go_stack, cb, closure);
}
    return v;
}


void* Go::ForEveryMappedGo (void cb(void*, Go*), void* closure)
{
    if (!mapped())
	return NULL;

    void* v;
{
    //due to a compiler bug which not call stack destructor;
    GStack go_stack;
    push_every_go(go_stack,
		  /*public*/ GFALSE,
		  /*mapped*/ GTRUE,
		  /*selected*/ GFALSE,
		  /*subtree*/ GTRUE,
		  /*sensor*/ GFALSE);

    v = execute_go_stack(&go_stack, cb, closure);
}
    return v;
}

void* Go::forEveryMappedGo (void cb(void*, Go*), void* closure)
{
    if (!mapped())
	return NULL;

    void* v;
{
    //due to a compiler bug which not call stack destructor;
    GStack go_stack;
    push_every_go(go_stack,
		  /*public*/ GTRUE,
		  /*mapped*/ GTRUE,
		  /*selected*/ GFALSE,
		  /*subtree*/ GTRUE,
		  /*sensor*/ GFALSE);

    v = execute_go_stack(&go_stack, cb, closure);
}
    return v;
}


void* Go::ForEverySelectedGo (void cb(void*, Go*), void* closure)
{
    if (!mapped())
	return NULL;

    void* v;
{
    //due to a compiler bug which not call stack destructor;
    GStack go_stack;
    push_every_go(go_stack,
		  /*public*/ GFALSE,
		  /*mapped*/ GTRUE,
		  /*selected*/ GTRUE,
		  /*subtree*/ GTRUE,
		  /*sensor*/ GFALSE);

    v = execute_go_stack(&go_stack, cb, closure);
}
    return v;
}

void* Go::forEverySelectedGo (void cb(void*, Go*), void* closure)
{
    if (!mapped())
	return NULL;

    void* v;
{
    //due to a compiler bug which not call stack destructor;
    GStack go_stack;
    push_every_go(go_stack,
		  /*public*/ GTRUE,
		  /*mapped*/ GTRUE,
		  /*selected*/ GTRUE,
		  /*subtree*/ GTRUE,
		  /*sensor*/ GFALSE);

    v = execute_go_stack(&go_stack, cb, closure);
}
    return v;
}


void* Go::ForEverySensor (void cb(void*, GoSensor*), void* closure)
{
    if (!mapped())
	return NULL;

    void* v;
{
    //due to a compiler bug which not call stack destructor;
    GStack go_stack;
    push_every_go(go_stack,
		  /*public*/ GFALSE,
		  /*mapped*/ GTRUE,
		  /*selected*/ GFALSE,
		  /*subtree*/ GTRUE,
		  /*sensor*/ GTRUE);

    v = execute_go_stack(&go_stack, (void(*)(void*, Go*))cb, closure);
}
    return v;
}

/*
 * Call proc on every selected go of this go subtree.
 * When a selected go is found, the recursion is not done on this
 * selected go subtree.
 * Proc is called with two parameters:
 *	proc(closure, go);
 *
 * See the Nota of the method `Go::forEveryGo()'.
 */
void* Go::ForEverySelectedSubtree (void cb(void*, Go*), void* closure)
{
    if (!mapped())
	return NULL;

    void* v;
{
    //due to a compiler bug which not call stack destructor;
    GStack go_stack; 
    push_every_go(go_stack,
		  /*public*/ GFALSE,
		  /*mapped*/ GTRUE,
		  /*selected*/ GTRUE,
		  /*subtree*/ GFALSE,
		  /*sensor*/ GFALSE);

    v = execute_go_stack(&go_stack, cb, closure);
}
    return v;
}

void* Go::forEverySelectedSubtree (void cb(void*, Go*), void* closure)
{
    if (!mapped())
	return NULL;

    void* v;
{
    //due to a compiler bug which not call stack destructor;
    GStack go_stack;
    push_every_go(go_stack,
		  /*public*/ GTRUE,
		  /*mapped*/ GTRUE,
		  /*selected*/ GTRUE,
		  /*subtree*/ GFALSE,
		  /*sensor*/ GFALSE);

    v = execute_go_stack(&go_stack, cb, closure);
}
    return v;
}


#ifndef STATIC_CONFUSION
static
#endif
void Go::push_every_go
    (GStack& go_stack,
     GBool public_only, GBool mapped_only, GBool selected_only,
     GBool on_subtree, GBool sensor_only)
{
    if (public_only && crossType() == GCROSS_AS_NOTHING)
	return;

    if (mapped_only && !mappable())
	return;

    if ((!selected_only || selected()) &&
	(!sensor_only || is_sensor()))
    {
	go_stack.push(this);
	if (!on_subtree)
	    return;
    }

    if (!public_only || crossType() != GCROSS_AS_LEAF)
    {
	Go* ch = Child();
	for (int i = 0; ch; ch = ch->Next(i++))
	    ch->push_every_go(go_stack, public_only, mapped_only,
			      selected_only, on_subtree, sensor_only);
    }
}




/*
 * Call proc on every go that contains `abspt'.
 * Proc is called with two parameters:
 *	proc(closure, go)
 *
 * See the Nota of the method `Go::forEveryGo()'.
 *
 * This method do not use a "go_stack", because it will be often
 * used with "stopRecursion()" on first call of "cb". Therefore
 * filling a stack with every Go would be a waste of time and space.
 * BUT, couterpart of that is that "cb" cannot modify the tree hierarchy.
 */
void* Go::forEveryGoContainingPoint
    (void cb(void*, Go*), void* closure,
     const GPoint& abspt, GNumber maxdist)
{
    if (!mapped()) // an unmapped go contains nothing;
	return NULL;

    RecursionStack().push(NULL); // recursion return value;
    RecursionStack().push((void*)GFALSE); // stop recursion ?;

    if (crossType() != GCROSS_AS_LEAF)
	Child()->for_every_go_containing_point_recur
		    (cb, closure, abspt, maxdist, 0, GTRUE);

    if (!RecursionStack().top() && contains (abspt, maxdist))
	cb (closure, this);

    RecursionStack().pop();
    return RecursionStack().pop();
}

void* Go::ForEveryGoContainingPoint
    (void cb(void*, Go*), void* closure,
     const GPoint& abspt, GNumber maxdist)
{
    if (!mapped()) // an unmapped go contains nothing;
	return NULL;

    RecursionStack().push(NULL); // recursion return value;
    RecursionStack().push((void*)GFALSE); // stop recursion ?;

    Child()->for_every_go_containing_point_recur
	(cb, closure, abspt, maxdist, 0, GFALSE);

    if (!RecursionStack().top() && Contains(abspt, maxdist))
	cb(closure, this);

    RecursionStack().pop();
    return RecursionStack().pop();
}


/* the ``CallbackStruct'' is used by the following two functions */
struct CallbackStruct {
    void(*f)(void*, GoSensor*);
    void* closure;
};


void* Go::ForEverySensorContainingPoint
    (void(*cb)(void*, GoSensor*), void* closure,
     const GPoint& abspt, GNumber maxdist)
{
    if (!mapped())
	return NULL;

    RecursionStack().push(NULL); //recursion return value;
    RecursionStack().push((void*)GFALSE); //stop recursion ?;

    Child()->for_every_sensor_containing_point_recur
	(cb, closure, abspt, maxdist, 0);

    if (!RecursionStack().top() && is_sensor() && Contains(abspt, maxdist))
	cb(closure, (GoSensor*)this);

    RecursionStack().pop();
    return RecursionStack().pop();
}


void Go::for_every_sensor_containing_point_recur
    (void (*cb)(void*, GoSensor*), void* closure,
     const GPoint& abspt, GNumber maxdist, int this_child_index)
{
    if (this == NULL)
	return;

    Next(this_child_index)->for_every_sensor_containing_point_recur
	(cb, closure, abspt, maxdist, this_child_index + 1);

    if (_mappable && !RecursionStack().top())
    {
	Child()->for_every_sensor_containing_point_recur
	    (cb, closure, abspt, maxdist, 0);
	if (!RecursionStack().top() && is_sensor() && Contains(abspt, maxdist))
	    cb(closure, (GoSensor*)this);
    }
}


void Go::for_every_go_containing_point_recur
    (void cb(void*, Go*), void* closure,
     const GPoint& abspt, GNumber maxdist, int this_child_index,
     GBool public_only)
{
    if (this == NULL)
	return;

    Next(this_child_index)->for_every_go_containing_point_recur
	(cb, closure, abspt, maxdist, this_child_index + 1, public_only);

    if (_mappable && !RecursionStack().top())
    {
	if (public_only && crossType() == GCROSS_AS_NOTHING)
	    return;
	
	STATIC_OBJECT(GRegion, region);
	region.setEmpty();
	addClip(region);
	if (region.contains(abspt, maxdist))
	{
	    if (!public_only || crossType() != GCROSS_AS_LEAF)
		Child()->for_every_go_containing_point_recur
		    (cb, closure, abspt, maxdist, 0, public_only);
	    if (!RecursionStack().top())
	    {
		if (public_only && contains (abspt, maxdist) ||
		    !public_only && Contains (abspt, maxdist))
		    cb (closure, this);
	    }
	}
    }
}


/*
 * Call proc on every selected go (of this go subtree) that contains `abspt'.
 * When such a go is found, the recursion is not done on this go subtree.
 * Proc is called with two parameters:
 *	proc(closure, go)
 *
 * See the Nota of the method `Go::forEveryGo()'.
 * See the Nota of the method forEveryGoContainingPoint(...).
 */
void* Go::forEverySelectedSubtreeContainingPoint (void cb(void*, Go*), void* closure,
						  const GPoint& abspt, GNumber maxdist)
{
    if (!mapped()) // an unmapped go contains nothing;
	return NULL;

    RecursionStack().push(NULL); // recursion return value;
    RecursionStack().push((void*)GFALSE); // stop recursion ?;

    for_every_selected_subtree_containing_point_recur
	(cb, closure, abspt, maxdist);

    RecursionStack().pop();
    return RecursionStack().pop();
}

void Go::for_every_selected_subtree_containing_point_recur
    (void cb(void*, Go*), void* closure, const GPoint& abspt, GNumber maxdist)
{
    if (_mappable)
    {
	STATIC_OBJECT(GRegion, region);
	region.setEmpty();
	addClip(region);
	if (region.contains(abspt, maxdist))
	{
	    if (_selected && Contains(abspt, maxdist))
		cb(closure, this);
	    else if (crossType() != GCROSS_AS_LEAF)
	    {
		Go* child = Child();
		for (int i = 0;
		     child && !RecursionStack().top();
		     child = child->Next(i++))
		{
		    if (child->crossType() != GCROSS_AS_NOTHING)
			child->
			    for_every_selected_subtree_containing_point_recur
				(cb, closure, abspt, maxdist);
		}
	    }
	}
    }
}


/*
 * This method will stop the recursion started by the last `forEvery...()' method.
 * `return_value' will be the value returned by the `forEvery...()' method.
 */
void Go::stopRecursion (void* return_value)
{
    RecursionStack().pop(); // stop recursion ?
    RecursionStack().top(return_value); // recursion value;
    RecursionStack().push((void*)GTRUE);
}


void Go::rel_reparent_setup_abscs
    (Go* oldpar, Go* newpar, GCoordSys& /*oldabs*/, GCoordSys& /*oldrel*/)
{
    /*
     * oldabs = oldparabs * oldrel
     * newabs = newparabs * newrel
     * 	and newrel == oldrel
     * ==>
     *	newabs = newparabs * ~oldparabs * oldabs
     */
    GCoordSys newparabs;
    if (newpar) newparabs = newpar->Abscs();

    GCoordSys mulLeft =
	newparabs * ~ ((oldpar) ? oldpar->Abscs() : GUnitCoordSys());

    relrep_setcs_on_subtree(mulLeft, newparabs);
}

void Go::relrep_setcs_on_subtree (GCoordSys& mulLeft, GCoordSys& parabs)
{
    GCoordSys newabs;

    if (d_store_abscs())
    {
	newabs = mulLeft * d_get_abscs(&parabs);
	d_set_coordsys(Relcs(parabs), newabs, parabs);
    }
    else
	newabs = parabs * d_get_relcs(&parabs);

    Go* child = Child();
    for (int i = 0; child; child = child->Next(i++))
	child->relrep_setcs_on_subtree(mulLeft, newabs);
}


GBool Go::call_interf_preconditions_on_subtree (const GoInterf* newinterf) const
{
    
    if (!call_interf_precond(newinterf) ||
	!call_class_interf_precond(newinterf))
	return GFALSE;

    Go* child = Child();
    for (int i = 0; child; child = child->Next(i++))
	if (!child->call_interf_preconditions_on_subtree(newinterf))
	    return GFALSE;

    return GTRUE;
}


void Go::call_interf_postconditions_on_subtree (GoInterf* oldinterf)
{
    assert(differ_destroy);
    
    call_interf_postcond(oldinterf);
    call_class_interf_postcond(oldinterf);

    Go* child = Child();
    for (int i = 0; child; child = child->Next(i++))
	child->call_interf_postconditions_on_subtree(oldinterf);

}


void Go::force_reparent (Go* newpar, Go* newnext)
{
    Go* oldpar = parent();

    int oldix = (oldpar) ? oldpar->childIndex(this) : 0;
    int newix = (newpar) ? newpar->childIndex(newnext) : 0;
    do_reparent(oldix, oldpar, Next(oldix), newix, newpar, newnext);
}


GBool Go::is_root () const
{
    return GFALSE;
}

GBool Go::is_sensor () const
{
    return GFALSE;
}


/*
 * Always-first/last management.
 */
class FirstLastInfo
{
  public:
    FirstLastInfo (Go* parent);
    static FirstLastInfo* get (Go* parent);
    GBool add (Go* child, GBool first);
    void rem (Go* child);
    
  private:
    static GAssoc* _assoc ();

    static GAssoc* assoc_ptr;
    Go* _parent;
    GSet first_set;
    GSet last_set;

    ~FirstLastInfo ();

    friend void first_last_info_add_child_post (void*, Go*, Go*);
};

GAssoc* FirstLastInfo::assoc_ptr = NULL;

GAssoc* FirstLastInfo::_assoc ()
{
    if (assoc_ptr == NULL)
	assoc_ptr = new GAssoc;

    return assoc_ptr;
}

void first_last_info_add_child_post (void* closure, Go*, Go* child)
{
    FirstLastInfo* info = (FirstLastInfo*) closure;

    if (info->first_set.contains (child))
    {
	if (child->Previous() == NULL ||
	    !info->first_set.contains (child->Previous()))
	    child->reparent (info->_parent, info->_parent->Child());
    }
    else if (info->last_set.contains (child))
    {
	if (child->Next() == NULL ||
	    !info->last_set.contains (child->Next()))
	    child->reparent (info->_parent);
    }
    else
    {
	if (info->first_set.contains (child->Next()))
	{
	    for (Go* goodNext = child->Next();
		 info->first_set.contains (goodNext);
		 goodNext = goodNext->Next());
	    child->reparent (info->_parent, goodNext);
	}
	else if (info->last_set.contains (child->Previous()))
	{
	    for (Go* goodNext = info->_parent->Child();
		 !info->last_set.contains (goodNext);
		 goodNext = goodNext->Next());
	    child->reparent (info->_parent, goodNext);
	}
    }
}

FirstLastInfo::FirstLastInfo (Go* parent)
{
    _parent = parent;
    assoc_ptr->set (parent, this);

    parent->addChildPostcond (first_last_info_add_child_post, this);
}

FirstLastInfo* FirstLastInfo::get (Go* parent)
{
    return (FirstLastInfo*) _assoc()->get (parent);
}

void FirstLastInfo::rem (Go* child)
{
    first_set.remove (child);
    last_set.remove (child);
}


GBool FirstLastInfo::add (Go* child, GBool first)
{
    if (first)
    {
	if (first_set.contains (child))
	    return GTRUE;
	else if (last_set.contains (child))
	{
	    GWarning ("GoTree.cc:GoAlwaysFirst(go): go is already always last!");
	    return GFALSE;
	}
	first_set.add (child);

	if (!first_set.contains (child->Previous()))
	    child->reparent (_parent, _parent->Child());
    }
    else
    {
	if (last_set.contains (child))
	    return GTRUE;
	else if (first_set.contains (child))
	{
	    GWarning ("GoTree.cc:GoAlwaysLast(go): go is already always first!");
	    return GFALSE;
	}
	last_set.add (child);
	if (!last_set.contains (child->next()))
	    child->reparent (_parent);
    }

    return GTRUE;
}

void go_first_last_reparent_post (Go* go, Go* oldParent, GBool first)
{
    Go* newParent = go->parent();

    if (oldParent == newParent)
	return;

    FirstLastInfo* info = FirstLastInfo::get (oldParent);
    if (info)
	info->rem (go);

    if (newParent)
    {
	info = FirstLastInfo::get (newParent);
	if (info == NULL)
	    info = new FirstLastInfo (newParent);
	info->add (go, first);
    }
}

void go_first_reparent_post (void*, Go* go, Go* oldParent, Go*)
{
    go_first_last_reparent_post (go, oldParent, GTRUE);
}

void go_last_reparent_post (void*, Go* go, Go* oldParent, Go*)
{
    go_first_last_reparent_post (go, oldParent, GFALSE);
}

void first_last_common (Go* go, GAddOrRemove add, GBool first)
{
    Go* parent = go->parent();
    FirstLastInfo* info = FirstLastInfo::get (parent);

    if (add == GREMOVE)
    {
	void (*proc) (void*, Go*, Go*, Go*) = go_first_reparent_post;
	if (!first)
	    proc = go_last_reparent_post;
	go->reparentPostcond (proc, NULL, GREMOVE);
	if (info)
	    info->rem (go);
    }
    else //add == GADD
    {
	if (parent)
	{
	    if (info == NULL)
		info = new FirstLastInfo (parent);
	    if (!info->add (go, first))
		return;
	}

	void (*proc) (void*, Go*, Go*, Go*) = go_first_reparent_post;
	if (!first)
	    proc = go_last_reparent_post;

	/* to avoid setting the postcond twice */
	go->reparentPostcond (proc, NULL, GREMOVE);

	go->reparentPostcond (proc, NULL);
    }
}



void GoAlwaysFirstChild (Go* go, GAddOrRemove add)
{
    first_last_common (go, add, GTRUE);
}

	
void GoAlwaysLastChild (Go* go, GAddOrRemove add)
{
    first_last_common (go, add, GFALSE);
}
