#ifdef REV_INFO
static char rcsid[] = "$Id: mGo.cc,v 1.3 1992/12/10 13:59:17 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  Go Kernel
FILE:    mGo.cc
AUTHOR:  Oliver Jojic
DATE:    April 91

DESCRIPTION:
  See the mGo.h header.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <Go/mGo.h>
#include <Go/Go.h>
#include <assert.h>

static GoMetaclass* go_metaclass = NULL;
GoMSet* GoMetaclass::meta_classes = NULL;

/*
 * The GoMetaclass constructor.
 * It put this metaclass in the metaclass_set.
 */
GoMetaclass::GoMetaclass ()
{
    if (meta_classes == NULL)
	meta_classes = new GoMSet();

    meta_classes->add(this);

    has_create_precond = GFALSE;
    has_dup_precond = GFALSE;
    has_destroy_precond = GFALSE;
    has_save_precond = GFALSE;
    has_shift_precond = GFALSE;
    has_reparent_precond = GFALSE;
    has_remove_child_precond = GFALSE;
    has_add_child_precond = GFALSE;
    has_root_precond = GFALSE;
    has_interf_precond = GFALSE;
    has_cross_type_precond = GFALSE;
    has_relcs_precond = GFALSE;
    has_abscs_precond = GFALSE;
    has_map_precond = GFALSE;
    has_select_precond = GFALSE;

    has_create_postcond = GFALSE;
    has_destroy_postcond = GFALSE;
    has_save_postcond = GFALSE;
    has_restore_postcond = GFALSE;
    has_dup_postcond = GFALSE;
    has_shift_postcond = GFALSE;
    has_reparent_postcond = GFALSE;
    has_remove_child_postcond = GFALSE;
    has_add_child_postcond = GFALSE;
    has_root_postcond = GFALSE;
    has_interf_postcond = GFALSE;
    has_cross_type_postcond = GFALSE;
    has_relcs_postcond = GFALSE;
    has_abscs_postcond = GFALSE;
    has_map_postcond = GFALSE;
    has_select_postcond = GFALSE;
    has_my_clip_postcond = GFALSE;
    has_clip_postcond = GFALSE;
}


GoMetaclass* Go::metaclass ()
{
    if (go_metaclass == NULL)
	go_metaclass = new GoMetaclass;

    return go_metaclass;
}

/*
 * The GoMetaclass destructor.
 * 	- delete all sub-metaclasses.
 * 	- delete all corresponding go class instances.
 */
GoMetaclass::~GoMetaclass ()
{
    GoMetaclass* sub;
    while (sub = sub_classes.first())
	delete sub;

    Go* go;
    while (go = _instances.first())
    {
	assert (go->getMetaclass() == this);
	if (!go->destroy())
	    GError("application error: delete of metaclass %s cannot destroy the instance %x",
		  go->getMetaclass()->classname(), go);
    }

    if (this == go_metaclass)
	go_metaclass = NULL;
}


/*
 * Return TRUE if this is a subclass of "metaclass" (or is equal to).
 */
GBool GoMetaclass::isSubclass (const GoMetaclass* metaclass) const
{
    for (const GoMetaclass* sup = this; sup; sup = sup->superclass())
	if (sup == metaclass)
	    return GTRUE;

    return GFALSE;
}


/*
 * Virtual m_classname method.
 * Each GoSubMetaclass must redefine this method.
 */
const char* GoMetaclass::m_classname () const
{
    return "Go";
}


/*
 * Virtual m_superclass method.
 * Each GoSubMetaclass must redefine this method.
 */
GoMetaclass* GoMetaclass::m_superclass () const
{
    return NULL;
}


static void _dup (Go* go)
{
    Go* copy = go->dup();
    if (copy)
	copy->reparent((Go*)go->root());
}

static void go_destroy (Go* go) { if (go) go->destroy(); }
static void go_raise (Go* go) { if (go) go->raise(); }
static void go_lower (Go* go) { if (go) go->lower(); }
static void go_leaf (Go* go) { if (go) go->leaf(GTRUE); }
static void go_unleaf (Go* go) { if (go) go->leaf(GFALSE); }
static void go_map (Go* go) { if (go) go->map(); }
static void go_unmap (Go* go) { if (go) go->unmap(); }
static void go_select (Go* go) { if (go) go->select(); }
static void go_unselect (Go* go) { if (go) go->unselect(); }


GActionRec go_actions[] =
{
  { "destroy()", 	(GActionProc) go_destroy },
  { "raise()", 		(GActionProc) go_raise },
  { "lower()",		(GActionProc) go_lower },
  { "leaf()",		(GActionProc) go_leaf },
  { "unleaf()",		(GActionProc) go_unleaf },
  { "map()",		(GActionProc) go_map },
  { "unmap()",		(GActionProc) go_unmap },
  { "select()",		(GActionProc) go_select },
  { "unselect()",	(GActionProc) go_unselect },
  { "dup()",		(GActionProc) _dup },
  { NULL, NULL },
};

/*
 * Virtual m_actions method.
 * Each GoSubMetaclass must redefine this method.
 */
const GActionRec* GoMetaclass::m_actions () const
{
    return go_actions;
}

Go* GoMetaclass::m_create () const
{
    GError("GoMetaclass::m_create()");
    return NULL;
}

GBool GoMetaclass::AddOrRemoveCondition
    (const char* type, void* cb, void* closure, GAddOrRemove add)
{
    GCallbackRec rec((GCBProc)cb, closure);
    return
	0 < ((add == GADD)
	     ? _assoc.add(type, sizeof rec, &rec)
	     : _assoc.rem(type, sizeof rec, &rec));
}

GBool GoMetaclass::CallPreconditions (const char* type, const Go* go) const
{
    int count;
    GCallbackRec* pcb = (GCallbackRec*)_assoc.get(type, count);

    for (int i = 0; i < count; i++, pcb++)
	if (!((GBool (*)(void*, const Go*))pcb->proc)
	    (pcb->closure, go))
	    return GFALSE;

    return GTRUE;
}

GBool GoMetaclass::CallPreconditions
    (const char* type, const Go* go, void* a) const
{
    int count;
    GCallbackRec* pcb = (GCallbackRec*)_assoc.get(type, count);

    for (int i = 0; i < count; i++, pcb++)
	if (!((GBool (*)(void*, const Go*, void*))pcb->proc)
	    (pcb->closure, go, a))
	    return GFALSE;

    return GTRUE;
}

GBool GoMetaclass::CallPreconditions
    (const char* type, const Go* go, void* a, void* b) const
{
    int count;
    GCallbackRec* pcb = (GCallbackRec*)_assoc.get(type, count);

    for (int i = 0; i < count; i++, pcb++)
	if (!((GBool (*)(void*, const Go*, void*, void*))pcb->proc)
	    (pcb->closure, go, a, b))
	    return GFALSE;

    return GTRUE;
}

GBool GoMetaclass::CallPreconditions
    (const char* type, const Go* go, void* a, void* b, void* c) const
{
    int count;
    GCallbackRec* pcb = (GCallbackRec*)_assoc.get(type, count);

    for (int i = 0; i < count; i++, pcb++)
	if (!((GBool (*)(void*, const Go*, void*, void*, void*))pcb->proc)
	    (pcb->closure, go, a, b, c))
	    return GFALSE;

    return GTRUE;
}

GBool GoMetaclass::CallPreconditions
    (const char* type, const Go* go, void* a, void* b, void* c, void* d) const
{
    int count;
    GCallbackRec* pcb = (GCallbackRec*)_assoc.get(type, count);

    for (int i = 0; i < count; i++, pcb++)
	if (!((GBool (*)(void*, const Go*, void*, void*, void*, void*))pcb->proc)
	    (pcb->closure, go, a, b, c, d))
	    return GFALSE;

    return GTRUE;
}

void GoMetaclass::CallPostconditions (const char* type, Go* go)
{
    int count;
    GCallbackRec* pcb = (GCallbackRec*)_assoc.get(type, count);

    GCallCallbacks(pcb, count, go);
}

void GoMetaclass::CallPostconditions (const char* type, Go* go, void* a)
{
    int count;
    GCallbackRec* pcb = (GCallbackRec*)_assoc.get(type, count);

    GCallCallbacks(pcb, count, go, a);
}

void GoMetaclass::CallPostconditions
    (const char* type, Go* go, void* a, void* b)
{
    int count;
    GCallbackRec* pcb = (GCallbackRec*)_assoc.get(type, count);

    GCallCallbacks(pcb, count, go, a, b);
}

void GoMetaclass::CallPostconditions
    (const char* type, Go* go, void* a, void* b, void* c)
{
    int count;
    GCallbackRec* pcb = (GCallbackRec*)_assoc.get(type, count);

    GCallCallbacks(pcb, count, go, a, b, c);
}

void GoMetaclass::CallPostconditions
    (const char* type, Go* go, void* a, void* b, void* c, void* d)
{
    int count;
    GCallbackRec* pcb = (GCallbackRec*)_assoc.get(type, count);

    GCallCallbacks(pcb, count, go, a, b, c, d);
}

GoMetaclass* Go::metaclass (const char* classname)
{
    for (GoMetaclass* mc = GoMetaclass::metaclasses().first();
	 mc;
	 mc = GoMetaclass::metaclasses().next())
	if (!strcmp(mc->classname(), classname))
	    return mc;
    return NULL; }

GoMetaclass* Go::m_metaclass () const
{
    return Go::metaclass();
}
