#ifdef REV_INFO
static char rcsid[] = "$Id: Array.cc,v 1.3 1992/12/10 13:32:49 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  Tools
FILE:    Array.cc
AUTHOR:  Philippe Smadja
DATE:    Dec 90

DESCRIPTION:
    Implementation of an array of elements. The elemsize
    is constant.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <assert.h>
#include <Go/Tools/Array.h>
#include <Go/Tools/Alloc.h>

/**
FUNCTION
    Constructors
DESCRIPTION

**/
GArray::GArray(int elemsize, int incr)
{
    _data = NULL ;
    _nelems = 0;
    _nelemsAvailable = 0;
    _elemsize = elemsize;
    _incr = incr < 1 ? 1 : incr;
    _firstIndex = 0;
}

GArray::GArray(const void* value, int nelems, int elemsize, int incr)
{
    assert (elemsize);
    _nelems = nelems;
    _elemsize = elemsize;
    _incr = incr < 1 ? 1 : incr;
    _firstIndex = 0;
    _nelemsAvailable   = GNB_AVAILABLE(nelems);

    if (_nelemsAvailable == 0)
	_data = NULL;
    else {
	_data = (uchar*)GMALLOC(_nelemsAvailable * _elemsize);
	if (value)
	    bcopy(value, _data, _nelems * _elemsize);
    }
}


/**
FUNCTION
    destructor.
DESCRIPTION
    delete the allocated data    
**/
GArray::~GArray()
{
    GFREE(_data);
}


/**
FUNCTION
    add
DESCRIPTION
    add nelems elems at index of the array. -1 means at the end.
**/
void GArray::add(const void* value, uint nelems_to_add, int index)
{
    if (nelems_to_add == 0)
	return;
    
    GBool alloc = GFALSE;
    int sz_to_add = nelems_to_add * _elemsize;
    uint n = GNB_AVAILABLE(nelems_to_add);

    if (_nelemsAvailable <  n) {
	_nelemsAvailable = n;
	alloc = GTRUE;
    }

    if (index == -1 || index == _nelems) {
	/**
	  put the new elem at the end
	  **/
	if (alloc)
	    GREALLOC(_data, _nelemsAvailable * _elemsize);

	if (GLAST_INDEX + nelems_to_add > _nelemsAvailable) {
	    /**
	      move to the top
	      **/
	    int oldfi = _firstIndex;
	    do  {
		assert (_firstIndex != 0);
		_firstIndex /= 2;
	    } while (GLAST_INDEX + nelems_to_add > _nelemsAvailable);
	    bcopy(_data + oldfi*_elemsize,
		  _data + _firstIndex*_elemsize,
		  _nelems*_elemsize);
	}
	/**
	  copy the new value
	  **/
	bcopy(value, _data + GLAST_INDEX*_elemsize, sz_to_add);

    } else {
	if (!alloc && index == 0) {
	    if (_firstIndex < nelems_to_add)  {
		/**
		  move to the bottom
		  **/
		int oldfi = _firstIndex;
		do {
		    assert (_firstIndex < _incr);
		    _firstIndex += _incr/2;
		} while (_firstIndex < nelems_to_add);
		while(GNB_AVAILABLE(_firstIndex) != n)
		    _firstIndex--;
		assert (_firstIndex >= nelems_to_add);
		bcopy(_data + oldfi*_elemsize,
		      _data + _firstIndex*_elemsize,
		      _nelems*_elemsize);
	    } 
	    _firstIndex -= nelems_to_add;
	    bcopy(value, _data + _firstIndex*_elemsize, sz_to_add);

	} else {
	    int s1 = index * _elemsize;
	    uchar* olddata = _data;
	    int deb = _firstIndex*_elemsize;
	    int newdeb = deb;

	    if (alloc) 
		_data = (uchar*)GMALLOC(_nelemsAvailable * _elemsize);

	    if (GLAST_INDEX+nelems_to_add > _nelemsAvailable) {
		/**
		  move to the top the data
		  from _firstIndex to _firstIndex + index
		  **/
		_firstIndex = 0;
		newdeb = 0;
	    } 
	    if (alloc || newdeb != deb)
		bcopy(olddata + deb, _data + newdeb,  s1);

	    bcopy(olddata + deb + s1, _data + newdeb + s1 + sz_to_add,
		 (_nelems-index)*_elemsize);
	    bcopy(value, _data + newdeb + s1, sz_to_add);

	    if (alloc)
		GFREE(olddata);

	}
    }
    _nelems += nelems_to_add;
    assert (GLAST_INDEX <= _nelemsAvailable);
}

/**
FUNCTION
    GArray::rem (index, nelems_to_free)
DESCRIPTION
    remove nelems from the index
    the index is from 0 to _nelems-1
    index is -1 means free the last elem.
**/

void GArray::rem(int index, uint nelems_to_free)
{
    if (nelems_to_free == 0)
	return;
    
    if (index == -1)
	index = _nelems - 1;
    assert (index + nelems_to_free <= _nelems);
    if (nelems_to_free == _nelems) {
	GFREE(_data);
	_nelems = _nelemsAvailable = 0;
	_firstIndex = 0;
	return;
    }


    uint n = GNB_AVAILABLE(-nelems_to_free);

    if (index == 0 && _firstIndex + _nelems <= n)
	_firstIndex += nelems_to_free;
    else {
	int size = _elemsize*(_nelems-(index+nelems_to_free));
	int dstfi = _firstIndex;
	if (_nelemsAvailable > n) {
	    dstfi = 0;
	    if (index)
		bcopy(_data + _elemsize*_firstIndex,
		      _data, _elemsize*index);
	}
	if (size)
	    bcopy(_data + _elemsize*(index+_firstIndex+nelems_to_free),
		  _data + _elemsize*(index+ dstfi),
		  size);
	_firstIndex = dstfi;
    }

    if (_nelemsAvailable > n) {
	_nelemsAvailable = n;
	GREALLOC(_data, _nelemsAvailable * _elemsize);
    }
    _nelems -= nelems_to_free;
    assert (GLAST_INDEX <= _nelemsAvailable);
}

/**
FUNCTION
    GArray::rem(const void* data, GBool only_first_element)
DESCRIPTION
    remove from the array the 1er element or all elements matching *value.
**/

void GArray::rem(const void* value, GBool only_first_element)
{
    uchar* Data;
    if (only_first_element) {
	Data = (uchar*)data();
	for(uint i =0; i < _nelems; i++, Data += _elemsize)
	    if (memcmp(value, Data, _elemsize) == 0) {
		rem(i,1);
		return;
	    }
    } else {
	GBool end = GFALSE;
	while (!end) {
	    Data = (uchar*)data();
	    for (uint i = 0; i < _nelems; i++, Data += _elemsize)
		if (memcmp(value, Data, _elemsize) == 0) {
		    rem(i,1);
		    break;
		} else
		    if (i == _nelems - 1)
			end = GTRUE;
	}
    }
}

/**
FUNCTION
    GPointerArray::add(const void* value, index = -1);
DESCRIPTION
    add the void* to the array at index. -1 means at the end
**/
void GPointerArray::add(const void* value, int index)
{
#define nelems_to_add 1
#define sz_to_add _elemsize

    GBool alloc = GFALSE;
    uint n = GNB_AVAILABLE(nelems_to_add);

    if (_nelemsAvailable <  n) {
	_nelemsAvailable = n;
	alloc = GTRUE;
    }

    if (index == -1 || index == _nelems) {
	/**
	  put the new elem at the end
	  **/
	if (alloc)
	    GREALLOC(_data, _nelemsAvailable * _elemsize);

	if (GLAST_INDEX + nelems_to_add > _nelemsAvailable) {
	    /**
	      move to the top
	      **/
	    int oldfi = _firstIndex;
	    do  {
		assert (_firstIndex != 0);
		_firstIndex /= 2;
	    } while (GLAST_INDEX + nelems_to_add > _nelemsAvailable);
	    bcopy(_data + oldfi*_elemsize,
		  _data + _firstIndex*_elemsize,
		  _nelems*_elemsize);
	}
	/**
	  copy the new value
	  **/
	*(ulong*)(_data + GLAST_INDEX*_elemsize) = (ulong)value;

    } else {
	if (!alloc && index == 0) {
	    if (_firstIndex < nelems_to_add)  {
		/**
		  move to the bottom
		  **/
		int oldfi = _firstIndex;
		do {
		    assert (_firstIndex < _incr);
		    _firstIndex += _incr/2;
		} while (_firstIndex < nelems_to_add);
		while(GNB_AVAILABLE(_firstIndex) != n)
		    _firstIndex--;
		assert (_firstIndex >= nelems_to_add);
		bcopy(_data + oldfi*_elemsize,
		      _data + _firstIndex*_elemsize,
		      _nelems*_elemsize);
	    } 
	    _firstIndex -= nelems_to_add;
	    /**
	      COPY the value
	      **/
	    *(ulong*)(_data + _firstIndex*_elemsize) = (ulong)value;

	} else {
	    int s1 = index * _elemsize;
	    uchar* olddata = _data;
	    int deb = _firstIndex*_elemsize;
	    int newdeb = deb;

	    if (alloc) 
		_data = (uchar*)GMALLOC(_nelemsAvailable * _elemsize);

	    if (GLAST_INDEX+nelems_to_add > _nelemsAvailable) {
		/**
		  move to the top the data
		  from _firstIndex to _firstIndex + index
		  **/
		_firstIndex = 0;
		newdeb = 0;
	    } 
	    if (alloc || newdeb != deb)
		bcopy(olddata + deb, _data + newdeb,  s1);

	    bcopy(olddata + deb + s1, _data + newdeb + s1 + sz_to_add,
		 (_nelems-index)*_elemsize);

	    /**
	      COPY the value
	      **/
	    *(ulong*)(_data + newdeb + s1) = (ulong)value;

	    if (alloc)
		GFREE(olddata);

	}
    }
    _nelems += nelems_to_add;
    assert (GLAST_INDEX <= _nelemsAvailable);
#undef nelems_to_add
#undef sz_to_add
}    


/**
FUNCTION
    GPointerArray::rem(const void* value, GBool only_first_element)
DESCRIPTION
    remove the array the 1er value or all values equal to value
**/
void GPointerArray::rem(const void* value, GBool only_first_element)
{
    ulong* Data;
    if (only_first_element)
    {
	Data = (ulong*)data();
	for(uint i =0; i < _nelems; i++)
	    if (value == (void*)*Data++){
		rem(i,1);
		return;
	    }
    }
    else
    {
	/**
	  compress the data to the top
	  **/
	uint i;
	int index = 0;
	Data = (ulong*)data(); /* the first available data */
	int nelems_to_remove = 0;
	for (i = 0; i < _nelems; i++)
	    if ((ulong)value == *Data++)
	    {
		nelems_to_remove++;
		index = i;
	    }
	
	if (nelems_to_remove == 0) return;
	else if (nelems_to_remove == 1)
	{
	    rem(index,1);
	    return;
	}
	/**
	  make a new data
	  **/
	ulong* _newData = (ulong*)GMALLOC(GNB_AVAILABLE(-nelems_to_remove)*
					  sizeof(void*));
	if (_newData == NULL)
	{
	    assert (_nelems == nelems_to_remove);
	}
	else
	{
	    ulong* newData = _newData;
	    Data = (ulong*)data();
	
	    for (i=0; i < _nelems; i++)
		if ((*newData = *Data++) != (ulong)value)
		    newData++;
	}
	
	_firstIndex = 0;
	_nelems -= nelems_to_remove;
	_nelemsAvailable = GNB_AVAILABLE(0);
	GFREE(_data);
	_data = (uchar*)_newData;
    }
}





