#ifdef REV_INFO
static char rcsid[] = "$Id: Assoc.cc,v 1.6 1992/09/25 14:02:47 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Tools
FILE:    Assoc.c
AUTHOR:  Olivier Jojic
DATE:    10 June 1989

DESCRIPTION:
    See <Go/Tools/Assoc.h>

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/machdef.h>
#include <assert.h>
#include <Go/Tools/Assoc.h>
#include <Go/Tools/Alloc.h>


class Entry
{
    ushort unit;
    ushort size;
    ushort count;
    char*  buf;

    void  incrSize ();
    void  incrSize (int newSize);

  public:
    const void* key;
    int elemSize () { return unit; }

    Entry(const void* tp);
    ~Entry();

    inline int getNumElem() { return count; };
    void  set (const void* value);
    void  set (int valueSize, const void* valueAddr);
    int   add (const void* elem);
    int   add (int elemSize, const void* elemAddr);
    void* get ();
    const void** get (int& numElems);
    void  rem ();
    int   rem (const void* elem);
    int   rem (int elemSize, const void* elemAddr);
};


class GLinearAssoc
{
    ushort  size;
    ushort  numEntries;
    Entry** entries;

    int keyToIndex (const void* key);
    int forceKeyToIndex (const void* key);
    void incrSize();

  public:
    GLinearAssoc();
    ~GLinearAssoc();

    int getNumEntries () { return numEntries; }
    Entry* entry(int i) { return entries[i]; }

    const void** getAndRemove (const void*& key, int& elemSize, int& numElems);

    void  set (const void* key, const void* value);
    void  set (const void* key, int size, const void* addr);
    int   add (const void* key, const void* elem);
    int   add (const void* key, int size, const void* addr);
    void* get (const void* key);
    const void** get (const void* key, int& numElems);
    void  rem (const void* key);
    int   rem (const void* key, const void* elem);
    int   rem (const void* key, int size, const void* addr);
};



Entry::Entry (const void* tp)
{
    key = tp;
    unit = 1;
    size = count = 0;
    buf = NULL;
}


Entry::~Entry ()
{
    if (size)
    {
	delete buf;
	buf = NULL;
    }
}


void Entry::incrSize ()
{
    size += 8 * unit;
    char* old = buf;
    buf = (char*)GMALLOC(size);
    memcpy(buf, old, count);
    GFREE(old);
}


void Entry::incrSize (int newSize)
{
    assert(count <= newSize);
    char* old = buf;
    size = newSize;
    buf = (char*)GMALLOC(size);
    memcpy(buf, old, count);
    GFREE(old);
}


void Entry::set (const void* value)
{
    unit = sizeof(const void*);
    count = 0;
    if (size < unit)
	incrSize (unit);
    count = unit;
    *((void**)buf) = (void*)value;
}


void Entry::set (int valueSize, const void* valueAddr)
{
    unit = valueSize;
    count = 0;
    if (size < unit)
	incrSize(unit);
    count = unit;
    memcpy(buf, valueAddr, unit);
}


int Entry::add (const void* elem)
{
    if (!count) unit = sizeof(long);
    else 	assert(unit == sizeof(long));

    if (count + unit > size)
	incrSize();
    *((void**)(buf+count)) = (void*)elem;
    count += unit;
    return count / unit;
}


int Entry::add (int elemSize, const void* elemAddr)
{
    if (!count)	unit = elemSize;
    else	assert(unit == elemSize);

    if (count + unit > size)
	incrSize();
    memcpy(buf+count, elemAddr, unit);
    count += unit;
    return count / unit;
}


void* Entry::get ()
{
    if (!count)
	return 0;
    assert(unit == sizeof(long));
    return (count) ? *((void**)(buf+count-unit)) : NULL;
}


const void** Entry::get (int& numElems)
{
    numElems = count / unit;
    if (numElems == 0)
	return NULL;
    return (const void**)buf;
}


void Entry::rem ()
{
    if (size)
	delete buf;
    buf = NULL;
    unit = 1;
    size = count = 0;
}


int Entry::rem (const void* elem)
{
    if (!count)
	return 0;
    assert(unit == sizeof(void*));
    for (int i = 0; i < count; i += unit)
	if (elem == *((void**)(buf+i)))
	{
	    count -= unit;
	    memcpy(buf + i, buf + i + unit, count - i);
	    break;
	}

    return count / unit;
}


int Entry::rem (int elemSize, const void* elemAddr)
{
    if (!count)
	return 0;
    assert(unit == elemSize);
    for (int i = 0; i < count; i += unit)
	if (!memcmp(elemAddr, buf+i, unit))
	{
	    count -= unit;
	    memcpy(buf+i, buf+i+unit, count - i);
	    break;
	}

    return count / unit;
}


GLinearAssoc::GLinearAssoc ()
{
    size = numEntries = 0;
    entries = 0;
}


GLinearAssoc::~GLinearAssoc ()
{
    for (int i = 0; i < numEntries; i++)
	delete entries[i];
    if (size)
	delete entries;
}


int GLinearAssoc::keyToIndex (const void* key)
{
    int lastEmpty = -1;

    for (int i = 0; i < numEntries; i++)
	if (entries[i]->key == key)
	    return i;
	else if (entries[i]->getNumElem() == 0)
	    lastEmpty = i;

    if (lastEmpty == -1)
	return -1;
    else
    {
	entries[lastEmpty]->key = key;
	return lastEmpty;
    }
}


void GLinearAssoc::incrSize ()
{
    const int incrUnit = 8;

    size += incrUnit;
    Entry** tmp = entries;
    entries = (Entry**)GMALLOC(size * sizeof(Entry*));
    memcpy(entries, tmp, numEntries * sizeof(Entry*));
    GFREE(tmp);
}


int GLinearAssoc::forceKeyToIndex (const void* key)
{
    int index = keyToIndex(key);
    if (index == -1)
    {
	if (numEntries >= size)
	    incrSize();
	index = numEntries++;
	entries[index] = new Entry(key);
    }
    return index;
}


void GLinearAssoc::set (const void* key, const void* value)
{
    entries[forceKeyToIndex(key)]->set(value);
}


void GLinearAssoc::set (const void* key, int size, const void* addr)
{
    entries[forceKeyToIndex(key)]->set(size, addr);
}


int GLinearAssoc::add (const void* key, const void* elem)
{
    return entries[forceKeyToIndex(key)]->add(elem);
}


int GLinearAssoc::add (const void* key, int size, const void* addr)
{
    return entries[forceKeyToIndex(key)]->add(size, addr);
}

void* GLinearAssoc::get (const void* key)
{
    int index = keyToIndex(key);
    return (index < 0) ? NULL : entries[index]->get();
}


const void** GLinearAssoc::get (const void* key, int& numElems)
{
    int index = keyToIndex(key);
    if (index >= 0)
	return entries[index]->get(numElems);
    else
    {
	numElems = 0;
	return NULL;
    }
}


void GLinearAssoc::rem (const void* key)
{
    int index = keyToIndex(key);
    if (index >= 0)
	entries[index]->rem();
}


int GLinearAssoc::rem (const void* key, const void* elem)
{
    int index = keyToIndex(key);
    if (index >= 0)
	return entries[index]->rem(elem);
    else
	return 0;
}


int GLinearAssoc::rem (const void* key, int size, const void* addr)
{
    int index = keyToIndex(key);
    if (index >= 0)
	return entries[index]->rem(size, addr);
    else
	return 0;
}


const void** GLinearAssoc::getAndRemove
    (const void*& key, int& elemSize, int& numElems)
{
    while (numEntries)
    {
	Entry* entry = entries[numEntries - 1];
        const void** addr = entry->get(numElems);
	if (numElems > 0)
	{
	    key = entry->key;
	    elemSize = entry->elemSize();
	    const void** result = (const void**)GMALLOC(numElems * elemSize);
	    memcpy((void*)result, (const void*)addr, numElems * elemSize);
	    delete entry;
	    --numEntries;
	    return result;
	}
	delete entry;
	--numEntries;
    }

    numElems = 0;
    return NULL;
}



int GAssoc::hash (const void* key) const
{
//  assert(sizeof(key) == 4);
    uchar* ptr = (uchar*)&key;
    return (ptr[0] + ptr[1] + ptr[2] + ptr[3]) & (arraySize - 1);
}



GAssoc::GAssoc ()
{
    arraySize = 1;
    array = new GLinearAssoc*[1];
    array[0] = NULL;
}


void GAssoc::reset ()
{
    for (int i = 0; i < arraySize; i++)
	if (array[i])
	    delete array[i];
    delete array;

    arraySize = 1;
    array = new GLinearAssoc*[1];
    array[0] = NULL;
}


GAssoc::~GAssoc ()
{
    for (int i = 0; i < arraySize; i++)
	if (array[i])
	    delete array[i];
    delete array;
}


inline void GAssoc::maybeSwitchToHash ()
{
    const int limitForHashUse = 4;
    const int defaultHashedArraySize = 256;

    if (arraySize == 1 && array[0]->getNumEntries() > limitForHashUse)
	setArraySize(defaultHashedArraySize);
}


int GAssoc::keyToIndex (const void* key) const
{
    return (arraySize == 1) ? 0 : hash(key);
}


void GAssoc::set (const void* key, const void* value)
{
    int index = keyToIndex(key);
    if (!array[index])
	array[index] = new GLinearAssoc;
    array[index]->set(key, value);

    maybeSwitchToHash();
}


void GAssoc::set (const void* key, int size, const void* addr)
{
    int index = keyToIndex(key);
    if (!array[index])
	array[index] = new GLinearAssoc;
    array[index]->set(key, size, addr);

    maybeSwitchToHash();
}


int GAssoc::add (const void* key, const void* elem)
{
    int index = keyToIndex(key);
    if (!array[index])
	array[index] = new GLinearAssoc;
    int numElem = array[index]->add(key, elem);

    maybeSwitchToHash();
    return numElem;
}


int GAssoc::add (const void* key, int size, const void* addr)
{
    int index = keyToIndex(key);
    if (!array[index])
	array[index] = new GLinearAssoc;
    int numElem = array[index]->add(key, size, addr);

    maybeSwitchToHash();
    return numElem;
}

void* GAssoc::get (const void* key) const
{
    int index = keyToIndex(key);
    return (!array[index]) ? NULL : array[index]->get(key);
}


const void* GAssoc::get (const void* key, int& numElems) const
{
    int index = keyToIndex(key);
    if (array[index])
	return array[index]->get(key, numElems);
    else
    {
	numElems = 0;
	return NULL;
    }
}


void GAssoc::rem (const void* key)
{
    int index = keyToIndex(key);
    if (array[index])
	array[index]->rem(key);
}


int GAssoc::rem (const void* key, const void* elem)
{
    int index = keyToIndex(key);
    if (array[index])
	return array[index]->rem(key, elem);
    else
	return 0;
}


int GAssoc::rem (const void* key, int size, const void* addr)
{
    int index = keyToIndex(key);
    if (array[index])
	return array[index]->rem(key, size, addr);
    else
	return 0;
}


void GAssoc::setArraySize (int newsize)
{
    assert(newsize <= 256 && 0 == (newsize & (newsize - 1))); //power of two;

    GLinearAssoc** oldarray = array;
    ushort oldsize = arraySize;
    arraySize = newsize;
    array = (GLinearAssoc**)GMALLOC(arraySize * sizeof(GLinearAssoc*));
    memset(array, 0, arraySize * sizeof(GLinearAssoc*));

    for (int i = 0; i < oldsize; i++)
	if (oldarray[i])
	{
	    for (;;)
	    {
		const void* key;
		int elemSize, numElems;
		char* addr =
		    (char*)oldarray[i]->getAndRemove(key, elemSize, numElems);
		if (!numElems)
		    break;
		for (int j = 0; j < numElems; j++, addr += elemSize)
		    add(key, elemSize, addr);
	    }
	    delete oldarray[i];
	}

    delete oldarray;
}


void** GAssoc::getKeys (int& count) const
{
    count = 0;
    for (int i = 0; i < arraySize; i++)
    {
	GLinearAssoc* l = array[i];
	if (l)
	    for (int j = 0; j < l->getNumEntries(); j++)
		if (l->entry(j)->getNumElem() > 0)
		    count++;
    }
    
    void** buffer = (void**)GMALLOC(count * sizeof(void*));
    int index = 0;
    for (i = 0; i < arraySize; i++)
    {
	GLinearAssoc* l = array[i];
	if (l)
	    for (int j = 0; j < l->getNumEntries(); j++) {
		Entry* entry = l->entry(j);
		if (entry->getNumElem() > 0)
		    buffer[index++] = (void*)entry->key;
	    }
    }
    return buffer;
}
