#ifdef REV_INFO
static char rcsid[] = "$Id: StringAssoc.cc,v 1.7 1992/09/30 13:59:05 kuta Exp $";
#endif
/*****************************************************************************
MODULE:  Association
FILE:    StringAssoc.cc
AUTHOR:  Olivier Jojic
DATE:    29 May 1990

DESCRIPTION:
    See StringAssoc.h

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/Tools/StringAssoc.h>
#include <Go/Tools/Alloc.h>
#include <assert.h>


class StringLinearAssoc;

class StringEntry
{
    ushort unit;
    ushort size;
    ushort count;
    char*  buf;

    void  incrSize ();
    void  incrSize (int newSize);

  public:
    char* key;
    int elemSize () const { return unit; }
    void setKey(const char* tp);
    char* getKey() const;

    StringEntry(const char* tp);
    ~StringEntry();

    inline int getNumElem() const { return count; };
    void  set (const void* value);
    void  set (int valueSize, const void* valueAddr);
    void  add (const void* elem);
    void  add (int elemSize, const void* elemAddr);
    void* get () const;
    const void* get (int& numElems) const;
    void  rem ();
    void  rem (const void* elem);
    void  rem (int elemSize, const void* elemAddr);
};

class StringLinearAssoc
{
    ushort  size;
    ushort  numEntries;
    class StringEntry** entries;

    int keyToIndex (const char* key) const;
    int forceKeyToIndex (const char* key);
    void incrSize();

  public:
    StringLinearAssoc();
    ~StringLinearAssoc();

    int getNumEntries () const { return numEntries; }
    StringEntry* entry(int i) const { return entries[i]; }

    void* getAndRemove (char*& key, int& elemSize, int& numElems);

    void  set (const char* key, const void* value);
    void  set (const char* key, int size, const void* addr);
    void  add (const char* key, const void* elem);
    void  add (const char* key, int size, const void* addr);
    void* get (const char* key) const;
    const void* get (const char* key, int& numElems) const;
    void  rem (const char* key);
    void  rem (const char* key, const void* elem);
    void  rem (const char* key, int size, const void* addr);
};


void StringEntry::setKey (const char* tp)
{
    GFREE(key);
    key = (char*)GMALLOC(1 + strlen(tp));
    strcpy(key, tp);
}


char* StringEntry::getKey () const
{
    char* tp = (char*)GMALLOC(1 + strlen(key));
    strcpy(tp, key);
    return tp;
}


StringEntry::StringEntry (const char* tp)
{
    key = (char*)GMALLOC(strlen(tp) + 1);
    strcpy(key, tp);
    unit = 1;
    size = count = 0;
    buf = NULL;
}


StringEntry::~StringEntry ()
{
    GFREE(key);
    if (size)
    {
	GFREE(buf);
	buf = NULL;
    }
}


void StringEntry::incrSize ()
{
    size += 8 * unit;
    char* old = buf;
    buf = (char*)GMALLOC(size);
    memcpy(buf, old, count);
    GFREE(old);
}


void StringEntry::incrSize (int newSize)
{
    assert(count <= newSize);
    char* old = buf;
    size = newSize;
    buf = (char*)GMALLOC(size);
    memcpy(buf, old, count);
    GFREE(old);
}


void StringEntry::set (const void* value)
{
    unit = sizeof(void*);
    count = 0;
    if (size < unit)
	incrSize (unit);
    count = unit;
    *((const void**)buf) = value;
}


void StringEntry::set (int valueSize, const void* valueAddr)
{
    unit = valueSize;
    count = 0;
    if (size < unit)
	incrSize(unit);
    count = unit;
    memcpy(buf, valueAddr, unit);
}


void StringEntry::add (const void* elem)
{
    if (!count) unit = sizeof(void*);
    else 	assert(unit == sizeof(void*));

    if (count + unit > size)
	incrSize();
    *((const void**)(buf+count)) = elem;
    count += unit;
}


void StringEntry::add (int elemSize, const void* elemAddr)
{
    if (!count)	unit = elemSize;
    else	assert(unit == elemSize);

    if (count + unit > size)
	incrSize();
    memcpy(buf+count, elemAddr, unit);
    count += unit;
}


void* StringEntry::get () const
{
    if (!count)
	return 0;
    assert(unit == sizeof(long));
    return (count) ? *((void**)(buf+count-unit)) : 0;
}


const void* StringEntry::get (int& numElems) const
{
    numElems = count / unit;
    if (numElems == 0)
	return NULL;
    return (void*)buf;
}


void StringEntry::rem ()
{
    if (size)
	GFREE(buf);
    buf = NULL;
    unit = 1;
    size = count = 0;
}


void StringEntry::rem (const void* elem)
{
    if (!count)
	return;
    assert(unit == sizeof(void*));
    for (int i = 0; i < count; i += unit)
	if (elem == *((void**)(buf+i)))
	{
	    count -= unit;
	    memcpy(buf + i, buf + i + unit, count - i);
	    return;
	}
}


void StringEntry::rem (int elemSize, const void* elemAddr)
{
    if (!count)
	return;
    assert(unit == elemSize);
    for (int i = 0; i < count; i += unit)
	if (!memcmp(elemAddr, buf+i, unit))
	{
	    count -= unit;
	    memcpy(buf+i, buf+i+unit, count - i);
	    return;
	}
}





StringLinearAssoc::StringLinearAssoc ()
{
    size = numEntries = 0;
    entries = 0;
}


StringLinearAssoc::~StringLinearAssoc ()
{
    for (int i = 0; i < numEntries; i++)
	delete entries[i];
    if (size)
	GFREE(entries);
}


int StringLinearAssoc::keyToIndex (const char* key) const
{
    int lastEmpty = -1;

    for (int i = 0; i < numEntries; i++)
	if (!strcmp((char*)entries[i]->key, key))
	    return i;
	else if (entries[i]->getNumElem() == 0)
	    lastEmpty = i;

    if (lastEmpty == -1)
	return -1;
    else
    {
	entries[lastEmpty]->setKey(key);
	return lastEmpty;
    }
}


void StringLinearAssoc::incrSize ()
{
    size += 8;
    StringEntry** tmp = entries;
    entries = (StringEntry**)GMALLOC(size * sizeof(StringEntry*));
    memcpy(entries, tmp, numEntries * sizeof(StringEntry*));
    GFREE(tmp);
}


int StringLinearAssoc::forceKeyToIndex (const char* key)
{
    int index = keyToIndex(key);
    if (index == -1)
    {
	if (numEntries >= size)
	    incrSize();
	index = numEntries++;
	entries[index] = new StringEntry(key);
    }
    return index;
}


void StringLinearAssoc::set (const char* key, const void* value)
{
    entries[forceKeyToIndex(key)]->set(value);
}


void StringLinearAssoc::set (const char* key, int size, const void* addr)
{
    entries[forceKeyToIndex(key)]->set(size, addr);
}


void StringLinearAssoc::add (const char* key, const void* elem)
{
    entries[forceKeyToIndex(key)]->add(elem);
}


void StringLinearAssoc::add (const char* key, int size, const void* addr)
{
    entries[forceKeyToIndex(key)]->add(size, addr);
}


void* StringLinearAssoc::get (const char* key) const
{
    int index = keyToIndex(key);
    return (index < 0) ? 0 : entries[index]->get();
}


const void* StringLinearAssoc::get (const char* key, int& numElems) const
{
    int index = keyToIndex(key);
    if (index >= 0)
	return entries[index]->get(numElems);
    else
    {
	numElems = 0;
	return NULL;
    }
}


void StringLinearAssoc::rem (const char* key)
{
    int index = keyToIndex(key);
    if (index >= 0)
	entries[index]->rem();
}


void StringLinearAssoc::rem (const char* key, const void* elem)
{
    int index = keyToIndex(key);
    if (index >= 0)
	entries[index]->rem(elem);
}


void StringLinearAssoc::rem (const char* key, int size, const void* addr)
{
    int index = keyToIndex(key);
    if (index >= 0)
	entries[index]->rem(size, addr);
}


void* StringLinearAssoc::getAndRemove (char*& key,
				       int& elemSize, int& numElems)
{
    while (numEntries)
    {
	StringEntry* entry = entries[numEntries - 1];
	const void* addr = entry->get(numElems);
	if (numElems > 0)
	{
	    key = entry->getKey();
	    elemSize = entry->elemSize();
	    char* result = (char*)GMALLOC(numElems * elemSize);
	    memcpy(result, addr, numElems * elemSize);
	    delete entry;
	    --numEntries;
	    return result;
	}
	delete entry;
	--numEntries;
    }

    numElems = 0;
    return NULL;
}



int GStringAssoc::hash (const char* key) const
{
    int h = strlen(key);
    int last_i = (h < 3) ? h : 3;

    for (int i = 0; i < last_i; i++)
	h += key[i];

    return i & (arraySize - 1);
}



GStringAssoc::GStringAssoc ()
{
    arraySize = 1;
    array = (StringLinearAssoc**)GMALLOC(sizeof(StringLinearAssoc*));
    array[0] = NULL;
}


GStringAssoc::~GStringAssoc ()
{
    for (int i = 0; i < arraySize; i++)
	if (array[i])
	    delete array[i];
    GFREE(array);
}


inline void GStringAssoc::maybeSwitchToHash ()
{
    const int limitForHashUse = 4;
    const int defaultHashedArraySize = 256;

    if (arraySize == 1 && array[0]->getNumEntries() > limitForHashUse)
	setArraySize(defaultHashedArraySize);
}


int GStringAssoc::keyToIndex (const char* key) const
{
    return (arraySize == 1) ? 0 : hash(key);
}


void GStringAssoc::set (const char* key, const void* value)
{
    int index = keyToIndex(key);
    if (!array[index])
	array[index] = new StringLinearAssoc;
    array[index]->set(key, value);

    maybeSwitchToHash();
}


void GStringAssoc::set (const char* key, int size, const void* addr)
{
    int index = keyToIndex(key);
    if (!array[index])
	array[index] = new StringLinearAssoc;
    array[index]->set(key, size, addr);

    maybeSwitchToHash();
}


void GStringAssoc::add (const char* key, const void* elem)
{
    int index = keyToIndex(key);
    if (!array[index])
	array[index] = new StringLinearAssoc;
    array[index]->add(key, elem);

    maybeSwitchToHash();
}


void GStringAssoc::add (const char* key, int size, const void* addr)
{
    int index = keyToIndex(key);
    if (!array[index])
	array[index] = new StringLinearAssoc;
    array[index]->add(key, size, addr);

    maybeSwitchToHash();
}


void* GStringAssoc::get (const char* key) const
{
    int index = keyToIndex(key);
    return (!array[index]) ? 0 : array[index]->get(key);
}


const void* GStringAssoc::get (const char* key, int& numElems) const
{
    int index = keyToIndex(key);
    if (array[index])
	return array[index]->get(key, numElems);
    else
    {
	numElems = 0;
	return NULL;
    }
}


void GStringAssoc::rem (const char* key)
{
    int index = keyToIndex(key);
    if (array[index])
	array[index]->rem(key);
}


void GStringAssoc::rem (const char* key, const void* elem)
{
    int index = keyToIndex(key);
    if (array[index])
	array[index]->rem(key, elem);
}


void GStringAssoc::rem (const char* key, int size, const void* addr)
{
    int index = keyToIndex(key);
    if (array[index])
	array[index]->rem(key, size, addr);
}


void GStringAssoc::setArraySize (int newsize)
{
    assert(newsize <= 256 && 0 == (newsize & (newsize - 1))); //power of two;

    StringLinearAssoc** oldarray = array;
    ushort oldsize = arraySize;
    arraySize = newsize;
    array = (StringLinearAssoc**)
	GMALLOC(arraySize * sizeof(StringLinearAssoc*));
    memset(array, 0, arraySize * sizeof(StringLinearAssoc*));

    for (int i = 0; i < oldsize; i++)
	if (oldarray[i])
	{
	    for (;;)
	    {
		char* key;
		int elemSize, numElems;
		char* addr = (char*)oldarray[i]->getAndRemove(key, elemSize, numElems);
		if (!numElems)
		    break;
		for (int j = 0; j < numElems; j++, addr += elemSize)
		    add(key, elemSize, addr);
	    }
	    delete oldarray[i];
	}

    GFREE(oldarray);
}

void GStringAssoc::reset ()
{
    for (int i = 0; i < arraySize; i++)
	if (array[i])
	    delete array[i];
    delete array;

    arraySize = 1;
    array = new StringLinearAssoc*[1];
    array[0] = NULL;
}


char** GStringAssoc::getKeys (int& count) const
{
    count = 0;
    for (int i = 0; i < arraySize; i++)
    {
	StringLinearAssoc* l = array[i];
	if (l)
	    for (int j = 0; j < l->getNumEntries(); j++)
		if (l->entry(j)->getNumElem() > 0)
		    count++;
    }

    char** buffer = (char**)GMALLOC(count * sizeof(char*));
    int index = 0;
    for (i = 0; i < arraySize; i++)
    {
	StringLinearAssoc* l = array[i];
	if (l)
	    for (int j = 0; j < l->getNumEntries(); j++) {
		StringEntry* entry = l->entry(j);
		if (entry->getNumElem() > 0)
		    buffer[index++] = entry->key;
	    }
    }

    return buffer;
}
