#ifdef REV_INFO
static char rcsid[] = "$Id: Color.cc,v 1.2 1992/09/25 14:05:25 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Tools
FILE:    Color.cc
AUTHOR:  Jacques Davy
DATE:    November 1990

DESCRIPTION:
    (see <Go/Color.h>

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/Tools/Color.h>
#include <Go/Tools/Pattern.h>
#include <assert.h>

/*
 * Callback types
 */
static char colorNmodifyPrecond[1];
static char colorNmodifyPostcond[1];
static char colorNdestroyPostcond[1];

typedef GBool (*PeProc)(void*, const GColor*, ulong&, GColorValues*);
typedef void  (*PsProc)(void*, GColor*, ulong);
typedef void  (*PdProc)(void*, const GColor*);

struct PrecondRec
{
    PeProc proc;
    void* closure;
};

struct ModifPostcondRec
{
    PsProc proc;
    void* closure;
};

struct DestroyPostcondRec
{
    PdProc proc;
    void* closure;
};


/*
 * Private methods
 * ---------------
 */
/**
FUNCTION
    Call_call_modify_postcond
DESCRIPTION
    This function is called when the NAMED ShColor is updated
**/
void _Call_call_modify_postcond (void* This, GShColor* sh_color,
				 ulong update_mask)
{
    Call_call_modify_postcond ((GColor*)This, sh_color, update_mask);
}

void Call_call_modify_postcond (GColor* This, const GShColor* sh_color,
				ulong update_mask)
{
    assert (sh_color == This->shColor());
    This->_call_modify_postcond (update_mask);
}

/*
 * Protected methods
 * -----------------
 */
/**/
int GColor::_addCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = &_assoc;
    GAssoc* cb = (GAssoc*)_cb;
    
    GCallbackRec cbrec (proc, closure);
    return cb->add(type, sizeof cbrec, (char*)&cbrec);
}

/**/
int GColor::_remCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = &_assoc;
    GAssoc* cb = (GAssoc*)_cb;

    GCallbackRec cbrec (proc, closure);
    return cb->rem(type, sizeof cbrec, (char*)&cbrec);
}

/**/
GBool GColor::_call_modify_precond (ulong& modify_mask, GColorValues* new_values)
{
    if (_flags._has_modify_precond)
    {
	int count;
        PrecondRec* cbrec = (PrecondRec*)_assoc.get (colorNmodifyPrecond, count);
	for (int i = 0; i < count; i++, cbrec++)
	    if (!cbrec->proc(cbrec->closure, this, modify_mask, new_values))
                return GFALSE;
    }
    return GTRUE;
}

/**/
void GColor::_call_modify_postcond (ulong modify_mask)
{
    if (_flags._has_modify_postcond)
    {
	int count;
	GCallbackRec* cbrec = (GCallbackRec*)_assoc.get (colorNmodifyPostcond, count);
	(void)GCallCallbacks(cbrec, count, this, modify_mask);
    }
}

/**/
void GColor::_call_destroy_postcond ()
{
    if (_flags._has_destroy_postcond)
    {
	int count;
	DestroyPostcondRec* cbrec = (DestroyPostcondRec*)_assoc.get (colorNdestroyPostcond, count);
	for (int i = 0; i < count; i++, cbrec++)
	     cbrec->proc(cbrec->closure, this);
    }
}

/*
 * Public methods
 * --------------
 */
GColor::GColor ()
{
    _sh_color = GCreateShColor();  /* default one */
}

GColor::GColor (const GColor* color)
{
    _sh_color = color->_sh_color->dup();
}

GColor::GColor (const GShColor* sh_color)
{
    _sh_color = sh_color->dup();
}

GColor::GColor (ulong mask, const GColorValues* values, const char* name)
{
    _sh_color = GCreateShColor (mask, values, name);
}

GColor::GColor (const GRgb* rgb, const char* name)
{
    _sh_color = GCreateShColor (rgb, name);
}

GColor::GColor (uchar gray, const char* name)
{
    _sh_color = GCreateShColor (gray, name);
}

GColor::GColor (const GShPattern* halftone, const char* name)
{
    _sh_color = GCreateShColor (halftone, name);
}

GColor::GColor (const char* name)
{
    _sh_color = GCreateShColor (name);
}

GColor::~GColor ()
{
    _call_destroy_postcond ();
    if (_sh_color->name() && _flags._has_modify_postcond)
        _sh_color->updatePostcond (_Call_call_modify_postcond, this, GREMOVE);
    _sh_color->destroy();
}

/**
FUNCTION
    GColor::modify
DESCRIPTION
    Modify the shared color and call pre/post condition.
**/

void GColor::modify (ulong mask, const GColorValues* v)
{
    GColorValues Values;
    Values = *v;
    if (mask && _call_modify_precond (mask, &Values))
    {
	// Remove postcond;
	if (_sh_color->isNamed() && _flags._has_modify_postcond)
	    _sh_color->updatePostcond (_Call_call_modify_postcond, this, GREMOVE);

	// Modify;
        GModify (_sh_color, mask, &Values);

	// Call postcond;
	_call_modify_postcond (mask);

	// Add postcond;
	if (_sh_color->isNamed() && _flags._has_modify_postcond)
	    _sh_color->updatePostcond (_Call_call_modify_postcond, this, GADD);
    }
}

void GColor::modify (const char* name)
{
    if (name && _sh_color->name() && (strcmp (name, _sh_color->name()) == 0))
        return;


    // Remove postcond if the color is already named.
    if (_sh_color->name() && _flags._has_modify_postcond)
        _sh_color->updatePostcond (_Call_call_modify_postcond, this, GREMOVE);

    // Get mask and values
    GShColor* save = _sh_color->dup();
    GRename (save, name);

    GColorValues Values;

    Values = *save->values();
    ulong mask = save->mask();

    if (mask && _call_modify_precond (mask, &Values))
    {
         _sh_color->destroy();
         _sh_color = GCreateShColor (mask, &Values, name);
         _call_modify_postcond (mask);
    }
    save->destroy();

    // Add postcond if the color is named.
    if (name && _flags._has_modify_postcond)
	_sh_color->updatePostcond (_Call_call_modify_postcond, this, GADD);
}

void GColor::modify(const GShColor* sh_color)
{
    if (_sh_color  == sh_color)
	return;

    // top of list.
    sh_color->raise(GShColor::_headOfList);
    
    if (sh_color->isNamed())
	modify(sh_color->name());
    else
    {
	GColorValues Values ;
	Values  = *sh_color->values();
	ulong mask = sh_color->mask();
	    

	// Remove postcond if the color is already named;
	if (_sh_color->name() && _flags._has_modify_postcond)
	    _sh_color->updatePostcond (_Call_call_modify_postcond, this, GREMOVE);

    	// unname _sh_color;
	GRename(_sh_color, NULL);
		
	if (mask && _call_modify_precond (mask, &Values))
	{
	    // Modify;
	    GModify (_sh_color, mask, &Values);

	    // Call postcond;
	    _call_modify_postcond (mask);
	}

	// Add postcond;
	if (_sh_color->isNamed() && _flags._has_modify_postcond)
	    _sh_color->updatePostcond (_Call_call_modify_postcond, this, GADD);
    }
}

/**
FUNCTION
    GColor::field_modification_methods
DESCRIPTION
    Each method modify one parameter of the Color.
**/

#define _modifyColorValue(bit_mask, field, value) \
    GColorValues values;\
    values.field = value;\
    modify (bit_mask, &values)

void GColor::rgb (const GRgb* rgb)
{
    GColorValues values;
    values.rgb = *rgb;
    modify (GCOLOR_RGB, &values);
}

void GColor::gray (uchar gray)
{
    _modifyColorValue(GCOLOR_GRAY, gray, gray);
}

/**/
void GColor::halftone (const GPattern* halftone)
{
    _modifyColorValue(GCOLOR_HALFTONE, halftone, (GShPattern*)halftone->shPattern());
}

void GColor::halftone (const GShPattern* halftone)
{
    _modifyColorValue(GCOLOR_HALFTONE, halftone, (GShPattern*)halftone);
}

void GColor::halftone (const char* name)
{
    GColorValues values;
    values.halftone = GCreateShPattern (name);
    modify (GCOLOR_HALFTONE, &values);
    values.halftone->destroy();
}
#undef _modifyColorValue

/**
FUNCTION
    GColor::add/rem pre/postconditions
DESCRIPTION
    Add or remove pre/post condition on modification and on destruction.

    Warning, if the ShColor is named, _call_call_modify_postcond
    is stored directly in the shared object and pre/condition
    may be not called!
**/

#define _add_rem_callback(type, flag) \
    (FLAGS->flag =\
         0 < ((add == GADD)\
	      ? _addCallback(type, (GCBProc)proc, closure)\
	      : _remCallback(type, (GCBProc)proc, closure)))

/**/
void GColor::modifyPrecond (PeProc proc, const void* closure, GAddOrRemove add) const
{
    const GColorCondFlags* flags = &_flags;
    GColorCondFlags* FLAGS = (GColorCondFlags*)flags;
    
   _add_rem_callback(colorNmodifyPrecond, _has_modify_precond);
}

/**/
void GColor::modifyPostcond (PsProc proc, const void* closure, GAddOrRemove add) const
{
    ushort flag = _flags._has_modify_postcond;

    const GColorCondFlags* flags = &_flags;
    GColorCondFlags* FLAGS = (GColorCondFlags*)flags;

    _add_rem_callback(colorNmodifyPostcond, _has_modify_postcond);

    // If the color is named, the modifyPostcond is called from the;
    // sh_color;
    if (_sh_color->isNamed())
    {
	if (!flag && _flags._has_modify_postcond)
	    _sh_color->updatePostcond (_Call_call_modify_postcond,
				       this, GADD);
	else
	    if (flag && !_flags._has_modify_postcond)
		_sh_color->updatePostcond (_Call_call_modify_postcond,
					   this, GREMOVE);
    }
}

/**/
void GColor::destroyPostcond (PdProc proc, const void* closure, GAddOrRemove add) const
{
    const GColorCondFlags* flags = &_flags;
    GColorCondFlags* FLAGS = (GColorCondFlags*)flags;

   _add_rem_callback(colorNdestroyPostcond, _has_destroy_postcond);
}

#undef _add_rem_callback
