#ifdef REV_INFO
static char rcsid[] = "$Id: CoordSys.cc,v 1.3 1992/05/26 16:29:40 benati Exp $";
#endif
/*****************************************************************************
MODULE:  Tools
FILE:    CoordSys.cc
AUTHOR:  Olivier Jojic
DATE:    Jan 90

DESCRIPTION:
    See the CoordSys.h header.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
*****************************************************************************/

#include <Go/Tools/CoordSys.h>
#include <Go/Tools/Matrix.h>

const GCoordSys& GUnitCoordSys() {
    static GCoordSys* unit_cs = NULL;

    if (! unit_cs)
	unit_cs = new GCoordSys;
    return *unit_cs;
}


GCoordSys::GCoordSys () : O(0,0), I(1, 0), J(0, 1)
{
}

GCoordSys::GCoordSys (const GCoordSys& cs) : O(cs.O), I(cs.I), J(cs.J)
{
}

GCoordSys::GCoordSys (const GPoint& o, const GVector& i, const GVector& j)
    : O(o), I(i), J(j)
{
}

GBool operator == (const GCoordSys& a, const GCoordSys& b)
{
    return a.O == b.O && a.I == b.I && a.J == b.J;
}

GVector GCoordSys::to (const GVector& v) const
{
    return v.x * I + v.y * J;
}

GPoint GCoordSys::to (const GPoint& p) const
{
    return O + to((GVector)p);
}

GPoint GCoordSys::from (const GPoint& p) const
{
    return from(p - O);
}

GCoordSys GCoordSys::from (const GCoordSys& cs) const
{
    return GCoordSys(from(cs.O),
		     from(cs.I),
		     from(cs.J));
}

GCoordSys GCoordSys::operator ~ () const
{
    GCoordSys cs = *this;
    return cs.invert();
}

void GCoordSys::angle (GNumber a)
{
    rotate(GPoint (0,0), a - I.angle());
}



GCoordSys::GCoordSys (const GPoint& a, const GPoint& b)
{
    O.x = min(a.x, b.x);
    O.y = min(a.y, b.y);

    I.x = (b.x > a.x) ? b.x - a.x : a.x - b.x;
    I.y = 0;

    J.x = 0;
    J.y = (b.y > a.y) ? b.y - a.y : a.y - b.y;
}

GVector GCoordSys::from (const GVector& v) const
{
    GNumber d = deter();
    if (d) return GVector((v.x*J.y - v.y*J.x)/d,
		         (v.y*I.x - v.x*I.y)/d);
    GVector r;
    if (I.null()) r = J;
    else          r = I;

    r.resize(v.size() / r.size());
    return r;
}


GBox GCoordSys::to (const GBox& b) const
{
    return GBox(to(GPoint(b.xmin, b.ymin)),
	       to(GPoint(b.xmax, b.ymax)),
	       to(GPoint(b.xmin, b.ymax)),
	       to(GPoint(b.xmax, b.ymin)));
}

GBox GCoordSys::from (const GBox& b) const
{
    return GBox(from(GPoint(b.xmin, b.ymin)),
	       from(GPoint(b.xmax, b.ymax)),
	       from(GPoint(b.xmin, b.ymax)),
	       from(GPoint(b.xmax, b.ymin)));
}


GRegion GCoordSys::to (const GRegion& r) const
{
    if (r.empty())
	return GRegion(r);

    return GRegion(GBox(to(GPoint(r.bigbox().xmin, r.bigbox().ymin)),
		      to(GPoint(r.bigbox().xmax, r.bigbox().ymax)),
		      to(GPoint(r.bigbox().xmin, r.bigbox().ymax)),
		      to(GPoint(r.bigbox().xmax, r.bigbox().ymin))));
}

GRegion GCoordSys::from (const GRegion& r) const
{
    if (r.empty())
	return r;

    return GRegion(GBox(from(GPoint(r.bigbox().xmin, r.bigbox().ymin)),
		      from(GPoint(r.bigbox().xmax, r.bigbox().ymax)),
		      from(GPoint(r.bigbox().xmin, r.bigbox().ymax)),
		      from(GPoint(r.bigbox().xmax, r.bigbox().ymin))));
}


void GCoordSys::rotate (const GPoint& center, GNumber angle)
{
    GPoint absC = to(center);
    I.rotate(angle);
    J.rotate(angle);
    O += absC - to(center);
}

GCoordSys& GCoordSys::invert ()
{
    GMatrix im = ~GMatrix(I, J);
    O = im * - O;
    I = im.vx();
    J = im.vy();
    return *this;
}    


GCoordSys GCoordSys::to (const GCoordSys& cs) const
{
    return GCoordSys(to(cs.O),
		    to(cs.I),
		    to(cs.J));
}


GBool GCoordSys::isOrthogonal () const
{
    GNumber ix = fabs(I.x);
    GNumber iy = fabs(I.y);
    if (ix > .001 && iy > .001)
	return GFALSE;

    GNumber jx = fabs(J.x);
    GNumber jy = fabs(J.y);
    if (jx > .001 && jy > .001)
	return GFALSE;

    if (ix > iy)
    {
	if (ix > iy * 1000 && jy > jx * 1000)
	    return GTRUE;
    }
    else
    {
	if (iy > ix * 1000 && jx > jy * 1000)
	    return GTRUE;
    }

    return GFALSE;
}
