#ifdef REV_INFO
static char rcsid[] = "$Id: Font.cc,v 1.2 1992/09/25 14:24:03 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Tools
FILE:    Font.cc
AUTHOR:  Jacques Davy
DATE:    November 1990

DESCRIPTION:
    (see <Go/Font.h>

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/Tools/Font.h>
#include <assert.h>

/*
 * GCallback types
 */
static char fontNmodifyPrecond[1];
static char fontNmodifyPostcond[1];
static char fontNdestroyPostcond[1];

typedef GBool (*PeProc)(void*, const GFont*, ulong&, GFontValues*);
typedef void (*PsProc)(void*, GFont*, ulong);
typedef void (*PdProc)(void*, const GFont*);

struct PrecondRec
{
    PeProc proc;
    void* closure;
};

struct ModifPostcondRec
{
    PsProc proc;
    void* closure;
};

struct DestroyPostcondRec
{
    PdProc proc;
    void* closure;
};


/*
 * Private methods
 * ---------------
 */
/**
FUNCTION
    _call_call_modify_postcond
DESCRIPTION
    This function is called when the NAMED GShFont is updated
**/
void _call_call_modify_postcond (void* This, GShFont* sh_font,
				 ulong update_mask)
{
    Call_call_modify_postcond ((GFont*)This, sh_font, update_mask);
}

void Call_call_modify_postcond (GFont* This,
				const GShFont* sh_font,
				ulong update_mask)
{
    assert (sh_font == This->shFont());
    This->_call_modify_postcond (update_mask);
}

/*
 * Protected methods
 * -----------------
 */
/**/
int GFont::_addCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = &_assoc;
    GAssoc* cb = (GAssoc*)_cb;
    
    GCallbackRec cbrec (proc, closure);
    return cb->add(type, sizeof cbrec, (char*)&cbrec);
}

/**/
int GFont::_remCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = &_assoc;
    GAssoc* cb = (GAssoc*)_cb;

    GCallbackRec cbrec (proc, closure);
    return cb->rem(type, sizeof cbrec, (char*)&cbrec);
}

/**/
GBool GFont::_call_modify_precond (ulong& modify_mask, GFontValues* values)
{
    if (_flags._has_modify_precond)
    {
	int count;
	PrecondRec* cbrec = (PrecondRec*)_assoc.get (fontNmodifyPrecond, count);
	for (int i = 0; i < count; i++, cbrec++)
	    if (!cbrec->proc(cbrec->closure, this, modify_mask, values))
                return GFALSE;
    }
    return GTRUE;
}

/**/
void GFont::_call_modify_postcond (ulong modify_mask)
{
    if (_flags._has_modify_postcond)
    {
	int count;
	ModifPostcondRec* cbrec = (ModifPostcondRec*)_assoc.get (fontNmodifyPostcond, count);
	for (int i = 0; i < count; i++, cbrec++)
	     cbrec->proc(cbrec->closure, this, modify_mask);
    }
}

/**/
void GFont::_call_destroy_postcond ()
{
    if (_flags._has_destroy_postcond)
    {
	int count;
	DestroyPostcondRec* cbrec = (DestroyPostcondRec*)_assoc.get (fontNdestroyPostcond, count);
	for (int i = 0; i < count; i++, cbrec++)
	     cbrec->proc(cbrec->closure, this);
    }
}

static GShFont* default_shfont = NULL;
/*
 * Public methods
 * --------------
 */
GFont::GFont ()
{
    if (default_shfont == NULL)
	default_shfont = GCreateShFont(0, NULL);
    assert (default_shfont->refcnt() >= 1);
    _sh_font = default_shfont->dup();
}


GFont::GFont (const GFont* font)
{
    _sh_font = font->_sh_font->dup();
}

GFont::GFont (const GShFont* sh_font)
{
    _sh_font = sh_font->dup();
}

GFont::GFont (ulong mask, const GFontValues* values, const char* name)
{
    _sh_font = GCreateShFont (mask, values, name);
}

GFont::GFont (const char* name)
{
    _sh_font = GCreateShFont (name);
}

GFont::~GFont ()
{
    _call_destroy_postcond ();
    if (_sh_font->name() && _flags._has_modify_postcond)
        _sh_font->updatePostcond (_call_call_modify_postcond, this, GREMOVE);
    _sh_font->destroy();
}

/**
FUNCTION
    GFont::modify
DESCRIPTION
    Modify the shared font and call pre/post condition.
**/

void GFont::modify (ulong mask, const GFontValues* values)
{
    GFontValues Values = *values;
    if (mask && _call_modify_precond (mask, &Values))
    {

	// Remove postcond;
	if (_sh_font->isNamed() && _flags._has_modify_postcond)
	    _sh_font->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

	// Modify. Do not call member function but ShFont *friend* function;
        ::GModify (_sh_font, mask, &Values);

	// Call postcond;
	_call_modify_postcond (mask);

	// Add postcond;
	if (_sh_font->isNamed() && _flags._has_modify_postcond)
	    _sh_font->updatePostcond (_call_call_modify_postcond, this, GADD);

    }
}

void GFont::modify (const char* name)
{
    if (name && _sh_font->name() && (strcmp (name, _sh_font->name()) == 0))
        return;

    // Remove postcond if the font is already named.
    if (_sh_font->name() && _flags._has_modify_postcond)
        _sh_font->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

    // Get mask and values
    GShFont* save = _sh_font->dup();
    GRename (save, name);
    GFontValues Values = *save->values();
    ulong mask = save->mask();

    if (mask && _call_modify_precond (mask, &Values))
    {
         _sh_font->destroy();
         _sh_font = GCreateShFont (mask, &Values, name);
         _call_modify_postcond (mask);
    }
    save->destroy();

    // Add postcond if the font is named.
    if (name && _flags._has_modify_postcond)
	_sh_font->updatePostcond (_call_call_modify_postcond, this, GADD);
}

void GFont::modify(const GShFont* sh_font)
{
    if (sh_font == _sh_font)
	return ;
    
    // top of list.
    sh_font->raise(GShFont::_headOfList);
    
    if (sh_font->isNamed())
	modify(sh_font->name());
    else {

	GFontValues Values ;
	Values  = *sh_font->values();
	ulong mask = sh_font->mask();
	    

	// Remove postcond if the font is already named;
	if (_sh_font->name() && _flags._has_modify_postcond)
	    _sh_font->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

    	// Rename to NULL;
	GRename(_sh_font, NULL);
		
	if (mask && _call_modify_precond (mask, &Values)) {

	    // Modify. Do not call member function but ShFont *friend* function;
	    ::GModify (_sh_font, mask, &Values);

	    // Call postcond;
	    _call_modify_postcond (mask);
	}


	// Add postcond;
	if (_sh_font->isNamed() && _flags._has_modify_postcond)
	    _sh_font->updatePostcond (_call_call_modify_postcond, this, GADD);
    }

}

/**
FUNCTION
    GFont::field_modification_methods
DESCRIPTION
    Each method modify one parameter of the GFont.
**/

#define _modifyFontValue(bit_mask, field, value) \
    GFontValues values;\
    values.field = value;\
    modify (bit_mask, &values)

#undef _modifyFontValue

/**
FUNCTION
    GFont::add/rem pre/postconditions
DESCRIPTION
    Add or remove pre/post condition on modification and on destruction.

    GWarning, if the GShFont is named, _call_call_modify_postcond
    is stored directly in the shared object and pre/condition
    may be not called!
**/

#define _add_rem_callback(type, flag) \
    (FLAGS->flag =\
         0 < ((add == GADD)\
	      ? _addCallback(type, (GCBProc)proc, closure)\
	      : _remCallback(type, (GCBProc)proc, closure)))

/**/
void GFont::modifyPrecond (PeProc proc, const void* closure, GAddOrRemove add) const
{
    const GFontCondFlags* flags = &_flags;
    GFontCondFlags* FLAGS = (GFontCondFlags*)flags;
    
   _add_rem_callback(fontNmodifyPrecond, _has_modify_precond);
}

/**/
void GFont::modifyPostcond (PsProc proc, const void* closure, GAddOrRemove add) const
{
   ushort flag = _flags._has_modify_postcond;

    const GFontCondFlags* flags = &_flags;
    GFontCondFlags* FLAGS = (GFontCondFlags*)flags;
    
   _add_rem_callback(fontNmodifyPostcond, _has_modify_postcond);

   // If the font is named, the modifyPostcond is called from the 
   // sh_font
   if (_sh_font->name())
   {
       if (!flag && _flags._has_modify_postcond)
           _sh_font->updatePostcond (_call_call_modify_postcond, this, GADD);
       else
           if (flag && !_flags._has_modify_postcond)
               _sh_font->updatePostcond (_call_call_modify_postcond, this, GREMOVE);
   }
}

/**/
void GFont::destroyPostcond (PdProc proc,
			    const void* closure, GAddOrRemove add) const
{
    const GFontCondFlags* flags = &_flags;
    GFontCondFlags* FLAGS = (GFontCondFlags*)flags;
    
   _add_rem_callback(fontNdestroyPostcond, _has_destroy_postcond);
}

#undef _add_rem_callback
