#ifdef REV_INFO
static char rcsid[] = "$Id: Brush.cc,v 1.2 1992/09/25 14:30:38 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  GoGC
FILE:    Brush.cc
AUTHOR:  Jacques Davy
DATE:    November 1990

DESCRIPTION:
    (see <Go/Brush.h>

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/Tools/Brush.h>
#include <Go/Tools/Color.h>
#include <Go/Tools/Pattern.h>
#include <assert.h>

/*
 * GCallback types
 */
static char brushNmodifyPrecond[1];
static char brushNmodifyPostcond[1];
static char brushNdestroyPostcond[1];

typedef GBool (*PeProc)(void*, const GBrush*, ulong&, GBrushValues*);
typedef void (*PsProc)(void*, GBrush*, ulong);
typedef void (*PdProc)(void*, const GBrush*);

struct PrecondRec
{
    PeProc proc;
    void* closure;
};

struct ModifPostcondRec
{
    PsProc proc;
    void* closure;
};

struct DestroyPostcondRec
{
    PdProc proc;
    void* closure;
};

/*
 * Private methods
 * ---------------
 */
/**
FUNCTION
    _call_call_modify_postcond
DESCRIPTION
    This function is called when the GShBrush is updated
**/
void _call_call_modify_postcond (void* This, GShBrush* sh_brush,
				 ulong update_mask)
{
    Call_call_modify_postcond ((GBrush*)This, sh_brush, update_mask);
}

void Call_call_modify_postcond (GBrush* This, const GShBrush* sh_brush,
				ulong update_mask)
{
    assert (sh_brush == This->shBrush());
    This->_call_modify_postcond (update_mask);
}


/*
 * Protected methods
 * -----------------
 */
/**/
int GBrush::_addCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = &_assoc;
    GAssoc* cb = (GAssoc*)_cb;
    
    GCallbackRec cbrec (proc, closure);
    return cb->add(type, sizeof cbrec, (char*)&cbrec);
}

/**/
int GBrush::_remCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = &_assoc;
    GAssoc* cb = (GAssoc*)_cb;
    
    GCallbackRec cbrec (proc, closure);
    return cb->rem(type, sizeof cbrec, (char*)&cbrec);
}

/**/
GBool GBrush::_call_modify_precond (ulong& modify_mask, GBrushValues* new_values)
{
    if (_flags._has_modify_precond)
    {
	int count;
	PrecondRec* cbrec = (PrecondRec*)_assoc.get (brushNmodifyPrecond, count);
	for (int i = 0; i < count; i++, cbrec++)
	    if (!cbrec->proc(cbrec->closure, this, modify_mask, new_values))
                return GFALSE;
    }
    return GTRUE;
}

/**/
void GBrush::_call_modify_postcond (ulong modify_mask)
{
    if (_flags._has_modify_postcond)
    {
	int count;
	GCallbackRec* cbrec = (GCallbackRec*)_assoc.get (brushNmodifyPostcond, count);
	(void)GCallCallbacks(cbrec, count, this, modify_mask);
    }
}

/**/
void GBrush::_call_destroy_postcond ()
{
    if (_flags._has_destroy_postcond)
    {
	int count;
	DestroyPostcondRec* cbrec = (DestroyPostcondRec*)_assoc.get (brushNdestroyPostcond, count);
	for (int i = 0; i < count; i++, cbrec++)
	     cbrec->proc(cbrec->closure, this);
    }
}

/*
 * Public methods
 * --------------
 */
GBrush::GBrush ()
{
    _sh_brush = GCreateShBrush(); /* default one */
}

GBrush::GBrush (const GBrush* brush)
{
    _sh_brush = brush->_sh_brush->dup();
}

GBrush::GBrush (const GShBrush* sh_brush)
{
    _sh_brush = sh_brush->dup();
}

GBrush::GBrush (ulong mask, const GBrushValues* values, const char* name)
{
    _sh_brush = GCreateShBrush (mask, values, name);
}

GBrush::GBrush (const char* name)
{
    _sh_brush = GCreateShBrush (name);
}

GBrush::~GBrush ()
{
    _call_destroy_postcond ();
    if (_flags._has_modify_postcond)
        _sh_brush->updatePostcond (_call_call_modify_postcond, this, GREMOVE);
    _sh_brush->destroy();
}

/**
FUNCTION
    GBrush::modify
DESCRIPTION
    Modify the shared brush and call pre/post condition.
**/

void GBrush::modify (ulong mask, const GBrushValues* values)
{
    GBrushValues Values = *values;
    if (mask && _call_modify_precond (mask, &Values))
    {
	// Remove postcond;
	if (_flags._has_modify_postcond)
	    _sh_brush->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

	// Modify;
        GModify(_sh_brush, mask, &Values);

	// Call postcond;
	_call_modify_postcond (mask);

	// Add postcond;
	if (_flags._has_modify_postcond)
	    _sh_brush->updatePostcond (_call_call_modify_postcond, this, GADD);
    }
}

void GBrush::modify (const char* name)
{
    if (name && _sh_brush->name() && (strcmp (name, _sh_brush->name()) == 0))
        return;

    // Remove postcond.
    if (_flags._has_modify_postcond)
        _sh_brush->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

    // Get mask and values
    GShBrush* save = _sh_brush->dup();

    GRename (save, name);

    GBrushValues Values = *save->values();
    ulong mask = save->mask();

    if (mask && _call_modify_precond (mask, &Values))
    {
         _sh_brush->destroy();
         _sh_brush = GCreateShBrush (mask, &Values, name);
         _call_modify_postcond (mask);
    }
    save->destroy();
    
    // Add postcond.
    if (_flags._has_modify_postcond)
	_sh_brush->updatePostcond (_call_call_modify_postcond, this, GADD);
}

void GBrush::modify (const GShBrush* sh_brush)
{
    if (sh_brush == _sh_brush)
        return;
    
    // top of list.
    sh_brush->raise(GShBrush::_headOfList);
    
    if (sh_brush->isNamed())
	modify(sh_brush->name());
    else 
    {
	GBrushValues Values = *sh_brush->values();
	ulong mask = sh_brush->mask();
	    
	// Remove postcond;
	if (_flags._has_modify_postcond)
	    _sh_brush->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

	// unname the _sh_brush;
	GRename(_sh_brush, NULL);
		
	if (mask && _call_modify_precond (mask, &Values))
	{
	    // Modify;
	    GModify (_sh_brush, mask, &Values);

	    // Call postcond;
	    _call_modify_postcond (mask);
	}

	// Add postcond;
	if (_flags._has_modify_postcond)
	    _sh_brush->updatePostcond (_call_call_modify_postcond, this, GADD);
    }

}

/**
FUNCTION
    GBrush::field_modification_methods
DESCRIPTION
    Each method modify one parameter of the GBrush.
**/

#define _modifyBrushValue(bit_mask, field, value) \
    GBrushValues values;\
    values.field = value;\
    modify (bit_mask, &values)

void GBrush::function (ulong function)
{
    _modifyBrushValue(GFUNCTION, function, function);
}

void GBrush::planeMask (ulong plane_mask)
{
    _modifyBrushValue(GPLANE_MASK, plane_mask, plane_mask);
}

/**/
void GBrush::color (const GColor* color)
{
    _modifyBrushValue(GCOLOR, color, (GShColor*)color->shColor());
}

void GBrush::color (const GShColor* color)
{
    _modifyBrushValue(GCOLOR, color, (GShColor*)color);
}

void GBrush::color (const char* name)
{
    GBrushValues values;
    values.color = GCreateShColor (name);
    modify (GCOLOR, &values);
    values.color->destroy();
}

/**/
void GBrush::altColor (const GColor* color)
{
    _modifyBrushValue(GALT_COLOR, alt_color, (GShColor*)color->shColor());
}

void GBrush::altColor (const GShColor* color)
{
    _modifyBrushValue(GALT_COLOR, alt_color, (GShColor*)color);
}

void GBrush::altColor (const char* name)
{
    GBrushValues values;
    values.alt_color = GCreateShColor (name);
    modify (GALT_COLOR, &values);
    values.alt_color->destroy();
}

/**/
void GBrush::fillStyle (ushort fill_style)
{
    _modifyBrushValue(GFILL_STYLE, fill.style, fill_style);
}

void GBrush::fillRule (ushort fill_rule)
{
    _modifyBrushValue(GFILL_RULE, fill.rule, fill_rule);
}

void GBrush::tile (const GPattern* tile)
{
    _modifyBrushValue(GTILE, tile, (GShPattern*)tile->shPattern());
}

void GBrush::tile (const GShPattern* tile)
{
    _modifyBrushValue(GTILE, tile, (GShPattern*)tile);
}

void GBrush::tile (const char* name)
{
    GBrushValues values;
    values.tile = GCreateShPattern(name);
    modify (GTILE, &values);
    values.tile->destroy();
}

void GBrush::stipple (const GPattern* stipple)
{
    _modifyBrushValue(GSTIPPLE, stipple, (GShPattern*)stipple->shPattern());
}

void GBrush::stipple (const GShPattern* stipple)
{
    _modifyBrushValue(GSTIPPLE, stipple, (GShPattern*)stipple);
}

void GBrush::stipple (const char* name)
{
    GBrushValues values;
    values.stipple = GCreateShPattern(name);
    modify (GSTIPPLE, &values);
    values.stipple->destroy();
}

void GBrush::tsOrigin (GNumber x_origin, GNumber y_origin)
{
    GBrushValues values;
    values.ts_x_origin = x_origin;
    values.ts_y_origin = y_origin;
    modify (GTILE_STIP_XORIGIN | GTILE_STIP_YORIGIN, &values);
}
#undef _modifyBrushValue

/**
FUNCTION
    GBrush::add/rem pre/postconditions
DESCRIPTION
    Add or remove pre/post condition on modification and on destruction.

    GWarning, if the GShBrush is named, _call_call_modify_postcond
    is stored directly in the shared object and pre/condition
    may be not called!
**/

#define _add_rem_callback(type, flag) \
    (FLAGS->flag =\
         0 < ((add == GADD)\
	      ? _addCallback(type, (GCBProc)proc, closure)\
	      : _remCallback(type, (GCBProc)proc, closure)))

/**/
void GBrush::modifyPrecond (PeProc proc,
			   const void* closure, GAddOrRemove add) const
{
    const GBrushCondFlags* flags = &_flags;
    GBrushCondFlags* FLAGS = (GBrushCondFlags*)flags;
    
   _add_rem_callback(brushNmodifyPrecond, _has_modify_precond);
}

/**/
void GBrush::modifyPostcond (PsProc proc, const void* closure, GAddOrRemove add) const
{
    ushort flag = _flags._has_modify_postcond;

    const GBrushCondFlags* flags = &_flags;
    GBrushCondFlags* FLAGS = (GBrushCondFlags*)flags;
    
    _add_rem_callback(brushNmodifyPostcond, _has_modify_postcond);

    /* The modifyPostcond is called from the sh_brush */

    if (!flag && _flags._has_modify_postcond)
	_sh_brush->updatePostcond (_call_call_modify_postcond, this, GADD);
    else
	if (flag && !_flags._has_modify_postcond)
	    _sh_brush->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

}

/**/
void GBrush::destroyPostcond (PdProc proc, const void* closure, GAddOrRemove add) const
{
    const GBrushCondFlags* flags = &_flags;
    GBrushCondFlags* FLAGS = (GBrushCondFlags*)flags;
    
   _add_rem_callback(brushNdestroyPostcond, _has_destroy_postcond);
}

#undef _add_rem_callback
