#ifdef REV_INFO
static char rcsid[] = "$Id: ShPen.cc,v 1.4 1992/12/10 13:33:29 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  GoGC
FILE:    ShPen.cc
AUTHOR:  Jacques Davy
DATE:    October 1990

DESCRIPTION:
    (See <Go/Tools/ShPen.h>

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  license from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/Tools/Error.h>
#include <Go/Tools/ShPen.h>
#include <Go/Tools/ShColor.h>
#include <Go/Tools/ShPattern.h>
#include <Go/Tools/ShDash.h>
#include <Go/GoStream.h>
#include <assert.h>

/*
 * static of the class
 */
GList*        GShPen::_headOfList;
GFunctionList GShPen::_funcList;

/*
 * Default values for ShPen constructor
 */

#define DEFAULT_FUNCTION    GCOPY
#define DEFAULT_PLANE_MASK  GALL_PLANES
#define DEFAULT_COLOR       GCreateShColor()
#define DEFAULT_ALT_COLOR   NULL

#define DEFAULT_LINE_WIDTH  0.3
#define DEFAULT_LINE_STYLE  GLINE_SOLID
#define DEFAULT_CAP_STYLE   GCAP_BUTT
#define DEFAULT_JOIN_STYLE  GJOIN_MITER
#define DEFAULT_MITER_LIMIT GNO_MITER_LIMIT
#define DEFAULT_DASH        NULL

#define DEFAULT_FILL_STYLE  GFILL_SOLID
#define DEFAULT_TILE        NULL
#define DEFAULT_STIPPLE     NULL
#define DEFAULT_TS_X_ORIGIN 0.0
#define DEFAULT_TS_Y_ORIGIN 0.0

#define DEFAULT_MASK GFUNCTION   | \
	             GPLANE_MASK  | \
	             GCOLOR      | \
	             GLINE_WIDTH  | \
	             GLINE_STYLE  | \
	             GCAP_STYLE   | \
	             GJOIN_STYLE  | \
	             GMITER_LIMIT | \
	             GFILL_STYLE


/*
 * static post conds
 */
static void _ColorIsUpdated(void* This, GShColor* color, ulong mask)
{
    ColorIsUpdated((GShPen*)This, color, mask);
}

void ColorIsUpdated(GShPen* This, const GShColor* color, ulong mask)
{
   const GPenValues* values = This->values();
   if (color == values->color)
       mask = GCOLOR;
   else if (color == values->alt_color)
       mask = GALT_COLOR;
   else 
	GError ("%s %s %s", "GShPen.cc", "ColorIsUpdated",
		"values and parameter incoherence");

   assert (color->isNamed());
   This->_call_update_postcond(mask);
}

static void _DashIsUpdated(void* This, GShDash* dash, ulong mask)
{
    DashIsUpdated((GShPen*)This, dash, mask);
}

void DashIsUpdated(GShPen* This, const GShDash* dash, ulong mask)
{
   assert (dash == This->dash());
   assert (dash->isNamed() && mask);
   This->_call_update_postcond(GDASH);
}

static   void _PatternIsUpdated(void* This, GShPattern* pattern,
				     ulong mask)
{
    PatternIsUpdated((GShPen*)This, pattern, mask);
}

void PatternIsUpdated(GShPen* This, const GShPattern* pattern,
		      ulong mask)
{
    const GPenValues* values = This->values();
    if (pattern == values->tile)
	mask = GTILE;
    else if (pattern == values->stipple)
	mask = GSTIPPLE;
    else
	GError ("%s %s %s", "GShPen.cc", "PatternIsUpdated",
		"values and parameter incoherence");
    
    This->_call_update_postcond(mask);
}

/*
 * Protected Methods
 */


int GShPen::save(const GShPen* sh, GoStream* gostream) {
    const GPenValues* val = sh->values();
    unsigned long mask = sh->mask();
    GBool named = sh->isNamed();
    
    if (!gostream->writeBool(named, "is named"))
	goto failure;
    if (named && !gostream->writeString(sh->name(), "name"))
	goto failure;
    if (!gostream->writeLong(sh->mask(), "mask"))
	goto failure;
    if ((mask & GFUNCTION) &&
	! gostream->writeLong(val->function, "function"))
	goto failure;
    if ((mask & GPLANE_MASK) &&
	! gostream->writeLong(val->plane_mask, "plane mask"))
	goto failure;
    if ((mask & GCOLOR) &&
	! gostream->saveObj(val->color, val->color->getClassname(), "color"))
	goto failure;
    if ((mask & GALT_COLOR) &&
	! gostream->saveObj(val->alt_color, val->alt_color->getClassname(), "alt color"))
	goto failure;
    if ((mask & GLINE_WIDTH) &&
	! gostream->writeNumber(val->line_width, "line width"))
	goto failure;
    if ((mask & GLINE_STYLE) &&
	! gostream->writeChar((char) val->line.style, "line style"))
	goto failure;
    if ((mask & GCAP_STYLE) &&
	! gostream->writeChar((char) val->line.cap, "line cap"))
	goto failure;
    if ((mask & GJOIN_STYLE) &&
	! gostream->writeChar((char) val->line.join, "line join"))
	goto failure;
    if ((mask & GMITER_LIMIT) &&
	! gostream->writeNumber(val->miter_limit, "miter limit"))
	goto failure;
    if ((mask & GDASH) &&
	! gostream->saveObj(val->dash, val->dash->getClassname(), "dash"))
	goto failure;
    if ((mask & GFILL_STYLE) &&
	! gostream->writeChar((char) val->fill.style, "fill style"))
	goto failure;
    if ((mask & GTILE) &&
	! gostream->saveObj(val->tile, val->tile->getClassname(), "tile"))
	goto failure;
    if ((mask & GSTIPPLE) &&
	! gostream->saveObj(val->stipple, val->stipple->getClassname(), "stipple"))
	goto failure;
    if ((mask & GTILE_STIP_XORIGIN) &&
	! gostream->writeNumber(val->ts_x_origin, "ts x origin"))
	goto failure;
    if ((mask & GTILE_STIP_YORIGIN) &&
	! gostream->writeNumber(val->ts_y_origin, "ts y origin"))
	goto failure;
    return GSharedObject::save(sh, gostream);

  failure:
    GWarning("GShPen save failure");
    return 0;
}

/**/
int GShPen::restore(GShPen** sh, GoStream* gostream) {
    long mask;
    GPenValues val;
    char c, str[256];
    char *name = str;
    GBool named;
    const char* class_name;
    if (!gostream->readBool(&named))
	goto failure;
    if (named && !gostream->readString(256, &name))
	goto failure;
    if (!gostream->readLong(&mask))
	goto failure;
    
    if ((mask & GFUNCTION) && ! gostream->readLong((long*) &val.function))
	goto failure;
    if ((mask & GPLANE_MASK) && ! gostream->readLong((long*) &val.plane_mask))
	goto failure;
    if ((mask & GCOLOR) &&
	! gostream->restoreObj((void**) &(val.color), &class_name))
	goto failure;
    if ((mask & GALT_COLOR) &&
	! gostream->restoreObj((void**) &(val.alt_color), &class_name))
	goto failure;
    if ((mask & GLINE_WIDTH) && ! gostream->readNumber(&val.line_width))
	goto failure;
    if (mask & GLINE_STYLE) {
	if (! gostream->readChar(&c))
	    goto failure;
	else val.line.style = c;
    }
    if (mask & GCAP_STYLE) {
	if (! gostream->readChar(&c))
	    goto failure;
	else val.line.cap = c;
    }
    if (mask & GJOIN_STYLE) {
	if (! gostream->readChar(&c))
	    goto failure;
	else val.line.join = c;
    }
    if ((mask & GMITER_LIMIT) && ! gostream->readNumber(&val.miter_limit))
	goto failure;
    if ((mask & GDASH) &&
	! gostream->restoreObj((void**) &(val.dash), &class_name))
	goto failure;
    if (mask & GFILL_STYLE) {
	if (! gostream->readChar(&c))
	    goto failure;
	else val.fill.style = c;
    }
    if ((mask & GTILE) &&
	! gostream->restoreObj((void**) &(val.tile), &class_name))
	goto failure;
    if ((mask & GSTIPPLE) &&
	! gostream->restoreObj((void**) &(val.stipple), &class_name))
	goto failure;
    if ((mask & GTILE_STIP_XORIGIN) && ! gostream->readNumber(&val.ts_x_origin))
	goto failure;
    if ((mask & GTILE_STIP_YORIGIN) && ! gostream->readNumber(&val.ts_y_origin))
	goto failure;

    if (*sh = GCreateShPen(mask, &val, named ? name : NULL))
    {
	if (GSharedObject::readParams(*sh, gostream))
	    return 1;
	else
	{
	    (*sh)->destroy();
	    *sh = NULL;
	}
    }
    
  failure:
    GWarning("GShPen restore failure");
    return 0;
}

/**/
int GShPen::readParams(GShPen*, GoStream*) {
    return 0;
}

/**/
GShPen* _newGShPen (GFunctionList* funcList, ulong mask, const void * values,
		  const char* name)
{
    return new GShPen (funcList, mask, (const GPenValues*) values, name);
}

/**/
GShPen::GShPen (GFunctionList* funcList, ulong mask, const GPenValues* values,
	      const char* name) : GSharedObject(funcList, mask, name)
{
#define _setValue(maskBit, field, default_value) \
    _values.field = (mask & maskBit ? \
		     values->field : default_value)
#define _setSharedValue(maskBit, field, default_value) \
    _values.field = (mask & maskBit ? \
		     values->field->dup() : default_value)

    _setValue (GFUNCTION, function, DEFAULT_FUNCTION);
    _setValue (GPLANE_MASK, plane_mask, DEFAULT_PLANE_MASK);
    _setSharedValue (GCOLOR, color, DEFAULT_COLOR);
    if (_values.color->isNamed())
	_values.color->updatePostcond(_ColorIsUpdated, this, GADD);

    _setSharedValue (GALT_COLOR, alt_color, DEFAULT_ALT_COLOR);
    if (_values.alt_color && _values.alt_color->isNamed())
	_values.alt_color->updatePostcond(_ColorIsUpdated, this, GADD);

    _setValue (GLINE_WIDTH, line_width, DEFAULT_LINE_WIDTH);
    _setValue (GLINE_STYLE, line.style, DEFAULT_LINE_STYLE);
    _setValue (GCAP_STYLE, line.cap, DEFAULT_CAP_STYLE);
    _setValue (GJOIN_STYLE, line.join, DEFAULT_JOIN_STYLE);
    _setValue (GMITER_LIMIT, miter_limit, DEFAULT_MITER_LIMIT);

    _setSharedValue (GDASH, dash, DEFAULT_DASH);
    if (_values.dash && _values.dash->isNamed())
	_values.dash->updatePostcond(_DashIsUpdated, this, GADD);
    
    _setValue (GFILL_STYLE, fill.style, DEFAULT_FILL_STYLE);
    _setSharedValue (GTILE, tile, DEFAULT_TILE);
    if (_values.tile)
	_values.tile->updatePostcond(_PatternIsUpdated, this, GADD);
    
    _setSharedValue (GSTIPPLE, stipple, DEFAULT_STIPPLE);
    if (_values.stipple)
	_values.stipple->updatePostcond(_PatternIsUpdated, this, GADD);

    _setValue (GTILE_STIP_XORIGIN, ts_x_origin, DEFAULT_TS_X_ORIGIN);
    _setValue (GTILE_STIP_YORIGIN, ts_y_origin, DEFAULT_TS_Y_ORIGIN);

    _mask = mask | DEFAULT_MASK;
#undef _setValue
#undef _setSharedValue
}

/**/
GShPen::~GShPen()
{
    if (_values.color->isNamed())
	_values.color->updatePostcond(_ColorIsUpdated, this, GREMOVE);
    _values.color->destroy();

    if (_mask & GALT_COLOR) {
	if (_values.alt_color->isNamed())
	    _values.alt_color->updatePostcond(_ColorIsUpdated, this, GREMOVE);
        _values.alt_color->destroy();
    }

    if (_mask & GTILE) {
	_values.tile->updatePostcond(_PatternIsUpdated, this, GREMOVE);
        _values.tile->destroy();
    }
    if (_mask & GSTIPPLE) {
	_values.stipple->updatePostcond(_PatternIsUpdated, this, GREMOVE);
	_values.stipple->destroy();
    }
    
    if (_mask & GDASH) {
	if (_values.dash->isNamed())
	    _values.dash->updatePostcond(_DashIsUpdated, this, GREMOVE);
	_values.dash->destroy();
    }
}

/**/
GBool GShPen::_match (ulong mask, const void * __values) const 
{
 const GPenValues* values =( const GPenValues* ) __values;

    ulong inter_mask = mask & _mask;

    if (!inter_mask)
        return GTRUE;

#define _matchValue(maskBit, field) \
    if (!inter_mask) return GTRUE; \
    if (inter_mask & maskBit)\
    {\
	inter_mask &= ~maskBit ;\
	if ( values->field != _values.field ) return GFALSE ;\
    }

    _matchValue (GFUNCTION, function);
    _matchValue (GPLANE_MASK, plane_mask);
    _matchValue (GCOLOR, color);
    _matchValue (GLINE_WIDTH, line_width);
    _matchValue (GLINE_STYLE, line.style);
    _matchValue (GCAP_STYLE, line.cap);
    _matchValue (GJOIN_STYLE, line.join);
    _matchValue (GFILL_STYLE, fill.style);
    _matchValue (GMITER_LIMIT, miter_limit);
    _matchValue (GTILE, tile);
    _matchValue (GSTIPPLE, stipple);
    _matchValue (GTILE_STIP_XORIGIN, ts_x_origin);
    _matchValue (GTILE_STIP_YORIGIN, ts_y_origin);
    _matchValue (GALT_COLOR, alt_color);
    _matchValue (GDASH, dash);
#undef _matchValue

    return GTRUE;
}

/**/
void GShPen::_makeNewValues (ulong mask, const void * __values,
			    ulong& newMask, void * __newValues) const
{
 const GPenValues* values =( const GPenValues* ) __values;
 GPenValues* newValues =(GPenValues* ) __newValues;

    ulong orMask = mask | _mask;
    newMask = orMask;

#define _makeNewValue(maskBit, field) \
    if (!orMask) return;\
    newValues->field = (mask & maskBit ?\
			values->field : _values.field);\
    orMask &= ~maskBit

/* GSharedObject field are not "dup()ed" because it is not necessary */
    _makeNewValue (GFUNCTION, function);
    _makeNewValue (GPLANE_MASK, plane_mask);
    _makeNewValue (GCOLOR, color);
    _makeNewValue (GLINE_WIDTH, line_width);
    _makeNewValue (GLINE_STYLE, line.style);
    _makeNewValue (GCAP_STYLE, line.cap);
    _makeNewValue (GJOIN_STYLE, line.join);
    _makeNewValue (GFILL_STYLE, fill.style);
    _makeNewValue (GTILE, tile);
    _makeNewValue (GSTIPPLE, stipple);
    _makeNewValue (GTILE_STIP_XORIGIN, ts_x_origin);
    _makeNewValue (GTILE_STIP_YORIGIN, ts_y_origin);
    _makeNewValue (GALT_COLOR, alt_color);
    _makeNewValue (GMITER_LIMIT, miter_limit);
    _makeNewValue (GDASH, dash);
#undef _makeNewValue
}

/**/
ulong GShPen::_update (ulong mask, const void * __values)
{
 const GPenValues* values =( const GPenValues* ) __values;

    ulong old_mask = _mask;
    ulong update_mask = 0;
    _mask |= mask;

#define _updatePrivateValue(maskBit, field)\
    if (!mask) return update_mask;\
    if (mask & maskBit) {\
	if  (!(old_mask & maskBit) || (_values.field != values->field)){\
           update_mask |= maskBit;\
	   _values.field = values->field;\
        }\
        mask &= ~maskBit;\
    }

#define _updateColorSharedValue(maskBit, field)\
    if (!mask) return update_mask;\
    if (mask & maskBit) {\
	if (!(old_mask & maskBit) || (_values.field != values->field)){\
           update_mask |= maskBit;\
           if (old_mask & maskBit) {\
	       if (_values.field->isNamed())\
		   _values.field->updatePostcond(_ColorIsUpdated, this, GREMOVE);\
              _values.field->destroy();\
	      }\
	   _values.field = values->field->dup();\
           if (_values.field->isNamed())\
	       _values.field->updatePostcond(_ColorIsUpdated, this, GADD);\
        }\
        mask &= ~maskBit;\
    }
    
#define _updateDashSharedValue(maskBit, field)\
    if (!mask) return update_mask;\
    if (mask & maskBit) {\
	if (!(old_mask & maskBit) || (_values.field != values->field)){\
           update_mask |= maskBit;\
           if (old_mask & maskBit) {\
	       if (_values.field->isNamed())\
		   _values.field->updatePostcond(_DashIsUpdated, this, GREMOVE);\
              _values.field->destroy();\
	      }\
	   _values.field = values->field->dup();\
           if (_values.field->isNamed())\
	       _values.field->updatePostcond(_DashIsUpdated, this, GADD);\
        }\
        mask &= ~maskBit;\
    }

#define _updatePatternSharedValue(maskBit, field)\
    if (!mask) return update_mask;\
    if (mask & maskBit) {\
	if (!(old_mask & maskBit) || (_values.field != values->field)){\
           update_mask |= maskBit;\
           if (old_mask & maskBit) {\
              _values.field->updatePostcond(_PatternIsUpdated, this, GREMOVE);\
              _values.field->destroy();\
	      }\
	   _values.field = values->field->dup();\
           _values.field->updatePostcond(_PatternIsUpdated, this, GADD);\
        }\
        mask &= ~maskBit;\
    }

    _updateColorSharedValue (GCOLOR, color);
    _updatePrivateValue (GLINE_WIDTH, line_width);
    _updateColorSharedValue (GALT_COLOR, alt_color);
    _updatePrivateValue (GFUNCTION, function);
    _updatePrivateValue (GFILL_STYLE, fill.style);
    _updatePatternSharedValue (GTILE, tile);
    _updatePatternSharedValue (GSTIPPLE, stipple);
    _updatePrivateValue (GTILE_STIP_XORIGIN, ts_x_origin);
    _updatePrivateValue (GTILE_STIP_YORIGIN, ts_y_origin);
    _updatePrivateValue (GLINE_STYLE, line.style);
    _updatePrivateValue (GCAP_STYLE, line.cap);
    _updatePrivateValue (GJOIN_STYLE, line.join);
    _updatePrivateValue (GMITER_LIMIT, miter_limit);
    _updateDashSharedValue (GDASH, dash);
    _updatePrivateValue (GPLANE_MASK, plane_mask);

#undef _updatePrivateValue
#undef _updateColorSharedValue
#undef _updateDashSharedValue
#undef _updatePatternSharedValue
    return update_mask;
}

/*
 * Public Methods
 */

/* static */
const char* GShPen::classname()
{
    return "GShPen";
}

/* virtual */
const char* GShPen::getClassname() const
{
    return GShPen::classname();
}

/**/
void GModify (const char* name, ulong mask, const GPenValues* values)
{
    GMODIFY_NAMED_SHAREDOBJECT (GShPen, mask, values, name);
}

static GShPen* default_shpen = NULL;
/**/
GShPen* GCreateShPen (ulong mask, const GPenValues* values, const char* name)
{
    if (mask == 0 && name == NULL)
    {
	if (default_shpen == NULL)
	{
	    GPenValues _values;
	    _values.function =    DEFAULT_FUNCTION;
	    _values.plane_mask =  DEFAULT_PLANE_MASK;
	    _values.color =       DEFAULT_COLOR;
	    _values.line_width =  DEFAULT_LINE_WIDTH;
	    _values.line.style =  DEFAULT_LINE_STYLE;
	    _values.line.cap =    DEFAULT_CAP_STYLE;
	    _values.line.join  =  DEFAULT_JOIN_STYLE;
	    _values.miter_limit = DEFAULT_MITER_LIMIT;
	    _values.fill.style  = DEFAULT_FILL_STYLE;
	    
	    default_shpen = GCREATE_SHAREDOBJECT(GShPen, DEFAULT_MASK, &_values, NULL);
	    _values.color->destroy();
	}
	return default_shpen->dup();
    }
    return GCREATE_SHAREDOBJECT (GShPen, mask, values, name);
}


/**/
GShPen* GCreateShPen (const char* name)
{
    return GCreateShPen(0, NULL, name);
}

/**/
GShPen* GCreateShPen ()
{
    return GCreateShPen(0, NULL, NULL);
}
