#ifdef REV_INFO
static char rcsid[] = "$Id: GeomUtil.cc,v 1.3 1992/12/10 13:33:34 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  Tools
FILE:    GeomUtil.cc
AUTHOR:  Olivier Jojic, Gilles Benati
DATE:    April 10th, 1991

DESCRIPTION:
  See Go/Tools/geomUtil.h header file

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/Tools/GeomUtil.h>
#include <Go/Tools/Matrix.h>
#include <assert.h>

#define SIN_11_DEG .1908089954

#ifndef M_PI_2
#define M_PI_2 1.57079632679489661923
#endif


GLine::GLine(const GPoint& p1, const GPoint& p2) {
    if (p1 == p2) {
	type = GPOINT_LINE;
	a = p1.x;
	b = p1.y;
    } else if (p1.x == p2.x) {
	type = GY_LINE;
	a = p1.x;
    } else {
	type = GNORMAL_LINE;
	a = (p2.y - p1.y) / (p2.x - p1.x);
	b = (p1.y*p2.x - p1.x*p2.y) / (p2.x - p1.x);
}   }

GBool GLine::setX(GPoint& p) {
    if (type == GPOINT_LINE) {
	p.x = a;
	return p.y == b;
    } else if (type == GY_LINE) {
	p.x = a;
	return GTRUE;
    } else if (a == 0) {
	return GFALSE; //???
    } else {
	p.x = (p.y - b) / a;
	return GTRUE;
}   }

GBool GLine::setY(GPoint& p) {
    if (type == GPOINT_LINE) {
	p.y = b;
	return p.x == a;
    } else if (type == GY_LINE) {
	if (p.x != a) return GFALSE;
	return GFALSE; //???
    } else {
	p.y = a*p.x + b;
	return GTRUE;
}   }


GBool GIntersect(const GPoint& a1, const GPoint& b1,
		 const GPoint& a2, const GPoint& b2, GBool& RtoL) {
    GVector i = b2 - a2;
    GCoordSys cs(a2, i, GVector(-i.y, i.x));  // j = i; j.rotate(M_PI_2);
    GCoordSys cs_inv = ~cs;
    GPoint ra1 = cs_inv.to(a1);
    GPoint rb1 = cs_inv.to(b1);
    RtoL = ((rb1.y - ra1.y) >= 0);
    if (ra1.y * rb1.y > 0)
 	return GFALSE;
    GLine rline1(ra1, rb1);
    GPoint p(0, 0);
    if (!rline1.setX(p))
 	return GFALSE;
    return (p.x >= 0 && p.x <= 1);
}

/**
FUNCTION
    void GRectangleClip(const GPoint& p1, const GPoint& p2, GBool filled,
    		        GNumber maxDist, GRegion& region)
    
DESCRIPTION
    This function works with the horizontal rectangle defined by the
    diagonal segment [p1, p2].
    It adds to region a Region containing all points P(x,y) such as :
    point P(x,y) is close to the rectangle.
    i.e. distance(P, R) <= maxDist.
    If filled is true, the function considers a filled rectangle
    else it considers an empty rectangle (outline only).
    The line_width **IS MEANINGLESS** in this context: maxDist must be set to
    an appropriate value to simulate the line_width.
**/
void GRectangleClip(const GPoint& p1, const GPoint& p2, GBool filled,
		    GNumber maxDist, GRegion& region) {
    if (maxDist < 0.)
	return;

    GPoint minP, maxP;
    
    if (p1.x > p2.x) {
	minP.x = p2.x;
	maxP.x = p1.x;
    }
    else {
	minP.x = p1.x;
	maxP.x = p2.x;
    }
    if (p1.y > p2.y) {
	minP.y = p2.y;
	maxP.y = p1.y;
    }
    else {
	minP.y = p1.y;
	maxP.y = p2.y;
    }

    if (filled) {
	if (maxDist != 0)
	    region += GBox(minP.x - maxDist, minP.y - maxDist,
			  maxP.x + maxDist, maxP.y + maxDist);
	else region += GBox(minP.x, minP.y, maxP.x, maxP.y);
	return;
    }
    else if (maxDist != 0) {
	region += GBox(minP.x - maxDist, minP.y - maxDist,
		      maxP.x + maxDist, minP.y + maxDist);
	region += GBox(minP.x - maxDist, maxP.y - maxDist,
		      maxP.x + maxDist, maxP.y + maxDist);
	region += GBox(minP.x - maxDist, minP.y - maxDist,
		      minP.x + maxDist, maxP.y + maxDist);
	region += GBox(maxP.x - maxDist, minP.y - maxDist,
		      maxP.x + maxDist, maxP.y + maxDist);
    }
    else {
	region += GBox(minP.x, minP.y, maxP.x, minP.y);
	region += GBox(minP.x, maxP.y, maxP.x, maxP.y);
	region += GBox(minP.x, minP.y, minP.x, maxP.y);
	region += GBox(maxP.x, minP.y, maxP.x, maxP.y);
    }
}


/**
FUNCTION
    GPolypointGetOutPoints(int nb, const GPoint *pts, &pmin, &pmax)
DESCRIPTION
    pts = array of nb GPoints
    Return in pmin/pmax the minimum/maximum x and y values of GPoints in pts.
    pts and nb *MUST* be non null.
**/
void GPolypointGetOutPoints(int nb, const GPoint *pts, GPoint& pmin,
			    GPoint& pmax) {
    assert(nb > 0 && pts != NULL);

    pmin = pts[0];
    pmax = pts[0];
    
    for (int i = 1; i < nb; i++) {
	if (pts[i].x < pmin.x)
	    pmin.x = pts[i].x;
	else if (pts[i].x > pmax.x)
	    pmax.x = pts[i].x;
	
	if (pts[i].y < pmin.y)
	    pmin.y = pts[i].y;
	else if (pts[i].y > pmax.y)
	    pmax.y = pts[i].y;
    }
}


/**
FUNCTION
    GPolypointGetAbsOutPoints(int nb, const GPoint *pts,
			      GPoint& pmin, GPoint& pmax, const GCoordSys *cs)
DESCRIPTION
    pts = array of nb GPoints
    cs = coordinate system in which the returned values are computed.
    Return in pmin/pmax the minimum/maximum x and y values in cs of GPoints in
    pts, relatively to cs.
    pts and nb *MUST* be non null.
**/
void GPolypointGetAbsOutPoints(const GCoordSys &cs, int nb, const GPoint *pts,
			       GPoint& pmin, GPoint& pmax) {
    assert(nb > 0 && pts != NULL);

    if (cs.isOrthogonal()) {
	GNumber tmp;
	GPolypointGetOutPoints(nb, pts, pmin, pmax);
	pmin = cs.to(pmin);
	pmax = cs.to(pmax);
	if (pmin.x > pmax.x) {
	    tmp = pmin.x;
	    pmin.x = pmax.x;
	    pmax.x = tmp;
	}
	if (pmin.y > pmax.y) {
	    tmp = pmin.y;
	    pmin.y = pmax.y;
	    pmax.y = tmp;
	}
    }
    else {
	GPoint p;
	pmin = cs.to(pts[0]);
	pmax = pmin;

	for (int i = 1; i < nb; i++) {
	    p = cs.to(pts[i]);
	    if (p.x < pmin.x)
		pmin.x = p.x;
	    else if (p.x > pmax.x)
		pmax.x = p.x;
	    if (p.y < pmin.y)
		pmin.y = p.y;
	    else if (p.y > pmax.y)
		pmax.y = p.y;
	}
    }
}

/**
FUNCTION
    GBool GPointInPolygon(int nb, const GPoint* pts,
                          const GPoint& point, GBool fill_winding)
DESCRIPTION
    Check whether point is within the area surrounded by the polygon defined
    by :
      o  nb, the number of summits, i.e. number of elements in pts[]
      o  pts[], the array of points

    fill_winding has the following meaning
      o  GFALSE <==> Winding rule : a point is "in" if an infinite ray with the
         point as origin is crossed and unequal number of times from left to
	 right and from right to left, as observed from the point, by the path
	 following the outline of the polygon.
	 directed path segments.
      o  GTRUE  <==> Odd-Even rule : a point is "in" if an infinite ray with
         the point as origin crosses the path an odd number of times.
**/
GBool GPointInPolygon(int nb, const GPoint* pts,
		      const GPoint& point, GBool fill_winding) {
    if (nb <= 2)
	return GFALSE;

    /*
     * For this algorithm, we need a point out of the polygon bounding box..
     */

    int count = 0;
    GPoint min_pt, max_pt;
    int i;

    GPolypointGetOutPoints(nb, pts, min_pt, max_pt);

    // if point is out of the bounding box...;
    if (point.x < min_pt.x || point.x > max_pt.x ||
	point.y < min_pt.y || point.y > max_pt.y)
	return(GFALSE);

    /*
     * Since the algorithm is going to deal with intersections between
     * the polygon path and an infinite radius, let us choose a radius that
     * cross a minimal part of the bounding box : we will hopefully minimize
     * the number of intersections that way.
     */

    // We first enlarge the box;
    min_pt = (min_pt - GPoint(10, 10)) - point;
    max_pt = (max_pt + GPoint(10, 10)) - point;

    // then store in min_pt the closest point to "point" on the bounding box;
    if (max_pt.x < - min_pt.x)
	min_pt.x = max_pt.x;
    if (max_pt.y < - min_pt.y)
	min_pt.y = max_pt.y;
    if (fabs(min_pt.x) < fabs(min_pt.y))
	min_pt = GPoint(point.x + min_pt.x, point.y);
    else min_pt = GPoint(point.x, point.y + min_pt.y);
    
    GBool RtoL;  // true if path crosses ray from right to left;
    
    for (i = 1; i < nb; i++) {
	if (GIntersect(point, min_pt, pts[i-1], pts[i], RtoL)) {
	    if (fill_winding && RtoL)
		count--;
	    else count++;
	}
		
    }
    if (GIntersect(point, min_pt, pts[nb-1], pts[0], RtoL)) {
	if (fill_winding && RtoL)
	    count--;
	else count++;
    }

    return(fill_winding ? count : count % 2);
}


/**
FUNCTION
    GBool GPointOnPolyline(const GCoordSys &cs, int nb, const GPoint* pts,
                           const GPoint& absGPoint, GNumber maxSquare,
			   GBool closed)
DESCRIPTION
    pts is an array of nb GPoints expressed in cs coordinate system.
    absGPoint is a GPoint expressed in the absolute coordinate system.
    Returns GTRUE if the square distance from absGPoint to one segment 
    pts[i] pts[i+1] is smaller than maxSquare,
    else returns GFALSE.
    If closed is GTRUE, we take in account the segment pts[nb-1] to pts[0] 
    (pts[] is a polygon), else we ignore it (pts[] is a polyline).
**/
GBool GPointOnPolyline(const GCoordSys &cs, int nb, const GPoint* pts,
		       const GPoint& absGPoint, GNumber maxSquare,
		       GBool closed) {
    if (nb <= 0 || maxSquare < 0. || ! pts)
	return GFALSE;

    GPoint prev = cs.to(pts[0]);

    if (nb == 1)
	return (GSquareDistance(absGPoint, prev) < maxSquare);

    int i;
    
    for (i = 1; i < nb ; i++) {
	GPoint crt = cs.to(pts[i]);
	if (GSquareDistance(absGPoint, prev, crt) < maxSquare)
	    return GTRUE;
	prev = crt;
    }
    if (closed && nb > 2)
	return(GSquareDistance(absGPoint, cs.to(pts[0]), prev) < maxSquare);
    return GFALSE;
}


/**
FUNCTION
    GBox GPolypointJoinMiterBox(const GCoordSys &cs, int nb, const GPoint *pts,
                                GNumber absWidth, GBool closed)
DESCRIPTION
    pts is an array of nb GPoints (**minimum 3**), expressed in the coordinate
    system cs and seen as a polygon if closed is GTRUE, else seen as a polyline.
    Returns the smallest box containing the poly(line/gon), including the 
    drawing extensions caused by the following attributes of pen():
	o  lineWidth = absWidth (assumed to be > 0)
        o  joinStyle = JoinMiter (the spiky one).
    All points are expressed in cs coordinate system and 
    absWidth is an absolute value.
    
**/
GBox GPolypointJoinMiterBox(const GCoordSys &cs, int nb, const GPoint *pts,
			    GNumber absWidth, GBool closed) {
    GPoint p1, p2, p3;      // at each step summit = p2;
    GVector u, v;           // u=GVector(p2,p1) v=GVector(p2,p3);
                            // angle between u and v , 0 <= alpha <= PI;
    GNumber du, dv; // ||u||, ||v|| i.e. du=dist(p1 p2), dv=dist(p2 p3);
    GNumber xmin, xmax, ymin, ymax; // summits of the returned box;
    GNumber xx, yy;         // the coordinates to be computed for each summit;
    GNumber coeff;          // to store an intermediate value;
    int i;

    assert(nb >= 3 && absWidth > 0);

    absWidth = absWidth / 2;
    p1 = cs.to(pts[0]);
    xmin = p1.x - absWidth;
    ymin = p1.y - absWidth;
    xmax = p1.x + absWidth;
    ymax = p1.y + absWidth;
    
    p2 = cs.to(pts[1]);
    u = (GVector) (p1 - p2);
    du = u.size();
    
    for (i = 1; i < nb - 1; i++) {
	p3 = cs.to(pts[i + 1]);
	v = (GVector) (p3 - p2);
	dv = v.size();
	coeff = fabs(v.x * u.y - v.y * u.x);  // vect_prod == du*dv*sin(alpha);
	if (coeff <= (SIN_11_DEG * (du * dv))) {
	    // alpha <= 11 deg;
	    xmin = min(xmin, p2.x - absWidth);
	    ymin = min(ymin, p2.y - absWidth);
	    xmax = max(xmax, p2.x + absWidth);
	    ymax = max(ymax, p2.y + absWidth);
	}
	else {
	    coeff = - absWidth / coeff;
	    xx = p2.x + coeff * (dv * u.x + du * v.x);
	    yy = p2.y + coeff * (dv * u.y + du * v.y);
	    if (xx > xmax)
		xmax = xx;
	    else if (xx < xmin)
		xmin = xx;
	    if (yy > ymax)
		ymax = yy;
	    else if (yy < ymin)
		ymin = yy;
	}
	p2 = p3;
	u = -v;
	du = dv;
    }
    if (closed) {
	// we are dealing with a polygon..
	// same old story for the angles on first and last point of pts[];
	v = (GVector) (p1 - p2);
	dv = v.size();
	for (i = 1; i <= 2; i++) {
	    coeff = fabs(v.x * u.y - v.y * u.x);
	    if (coeff <= (SIN_11_DEG * (du * dv))) {
		// alpha <= 11 deg;
		xmin = min(xmin, p2.x - absWidth);
		ymin = min(ymin, p2.y - absWidth);
		xmax = max(xmax, p2.x + absWidth);
		ymax = max(ymax, p2.y + absWidth);
	    }
	    else {
		coeff = - absWidth / coeff;
		xx = p2.x + coeff * (dv * u.x + du * v.x);
		yy = p2.y + coeff * (dv * u.y + du * v.y);
		if (xx > xmax)
		    xmax = xx;
		else if (xx < xmin)
		    xmin = xx;
		if (yy > ymax)
		    ymax = yy;
		else if (yy < ymin)
		    ymin = yy;
	    }
	    if (i == 1) {
		p2 = p1;
		u  = -v;
		du = dv;
		v  = (GVector) (cs.to(pts[1]) - p1);
		dv = v.size();
	    }
	}
    }
    else {
	// simple polyline; check if last point is out of current box.
	xmin = min(xmin, p2.x - absWidth);
	ymin = min(ymin, p2.y - absWidth);
	xmax = max(xmax, p2.x + absWidth);
	ymax = max(ymax, p2.y + absWidth);
    }

    return GBox(xmin, ymin, xmax, ymax);
}



#define SET_MIN_MAX(m, M, a, b) { \
	if (a < b) { m = a; M = b; } \
	else { m = b; M = a; } \
	}
#define UPDATE_MIN_MAX(m, M, a) { \
    	if (m > a) m = a; \
	else if (M < a) M = a; \
	}

/**
FUNCTION
    GNumber GModulo(GNumber x, GNumber y = 2*PI)
DESCRIPTION
    Returns x modulo y. Error if y == 0.
**/
GNumber GModulo(GNumber x, GNumber y) {
    assert(y != 0.);   // REMOVE ASSERT WHEN SAFE;
    if (y < 0.)
	y = -y;
    if (x < 0.) {
	if (x < -y)
	    x += ((int) (-x / y)) * y;
	return((x == 0.) ? 0. : x + y);
    }
    return((x < y) ? x : x - ((int) (x / y)) * y);
}


/**
FUNCTION
    GOrientedAngles(GNumber amplitude, start, *inf, *sup);
DESCRIPTION
    convert arc format from "start-amplitude" model to "limInf-limSup" model.
**/
void GOrientedAngles(GNumber amplitude, GNumber start,
		     GNumber* inf, GNumber* sup) {
    if (amplitude < 0) {
	*inf = GModulo(start + amplitude);
	*sup = *inf - amplitude;
    }
    else {
	*inf = GModulo(start);
	*sup = *inf + amplitude;
    }
}


/**
FUNCTION
    GNumber GEllipseGetAngle(GNumber radX, GNumber radY, GPoint& pt);
DESCRIPTION
    Let E be the ellipse defined by radx and rady (both positive).
    A ray from the center of E (0, 0) and passing through pt will cross E
    in an unique point f(pt). Each point of E is associated with an unique
    (modulo 2*PI) angle alpha such as
          (x, y) == (radX*cos(alpha), radY*sin(alpha))
    The current function intends to compute this angle alpha associated with
    f(pt)   i.e. pt ---> f(pt) on E ---> alpha

    In other words :
    
    Let g be the bijection from [0, 2*PI[ to E defined as
      g(alpha) == (radX*cos(alpha), radY*sin(alpha))
    The current function intends to compute this angle alpha such as
      pt, g(alpha) and (0, 0) the center of E are aligned.
**/

GNumber GEllipseGetAngle(GNumber radX, GNumber radY, GPoint& pt) {
    assert(radX >= 0. && radY > 0.);
    GNumber ret;
    if (pt.x == 0.)
	ret = (pt.y >= 0.) ? M_PI_2 : -M_PI_2;
    else if (pt.y == 0.)
	ret = (pt.x > 0.) ? 0. : PI;
    else {
	ret = atan((pt.y*radX)/(pt.x*radY));
	if (pt.x < 0.)
	    ret += PI;
	if (ret < 0)
	    ret += 2*PI;
    }
    return ret;
}

/**
FUNCTION
    GPoint GEllipseGetTangentPoint(GNumber radX, radY, const GVector& v)
DESCRIPTION
    Considering the ellipse defined by the radii radX and radY, this function
    returns tangentPt, point of the ellipse,
    such as the tangent in tangentPt is parallel to vector v.
    We consider the upper half of the ellipse (y >= 0). There is another point
    with the same property : -tangentPt.
    The half_ellipse equation is y = f(x) = radY * sqrt(1 - x^2/radX^2)
    Derivate function is f'(x) = - (radY/radX) * x /sqrt(radX^2 - x^2)
    We have to find X such as f'(X) = v.y / v.x;
    Then, tangentPt = (X, f(X)).
**/
GPoint GEllipseGetTangentPoint(GNumber radX, GNumber radY, const GVector& v) {
    if (v.x == 0)
	return GPoint(radX, 0);
    else if (v.y == 0)
	return GPoint(0, radY);
    else {
	GNumber slope = v.y / v.x;
	GNumber constant = sqrt(radY*radY + slope*slope*radX*radX);
	return GPoint(- (slope*radX*radX)/constant, radY*radY / constant);
    }
}


/**
FUNCTION
    void GEllipseGetOutPoints(GNumber radX, radY, angle1, angle2, 
                              GPoint& pmin, GPoint& pmax)
DESCRIPTION
    Considering the elliptical arc defined by :
      o the ellipse radii radX and radY
      o the angles angle1 < angle2, delimiting the arc
    this function returns the arc bounding box in its own coordinate system
    i.e. gets the minimum and maximum values of x and y for points on the arc.
**/
void GEllipseGetOutPoints(GNumber radX, GNumber radY,
			  GNumber angle1, GNumber angle2, 
			  GPoint& pmin, GPoint& pmax) {
    if (angle2 - angle1 >= 2*PI) {
	pmin.x = -radX;
	pmax.x = radX;
	pmin.y = -radY;
	pmax.y = radY;
    }
    else {
	GNumber a, b;
	    
	a = radX * cos(angle1);     // a = arc starting point.x;
	b = radX * cos(angle2);     // b = arc ending point.x;
	SET_MIN_MAX(pmin.x, pmax.x, a, b);
	a = radY * sin(angle1);     // a = arc starting point.y;
	b = radY * sin(angle2);     // b = arc ending point.y;
	SET_MIN_MAX(pmin.y, pmax.y, a, b);

	// check if arc contains PI/2;
	if ((angle1 <= M_PI_2 && M_PI_2 <= angle2) ||
	    (angle1 <= M_PI_2 + 2*PI && M_PI_2 + 2*PI <= angle2)) {
	    pmax.y = radY;
	    UPDATE_MIN_MAX(pmin.x, pmax.x, 0);
	}
	// check if arc contains PI;
	if ((angle1 <= PI && PI <= angle2) ||
	    (angle1 <= 3*PI && 3*PI <= angle2)) {
	    pmin.x = -radX;
	    UPDATE_MIN_MAX(pmin.y, pmax.y, 0);
	}
	// check if arc contains 3PI/2;
	if ((angle1 <= 3*M_PI_2 && 3*M_PI_2 <= angle2) ||
	    (angle1 <= 7*M_PI_2 && 7*M_PI_2 <= angle2)) {
	    pmin.y = -radY;
	    UPDATE_MIN_MAX(pmin.x, pmax.x, 0);
	}
	// check if arc contains 2PI;
	if (angle1 <= 2*PI && 2*PI <= angle2) {
	    pmax.x = radX;
	    UPDATE_MIN_MAX(pmin.y, pmax.y, 0);
	}
    }
}


/**
FUNCTION
    GBox GEllipseGetBox(const GCoordSys& cs, GNumber radX, radY, ang1, ang2)
DESCRIPTION
    
**/
GBox GEllipseGetBox(const GCoordSys& cs, GNumber radX, GNumber radY,
		    GNumber angle1, GNumber angle2) {
    if (cs.isOrthogonal()) {
	GPoint pmin, pmax;
	GEllipseGetOutPoints(radX, radY, angle1, angle2, pmin, pmax);
	return GBox(cs.to(GPoint(pmax.x, pmin.y)),
		    cs.to(GPoint(pmin.x, pmax.y)),
		    cs.to(pmin), cs.to(pmax));
    }

    GMatrix inverted = ~GMatrix(cs.I, cs.J);
    GPoint p1 = GEllipseGetTangentPoint(radX, radY, inverted.vx());
    GPoint p2 = GEllipseGetTangentPoint(radX, radY, inverted.vy());
    // p1 and -p1 (resp p2 and -p2) are point of the ellipse..;
    // with horizontal (resp vertical) tangent.;
    // They are expressed in the relative coordinate system;
    
    if (angle2 - angle1 >= 2*PI)
	return GBox(cs.to(p1), cs.to(GPoint(-p1.x, -p1.y)),
		    cs.to(p2), cs.to(GPoint(-p2.x, -p2.y)));
    
    GNumber angleP1 = p1.angle(); // 0 <= angleP1 <= PI;
    GNumber angleP2 = p2.angle(); // 0 <= angleP2 <= PI;
    GPoint pmin, pmax, pstart, pend, pcur;
    
    // pstart = arc starting point;
    pstart = cs.to(GPoint(radX * cos(angle1), radY * sin(angle1)));
    
    // pend = arc ending point;
    pend = cs.to(GPoint(radX * cos(angle2), radY * sin(angle2)));
    
    SET_MIN_MAX(pmin.x, pmax.x, pstart.x, pend.x);
    SET_MIN_MAX(pmin.y, pmax.y, pstart.y, pend.y);
    
    // check if p1 is in arc;
    if ((angle1 <= angleP1 && angleP1 <= angle2) ||
	(angle1 <= angleP1 + 2*PI && angleP1 + 2*PI <= angle2)) {
	pcur = cs.to(p1);
	UPDATE_MIN_MAX(pmin.x, pmax.x, pcur.x);
	UPDATE_MIN_MAX(pmin.y, pmax.y, pcur.y);
    }
    
    // check if -p1 is in arc;
    if ((angle1 <= angleP1 + PI && angleP1 + PI <= angle2) ||
	(angle1 <= angleP1 + 3*PI && angleP1 + 3*PI <= angle2)) {
	pcur = cs.to(GPoint(-p1.x, -p1.y));
	UPDATE_MIN_MAX(pmin.x, pmax.x, pcur.x);
	UPDATE_MIN_MAX(pmin.y, pmax.y, pcur.y);
    }
    
    // check if p2 is in arc;
    if ((angle1 <= angleP2 && angleP2 <= angle2) ||
	(angle1 <= angleP2 + 2*PI && angleP2 + 2*PI <= angle2)) {
	pcur = cs.to(p2);
	UPDATE_MIN_MAX(pmin.x, pmax.x, pcur.x);
	UPDATE_MIN_MAX(pmin.y, pmax.y, pcur.y);
    }
    
    // check if -p2 is in arc;
    if ((angle1 <= angleP2 + PI && angleP2 + PI <= angle2) ||
	(angle1 <= angleP2 + 3*PI && angleP2 + 3*PI <= angle2)) {
	pcur = cs.to(GPoint(-p2.x, -p2.y));
	UPDATE_MIN_MAX(pmin.x, pmax.x, pcur.x);
	UPDATE_MIN_MAX(pmin.y, pmax.y, pcur.y);
    }
    return GBox(pmin.x, pmin.y, pmax.x, pmax.y);
}


int GEllipseNbLines(const GCoordSys& cs, GNumber radiusX, GNumber radiusY,
		    GNumber &angle1, GNumber &angle2, GQuality) {
    // WATCH OUT for side effects on angle1 and angle2 !! ;
    int nb_line;
    
    // Adjust angles;
    if (angle1 == 0.0 && angle2 == 0.0)
	angle2 = 2*PI;
    while (angle2 < angle1) angle2 += 2*PI;

    // Compute npoint;
    GNumber R = (cs.I.size() * radiusX + cs.J.size() * radiusY) / 2;
    if (R <= 4)
	nb_line = 1 << (int)(R + 1);
    else {
	nb_line = 32 + (int) (R / 10);
	if (nb_line > GMAX_ELLIPSE_POINTS - 1)
	    nb_line = GMAX_ELLIPSE_POINTS - 1;
    }
    return(1 + (int) (nb_line * (angle2 - angle1) / (2*PI)));
}

/**
FUNCTION
    GEllipseComputePoints(const GCoordSys& cs,
                          GNumber radiusX, radiusY, angle1, angle2,
			  GQuality quality, int& npoint, GPoint* points)
DESCRIPTION
    "points" must point to at least GMAX_ELLIPSE_POINTS GPoint.
    Returns npoint and points, a "polyline approximation" of the elliptical
    arc defined by radiusX, radiusY, angle1 and angle2,
    
**/

void GEllipseComputePoints(const GCoordSys& cs, GNumber radiusX,
			   GNumber radiusY, GNumber angle1, GNumber angle2,
			   GQuality quality, int& npoint, GPoint* points) {
    // Compute the number of points needed;

    // we cannot use references or addresses of parameter of type GNumber;
    // This is due to the way parameters are stored in old fashion..;
    // C compilers;
    GNumber a  = angle1;
    GNumber da = angle2;
    
    // WATCH OUT for side effects on a and da !! ;
    npoint = GEllipseNbLines(cs, radiusX, radiusY, a, da, quality);
    // actually, the real number of points we are looking for is npoint+1;
    
    // Compute last point;
    points[npoint].x = radiusX * cos(da);
    points[npoint].y = radiusY * sin(da);

    // Compute other points;
    da = (da - a) / npoint;
    for (int i = 0; i < npoint; i++, a += da) {
	points[i].x = radiusX * cos(a);
	points[i].y = radiusY * sin(a);
    }
    npoint++;
}

int GEllipseArcLimits(const GCoordSys& cs, GNumber radX, GNumber radY,
		      GNumber angle1, GNumber angle2, GPoint &start,
		      GPoint &end, GPoint &near_start, GPoint &near_end,
		      GQuality quality) {
    
    // check if we need the 2 extra points near_start and near_end;

    // we cannot use references or addresses of parameter of type GNumber;
    // This is due to the way parameters are stored in old fashion..;
    // C compilers;
    GNumber ang1 = angle1;
    GNumber ang2 = angle2;

    // WATCH OUT for side effects on ang1 and ang2 !! ;
    int nb_point = GEllipseNbLines(cs, radX, radY, ang1, ang2, quality);

    // set chord segment ;
    start.x = radX * cos(ang1);
    start.y = radY * sin(ang1);
    end.x   = radX * cos(ang2);
    end.y   = radY * sin(ang2);
    
    if (nb_point >= 4) {
	GNumber delta = (ang2 - ang1) / nb_point;
	near_start.x = radX * cos(ang1 + delta);
	near_start.y = radY * sin(ang1 + delta);
	near_end.x   = radX * cos(ang2 - delta);
	near_end.y   = radY * sin(ang2 - delta);
	return nb_point;
    }
    else return 0;
}


static GNumber binomCoeff(int n, int i) {
    static int BCTable [] = {
	1, 
	1,   1, 
	1,   2,   1,
	1,   3,   3,   1,
	1,   4,   6,   4,   1,
	1,   5,  10,  10,   5,   1,
	1,   6,  15,  20,  15,   6,   1,
	1,   7,  21,  35,  35,  21,   7,   1,
	1,   8,  28,  56,  70,  56,  28,   8,   1,
	1,   9,  36,  84, 126, 126,  84,  36,   9,   1};

    if ( i >n)
	return 0;

    if ( n < 10)
	return(BCTable[n* (n+1)/2 + i]);

    GNumber bc = 1.0;
    int j;
    for (j = n -i +1; j <= n; j++)
	bc = bc * j;
    for (j = 1; j <= i; j++)
	bc =  bc / j;
    return bc;
}
	
static GNumber bezierBlend(int i,int n,GNumber u) {
    double b = binomCoeff(n,i);
    int j;
    for (j = 0; j < i; j++)
        b = b * u;

    for (j = 0; j < (n-i); j++)
  	b = b * (1 -u);
  
    if ( b <= (GNumber) 1e-30)
	return 0;
    if ( b >= (GNumber) 1e+30)
	return 1;
    return (b);
}


/**
FUNCTION
    GBezierComputePoints().
DESCRIPTION
    A bezier curve of  n-th order is characterised by n+1 control points.
    In general the curve does not pass through all control points,
    but it passses through the two end points, and the is tangent at the end 
    points to the corresponding edge of the polygon of the control points.
    The curve will always lie within the convex hull of the control points.
    The curve does not provide localized control: moving a control point will
    change the shape of every part of the curve.
    "points" must points to at least MaxBezierGPoints(nctrl) GPoint.
**/
void GBezierComputePoints(int nctrl, const GPoint* ctrl, GQuality quality,
			  int& npoint, GPoint* points) {
    int interval = (GMAX_BEZIER_INTERVAL / GN_QUALITY) * (quality + 1);
    GNumber u = 0;
    GPoint p;
    npoint = (nctrl * interval) + 1;
    GNumber du = 1.0 / (GNumber) (npoint - 1);
    for (int i = 0; i < npoint; i++, u += du)
    {
	p.x = 0; p.y = 0;
	for (int j = 0; j < nctrl; j++)
	    p += ctrl[j] * bezierBlend (j, nctrl-1, u);
        points[i] = p;
    }
}



/**
FUNCTION
    GCubicCurveComputePoints().
DESCRIPTION
    "points" must points to at least MaxCubicGPoints(nctrl,evalshift) GPoint.
**/
void GCubicCurveComputePoints(int nctrl, const GPoint* ctrl,
			      const GNumber (*basisMatrix)[4] /* [4][4] */,
			      int evalShift, GQuality quality,
			      int& npoint, GPoint* points) {
    int numseg = (nctrl - 4 + evalShift) / evalShift;
    if (numseg <= 0) {
        npoint = 0;
	return;
    }

    int interval = (GMAX_CUBIC_INTERVAL / GN_QUALITY) * (quality + 1);
    npoint = interval * numseg + 1;

    int k = 0;
    GNumber dt  = 1.0 / interval;
    GNumber dt2 = dt * dt;
    GNumber dt3 = dt2 * dt;

    GNumber dfx = 0, dfy = 0;
    GNumber df2x, df2y, df3x, df3y;

    int si = 0;
    for (int sn = 0; sn < numseg; sn++, si += evalShift) {
	GNumber ax = 0, ay = 0, bx = 0, by = 0;
	GNumber cx = 0, cy = 0, dx = 0, dy = 0;
	
	for (int i = 0, ci = si; i < 4; i++,ci++) {
	    ax += (ctrl[ci].x * basisMatrix[0][i]);
	    ay += (ctrl[ci].y * basisMatrix[0][i]);
	    bx += (ctrl[ci].x * basisMatrix[1][i]);
	    by += (ctrl[ci].y * basisMatrix[1][i]);
	    cx += (ctrl[ci].x * basisMatrix[2][i]);
	    cy += (ctrl[ci].y * basisMatrix[2][i]);
	    dx += (ctrl[ci].x * basisMatrix[3][i]);
	    dy += (ctrl[ci].y * basisMatrix[3][i]);
	}

	points[k].x = dx;
	points[k].y = dy;
	k++;
	dfx  = ax * dt3 + bx * dt2 + cx * dt;
        dfy  = ay * dt3 + by * dt2 + cy * dt;
	df3x = 6 * ax * dt3;
	df3y = 6 * ay * dt3;
	df2x = df3x + 2 * bx * dt2;
	df2y = df3y + 2 * by * dt2;

	for (i = 1; i < interval; i++, k++) {
	    points[k].x = points[k-1].x + dfx;
	    points[k].y = points[k-1].y + dfy;
	    dfx  += df2x;
     	    dfy  += df2y;
	    df2x += df3x;
	    df2y += df3y;
	}
    }
    points[k].x = points[k-1].x + dfx;
    points[k].y = points[k-1].y + dfy;
    k++;
}
