#ifdef REV_INFO
static char rcsid[] = "$Id: Pattern.cc,v 1.2 1992/09/25 14:50:22 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Tools
FILE:    Pattern.cc
AUTHOR:  Jacques Davy
DATE:    November 1990

DESCRIPTION:
    (see <Go/Tools/Pattern.h>

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/Tools/Pattern.h>
#include <assert.h>

/*
 * GCallback types
 */
static char patternNmodifyPrecond[1];
static char patternNmodifyPostcond[1];
static char patternNdestroyPostcond[1];

typedef GBool (*PeProc)(void*, const GPattern*, ulong&, GPatternValues*);
typedef void (*PsProc)(void*, GPattern*, ulong);
typedef void (*PdProc)(void*, const GPattern*);

struct PrecondRec
{
    PeProc proc;
    void* closure;
};

struct ModifPostcondRec
{
    PsProc proc;
    void* closure;
};

struct DestroyPostcondRec
{
    PdProc proc;
    void* closure;
};

/*
 * Private methods
 * ---------------
 */
/**
FUNCTION
    _call_call_modify_postcond
DESCRIPTION
    This function is called when the GShPattern is updated
**/
void _call_call_modify_postcond (void* This, GShPattern* sh_Pattern,
				 ulong update_mask)
{
    Call_call_modify_postcond ((GPattern*)This, sh_Pattern, update_mask);
}

void Call_call_modify_postcond (GPattern* This,
				const GShPattern* sh_Pattern,
				ulong update_mask)
{
    assert (sh_Pattern == This->shPattern());
    This->_call_modify_postcond (update_mask);
}


/*
 * Protected methods
 * -----------------
 */
/**/
int GPattern::_addCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = (GAssoc*)&_assoc;
    GAssoc* cb = (GAssoc*)_cb;
    
    GCallbackRec cbrec (proc, closure);
    return cb->add(type, sizeof cbrec, (char*)&cbrec);
}

/**/
int GPattern::_remCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = (GAssoc*)&_assoc;
    GAssoc* cb = (GAssoc*)_cb;

    GCallbackRec cbrec (proc, closure);
    return cb->rem(type, sizeof cbrec, (char*)&cbrec);
}

/**/
GBool GPattern::_call_modify_precond (ulong& modify_mask, GPatternValues* new_values)
{
    if (_flags._has_modify_precond)
    {
	int count;
	PrecondRec* cbrec = (PrecondRec*)_assoc.get (patternNmodifyPrecond, count);
	for (int i = 0; i < count; i++, cbrec++)
	    if (!cbrec->proc(cbrec->closure, this, modify_mask, new_values))
		return GFALSE;
    }
    return GTRUE;
}

/**/
void GPattern::_call_modify_postcond (ulong modify_mask)
{
    if (_flags._has_modify_postcond)
    {
	int count;
	GCallbackRec* cbrec = (GCallbackRec*)_assoc.get (patternNmodifyPostcond, count);
	(void)GCallCallbacks(cbrec, count, this, modify_mask);
    }
}

/**/
void GPattern::_call_destroy_postcond ()
{
    if (_flags._has_destroy_postcond)
    {
	int count;
	DestroyPostcondRec* cbrec = (DestroyPostcondRec*)_assoc.get (patternNdestroyPostcond, count);
	for (int i = 0; i < count; i++, cbrec++)
	     cbrec->proc(cbrec->closure, this);
    }
}

/*
 * Public methods
 * --------------
 */
GPattern::GPattern ()
{
    _sh_pattern = GCreateShPattern(); /* the default one */
}

GPattern::GPattern (const GPattern* pattern)
{
    _sh_pattern = pattern->_sh_pattern->dup();
}

GPattern::GPattern (const GShPattern* sh_pattern)
{
    _sh_pattern = sh_pattern->dup();
}

GPattern::GPattern (ulong mask, const GPatternValues* values, const char* name)
{
    _sh_pattern = GCreateShPattern (mask, values, name);
}

GPattern::GPattern (const char* name)
{
    _sh_pattern = GCreateShPattern (name);
}

GPattern::~GPattern ()
{
    _call_destroy_postcond ();
    if (_flags._has_modify_postcond)
        _sh_pattern->updatePostcond (_call_call_modify_postcond,
				     this, GREMOVE);
    _sh_pattern->destroy();
}

/**
FUNCTION
    GPattern::modify
DESCRIPTION
    Modify the shared pattern and call pre/post condition.
**/

void GPattern::modify (ulong mask, const GPatternValues* values)
{
    GPatternValues Values = *values ;
    if (mask && _call_modify_precond (mask, &Values))
    {
	// Remove postcond;
	if (_flags._has_modify_postcond)
	    _sh_pattern->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

	// Modify;
        ::GModify(_sh_pattern, mask, &Values);

	// Call postcond;
	_call_modify_postcond (mask);

	// Add postcond;
	if (_flags._has_modify_postcond)
	    _sh_pattern->updatePostcond (_call_call_modify_postcond, this, GADD);
    }
}

void GPattern::modify (const char* name)
{
    if (name && _sh_pattern->name() && (strcmp (name, _sh_pattern->name()) == 0))
        return;
    
    // Remove postcond;
    if (_flags._has_modify_postcond)
        _sh_pattern->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

    // Get mask and values
    GShPattern* save = _sh_pattern->dup();
    GRename (save, name);
    GPatternValues Values = *save->values();
    ulong mask = save->mask();

    if (mask && _call_modify_precond (mask, &Values))
    {
         _sh_pattern->destroy();
         _sh_pattern = GCreateShPattern (mask, &Values, name);
         _call_modify_postcond (mask);
    }
    save->destroy();

    if (_flags._has_modify_postcond)
	_sh_pattern->updatePostcond (_call_call_modify_postcond, this, GADD);
}

void GPattern::modify (const GShPattern* sh_pattern)
{
    if (sh_pattern == _sh_pattern)
        return;

    // top of list;
    sh_pattern->raise(GShPattern::_headOfList);
    
    if (sh_pattern->isNamed())
	modify(sh_pattern->name());
    else {

	GPatternValues Values = *sh_pattern->values();
	ulong mask = sh_pattern->mask();
	    
	// Remove postcond;
	if (_flags._has_modify_postcond)
	    _sh_pattern->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

	// Rename to NULL;
	GRename(_sh_pattern, NULL);
		
	if (mask && _call_modify_precond (mask, &Values)) {

	    // Modify;
	    ::GModify (_sh_pattern, mask, &Values);

	    // Call postcond;
	    _call_modify_postcond (mask);
	}

	// Add postcond;
	if (_flags._has_modify_postcond)
	    _sh_pattern->updatePostcond (_call_call_modify_postcond, this, GADD);
    }
}


/**
FUNCTION
    GPattern::field_modification_methods
DESCRIPTION
    Each method modify one parameter of the GPattern.
**/

#define _modifyPatternValue(bit_mask, field, value) \
    GPatternValues values;\
    values.field = value;\
    modify (bit_mask, &values)

#undef _modifyPatternValue

/**
FUNCTION
    GPattern::add/rem pre/postconditions
DESCRIPTION
    Add or remove pre/post condition on modification and on destruction.

    GWarning: _call_call_modify_postcond
    is stored directly in the shared object and pre/condition
    may be not called!
**/

#define _add_rem_callback(type, flag) \
    (FLAGS->flag =\
         0 < ((add == GADD)\
	      ? _addCallback(type, (GCBProc)proc, closure)\
	      : _remCallback(type, (GCBProc)proc, closure)))

/**/
void GPattern::modifyPrecond (PeProc proc,
			     const void* closure, GAddOrRemove add) const
{
    const GPatternCondFlags* flags = &_flags;
    GPatternCondFlags* FLAGS = (GPatternCondFlags*)flags;

   _add_rem_callback(patternNmodifyPrecond, _has_modify_precond);
}

/**/
void GPattern::modifyPostcond (PsProc proc,
			      const void* closure, GAddOrRemove add) const
{
    ushort flag = _flags._has_modify_postcond;

    const GPatternCondFlags* flags = &_flags;
    GPatternCondFlags* FLAGS = (GPatternCondFlags*)flags;

    _add_rem_callback(patternNmodifyPostcond, _has_modify_postcond);

    // The modifyPostcond is called from the sh_pattern;

    if (!flag && _flags._has_modify_postcond)
	_sh_pattern->updatePostcond (_call_call_modify_postcond, this, GADD);
    else
	if (flag && !_flags._has_modify_postcond)
	    _sh_pattern->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

}

/**/
void GPattern::destroyPostcond (PdProc proc,
			       const void* closure, GAddOrRemove add) const
{
    const GPatternCondFlags* flags = &_flags;
    GPatternCondFlags* FLAGS = (GPatternCondFlags*)flags;

   _add_rem_callback(patternNdestroyPostcond, _has_destroy_postcond);
}

#undef _add_rem_callback
