#ifdef REV_INFO
static char rcsid[] = "$Id: PxlArray.cc,v 1.2 1992/09/25 14:52:16 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Tools
FILE:    PixelArray.cc
AUTHOR:  Jacques Davy
DATE:    November 1990

DESCRIPTION:
    (see <Go/Tools/PixelArray.h>

    It is provided a global function "GeneralImageDeformation"
    to deform a pixel array.

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/Tools/PixelArray.h>
#include <assert.h>

/*
 * GCallback types
 */
static char pxlArrayNmodifyPrecond[1];
static char pxlArrayNmodifyPostcond[1];
static char pxlArrayNdestroyPostcond[1];

typedef GBool (*PeProc)(void*, const GPixelArray*, ulong&, GPixelArrayValues*);
typedef void (*PsProc)(void*, GPixelArray*, ulong);
typedef void (*PdProc)(void*, const GPixelArray*);
typedef void (*PshProc)(void*, GShPixelArray*, ulong);

struct PrecondRec
{
    PeProc proc;
    void* closure;
};

struct ModifPostcondRec
{
    PsProc proc;
    void* closure;
};

struct DestroyPostcondRec
{
    PdProc proc;
    void* closure;
};

/*
 * Private methods
 * ---------------
 */
/**
FUNCTION
    _call_call_modify_postcond
DESCRIPTION
    This function is called when the NAMED GShPixelArray is updated
**/
void _call_call_modify_postcond (void* This, GShPixelArray* sh_PixelArray,
				 ulong update_mask)
{
    Call_call_modify_postcond ((GPixelArray*)This, sh_PixelArray, update_mask);
}

void Call_call_modify_postcond (GPixelArray* This,
				const GShPixelArray* sh_PixelArray,
				ulong update_mask)
{
    assert (sh_PixelArray == This->shPixelArray());
    This->_call_modify_postcond (update_mask);
}

/*
 * Protected methods
 * -----------------
 */
/**/
int GPixelArray::_addCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = &_assoc;
    GAssoc* cb = (GAssoc*)_cb;
    
    GCallbackRec cbrec (proc, closure);
    return cb->add(type, sizeof cbrec, (char*)&cbrec);
}

/**/
int GPixelArray::_remCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = &_assoc;
    GAssoc* cb = (GAssoc*)_cb;


    GCallbackRec cbrec (proc, closure);
    return cb->rem(type, sizeof cbrec, (char*)&cbrec);
}

/**/
GBool GPixelArray::_call_modify_precond (ulong& modify_mask, GPixelArrayValues* new_values)
{
    if (_flags._has_modify_precond)
    {
	int count;
	PrecondRec* cbrec = (PrecondRec*)_assoc.get (pxlArrayNmodifyPrecond, count);
	for (int i = 0; i < count; i++, cbrec++)
	    if (!cbrec->proc(cbrec->closure, this, modify_mask, new_values))
                return GFALSE;
    }
    return GTRUE;
}

/**/
void GPixelArray::_call_modify_postcond (ulong modify_mask)
{
    if (_flags._has_modify_postcond)
    {
	int count;
	GCallbackRec* cbrec = (GCallbackRec*)_assoc.get (pxlArrayNmodifyPostcond, count);
	(void)GCallCallbacks(cbrec, count, this, modify_mask);
    }
}

/**/
void GPixelArray::_call_destroy_postcond ()
{
    if (_flags._has_destroy_postcond)
    {
	int count;
	DestroyPostcondRec* cbrec = (DestroyPostcondRec*)_assoc.get (pxlArrayNdestroyPostcond, count);
	for (int i = 0; i < count; i++, cbrec++)
	     cbrec->proc(cbrec->closure, this);
    }
}

/*
 * Public methods
 * --------------
 */
GPixelArray::GPixelArray ()
{
    _sh_pixelArray = GCreateShPixelArray();
}

GPixelArray::GPixelArray (const GPixelArray* pixelArray)
{
    _sh_pixelArray = pixelArray->_sh_pixelArray->dup();
}

GPixelArray::GPixelArray (const GShPixelArray* sh_pixelArray)
{
    _sh_pixelArray = sh_pixelArray->dup();
}

GPixelArray::GPixelArray (ulong mask, const GPixelArrayValues* values, const char* name,
			  GDataSpace space)
{
    _sh_pixelArray = GCreateShPixelArray (mask, values, name, space);
}

GPixelArray::GPixelArray (const char* name)
{
    _sh_pixelArray = GCreateShPixelArray (name);
}

GPixelArray::~GPixelArray ()
{
    _call_destroy_postcond ();
    if (_sh_pixelArray->isNamed() && _flags._has_modify_postcond)
        _sh_pixelArray->updatePostcond (_call_call_modify_postcond, this, GREMOVE);
    _sh_pixelArray->destroy();
}

/**
FUNCTION
    GPixelArray::modify
DESCRIPTION
    Modify the shared pixelArray and call pre/post condition.
**/

void GPixelArray::modify (ulong mask, const GPixelArrayValues* values)
{
    GPixelArrayValues Values = *values;
    if (mask && _call_modify_precond (mask, &Values))
    {
	// Remove postcond;
	if (_sh_pixelArray->isNamed() && _flags._has_modify_postcond)
	    _sh_pixelArray->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

	// Modify;
        GModify (_sh_pixelArray, mask, &Values);

	// Call postcond;
	_call_modify_postcond (mask);

	// Add postcond;
	if (_sh_pixelArray->isNamed() && _flags._has_modify_postcond)
	    _sh_pixelArray->updatePostcond (_call_call_modify_postcond, this, GADD);

    }
}

void GPixelArray::modify (const char* name)
{
    if (name && _sh_pixelArray->isNamed() && (strcmp (name, _sh_pixelArray->name()) == 0))
        return;


    // Remove postcond if the pixelArray is already named.
    if (_sh_pixelArray->name() && _flags._has_modify_postcond)
        _sh_pixelArray->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

    // Get mask and values
    GShPixelArray* save = _sh_pixelArray->dup();

    GRename (save, name);

    GPixelArrayValues Values = *save->values();
    ulong mask = save->mask();

    if (mask && _call_modify_precond (mask, &Values))
    {
         _sh_pixelArray->destroy();
         _sh_pixelArray = GCreateShPixelArray (mask, &Values, name);
         _call_modify_postcond (mask);
    }
    save->destroy();

    // Add postcond if the pixelArray is named.
    if (name && _flags._has_modify_postcond)
	_sh_pixelArray->updatePostcond (_call_call_modify_postcond, this, GADD);
}


void GPixelArray::modify(const GShPixelArray* sh_pixelArray)
{
    if (_sh_pixelArray  == sh_pixelArray)
	return;

    // top of list.
    sh_pixelArray->raise(GShPixelArray::_headOfList);
    
    if (sh_pixelArray->isNamed())
	modify(sh_pixelArray->name());
    else {

	GPixelArrayValues Values ;
	Values  = *sh_pixelArray->values();
	ulong mask = sh_pixelArray->mask();
	    

	// Remove postcond if the pixelArray is already named;
	if (_sh_pixelArray->name() && _flags._has_modify_postcond)
	    _sh_pixelArray->updatePostcond (_call_call_modify_postcond, this, GREMOVE);

    	// Rename to NULL;
	GRename(_sh_pixelArray, NULL);
		
	if (mask && _call_modify_precond (mask, &Values)) {

	    // Modify;
	    GModify (_sh_pixelArray, mask, &Values);

	    // Call postcond;
	    _call_modify_postcond (mask);
	}


	// Add postcond;
	if (_sh_pixelArray->isNamed() && _flags._has_modify_postcond)
	    _sh_pixelArray->updatePostcond (_call_call_modify_postcond, this, GADD);
    }
}

/**
FUNCTION
    GPixelArray::field_modification_methods
DESCRIPTION
    Each method modify one parameter of the GPixelArray.
**/

#define _modifyPixelArrayValue(bit_mask, field, value) \
    GPixelArrayValues values;\
    values.field = value;\
    modify (bit_mask, &values)

#undef _modifyPixelArrayValue

/**
FUNCTION
    GPixelArray::add/rem pre/postconditions
DESCRIPTION
    Add or remove pre/post condition on modification and on destruction.

    GWarning, if the GShPixelArray is named, _call_call_modify_postcond
    is stored directly in the shared object and pre/condition
    may be not called!
**/

#define _add_rem_callback(type, flag) \
    (FLAGS->flag =\
         0 < ((add == GADD)\
	      ? _addCallback(type, (GCBProc)cb, closure)\
	      : _remCallback(type, (GCBProc)cb, closure)))

/**/
void GPixelArray::modifyPrecond (PeProc cb, const void* closure, GAddOrRemove add) const
{
    const GPxlArrayCondFlags* flags = &_flags;
    GPxlArrayCondFlags* FLAGS = (GPxlArrayCondFlags*)flags;

    _add_rem_callback(pxlArrayNmodifyPrecond, _has_modify_precond);
}

/**/
void GPixelArray::modifyPostcond (PsProc cb, const void* closure, GAddOrRemove add) const
{
    ushort flag = _flags._has_modify_postcond;

    const GPxlArrayCondFlags* flags = &_flags;
    GPxlArrayCondFlags* FLAGS = (GPxlArrayCondFlags*)flags;
    
    _add_rem_callback(pxlArrayNmodifyPostcond, _has_modify_postcond);

   // If the pixelArray is named, the modifyPostcond is called from the 
   // sh_pixelArray
   if (_sh_pixelArray->isNamed())
   {
       if (!flag && _flags._has_modify_postcond)
           _sh_pixelArray->updatePostcond (_call_call_modify_postcond, this, GADD);
       else
           if (flag && !_flags._has_modify_postcond)
               _sh_pixelArray->updatePostcond (_call_call_modify_postcond, this, GREMOVE);
   }
}

/**/
void GPixelArray::destroyPostcond (PdProc cb,
				  const void* closure, GAddOrRemove add) const
{
    const GPxlArrayCondFlags* flags = &_flags;
    GPxlArrayCondFlags* FLAGS = (GPxlArrayCondFlags*)flags;

   _add_rem_callback(pxlArrayNdestroyPostcond, _has_destroy_postcond);
}

#undef _add_rem_callback







