#ifdef REV_INFO
static char rcsid[] = "$Id: Region.cc,v 1.7 1993/01/04 15:34:48 davy Exp $";
#endif
/*****************************************************************************
MODULE:  Go Tools
FILE:    Region.cc
AUTHOR:  Oliver Jojic
DATE:    July 90

DESCRIPTION:
    See Go/Tools/Region.h

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/
#include <Go/machdef.h>
#include <Go/Tools/Region.h>
#include <Go/Tools/Error.h>
#include <Go/Tools/Alloc.h>


#ifndef MINSHORT
#define MINSHORT -MAXSHORT
#endif


const GRegion& GEmptyRegion() {
    static GRegion* empty_region = NULL;

    if (! empty_region)
	empty_region = new GRegion;

    return *empty_region;
}

const GRegion& GBiggestRegion() {
    static GRegion* biggest_region = NULL;

    if (! biggest_region)
	biggest_region = new GRegion(GBiggestBox());

    return *biggest_region;
}


void GRegion::initialize (int initial_size)
{
    count = 0;
    size = initial_size;
    big = GEmptyBox();
    boxes = (size == 0) ? NULL : (GBox*)GMALLOC(size * sizeof(GBox));
}

GRegion::GRegion ()
{
    initialize(0);
}

GRegion::GRegion (int initial_size)
{
    initialize(initial_size);
}


GRegion::GRegion (const GBox& box)
{
    initialize(1);
    *this = box;
}


GRegion::GRegion (const GRegion& region)
{
    initialize(region.count);
    *this = region;
}

void GRegion::recompute_bigbox ()
{
    if (count == 0)
    {
	big = GEmptyBox();
	return;
    }

    big.ymin = boxes[0].ymin;
    big.ymax = boxes[count - 1].ymax;

    big.xmin = boxes[0].xmin;
    big.xmax = boxes[0].xmax;
    for (int i = 1; i < count; i++)
    {
	if (big.xmin > boxes[i].xmin)
	    big.xmin = boxes[i].xmin;
	if (big.xmax < boxes[i].xmax)
	    big.xmax = boxes[i].xmax;
    }
}

void GRegion::operator = (const GRegion& r)
{
    if (this == &r)
	return;

    if (size < r.count)
    {
	size = r.count;
	GFREE(boxes);
	boxes = (GBox*)GMALLOC(sizeof(GBox)*size);
    }

    big = r.big;
    count = r.count;
    bcopy(r.boxes, boxes, sizeof(GBox)*count);
}


void GRegion::operator = (const GBox& b)
{
    if (b.empty())
    {
	count = 0;
	big = GEmptyBox();
	return;
    }
    big = b;
    count = 1;

    if (size == 0)
    {
	size = 1;
	boxes = (GBox*)GMALLOC(sizeof(GBox));
    }
    boxes[0] = b;
}

GBool operator == (const GRegion& a, const GRegion& b)
{
    if (a.count != b.count || a.big != b.big)
	return GFALSE;

    for (int i = 0; i < a.count; i++)
	if (!GEqual(a.boxes[i], b.boxes[i]))
	    return GFALSE;

    return GTRUE;
}


static GBool intersectP
    (const GBox* pA, const GBox* pA_end,
     const GBox* pB, const GBox* pB_end)
{
    for (;;)
    {
	if (pA->ymax < pB->ymin)
	{
	    pA++;
	    if (pA == pA_end) return GFALSE;
	}
	else if (pB->ymax < pA->ymin)
	{
	    pB++;
	    if (pB == pB_end) return GFALSE;
	}
	else 			/* pA and pB intersect on y */
	{
	    const GBox* save_pA = pA;
	    const GBox* save_pB = pB;

	    for (;;)
	    {
		if (pA->xmax < pB->xmin)
		{
		    pA++;
		    if (pA >= pA_end || pA->ymin != save_pA->ymin)
			break;
		}
		else if (pB->xmax < pA->xmin)
		{
		    pB++;
		    if (pB >= pB_end || pB->ymin != save_pB->ymin)
			break;
		}
		else
		    return GTRUE;
	    }

	    if (save_pA->ymax > save_pB->ymax)
	    {
		pA = save_pA;
		for (;;)
		{
		    if (pB == pB_end) return GFALSE;
		    if (pB->ymin != save_pB->ymin) break;
		    pB++;
		}
	    }
	    else if (save_pB->ymax > save_pA->ymax)
	    {
		pB = save_pB;
		for (;;)
		{
		    if (pA == pA_end) return GFALSE;
		    if (pA->ymin != save_pA->ymin) break;
		    pA++;
		}
	    }
	    else /* save_pA->ymax == save_pB->ymax */
	    {
		for (;;)
		{
		    if (pA == pA_end) return GFALSE;
		    if (pA->ymin != save_pA->ymin) break;
		    pA++;
		}
		for (;;)
		{
		    if (pB == pB_end) return GFALSE;
		    if (pB->ymin != save_pB->ymin) break;
		    pB++;
		}
	    }
	}
    }
}
	    

GBool operator && (const GRegion& a, const GRegion& b)
{
    if (!(a.big && b.big))
	return GFALSE;

    return intersectP(a.boxes, a.boxes + a.count,
		      b.boxes, b.boxes + b.count);
}


GBool GRegion::contains (const GPoint& p, GNumber maxdist) const
{
    if (maxdist == 0)
	maxdist = .01; /* BUG: what else ? */
    
    return *this && GBox(p.x - maxdist, p.y - maxdist,
			p.x + maxdist, p.y + maxdist);
}


/* this macro is used both by common_GAdd and by common_GSub */

#define ADD_TO_PR(xm, ym, xM, yM)\
{\
    if (pR == pR_end)\
	GError("Region.cc: too small computed space");\
    pR->xmin = (xm); pR->ymin = (ym); pR->xmax = (xM); pR->ymax = (yM);\
    pR++;\
}


#ifdef NO_ALLOCA

#if defined(__BORLANDC__) || defined(__MSC7__) 
static const int tempDataSize = 1 * 1024;
#else
static const int tempDataSize = 16 * 1024;
#endif

static GBox tempData[tempDataSize];
#endif

void common_GAdd
    (const GBox* pA, const GBox* pA_end,
     const GBox* pB, const GBox* pB_end,
     GRegion& R)
{
    GNumber ymin, ymax, pBymin, pAymin;
    GBox *start_pR, *pR, *pR_end, *prev_pR, *middle_pR;
    const GBox *save_pA, *save_pB;
    GBool pA_is_done_but_not_pB, concat;
    int biggest_size;
    
    /* to be shure not to realloc too often or not to alloc a too big
       memory for a small region, we use temporary R boxes on stack.
       (Of course this is not true for poor people defining NO_ALLOCA).
       I am not sure that variable declaration works fine on all machines
       after an alloca call. This is why I put all them above. */

#if defined(__BORLANDC__) || defined(__MSC7__) 
    biggest_size = 1 * 1024; /*BUG: difficult to compute*/
#else
    biggest_size = 16 * 1024; /*BUG: difficult to compute*/
#endif

#ifndef NO_ALLOCA
    start_pR = (GBox*)alloca(biggest_size * sizeof(GBox));
#else /* sorry for you, poor people ... */
    if (biggest_size > tempDataSize)
	GError("Region.cc: too small temp data size");
    start_pR = tempData;
#endif
    pR = start_pR;
    pR_end = start_pR + biggest_size;

    ymin = min(pA->ymin, pB->ymin);

    prev_pR = NULL;
    middle_pR = pR;

    for (;;) 			/* ymin is already computed here */
    {
	if (pB->ymin > ymin) 		/* then pA->ymin <= ymin */
	{
	    save_pA = pA;
	    ymax = min(pA->ymax, pB->ymin);

	    do {
		ADD_TO_PR(pA->xmin, ymin, pA->xmax, ymax)
		pA++;
	    } while (pA < pA_end && pA->ymin == save_pA->ymin);

	    if (save_pA->ymax > ymax)
	    	{ pA = save_pA; ymin = ymax; }
	    else if (pA == pA_end)
	    	ymin = pB->ymin;
	    else 
		ymin = min(pA->ymin, pB->ymin);
	}
	else if (pA->ymin > ymin) 	/* then pB->ymin <= ymin */
	{
	    save_pB = pB;
	    ymax = min(pB->ymax, pA->ymin);

	    do {
		ADD_TO_PR(pB->xmin, ymin, pB->xmax, ymax)
		pB++;
	    } while (pB < pB_end && pB->ymin == save_pB->ymin);

	    if (save_pB->ymax > ymax)
		{ pB = save_pB; ymin = ymax; }
	    else if (pB == pB_end)
		ymin = pA->ymin;
	    else
		ymin = min(pA->ymin, pB->ymin);
	}
	else				/* pA and pB intersect on y */
	{
	    ymax = min(pA->ymax, pB->ymax);
	    save_pA = pA;
	    save_pB = pB;

	    /* we put a first x item into pR before entering the loop */
	    if (pA->xmin <= pB->xmin)
		ADD_TO_PR(pA->xmin, ymin, pA->xmax, ymax)
	    else
		ADD_TO_PR(pB->xmin, ymin, pB->xmax, ymax)

	    for (;;) 			/* one x item is already inside pR */
	    {
		if (pA->xmin <= pB->xmin)
		{
		    if (pA->xmin > pR[-1].xmax)
			ADD_TO_PR(pA->xmin, ymin, pA->xmax, ymax)
		    else
		    {
			if (pA->xmax > pR[-1].xmax)
			    pR[-1].xmax = pA->xmax;
		    }

		    pA++;
		    if (pA == pA_end || pA->ymin != save_pA->ymin)
		    {
			pA_is_done_but_not_pB = GTRUE;
			break;
		    }
		}
		else /* pB->xmin < pA->xmin */
		{
		    if (pB->xmin > pR[-1].xmax)
			ADD_TO_PR(pB->xmin, ymin, pB->xmax, ymax)
		    else
		    {
			if (pB->xmax > pR[-1].xmax)
			    pR[-1].xmax = pB->xmax;
		    }

		    pB++;
		    if (pB == pB_end || pB->ymin != save_pB->ymin)
		    {
			pA_is_done_but_not_pB = GFALSE;
			break;
		    }
		}
	    }

	    if (pA_is_done_but_not_pB)
	    {
		while (pB->xmax <= pR[-1].xmax)
		{
		    pB++;
		    if (pB == pB_end || pB->ymin != save_pB->ymin)
			goto end_of_this_scanline; /* not to waste time */
		}

		/* here pB->xmax > pR[-1].xmax */
		if (pB->xmin <= pR[-1].xmax)
		{
		    pR[-1].xmax = pB->xmax;
		    pB++;
		}

		while (pB < pB_end && pB->ymin == save_pB->ymin)
		{
		    ADD_TO_PR(pB->xmin, ymin, pB->xmax, ymax)
		    pB++;
		}
	    }
	    else /* pB is done but not pA */
	    {
		while (pA->xmax <= pR[-1].xmax)
		{
		    pA++;
		    if (pA == pA_end || pA->ymin != save_pA->ymin)
			goto end_of_this_scanline; /* not to waste time */
		}

		/* here pA->xmax > pR[-1].xmax */
		if (pA->xmin <= pR[-1].xmax)
		{
		    pR[-1].xmax = pA->xmax;
		    pA++;
		}

		while (pA < pA_end && pA->ymin == save_pA->ymin)
		{
		    ADD_TO_PR(pA->xmin, ymin, pA->xmax, ymax)
		    pA++;
		}
	    }

	  end_of_this_scanline:
	    if (save_pA->ymax > ymax)
	    	{ pA = save_pA; ymin = ymax; }
	    else if (save_pB->ymax > ymax)
	    	{ pB = save_pB; ymin = ymax; }
	    else if (pA < pA_end && pB < pB_end)
		ymin = min(pA->ymin, pB->ymin);
	    else if (pA == pA_end)
	    {
		if (pB < pB_end) //otherwise, ymin is no more used.
		    ymin = pB->ymin;
	    }
	    else
		ymin = pA->ymin;
	}

	/* try to concat this with previous scanline */
      try_to_concat_scanlines:
	concat = GFALSE;
	if (pR - middle_pR == middle_pR - prev_pR &&
	    middle_pR->ymin == prev_pR->ymax)
	{
	    concat = GTRUE;
	    for (int i = pR - middle_pR - 1; i >= 0; --i)
		if (middle_pR[i].xmin != prev_pR[i].xmin ||
		    middle_pR[i].xmax != prev_pR[i].xmax)
		{
		    concat = GFALSE;
		    break;
		}

	    if (concat)
	    {
		for (i = pR - middle_pR - 1; i >= 0; --i)
		    prev_pR[i].ymax = middle_pR->ymax;
		pR = middle_pR;
	    }
	}

	if (!concat)
	{
	    prev_pR = middle_pR;
	    middle_pR = pR;
	}

	/* now test for the end before reentering the loop */
	if (pA == pA_end)
	{
	    if (pB < pB_end)
	    {
		/* before simply adding the rest of pB, we must care if
		   the first pB "scanline" may be concated with the last one
		   inside pR */
		if (pB->ymin <= ymin)
		{
		    pBymin = pB->ymin;
		    do {
			ADD_TO_PR(pB->xmin, ymin, pB->xmax, pB->ymax)
			pB++;
		    } while (pB < pB_end && pB->ymin == pBymin);

		    goto try_to_concat_scanlines;

		    /* next time we will get "pB == pB_end" or
		       "pB->ymin > ymin". Therefore we will not do this
		       bloc twice */
		}

		while (pB < pB_end)
		{
		    ADD_TO_PR(pB->xmin, pB->ymin, pB->xmax, pB->ymax)
		    pB++;
		}
	    }

	    break;
	}
	else if (pB == pB_end)	    	/* pA < pA_end */
	{
	    /* before simply adding the rest of pA, we must care if
	       the first pA "scanline" may be concated with the last one
	       inside pR */
	    if (pA->ymin <= ymin)
	    {
		pAymin = pA->ymin;
		do {
		    ADD_TO_PR(pA->xmin, ymin, pA->xmax, pA->ymax)
		    pA++;
		} while (pA < pA_end && pA->ymin == pAymin);
		
		goto try_to_concat_scanlines;

		/* next time we will get "pA == pA_end" or
		   "pA->ymin > ymin". Therefore we will not do this
		   bloc twice */
	    }

	    while (pA < pA_end)
	    {
		ADD_TO_PR(pA->xmin, pA->ymin, pA->xmax, pA->ymax)
		pA++;
	    }

	    break;
	}
    }

    /* it is finally done ! */

    R.count = pR - start_pR;

    if (R.size < R.count)
    {
	GFREE(R.boxes);
	R.size = R.count;
	R.boxes = (GBox*)GMALLOC(R.size * sizeof(GBox));
    }

    bcopy(start_pR, R.boxes, R.count * sizeof(GBox));
} /* end of "common_GAdd", a slightly big function, isn'it ? */


void common_GSub
    (const GBox* pA, const GBox* pA_end,
     const GBox* pB, const GBox* pB_end,
     GRegion& R)
{
    GBox *start_pR, *pR, *pR_end, *middle_pR, *prev_pR;
    const GBox *save_pA, *save_pB;
    GNumber xmin, ymin, ymax, pAymin;
    int	biggest_size;
    GBool concat;

    /* to be shure not to realloc too often or not to alloc a too big
       memory for a small region, we use temporary R boxes on stack.
       (Of course this is not true for poor people defining NO_ALLOCA).
       I am not sure that variable declaration works fine on all machines
       after an alloca call. This is why I put all them above. */

#if defined(__BORLANDC__) || defined(__MSC7__)  
    biggest_size = 1 * 1024; /*BUG: difficult to compute*/
#else
    biggest_size = 16 * 1024;
#endif

    /* BUG: bigger than 2 + (pA_end - pA) * (1 + (pB_end - pB))
       but difficult to compute exactly */

#ifndef NO_ALLOCA
    start_pR = (GBox*)alloca(biggest_size * sizeof(GBox));
#else
    if (tempDataSize < biggest_size)
	GError("Region.cc: too small temp data size");
    start_pR = (GBox*)tempData;
#endif
    pR = start_pR;
    pR_end = start_pR + biggest_size;
    middle_pR = pR;
    prev_pR = NULL;

    ymin = pA->ymin;

    for (;;)
    {
	if (pB->ymin > ymin)
	{
	    ymax = (pB->ymin < pA->ymax) ? pB->ymin : pA->ymax;
	    save_pA = pA;
	    do {
		ADD_TO_PR(pA->xmin, ymin, pA->xmax, ymax)
		pA++;
	    } while (pA < pA_end && pA->ymin == save_pA->ymin);

	    if (save_pA->ymax > ymax)
	    {
		pA = save_pA;
		ymin = ymax;
	    } else 
		if (pA < pA_end)
		    ymin = pA->ymin;
	}
	else if (ymin >= pB->ymax)
	{
	    do
		pB++;
	    while (pB < pB_end && pB->ymax <= ymin);
	}
	else /* pA and pB intersect on y */
	{
	    save_pA = pA;
	    save_pB = pB;
	    ymax = min(pA->ymax, pB->ymax);
	    xmin = pA->xmin;
	    for (;;)
	    {
		if (xmin < pB->xmin)
		{
		    if (pA->xmax <= pB->xmin)
		    {
			ADD_TO_PR(xmin, ymin, pA->xmax, ymax)
			pA++;
			if (pA == pA_end || pA->ymin != save_pA->ymin)
			    break;
			xmin = pA->xmin;
		    }
		    else
		    {
			ADD_TO_PR(xmin, ymin, pB->xmin, ymax)
			xmin = pB->xmin;
		    }
		}
		else if (pB->xmax <= xmin)
		{
		    pB++;
		    if (pB == pB_end || pB->ymin != save_pB->ymin)
			break;
		}
		else if (pB->xmax < pA->xmax)
		{
		    xmin = pB->xmax;
		    pB++;
		    if (pB == pB_end || pB->ymin != save_pB->ymin)
			break;
		}
		else
		{
		    pA++;
		    if (pA == pA_end || pA->ymin != save_pA->ymin)
			break;
		    xmin = pA->xmin;
		}
	    }

	    if (pA != pA_end && pA->ymin == save_pA->ymin)
	    {
		ADD_TO_PR(xmin, ymin, pA->xmax, ymax)
		pA++;
		while (pA != pA_end && pA->ymin == save_pA->ymin)
		{
		    ADD_TO_PR(pA->xmin, ymin, pA->xmax, ymax)
		    pA++;
		}
	    }

	    if (save_pA->ymax > ymax)
	    {
		pA = save_pA;
		ymin = ymax;
	    }
	    else if (save_pB->ymax > ymax)
	    {
		if (pA < pA_end)
		{
		    pB = save_pB;
		    ymin = pA->ymin;
		}
	    }
	    else /* save_pA->ymax == save_pB->ymax */
		if (pA < pA_end)
		    ymin = pA->ymin;
	}

      try_to_concat_scanlines:
	/* try to concat this with previous "scanline" */

	if (pR - middle_pR == middle_pR - prev_pR &&
	    middle_pR->ymin == prev_pR->ymax)
	{
	    concat = GTRUE;
	    for (int i = pR - middle_pR - 1; i >= 0; --i)
		if (middle_pR[i].xmin != prev_pR[i].xmin ||
		    middle_pR[i].xmax != prev_pR[i].xmax)
		{
		    concat = GFALSE;
		    break;
		}
	    if (concat)
	    {
	        for (i = pR - middle_pR - 1; i >= 0; --i)
		    prev_pR[i].ymax = middle_pR->ymax;
		pR = middle_pR;
	    }
	}

	/* Explanation of the following test "(pR > middle_pR)":
	   1) if this loop has not added any box to pR (i.e. pR == middle_pR)
	      we do not change "scanline" pointers.
	   2) If above the concat was successfull, pR became equal to
	      middle_pR, and will not be set again bellow. */

	if (pR > middle_pR)
	{
	    prev_pR = middle_pR;
	    middle_pR = pR;
	}

	/* before restarting a new "scanline" loop check for the end */

	if (pA == pA_end)
	    break;
	else if (pB == pB_end)
	{
	    if (pA->ymin < ymin)
	    {
		/* before simply adding the rest of pA, we must care if
		   the first pA "scanline" may be concated with the last
		   one inside pR */

		pAymin = pA->ymin;
		do {
		    ADD_TO_PR(pA->xmin, ymin, pA->xmax, pA->ymax)
		    pA++;
		} while (pA < pA_end && pA->ymin == pAymin);

		goto try_to_concat_scanlines;

		/* next time we will get "pA == pA_end" or
		   "pA->ymin > ymin". Therefore we will not do this bloc
		   twice. */
	    }

	    while (pA < pA_end)
	    {
		ADD_TO_PR(pA->xmin, pA->ymin, pA->xmax, pA->ymax)
		pA++;
	    }

	    break;
	}
    }

    /* it's finally done ! */

    R.count = pR - start_pR;

    if (R.size < R.count)
    {
	GFREE(R.boxes);
	R.size = R.count;
	R.boxes = (GBox*)GMALLOC(R.size * sizeof(GBox));
    }

    bcopy(start_pR, R.boxes, R.count * sizeof(GBox));

    R.recompute_bigbox();
} /* end of "common_GSub" */


void GAdd (const GRegion& A, const GBox& box, GRegion& R)
{
    /* We try to avoid calling the heavy function "common_GAdd",
       when adding a box to a region (when possible) */

    if (box.empty())
        { R = A; return; }
    if (A.big.empty())
    	{ R = box; return; }
    
    if (box.ymax == A.big.ymin
	&& (box.xmin == A.boxes[0].xmin && box.xmax == A.boxes[0].xmax)
	||
	box.ymin == A.big.ymax
	&& (box.xmin == A.boxes[A.count-1].xmin
	    && box.xmax == A.boxes[A.count-1].xmax))
    {
	goto too_difficult;
    }
    
    if (box.ymin >= A.big.ymax || box.ymax <= A.big.ymin)
    {
	/* box is "outside" A, its easy */

	if (&R == &A)
	{
	    if (box.ymin >= R.big.ymax)
	    {
		if (R.size == R.count)
		{
		    R.size += 8;
		    GREALLOC(R.boxes, R.size * sizeof(GBox));
		}
		R.boxes[R.count] = box;
	    }
	    else /* box.ymax <= R.big.ymin */
	    {
		if (R.size > R.count)
		    bcopy(R.boxes, R.boxes + 1, R.count * sizeof(GBox));
		else
		{
		    GBox* save = R.boxes;
		    R.size += 8;
		    R.boxes = (GBox*)GMALLOC(R.size * sizeof(GBox));
		    bcopy(save, R.boxes + 1, R.count * sizeof(GBox));
		    GFREE(save);
		}
		R.boxes[0] = box;
	    }
	}
	else /* &R != &A */
	{
	    if (R.size < A.count + 1)
	    {
		GFREE(R.boxes);
		R.size = A.count + 8;
		R.boxes = (GBox*)GMALLOC(R.size * sizeof(R.boxes));
	    }

	    if (box.ymin >= A.big.ymax)
	    {
		bcopy(A.boxes, R.boxes, A.count * sizeof(GBox));
		R.boxes[A.count] = box;
	    }
	    else /* box.ymax <= A.big.ymin */
	    {
		R.boxes[0] = box;
		bcopy(A.boxes, R.boxes + 1, A.count * sizeof(GBox));
	    }
	}

	R.big = A.big + box;
	R.count = A.count + 1;
	return;
    }


  too_difficult:    		/* we use the standard way: common_GAdd */
    
    R.big = A.big + box;
    common_GAdd(A.boxes, A.boxes + A.count, &box, &box + 1, R);
}


void GAdd (const GRegion& A, const GRegion& B, GRegion& R)
{
    if (A.count == 0)
    	{ R = B; return; }
    if (B.count == 0)
    	{ R = A; return; }
    if (&A == &B)
        { R = A; return; }

    R.big = A.big + B.big;
    common_GAdd(A.boxes, A.boxes + A.count, B.boxes, B.boxes + B.count, R);
}


GBool operator && (const GRegion& r, const GBox& box)
{
    if (r.big && box)
    {
	GBox* pB = r.boxes;
	GBox* pB_end = pB + r.count;

	while (pB->ymax <= box.ymin)
	    pB++; 			/* must reach one before pB_end
					   because r.big intersect box */

	while (pB->ymin < box.ymax)
	{
	    /* pB and box intersect on y */
	    if (pB->xmin < box.xmax && pB->xmax > box.xmin)
		return GTRUE;
	    pB++;
	    if (pB == pB_end)
		return GFALSE;
	}
    }

    return GFALSE;
}


void GSub (const GRegion& A, const GRegion& B, GRegion& R)
{
    if (A.count == 0 || B.count == 0 || !GIntersect(A.big, B.big))
    {
	R = A;
	return;
    }

    if (&A == &B)
    {
	R = GEmptyRegion();
	return;
    }

    common_GSub(A.boxes, A.boxes + A.count, B.boxes, B.boxes + B.count, R);
}


void GSub (const GRegion& A, const GBox& box, GRegion& R)
{
    if (A.count == 0 || box.empty() || !GIntersect(A.big, box))
    {
	R = A;
	return;
    }

    common_GSub(A.boxes, A.boxes + A.count, &box, &box + 1, R);
}


void GSub (const GBox& box, const GRegion& A, GRegion& R)
{
    if (A.count == 0 || box.empty() || !GIntersect(A.big, box))
    {
	R = box;
	return;
    }

    common_GSub(&box, &box + 1, A.boxes, A.boxes + A.count, R);
}


int  GPositionToPixel (GNumber mm, GNumber p_size)
{
    if (fabs(mm) == MAXFLOAT)
	return MAXSHORT;
    
    GNumber tmp = mm / p_size + 0.499;
    if (tmp < 0)
	tmp -= 0.999;

    if (tmp > MAXSHORT)
	tmp = MAXSHORT;
    else if (tmp < MINSHORT)
	tmp = MINSHORT;

    return (int)tmp;
}


GIntBox* GRegion::pixelBoxes
    (GNumber pixel_width, GNumber pixel_height, int& num_boxes) const
{
    if (count == 0)
    {
	num_boxes = 0;
	return NULL;
    }

    GIntBox* pixboxes_start = (GIntBox*)GMALLOC(count * sizeof(GIntBox));
    GIntBox* pixboxes = pixboxes_start;

    GBox* pB = boxes;
    GBox* pB_end = boxes + count;

    do {
	pixboxes->x = GPositionToPixel(pB->xmin, pixel_width);
	pixboxes->y = GPositionToPixel(pB->ymin, pixel_height);
	pixboxes->width =
	    GPositionToPixel(pB->xmax, pixel_width) - pixboxes->x;
	pixboxes->height =
	    GPositionToPixel(pB->ymax, pixel_height) - pixboxes->y;
	if (pixboxes->width > 0 && pixboxes->height > 0)
	    pixboxes++;
	pB++;
    } while (pB < pB_end);

    num_boxes = pixboxes - pixboxes_start;
    return pixboxes_start;
}


void GIntersect (const GRegion& A, const GRegion& B, GRegion& R)
{
    /* A inter B == A minus not B,
       therefore: */

    GRegion notB;
    GSub(GBiggestBox(), B, notB);
    GSub(A, notB, R);
}


void GIntersect (const GRegion& A, const GBox& B, GRegion& R)
{
    GRegion notB(GBiggestBox());
    GSub(notB, B, notB);
    GSub(A, notB, R);
}


GNumber GRegion::squareArea () const
{
    GNumber sq_area = 0;

    GBox* pB_end = boxes + count;

    for (GBox* pB = boxes; pB < pB_end; pB++)
	sq_area += pB->width() * pB->height();

    return sq_area;
}

GBool GRegion::squareAreaBiggerThan (GNumber area) const
{
    GNumber computed = 0;
    GBox* pB_end = boxes + count;

    for (GBox* pB = boxes; pB < pB_end; pB++)
    {
	computed += pB->width() * pB->height();
	if (computed >= area)
	    return GTRUE;
    }

    return GFALSE;
}


void GAdd (const GRegion& A, const GRegion& B, GRegion& R, GNumber)
{
    GAdd(A, B, R);
}


GRegion::~GRegion()
{
    GFREE(boxes);

    /* The unbearable order of global destructors make possible (shure,
       it must be our bug) the access of a global Region even if its
       destructor was already called. Therefore, to avoid a bus error
       after exit, we reset also its size. */
    size = 0;
}


GBool operator >= (const GRegion& A, const GRegion& B)
{
    if (B.count == 0) return GTRUE;
    if (A.count == 0) return GFALSE;

    if (A.big.xmin > B.big.xmin || A.big.ymin > B.big.ymin ||
	A.big.xmax < B.big.xmax || A.big.ymax < B.big.ymax)
	return GFALSE;

    GBox* b = B.boxes;
    GBox* endb = b + B.count;
    GBox* a = A.boxes;
    GBox* enda = a + A.count;

    GNumber ymin = b->ymin;
    for (;;)
    {
	if (ymin < a->ymin)
	    return GFALSE;

	if (a->ymax <= ymin)
	    do {
		a++;
		if (a == enda) return GFALSE;
	    } while (a->ymax <= ymin);

	GBox* inia = a;
	GBox* inib = b;

	for (;;)
	{
	    while (a->xmax <= b->xmin)
	    {
		a++;
		if (a == enda || a->ymin != inia->ymin)
		    return GFALSE;
	    }
	    if (b->xmin < a->xmin || b->xmax > a->xmax)
		return GFALSE;
	    b++;
	    if (b == endb)
		return GTRUE;
	    if (b->ymin != inib->ymin)
		break;
	}

	if (inib->ymax > inia->ymax)
	{
	    b = inib;
	    ymin = inia->ymax;
	}
	else if (inia->ymax > inib->ymax)
	{
	    a = inia;
	    ymin = b->ymin;
	}
    }
}


void GRegion::operator += (const GBox& b)
{
    GAdd(*this, b, *this);
}


void GRegion::setEmpty ()
{
    count = 0;
    big = GEmptyBox();
}

GBool operator != (const GRegion& a, const GRegion& b)
{
    return !(a == b);
}
