#ifdef REV_INFO
static char rcsid[] = "$Id: SharedObj.cc,v 1.3 1992/09/25 15:09:34 smadja Exp $";
#endif
/*****************************************************************************
MODULE:  Tools (Shared Object Management)
FILE:    SharedObj.cc
AUTHOR:  Jacques Davy, Philippe Smadja
DATE:    August 1990
DESCRIPTION:
    See Go/Tools/SharedObj.h

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/machdef.h>
#include <Go/Tools/SharedObj.h>
#include <Go/Tools/Alloc.h>
#include <Go/Tools/Error.h>
#include <Go/GoStream.h>

/*
 * GCallback types
 */
static char shoNupdatePostcond[1];
static char shoNdestroyPostcond[1];
static char shoNincrRefPostcond[1];
static char shoNdecrRefPostcond[1];

/*
 * Private Methods
 * ---------------
 */
/**/
GSharedObject* _findAndModifyUnnamed (GList*& head, ulong mask, const void * values)
{
    GSharedObject* obj = (GSharedObject*) head;
    for ( ; obj != NULL; obj = (GSharedObject*)(obj->next()) )
        if (!(obj->_name) && (obj->_match (mask, values)) ) {
	    obj->_refcnt++;
	    obj->_call_incrRef_postcond ();
	    obj->raise (head);
	    if (obj->_mask != mask) {
                ulong update_mask = obj->_update (mask, values);
		obj->_mask |= mask ;
		if (update_mask)
		    obj->_call_update_postcond (update_mask);
	    }
	    return obj;
        }
    return (GSharedObject*)NULL;
}

/**/
GSharedObject* _findAndModifyNamed (GList*& head, ulong mask, const void * values, const char* name)
{
    GSharedObject* obj = (GSharedObject*)head;
    for ( ; obj != NULL; obj = (GSharedObject*)(obj->next()) )
        if (obj->_name && (strcmp (obj->_name, name) == 0)) {
            obj->_refcnt++;
            obj->_call_incrRef_postcond();
            obj->raise (head);
            if (mask) {
                ulong update_mask = obj->_update (mask, values);
		obj->_mask |= mask;
		if (update_mask)
		    obj->_call_update_postcond (update_mask);
	    }
	    return obj;
        }
    return (GSharedObject*)NULL;
}

/*
 * Protected Methods
 * -----------------
 */

/**/
int GSharedObject::_addCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = (GAssoc*)&_callbacks;
    GAssoc* cb = (GAssoc*)_cb;
    
    GCallbackRec cbrec (proc, closure);
    return cb->add(type, sizeof cbrec, (char*)&cbrec);
}

/**/
int GSharedObject::_remCallback (const char* type, GCBProc proc, const void* closure) const
{
    const GAssoc* _cb = (GAssoc*)&_callbacks;
    GAssoc* cb = (GAssoc*)_cb;

    GCallbackRec cbrec (proc, closure);
    return cb->rem(type, sizeof cbrec, (char*)&cbrec);
}

/**/
void GSharedObject::_call_update_postcond (ulong update_mask)
{
    if (_flags._has_update_postcond)
    {
	int count;
	GCallbackRec* cbrec = (GCallbackRec*)_callbacks.get (shoNupdatePostcond, count);
	(void)GCallCallbacks(cbrec, count, this, update_mask);
    }
}

/**/
void GSharedObject::_call_destroy_postcond ()
{
    if (_flags._has_destroy_postcond)
    {
	int count;
	GCallbackRec* cbrec = (GCallbackRec*)_callbacks.get (shoNdestroyPostcond, count);
	(void)GCallCallbacks(cbrec, count, this);
    }
}

/**/
void GSharedObject::_call_incrRef_postcond ()
{
    if (_flags._has_incrRef_postcond)
    {
	int count;
	GCallbackRec* cbrec = (GCallbackRec*)_callbacks.get (shoNincrRefPostcond, count);
	(void)GCallCallbacks(cbrec, count, this);
    }
}

/**/
void GSharedObject::_call_decrRef_postcond ()
{
    if (_flags._has_decrRef_postcond)
    {
	int count;
	GCallbackRec* cbrec = (GCallbackRec*)_callbacks.get (shoNdecrRefPostcond, count);
	(void)GCallCallbacks(cbrec, count, this);
    }
}

/**/
GSharedObject::GSharedObject (GFunctionList* funcList, ulong mask, const char* name) : GList(((*(GSharedObject*)NULL).*(funcList->head))())
{
    __funcList = funcList;
    _refcnt = 1;
    _mask = mask;
    if (name)
    {
       _name = new char[strlen(name)+1];
        strcpy (_name, name);
    } else
        _name = NULL;
}

/**/
GSharedObject::~GSharedObject()
{
    remove (_head());
    if (_name)
        delete _name;
}

/**/
GSharedObject* GSharedObject::_create(ulong mask, const void * values, const char* name)
{
    return (*__funcList->create)(__funcList, mask, values, name);
}

/*
 * Public Methods
 * --------------
 */

/* static */
const char* GSharedObject::classname()
{
    return "GSharedObject";
}

/* virtual */
const char* GSharedObject::getClassname() const
{
    return GSharedObject::classname();
}

/**/
void GSharedObject::modify (GSharedObject*& current, ulong mask, const void * values)
{
    if ( mask == 0 )
        return ;

    GSharedObject* obj;
    current->raise (current->_head());

    // Is there any matching object?
    ulong newMask;
#ifdef NO_ALLOCA
    char* newValues = (char*)GMALLOC (current->_valueSize());
#else
    char* newValues = (char*)alloca ((int) current->_valueSize());
#endif
    current->_makeNewValues (mask, values, newMask, newValues);

    obj = (current->_name ?
	   _findAndModifyNamed (current->_head(), newMask, newValues, current->_name) :
	   _findAndModifyUnnamed (current->_head(), newMask, newValues));

    if (obj)
    {
        current->destroy();
        current = obj;
#ifdef NO_ALLOCA
	GFREE (newValues);
#endif
        return;
    }

    // No match, update the current unnamed object
    if (_refcnt == 1 ) {
        ulong update_mask = current->_update (mask, values);
	_mask = newMask;
	if (update_mask)
            current->_call_update_postcond (update_mask);
    } else {
        current->destroy ();
	current = _create (newMask, newValues);
    }
#ifdef NO_ALLOCA
    GFREE (newValues);
#endif
}

/**/
void GSharedObject::rename (GSharedObject*& current, const char* name)
{
    if (name && current->_name && (strcmp (name, current->_name) == 0))
        return;

    GSharedObject* obj = NULL;

    // Find out a named object
    if (name)
        obj = _findAndModifyNamed (current->_head(), 0, NULL, name);

    // If no named object exists, a new one is created with current values
    if (!obj)
    {
        // Getting current values
        ulong mask;
#ifdef NO_ALLOCA
	char* values = (char*) GMALLOC (current->_valueSize());
#else
	char* values = (char*) alloca ((int) current->_valueSize());
#endif
	current->_makeNewValues (0, NULL, mask, values);
        obj = ModifyOrCreateShO(current->__funcList, mask, values, name);
#ifdef NO_ALLOCA
	GFREE (values);
#endif
    }
    current->destroy();
    current = obj;
}

/**/
GSharedObject* GSharedObject::dup ()
{
    _refcnt++;
    _call_incrRef_postcond ();
    raise(_head());

    return this;
}

/**/
#define _add_rem_callback(type, flag) \
    (FLAGS->flag =\
         0 < ((add == GADD)\
	      ? _addCallback(type, (GCBProc)cb, closure)\
	      : _remCallback(type, (GCBProc)cb, closure)))

/**/
void GSharedObject::updatePostcond (void cb (void*, GSharedObject*, ulong update_mask), const void* closure, GAddOrRemove add) const
{
    const GShoPostcondFlags* flags = &_flags;
    GShoPostcondFlags* FLAGS = (GShoPostcondFlags*)flags;
    _add_rem_callback(shoNupdatePostcond, _has_update_postcond);
}

/**/
void GSharedObject::destroyPostcond (void cb (void*, const GSharedObject*), const void* closure, GAddOrRemove add) const
{
    const GShoPostcondFlags* flags = &_flags;
    GShoPostcondFlags* FLAGS = (GShoPostcondFlags*)flags;
    _add_rem_callback(shoNdestroyPostcond, _has_destroy_postcond);
}

/**/
void GSharedObject::incrRefPostcond (void cb (void*, GSharedObject*), const void* closure, GAddOrRemove add) const
{
    const GShoPostcondFlags* flags = &_flags;
    GShoPostcondFlags* FLAGS = (GShoPostcondFlags*)flags;
    _add_rem_callback(shoNincrRefPostcond, _has_incrRef_postcond);
}

/**/
void GSharedObject::decrRefPostcond (void cb (void*, GSharedObject*), const void* closure, GAddOrRemove add) const
{
    const GShoPostcondFlags* flags = &_flags;
    GShoPostcondFlags* FLAGS = (GShoPostcondFlags*)flags;
    _add_rem_callback(shoNdecrRefPostcond, _has_decrRef_postcond);
}
#undef _add_rem_callback


/**/
void GSharedObject::destroy ()
{
    --_refcnt;
    _call_decrRef_postcond ();

    if (_refcnt == 0) {
	_call_destroy_postcond ();
        delete this;
    }
}

/*
 * Friend functions
 * ----------------
 */

/**/
GSharedObject* ModifyOrCreateShO (GFunctionList* funcList, ulong mask, const void * values, const char* name)
{
    GList*& head = ((*(GSharedObject*)NULL).*(funcList->head))();
    GSharedObject* obj = (name ?
			 _findAndModifyNamed (head, mask, values, name) :
			 _findAndModifyUnnamed(head, mask, values));

    if (!obj)
    {
        // Not found ... create a new one
        obj = (*funcList->create)(funcList, mask, values, name);
    }
    return obj;
}


/**/
void ModifyNamedShO (GFunctionList* funcList, ulong mask, const void * values, const char* name)
{
    if (!name)
	GError ("Enable to modify a nameless named shared object\n");

    GSharedObject* obj = ModifyOrCreateShO (funcList, mask, values, name);
    if (obj->refcnt() > 1)
    {
	// It was just a modification, not a new creation (==> refcnt--);
	obj->destroy();
    }
}



/**/
GList*& GSharedObject::_head()
{
#ifdef __MSC7__
    GShoHeadList tmpFunc = __funcList->head;
    return (this->*tmpFunc)();
#else
    return ((*this).*(__funcList->head))();
#endif
}

/**/
GSharedObject* GSharedObject::headOfList() const
{
    return (GSharedObject*) ((GSharedObject*)this)->_head();
}

/**/
GSharedObject* GSharedObject::next() const
{
#ifdef __MSC7__
	 GList * tmp =  GList::next();
    return (GSharedObject*)tmp;	
#else
    return (GSharedObject*) ((GList*)this)->next();
#endif
}

/**/
int GSharedObject::_valueSize()
{
#ifdef __MSC7__
    GShoValueSize tmpFunc = __funcList->value_size;
    return (this->*tmpFunc)();
#else
    return ((*this).*(__funcList->value_size))();
#endif
}
