#ifdef REV_INFO
static char rcsid[] = "$Id: Undo.cc,v 1.4 1992/12/10 13:34:14 jojic Exp $";
#endif
/*****************************************************************************
         Copyright (c) BULL S.A. 1990

MODULE:  Go Tools
FILE:    Undo.cc
AUTHOR:  Oliver Jojic
DATE:    Nov 26, 1991

DESCRIPTION:
    See Undo.h

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/Tools/Undo.h>
#include <Go/Tools/Error.h>

const int gDefaultUndoActionsMaxCount = 128;

#define GOPEN_MARK ((void*)1)
#define GCLOSE_MARK ((void*)2)
#define	GSEPARATOR_MARK ((void*)3)


GUndoManager::GUndoManager ()
{
    reset();
}


GUndoManager::~GUndoManager ()
{
    reset();
}


void GUndoManager::reset ()
{
    reset_redo_queue();
    reset_undo_queue();

    _state = outside;

    unclosed_mark_count = 0;

    max_count = gDefaultUndoActionsMaxCount;
    _count = 0;
}


void GUndoManager::openMark ()
{
    if (_state == outside)
	reset_redo_queue();

    ((_state == inside_undo) ? &redo_queue : &undo_queue)
	->putlast(GOPEN_MARK);

    unclosed_mark_count += 1;
}


void GUndoManager::closeMark ()
{
    if (unclosed_mark_count == 0)
    {
	GWarning("someone try to close an undo mark which is not opened");
	return;
    }

    GQueue* q = (_state == inside_undo) ? &redo_queue : &undo_queue;

    unclosed_mark_count -= 1;

    void* top = q->getlast();
    if (top != GOPEN_MARK)
    {
	q->putlast(top);
	q->putlast(GCLOSE_MARK);
	if (unclosed_mark_count == 0 && q == &undo_queue)
	    incr_count();
    }
}


void GUndoManager::call_undo ()
{
    int mark_count = 0;
    do
    {
	GUndoProc proc, fproc;
	void *closure, *fclosure;

	int save_mark_count = mark_count;
	get_last_undo((_state != inside_undo) ? &redo_queue : &undo_queue,
		      mark_count, proc, closure, fproc, fclosure);
	if (proc != NULL)
	    proc(closure);
	else if (mark_count > save_mark_count)
	    openMark();
	else if (mark_count < save_mark_count)
	    closeMark();
    }
    while (mark_count > 0);

    if (unclosed_mark_count != 0)
    {
	GWarning("an undo mark was not closed, I close it");
	while (unclosed_mark_count > 0)
	    closeMark();
    }
}


void GUndoManager::undo ()
{
    if (_state != outside)
    {
	GWarning("a try to undo from an already running undo");
	return;
    }

    if (unclosed_mark_count > 0)
    {
	GWarning("a try to undo while a marked pushundo is not closed");
	return;
    }

    /*
      { _state == outside  	unclosed_mark_count == 0 }
     */

    _state = inside_undo;
    call_undo();
    /* { unclosed_mark_count == 0 } */
    _state = outside;

    if (_count > 0)
	_count -= 1;
}


void GUndoManager::redo ()
{
    if (_state != outside)
    {
	GWarning("someone try to redo from an already running undo");
	return;
    }

    if (unclosed_mark_count > 0)
    {
	GWarning("someone try to redo while an undo mark is not closed");
	return;
    }

    /*
      { _state == outside	unclosed_mark_count == 0 }
     */

    _state = inside_redo;
    call_undo();
    /* { unclosed_mark_count = 0 } */
    _state = outside;
}


void GUndoManager::push
    (GUndoProc proc, void* closure, GUndoProc fproc, void* fclosure)
{
    if (_state == outside)
	reset_redo_queue();

    add_last_undo(proc, closure, fproc, fclosure);

    if (_state != inside_undo && unclosed_mark_count == 0)
	incr_count();
}


GBool GUndoManager::undoEmpty () const
{
    return _count <= 0;
}


GBool GUndoManager::redoEmpty () const
{
    return redo_queue.empty();
}


void GUndoManager::maxCount (int _max)
{
    if (_max < 0)
	_max = 0;
    
    max_count = _max;

    while (_count > max_count)
	remove_first_undo();
}


void GUndoManager::remove_first_undo ()
{
    int mark_count = 0;
    do
    {
	GUndoProc fproc;
	void *fclosure;
	get_first_undo(mark_count, fproc, fclosure);
	if (fproc != NULL)
	    fproc(fclosure);
    }
    while (mark_count > 0);

    _count -= 1;
}


void GUndoManager::incr_count ()
{
    _count += 1;

    if (_count > max_count)
	remove_first_undo();
}


void GUndoManager::get_first_undo
    (int& mark_count, void (*&fproc)(void*), void *&fclosure)
{
    fproc = NULL;
    fclosure = NULL;

    void* top = undo_queue.getfirst();

    if (top == GOPEN_MARK)
	mark_count += 1;
    else if (top == GCLOSE_MARK)
	mark_count -= 1;
    else //top == GSEPARATOR_MARK
    {
	fclosure = undo_queue.getfirst();
	fproc = (void(*)(void*)) undo_queue.getfirst();

	if (undo_queue.empty())
	{
	    fproc = NULL;
	    fclosure = NULL;
	}
	else
	{
	    top = undo_queue.getfirst();
	    if (top == GOPEN_MARK ||
		top == GCLOSE_MARK ||
		top == GSEPARATOR_MARK)
	    {
		undo_queue.putfirst(top);
		fproc = NULL;
		fclosure = NULL;
	    }
	    else
		(void)undo_queue.getfirst();
	}
    }
}


void GUndoManager::get_last_undo
    (GQueue* q, int& mark_count,
     void (*&proc)(void*), void *&closure,
     void (*&fproc)(void*), void *&fclosure)
{
    proc = fproc = NULL;
    closure = fclosure = NULL;

    void* top = q->getlast();

    if (top == GOPEN_MARK)
	mark_count -= 1;
    else if (top == GCLOSE_MARK)
	mark_count += 1;
    else
    {
	proc = (void(*)(void*)) top;
	closure = q->getlast();

	top = q->getlast();
	if (top != GSEPARATOR_MARK)
	{
	    fproc = (void(*)(void*)) top;
	    fclosure = q->getlast();
	    q->getlast(); //GSEPARATOR_MARK;
	}
    }
}


void GUndoManager::add_last_undo
    (GUndoProc proc, void *closure, GUndoProc fproc, void *fclosure)
{
    GQueue* q = (_state == inside_undo) ? &redo_queue : &undo_queue;

    q->putlast(GSEPARATOR_MARK);

    if (fproc != NULL)
    {
	q->putlast((fclosure == NULL) ? closure : fclosure);
	q->putlast(fproc);
    }

    q->putlast(closure);
    q->putlast(proc);
}


void GUndoManager::reset_undo_queue ()
{
    int mark_count = 0;
    while (!undo_queue.empty())
    {
	GUndoProc proc, fproc;
	void *closure, *fclosure;

	get_last_undo(&undo_queue, mark_count, proc, closure, fproc, fclosure);
	if (fproc != NULL)
	    fproc(fclosure);
    }

    _count = 0;
}


void GUndoManager::reset_redo_queue ()
{
    int mark_count = 0;
    while (!redo_queue.empty())
    {
	GUndoProc proc, fproc;
	void *closure, *fclosure;

 	get_last_undo(&redo_queue, mark_count,
		      proc, closure, fproc, fclosure);
	if (fproc != NULL)
	    fproc(fclosure);
    }
}


void GUndoManager::removeOneUndo ()
{
    if (_state != outside)
    {
	GWarning("a try to remove undo from an already running undo");
	return;
    }

    if (unclosed_mark_count > 0)
    {
	GWarning("a try to remove undo while a marked pushundo is not closed");
	return;
    }

    /*
      { _state == outside  	unclosed_mark_count == 0 }
     */

    int mark_count = 0;
    do 
    {
	GUndoProc proc, fproc;
	void *closure, *fclosure;
	get_last_undo (&undo_queue, mark_count, proc, closure, fproc, fclosure);
	if (fproc != NULL)
	    fproc (fclosure);
    } while (mark_count > 0);

    if (_count > 0)
	_count -= 1;
}


void GUndoManager::removeOneRedo ()
{
    if (_state != outside)
    {
	GWarning("a try to remove redo from an already running undo");
	return;
    }

    if (unclosed_mark_count > 0)
    {
	GWarning("a try to remove redo while a marked pushundo is not closed");
	return;
    }

    /*
      { _state == outside  	unclosed_mark_count == 0 }
     */

    int mark_count = 0;
    do 
    {
	GUndoProc proc, fproc;
	void *closure, *fclosure;
	get_last_undo (&redo_queue, mark_count, proc, closure, fproc, fclosure);
	if (fproc != NULL)
	    fproc (fclosure);
    } while (mark_count > 0);
}
