#ifdef REV_INFO
static char rcsid[] = "$Id: GoUI.cc,v 1.5 1992/12/31 13:08:43 jojic Exp $";
#endif
/*****************************************************************************
MODULE:  Go Kernel (User Interface)
FILE:    GoUI.cc
AUTHOR:  Oliver Jojic
DATE:    January 91

COPYRIGHT:
  Copyright (c) 1990, 1991 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.


DESCRIPTION:

  The GoUI object gets events from its GoRoot associated GoInterf
  and dispatches them through GoSensors.

  We will describe following GoUI actions:
  	1) getting events from the associated GoInterf.
	2) sending events to sensors.
		2a) filter management.
		2b) choice of the right sensor and enter/leave management.
        3) management of a cursor stack.

				---------

  1) GETTING EVENTS FROM THE ASSOCIATED GOINTERF

  	GoUI sets an ``eventPostcond'' on the associated GoInterf:
		event_postcond_proc(goui, gointerf, event);
	The friend private function event_postcond_proc simply calls
		goui->dispatch(event);

	GoInterf calls event postconditions each time it gets an event which
	type matches the event interest mask. Therefore this
	event mask must contain all event types needed by currently active
	user interactions. Sensors must tell GoUI which event types
	interest them, and GoUI will set the event mask of GoInterf.
	If the associated GoInterf of GoUI changes, GoUI will remove its
	mask from the old GoInterf and add it to the new GoInterf. Localy,
	GoUI stores a counter for each event type, i.e. if an add is done
	3 times for one event mask, the remove must be done 3 times before
	GoUI decides to remove this event mask from GoInterf.
	GoUI makes a special management with enter/leave sensor mask. In that
	case it set the compress-motion mask to GoInterf.

	Therefore, if we remove the GoInterf of a GoUI, we must:
		- remove the eventPostcond ``event_postcond_proc'',
		- remove the event type interest mask,
		- eventually remove the compress-motion-mask.

	When a sensor grabs, to make overall drawings (independently of
	GoRoot), GoUI must do a special management of expose events:
		- inside the exposed area, everything has the background color.
		- the GoRoot itself must handle the expose.
		- GoUI must call ``goroot->flush()'', therefore the exposed
		  area contains the GoRoot tree.
		- GoUI must send the expose event to grabbing sensors, which
		  must restore their overall drawings.

	From the above, we can see that GoInterf must send the expose event
	to GoRoot and only after to GoUI. This order will be kept only
	if GoUI sets its exposeCallback only after GoRoot's one. 
	It is sufficient that GoUI sets the exposePostcond when the first
	sensor grabs (at this time GoRoot has already setted
	exposePostcond). GoUI must also remove the exposePostcond
	when it has no more grabbing sensors.


  2) SENDING EVENTS TO SENSORS

    2a) FILTER MANAGEMENT

     	GoUI stores a list of filters. The dispatch(event) method starts
	by calling all filters. The event returned by a filter became the
	input event of the following filter. The dispatch method itself
	deletes events generated by filters. If a filter returns NULL,
	no events are dispatched at all.
	Events generated by GoUI, i.e. enter/leave sensor, are not filtered.
	GoUI does not provide any ordering functionality onto these filters
	(this could be done if needed).
	Therefore, the event returned by the last filter became the dispatched
	event.

    2b) CHOICE OF THE RIGHT SENSOR AND ENTER/LEAVE MANAGEMENT

    	If the grabbing sensor stack is not empty, all events are send to
	the top of the grabbing sensor stack.
	Otherwise, if it is a char event and if there is a focussing sensor,
	the char event is sent to the focussing	sensor.

	Otherwise, GoUI must go through sensors, from top to bottom, to find
	the first one containing the event position and interested by this
	event. GoUI must also generate enter/leave sensor events (if it has
	the event/leave sensor mask in its event interest mask).

	Let us call ``current_sensor'' the last sensor having accepted an
	enter sensor event.

	The cross over sensors is done by using the
		goroot->ForEverySensorContainingPoint
			       (dispatch_recur, goui,
				event->position(), max_distance)
	method. ``dispatch_proc'' will then be called recursively with
	each sensor containing event position, starting with the top one.

	The dispatch_recur function will then do the following:
	If the sensor is equal to the current_sensor, GoUI do not care
	any more about enter/leave.
	Otherwise GoUI asks this sensor if it is interested with an enter
	sensor event. If it is, GoUI sends a leave sensor event to the
	current_sensor, and sends an enter sensor event to the new
	sensor, which became itself the current_sensor. After that, GoUI
	no more handles enter/leave.
	If the sensor is not interested with an enter sensor event, GoUI
	will continue this enter/leave management with the following sensor.
	
	If the current sensor will be destroyed, goui must find the new
	current sensor. This is implemented by setting a destroy postcondition
	onto that sensor.

	In all cases, GoUI asks the sensor if it is interested with the
	current event. If it is, GoUI sends the event to the sensor and
	stops the current event management.

	Of course, if both the current event and the enter/leave managements
	are terminated, GoUI stops the recursion on sensors.

******************************************************************************/


#include <Go/GoUI.h>
#include <Go/GoInterf.h>
#include <Go/Sensor.h>
#include <Go/GoRoot.h>
#include <Go/Tools/Set.h>
#include <Go/GoInDisp.h>
#include <Go/Interactor.h>
#include <Go/Tools/Bind.h>


/*
---------------------------
Description of private data
---------------------------

  GoRoot* _root;
  	The GoRoot associated to that GoUI.

  GEventMask _event_mask;
  ushort _event_mask_count[GN_EVENT_MASK];
  	GoUI stores not only an event mask, but also the number of
	``persons'' (e.g. sensors) which are interested for each event type.
	This allows users of GoUI to add or to remove an event type from
event	the mask without caring about other GoUI users. GoUI sets the
	same mask on its associated GoInterf.

  GNumber _max_distance;
  	This value is used to pick easy small graphic objects. If the pointer
	is less far than ``_max_distance'' from a Go, then it is considered
	as inside.

  GStack _grabbing_sensors;
  	GoUI stores a grabbing sensor stack. When this stack is not empty,
	the top of the stack receives all events, and enter/leave sensor
	events are no more generated.

  GSensor* _focusing_sensor;
	If ``_focusing_sensor'' is not NULL, then all char events are send
	to this sensor.
*/

/*
 All instances of GoUI are stored in a set. Using this set, it is
 easy to retreive which GoUI is associated to a given GoRoot.
 */
static GSet* goui_set();



/*
-------------------
internal structures
-------------------
*/

/* a FilterList is used to store event filters */
struct FilterList
{
    FilterList* next;
    GEvent* (*proc)(void*, GoUI*, const GEvent*);
    void* closure;
    
    FilterList(GEvent* (*f)(void*, GoUI*, const GEvent*),
	       void* cl, FilterList* n)
    {
	proc = f;
	closure = cl;
	next = n;
    }
};


/* This function is used by the next one */
void event_postcond_proc (void* closure, GoInterf*, const GEvent* event)
{
    GoUI* goui = (GoUI*)closure;

    goui->dispatch (event);
}


void goui_roots_interf_changed
    (void* closure, Go*, GoInterf* oldInterf)
{
    GoUI* goui = (GoUI*) closure;

    if (oldInterf != GoDefaultGoInterf())
    {
	oldInterf->removeInputMask (goui->_event_mask);
	oldInterf->eventPostcond (&event_postcond_proc, goui, GREMOVE);
    }

    if (goui->_root->interf() != GoDefaultGoInterf())
    {
	goui->_root->interf()->addInputMask (goui->_event_mask);
	goui->_root->interf()->eventPostcond (&event_postcond_proc, goui);
    }
}


void goui_sensors_root_changed (void* closure, Go* go, Go* oldRoot)
{
    GoUI* goui = (GoUI*) closure;
    GoSensor* sensor = (GoSensor*) go;

    if (oldRoot == (Go*) goui->_root)
    {
	if (sensor == goui->_focusing_sensor)
	    goui->focus (NULL);

	if (sensor == goui->_current_sensor)
	{
	    if (goui->_grabbing_sensors.empty())
		goui->recompute_current_sensor();
	    else
		goui->set_current_sensor (NULL);
	}
    }
    else if (sensor->root() == goui->_root)
	goui->recompute_current_sensor (sensor);
}



/*
----------------------------
public methods and functions
----------------------------
*/


void goui_focus (GoUI* goui, GoSensor* sor)
{
    if (goui != NULL)
	goui->focus (sor);
}
	


/*
 * The GoUI constructor initializes all internal data
 * and sets its _root by calling the ``root(GoRoot*)'' method.
 * It adds itself in the GoUI set returned by the ``goui_set()'' function.
 */
GoUI::GoUI (GoRoot* goroot)
{
    static GBool notFirstCall = GFALSE;

    _pointer_pos.x = _pointer_pos.y = -1000;

    if (!notFirstCall)
    {
	const char* typeFunction = "Function";

	GBind ("GoUI::focus", &goui_focus, typeFunction);

	notFirstCall = GTRUE;
    }

    _focusing_sensor = NULL;
    _filter_list = NULL;
    _enter_leave_already_managed = _event_already_managed = GFALSE;
    _event = NULL;
    
    _reset();
    
    goui_set()->add (this);
    root (goroot);

    GoSensor::metaclass()->rootPostcond (goui_sensors_root_changed, this);
}


/*
 * The GoUI destructor removes the connexion with its _root by calling
 * the ``interf'' method (with a NULL parameter), and free all internal data.
 * It removes itself from the GoUI set returned by the ``goui_set()'' function.
 */
GoUI::~GoUI ()
{
    root (NULL);

    if (_current_sensor)
	set_current_sensor (NULL);

    while (_filter_list != NULL)
    {
	FilterList* filter = _filter_list;
	_filter_list = filter->next;
	delete filter;
    }

    goui_set()->remove (this);

    GoSensor::metaclass()->rootPostcond
	(goui_sensors_root_changed, this, GREMOVE);
}


/*
 * The 2 methods add- and remove- InputMask modify the mask of events
 * managed by this GoUI. GoUI stores a counter per event type. I.e. if
 * there are 3 adds for one specific event type, it must be 3 removes on that
 * event type before it is removed.
 * GoUI then sets the same mask on its GoInterf, but only once.
 * For enter and leave sensor event masks, GoUI sets a compress-motion event
 * mask to its GoInterf.
 */
void GoUI::addInputMask (GEventMask added_mask)
{
    GEventMask old_mask = _event_mask;

    incr_or_decr_event_mask_count(added_mask, GADD);

    if (_event_mask == old_mask ||
	_root == NULL ||
	_root->interf() == GoDefaultGoInterf())
	return;

    _root->interf()->removeInputMask (old_mask);
    _root->interf()->addInputMask (_event_mask);
}

void GoUI::removeInputMask (GEventMask removed_mask)
{
    GEventMask old_mask = _event_mask;

    incr_or_decr_event_mask_count(removed_mask, GREMOVE);

    if (_event_mask == old_mask ||
	_root == NULL ||
	_root->interf() == GoDefaultGoInterf())
	return;

    _root->interf()->removeInputMask (old_mask);
    _root->interf()->addInputMask (_event_mask);
}


/*
 * This private method increment (if add == GADD) or decrement
 * (if add == GREMOVE) counters inside ``_event_mask_count'', following
 * bits inside ``mask_extension''.
 *
 * side effect: the BITOR of these counters, which is stored inside
 *``_event_mask'' is also set up to date.
 */
void GoUI::incr_or_decr_event_mask_count
    (GEventMask mask_extension, GAddOrRemove add)
{
    if (mask_extension & (GENTER_SENSOR_MASK|GLEAVE_SENSOR_MASK))
    {
	if (add)
	{
	    _event_mask_count[GENTER_SENSOR_MASK_INDEX] += 1;
	    if (_event_mask_count[GENTER_SENSOR_MASK_INDEX] == 1)
		recompute_current_sensor ();
	}
	else
	{
	    _event_mask_count[GENTER_SENSOR_MASK_INDEX] -= 1;

	    if (_event_mask_count[GENTER_SENSOR_MASK_INDEX] < 0)
	    {
		GWarning("GoUI.cc: enter sensor mask count less than zero !");
		GWarning("I reset it to zero.");
		_event_mask_count[GENTER_SENSOR_MASK_INDEX] = 0;
	    }

	    if (_event_mask_count[GENTER_SENSOR_MASK_INDEX] == 0 &&
		_current_sensor != NULL)
	    {
		set_current_sensor (NULL);
	    }
	}

	mask_extension &= ~(GENTER_SENSOR_MASK|GLEAVE_SENSOR_MASK);
	mask_extension |= GCOMPRESS_MOTION_MASK;
    }

    for (int index = 0; index < GN_EVENT_MASK; index++)
    {
	if (mask_extension == 0)
	    break;

	if (mask_extension & (1 << index))
	{
	    mask_extension &= ~(1 << index);

	    if (add)
		_event_mask_count[index] += 1;
	    else
		_event_mask_count[index] -= 1;

	    if (_event_mask_count[index] < 0)
	    {
		GWarning("GoUI.cc: event mask counter less than zero !");
		GWarning("for event index %d", index);
		GWarning("I reset it to zero.");
		_event_mask_count[index] = 0;
	    }

	    if (_event_mask_count[index])
		_event_mask |= (1 << index);
	    else
		_event_mask &= ~(1 << index);
	}
    }
}



/*
 * ``dispatch'' is the main method of GoUI. It is responsible to find the
 * right sensor, i.e. the one wich must handle ``event''.
 * This method is virtual, allowing GoUI subclasses to modify the event
 * dispatching rule.
 * The rule implemented in this method is described in the DESCRIPTION part
 * of the first comment of this file.
 */
void GoUI::dispatch (const GEvent* event)
{
    if (event == NULL)
	return;

    if (event->type() == GMOTION_EVENT)
	_pointer_pos = event->position();

    /* call filters */

    const GEvent* original_event = event;
    for (FilterList* l = _filter_list; l != NULL; l = l->next)
    {
	GEvent* new_event = l->proc(l->closure, this, event);

	if (new_event == NULL)
	{
	    if (event != original_event)
		delete (GEvent*)event;
	    return;
	}

	if (new_event != event)
	{
	    if (event != original_event)
		delete (GEvent*)event;
	    event = new_event;
	}
    }
    
    //here we must dispatch ``event'', which is not NULL */

    GoRoot* goroot = root();
    if (goroot == NULL)
	return;

    if (!_grabbing_sensors.empty())
    {
	/* grabbing sensors */

	if (event->type() != GEXPOSE_EVENT)
	{
	    GoSensor* sensor = (GoSensor*)_grabbing_sensors.top();
	    if (sensor->contains (event->position(), _max_distance))
		sensor->handle(event);
	}
	else
	{   /* management of an expose event */

	    /* The associated GoRoot has already received the expose event.
	       We simply tell it to flush its drawings. */

	    goroot->flush();
	    
	    /* And now we send the expose event to all grabbing sensors,
	       (after setting the clip onto the associated GoInterf). */

	    _root->interf()->setClip (event->region());

	    int count;
	    for (GoSensor** pS = (GoSensor**)_grabbing_sensors.data (count);
		 count > 0;
		 --count, ++pS)
	    {
		(*pS)->handle (event);
	    }

	    _root->interf()->resetClip();
	}
    }
    else if (_focusing_sensor != NULL &&
	     (event->type() == GCHAR_EVENT ||
	      event->type() == GFUNCTION_KEY_EVENT) &&
	     _focusing_sensor->careAbout (event))
    {
	/* focussing sensor */

	_focusing_sensor->handle (event);
    }
    else
    {
	/* true dispatching */

	if (event->type() != GEXPOSE_EVENT)
	{
	    _enter_leave_already_managed =
		!(_event_mask_count[GENTER_SENSOR_MASK_INDEX] > 0);
	    _event_already_managed = GFALSE;
	    _event = event;

	    /* inside dispatch_recur we will handle LeaveSensorEvent,
	       EnterSensorEvent and event itself. The LeaveSensorEvent
	       management may destroy (why not ?) the sensor which will
	       manage event or the sensor which will manage
	       the EnterSensorEvent. Therefore it is usefull to use
	       differDestroy and handleDestroy.
	       */
  	    Go::differDestroy();
	    root()->ForEverySensorContainingPoint
		(dispatch_recur, this, event->position(), _max_distance);
	    if (!_enter_leave_already_managed &&
		_current_sensor != NULL &&
		!_current_sensor->contains (event->position(), _max_distance))
	    {
		set_current_sensor (NULL, event->time ());
	    }
	    
	    Go::handleDestroy();
	}
    }

    //if event was created by filters we must delete it;
    if (event != original_event)
	delete (GEvent*)event;
}


/*
 * See the description of this function
 * in the DESCRIPTION section at top of this file.
 */
void dispatch_recur (void* closure, GoSensor* sensor)
{
    GoUI* goui = (GoUI*)closure;

    if (goui->_enter_leave_already_managed &&
	goui->_event_already_managed)
    {
	sensor->root()->stopRecursion (NULL);
	return;
    }

    if (!goui->_enter_leave_already_managed)
    {
	if (sensor == goui->_current_sensor)
	    goui->_enter_leave_already_managed = GTRUE;
	else
	{
	    if (sensor->interactor() != NULL &&
		(sensor->interactor()->inputMask() & GENTER_SENSOR_EVENT) != 0)
	    {
		goui->set_current_sensor (sensor, goui->_event->time ());
		goui->_enter_leave_already_managed = GTRUE;
	    }
	}
    }

    if (!goui->_event_already_managed)
    {
	if (sensor->careAbout (goui->_event))
	{
	    sensor->handle(goui->_event);
	    goui->_event_already_managed = GTRUE;
	}
    }
}


/*
 * This method allows you to add or to remove event filters.
 */
void GoUI::filter
    (GEvent* filter_proc(void*, GoUI*, const GEvent*),
     void* closure, GAddOrRemove add)
{
    if (add)
	_filter_list = new FilterList(filter_proc, closure, _filter_list);
    else
	for (FilterList** pl = &_filter_list;
	     *pl;
	     pl = &(*pl)->next)
	{
	    if ((*pl)->proc == filter_proc &&
		(*pl)->closure == closure)
	    {
		FilterList* list_to_remove = *pl;
		(*pl) = list_to_remove->next;
		delete list_to_remove;
	    }
	}
}


/*
 * These methods push/pop grabbing sensors.
 * A Stack object is used to manage the grabbing sensor stack.
 * We care about sending expose events to grabbing sensors.
 */
void GoUI::grab (GoSensor* sensor)
{
    if (_grabbing_sensors.empty())
    {
	//grabbing sensors may be interested by an expose event;
	addInputMask (GEXPOSE_MASK);
    }
    
    _grabbing_sensors.push(sensor);
}

void GoUI::ungrab (GoSensor* sensor)
{
    if (_grabbing_sensors.empty())
	return;
    
    if (sensor == NULL)
	sensor = (GoSensor*)_grabbing_sensors.top();

    GStack tmpStack;
    do
    {
	GoSensor* sor = (GoSensor*)_grabbing_sensors.pop();
	if (sor == sensor)
	    break;
	tmpStack.push (sor);
    } while (!_grabbing_sensors.empty());
    
    while (!tmpStack.empty())
	_grabbing_sensors.push (tmpStack.pop());

    if (_grabbing_sensors.empty())
    {
	//we are no more interested with expose events;
	removeInputMask (GEXPOSE_MASK);

	recompute_current_sensor ();
    }
}

GoSensor* GoUI::grabbingSensor () const
{
    if (_grabbing_sensors.empty())
	return NULL;
    else
	return (GoSensor*)_grabbing_sensors.top();
}


/*
 * This method allows you to set (or to remove if sensor is NULL)
 * a focusing sensor.
 */
void GoUI::focus (GoSensor* sensor)
{
    if (sensor == _focusing_sensor)
	return;

    if (_focusing_sensor != NULL)
    {
	GFocusOutEvent focusOut (_focusing_sensor, _root->interf());
	GoSensor* wasFocusing = _focusing_sensor;
	_focusing_sensor = NULL;
	wasFocusing->handle (&focusOut);
    }
    
    _focusing_sensor = sensor;

    if (_focusing_sensor != NULL)
    {
	GFocusInEvent focusIn (_focusing_sensor, _root->interf());
	_focusing_sensor->handle (&focusIn);
    }
}


GIdle refresh_goroot (void* closure)
{
    GoUI* goui = (GoUI*) closure;
    goui->root()->flush();

    return GIDLE_AFTER_NEXTEVENT;
}


void GoUI::_reset ()
{
    _root = NULL;

    _event_mask = GCOMPRESS_MOTION_MASK;
    for (int i = 0; i < GN_EVENT_MASK; i++)
	_event_mask_count[i] = 0;
    _event_mask_count[GCOMPRESS_MOTION_MASK_INDEX] = 1;
    
    _max_distance = .6;

    _current_sensor = NULL;

    (void) Go::differDestroy ();

    if (_focusing_sensor != NULL)
	focus (NULL);

    while (!_grabbing_sensors.empty())
    {
	GoSensor* sor = (GoSensor*)_grabbing_sensors.pop();
	sor->destroy();
    } 

    (void) Go::handleDestroy();
}


void goui_new_root_handle_sensors (void* closure, GoSensor* sensor)
{
    GoUI* goui = (GoUI*) closure;

    if (sensor->interactor() != NULL &&
	sensor->interactor()->state() == GITOR_RUNNING)
    {
	goui->addInputMask (sensor->interactor()->inputMask());
    }	
}



/*
 * This method modifies the GoRoot associated to this GoUI.
 * See precise description in the header of this file.
 */
void GoUI::root (GoRoot* goroot)
{
    if (_root)
    {
	if (_root->interf() != GoDefaultGoInterf())
	{
	    _root->interf()->removeInputMask (_event_mask);
	    _root->interf()->eventPostcond
		(&event_postcond_proc, this, GREMOVE);
	}
	
	_root->interfPostcond (goui_roots_interf_changed, this, GREMOVE);

	if (GInputDispatcher() != NULL)
	    GInputDispatcher()->idleProc (&refresh_goroot, this, GREMOVE);

	_reset();
    }

    _root = goroot;

    if (_root)
    {
	root()->ForEverySensor (goui_new_root_handle_sensors, this);
	recompute_current_sensor ();

	if (_root->interf() != GoDefaultGoInterf())
	{
	    _root->interf()->addInputMask (_event_mask);
	    _root->interf()->eventPostcond (&event_postcond_proc, this);
	}
	
	_root->interfPostcond (goui_roots_interf_changed, this);

	if (GInputDispatcher() == NULL)
	    GWarning("GoUI.cc: a GoUI is created without an %s",
"GoInputDispatcher so: the idleProc is not handled (refresh GoRoot)");
	else
	    GInputDispatcher()->idleProc (&refresh_goroot, this);
    }
}


/*
 * This method returns the GoUI which associated GoInterf is ``gointerf''.
 * It uses the set of all GoUIs returned by the goui_set() function.
 */
GoUI* GoUIFrom (const GoRoot* goroot)
{
    GSet* all_gouis = goui_set();
    for (GoUI* goui = (GoUI*)all_gouis->first();
	 goui != NULL;
	 goui = (GoUI*)all_gouis->next())
    {
	if (goui->_root == goroot)
	    return goui;
    }

    return NULL;
}



/*
-----------------------------
private methods and functions
-----------------------------
*/


/*
 * This method returns the set of all instances of GoUI.
 * This set, i.e. a static global variable, is accessible only through
 * this function, allowing special initializations needed for global
 * variables on Windows.
 */
GSet* goui_set ()
{
    static GSet* set;

    if (set == NULL)
	set = new GSet;

    return set;
}



/*
 * This method sets the ``_current_sensor'' field of goui. That means the
 * last sensor having accepted an enter sensor event.
 */
void GoUI::set_current_sensor (GoSensor* sensor, GTimeStamp time)
{
    if (sensor == _current_sensor)
	return;

    if (_current_sensor != NULL)
    {
	GLeaveSensorEvent leaveEvent (sensor, _root->interf(), time);
	GoSensor* tmp = _current_sensor;
	_current_sensor = NULL;
	tmp->handle (&leaveEvent);
    }

    if (sensor != NULL)
    {
	GEnterSensorEvent enterEvent (sensor, _root->interf(), time);
	sensor->handle (&enterEvent);
	_current_sensor = sensor;
    }
}


void GoUI::recompute_current_sensor (GoSensor* concernedSensor)
{
    if (!_grabbing_sensors.empty())
	return;

    if (concernedSensor != NULL)
    {
	if (concernedSensor->interactor() == NULL ||
	    (concernedSensor->interactor()->inputMask()
	     & GENTER_SENSOR_MASK) == 0 ||
	    !concernedSensor->contains (_pointer_pos, _max_distance))
	{
	    return;
	}
    }

    if (_event_mask_count[GENTER_SENSOR_MASK_INDEX])
    {
	GoSensor* new_current = NULL;

	root()->ForEverySensorContainingPoint
	    (recompute_sensor_recur, &new_current, _pointer_pos, _max_distance);

	set_current_sensor (new_current);
    }
}


void recompute_sensor_recur (void* closure, GoSensor* sensor)
{
    GoSensor** pNewCurrent = (GoSensor**) closure;

    if (sensor->interactor() != NULL &&
	(sensor->interactor()->inputMask() & GENTER_SENSOR_MASK) != 0)
    {
	sensor->root()->stopRecursion (NULL);
	*pNewCurrent = sensor;
    }
}


void GoUI::cancelGrabbers ()
{
    for (;;)
    {
	GoSensor* sor= grabbingSensor();
	if (sor == NULL)
	    return;
	sor->destroy();
    }
}
