#ifdef REV_INFO
static char rcsid[] = "$Id: Sensor.cc,v 1.6 1993/01/04 14:04:35 jacomet Exp $";
#endif
/*****************************************************************************
MODULE:  Graphic Objects
FILE:    Sensor.cc
AUTHOR:  jojic
DATE:    Jun 29 1992

DESCRIPTION:
  See also the Sensor.h header.
  
COPYRIGHT:
  Copyright (c) 1992 Bull S.A.
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull S.A. may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull S.A..
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull S.A. under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull S.A..

  Bull S.A. assumes no responsibility for the use or inability to use
  any of its software. The software is provided "AS IS" without warranty
  of any kind, and Bull S.A. expressly disclaims all implied warranties,
  including but not limited to the implied warranties of merchantability
  and fitness for a particular purpose.

  In no event shall Bull S.A. be liable for any special, incidental,
  indirect or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.
******************************************************************************/

#include <Go/Sensor.h>
#include <Go/GoUI.h>
#include <Go/Interactor.h>
#include <Go/Tools/Error.h>
#include <Go/GoStream.h>
#include <Go/GoRoot.h>
#include <Go/GoInterf.h>
#include <Go/Tools/Bind.h>

#include <ctype.h>
#include <assert.h>


static inline char* StringCopy (const char* s)
{
    int n = 1 + strlen(s);
    char* d = (char*)GMALLOC(n);
    bcopy(s, d, n);
    return d;
}



/*
 *-------------------------------------------------------
 * pre- and post- conditions
 *-------------------------------------------------------
 */


inline GBool GoSensor::call_program_precond (const char[256])
{
    /* TO BE IMPLEMENTED */
    return GTRUE;
}


inline GBool GoSensor::call_class_program_precond (const char[256])
{
    /* TO BE IMPLEMENTED */
    return GTRUE;
}


inline void GoSensor::call_program_postcond (const char*)
{
    /* TO BE IMPLEMENTED */
}


inline void GoSensor::call_class_program_postcond (const char*)
{
    /* TO BE IMPLEMENTED */
}



/*
 *-------------------------------------------------------
 * GoSensor public methods
 *-------------------------------------------------------
 */


/*
 * Sets the program of the sensor.
 * Of course, changing the program of the sensor, changes its associated
 * interactor.
 * This method returns GTRUE if a precondition refuses the program change.
 */
GBool GoSensor::program (const char* nm)
{
    char new_program[256];
    if (nm)
	strcpy(new_program, nm);
    else
	new_program[0] = 0;

    if (!call_program_precond(new_program) ||
	!call_class_program_precond(new_program))
	return GFALSE;

    remove_interactor();

    const char* old_program = _program;
    if (new_program[0])
	_program = StringCopy(new_program);
    else
	_program = NULL;
    
    create_interactor();

    call_program_postcond(old_program);
    call_class_program_postcond(old_program);
    GFREE(old_program);

    return GTRUE;
}


void GoSensor::activate ()
{
    if (!_active)
    {
	_active = GTRUE;
	create_interactor ();
    }
}


void GoSensor::disactivate ()
{
    if (_active)
    {
	_active = GFALSE;
	remove_interactor ();
    }
}

    

/*
 * Returns GTRUE if the associated interaction care about "event"
 * (i.e. will do something if we pass it the event).
 */
GBool GoSensor::careAbout (const GEvent* event) const
{
    return (_interactor == NULL)
	? GFALSE
	    : _interactor->careAbout(event);
}
   

/*
 * Handles ``event''.
 * I.e. passes it to the associated interactor.
 */
GBool GoSensor::handle (const GEvent* event)
{
    if (_interactor == NULL)
	return GFALSE;

    Go::differDestroy();
    if (!_interactor->handle(event))
    {
	Go::handleDestroy();
	return GFALSE;
    }
    
    if (_interactor && _interactor->state() != GITOR_RUNNING)
    {
	GBool destroyP = GTRUE;

	if (_postcond != NULL)
	{
	    void (*tmp) (void*, GoSensor*) = _postcond;
	    _postcond = NULL;
	    (*tmp) (_closure, this);

	    destroyP =
		_interactor == NULL ||
		    _interactor->state() != GITOR_RUNNING;
	}
	
	if (destroyP)
	{
	    remove_interactor();
	    this->destroy();
	}
    }

    Go::handleDestroy();
	
    return GTRUE;
}



/*
 *-------------------------------------------------------
 * GoSensor private methods
 *-------------------------------------------------------
 */


/*
 * This method finds the associated interactor from the sensor program.
 * It also returns GFALSE if it cannot find it.
 */
void GoSensor::create_interactor ()
{
    remove_interactor();
    
    if (_program == NULL ||
	!_active ||
	parent() == NULL ||
	interf() == GoDefaultGoInterf())
	return;

    _interactor = GCreateInteractor(_program, start_event, parent());
    if (_interactor == NULL)
	return;
    start_event = NULL;

    GoUI* goui = GoUIFrom (root());
    if (goui != NULL)
	goui->addInputMask(_interactor->inputMask());
}



/*
 * This method removes its associated interactor.
 */
void GoSensor::remove_interactor ()
{
    if (_interactor == NULL)
	return;

    if (_interactor->state() == GITOR_RUNNING)
	_interactor->cancel();

    unsigned int mask = _interactor->inputMask();

    delete _interactor;
    _interactor = NULL;
    GoUI* goui = GoUIFrom (root());

    if (goui != NULL)
	goui->removeInputMask (mask);
}


void GoSensor::bindActions (const GoMetaclass* mclass)
{
    for (; mclass != NULL; mclass = mclass->superclass())
    {
	const GActionRec* actions = mclass->actions();
	if (actions == NULL)
	    continue;
	
	for (const GActionRec* pA = actions; pA->name != NULL; pA++)
	{
	    for (const char* s = pA->name;
		 *s && !isspace(*s) && *s != '(';
		 s++);
	    int nC = s - pA->name;
	    char methodName[80];
	    assert (80 >= nC + 1);
	    strncpy (methodName, pA->name, nC);
	    methodName[nC] = 0;
	    char procName[128];
	    assert (128 >= strlen (mclass->classname()) + 2 + nC + 1);
	    sprintf (procName, "%s::%s", mclass->classname(), methodName);

	    if (pA == actions && GBinded (procName))
		return;

	    GBind (procName, pA->proc, "GFunction");
	}
    }
}


void parent_changed (void*, Go* go, Go* /* old_parent */, Go*)
{
    GoSensor* sor = (GoSensor*) go;

    sor->remove_interactor();

    if (sor->parent())
    {
	GoSensor::bindActions (sor->parent()->getMetaclass());
	sor->create_interactor();
    }
}


GBool gosensor_add_child_precond
    (void*, const Go*, const Go*, const Go*(*))
{
    return GFALSE;
}


GBool gosensor_relcs_precond
    (void*, const Go*,
     GCoordSys* newrelcs, const GCoordSys*, const GCoordSys*)
{
    GCoordSys unit_cs;
    *newrelcs = unit_cs;
    return GFALSE;
}



/*
 *-----------------------------
 * GoSensorMetaclass methods
 *-----------------------------
 */

/*
 This global variable is used to store the unique instance
 of the GoSensorMetaclass
 */
static GoSensorMetaclass* Sensor_metaclass = NULL;


void sensor_destroyed (void*, const Go* go)
{
    GoSensor* sor = (GoSensor*) go;

    if (sor->_interactor)
    {
	if (sor->_interactor->state() == GITOR_RUNNING)
	    sor->_interactor->cancel();

	if (sor->_postcond != NULL)
	    (*sor->_postcond) (sor->_closure, sor);

	sor->remove_interactor();
    }
}


/*
 * The GoSensorMetaclass constructor.
 */
GoSensorMetaclass::GoSensorMetaclass ()
{
}


/*
 * The GoSensorMetaclass destructor.
 */
GoSensorMetaclass::~GoSensorMetaclass ()
{
    /* If we are deleting the unique GoSensorMetaclass instance
       (and not some subclass), we must remove ``this'' from the
       set of subclasses in the superclass, and set NULL to the
       global pointer ``Sensor_metaclass''.
       Quid about GoSensor instances ? should we delete them ?
       */
    if (this == Sensor_metaclass)
    {
        superclass()->RemoveSubclass(this);
	Sensor_metaclass = NULL;
    }
}


/*
 * The virtual method ``m_classname'' returns the name (string) of
 * this class, i.e. "GoSensor".
 */
const char* GoSensorMetaclass::m_classname () const
{
    return "GoSensor";
}


/*
 * The virtual method ``m_superclass'' returns the unique instance of
 * the superclass of this GoSensorMetaclass class.
 */
GoMetaclass* GoSensorMetaclass::m_superclass () const
{
    return Go::metaclass();
}


/* This static method must be used to create a new GoSensor. */
GoSensor* GoSensor::create
    (const char* program, Go* parent, const GEvent* startEvent,
     void (*postcond)(void*, GoSensor*), void* closure)
{
    GoSensor* sor = GoSensor::create();
    sor->program(program);
    sor->start_event = startEvent;
    sor->_postcond = postcond;
    sor->_closure = closure;
    sor->reparent(parent); //should be last, because it may start an interactor;
    return sor;
}


void create_sensor_action (const char* program, Go* parent, const GEvent* startEvent)
{
    if (program != NULL && parent != NULL)
	(void) GoSensor::create (program, parent, startEvent);
}

/*
 * The action table of GoSensor objects.
 * Functions inside this table will be accessible through
 * their name (i.e. a string). This is usefull to implement some
 * interpreted languages dealing with graphic objects.
 */
static GActionRec Sensor_actions[] = {
    {"create(program, parent, event)", (GActionProc)create_sensor_action},
    {NULL, NULL} //this must be the last line;
};


/*
 * This method returns the action table of GoSensor objects.
 */
const GActionRec* GoSensorMetaclass::m_actions () const
{
    return Sensor_actions;
}


/*
 * This method creates and returns a new GoSensor instance.
 * This allows the creation of a new instance of an object, having only
 * the name of the object class.
 */
Go* GoSensorMetaclass::m_create () const
{
    return new GoSensor;
}



/*
 *------------------------------------
 * GoSensor methods (derived from Go)
 *------------------------------------
 */


/*
 * This static method returns the unique GoSensorMetaclass instance.
 * If this instance does not already exist, it is created. In such a case
 * the static global variable ``Sensor_metaclass'' is set up and this
 * metaclass instance is added to the set of subclass instances of the
 * superclass.
 */
GoSensorMetaclass* GoSensor::metaclass ()
{
    if (Sensor_metaclass == NULL)
    {
	Sensor_metaclass = new GoSensorMetaclass;
        Sensor_metaclass->superclass()->AddSubclass(Sensor_metaclass);

	/* When a sensor is reparented we must remove its associated
	   interactor */
	Sensor_metaclass->reparentPostcond (&parent_changed, NULL);
	Sensor_metaclass->destroyPostcond (&sensor_destroyed, NULL);

      GoStream::bind(GoSensor::metaclass()->classname(),
		     (int (*) (const void*, GoStream*)) GoSensor::save,
		     (int (*) (void**, GoStream*)) GoSensor::restore);
	GoSensor::bindActions (Sensor_metaclass);
    }

    return Sensor_metaclass;
}


/*
 * This is the GoSensor constructor.
 */
GoSensor::GoSensor ()
{
    _program = NULL;
    _interactor = NULL;
    _parent = NULL;
    _next = NULL;
    start_event = NULL;
    _postcond = NULL;
    _closure = NULL;
    _active = GTRUE;

    crossType (GCROSS_AS_NOTHING);

    /* A basic GoSensor cannot have children */
    addChildPrecond(&gosensor_add_child_precond, NULL);
    /* A basic GoSensor cannot change its parent relative coordinate system */
    relcsPrecond(&gosensor_relcs_precond, NULL);
}


/*
 * This is the GoSensor destructor.
 */
GoSensor::~GoSensor ()
{
    if (_interactor)
    {
	GWarning("GoSensor::~GoSensor(): an interactor still exists !");
	remove_interactor();
    }
    GFREE(_program);
}


/*
 * This virtual method returns GTRUE if ``p'' is inside the drawing of this
 * object. With an accepted distance ``maxd''.
 *
 * For the basic GoSensor class, the drawing of its parent is taken.
 */
GBool GoSensor::m_contains (const GPoint& p, GNumber maxd) const
{
    return
	(_parent != NULL)
	    ? _parent->contains(p, maxd)
		: GFALSE;
}


/*
 * This private virtual method must copy its own class private data
 * from ``src'' to ``this''.
 * This method is used by the ``dup(go)'' method. The dup method returns a
 * copy of the ``go'' subtree.
 *
 * If dup was called on ``this'', then this next brother must not be dupped,
 * ``donext'' is then FALSE.
 * If dup was called on some ancestor of ``this'', then this next brother
 * must also be dupped. The flag ``donext'' is TRUE. In that case, IF THIS
 * CLASS stores a ``_next'' field, it must call the recursive Dup on the
 * next brother.
 */
void GoSensor::m_dup
    (const Go* src, Go* parcopy, Go* prevcopy, GBool donext, GBool dochildren)
{
      Go::m_dup(src, parcopy, prevcopy, donext, dochildren);

      const GoSensor* orig = (GoSensor*)src;

      _parent = parcopy;

      if (orig->_program != NULL)
	  _program = StringCopy(orig->_program);
      _active = orig->_active;

    if (donext)
	for (Go* orig_nxt = orig->_next; orig_nxt; orig_nxt = orig_nxt->Next())
	    if (_next = orig_nxt->Dup(parcopy, this, GTRUE, dochildren))
		break;
}
	

/*
 * This private virtual method returns the metaclass of ``this''.
 */
GoMetaclass* GoSensor::m_metaclass () const
{
    return GoSensor::metaclass();
}


/*
 * This is a private virtual ``data storage'' method.
 * It returns a pointer to the GAssoc associated to ``this''.
 */
GAssoc* GoSensor::d_get_assoc () const
{
    const GAssoc* assoc =&_assoc;
    return (GAssoc*)assoc;
}


/*
 * This is a private virtual ``data storage'' method.
 * It returns the parent of ``this''.
 */
Go* GoSensor::d_get_parent () const
{
    return _parent;
}


/*
 * This is a private virtual ``data storage'' method.
 * It returns the previous brother of ``this''.
 */
Go* GoSensor::d_get_previous () const
{
    return (_parent)
	? _parent->Child(_parent->childIndex(this) - 1)
	    : NULL;
}


/*
 * This is a private virtual ``data storage'' method.
 * It returns the next brother of ``this''.
 */
Go* GoSensor::d_get_next () const
{
    return _next;
}


/*
 * This is a private virtual ``data storage'' method.
 * A basic GoSensor cannot have children.
 */
Go* GoSensor::d_get_first_child () const
{
    return NULL;
}


/*
 * This is a private virtual "data storage" method.
 * It returns TRUE if this class stores an array of children.
 *
 * (Experience has shown us, that if a class stores a children
 * array, it will have good performances only if this class has
 * only some specific children which store their index into the
 * parent array).
 */
GBool GoSensor::d_store_children_array () const
{
    return GFALSE;
}


/*
 * This is a private virtual "data storage" method.
 * It sets the "previous brother" field (if any).
 */
void GoSensor::d_set_previous (const Go*)
{
}


/*
 * This is a private virtual ``data storage'' method.
 * It sets the ``next brother'' field (if any).
 */
void GoSensor::d_set_next (const Go* go)
{
    _next = (Go*)go;
}


/*
 * This is a private virtual ``data storage'' method.
 * A basic GoSensor cannot have children.
 */
void GoSensor::d_remove_child (const Go*, int)
{
    GError("GoSensor::d_remove_child() called.");
}


/*
 * This is a private virtual ``data storage'' method.
 * It adds a new child to the data structure used to store children.
 */
void GoSensor::d_add_child (const Go*, int)
{
    GError("GoSensor::d_add_child() called");
}


/*
 * This is a private virtual ``data storage'' method.
 * It must reset tree informations (i.e. the parent, previous and
 * next brothers).
 */
void GoSensor::d_set_tree
    (const Go* parent, const Go* /*previous*/, const Go* next)
{
    _parent = (Go*)parent;
    _next = (Go*)next;
}


/*
 * This is a private virtual ``data storage'' method.
 * It returns TRUE if this class stores an absolute coordinate system.
 */
GBool GoSensor::d_store_abscs () const
{
    return GFALSE;
}


/*
 * This is a private virtual ``data storage'' method.
 * It returns TRUE if this class stores a parent relative
 * coordinate system.
 */
GBool GoSensor::d_store_relcs () const
{
    return GTRUE;
}


/*
 * This is a private virtual ``data storage'' method.
 * It returns the parent relative coordinate system.
 * The Go kernel will never call this method if ``d_store_relcs()''
 * returns FALSE (you do not have to implement it then).
 */
const GCoordSys& GoSensor::d_get_relcs
    (const GCoordSys* /*parent_abscs*/) const
{
    STATIC_OBJECT(GCoordSys, unit_cs);
    return unit_cs;
}


/*
 * This is a private virtual ``data storage'' method.
 * It must set up data used to retrieve the coordinate system.
 */
void GoSensor::d_set_coordsys
    (const GCoordSys& /*relcs*/, const GCoordSys& /*abscs*/,
     const GCoordSys& /*parent_abscs*/)
{
}


/*
 * This is a private virtual ``data storage'' method.
 * It returns a pointer to the stored clip of this object.
 *
 * The stored clip, if it exists, is used by the Go kernel to
 * store the clip of this go subtree. Therefore it will not recompute it
 * each time it is needed. On the other hand, storing the go subtree clip
 * is space consuming.
 *
 * If you decide not to store the go subtree clip, you do not have to
 * implement this method.
 */
GRegion* GoSensor::d_get_clip () const
{
    return NULL;
}


/*
 * This static method is responsible to save ``this'' go on ``stream''.
 * It must write its private data and then call the save of its superclass.
 */
int GoSensor::save (const GoSensor* go, GoStream* stream)
{
    return
	stream->writeString (go->_program, "program") &&
	stream->writeBool (go->_active, "active") &&
	Go::save(go, stream);
}


/*
 * This static method is responsible to restore ``this'' go from a stream.
 * For that purpose, it reads its private data and then calls the
 * readParams of its superclass.
 */
int GoSensor::readParams (GoSensor* go, GoStream* stream)
{
    return
	stream->readString (0, (char**)&go->_program) &&
	    stream->readBool (&go->_active) &&
		Go::readParams (go, stream);
}

/**
FUNCTION
    int restore();
DESCRIPTION
    
**/
int GoSensor::restore(GoSensor** sensor, GoStream* gostream) {
    if (*sensor = GoSensor::create())
	return GoSensor::readParams(*sensor, gostream);
    return GFALSE;
}


GBool GoSensor::is_sensor () const
{
    return GTRUE;
}
