#ifdef REV_INFO
static char rcsid[] = "$Id: Translator.cc,v 1.3 1992/12/30 09:51:52 song Exp $";
#endif
/*****************************************************************************
MODULE:  Go Kernel (User Interface)
FILE  :  Translator.cc  
AUTHOR:  Serge Mbikina, Olivier Jojic, Philippe Smadja.
DATE  :  July 92

DESCRIPTION:
  See the Translator.h header.

COPYRIGHT:
  Copyright (c) 1990, 1991, 1992 Bull S.A., Paris, France
 
  Permission to use, copy, modify, distribute FREE OF CHARGE this software
  and its documentation for any purpose is hereby granted without fee, provided
  that (i) the above copyright notices and this permission notice appear in
  all copies of the software and related documentation, and (ii) the name of
  Bull may not be used in any advertising or publicity relating to the
  software without the specific, prior written permission of Bull.
  
  The permission to distribute or sublicense FOR CHARGE either this software
  (all or part) or products derived from this software is subject to prior
  licence from Bull under separate agreement.

  The information in this software is subject to change without
  notice and should not be construed as a commitment by Bull.

  Bull assumes no responsibility for the use or inability to use any of its
  software. The software is provided "AS IS" without warranty of any
  kind, and Bull expressly disclaims all implied warranties, including
  but not limited to the implied warranties of merchantability and
  fitness for a particular purpose.

  In no event shall Bull be liable for any special, incidental, indirect
  or consequential damages of any kind, or any damages whatsoever
  resulting from loss of use, data or profits, whether or not advised of
  the possibility of damage, and on any theory of liability, arising out
  of or in connection with the use or performance of this software.

*****************************************************************************/


#include <assert.h>
#include <Go/Tools/Array.h>
#include <Go/Translator.h>
#include <Go/GoInterf.h>
#include <Go/Sensor.h>
#include <Go/GoUI.h>
#include <Go/GoRoot.h>
#include <Go/Tools/Error.h>
#include <Go/Tools/Color.h>
#include <Go/Tools/Queue.h>
#include <Go/Tools/Bind.h>
#include <Go/mGo.h>
    
#include <stdio.h>
#include <ctype.h>


static inline char* StringCopy (const char* s)
{
    char* d = (char*)GMALLOC(1 + strlen(s));
    strcpy(d, s);
    return d;
}


static GBool anErrorOccured;
static GBool checkOnly;
static char checkErrorMsg[256];



/*
 *-------------------------------------------------------
 * The Handler class
 *-------------------------------------------------------
 */
class Handler
{
  private:
    char*    _functionCall;
    char*    _eventMask   ;
    char*    _methodName  ;
    struct EvInfo
    {
	short type;
	GBool testButton     : 1,
              testKey        : 1,
              testFunctionKey: 1,
	      testModifiers  : 1,
	      testMessage    : 1;
	union {
	    uchar    button;
	    GQueue*  keyList;
	    int      functionKey;
	    char     *message;
	} detail;
	GQueue* modifiersList;
    } evinfo;

    GQueue* params;
    void* func;

    GBool   parseEvent(const char* event);
    GBool   parseMethod(char*, char*);
    GBool   ParseCharEvent(const char *&s,const char *event);
    GBool   ParseFunctionEvent(const char *&s);
    GBool   ParseButtonEvent(const char *&s);
    GBool   ParseMessageEvent(const char *&s);
    GBool   ParseNoDetail(const char *&s);
    GBool   ParseModifiers(const char *&s);
    GBool   ParseEvent (const char* event);

  public:
    Handler ();
    Handler (const Handler&);
    ~Handler ();
    GBool Parse(const char* event, const char* method, const char* parameters);
    inline  int    type()             { return evinfo.type;        }
    inline  char   *getFunctionCall() { return _functionCall;      }
    inline  char   *getEventMask()    { return _eventMask   ;      }
    GBool   match (const GEvent*) const;
    void    call(Go*, const GEvent*, GTranslator*) const;
};




// warning the 3 following structures are copied in CallGoUIProc.c
// do not forget to modify the corresponding struct at any change;
enum ParamType {
    ParamUndef,  ParamLong,  ParamInputChar,ParamNumber,
    ParamPointer,ParamChar,  ParamGo,       ParamGoAtPos,
    ParamGoUI,   ParamSensor,ParamEvent,    ParamPos,
    ParamColor,  ParamString 
};



union ParamValue
{
    char c;
    long l;
    GNumber n;
    void* p;
    inline ParamValue() {};
    inline ParamValue(char par) { c = par;};
    inline ParamValue(long par) { l = par;};
    inline ParamValue(GNumber par) { n = par;};
    inline ParamValue(void* par) { p = par;};
};
	


struct ParamRec
{
    ParamType type;
    ParamValue value;
    
    ParamRec ()   { type = ParamUndef; }
    ParamRec (ParamType tp) { type = tp; }
    ParamRec (ParamType tp, ParamValue v) { type = tp; value = v; }
    ParamRec (const ParamRec& pr) { type = pr.type; value = pr.value; }
};
// end of copy;



struct CharBuf
{
    char cmin, cmax;
    CharBuf(char a, char b) { cmin = a; cmax = b; }
    CharBuf(const CharBuf& cb) { cmin = cb.cmin; cmax = cb.cmax; }
};




static GBool _parse_handler_line (const char* line);
static GBool _parse_class_sensor_names (const char*& line,
					char *translator_name,
					char* class_name, char* sensor_name);
static GBool _parse_event_info (const char*& line, char* evinfo_string);
static GBool _parse_function_call (const char*& line, char* function_call);
static GBool _parse_method_name (const char*& line, char* method_name);
static GBool _parse_params (const char*& line, char* param_string);
static GBool  PeekChar (const char* string, const char* event,char &ch);
static GBool  GetChar (const char*& s, const char* event,char &c);
static GBool  ParseKey (const char*& s, const char* event,GQueue &q);
static GBool  modifierExist(GQueue &q,uchar c);
static GBool  read_ident (const char*& line,char* ident) ;
static GBool  read_string(const char*& line,char *string);
static GBool  is_ident(char *string);
static GBool  is_string(char *str)  ;
static GBool  is_long(char *string) ;
static GBool  is_float(char *string);
static GBool  read_method_name(char *&s,char*&goclass,char* &function);
static GBool  read_next_parameter(char *&s,char *&param);




/*  Global Variables  */
const char*  file_name;
int    line_number;
GBool  bNoFctCall ;


GPointerArray*  GTranslatorPgm::_instances;

GTranslatorPgm::GTranslatorPgm (const char* name) : GInteractorPgm(name)
{
    if (_instances == NULL)
	_instances = new GPointerArray();
    _instances->add(this);
}


GTranslatorPgm::~GTranslatorPgm ()
{
    while (!_handlers.empty())
    {
	Handler* h = (Handler*)_handlers.getfirst();
	delete h;
    }
    _instances->rem(this);
    if (_instances->nelems() == 0)
    {
	delete _instances;
	_instances = NULL;
    }
}


GTranslator::GTranslator (const GTranslatorPgm* pgm, Go* go) : GInteractor(pgm, go)
{
}


GBool GTranslatorPgm::remove (const char* eventMask, const char *functionCall)
{
    checkOnly = GFALSE;
    anErrorOccured = GFALSE;

    GQueue  *q = new GQueue;
    Handler *h;
    GQueue  *listHandlers = &_handlers;
    while(! (listHandlers->empty()))
    {
	h =  (Handler*)listHandlers->getfirst();
	if  ((strcmp(h->getEventMask(),eventMask)==0))
	{
	    if (functionCall==NULL)
		delete h ;
	    else if ((strcmp(h->getFunctionCall(),functionCall)==0))
	    { delete h ; break ; }
	    else  q->putlast(h);	   
	}
	else  q->putlast(h);	   
    }
    if (!q->empty())
	listHandlers->putfirst(*q);
    delete q ;
    return GTRUE ;
}




GBool GTranslatorPgm::add (const char *eventMask,const char *functionCall)
{
    checkOnly = GFALSE;
    anErrorOccured = GFALSE;

    char  method_name[128] ;
    char  param_string[256];

    if (!_parse_method_name(functionCall, method_name))
	return GFALSE;
    if (!_parse_params(functionCall,param_string))
	return GFALSE;
    Handler *h = new Handler ;

    if (h->Parse(eventMask,method_name,param_string))
    {
	this->_handlers.putlast(h);
    }
    else
    {
	delete h ;
	return GFALSE;
    }
    return GTRUE ;
}




GBool GLoadTranslatorPgms(const char* filename)
{
    checkOnly = GFALSE;
    anErrorOccured = GFALSE;

    file_name = filename;

    FILE* file = fopen(filename, "r");
    if (file == NULL)
    {
	GWarning("cannot open handlers file %s", filename);
	return GFALSE;
    }
    
    line_number = 0;

    while (!feof(file))
    {
	char buf[512];
	(void)fgets(buf, 512, file);
	line_number++;
	_parse_handler_line(buf);
    }

    fclose(file);
	
    return GTRUE;
}

    
    
    
#define error(string)\
if (!checkOnly)\
{ GWarning("Translator::parsing handler file %s on line %d %s\n",\
	   file_name, line_number, string) ; return GFALSE ; }\
else\
{ anErrorOccured = GTRUE;\
  strcpy (checkErrorMsg, string);\
  return GFALSE; }


GBool _parse_handler_line (const char* line)
{
    char  translator_name[256];
    char  class_name[128];
    char  sensor_name[128];
    char  evinfo_string[128];
    char  function_call[512];
    char  include_word[256];
    char  include_file[256];
    GBool bStar = GFALSE   ;
    GBool bNew  = GFALSE   ;
    
    bNoFctCall  = GFALSE   ;
    for (; isspace(*line); line++);
    if (*line == 0 )
	return GTRUE;

    if (*line == '#')
    {
	if (!read_ident(line,include_word)) return GTRUE; //comment
	if (strcmp(include_word,"include")==0)
	{
	    if (!read_string(line,include_file))
		error("bad include file name");
	    const char *old_file_name = file_name ;
	    int old_line_number = line_number;
	    if (!checkOnly)
		GLoadTranslatorPgms(include_file)   ;
	    else
	    {
		int notUsed;
		GTranslatorPgm::checkFileSyntax (include_file, notUsed);
		if (anErrorOccured)
		{
		    char msg[512];
		    sprintf (msg, "(in included file \"%s\") %s", include_file, checkErrorMsg);
		    strcpy (checkErrorMsg, msg);
		    return GFALSE;
		}
	    }
	    
	    file_name   = old_file_name      ;
	    line_number = old_line_number    ;
	}
	return GTRUE;
    }
    if (*line == '*')
    {
	line++ ;
	bStar  = GTRUE ;
    }

    if (!_parse_class_sensor_names(line,translator_name,class_name, sensor_name))
	return GFALSE ;
    if (!_parse_event_info(line, evinfo_string))
	return GFALSE ;
    if (bStar)
    {
	for (; *line && isspace(*line); line++);
	if (*line) {
	    if (!_parse_function_call(line,function_call))
		return GFALSE ;
	}
	else bNoFctCall= GTRUE ;
    }    
    else if (!_parse_function_call(line,function_call))
	return GFALSE;
    
    for (; isspace(*line); line++);
    if (!(*line == 0 || *line == '#'))
	error("extra character at end of line");
    
    if (checkOnly)
    {
	GTranslatorPgm::checkSyntax (evinfo_string, function_call);
	return !anErrorOccured;
    }
    
    GTranslatorPgm *trpgm = (GTranslatorPgm*)
	GInteractorPgmFromName(translator_name);
    if (trpgm==NULL)
    {
	trpgm = GTranslatorPgm::create (translator_name);
	bNew = GTRUE ;
    }
    GBool bRes = GFALSE;
    if (bStar)
    {
	if (bNoFctCall)
	    bRes = trpgm->remove(evinfo_string,NULL);
	else
	    bRes = trpgm->remove(evinfo_string,function_call);
    }
    else  bRes = trpgm->add(evinfo_string,function_call);
    if ((!bRes) && bNew)
        delete trpgm;
    return GTRUE ;
}




GBool _parse_class_sensor_names (const char*& line,char *translator_name,char* class_name, char* sensor_name)
{
    for (; isspace(*line); line++);
    if (!isalpha(*line))
	error("bad first non space character");
    
    const char* start = line;
    for (line++; isalnum(*line); line++);
    int count = line - start;
    memcpy(sensor_name, start, count);
    sensor_name[count] = 0;
    strcpy(translator_name,sensor_name);
    
    for (; isspace(*line); line++);
    if (*line == '.')
    {
	strcpy(class_name, sensor_name);
	for (line++; isspace(*line); line++);
	if (!isalpha(*line))
	    error("bad character after '.'");
	start = line;
	for (line++; isalnum(*line); line++);
	count = line - start;
	memcpy(sensor_name, start, count);
	sensor_name[count] = 0;
	strcat(translator_name,".");
	strcat(translator_name,sensor_name);
    }
    return GTRUE ;
}


GBool _parse_event_info (const char*& line, char* evinfo_string)
{
    for (; isspace(*line); line++);
    const char* start = line;
    if (*line == '*')
	line++;
    if (!isalpha(*line))
	error("bad first event info character");
    for (; *line != 0 && !isspace(*line); line++)
	if (*line == '\\')
	    line++;
    int count = line - start;
    memcpy(evinfo_string, start, count);
    evinfo_string[count] = 0;
    return GTRUE ;
}




GBool _parse_function_call (const char*& line, char* function_call)
{
    for (; isspace(*line) ; line++);
    const char* start = line;
    for (line++; *line && (*line !='#') ; line++);
    int count = line - start;
    memcpy(function_call, start, count);
    function_call[count] = 0;
    return GTRUE ;
}




GBool _parse_method_name (const char*& line, char* method_name)
{
    for (; isspace(*line) ; line++);
    if (!isalnum(*line))
	error("bad method first character");
    const char* start = line;
    for (line++; isalnum(*line); line++);
    int count = line - start;
    memcpy(method_name, start, count);
    method_name[count] = 0;
    for (; isspace(*line); line++);
    if (*line == ':')
    {
	if (*++line != ':')
	    error("there must be two semicolons '::'");
	for (line++; isspace(*line); line++);
	start = line;
	if (!isalpha(*line))
	    error("bad first method character");
	for (line++; isalnum(*line); line++);
	int new_count = line - start;
	method_name[count] = method_name[count+1] = ':';
	memcpy(method_name + count + 2, start, new_count);
	method_name[count + 2 + new_count] = 0;
    }
    return GTRUE ;
}




GBool _parse_params (const char*& line, char* param_string)
{
    for (; isspace(*line); line++);
    if (*line != '(')
	error("'(' is missing");
	*param_string++ = '(';
			    
    const char* start;
    int count;

    for (line++; isspace(*line); line++);
    if (*line == ')')
	goto end_label;

    start_label:
	if ( !isalnum(*line) && *line != '"' && *line != '-')
	    error("bad param's first character");
	start = line;
	if (*line != '"')
  	    for (line++; isalnum(*line) || *line == '-' || *line == '.'; line++);
	else {
	    for (line++; *line && *line != '"'; line++);
   	    line++;
        }
	count = line - start;
	memcpy(param_string, start, count);
	param_string += count;
	for (; isspace(*line); line++);
	if (*line == ')')
 	    goto end_label;
        if (*line != ',')
	    error("bad parameter syntax");
        *param_string++ = ',';
        for (line++; isspace(*line); line++);
        goto start_label;

  end_label:
    *param_string++ = ')';
    *param_string = 0;
    line++;

  return GTRUE ;
}



Handler::Handler ()
{
    _functionCall = NULL;
    _eventMask = NULL;
    _methodName = NULL;
    evinfo.type = -1;
    evinfo.testButton =
	evinfo.testKey =
	    evinfo.testFunctionKey =
		evinfo.testModifiers =
		    evinfo.testMessage = GFALSE;
    evinfo.detail.message = NULL;
    evinfo.modifiersList = NULL;
    params = NULL;
    func = NULL;
}


Handler::Handler (const Handler& h)
{
    _functionCall = StringCopy(h._functionCall);
    _eventMask = StringCopy(h._eventMask);
    _methodName = StringCopy(h._methodName);
    evinfo.type = h.evinfo.type;
    evinfo.testButton = h.evinfo.testButton;
    evinfo.testKey = h.evinfo.testKey;
    evinfo.testFunctionKey = h.evinfo.testFunctionKey;
    evinfo.testModifiers = h.evinfo.testModifiers;
    evinfo.testMessage = h.evinfo.testMessage;

    switch (evinfo.type)
    {
      case GBUTTON_PRESS_EVENT:
      case GBUTTON_RELEASE_EVENT:
      case GDOUBLE_CLICK_EVENT:
      case GTRIPLE_CLICK_EVENT:
	evinfo.detail.button = h.evinfo.detail.button;
	break;
      case GCHAR_EVENT: {
	evinfo.detail.keyList = new GQueue;
	int nCB;
	CharBuf** pCB = (CharBuf**)
	    h.evinfo.detail.keyList->arrayFromFirst(nCB);
	for (int i = 0; i < nCB; i++)
	    evinfo.detail.keyList->putlast(new CharBuf(*pCB[i]));
	}
	break;
      case GFUNCTION_KEY_EVENT:
	evinfo.detail.functionKey = h.evinfo.detail.functionKey;
	break;
      case GMESSAGE_EVENT:
	evinfo.detail.message = StringCopy(h.evinfo.detail.message);
	break;
    }

    evinfo.modifiersList = NULL;
    if (evinfo.testModifiers && h.evinfo.modifiersList != NULL)
    {
	evinfo.modifiersList = new GQueue;
	int nM;
	uchar** pM = (uchar**)h.evinfo.modifiersList->arrayFromFirst(nM);
	for (int i = 0; i < nM; i++)
	    evinfo.modifiersList->putlast(new uchar(*pM[i]));
    }

    if (h.params == NULL)
	params = NULL;
    else
    {
	params = new GQueue;
	int nP;
	ParamRec** pP = (ParamRec**)h.params->arrayFromFirst(nP);
	for (int i = 0; i < nP; i++)
	    params->putlast(new ParamRec(*pP[i]));
    }

    func = h.func;
}


GBool Handler::Parse(const char* event, const char* method, const char* parameters)
{
    int n = 1 + strlen(method);
    this->_methodName   = (char*)GMALLOC(n);
    this->_functionCall = new char[(n + strlen(parameters))];
    this->_eventMask    = new char[(strlen(event)+1)];
    strcpy(this->_methodName, method);
    strcpy(this->_eventMask , event );
    strcpy(this->_functionCall,method);
    strcat(this->_functionCall,parameters);
    
    this->evinfo.testButton      = GFALSE;
    this->evinfo.testFunctionKey = GFALSE;
    this->evinfo.testKey         = GFALSE;
    this->evinfo.testModifiers   = GFALSE;
    this->evinfo.detail.keyList  = NULL;
    // this->procRec.procedure   = NULL;

    if (!this->ParseEvent(event))  return GFALSE;

    if (bNoFctCall==GFALSE)
    {
#ifndef NO_ALLOCA
    char* meth = (char*)alloca(strlen(method)+1);
    char* param = (char*)alloca(strlen(parameters)+1);
#else
    char* meth = (char*)GMALLOC(strlen(method)+1);
    char* param = (char*)GMALLOC(strlen(parameters)+1);
#endif
    strcpy(meth, method);
    strcpy(param, parameters);
    
    GBool bRes = this->parseMethod(meth, param) ;
    
#ifdef NO_ALLOCA
    GFREE(meth);
    GFREE(param);
#endif
    return bRes;
    }
 return GTRUE ;
}




Handler::~Handler ()
{
    GFREE(_functionCall);
    GFREE(_eventMask);
    GFREE(_methodName);

    GQueue *q = NULL;
    int count = 0 ;
    
    if (this->evinfo.testKey)
    {
	q = this->evinfo.detail.keyList;
	uchar **pt = (uchar **)q->arrayFromFirst(count);
	if ((count != 0) && (pt))
	{
	    uchar* t = *pt ;
	    for (int i =0 ; i < count ; i++, t = *(++pt))
		delete t ;
	}
	delete this->evinfo.detail.keyList;
    }

    if (this->evinfo.modifiersList)
    {
	q = this->evinfo.modifiersList;
	uchar **pt = (uchar **)q->arrayFromFirst(count);
	if ((count != 0) && (pt))
	{
	    uchar* t = *pt ;
	    for (int i =0 ; i < count ; i++, t = *(++pt))
		delete t ;
	}
	delete this->evinfo.modifiersList;
    }

    if (this->params)
    {
	q = this->params;
	ParamRec **pt = (ParamRec **)q->arrayFromFirst(count);
	if ((count != 0) && (pt))
	{
	    ParamRec* t = *pt;
	    for (int i =0 ; i < count ; i++, t = *(++pt))
		delete t ;
	}
	delete this->params  ;
    } 
}





static int charSize;

inline int XDigit (char c)
{
    if (isdigit(c)) return c - '0';
    else return c - 'A' + 10;
}




inline GBool myisxdigit (char c)
{
    return isdigit(c) || c >= 'A' && c <= 'F';
}




static GBool PeekChar (const char* string, const char* /* event */,
		       char &ch)
{
    const char* s = string;
    if      (*s == '>')  { charSize = 0; return GTRUE; }
    else if (*s == '^') {
	if (islower(s[1])) {
	    ch = s[1] - 'a' + 1;
	    charSize = 2;
	} else if (isupper(s[1])) {
	    ch = s[1] - 'A' + 1;
	    charSize = 2;
	} else {
	    ch = s[0];
	    charSize = 1;
	}
    }
    else if (*s == '\\') {
	if      (s[1] == '\\') { charSize = 2; ch = '\\';  }
	else if (s[1] == '-')  { charSize = 2; ch =  '-' ; }
	else if (s[1] == '>')  { charSize = 2; ch =  '>' ; }
	else if (s[1] == '^') { charSize = 2; ch = '^'; }
	else if (myisxdigit(s[1]) && myisxdigit(s[2]))
	{
	    charSize = 3;
	    ch = 16*XDigit(s[1]) + XDigit(s[2]);
	}
	else error("bad \\ key syntax"); 
    }
    else
    {
	charSize = 1;
	ch = *s;
    }
    return GTRUE;
}




static GBool GetChar (const char*& s, const char* event,char &c)
{
    if (!PeekChar(s, event,c))
	return GFALSE ;
    if (charSize == 0)
	error("something is missing before '>'");
    s += charSize;
    return GTRUE;
}




GBool ParseKey (const char*& s, const char* event,GQueue &q)
{
    //<x> <xxx> <x-x>;
    //with x: a \- \1F >;
    while ( *s != '>') {
	char c ;
	if (!GetChar(s,event,c))
	    return GFALSE ;
	char c2 = c;
	char nextc = 0;
	if (!PeekChar(s, event,nextc))
	    return GFALSE ;
	if ( (nextc == '-') && (*s != '\\'))
	{
	    char ch;
	    if (!GetChar(s,event,ch))
		return GFALSE;
	    if (!GetChar(s, event,c2))
		return GFALSE;
	    if (c2 < c)
		error("bad <x-y> key syntax");
	}
	q.putlast(new CharBuf(c, c2));
    }
    return GTRUE;
}




#define checkString(str,tp)\
    (!strncmp(s,str,-1+sizeof(str)))\
    {\
	 this->evinfo.type = tp;\
	 s += -1+sizeof(str);\
    }




#define check(name, value)\
    if (!memcmp(s,name,sizeof(name)-1)) {\
        s += sizeof(name)-1;\
        this->evinfo.detail.functionKey = value;\
    }



GBool Handler::ParseCharEvent(const char *&s,const char *event)
{
    this->evinfo.detail.keyList = new GQueue ;
    if (*s++ != '<')
	error("bad detail syntax");
    this->evinfo.testKey        = GTRUE;
//  this->evinfo.testModifiers  = GFALSE;
    if (!ParseKey(s,event,*this->evinfo.detail.keyList))
	return GFALSE;
    
    if (*s++ != '>')
	error("bad detail syntax");
    return GTRUE ;
}




GBool Handler::ParseFunctionEvent(const char *&s)
{
    if (*s++ != '<')
	error("bad detail syntax");
    this->evinfo.testFunctionKey = GTRUE;
    switch (s[0])
    {
      case 'u':
	check("up", GKEY_UP)
 		else goto err;
	break;
      case 'd':
	check("down", GKEY_DOWN)
		else check("delete", GKEY_DELETE)
		else goto err;
	break;
      case 'l':
	check("left", GKEY_LEFT)
		else goto err;
	break;
      case 'r':
	check("right", GKEY_RIGHT)
		else goto err;
	break;
      case 'i':
	check("insert", GKEY_INSERT)
		else goto err;
	break;
      case 'h':
	check("home", GKEY_HOME)
		else goto err;
	break;
      case 'e':
	check("end", GKEY_END)
		else goto err;
	break;
      case 'p':
	check("pageup", GKEY_PAGE_UP)
		else check("pagedown", GKEY_PAGE_DOWN)
		else goto err;
	break;
      case 'f':
	        check("f1", GKEY_F1)
		else check("f2", GKEY_F2)
		else check("f3", GKEY_F3)
		else check("f4", GKEY_F4)
		else check("f5", GKEY_F5)
		else check("f6", GKEY_F6)
		else check("f7", GKEY_F7)
		else check("f8", GKEY_F8)
		else check("f9", GKEY_F9)
		else check("f10", GKEY_F10)
		else goto err;
	break;
      default:
      err:
	error("bad function key");
    }
    if (*s++ != '>')
	error("bad detail syntax");
    return GTRUE ;
}




// TEST : OK
GBool Handler::ParseButtonEvent(const char *&s)
{
    if (*s++ != '<')
	error("bad detail syntax");    
    this->evinfo.testButton    = GTRUE;

    char c ;
    this->evinfo.detail.button = 0 ;
    for(int i =0 ; i < 8 && isdigit(*s) ; i++)
    {
	c = *s -'0' ;
	if (c < 1 || c > 8)
	    error("bad button");
	this->evinfo.detail.button |= ( 1 << (c - 1) );
	s++ ;
    }
    if (*s++ != '>')
	error("bad detail syntax");
    return GTRUE ;
}




// TEST : OK
GBool Handler::ParseMessageEvent(const char *&s)
{
    if (*s++ != '<')
	error("bad detail syntax");
    this->evinfo.testMessage  = GTRUE;
    const char *start = s ;
    for (s++; isalnum(*s) || *s == '*' || *s == '?' ; s++);
    int count = s - start;
    this->evinfo.detail.message = new char[count + 1];
    memcpy(this->evinfo.detail.message, start,count);
    this->evinfo.detail.message[count] = 0;
    if (*s++ != '>')
	error("bad detail syntax");
    return GTRUE ;
}





GBool Handler::ParseNoDetail(const char *&s)
{
    if (*s++ != '<')
	error("bad detail syntax");
    if (*s++ != '>')
	error("bad detail syntax");
    return GTRUE ;
}




 
GBool modifierExist(GQueue &q,uchar c)
{
    int count = 0 ;
    uchar **pt = (uchar **)q.arrayFromFirst(count);
    if (count == 0)
	return GFALSE ;
    if (pt) {
	uchar* t = *pt;
	for (int i =0 ; i < count ; i++, t = *(++pt))
	    if ( *t== c)
		return GTRUE ;
    }
    return GFALSE ;
}





GBool Handler::ParseModifiers(const char *&s)
{
    //modifiers:
    GBool  flagLast,flagFirst,flagMiddle   ; //to detect an empty modifier;
    flagLast=flagFirst=flagMiddle = GFALSE ; 
    GQueue *queue  = new GQueue  ;
    uchar   modf   = 0, *C ;
    for (; *s && !isspace(*s); s++)
    {
	if   (*s== 'c')
	{
	    modf   |= GCONTROL;
	    flagLast = GFALSE  ;
	}
	else if (*s== 's')
	{
	    modf   |= GSHIFT  ;
	    flagLast = GFALSE  ;
	}
	else if (*s== 'a')
	{
	    modf   |= GALTERNATE;
	    flagLast = GFALSE    ;
	}
	else if (*s=='|')
	{
	    if (*(s-1)=='|')
		flagMiddle= GTRUE ;
	    else if ( (*(s-1)!='c') && (*(s-1)!='a') && (*(s-1)!='s'))
		flagFirst = GTRUE ;
	    flagLast = GTRUE ;

	    if ((*(s-1)=='c')|| (*(s-1)=='a') || (*(s-1)=='s'))
		{
		    if (!modifierExist(*queue,modf))
			{
			    C  = new uchar ;
			   *C  = modf      ;
			    queue->putlast(C) ;
			    modf = 0 ;
			}
		}
	}   
	else  error("bad modifier");
    }
    if ((*(s-1)=='c')|| (*(s-1)=='a') || (*(s-1)=='s'))
	if (!modifierExist(*queue,modf))
	    {
		 C  = new uchar ;
		*C  = modf     ;
		queue->putlast(C) ;
	    }
    if (flagFirst || flagLast || flagMiddle || queue->empty())
    {
	C = new uchar ; *C = GNONE ;
	queue->putlast(C) ;
    }
    this->evinfo.modifiersList = queue ;
    return GTRUE ;
}




/*
 * event syntax:
 * 		press<1>s   //shift
 *		release<2>c //control
 *		key<a>a     //alternate
 *		fun<up>s
 *		motion<>
 */
GBool Handler::ParseEvent (const char* event)
{
    const char* s = event;

    // compatibilite ascendante avec version prec de goui
    if (*s == '*')       
	s++;
    
    if      checkString("press",   GBUTTON_PRESS_EVENT) 
    else if checkString("release", GBUTTON_RELEASE_EVENT)
    else if checkString("double",  GDOUBLE_CLICK_EVENT)  
    else if checkString("triple",  GTRIPLE_CLICK_EVENT)  
    else if checkString("key",     GCHAR_EVENT) 	  
    else if checkString("fun",     GFUNCTION_KEY_EVENT)  
    else if checkString("enter",   GENTER_SENSOR_EVENT)  
    else if checkString("leave",   GLEAVE_SENSOR_EVENT)  
    else if checkString("focusIn", GFOCUS_IN_EVENT)
    else if checkString("focusOut",GFOCUS_OUT_EVENT)
    else if checkString("motion",  GMOTION_EVENT)
    else if checkString("expose",  GEXPOSE_EVENT) 	  
    else if checkString("message", GMESSAGE_EVENT)       
    else    error("bad event type");
    
    if (s == event)
	error("bad event type");

    this->evinfo.testButton = this->evinfo.testKey = GFALSE;
    this->evinfo.testFunctionKey = GFALSE;
    if (this->evinfo.type == GBUTTON_PRESS_EVENT   ||
	this->evinfo.type == GBUTTON_RELEASE_EVENT ||
	this->evinfo.type == GDOUBLE_CLICK_EVENT   ||
	this->evinfo.type == GTRIPLE_CLICK_EVENT   ||
	this->evinfo.type == GCHAR_EVENT           ||
	this->evinfo.type == GFUNCTION_KEY_EVENT   ||
        this->evinfo.type == GMESSAGE_EVENT)
    {
	if (evinfo.type != GMESSAGE_EVENT)
	    evinfo.testModifiers = GTRUE;

	if  (this->evinfo.type == GCHAR_EVENT)
	    {
		if (!this->ParseCharEvent(s,event))
		    return GFALSE ;
	    }
	else if (this->evinfo.type == GFUNCTION_KEY_EVENT)
	     {
		 if (!this->ParseFunctionEvent(s))
		     return GFALSE ;
	     }
	else if (this->evinfo.type == GMESSAGE_EVENT)
	     {
		 if (!this->ParseMessageEvent(s))
		     return GFALSE ;
	     }
	else    //evinfo.type == goButton Press, Release, double, triple Event
	    if (!this->ParseButtonEvent(s))
		return GFALSE; 
    }
    else if (!this->ParseNoDetail(s))
	     return GFALSE;
    if (!this->ParseModifiers(s))
	     return GFALSE ;

    for (; isspace(*s); s++);
    if (*s)
	error("scories after a good event syntax");
    return GTRUE ;
}




#undef checkString
#undef check


static GBool ouvFlag = GFALSE ;
static GBool ferFlag = GFALSE ;



/*
 * --------------------------------------------------------
 * method syntax:
 *		proc(params)
 *		GoClass::proc(params)
 *
 * params:
 *		goui, go
 * --------------------------------------------------------
 */
GBool Handler::parseMethod (char* method, char* parameters)
{
    char*    goclass = NULL;
    char*    function;
    char*    s = method; 
    GQueue*  queue = new GQueue ;  
    read_method_name(s,goclass,function);

    s    = parameters;
    char*  param = s ;
    while (read_next_parameter(s,param))
    {
	ParamRec* tmp;
	if      (is_long(param))
	         tmp = new ParamRec(ParamLong,(long)atoi(param));
	else if (is_float(param))
	        tmp  = new ParamRec(ParamNumber,(GNumber)atof(param));
	else if (is_ident(param))
	{
	    if      (!strcmp("NULL", param))
		    tmp = new ParamRec(ParamPointer,(void*)NULL);
	    else if (!strcmp("GTRUE", param))
		    tmp = new ParamRec(ParamChar,(char)GTRUE);
	    else if (!strcmp("GFALSE", param))
		    tmp = new ParamRec(ParamChar,(char)GFALSE);
	    else if (!strcmp("go", param))
		    tmp = new ParamRec(ParamGo);
	    else if (!strcmp("goAtPos",param))
		    tmp = new ParamRec(ParamGoAtPos);
	    else if (!strcmp("goui", param))
		    tmp = new ParamRec(ParamGoUI);
	    else if (!strcmp("sensor", param))
		    tmp = new ParamRec(ParamSensor);
	    else if (!strcmp("event", param))
		    tmp = new ParamRec(ParamEvent);
	    else if (!strcmp("pos", param))
		    tmp = new ParamRec(ParamPos);
	    else if (!strcmp("char", param))
		    tmp = new ParamRec(ParamInputChar);
	    else if ((!strncmp("COLOR",param,5)) || (!strncmp("color",param,5)))
	    {
		param += 5;
		GShColor* color = GoDefaultGoInterf()->colorFromName(param);
		if (!color) GWarning("Handler.cc: bad color name: %s", param);
		tmp = new ParamRec(ParamColor,color); 
	    }
	    else {ouvFlag=ferFlag = GFALSE ;error("type unknown");}
	}
	else if  (is_string(param))
	{
	    param++;
	    for (char* s = param; *s != '"'; s++);
	    char* str = new char[s - param + 1];
	    memcpy(str, param, s - param);
	    str[s - param] = 0;
	    tmp = new ParamRec(ParamString,str);
	}
        else { ouvFlag = ferFlag = GFALSE ; error("type unknown");}
	queue->putlast(tmp);
    }
    this->params = queue ;
    if (*s && *s != ')')
	error("scories after a good method syntax");
    return GTRUE ;
}



    
GBool  read_ident (const char*& line,char* ident)
{
    for (line++; isspace(*line); line++);
    const char* start = line;
    for (line++; isalnum(*line); line++);
    int count = line - start;
    memcpy(ident, start, count);
    ident[count] = 0;
    return GTRUE ;
}




GBool  read_string(const char*& line,char *string)
{
    for (; isspace(*line); line++);
    if  (*line != '"')
    {
	anErrorOccured = GTRUE;
        strcpy (checkErrorMsg, "missing string");
	return GFALSE ;
    }
    const char *start = ++line;
    for (line++; *line && *line != '"'; line++);
    int count = line - start;
    memcpy(string, start, count);
    string[count] = 0;
    return GTRUE ;
}


GBool  is_ident (char* ident)
{
    char *str = ident ;
    for (; isspace(*str); str++);
    for (; isalnum(*str); str++);
    return !(*str);
}


GBool is_string(char *str)
{
    if (str[0] == '"')
	for (str++; *str ; str++)
	    if (*str != '"')  return GTRUE ;
    return GFALSE ;
}




GBool is_long(char *string)
{
    char *str = string ;
    if (*str == '-')
	str++;
    for (;*str >= '0' && *str <= '9' ; str++);
    return !*str ;
}




GBool is_float(char *string)
{
    GBool floatNumber = GFALSE;
    char* str = string ;
    if (*str == '-')
	str++;
    
    if (*str >= '0' && *str <= '9')
    {
	for (;*str >= '0' && *str <= '9' || *str == '-'||
	      *str == '.' || *str == 'e';str++)
	    if (*str == '.')
		floatNumber = GTRUE;
	if (*str)
	    return GFALSE ;
    }
    if (floatNumber)
	return GTRUE ;
    return GFALSE;
}




GBool read_method_name(char*& s, char*& goclass, char*& function)
{
    char *method = s ;
    for  (; isalnum(*s); s++);
    if   (*s != ':')
    {
	function = method;
	goclass  = NULL  ;
    }
    else
    {
	goclass = method;
	*s++ = 0;
	if (*s++ != ':')
	    error("two semi-colons are needed for goclass methods");
	function = s;
	for (; isalnum(*s); s++);
    }
    if (s == function)
	error("method is missing");
    if (*s != 0)
	error("bad function name");
    return GTRUE ;
}




GBool read_next_parameter(char *&s,char *&param)
{
    if (ouvFlag==GFALSE)
    {
	for (; isspace(*s); s++);
	if (*s != '(')
	    error("'(' is missing")
	else
	{
	    ouvFlag = GTRUE ;
	    *s++ = 0 ;
	    if (*s == ')')
		ferFlag = GTRUE ;
	}
    }
    if (ferFlag==GFALSE)
    {
	for (; isspace(*s); s++);
	if (*s == '"')   // cas string
	{
	    param = s ;
	    for (s++; *s && *s != '"'; s++);
	    s++;
	}
	else for (param=s; isalnum(*s) || *s == '.' || *s == '-'; s++); // cas ident,real,int
	if  (isspace(*s)) *s++ = 0 ;
	for (; isspace(*s); s++)   ;
	if      (*s == ')')
	        { *s++ = 0 ; ferFlag = GTRUE ; }
	else if (*s == ',')
	        { *s++ = 0 ; for (;isspace(*s); s++);}
	else error("bad method's param syntax");
	return GTRUE ;
    }
    ouvFlag = ferFlag = GFALSE ;
    return GFALSE ;
}

#undef error



/*
 *-------------------
 * execution methods
 *-------------------
 */


GInteractor* GTranslatorPgm::start (const GEvent*, Go* go)
{
    GInteractor* itor = new GTranslator (this, go);
    Initialize(itor);
    return itor;
}


GEventMask GTranslatorPgm::inputMask () const
{
    ulong mask = 0;
    int count;
    for (Handler** pH = (Handler**)GetHandlers()->arrayFromFirst(count);
	 count > 0; --count, pH++)
    {
	mask |= (((uint)1) << (*pH)->type());
    }
    
    // append the inputamsk of the super program;
    GTranslatorPgm* super = superTranslatorPgm();
    if (super)
	return mask | super->inputMask();
    return mask;
}

/**
FUNCTION
    GTranslatorPgm* superTranslatorPgm() const
DESCRIPTION
    if The name of the program is something like GoClass.name;
    if return the super translator  GoSuperclass.name (if any).
**/
GTranslatorPgm* GTranslatorPgm::superTranslatorPgm() const
{
    // left part: classname;
    // find right part (<--> method name);
    
    for (const char* right_part = name(); *right_part != '0'; right_part++)
	if (*right_part == '.')
	    break;

    if (*right_part != '.')
	return NULL;
    
    char tmp_name[128];
    uint len = right_part - name();
    if (len >= sizeof(tmp_name))
    {
	GWarning("GTranslatorPgm::superTranslatorPgm(): name too long %s %s %s"
		 "subclass of the program:", name(), " not handled\n");
	return NULL;
    }

    // copy classname;
    strncpy(tmp_name, name(), len);
    tmp_name[len] ='\0';

    // find the corresponding metaclass;
    GoMetaclass* mclass = Go::metaclass(tmp_name);
    if (!mclass)
	return NULL;

    GoMetaclass* superclass;
    while ((superclass = mclass->superclass()) != NULL)
    {
	sprintf(tmp_name, superclass->classname());
	strcat(tmp_name, right_part);

	// read the translators list;
	GTranslatorPgm** superpgm = (GTranslatorPgm**)_instances->data();
	for (int i = 0 ; i < _instances->nelems(); i++, superpgm++)
	    if (strcmp((*superpgm)->name(), tmp_name) == 0)
		return *superpgm;

	// continue with the superclass;
	mclass = superclass;
    }
    return NULL;
}




GBool Handler::match (const GEvent* event) const
{
    if (evinfo.type != event->type())
	return GFALSE;

    if (evinfo.testButton &&
	((evinfo.detail.button & (1 << (event->button() - 1))) == 0))
	return GFALSE;

    if (evinfo.testKey)
    {
	int nI;
	CharBuf** pI = (CharBuf**)evinfo.detail.keyList->arrayFromFirst(nI);
	GBool found = GFALSE;
	for (; nI > 0; --nI, pI++)
	    if (event->character() >= (*pI)->cmin &&
		event->character() <= (*pI)->cmax)
	    {
		found = GTRUE;
		break;
	    }

	if (!found)
	    return GFALSE;
    } 

    if (evinfo.testFunctionKey &&
	evinfo.detail.functionKey != event->functionKey())
	return GFALSE;

    if (evinfo.testModifiers)
    {
	int nM;
	uchar** pM = (uchar**)evinfo.modifiersList->arrayFromFirst(nM);
	GBool found = GFALSE;
	for (; nM > 0; --nM, pM++)
	    if (**pM == event->modifiers())
	    {
		found = GTRUE;
		break;
	    }

	if (!found)
	    return GFALSE;
    }

    return GTRUE;
}


GBool GTranslator::careAbout (const GEvent* event) const
{
    int count;
    Handler** pH;
    for (pH = (Handler**)Pgm()->GetHandlers()->arrayFromFirst(count);
	 count > 0; --count, pH++)
    {
	if ((*pH)->match(event))
	    return GTRUE;
    }

    // try with all the super programs;
    GTranslatorPgm* super;
    GTranslatorPgm* current = Pgm();
    
    while ((super = current->superTranslatorPgm()) != NULL)
    {
	for (pH = (Handler**)super->GetHandlers()->arrayFromFirst(count);
	     count > 0; --count, pH++)
	{
	    if ((*pH)->match(event))
		return GTRUE;
	}
	current = super;
    }
	
    return GFALSE;
}


GBool GTranslator::handle (const GEvent* event)
{
    GBool found = GFALSE;
    int count;
    Handler** pH;

    for (pH = (Handler**)Pgm()->GetHandlers()->arrayFromFirst(count);
	 count > 0;  --count, pH++)
	if ((*pH)->match(event))
	{
	    found = GTRUE;
	    (*pH)->call(_go, event, this);
	}
    
    // ONLY IF NOT FOUND;
    // try with all the super programs;

    GTranslatorPgm* super;
    GTranslatorPgm* current = Pgm();
    while (!found && (super = current->superTranslatorPgm()) != NULL)
    {
	for (pH = (Handler**)super->GetHandlers()->arrayFromFirst(count);
	     count > 0; --count, pH++)
	    if ((*pH)->match(event))
	    {
		found = GTRUE;
		(*pH)->call(_go, event, this);
	    }
	// reset the current program;
	current = super;
    }
    return found;
}


void Handler::call (Go* go, const GEvent* event, GTranslator* translator) const
{
    int nP;
    ParamRec** pP = (ParamRec**)params->arrayFromFirst(nP);

    if (func == NULL)
    {
	void* ptr;
	if (!GBinded(_methodName, ptr))
	{
	    GWarning("Translator.cc: undefined function %s", _methodName);
	    return;
	}

	((Handler*)this)->func = ptr;
    }

    /* we keep only those parameter types (to simplify swithes below):
       		char
		long
		number
		pointer
		*/
    ParamRec args[4];
    GPoint position;
    for (int i = 0; i < nP && i < 4; i++)
    {
	switch ((int)pP[i]->type)
	{
	  case ParamLong:
	  case ParamNumber:
	  case ParamPointer:
	  case ParamChar:
	    args[i] = *pP[i];
	    break;
	  case ParamInputChar:
	    args[i].type = ParamChar;
	    args[i].value.c = event->character();
	    break;
	  case ParamGo:
	    args[i].type = ParamPointer;
	    args[i].value.p = go;
	    break;
	  case ParamGoAtPos:
	    args[i].type = ParamPointer;
	    args[i].value.p = go->root()->goAt(event->position(), .6);
	    break;
	  case ParamGoUI:
	    args[i].type = ParamPointer;
	    args[i].value.p = GoUIFrom(go->root());
	    break;
	  case ParamSensor: {
	    args[i].type = ParamPointer;
	    args[i].value.p = NULL;
	    for (Go* child = go->Child(); child; child = child->Next())
		if (child->getMetaclass()->isSubclass(GoSensor::metaclass()) &&
		    ((GoSensor*)child)->interactor() == translator)
		{
		    args[i].value.p = child;
		    break;
		}
	    }	
	    break;
	  case ParamEvent:
	    args[i].type = ParamPointer;
	    args[i].value.p = (GEvent*)event;
	    break;
	  case ParamPos:
	    args[i].type = ParamPointer;
	    position = event->position();
	    args[i].value.p = &position;
	    break;
	  case ParamColor:
	    args[i].type = ParamPointer;
	    args[i].value.p = pP[i]->value.p;
	    break;
	  case ParamString:
	    args[i].type = ParamPointer;
	    args[i].value.p = pP[i]->value.p;
	    break;
	  default:
	    GError("Translator.cc: internal error in Handler::call()");
	}
    }

    /* Now the big switch through args count and types */

#define SWITCHCALL(num,call) \
    switch ((int)args[num].type) \
    { \
      case ParamLong: call(long, args[num].value.l); break; \
      case ParamChar: call(uchar, args[num].value.c); break; \
      case ParamNumber: call(GNumber, args[num].value.n); break; \
      case ParamPointer: call(void*, args[num].value.p); break; \
    }
    
    switch (nP)
    {
      case 0:
	(*(void(*)())func)();
	break;

      case 1:
#define CALL(t,a) (*(void(*)(t))func)((t)a)
	SWITCHCALL(0, CALL); break;

      case 2:
	switch ((int)args[0].type)
	{
	  case ParamLong:
#define CALL_L(t,a) (*(void(*)(long, t))func)(args[0].value.l, a)
	    SWITCHCALL(1, CALL_L); break;
	  case ParamChar:
#define CALL_C(t,a) (*(void(*)(uchar, t))func)(args[0].value.c, a)
	    SWITCHCALL(1, CALL_C); break;
	  case ParamNumber:
#define CALL_N(t,a) (*(void(*)(GNumber, t))func)(args[0].value.n, a)
	    SWITCHCALL(1, CALL_N); break;
	  case ParamPointer:
#define CALL_P(t,a) (*(void(*)(void*, t))func)(args[0].value.p, a)
	    SWITCHCALL(1, CALL_P); break;
	}
	break;

      case 3:
	switch ((int)args[0].type)
	{
	  case ParamLong:
	    switch ((int)args[1].type)
	    {
	      case ParamLong:
#define CALL_LL(t,a) (*(void(*)(long, long, t))func)(args[0].value.l, args[1].value.l, a)
		SWITCHCALL(2, CALL_LL); break;
	      case ParamChar:
#define CALL_LC(t,a) (*(void(*)(long, uchar, t))func)(args[0].value.l, args[1].value.c, a)
		SWITCHCALL(2, CALL_LC); break;
	      case ParamNumber:
#define CALL_LN(t,a) (*(void(*)(long,GNumber,t))func)(args[0].value.l, args[1].value.n, a)
		SWITCHCALL(2, CALL_LN); break;
	      case ParamPointer:
#define CALL_LP(t,a) (*(void(*)(long,void*,t))func)(args[0].value.l, args[1].value.p, a)
		SWITCHCALL(2, CALL_LP); break;
	    }
	    break;
	    
	  case ParamChar:
	    switch ((int)args[1].type)
	    {
	      case ParamLong:
#define CALL_CL(t,a) (*(void(*)(uchar,long,t))func)(args[0].value.c, args[1].value.l, a)
		SWITCHCALL(2, CALL_CL); break;
	      case ParamChar:
#define CALL_CC(t,a) (*(void(*)(uchar,uchar,t))func)(args[0].value.c, args[1].value.c, a)
		SWITCHCALL(2, CALL_CC); break;
	      case ParamNumber:
#define CALL_CN(t,a) (*(void(*)(uchar,GNumber,t))func)(args[0].value.c, args[1].value.n, a)
		SWITCHCALL(2, CALL_CN); break;
	      case ParamPointer:
#define CALL_CP(t,a) (*(void(*)(uchar,void*,t))func)(args[0].value.c, args[1].value.p, a)
		SWITCHCALL(2, CALL_CP); break;
	    }
	    break;
	    
	  case ParamNumber:
	    switch ((int)args[1].type)
	    {
	      case ParamLong:
#define CALL_NL(t,a) (*(void(*)(GNumber,long,t))func)(args[0].value.n, args[1].value.l, a)
		SWITCHCALL(2, CALL_NL); break;
	      case ParamChar:
#define CALL_NC(t,a) (*(void(*)(GNumber,uchar,t))func)(args[0].value.n, args[1].value.c, a)
		SWITCHCALL(2, CALL_NC); break;
	      case ParamNumber:
#define CALL_NN(t,a) (*(void(*)(GNumber,GNumber,t))func)(args[0].value.n,args[1].value.n,a)
		SWITCHCALL(2, CALL_NN); break;
	      case ParamPointer:
#define CALL_NP(t,a) (*(void(*)(GNumber,void*,t))func)(args[0].value.n, args[1].value.p, a)
		SWITCHCALL(2, CALL_NP); break;
	    }
	    break;

	  case ParamPointer:
	    switch ((int)args[1].type)
	    {
	      case ParamLong:
#define CALL_PL(t,a) (*(void(*)(void*,long,t))func)(args[0].value.p, args[1].value.l, a)
		SWITCHCALL(2, CALL_PL); break;
	      case ParamChar:
#define CALL_PC(t,a) (*(void(*)(void*,uchar,t))func)(args[0].value.p, args[1].value.c, a)
		SWITCHCALL(2, CALL_PC); break;
	      case ParamNumber:
#define CALL_PN(t,a) (*(void(*)(void*,GNumber,t))func)(args[0].value.p,args[1].value.n,a)
		SWITCHCALL(2, CALL_PN); break;
	      case ParamPointer:
#define CALL_PP(t,a) (*(void(*)(void*,void*,t))func)(args[0].value.p, args[1].value.p, a)
		SWITCHCALL(2, CALL_PP); break;
	    }
	    break;
	}
	break;

      case 4:
	switch ((int)args[0].type)
	{
	  case ParamLong:
	    switch ((int)args[1].type)
	    {
	      case ParamLong:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_LLL(t,a) \
  (*(void(*)(long,long,long,t))func)(args[0].value.l, args[1].value.l, args[2].value.l, a)
		    SWITCHCALL(3, CALL_LLL); break;
		  case ParamChar:
#define CALL_LLC(t,a) \
  (*(void(*)(long,long,uchar,t))func)(args[0].value.l, args[1].value.l, args[2].value.c, a)
	            SWITCHCALL(3, CALL_LLC); break;
		  case ParamNumber:
#define CALL_LLN(t,a) \
  (*(void(*)(long,long,GNumber,t))func)(args[0].value.l, args[1].value.l, args[2].value.n, a)
	            SWITCHCALL(3, CALL_LLN); break;
		  case ParamPointer:
#define CALL_LLP(t,a) \
  (*(void(*)(long,long,void*,t))func)(args[0].value.l, args[1].value.l, args[2].value.p, a)
	            SWITCHCALL(3, CALL_LLP); break;
		}
		break;
		    
	      case ParamChar:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_LCL(t,a) \
  (*(void(*)(long,uchar,long,t))func)(args[0].value.l, args[1].value.c, args[2].value.l, a)
		    SWITCHCALL(3, CALL_LCL); break;
		  case ParamChar:
#define CALL_LCC(t,a) \
  (*(void(*)(long,uchar,uchar,t))func)(args[0].value.l, args[1].value.c, args[2].value.c, a)
	            SWITCHCALL(3, CALL_LCC); break;
		  case ParamNumber:
#define CALL_LCN(t,a) \
  (*(void(*)(long,uchar,GNumber,t))func)(args[0].value.l, args[1].value.c, args[2].value.n, a)
	            SWITCHCALL(3, CALL_LCN); break;
		  case ParamPointer:
#define CALL_LCP(t,a) \
  (*(void(*)(long,uchar,void*,t))func)(args[0].value.l, args[1].value.c, args[2].value.p, a)
	            SWITCHCALL(3, CALL_LCP); break;
		}
		break;
		    
	      case ParamNumber:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_LNL(t,a) \
  (*(void(*)(long,GNumber,long,t))func)(args[0].value.l, args[1].value.n, args[2].value.l, a)
		    SWITCHCALL(3, CALL_LNL); break;
		  case ParamChar:
#define CALL_LNC(t,a) \
  (*(void(*)(long,GNumber,uchar,t))func)(args[0].value.l, args[1].value.n, args[2].value.c, a)
	            SWITCHCALL(3, CALL_LNC); break;
		  case ParamNumber:
#define CALL_LNN(t,a) \
  (*(void(*)(long,GNumber,GNumber,t))func)(args[0].value.l, args[1].value.n, args[2].value.n, a)
	            SWITCHCALL(3, CALL_LNN); break;
		  case ParamPointer:
#define CALL_LNP(t,a) \
  (*(void(*)(long,GNumber,void*,t))func)(args[0].value.l, args[1].value.n, args[2].value.p, a)
	            SWITCHCALL(3, CALL_LNP); break;
		}
		break;
		    
	      case ParamPointer:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_LPL(t,a) \
  (*(void(*)(long,void*,long,t))func)(args[0].value.l, args[1].value.p, args[2].value.l, a)
		    SWITCHCALL(3, CALL_LPL); break;
		  case ParamChar:
#define CALL_LPC(t,a) \
  (*(void(*)(long,void*,uchar,t))func)(args[0].value.l, args[1].value.p, args[2].value.c, a)
	            SWITCHCALL(3, CALL_LPC); break;
		  case ParamNumber:
#define CALL_LPN(t,a) \
  (*(void(*)(long,void*,GNumber,t))func)(args[0].value.l, args[1].value.p, args[2].value.n, a)
	            SWITCHCALL(3, CALL_LPN); break;
		  case ParamPointer:
#define CALL_LPP(t,a) \
  (*(void(*)(long,void*,void*,t))func)(args[0].value.l, args[1].value.p, args[2].value.p, a)
	            SWITCHCALL(3, CALL_LPP); break;
		}
		break;
	    }
	    break;
	    
	  case ParamChar:
	    switch ((int)args[1].type)
	    {
	      case ParamLong:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_CLL(t,a) \
  (*(void(*)(uchar,long,long,t))func)(args[0].value.c, args[1].value.l, args[2].value.l, a)
		    SWITCHCALL(3, CALL_CLL); break;
		  case ParamChar:
#define CALL_CLC(t,a) \
  (*(void(*)(uchar,long,uchar,t))func)(args[0].value.c, args[1].value.l, args[2].value.c, a)
	            SWITCHCALL(3, CALL_CLC); break;
		  case ParamNumber:
#define CALL_CLN(t,a) \
  (*(void(*)(uchar,long,GNumber,t))func)(args[0].value.c, args[1].value.l, args[2].value.n, a)
	            SWITCHCALL(3, CALL_CLN); break;
		  case ParamPointer:
#define CALL_CLP(t,a) \
  (*(void(*)(uchar,long,void*,t))func)(args[0].value.c, args[1].value.l, args[2].value.p, a)
	            SWITCHCALL(3, CALL_CLP); break;
		}
		break;

	      case ParamChar:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_CCL(t,a) \
  (*(void(*)(uchar,uchar,long,t))func)(args[0].value.c, args[1].value.c, args[2].value.l, a)
		    SWITCHCALL(3, CALL_CCL); break;
		  case ParamChar:
#define CALL_CCC(t,a) \
  (*(void(*)(uchar,uchar,uchar,t))func)(args[0].value.c, args[1].value.c, args[2].value.c, a)
	            SWITCHCALL(3, CALL_CCC); break;
		  case ParamNumber:
#define CALL_CCN(t,a) \
  (*(void(*)(uchar,uchar,GNumber,t))func)(args[0].value.c, args[1].value.c, args[2].value.n, a)
	            SWITCHCALL(3, CALL_CCN); break;
		  case ParamPointer:
#define CALL_CCP(t,a) \
  (*(void(*)(uchar,uchar,void*,t))func)(args[0].value.c, args[1].value.c, args[2].value.p, a)
	            SWITCHCALL(3, CALL_CCP); break;
		}
		break;

	      case ParamNumber:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_CNL(t,a) \
  (*(void(*)(uchar,GNumber,long,t))func)(args[0].value.c, args[1].value.n, args[2].value.l, a)
		    SWITCHCALL(3, CALL_CNL); break;
		  case ParamChar:
#define CALL_CNC(t,a) \
  (*(void(*)(uchar,GNumber,uchar,t))func)(args[0].value.c, args[1].value.n, args[2].value.c, a)
	            SWITCHCALL(3, CALL_CNC); break;
		  case ParamNumber:
#define CALL_CNN(t,a) \
  (*(void(*)(uchar,GNumber,GNumber,t))func)(args[0].value.c, args[1].value.n, args[2].value.n, a)
	            SWITCHCALL(3, CALL_CNN); break;
		  case ParamPointer:
#define CALL_CNP(t,a) \
  (*(void(*)(uchar,GNumber,void*,t))func)(args[0].value.c, args[1].value.n, args[2].value.p, a)
	            SWITCHCALL(3, CALL_CNP); break;
		}
		break;

	      case ParamPointer:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_CPL(t,a) \
  (*(void(*)(uchar,void*,long,t))func)(args[0].value.c, args[1].value.p, args[2].value.l, a)
		    SWITCHCALL(3, CALL_CPL); break;
		  case ParamChar:
#define CALL_CPC(t,a) \
  (*(void(*)(uchar,void*,uchar,t))func)(args[0].value.c, args[1].value.p, args[2].value.c, a)
	            SWITCHCALL(3, CALL_CPC); break;
		  case ParamNumber:
#define CALL_CPN(t,a) \
  (*(void(*)(uchar,void*,GNumber,t))func)(args[0].value.c, args[1].value.p, args[2].value.n, a)
	            SWITCHCALL(3, CALL_CPN); break;
		  case ParamPointer:
#define CALL_CPP(t,a) \
  (*(void(*)(uchar,void*,void*,t))func)(args[0].value.c, args[1].value.p, args[2].value.p, a)
	            SWITCHCALL(3, CALL_CPP); break;
		}
		break;
	    }
	    break;

	  case ParamNumber:
	    switch ((int)args[1].type)
	    {
	      case ParamLong:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_NLL(t,a) \
  (*(void(*)(GNumber,long,long,t))func)(args[0].value.n, args[1].value.l, args[2].value.l, a)
		    SWITCHCALL(3, CALL_NLL); break;
		  case ParamChar:
#define CALL_NLC(t,a) \
  (*(void(*)(GNumber,long,uchar,t))func)(args[0].value.n, args[1].value.l, args[2].value.c, a)
	            SWITCHCALL(3, CALL_NLC); break;
		  case ParamNumber:
#define CALL_NLN(t,a) \
  (*(void(*)(GNumber,long,GNumber,t))func)(args[0].value.n, args[1].value.l, args[2].value.n, a)
	            SWITCHCALL(3, CALL_NLN); break;
		  case ParamPointer:
#define CALL_NLP(t,a) \
  (*(void(*)(GNumber,long,void*,t))func)(args[0].value.n, args[1].value.l, args[2].value.p, a)
	            SWITCHCALL(3, CALL_NLP); break;
		}
		break;

	      case ParamChar:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_NCL(t,a) \
  (*(void(*)(GNumber,uchar,long,t))func)(args[0].value.n, args[1].value.c, args[2].value.l, a)
		    SWITCHCALL(3, CALL_NCL); break;
		  case ParamChar:
#define CALL_NCC(t,a) \
  (*(void(*)(GNumber,uchar,uchar,t))func)(args[0].value.n, args[1].value.c, args[2].value.c, a)
	            SWITCHCALL(3, CALL_NCC); break;
		  case ParamNumber:
#define CALL_NCN(t,a) \
  (*(void(*)(GNumber,uchar,GNumber,t))func)(args[0].value.n, args[1].value.c, args[2].value.n, a)
	            SWITCHCALL(3, CALL_NCN); break;
		  case ParamPointer:
#define CALL_NCP(t,a) \
  (*(void(*)(GNumber,uchar,void*,t))func)(args[0].value.n, args[1].value.c, args[2].value.p, a)
	            SWITCHCALL(3, CALL_NCP); break;
		}
		break;

	      case ParamNumber:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_NNL(t,a) \
  (*(void(*)(GNumber,GNumber,long,t))func)(args[0].value.n, args[1].value.n, args[2].value.l, a)
		    SWITCHCALL(3, CALL_NNL); break;
		  case ParamChar:
#define CALL_NNC(t,a) \
  (*(void(*)(GNumber,GNumber,uchar,t))func)(args[0].value.n, args[1].value.n, args[2].value.c, a)
	            SWITCHCALL(3, CALL_NNC); break;
		  case ParamNumber:
#define CALL_NNN(t,a) \
  (*(void(*)(GNumber,GNumber,GNumber,t))func)(args[0].value.n,args[1].value.n,args[2].value.n, a)
	            SWITCHCALL(3, CALL_NNN); break;
		  case ParamPointer:
#define CALL_NNP(t,a) \
  (*(void(*)(GNumber,GNumber,void*,t))func)(args[0].value.n, args[1].value.n, args[2].value.p, a)
	            SWITCHCALL(3, CALL_NNP); break;
		}
		break;

	      case ParamPointer:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_NPL(t,a) \
  (*(void(*)(GNumber,void*,long,t))func)(args[0].value.n, args[1].value.p, args[2].value.l, a)
		    SWITCHCALL(3, CALL_NPL); break;
		  case ParamChar:
#define CALL_NPC(t,a) \
  (*(void(*)(GNumber,void*,uchar,t))func)(args[0].value.n, args[1].value.p, args[2].value.c, a)
	            SWITCHCALL(3, CALL_NPC); break;
		  case ParamNumber:
#define CALL_NPN(t,a) \
  (*(void(*)(GNumber,void*,GNumber,t))func)(args[0].value.n,args[1].value.p,args[2].value.n, a)
	            SWITCHCALL(3, CALL_NPN); break;
		  case ParamPointer:
#define CALL_NPP(t,a) \
  (*(void(*)(GNumber,void*,void*,t))func)(args[0].value.n, args[1].value.p, args[2].value.p, a)
	            SWITCHCALL(3, CALL_NPP); break;
		}
		break;
	    }
	    break;

	  case ParamPointer:
	    switch ((int)args[1].type)
	    {
	      case ParamLong:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_PLL(t,a) \
  (*(void(*)(void*,long,long,t))func)(args[0].value.p, args[1].value.l, args[2].value.l, a)
		    SWITCHCALL(3, CALL_PLL); break;
		  case ParamChar:
#define CALL_PLC(t,a) \
  (*(void(*)(void*,long,uchar,t))func)(args[0].value.p, args[1].value.l, args[2].value.c, a)
	            SWITCHCALL(3, CALL_PLC); break;
		  case ParamNumber:
#define CALL_PLN(t,a) \
  (*(void(*)(void*,long,GNumber,t))func)(args[0].value.p,args[1].value.l,args[2].value.n, a)
	            SWITCHCALL(3, CALL_PLN); break;
		  case ParamPointer:
#define CALL_PLP(t,a) \
  (*(void(*)(void*,long,void*,t))func)(args[0].value.p, args[1].value.l, args[2].value.p, a)
	            SWITCHCALL(3, CALL_PLP); break;
		}
		break;

	      case ParamChar:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_PCL(t,a) \
  (*(void(*)(void*,uchar,long,t))func)(args[0].value.p, args[1].value.c, args[2].value.l, a)
		    SWITCHCALL(3, CALL_PCL); break;
		  case ParamChar:
#define CALL_PCC(t,a) \
  (*(void(*)(void*,uchar,uchar,t))func)(args[0].value.p, args[1].value.c, args[2].value.c, a)
	            SWITCHCALL(3, CALL_PCC); break;
		  case ParamNumber:
#define CALL_PCN(t,a) \
  (*(void(*)(void*,uchar,GNumber,t))func)(args[0].value.p,args[1].value.c,args[2].value.n, a)
	            SWITCHCALL(3, CALL_PCN); break;
		  case ParamPointer:
#define CALL_PCP(t,a) \
  (*(void(*)(void*,uchar,void*,t))func)(args[0].value.p, args[1].value.c, args[2].value.p, a)
	            SWITCHCALL(3, CALL_PCP); break;
		}
		break;

	      case ParamNumber:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_PNL(t,a) \
  (*(void(*)(void*,GNumber,long,t))func)(args[0].value.p, args[1].value.n, args[2].value.l, a)
		    SWITCHCALL(3, CALL_PNL); break;
		  case ParamChar:
#define CALL_PNC(t,a) \
  (*(void(*)(void*,GNumber,uchar,t))func)(args[0].value.p, args[1].value.n, args[2].value.c, a)
	            SWITCHCALL(3, CALL_PNC); break;
		  case ParamNumber:
#define CALL_PNN(t,a) \
  (*(void(*)(void*,GNumber,GNumber,t))func)(args[0].value.p,args[1].value.n,args[2].value.n, a)
	            SWITCHCALL(3, CALL_PNN); break;
		  case ParamPointer:
#define CALL_PNP(t,a) \
  (*(void(*)(void*,GNumber,void*,t))func)(args[0].value.p, args[1].value.n, args[2].value.p, a)
	            SWITCHCALL(3, CALL_PNP); break;
		}
		break;

	      case ParamPointer:
		switch ((int)args[2].type)
		{
		  case ParamLong:
#define CALL_PPL(t,a) \
  (*(void(*)(void*,void*,long,t))func)(args[0].value.p, args[1].value.p, args[2].value.l, a)
		    SWITCHCALL(3, CALL_PPL); break;
		  case ParamChar:
#define CALL_PPC(t,a) \
  (*(void(*)(void*,void*,uchar,t))func)(args[0].value.p, args[1].value.p, args[2].value.c, a)
	            SWITCHCALL(3, CALL_PPC); break;
		  case ParamNumber:
#define CALL_PPN(t,a) \
  (*(void(*)(void*,void*,GNumber,t))func)(args[0].value.p,args[1].value.p,args[2].value.n, a)
	            SWITCHCALL(3, CALL_PPN); break;
		  case ParamPointer:
#define CALL_PPP(t,a) \
  (*(void(*)(void*,void*,void*,t))func)(args[0].value.p, args[1].value.p, args[2].value.p, a)
	            SWITCHCALL(3, CALL_PPP); break;
		}
		break;
	    }
	    break;
	}
	break;

      default:
	GWarning("Translator.cc: too many arguments: %d", nP);
	return;
    }
}


GTranslatorPgm* GTranslatorPgm::create (const char* name)
{
    void* value;
    const char* type;

    if (GBinded (name, value, type))
    {
	if (strcmp (type, GInteractorPgm::type()))
	{
	    GWarning ("GTranslatorPgm::create(): the name \"%s\" is already used as a \"%s\" type.",
		      name, (type == NULL) ? "(NULL)" : type);
	    return NULL;
	}

	return (GTranslatorPgm*) value;
    }

    return new GTranslatorPgm (name);
}


GTranslator::~GTranslator ()
{
}


const char* GTranslatorPgm::checkFileSyntax (const char* filename, int& wrongLineNumber)
{
    checkOnly = GTRUE;
    anErrorOccured = GFALSE;

    wrongLineNumber = 0;

    FILE* file = fopen (filename, "r");
    if (file == NULL)
    {
	sprintf (checkErrorMsg, "cannot open file \"%s\"", filename);
	return checkErrorMsg;
    }

    while (!feof(file))
    {
	char buf[512];
	(void) fgets (buf, 512, file);
	wrongLineNumber++;
	checkFileLineSyntax (buf);
	if (anErrorOccured)
	    break;
    }
    
    fclose (file);
    
    if (!anErrorOccured)
	return NULL;
    else
	return checkErrorMsg;
}


const char* GTranslatorPgm::checkSyntax (const char* event_mask, const char* function_to_call)
{
    checkOnly = GTRUE;
    anErrorOccured = GFALSE;
    
    char methodName[128];
    if (!_parse_method_name (function_to_call, methodName))
	return checkErrorMsg;
    
    char params[512];
    if (!_parse_params (function_to_call, params))
	return checkErrorMsg;

    Handler h;
    h.Parse (event_mask, methodName, params);

    return anErrorOccured ? checkErrorMsg : NULL;
}


const char* GTranslatorPgm::checkFileLineSyntax (const char* line)
{
    checkOnly = GTRUE;
    anErrorOccured = GFALSE;

    _parse_handler_line (line);
    return (anErrorOccured) ? checkErrorMsg : NULL;
}



